/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Odonata.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.bridge;

import java.io.*;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.sql.SQLException;
import java.util.*;

import javax.sql.DataSource;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParserFactory;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;
import org.xquark.jdbc.datasource.JDBCDataSourceFactory;
import org.xquark.serialize.XMLSerializer;
import org.xquark.xml.xdbc.*;

/**
 * This class is the command-line launcher for Bridge. It is mostly useful
 * for test and demo purposes, as the JVM and connection setup overhead
 * are prohibitive in most cases.
 */
public class Main {

	private static final String RCSRevision = "$Revision: 1.3 $";
	private static final String RCSName = "$Name: XQ1_1_1 $";

	private static final String HELP_OPT1 = "-help";
	private static final String HELP_OPT2 = "-?";
	private static final String CONF_OPT = "-conf";
	private static final String PROP_OPT1 = "-p";
	private static final String PROP_OPT2 = "-properties";
	private static final String DRIVER_OPT = "-driver";
	private static final String JDBC_OPT = "-db";
	private static final String JDBC_OPT2 = "-url";
	private static final String USER_OPT = "-user";
	private static final String PASSWORD_OPT = "-password";
	private static final String QUERY_OPT1 = "-query";
	private static final String QUERY_OPT2 = "-q";
	private static final String INSERT_OPT1 = "-insert";
	private static final String INSERT_OPT2 = "-i";
	private static final String FILE_OPT1 = "-file";
	private static final String FILE_OPT2 = "-f";
	private static final String DEBUG_OPT = "-debug";
	private static final String CLASSPATH_OPT1 = "-cp";
	private static final String CLASSPATH_OPT2 = "-classpath";

	private static SAXParserFactory parserFactory;

	static {
		parserFactory = SAXParserFactory.newInstance();
		parserFactory.setNamespaceAware(true);
	}

	/* This part of the file deals with command-line execution */

	/**
	 * The method called from the command line to run the XQBridge as a standalone program.
	 * Use the -help option to display a list of available options.
	 * @param args the command line parameters
	 */
	public static void main(String[] args) {
		String propFile = "XQBridge.properties";
		boolean defaultPropFile = true;
		String configFile = null;
		String batchFile = null;
		String driver = null;
		String url = null;
		String user = null;
		String password = null;
		String classpath = null;
		ClassLoader classloader = XQBridge.class.getClassLoader();
		boolean isQuery = false;
		boolean isInsert = false;
		ArrayList parameters = new ArrayList();

		for (int i = 0; i < args.length; i++) {
			if (!args[i].startsWith("-")) {
				// not an option
				parameters.add(args[i]);
			} else {
				// Check options are in first position
				if (parameters.size() > 0) {
					usage();
					System.err.println("Options must appear before the command parameters.");
					System.exit(-1);
				}
				// one-parameter option
				if (HELP_OPT1.equals(args[i]) || HELP_OPT2.equals(args[i])) {
					usage();
					System.exit(0);
				} else if (QUERY_OPT1.equals(args[i]) || QUERY_OPT2.equals(args[i])) {
					isQuery = true;
				} else if (INSERT_OPT1.equals(args[i]) || INSERT_OPT2.equals(args[i])) {
					isInsert = true;
				} else {
					// two-parameter option
					if (i + 1 == args.length) {
						System.err.println("Missing value for option " + args[i]);
						usage();
						System.exit(-1);
					}
					if (CONF_OPT.equals(args[i])) {
						configFile = args[++i];
					} else if (PROP_OPT1.equals(args[i]) || PROP_OPT2.equals(args[i])) {
						propFile = args[++i];
						defaultPropFile = false;
					} else if (DEBUG_OPT.equals(args[i])) {
						Logger.getRootLogger().setLevel(Level.DEBUG);
					} else if (FILE_OPT1.equals(args[i]) || FILE_OPT2.equals(args[i])) {
						batchFile = args[++i];
					} else if (DRIVER_OPT.equals(args[i])) {
						driver = args[++i];
					} else if (JDBC_OPT.equals(args[i]) || JDBC_OPT2.equals(args[i])) {
						url = args[++i];
					} else if (USER_OPT.equals(args[i])) {
						user = args[++i];
					} else if (PASSWORD_OPT.equals(args[i])) {
						password = args[++i];
					} else if (CLASSPATH_OPT1.equals(args[i]) || CLASSPATH_OPT2.equals(args[i])) {
						classpath = args[++i];
					} else {
						// Unknown option
						System.err.println("Invalid option " + args[i]);
						usage();
						System.exit(-1);
					}
				}
			}
		}
		if (isQuery == isInsert) {
			System.err.println("One and only one of -q(uery) and -i(nsert) must be specified");
			System.exit(-1);
		}

		Properties props = new Properties();
		try {
			props.load(new FileInputStream(propFile));
		} catch (IOException e) {
			InputStream inStream = null;
			try {
				inStream = XQBridge.class.getResourceAsStream("/" + propFile);
				if (inStream != null)
					props.load(inStream);
			} catch (IOException ex) {
				inStream = null;
			}
			if (configFile == null && inStream == null && (url == null || user == null || password == null)) {
				System.err.println("error : configuration file " + propFile + " couldn't be found.");
				System.err.println("(neither using file name nor in CLASSPATH root)");
				System.exit(-1);
			}
		}

		if (configFile == null || !defaultPropFile) {
			if (driver == null)
				driver = props.getProperty("database.driver");
			if (url == null)
				url = props.getProperty("database.url");
			if (user == null)
				user = props.getProperty("database.user");
			if (password == null)
				password = props.getProperty("database.password");
		}

		if (classpath == null)
			classpath = props.getProperty("xqbridge.classpath");

		if (classpath != null) {
			classloader = new URLClassLoader(parseClassPath(classpath), classloader);
		}

		DataSource ds = null;
		if (url != null && user != null && password != null) {
			try {
				JDBCDataSourceFactory dsf = JDBCDataSourceFactory.newInstance(classloader);
				ds = dsf.newDataSource(driver, url, user, password);
			} catch (SQLException ex) {
				ds = null;
			}
		}

		XQBridge bridge = null;
		if (ds == null) {
			if (configFile == null) {
				System.err.println("Could not connect to database " + url + ". Check your JDBC parameters.");
				System.exit(-1);
			}
			bridge = new XQBridge(new File(configFile));
		} else {
			if (configFile == null) {
				bridge = new XQBridge(ds);
			} else {
				bridge = new XQBridge(ds, new File(configFile));
			}
		}

		bridge.setClassLoader(classloader);

		List parametersList = null;
		if (batchFile != null) {
			parametersList = parseBatchFile(batchFile, isQuery);
		} else {
			checkParameters(parameters, isQuery);
			parametersList = new ArrayList(1);
			parametersList.add(parameters);
		}
		int result = 0;
		if (isQuery) {
			result = executeQuery(bridge, parametersList);
		} else {
			result = executeInsert(bridge, parametersList);
		}
		System.exit(result);
	}

	private static void checkParameters(List params, boolean isQuery) {
		int size = params.size();
		if (isQuery) {
			if ((isQuery && (size < 1 || size > 2)) || (!isQuery && size < 2)) {
				System.err.println("Wrong number of parameters.");
				System.exit(-1);
			}
		}
	}

	private static List parseBatchFile(String batchFile, boolean isQuery) {
		ArrayList result = new ArrayList();
		BufferedReader reader = null;
		try {
			File f = new File(batchFile);
			reader = new BufferedReader(new FileReader(f));
		} catch (IOException ex) {
			System.err.println("Could not find specified batch file");
			System.exit(-1);
		}
		try {
			while (true) {
				String line = reader.readLine();
				if (line == null)
					break;
				if (line.startsWith("#") || line.trim().length() == 0)
					continue;
				java.util.StringTokenizer it = new java.util.StringTokenizer(line);
				ArrayList params = new ArrayList();
				while (it.hasMoreTokens()) {
					params.add(it.nextToken());
				}
				checkParameters(params, isQuery);
				result.add(params);
			}
		} catch (IOException e) {
			System.err.println("Error while reading specified batch file");
			System.exit(-1);
		}
		return result;
	}

	private static String readQueryFile(String fileName) {
		try {
			StringBuffer retVal = new StringBuffer();
			String tmpStr;
			BufferedReader in = new BufferedReader(new FileReader(fileName));
			while ((tmpStr = in.readLine()) != null) {
				if (retVal.length() > 0)
					retVal.append('\n');
				retVal.append(tmpStr);
			}
			in.close();
			return retVal.toString();
		} catch (FileNotFoundException ex) {
			System.err.println("Could not find query file " + fileName);
		} catch (IOException ex) {
			System.err.println("Error while reading query file " + fileName);
		}
		return null;
	}

	private static int executeQuery(XQBridge bridge, List parametersList) {
		int result = 0;
		XMLConnection connection = null;
		XMLStatement stat = null;
		try {
			connection = bridge.getXMLConnection();
			stat = connection.createStatement();
		} catch (XMLDBCException ex) {
			System.err.println("Error while creating XQuery statement");
			printError(ex);
			result = -1;
		}
		Iterator it = parametersList.iterator();
		while (it.hasNext()) {
			List params = (List) it.next();
			String queryFile = (String) params.get(0);
			String resultFile = null;
			if (params.size() > 1)
				resultFile = (String) params.get(1);
			String query = readQueryFile(queryFile);
			if (query != null) {
				OutputStream out = System.out;
				try {
					if (resultFile != null) {
						out = new BufferedOutputStream(new FileOutputStream(new File(resultFile)));
					}
				} catch (IOException ex) {
					out = null;
					System.err.println("Could not open result file " + resultFile);
					result = -1;
				}
				if (out != null) {
					try {
						stat.setBaseURI(new File(queryFile).toURL().toString());
						XMLResultSet rs = stat.executeQuery(query);
						XMLSerializer handler = new XMLSerializer(out);
						rs.setContentHandler(handler);
						handler.startDocument();
						long rtime = System.currentTimeMillis();
						while (rs.hasNext())
							rs.nextAsSAX();
						handler.endDocument();
						rs.close();
						if (out != System.out)
							out.close();
					} catch (MalformedURLException e) {
						System.err.println("Error while accessing query file " + queryFile);
						printError(e);
						result = -1;
					} catch (XMLDBCException e) {
						System.err.println("Error while processing query file " + queryFile);
						printError(e);
						result = -1;
					} catch (UnsupportedEncodingException e) {
						System.err.println("Illegal encoding for query result for " + queryFile);
						printError(e);
						result = -1;
					} catch (SAXException e) {
						System.err.println("Error while processing query result for " + queryFile);
						printError(e);
						result = -1;
					} catch (IOException e) {
						System.err.println("Error while closing result stream");
						printError(e);
						result = -1;
					}
				}
			} else {
				result = -1;
			}
		}
		try {
			if (stat != null)
				stat.close();
		} catch (XMLDBCException ex) {
			System.err.println("Error while closing XQuery statement");
			printError(ex);
			result = -1;
		}
		return result;
	}

	private static int executeInsert(XQBridge bridge, List parametersList) {
		int result = 0;
		Iterator it = parametersList.iterator();
		while (it.hasNext()) {
			List params = (List) it.next();
			String mapFile = (String) params.get(0);
			Mapping mapping = null;
			try {
				mapping = bridge.getMapping(mapFile);
			} catch (XMLDBCException ex) {
				System.err.println("Error while loading mapping file " + mapFile);
				printError(ex);
				result = -1;
			}
			if (mapping != null) {
				String xmlFile = null;
				try {
					Mapper mapper = mapping.getMapper();
					XMLReader reader = parserFactory.newSAXParser().getXMLReader();
					reader.setContentHandler(mapper);
					for (int i = 1; i < params.size(); i++) {
						xmlFile = (String) params.get(i);
						reader.parse(xmlFile);
					}
					mapper.close();
					mapping.close();
				} catch (XMLDBCException ex) {
					System.err.println("Error while accessing mapping information for " + xmlFile);
					printError(ex);
					result = -1;
				} catch (SAXException ex) {
					System.err.println("Error while processing XML file " + xmlFile);
					printError(ex);
					result = -1;
				} catch (ParserConfigurationException ex) {
					System.err.println("Cannot create XML parser for " + xmlFile);
					printError(ex);
					result = -1;
				} catch (IOException ex) {
					System.err.println("Error while accessing XML file " + xmlFile);
					printError(ex);
					result = -1;
				}
			}
		}
		return result;
	}

	/**
	 * Print an error message to the standard error output.
	 */
	private static void printError(Throwable e) {
		XMLDBCException.printError(e, Logger.getRootLogger().isDebugEnabled());
	}

	private static URL[] parseClassPath(String classpath) {
		String pathSeparator = System.getProperty("path.separator");
		ArrayList urls = new ArrayList();
		if (classpath != null) {
			StringTokenizer tokenizer = new StringTokenizer(classpath, pathSeparator);
			while (tokenizer.hasMoreTokens()) {
				File file = new File(tokenizer.nextToken());
				try {
					urls.add(file.getCanonicalFile().toURL());
				} catch (MalformedURLException e) {
				} catch (IOException e) {
				}
			}
		}
		return (URL[]) urls.toArray(new URL[0]);
	}

	private static void usage() {
		System.out.println("Usage: java org.xquark.bridge.XQBridge -q(uery) <common options> <file1> [<file2>]");
		System.out.println("       java org.xquark.bridge.XQBridge -i(nsert) <common options> <file1> <file2> [filei]*");
		System.out.println();
		System.out.println("-q, -query:  query mode");
		System.out.println("file1:       XQuery file");
		System.out.println("file2:       optional output file, output to stdout by default");
		System.out.println();
		System.out.println("-i, -insert: mapping mode");
		System.out.println("file1:       mapping file");
		System.out.println("file2:       XML document to be inserted according to the mapping");
		System.out.println("filei:       optional additional XML documents to be inserted according to the mapping");
		System.out.println();
		System.out.println("Common options:");
		System.out.println("-?, -help:       this message");
		System.out.println("-conf:           XQBridge configuration file (see manual for details).");
		System.out.println("-p, properties:  properties file containing database connection information.");
		System.out.println("-driver:         JDBC driver.");
		System.out.println("-db, -url:       JDBC database URL.");
		System.out.println("-user:           database user name.");
		System.out.println("-password:       database user password.");
		System.out.println("-cp, -classpath: classpath used to load JDBC drivers and user generators.");
		System.out.println("-f, file:        file containing a list of files to be processed.");
		System.out.println("                 Each line of the file has the number of files described above, according to the mode.");
		System.out.println("                 Lines starting with # are comments");
	}



}
