/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.bridge;

import java.sql.Connection;

import org.xml.sax.ContentHandler;
import org.xml.sax.ErrorHandler;
import org.xquark.xml.xdbc.XMLDBCException;
import org.xquark.xml.xdbc.XMLErrorHandler;

/**
 * Objects implementing this class are used to store XML data
 * into a database according to the parent mapping.<BR>
 * An object of this class cannot be used in multiple threads concurrently.
 * Different Mappers can be used concurrently by different threads.<BR>
 * A Mapper may be used multiple times, to store several documents based
 * on the same mapping.<BR>
 * A Mapper provides three different interfaces to store XML data: the most
 * efficient is based on SAX2, as the Mapper can be plugged directly as a
 * org.xml.sax.ContentHandler into any XML parser implementing the org.xml.sax.XMLReader
 * interface. <B>Note that with this method the error handler must also be
 * plugged so that the mapper can have a consistent behavior in case of parsing
 * error due to a malformed XML document (perform a rollback).</B>
 * For convenience, a DOM2-based and a string-based interfaces are also
 * provided.<BR>
 * A Mapper object must be explicitly closed after use, in order to free its associated database
 * resources.
 * @see org.xml.sax.XMLReader
 */
public interface Mapper extends ContentHandler, ErrorHandler
{
	public static final String 	RCSRevision 	= "$Revision: 1.5 $";
	public static final String 	RCSName     	= "$Name: XQ1_1_1 $";
    
    /**
     * Returns the mapping used by this object
     * @return the parent mapping
     */
    public Mapping getMapping();
    
    /**
     * Closes the object and frees its associated resources.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void close() throws XMLDBCException;
    
    /**
     * Inserts an XML document in the database using the SAX method.
     * The SAX parser is automatically allocated using JAXP.
     * @param input the SAX InputSource.
     * @throws XMLDBCException if a data source access error occurs.
     * @throws SAXException if a parsing error occurs.
     */
    public void insertDocument(org.xml.sax.InputSource input)
    throws XMLDBCException, org.xml.sax.SAXException;
    
    /**
     * Inserts an XML document in database using the SAX method.
     * @param parser the SAX 2 XMLReader implementation.
     * @param input the SAX InputSource.
     * @throws XMLDBCException if a data source access error occurs.
     * @throws SAXException if a parsing error occurs.
     */
    public void insertDocument(org.xml.sax.XMLReader parser, org.xml.sax.InputSource input)
    throws XMLDBCException, org.xml.sax.SAXException;
    
    /**
     * Inserts the XML data contained in the DOM2 document in the database.
     * @param doc the XML document as a DOM2 org.w3c.dom.Document.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void insertDocument(org.w3c.dom.Document doc) throws XMLDBCException;
    
    /**
     * Inserts the XML data contained in the string in the database.
     * @param doc a string containing the XML document.
     * @throws XMLDBCException if a data source access error occurs.
     * @throws org.xml.sax.SAXException if a parse exception occurs.
     */
    public void insertDocument(String doc) throws XMLDBCException, org.xml.sax.SAXException;
    
    /**
     * Return the JDBC connection used by this object. It can be used to perform
     * custom queries in the relational database.
     *
     * @return A JDBC connection.
     */
    public Connection getConnection();
    
    /**
     * Return the current error handler.
     *
     * @return The current error handler, or null if none
     *         has been registered.
     * @see #setErrorHandler
     */
    public XMLErrorHandler getErrorHandler();
    
  /**
   * Allows an application to register an error event handler for non-fatal
   * errors.
   *
   * <p>If the application does not register an error handler, all warning
   * and non-fatal error events reported by the filer will be silently
   * ignored and storage will continue.</p>
   *
   * <p>Applications may register a new or different handler in the
   * middle of a storage operation, and the filer must begin using the new
   * handler immediately.</p>
   *
   * @param handler The error handler.
   * @exception java.lang.NullPointerException If the handler 
   *            argument is null.
   * @see #getErrorHandler
   */
    public void setErrorHandler(XMLErrorHandler handler);
  
  /**
   * This method is to be used when the document stored contains an 'xsi:schemaLocation'
   * attribute (or any element or attribute with the XML Schema type 'anyURI')
   * that uses relatives URIs and when the API used to store documents does not
   * use an InputSource with the SystemID set (e.g. DOM or String ones). The document
   * base provided is then used to resolve the relative URI into an absolute one.
   *
   * @param URI The URI to be used as base during document storage.
   * @see #getDocumentBase
   */
    public void setDocumentBase(String URI);
  
  /**
   * Retrieves the current document base used by this object.
   *
   * @return an URI.
   * @see #setDocumentBase(String)
   */
    public String getDocumentBase();
  
  /**
   * Turns on or off the automatic buffering mode. If this mode is off
   * XML documents are buffered until the flushBuffer() method is
   * called, providing a kind of 'manual' mode for controlling data flush.
   * If the automatic mode is on (the default) data is automatically flushed after
   * each document insertion.
   * <p><b>Important note: </b>The manual mode is only useful if the JDBC connection
   * autoCommit mode is false (otherwise, a flushBuffer() and commit() are automatically
   * called at the end of each document insertion, regardless of the autoFlush mode). When
   * both autoCommit and autoFlush are turned off, the user MUST call flushBuffer() before
   * committing the changes. 
   * @param activated 'true' is the default and means that autoflush is on.
   * @throws XMLDBCException if a data source access error occurs.
   */
  public void setAutoFlush(boolean activated) throws XMLDBCException;
  
  /**
   * Retrieves the current buffering mode.
   * @see #setAutoFlush(boolean)
   * @return true if the 'automatic' buffering mode is activated.
   */
  public boolean getAutoFlush();
  
  /**
   * Sends buffered XML data to the data source. This method is to be called
   * only when the automatic buffering mode is off.
   * @see #setAutoFlush(boolean)
   * @throws XMLDBCException if a data source access error occurs.
   */
  public void flushBuffer() throws XMLDBCException;

  /**
   * Discards the XML data buffered by the filer since the last call to
   * flushBuffer() method. This method is to be called only when the automatic
   * buffering mode is off.
   * @see #setAutoFlush(boolean)
   * @throws XMLDBCException if a data source access error occurs.
   */
  public void clearBuffer() throws XMLDBCException;

}


