/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.bridge;

import java.sql.Connection;
import java.sql.SQLException;

import org.w3c.dom.Document;
import org.xml.sax.*;
import org.xml.sax.ext.LexicalHandler;
import org.xquark.mapper.dbms.AbstractConnection;
import org.xquark.mapper.storage.SAXHandler;
import org.xquark.mapper.storage.XDBCMapperAdapter;
import org.xquark.schema.validation.ValidatingSchemaHandler;
import org.xquark.xml.xdbc.XMLDBCException;

/**
 * Objects implementing this class are used to store XML data
 * into a database according to the parent mapping.<BR>
 * An object of this class cannot be used in multiple threads concurrently.
 * Different Mappers can be used concurrently by different threads.<BR>
 * A Mapper may be used multiple times, to store several documents based
 * on the same mapping.<BR>
 * A Mapper provides three different interfaces to store XML data: the most
 * efficient is based on SAX2, as the Mapper can be plugged directly as a
 * org.xml.sax.ContentHandler into any XML parser implementing the org.xml.sax.XMLReader
 * interface. For convenience, a DOM2-based and a string-based interfaces are also
 * provided.<BR>
 * A Mapper object must be explicitly closed after use, in order to free its associated database
 * resources.
 * @see org.xml.sax.XMLReader
 */
class MapperImpl extends XDBCMapperAdapter implements Mapper
{
	private static final String 	RCSRevision 	= "$Revision: 1.8 $";
	private static final String 	RCSName     	= "$Name: XQ1_1_1 $";
    
    private Mapping mapping = null;
    private AbstractConnection connection;
    private ValidatingSchemaHandler validatingContentHandler = null;
    
    /**
     * Package-private constructor
     * @param mapping the parent mapping
     */
    MapperImpl(Mapping mapping, AbstractConnection connection, ValidatingSchemaHandler validatingContentHandler,
               LexicalHandler lexicalHandler, ErrorHandler errorHandler, SAXHandler handler)
        throws XMLDBCException
    {
        super(validatingContentHandler, lexicalHandler, errorHandler, handler);
        this.mapping = mapping;
        this.connection = connection;
        this.validatingContentHandler = validatingContentHandler;
    }
    
    /**
     * Returns the mapping used by this object
     * @return the parent mapping
     */
    public Mapping getMapping()
    {
        return mapping;
    }
    
    /**
     * Closes the object and frees its associated resources.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void close() throws XMLDBCException
    {
		super.close();
    	try {
			connection.close();
    	} catch (SQLException ex) {
    		throw new XMLDBCException(ex.getMessage(), ex);
    	}
        mapping.remove(this);
    }
    
    ///////////////////////////////////////////////////////////////////////////
    //                      Mapper implementation
    ///////////////////////////////////////////////////////////////////////////
    public void insertDocument(InputSource input)
    throws XMLDBCException, SAXException
    {
        insertXMLDocument(input);
    }
    
    public void insertDocument(XMLReader parser, InputSource input)
    throws XMLDBCException, SAXException
    {
        insertXMLDocument(parser, input);
    }
    
    public void insertDocument(String doc) throws XMLDBCException, org.xml.sax.SAXException
    {
        insertXMLDocument(doc);
    }
    
    public void insertDocument(Document document) throws XMLDBCException
    {
        insertXMLDocument(document);
    }
    
    /**
     * Retrieves the current document base used by this object.
     *
     * @return an URI.
     * @see #setDocumentBase(String)
     */
    public String getDocumentBase()
    {
        return validatingContentHandler.getDocumentBase();
    }
    
    /**
     * This method is to be used when the document stored contains an 'xsi:schemaLocation'
     * attribute (or any element or attribute with the XML Schema type 'anyURI')
     * that uses relatives URIs and when the API used to store documents does not
     * use an InputSource with the SystemID set (e.g. DOM or String ones). The document
     * base provided is then used to resolve the relative URI into an absolute one.
     *
     * @param URI The URI to be used as base during document storage.
     * @see #getDocumentBase
     */
    public void setDocumentBase(String URI)
    {
        validatingContentHandler.setDocumentBase(URI);
    }

    /**
     * Returns the JDBC connection used by this object.
     * @return a JDBC connection.
     */
    public Connection getConnection()
    {
        return connection.getConnection();
    }
    
}


