/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.bridge;

import java.io.CharArrayWriter;
import java.io.IOException;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;

import javax.sql.DataSource;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParserFactory;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.xml.sax.*;
import org.xquark.mapper.dbms.AbstractConnection;
import org.xquark.mapper.dbms.AbstractConnectionFactory;
import org.xquark.mapper.mapping.*;
import org.xquark.mapper.storage.MapperTupleBuilder;
import org.xquark.mapper.storage.SAXHandler;
import org.xquark.mapper.storage.ValidatingFilerAdapter;
import org.xquark.mapper.util.RepositoryProperties;
import org.xquark.schema.SchemaManager;
import org.xquark.schema.validation.SchemaValidationContext;
import org.xquark.schema.validation.ValidatingSchemaFilter;
import org.xquark.serialize.XMLSerializer;
import org.xquark.util.SAXConstants;
import org.xquark.xml.xdbc.XMLDBCException;
import org.xquark.xpath.XTreeReader;

/**
 * An object implementing this class represents a processed mapping file.
 * Objects of this class are thread-safe and can be reused multiple times.<BR>
 * This class uses a cache for loaded mapping files, so that the loading overhead is
 * reduced. The cache mechanism can be overriden to reload a modified file.
 */

public class Mapping
{
	private static final String 	RCSRevision 	= "$Revision: 1.13 $";
	private static final String 	RCSName     	= "$Name: XQ1_1_1 $";
    
	private static Log log = LogFactory.getLog(Mapping.class);
	private static SAXParserFactory parserFactory;
	static
	{
		parserFactory = SAXParserFactory.newInstance();
		parserFactory.setNamespaceAware(true);
        
		/* load configuration information for batch for instance */
		RepositoryProperties.load();
	}
    
    private DataSource dataSource;
    private String systemId = null;
    private RepositoryMapping metadata = null;
    private HashSet mappers = new HashSet();
    private SchemaManager schemaManager;  // this is to be independant if MappingMetadata is reset
    
	/**
	 * Constructor.
	 * @param dataSource a JDBC data source
	 * @param source the SAX InputSource encapsulating the mapping file
	 * @param reload true means that the mapping file should be reloaded, even
	 * if it is present in the cache
	 * @param classLoader ClassLoader used for loading user generators
	 * @param schemaManager SchemaManager for schemas
	 * @param mappingMetadata if not null, a HashMap holding previously loaded
	 * mapping metadata
     * @throws XMLDBCException if an error occured while loading the mapping
     * file. The underlying exception provides the root cause for the exception
	 */
	public Mapping(DataSource dataSource, InputSource source, boolean reload,
			ClassLoader classLoader, SchemaManager schemaManager, HashMap mappingMetadata)
	throws XMLDBCException {
		this.dataSource = dataSource;
		this.schemaManager = schemaManager;
		Connection jdbcConn = null;
		AbstractConnection connection = null;
		try {
			jdbcConn = dataSource.getConnection();
			connection = AbstractConnectionFactory.newConnection(jdbcConn, true);
			systemId = source.getSystemId();
			if (mappingMetadata != null && systemId != null && !reload) {
				metadata = (RepositoryMapping) mappingMetadata.get(systemId);
			}
			if (metadata == null) {
				XMLReader reader = new ValidatingSchemaFilter(parserFactory.newSAXParser().getXMLReader(),
															  new SchemaValidationContext(schemaManager));
				reader.setErrorHandler(new ValidationErrorHandler());
				Loader loader = new Loader(reader, connection, schemaManager, classLoader, false);
				MappingFactory mappingFactory = new MappingFactory(schemaManager, schemaManager);
				mappingFactory.createTree();
				metadata = loader.load(source, mappingFactory, true);
				if (log.isDebugEnabled()) {
					XMLReader parser = new XTreeReader(metadata);
					parser.setFeature(SAXConstants.TRACE_FEATURE, true);
					CharArrayWriter caw = new CharArrayWriter();
					parser.setContentHandler(new XMLSerializer(caw));
					parser.parse("");
					log.debug(caw);
				}
				if (mappingMetadata != null && systemId != null) {
					mappingMetadata.put(systemId, metadata);
				}
			}
		} catch (ParserConfigurationException ex) {
			throw new XMLDBCException("Could not load SAX parser", ex);
		} catch (SQLException e) {
			throw new XMLDBCException("Error while loading mapping", e); 
		} catch (IOException e) {
			throw new XMLDBCException("Error while loading mapping", e); 
		} catch (SAXException e) {
			throw new XMLDBCException("Error while loading mapping", e); 
		} finally {
			try {
				if (connection != null) connection.close(); // this closes also jdbcConn
				else if (jdbcConn != null) jdbcConn.close();
			} catch (SQLException e) {
			}
		}
	}
    
    /**
     * Frees all resources associated to the object
     * @throws XMLDBCException if a database connection error occured.
     */
    public synchronized void close() throws XMLDBCException
    {
        Iterator it = mappers.iterator();
        while (it.hasNext()) ((Mapper)it.next()).close();
    }
    
    /**
     * Creates a new Mapper object for storing a set of XML documents
     * in the database
     * @return a Mapper object
     * @throws XMLDBCException if a database connection error occured.
     * @see Mapper
     */
    public Mapper getMapper() throws XMLDBCException
    {
        SchemaValidationContext validationContext = new SchemaValidationContext(schemaManager);
        ValidatingFilerAdapter validator = new ValidatingFilerAdapter(validationContext, schemaManager);
		Connection jdbcConn = null;
		AbstractConnection connection = null;
		try {
			jdbcConn = dataSource.getConnection();
			connection = AbstractConnectionFactory.newConnection(jdbcConn, true);
		} catch (SQLException ex) {
			throw new XMLDBCException("Could not create connection", ex);		
		}
        SAXHandler handler = new SAXHandler(new MappingIterator(metadata),
        									connection,
       	 									new MapperTupleBuilder(connection,
        														   metadata,
        									                       validationContext)
        									);
        validator.setSAXHandler(handler);
        Mapper result = new MapperImpl(this, connection, validator, handler, handler, handler);
        synchronized (mappers) {
			mappers.add(result);
        }
        return result;
    }
    
    /**
     * Returns the system id (the URL) of the mapping file, if known.
     * @return the system id of the mapping file, or null if unknown.
     */
    public String getSystemId() { return systemId;}
    
    /**
     * Returns the JDBC data source used by this mapping.
     * @return a JDBC datasource.
     */
    public DataSource getDataSource() { return dataSource;}
    
    /**
     * Package-private method returning the internal description of the
     * mapping file
     * @return the internal representation of the mapping file
     */
    RepositoryMapping getMetaData() { return metadata;}
    
    /**
     * Package-private maethod used to remove a Mapper from its parent list
     */
    synchronized void remove(MapperImpl mapper) { mappers.remove(mapper);}

	private class ValidationErrorHandler implements ErrorHandler
	{
		public static final String 	RCSRevision 	= "$Revision: 1.13 $";
		public static final String 	RCSName     	= "$Name: XQ1_1_1 $";
        
		////////////////////////////////////////////////////////////////////
		//    Implementation of ErrorHandler interface.
		////////////////////////////////////////////////////////////////////
		public void warning(SAXParseException ex) throws SAXException {
			throw ex;
		}
        
		public void error(SAXParseException ex) throws SAXException {
			throw ex;
		}
        
		public void fatalError(SAXParseException ex) throws SAXException {
			throw ex;
		}
	}
}


