/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.extractor;

import java.io.IOException;
import java.net.URL;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xquark.extractor.common.MessageLibrary;
import org.xquark.extractor.metadata.MetaDataManager;
import org.xquark.extractor.oracle.OracleQueryFactory;
import org.xquark.extractor.runtime.ExtractorStatement;
import org.xquark.extractor.runtime.PreparedExtractorStatement;
import org.xquark.extractor.runtime.QueryFactory;
import org.xquark.xml.xdbc.*;
import org.xquark.xquery.metadata.resolver.MetadataAccess;
import org.xquark.xquery.parser.XQueryException;
import org.xquark.xquery.xdbc.XMLDataSourceMetaDataConnectionWrapper;

/**
 * Extractor implementation of {@link org.xquark.xml.xdbc.XMLConnection}.
 */
public class ExtractorConnection extends DefaultReadOnlyXMLConnection implements XMLConnection {

	private static final String RCSRevision = "$Revision: 1.10 $";
	private static final String RCSName = "$Name: XQ1_1_1 $";

	private Connection _jdbcConnection;
	private Extractor _extractor = null;
	private MetaDataManager metaDataManager = null;
	private XMLDataSourceMetaData metaData = null;

	/* */
	private boolean _isClosed = false;
	private String _baseURI = null;

	/* a list all active XMLStatement that this XMLConnection has created */
	private List _activeStatementList = new ArrayList();

	private int stmtCounter = 0; // TODO: recycle

	ExtractorConnection(Extractor extractor, Connection jdbcConnection) {
		this._extractor = extractor;
		this._jdbcConnection = jdbcConnection;
	}

	public Connection getJdbcConnection() {
		return _jdbcConnection;
	}

	public Extractor getExtractor() {
		return _extractor;
	}

	public synchronized void close() throws XMLDBCException {
		if (!_isClosed) {
			try {
				_isClosed = true;
				XMLStatement statement;
				for (int i = _activeStatementList.size() - 1; i > -1; i--) {
					statement = (XMLStatement) _activeStatementList.get(i);
					if (statement != null)
					    statement.close();
				}
				_jdbcConnection.close();
			} catch (SQLException ex) {
				throw new XMLDBCException("Error while closing connection", ex);
			}
		}
	}

	public void commit() throws XMLDBCException, XMLDBCNotSupportedException {
		if (!_isClosed) {
			try {
				_jdbcConnection.commit();
			} catch (SQLException ex) {
				throw new XMLDBCException(ex.getMessage(), ex);
			}
		} else
			throw new XMLDBCException("XMLConnection is already closed.");
	}

	public boolean isClosed() throws XMLDBCException {
		return _isClosed;
	}

	public String getUserName() throws XMLDBCException {
		return _extractor.getUserName();
	}

	public String getURL() throws XMLDBCException {
		return _extractor.getURL();
	}

	private QueryFactory createQueryFactory(String baseURI) throws XMLDBCException {
		if (isClosed())
			throw new XMLDBCException(MessageLibrary.getMessage("CON_CLOSED"));

		QueryFactory queryFactory = null;
		String dbName = _extractor.getDatabaseName();
		String dbVersion = _extractor.getDatabaseVersion();

		/* DBMS version release switch (Statement implementations */
		if (-1 < dbName.indexOf("Oracle")) {
			if (-1 < dbVersion.indexOf("Release 8.1"))
				queryFactory = new OracleQueryFactory(getMetaDataManager(), getMetadataAccess());
			else if (-1 < dbVersion.indexOf("Release 7.3.4"))
				queryFactory = new org.xquark.extractor.oracle.Oracle7QueryFactory(getMetaDataManager(), getMetadataAccess());
			else // default Oracle is Oracle 8i
				queryFactory = new org.xquark.extractor.oracle.OracleQueryFactory(getMetaDataManager(), getMetadataAccess());
		} else if (-1 < dbName.indexOf("Microsoft"))
			queryFactory = new org.xquark.extractor.microsoft.MicrosoftQueryFactory(getMetaDataManager(), getMetadataAccess());
		else if (-1 < dbName.indexOf("Adaptive Server Enterprise"))
			queryFactory = new org.xquark.extractor.sybase.SybaseQueryFactory(getMetaDataManager(), getMetadataAccess());
		else if (-1 < dbName.indexOf("PROGRESS"))
			queryFactory = new org.xquark.extractor.progress.ProgressQueryFactory(getMetaDataManager(), getMetadataAccess());
		else if (-1 < dbName.indexOf("MySQL"))
			queryFactory = new org.xquark.extractor.mysql.MysqlQueryFactory(getMetaDataManager(), getMetadataAccess());
		else
			throw new XMLDBCNotSupportedException(MessageLibrary.getMessage("D_N_SUP_DB_VENDOR", dbName));
		
		queryFactory.setBaseURI(baseURI);
		return queryFactory;
	}

	public XMLStatement createStatement(short queryType) throws XMLDBCException {
		XMLStatement retVal = new ExtractorStatement(this, stmtCounter++, createQueryFactory(_baseURI), getMetaDataManager());
		synchronized (this) {
		    _activeStatementList.add(retVal);
		}
		return retVal;
	}

	public XMLStatement createStatement() throws XMLDBCException {
		return createStatement(XQUERY_STRING_TYPE);
	}

	public PreparedXMLStatement prepareStatement(String query) throws XMLDBCException, XMLDBCNotSupportedException {
		PreparedXMLStatement retVal = new PreparedExtractorStatement(this, stmtCounter++, createQueryFactory(_baseURI), getMetaDataManager(), query, _extractor.getCachedStatements());
		synchronized (this) {
		    _activeStatementList.add(retVal);
		}
		return retVal;
	}

	/**
	 * Gets the metadata regarding this connection's data source. A
	 * connection's data source is able to provide information
	 * describing itself.
	 * @return an XMLDataSourceMetaData object for this connection.
	 * @throws XMLDBCException if a data source access error occurs.
	 */
	public XMLDataSourceMetaData getMetaData() throws XMLDBCException {
		return getMetaData(false);
	}

	public XMLDataSourceMetaData getMetaData(boolean refresh) throws XMLDBCException {
		if (metaData == null || refresh) {
			metaData = new XMLDataSourceMetaDataConnectionWrapper(this, _extractor.getDataSourceMetaData(_jdbcConnection, refresh));
			metaDataManager = _extractor.getMetaDataManager(_jdbcConnection, false);
		}
		return metaData;
	}

	MetaDataManager getMetaDataManager() throws XMLDBCException {
		return getMetaDataManager(false);
	}

	MetaDataManager getMetaDataManager(boolean refresh) throws XMLDBCException {
		if (metaDataManager == null || refresh) {
			metaDataManager = _extractor.getMetaDataManager(_jdbcConnection, refresh);
			metaData = new XMLDataSourceMetaDataConnectionWrapper(this, _extractor.getDataSourceMetaData(_jdbcConnection, false));
		}
		return metaDataManager;
	}

	MetadataAccess getMetadataAccess() throws XMLDBCException {
		return _extractor.getDataSourceMetaData(_jdbcConnection, false);
	}

	public synchronized void statementClosed(XMLStatement closedStatement) throws XMLDBCException {
		_activeStatementList.remove(closedStatement);
	}

	protected void finalize() throws Exception {
		close();
	}

	// add LARS 18/05/04
	public void loadModule(String moduleURL) throws XMLDBCException {
		try {
			URL url = new URL(moduleURL);
			InputSource source = new InputSource(url.openStream());
			source.setSystemId(url.toString());
			getMetadataAccess().getModuleManager().loadModule(source, getMetadataAccess(), getMetadataAccess().getSchemaManager());
		} catch (IOException xqe) {
			throw new XMLDBCException("Could not read module file", xqe);
		} catch (XQueryException xqe) {
			throw new XMLDBCException("Could not read module file", xqe);
		}
	}
	
	public void loadSchema(InputSource source) throws SAXException, XMLDBCException {
		getMetadataAccess().getSchemaManager().loadSchema(source);
	}
	
	/**
	 * Sets the base URI for documents.
	 * @return nothing 
	 */
	public void setBaseURI(String baseURI) {
		this._baseURI = baseURI;
	}

}
