/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.extractor.algebra;

import java.util.*;

import org.xquark.extractor.common.Debug;
import org.xquark.extractor.common.SqlWrapperException;

public final class AddKeys extends DefaultAlgebraVisitor {
	private static final String RCSRevision = "$Revision: 1.9 $";
	private static final String RCSName = "$Name: XQ1_1_1 $";

	private Stack _returnStack = new Stack();
	private boolean _cleBidon = false;

	public AddKeys() {
	}

	public void reinit() {
		_returnStack.clear();
		_cleBidon = false;
	}

	public boolean isHeavyEliminationNeeded() {
		return _cleBidon;
	}

	/**
	 * @return a list of keys, each key is a Expression
	 */
	public List getKeys() {
		List retVal = (List) _returnStack.peek();
		Debug.assertTrue(retVal.size() > 0, "retVal.size()>0");
		return retVal;
	}

	public void visit(DummyTable arg) throws SqlWrapperException {
		// TODO: Change it ?: keys == null is used to say if a relation is Dummy by MySQL rewriting
		//		throw new SqlWrapperException(MessageLibrary.getMessage("IE_ERR","DummyTable can not provide key."));
	}

	public void visit(Expression arg) throws SqlWrapperException {
		Debug.assertTrue(false, "AddKeys.visit() not implemented for " + arg.pprint());
	}

	public void visit(BinOpOuterJoin arg) throws SqlWrapperException {
		List allKeys = new ArrayList();

		List keys;
		Expression operand;

		operand = arg.getLeftOperand();
		operand.accept(this);
		keys = (List) _returnStack.pop();
		allKeys.addAll(keys);

		operand = arg.getRightOperand();
		operand.accept(this);
		keys = (List) _returnStack.pop();
		allKeys.addAll(keys);

		arg.setKeys(allKeys);
		_returnStack.push(allKeys);
	}

	public void visit(Join arg) throws SqlWrapperException {
		List allKeys = new ArrayList();
		List keys;

		Iterator iter = arg.getOperandList().iterator();
		Expression operand = null;
		while (iter.hasNext()) {
			operand = (Expression) iter.next();
			operand.accept(this);
			keys = (List) _returnStack.pop();
			allKeys.addAll(keys);
		}

		arg.setKeys(allKeys);
		_returnStack.push(allKeys);
	}

	public void visit(RenameRelation arg) throws SqlWrapperException {
		arg.getOperand().accept(this);
		List allKeys = (List) _returnStack.peek();

		List keys = null;
		try {
			keys = AlgebraTools.clone(allKeys); // QUESTION: why cloning ?
		} catch (CloneNotSupportedException ex) {
			throw new SqlWrapperException(ex.getMessage(), ex);
		}

		/* set the _tableInstance property of all AttributeExpression to current node */
		Iterator iter = keys.iterator();
		while (iter.hasNext()) {
			((AttributeExpression) iter.next()).setTableInstance(arg);
		}

		arg.setKeys(keys);
		_returnStack.push(keys);
	}

	public void visit(Table arg) throws SqlWrapperException {
		List allKeys = new ArrayList();
		/** @todo  from non-oracle DB */
		List keysList = arg.getKeys();
		List key = null;
		if (null == keysList || keysList.isEmpty()) {
			key = arg.getAttributes();
			_cleBidon = true;
		} else
			key = (List) keysList.get(0); // TODO: do not take the pseudo-id if PK is a single column ? (to limit column retrieval)

		arg.setUsedKeyAttributeList(key);

		Attribute keyAttribute = null;
		AttributeExpression keyAttributeExpression = null;

		for (int i = 0; i < key.size(); i++) {
			keyAttribute = (Attribute) key.get(i);
			if (keyAttribute instanceof PseudoAttribute) {
				keyAttributeExpression = new Rowid(null, keyAttribute.getName());
			} else {
				keyAttributeExpression = new AttributeExpression(null, keyAttribute.getName());
			}
			keyAttributeExpression.setUnderlyingExpr(keyAttribute);
			allKeys.add(keyAttributeExpression);
		}

		_returnStack.push(allKeys);
	}

	public void visit(UnaryOperator arg) throws SqlWrapperException {
		arg.getOperand().accept(this);
	}

	public void visit(UnOpProject arg) throws SqlWrapperException {
		List retVal = null;
		AttributeExpression newAttrItem;

		if (arg.getDistinct()) {
			/* if the UnOproject is SELECT DISTINCT:
			* 1 we can not add other columns to the select List .
			* 2 there will not be duplicate tuples in the rerult ,
			* so the select List can play the role of key
			*/
			List selectList = arg.getItemList();
			List allKeys = new ArrayList();
			Iterator iter = selectList.iterator();
			Expression item;
			while (iter.hasNext()) {
				item = (Expression) iter.next();
				newAttrItem = new AttributeExpression(null, item.getName());
				newAttrItem.setUnderlyingExpr(item);
				allKeys.add(newAttrItem);
			}
			retVal = allKeys;
		} else {
			/* if the UnOproject is SELECT ALL:
			* add key columns to the select List .
			*/
			if (null == arg.getKeys()) {
				arg.getOperand().accept(this);
				List keys = (List) _returnStack.pop();

				List newKeys = new ArrayList();

				Iterator iter = keys.iterator();
				AttributeExpression newItem;
				Expression selectItem, keyItem;
				List selectList = arg.getItemList();
				while (iter.hasNext()) {
					keyItem = (AttributeExpression) iter.next();
					selectItem = UnOpProject.identicalExpr(selectList, (AttributeExpression) keyItem);
					if (null == selectItem) {
						keyItem = arg.addItem(keyItem);
					} else {
						keyItem = selectItem;
					}
					newAttrItem = new AttributeExpression(null, keyItem.getName());
					newAttrItem.setUnderlyingExpr(keyItem);

					newKeys.add(newAttrItem);
				}
				arg.setKeys(newKeys);
				retVal = newKeys;
			} else {
				retVal = arg.getKeys();
			}
		}

		_returnStack.push(retVal);
	}

	public void visit(UnOpGroup arg) throws SqlWrapperException {
		List keys = null;

		try {
			keys = AlgebraTools.clone(arg.getKeys());
		} catch (CloneNotSupportedException ex) {
			throw new SqlWrapperException(ex.getMessage(), ex);
		}

		_returnStack.push(keys);
	}
}
