/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.extractor.algebra;

import java.util.*;

import org.xquark.extractor.common.SqlWrapperException;
import org.xquark.extractor.xfunctions.XFunction;

/**
 * Set the {@link org.xquark.extractor.algebra.Expression#_referredAttributes}
 * field for an algebra tree with the attributes encountered during the subtree
 * with the exception of {@link Relation#visibleTableInstances()} which are
 * removed.
 */
public final class BuildReferredAttributes extends  DefaultCompleteVisitor {

    private static final String RCSRevision = "$Revision: 1.6 $";
    private static final String RCSName     = "$Name: XQ1_1_1 $";


    private Stack _returnStack = new Stack();

    public BuildReferredAttributes() {
    }

    public void visit(Expression arg) throws SqlWrapperException {
        _returnStack.push(null);
    }

    public void visit(BinaryAlgebra arg) throws SqlWrapperException {
        Set tmpCol = new HashSet();

        arg.getLeftOperand().accept(this);
        Set  popped = (Set)_returnStack.pop();
        if ( null != popped ) {
            tmpCol.addAll(popped);
        }

        arg.getRightOperand().accept(this);
        popped = (Set)_returnStack.pop();
        if ( null != popped ) {
            tmpCol.addAll(popped);
        }

        arg.setReferredAttributes(tmpCol);
        _returnStack.push(tmpCol);
    }

    public void visit(Join arg) throws SqlWrapperException {
        Set retVal = new HashSet();
        Set tmpSet = null;
        List list = arg.getOperandList();
        Expression expr = null;
        if (null != list) {
            for (int i = 0; i < list.size(); i++) {
                expr = (Expression)list.get(i);
                expr.accept(this);
                tmpSet = (Set)_returnStack.pop();
                if (null != tmpSet) {
                    retVal.addAll(tmpSet);
                }
            }
        }

        list = arg.getPredicateList();
        if (null != list) {
            for (int i = 0; i < list.size(); i++) {
                expr = (Expression)list.get(i);
                expr.accept(this);
                tmpSet = (Set)_returnStack.pop();
                if (null != tmpSet) {
                    retVal.addAll(tmpSet);
                }
            }
        }

        List localAttributeList = new ArrayList();
        Set  localTableInstance = ((Relation)arg).visibleTableInstances();
        AttributeExpression attr = null;
        for (Iterator i = retVal.iterator(); i.hasNext(); ) {
            attr = (AttributeExpression)i.next();
            if (localTableInstance.contains(attr.getTableInstance())) {
                localAttributeList.add(attr);
            }
        }

        retVal.removeAll(localAttributeList);

        if (retVal.isEmpty()) {
            retVal = null;
        }

        _returnStack.push(retVal);
    }


    public void visit(BinOpOuterJoin arg) throws SqlWrapperException {
        Set retVal = new HashSet();
        Set tmpSet = null;

        arg.getLeftOperand().accept(this);
        tmpSet = (Set)_returnStack.pop();
        if (null != tmpSet) {
            retVal.addAll(tmpSet);
        }

        arg.getRightOperand().accept(this);
        tmpSet = (Set)_returnStack.pop();
        if (null != tmpSet) {
            retVal.addAll(tmpSet);
        }


        List list = arg.getPredicateList();
        Expression expr = null;
        if (null != list) {
            for (int i = 0; i < list.size(); i++) {
                expr = (Expression)list.get(i);
                expr.accept(this);
                tmpSet = (Set) _returnStack.pop();
                if (null != tmpSet) {
                    retVal.addAll(tmpSet);
                }
            }
        }

        List localAttributeList = new ArrayList();
        Set  localTableInstance = ((Relation)arg).visibleTableInstances();
        AttributeExpression attr = null;
        for (Iterator i = retVal.iterator(); i.hasNext(); ) {
            attr = (AttributeExpression)i.next();
            if (localTableInstance.contains(attr.getTableInstance())) {
                localAttributeList.add(attr);
            }
        }

        retVal.removeAll(localAttributeList);

        if (retVal.isEmpty()) {
            retVal = null;
        }

        _returnStack.push(retVal);
    }

    public void visit(UnaryAlgebra arg) throws SqlWrapperException {
        arg.getOperand().accept(this);
        Set retVal = (Set)_returnStack.pop();

        if (null == retVal) {
            retVal = new HashSet();
        }

        List list = arg.getParameterList();
        Expression expr = null;
        if (null != list) {
            for (int i = 0; i < list.size(); i++) {
                expr = (Expression)list.get(i);
                expr.accept(this);
                Set tmpSet = (Set) _returnStack.pop();
                if (null != tmpSet) {
                    retVal.addAll(tmpSet);
                }
            }
        }

        List localAttributeList = new ArrayList();
        Set  localTableInstance = ((Relation)arg).visibleTableInstances();
        AttributeExpression attr = null;
        for (Iterator i = retVal.iterator(); i.hasNext(); ) {
            attr = (AttributeExpression)i.next();
            if (localTableInstance.contains(attr.getTableInstance()) ||
				attr.getTableInstance() == null ) {
                localAttributeList.add(attr);
            }
        }

        retVal.removeAll(localAttributeList);

        _returnStack.push(retVal);
    }

    public void visit(AttributeExpression arg) throws SqlWrapperException {
        Set tmpCol = new HashSet();
        tmpCol.add(arg);
        arg.setReferredAttributes(tmpCol);

        _returnStack.push(tmpCol);
    }

    public void visit(BinaryAtomicOp arg) throws SqlWrapperException {
        Set tmpCol = new HashSet();

        arg.getLeftOperand().accept(this);
        Set  popped = (Set)_returnStack.pop();
        if ( null != popped ) {
            tmpCol.addAll(popped);
        }

        arg.getRightOperand().accept(this);
        popped = (Set)_returnStack.pop();
        if ( null != popped ) {
            tmpCol.addAll(popped);
        }

        arg.setReferredAttributes(tmpCol);
        _returnStack.push(tmpCol);
    }

    public void visit(OuterJoinPredicate arg) throws SqlWrapperException {
        arg.getPredicate().accept(this);

        Set tmpCol = (Set)_returnStack.peek();
        arg.setReferredAttributes(tmpCol);
    }

    public void visit(UnaryAtomicOp arg) throws SqlWrapperException {
        arg.getOperand().accept(this);

        Set tmpCol = (Set)_returnStack.peek();
        arg.setReferredAttributes(tmpCol);
    }

    public void visit(XFunction arg) throws SqlWrapperException
    {
        Set retVal = null;
        List argList = arg.getArguments();
        if (null != argList){
            retVal= new HashSet();
            Object aArg = null;
            for (int i = 0; i < argList.size(); i++) {

                aArg = argList.get(i);
                if (aArg instanceof Expression) {
                    ((Expression)aArg).accept(this);
                    Set  popped = (Set)_returnStack.pop();
                    if ( null != popped ) {
                        retVal.addAll(popped);
                    }
                }
            }
            if (retVal.isEmpty()) {
                retVal = null;
            }
        }

        arg.setReferredAttributes(retVal);
        _returnStack.push(retVal);
    }
}
