/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.extractor.algebra;

import java.util.*;

import org.xquark.extractor.common.Debug;
import org.xquark.extractor.common.SqlWrapperException;
import org.xquark.xquery.parser.XQueryExpression;

/**
 * This visitor seems to clean useless projection and rename nodes that are
 * systematically generated from Qdb.
 * TODO: check the precise behavior. 
 */
public final class RemoveIntermediateProjectVisitor extends DefaultCompleteVisitor {
	private static final String RCSRevision = "$Revision: 1.8 $";
	private static final String RCSName = "$Name: XQ1_1_1 $";

	private TopDownReplacer _topDownReplacer;
	private ReplaceTractor _replaceTractor;

	private Stack _scopeStack = new Stack();

	private RemoveProjectVoteVisitor _voteVisitor = new RemoveProjectVoteVisitor();

	public void reinit() {
		_scopeStack.clear();
		_voteVisitor.reinit();
		createScope();
	}

	public RemoveIntermediateProjectVisitor() {
	}

	public TopDownReplacer getReplacerInstance() {
		if (null == _topDownReplacer) {
			_topDownReplacer = new TopDownReplacer();
		}
		return _topDownReplacer;
	}

	public ReplaceTractor getTractorInstance() {
		if (null == _replaceTractor) {
			_replaceTractor = new ReplaceTractor();
		}
		return _replaceTractor;
	}

	private boolean isDeletePermitted() {
		//        Trace.enter(this, "isDeletePermitted()");
		boolean retVal = ((Scope) _scopeStack.peek())._deletePermited;
		//        Trace.exit(this, "isDeletePermitted()", retVal.toString());
		return retVal;
	}

	private void permitDelete(Expression node, boolean permit) {
		//Trace.enter(this, "permitDelete(boolean permit)", node.toString() + "\t" + (permit ? "true" : "false"));
		Scope currentScope = (Scope) _scopeStack.peek();
		currentScope._deletePermited = permit;
		currentScope._node = node;
		//Trace.exit(this, "permitDelete(boolean permit)", node.toString() + "\t" + (permit ? "true" : "false"));
	}

	private Expression currentPermiter() {
		//Trace.enter(this, "currentPermiter()");
		Scope currentScope = (Scope) _scopeStack.peek();
		Debug.assertTrue(currentScope._deletePermited, "currentScope._deletePermited");
		//Trace.exit(this, "currentPermiter()");
		return currentScope._node;
	}

	private void createScope() {
		_scopeStack.push(new Scope());
	}

	private Scope currentScope() {
		return (Scope) _scopeStack.peek();
	}

	private void leaveScope() {
		//Trace.enter(this, "leaveScope()");
		RenameRelation rr = currentScope()._unresolvedRenameRelation;
		if (null != rr)
			removeUnresolvedRename(rr);

		_scopeStack.pop();
		//Trace.exit(this, "leaveScope()");
	}

	private void removeUnresolvedRename(RenameRelation rr) {
		if (null != rr.getFather() && !(rr.getOperand() instanceof Table)) {
			/* remove this RenameRelation*/
			//Trace.message(this, "leaveScope()", "Removing  unresolved RenameRelation...");
			/* make up map*/
			HashMap map = new HashMap();
			Object tableInstance = null;
			Set vtis = rr.visibleTableInstances();
			int visibleTableInstancesNumber = vtis.size();

			if (0 == visibleTableInstancesNumber) {
				Debug.assertTrue(false, "internal logic error");
			} else if (1 == visibleTableInstancesNumber) {
				tableInstance = (Expression) vtis.toArray()[0];
				if (!(tableInstance instanceof RenameRelation)) {
					tableInstance = new NullPointer();
				}
				map.put(rr, tableInstance);
			} else {
				Debug.nyi("remove RenameRelation");
			}

			/* */
			TopDownReplacer replacer = getReplacerInstance();
			replacer.setReplaceMap(map);

			ReplaceTractor tractor = getTractorInstance();
			tractor.setVisitor(replacer);
			tractor.setEndNode(currentPermiter());
			//Trace.message(this, "replace map", Debug.pprint(map));

			Expression father = rr.getFather();
			boolean replaced = father.replaceChild(rr, rr.getOperand());
			Debug.assertTrue(replaced, "replaced");
			rr.setFather(null);
			//Trace.message(this, "leaveScope()", "RenameRelation removed.");
			father.accept(tractor);
		}
	}

	//public void visit(Expression arg) throws SqlWrapperException;

	public void visit(BinaryAlgebra arg) throws SqlWrapperException {
		//Trace.enter(this, "visit(BinaryAlgebra arg)", arg);
		/* implementation or Merg, Union, Intersect and difference, but not for OuterJoin */
		/* here the process of deepth-first serach re-start */
		createScope();
		arg.getLeftOperand().accept(this);
		leaveScope();

		createScope();
		arg.getRightOperand().accept(this);
		leaveScope();

		//Trace.exit(this, "visit(BinaryAlgebra arg)", arg);
	}

	public void visit(BinaryAtomicOp arg) throws SqlWrapperException {
		//Trace.enter(this, "visit(BinaryAtomicOp arg)", arg);

		createScope();
		arg.getLeftOperand().accept(this);
		leaveScope();

		createScope();
		arg.getRightOperand().accept(this);
		leaveScope();

		//Trace.exit(this, "visit(BinaryAtomicOp arg)", arg);
	}

	public void visit(BinOpOuterJoin arg) throws SqlWrapperException {
		//Trace.enter(this, "visit(BinOpOuterJoin arg)", arg);

		arg.getLeftOperand().accept(this);
		arg.getRightOperand().accept(this);

		//Trace.exit(this, "visit(BinOpOuterJoin arg)", arg);
	}

	public void visit(SortSpecification arg) throws SqlWrapperException {
		//Trace.enter(this, "visit(SortSpecification arg)", arg);
		Expression sortExpr = arg.getSortExpression();
		sortExpr.accept(this);
		//Trace.exit(this, "visit(SortSpecification arg)", arg);
	}

	public void visit(RenameRelation arg) throws SqlWrapperException {
		//Trace.enter(this, "visit(RenameRelation arg)", arg);

		/** @todo within a scope, there might be serverl  unresolved RenameRelation.
		 *  Scope class should keep a list of  unresolved RenameRelation  instead of one*/
		if (!(arg.getOperand() instanceof Table || arg.getOperand() instanceof UnOpProject)) {
			/* when the operand is Table, this RenameRelation should not be removed
			*  when the operand is UnOpProject, this RenameRelation should have been
			*  removed with the UnOpProject */
			currentScope()._unresolvedRenameRelation = arg;
			//Trace.message(this, "visit(RenameRelation arg)", "unresolved RenameRelation added to current scope");
		}

		Expression operand = arg.getOperand();
		operand.accept(this);

		//Trace.exit(this, "visit(RenameRelation arg)", arg);
	}

	public void visit(Table arg) throws SqlWrapperException {
		//Trace.enter(this, "visit(Table arg)", arg);
		/* here, the process reach a leaf node */
		//Trace.exit(this, "visit(Table arg)", arg);
	}

	public void visit(Join arg) throws SqlWrapperException {
		//Trace.enter(this, "visit(Join arg)", arg);

		List list = arg.getPredicateList();
		if (null != list) {
			Expression expr = null;
			for (int i = 0; i < list.size(); i++) {
				expr = (Expression) list.get(i);
				createScope();
				expr.accept(this);
				leaveScope();
			}
		}

		list = arg.getOperands();
		Debug.assertTrue(null != list, "null!=list");
		Expression expr = null;
		for (int i = 0; i < list.size(); i++) {
			expr = (Expression) list.get(i);
			expr.accept(this);
		}

		//Trace.exit(this, "visit(Join arg)", arg);
	}

	public void visit(UnaryAlgebra arg) throws SqlWrapperException {
		//Trace.enter(this, "visit(UnaryAlgebra arg)", arg);

		List list = arg.getParameterList();
		if (null != list) {
			Expression expr = null;
			for (int i = 0; i < list.size(); i++) {
				expr = (Expression) list.get(i);
				createScope();
				expr.accept(this);
				leaveScope();
			}
		}

		Expression operand = arg.getOperand();
		if (isDeletePermitted()) {
			/* check if this operater permet to delete a underlying UnOpProject*/
			List opposers = vote(arg.getParameterList());
			if (null == opposers || opposers.isEmpty()) {
				/* do nothing */
			} else {
				permitDelete(arg, false);
			}
		}
		operand.accept(this);

		//Trace.exit(this, "visit(UnaryAlgebra arg)", arg);
	}

	public void visit(UnaryAtomicOp arg) throws SqlWrapperException {
		//Trace.enter(this, "visit(UnaryAtomicOp arg)", arg);
		arg.getOperand().accept(this);
		//Trace.exit(this, "visit(UnaryAtomicOp arg)", arg);
	}

	public void visit(UnOpAggregate arg) throws SqlWrapperException {
		//Trace.enter(this, "visit(UnOpAggregate arg)", arg);

		Expression operand = arg.getOperand();

		/* check if this UnOpAggregate permet to delete a underlying UnOpProject*/
		createScope();
		List opposers = vote(arg.getItemList());
		if (null == opposers || opposers.isEmpty()) {
			permitDelete(arg, true);
			//Trace.message(this, "visit(UnOpAggregate arg)", "removal of project is permitted !");
		}
		operand.accept(this);
		leaveScope();

		List itemList = arg.getItemList();
		Expression item = null;
		;
		for (int i = 0; i < itemList.size(); i++) {
			createScope();
			item = (Expression) itemList.get(i);
			item.accept(this);
			leaveScope();
		}

		//Trace.exit(this, "visit(UnOpAggregate arg)", arg);
	}

	public void visit(UnOpExists arg) throws SqlWrapperException {
		//Trace.enter(this, "visit(UnOpExists arg)", arg);
		createScope();
		boolean doSpecial = false;
		Expression op = arg.getOperand();
		if (op instanceof UnOpProject) {
			if (((UnOpProject)op).getOperand() instanceof Join)
				doSpecial = true;
		}
		// change LARS 30/10/03 true to false
		if (doSpecial)
			permitDelete(arg, false);
		else
			permitDelete(arg, true);
		arg.getOperand().accept(this);
		leaveScope();
		//Trace.exit(this, "visit(UnOpExists arg)", arg);
	}

	public void visit(UnOpProject arg) throws SqlWrapperException {
		//Trace.enter(this, "visit(UnOpProject arg)", arg);

		Expression operand = arg.getOperand();

		/* visit its operand */
		//Trace.message(this, "visit(UnOpProject arg)", "Visiting its operand...");
		createScope();
		/* check
		
		if this UnOpProject permet to delete a underlying UnOpProject*/
		List opposers = vote(arg.getItemList());
		if (null == opposers || opposers.isEmpty()) {
			permitDelete(arg, true);
			//Trace.message(this, "visit(UnOpProject arg)", "removal of project is permitted !");
		}
		operand.accept(this);
		leaveScope();

		/* visit its items */
		//Trace.message(this, "visit(UnOpProject arg)", "Visiting its item list...");
		List itemList = arg.getItemList();
		Expression item = null;
		boolean hasAggregateItem = false;
		if (itemList != null)
			for (int i = 0; i < itemList.size(); i++) {
				createScope();
				item = (Expression) itemList.get(i);
				item.accept(this);
				// Check if this Project contains Aggregate items
				hasAggregateItem = (item instanceof RenameItem && ((RenameItem)item).getOperand() instanceof UnOpAggregate);
			 	leaveScope();
			}

		/* remove itself if possible */
		if (isDeletePermitted()) {
			if (currentPermiter() instanceof UnOpAggregate) {
				// Aggregate items prevent the Project to be removed in an Aggregate
				if (!hasAggregateItem) {
					if (!arg.getDistinct()) {
						removeProject(arg, currentPermiter());
					} else {
						/* @todo verify that currentPermiter contains only aggregation funtions like :
						 *  aggrFunc(attr), but not aggrFunc(*)
						 */
						if (1 == arg.getItemList().size()) {
							removeProject(arg, currentPermiter());
							((UnOpAggregate) currentPermiter()).distinctAggregation();

						} else {
							Debug.nyi("1 != arg.getItemList().size()");
						}
					}
				}
			} else if (!arg.getDistinct()) {
				removeProject(arg, currentPermiter());
			}
		}

		//Trace.exit(this, "visit(UnOpProject arg)", arg);
	}

	private void removeProject(UnOpProject project, Expression permittingNode) {
		//Trace.enter(this, "removeProject(UnOpProject project, Expression permittingNode)");

		//Trace.message(this, "visit(UnOpProject arg)", "starting GoUpVisitor..");
		HashMap map = makeMap(project);

		/* */
		TopDownReplacer replacer = getReplacerInstance();
		replacer.setReplaceMap(map);

		ReplaceTractor tractor = getTractorInstance();
		tractor.setVisitor(replacer);
		tractor.setEndNode(currentPermiter());
		//Trace.message(this, "replace map", Debug.pprint(map));

		Expression father = project.getFather();

		boolean replaced = father.replaceChild(project, project.getOperand());
		Debug.assertTrue(replaced, "replaced");

		project.setFather(null);
		//Trace.message(this, "visit(UnOpProject arg)", "project removed.");
		father.accept(tractor);

		//Trace.exit(this, "removeProject(UnOpProject project, Expression permittingNode)");
	}

	private List vote(List exprList) {
		return null;
		//        Trace.enter(this,"vote(List exprList)");
		//
		//		_voteVisitor.reinit();
		//		Expression expr = null;
		//		for (int i = 0; i < exprList.size(); i++) {
		//			expr = (Expression)exprList.get(i);
		//			expr.accept(_voteVisitor);
		//		}
		//        List retVal = _voteVisitor.getOpposerList();
		//
		//        Trace.exit(this,"vote(List exprList)");
		//        return retVal;
	}

	private HashMap makeMap(UnOpProject arg) {
		//Trace.enter(this, "makeMap(UnOpProject arg))", arg.pprint());

		HashMap retVal = new HashMap();
		List itemList = arg.getItemList();
		Expression item = null;
		String attrName = null;
		AttributeExpression attrExpr = null;
		Expression replacer = null;
		XQueryExpression originalXExpr = null;

		if (itemList != null)
			for (int i = 0; i < itemList.size(); i++) {
				item = (Expression) itemList.get(i);
				attrName = item.getName();
				attrExpr = new AttributeExpression(null, attrName);

				//            originalXExpr = item.getOrginalXExpr();
				//            Debug.assertTrue(null != originalXExpr, "null != originalXExpr" + item.pprint());

				replacer = findReplacer(item);
				retVal.put(attrExpr, replacer);
			}

		/* create a map item for "*" */
		Set vtiSet = arg.visibleTableInstances();
		if (1 == vtiSet.size()) {
			Expression vti = (Expression) vtiSet.toArray()[0];
			if (vti instanceof RenameRelation) {
				attrExpr = new AttributeExpression(null, "*");
				replacer = new AttributeExpression((RenameRelation) vti, "*");
				retVal.put(attrExpr, replacer);
			}
		}

		//Trace.exit(this, "makeMap(UnOpProject arg))", arg.pprint());
		return retVal;
	}

	private Expression findReplacer(Expression item) {
		//Trace.enter(this, "findReplacer(Expression item)", item.pprint());

		Expression retVal = null;

		if (item instanceof RenameItem) {
			retVal = ((RenameItem) item).getOperand();
			retVal = findReplacer(retVal);
		}
		//        else  if (item instanceof AttributeExpression) {
		//            retVal = ((AttributeExpression)item).getUnderlyinExpr();
		//            if (retVal instanceof Attribute) {
		//                /* item is the first AttributeExpression who refer a Attribute */
		//                retVal = item;
		//            }
		//            retVal = item;
		//        }
		else {
			retVal = item;
			//            Debug.assertTrue(false,"internal logic error");
		}

		//Trace.exit(this, "findReplacer(Expression item)", item.pprint());
		return retVal;
	}

	class Scope {
		private static final String RCSRevision = "$Revision: 1.8 $";
		private static final String RCSName = "$Name: XQ1_1_1 $";

		boolean _deletePermited = false;
		Expression _node = null;
		RenameRelation _unresolvedRenameRelation = null;
		Scope() {
		}
		Scope(Expression node, boolean deletePermited) {
			_node = node;
			_deletePermited = deletePermited;
		}
	}

	final class ReplaceTractor extends BottomUpTractor {
		ReplaceTractor() {
			super();
		}

		public void visit(RenameRelation arg) throws SqlWrapperException {
			//Trace.enter(this, "visit(RenameRelation arg)", arg);

			/* set arg to "TableInstance" attribute for all AttributExpression*/
			Map map = ((TopDownReplacer) getVisitor()).getReplaceMap();

			HashMap newMap = new HashMap();
			Set keySet = map.keySet();
			Iterator iter = keySet.iterator();
			AttributeExpression item = null;
			Object value = null;
			while (iter.hasNext()) {
				item = (AttributeExpression) iter.next();
				value = map.get(item);

				item.setTableInstance(arg);
				newMap.put(item, value);
			}

			((TopDownReplacer) getVisitor()).setReplaceMap(newMap);

			//Trace.message(this, "visit(RenameRelation arg)", "updated replac map :" + Debug.pprint(newMap));

			/* remove this node */
			Expression father = arg.getFather();
			boolean replaced = father.replaceChild(arg, arg.getOperand());
			Debug.assertTrue(replaced, "replaced");
			arg.setFather(null);
			//Trace.message(this, "visit(RenameRelation arg)", "RenameRelation removed");

			if (getEndNode() != arg) {
				father.accept(this);
			}

			//Trace.exit(this, "visit(RenameRelation arg)", arg);
		}

		public void visit(UnOpProject arg) throws SqlWrapperException {
			//Trace.enter(this, "UnOpProject(Join arg)", arg.pprint());

			List list = arg.getParameterList();

			if (null != list) {
				treat(list);
			}

			UnOpSort sort = arg.getAssociatedSort();
			if (null != sort) {
				SortSpecification spec = null;
				list = sort.getSortSpecificationList();
				treat(list);
			}

			visitFather(arg);

			//Trace.exit(this, "UnOpProject(Join arg)", arg.pprint());
		}
	}
}
