/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * Copyright (C) 2004 Odonata.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.extractor.algebra;

import java.sql.Types;
import java.util.*;

import org.xquark.extractor.common.Debug;
import org.xquark.jdbc.typing.DbType;
import org.xquark.schema.*;
import org.xquark.xquery.parser.QName;
import org.xquark.xquery.parser.XQueryException;
import org.xquark.xquery.parser.XQueryExpression;
import org.xquark.xquery.typing.*;

/**
 * This class is used by the GenAlgebraVisitor to perform a "quick" translation
 * of XML typing coming from the query into an SQL (JDBC) data type. The mapping
 * is defined statically and used by {@link org.xquark.extractor.algebra.GenAlgebraVisitor}
 * to check the query is consistent from a SQL type point of view and mostly to
 * guide algebra construction during operator generation, e.g. according to the
 * fact that one operator is an atom or not (a constant or a column).
 */
// TODO: check it is not redundant with front-end processing. Especially when multiplicity
// if set by the new QType flag.
public final class TypeInterpreter extends DefaultQTypeVisitor
{
	private static final String RCSRevision = "$Revision: 1.11 $";
	private static final String RCSName = "$Name: XQ1_1_1 $";

	/** The static type mapping table between XML and SQL atom types */
	private static final int[] CODE_TABLE = { -1, Types.VARCHAR, Types.BIT, Types.DECIMAL, Types.DOUBLE, Types.FLOAT, -1, Types.TIMESTAMP, Types.TIME, Types.DATE, -1, -1, -1, -1, -1 - 1, -1, -1, -1, -1, -1 };

	private SqlType _type = null;
	
	public TypeInterpreter() {}

	private int sqlTypeCode(int schemaTypeCode) {
		int retVal = -1;
		if (schemaTypeCode < 1 || schemaTypeCode > 20)
			retVal = -1;
		else
			retVal = CODE_TABLE[schemaTypeCode];

		return retVal;
	}

	public SqlType interprete(QType qtype) throws XQueryException {
		//Trace.enter(this, "interprete (QType qtype)");
		Debug.assertTrue(null != qtype, "null != qtype");

		qtype.accept(this);
		
		if (_type != null) {
			if (qtype.isMultiple())
				_type.setMultiplicity(SqlType.MANY);
			else
				_type.setMultiplicity(SqlType.ONE);
		}

		//Trace.exit(this, "interprete (QType qtype)");
		return _type;
	}

	public void visit(QTypeAtom qtype) throws XQueryException {
		//Trace.enter(this, "interprete (QTypeAtom qtype)");

		_type = interprete(qtype.getSimpleType());

		//Trace.exit(this, "interprete (QTypeAtom qtype)");
	}

	public void visit(QTypeText qtype) throws XQueryException {
		//Trace.enter(this, "interprete (QTypeText qtype)");

		_type = interprete(qtype.getSimpleType());

		//Trace.exit(this, "interprete (QTypeText qtype)");
	}

	public void visit(QTypeElement qte) throws XQueryException {
		//Trace.enter(this, "interprete(QTypeElement qte)");
		Type type = qte.getType();
		if (type instanceof SimpleType) { // qtype contains a SimpleType
			SqlTypeAtom sqlAtom;
			sqlAtom = (SqlTypeAtom) interprete((SimpleType) type);
			if (null != sqlAtom) {
				// set name
				XQueryExpression name = qte.getName();
				Debug.assertTrue((name instanceof QName), "(name instanceof QName)");
				sqlAtom.setName(((QName) name).getName());
				_type = sqlAtom;
			}
		}
		else // QTypeElement contains a ComplexType
			_type = interprete((ComplexType) type);

		//Trace.exit(this, "interprete (QTypeElement qtype)");
	}

	public void visit(QTypeSequence qtype) throws XQueryException {
		//Trace.enter(this, "interprete(QTypeSequence qtype)");

		SqlType sqlType = interprete((QType) qtype.getList().get(0));
		if (sqlType.isRelation()) {
			_type = VerifyRelation(qtype);
		} else if (sqlType.isAtom()) {
			_type = VerifyColumn(qtype);
		} else {
			Debug.assertTrue(false, "NYI");
		}

		//Trace.enter(this, "interprete(QTypeSequence qtype)");
	}

	public void visit(QTypeDocument qtype) throws XQueryException {
		//Trace.enter(this, "interprete(QTypeDocument qtype)");

		// change LARS 29/06/04
		if (qtype.isFromDocument())
			throw new XQueryException("Function doc is not supported");
		_type = interprete(qtype.getQType());

		//Trace.exit(this, "interprete(QTypeDocument qtype)");
	}

	private SqlType VerifyColumn(QTypeSequence qtype) throws XQueryException {
		//Trace.enter(this, "VerifyColumn(QTypeSequence qtype)");

		SqlTypeStructure retVal = new SqlTypeStructure(null);

		SqlType sqlType = null;
		Iterator iter = qtype.getList().iterator();
		while (iter.hasNext()) {
			//Debug.assertTrue(.........);
			sqlType = interprete((QType) iter.next());
			if (null == sqlType) {
				retVal = null;
				break;
			}
			if (sqlType.isAtom()) {
				retVal.addAttribute((SqlTypeAtom) sqlType);
			} else if (sqlType.isTuple()) {
				retVal.addAttributes(((SqlTypeStructure) sqlType).getAttributes());
			} else {
				retVal = null;
				break;
			}
		}

		//Trace.exit(this, "VerifyColumn(QTypeSequence qtype)");
		return retVal;
	}

	private SqlType VerifyRelation(QTypeSequence qtype) throws XQueryException {
		//Trace.enter(this, "VerifyRelation(QTypeSequence qtype)");
		SqlType retVal = null;

		SqlType sqlType = null;
		Iterator iter = qtype.getList().iterator();
		while (iter.hasNext()) {
			//Debug.assertTrue(.........);
			sqlType = interprete((QType) iter.next());
			if (null == sqlType && !sqlType.isRelation()) {
				retVal = null;
				break;
			} else if (null == retVal) {
				retVal = sqlType;
			}

		}

		//Trace.exit(this, "VerifyRelation(QTypeSequence qtype)");
		return retVal;
	}

	private SqlType interprete(SimpleType stype) {
		//Trace.enter(this, "iinterprete(SimpleType stype)");
		SqlTypeAtom retVal;
		Debug.assertTrue(SimpleType.VARIETY_ATOMIC == stype.getVariety(), "SimpleType.VARIETY_ATOMIC==stype.getVariety()");

		DbType dt = new DbType(sqlTypeCode(stype.getPrimitive().getType()));
		retVal = new SqlTypeAtom(dt, null, SqlType.ONE);

		//Trace.exit(this, "interprete", "SimpleType");
		return retVal;
	}

	private SqlType interprete(ComplexType stype) {
		//Trace.enter(this, "interprete(SimpleType stype)");

		SqlTypeStructure retVal = null;
		retVal = new SqlTypeStructure(null, SqlType.ONE);

		Collection elements = getElementDeclarations(stype);
		ElementDeclaration element;
		SqlType elementType;
		Iterator iter = elements.iterator();
		Debug.assertTrue(iter.hasNext(), "iter.hasNext()");
		while (iter.hasNext()) {
			element = (ElementDeclaration) iter.next();
			elementType = interprete(element);
			if (elementType != null && elementType.isAtom()) {
				retVal.addAttribute((SqlTypeAtom) elementType);
			} else {
				retVal = null;
				break;
			}
		}
		//Trace.exit(this, "interprete(SimpleType stype)");
		return retVal;
	}

	private SqlType interprete(ElementDeclaration stype) {
		//Trace.enter(this, "interprete(ElementDeclaration stype)");

		SqlType retVal = null;

		Type schemaType = stype.getType();
		if (schemaType.isSimpleType()) {
			retVal = interprete((SimpleType) schemaType);
			((SqlTypeAtom) retVal).setName(stype.getName());
		} else {
			retVal = interprete((ComplexType) schemaType);
		}

		//Trace.exit(this, "interprete(ElementDeclaration stype)");
		return retVal;
	}

	private List getElementDeclarations(ComplexType complextype) {
		//Trace.enter(this, "getElementDeclarations(ComplexType complextype)");

		List retVal = new ArrayList();
		ContentModel contentmodel = complextype.getContentModel();
		Particle particle = (Particle) contentmodel.getModel();
		Object term = particle.getTerm();
		if (term instanceof ElementDeclaration)
			retVal.add(term);
		else if (term instanceof ModelGroup) {
			ModelGroup modelgroup = (ModelGroup) term;
			int compositor = modelgroup.getCompositor();
			Debug.assertTrue(compositor == ModelGroup.SEQUENCE, "compositor == ModelGroup.SEQUENCE");

			for (int i = 0; i < modelgroup.size(); i++) {
				Particle part = (Particle) modelgroup.get(i);
				ElementDeclaration elemDecl = (ElementDeclaration) part.getTerm();
				retVal.add(elemDecl);
			}
		}

		//Trace.exit(this, "getElementDeclarations(ComplexType complextype)");
		return retVal;
	}
}
