/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.extractor.algebra;

import java.util.*;

import org.xquark.extractor.common.Debug;
import org.xquark.extractor.common.SqlWrapperException;
import org.xquark.extractor.runtime.IDProvider;
import org.xquark.extractor.sql.SqlExpression;
import org.xquark.extractor.xfunctions.AfTrim;

public class UnOpProject extends UnaryAlgebra {
	private static final String RCSRevision = "$Revision: 1.15 $";
	private static final String RCSName = "$Name: XQ1_1_1 $";

	protected boolean _distinct = false;
	protected List _itemList = null;
	protected List _itemNames = null;
	protected UnOpSort _associatedSort = null;
	protected IDProvider _attIdProvider = null;

	private Set _providedTableInstances = Collections.singleton(this);


	public UnOpProject(IDProvider attIdProvider) {
		super();
		_attIdProvider = attIdProvider;
	}

	public UnOpProject(Expression operand, IDProvider attIdProvider) {
		super(operand);
		_attIdProvider = attIdProvider;
	}

	public UnOpProject(Expression operand, Expression item, IDProvider attIdProvider) {
		this(operand, attIdProvider);
		addItem(item);
	}

	public UnOpProject(Expression operand, List itemList, boolean distinct, IDProvider attIdProvider) {
		this(operand, attIdProvider);
		setItemList(itemList);
		setDistinct(distinct);
	}

	synchronized Object clone(Map clonedExprs) throws CloneNotSupportedException {
		UnOpProject retVal = (UnOpProject) super.clone(clonedExprs);
		if (null != _itemList) {
			List itemList = AlgebraTools.clone(_itemList, clonedExprs);
			List itemNames = AlgebraTools.clone(_itemNames);
			Expression item = null;
			for (int i = 0; i < itemList.size(); i++) {
				item = (Expression) itemList.get(i);
				item.setFather(retVal);
			}
			retVal._itemList = itemList;
			retVal._itemNames = itemNames;
		}

		clonedExprs.put(this, retVal);
		return retVal;
	}

	/**
	 * Determines if the _distinct property is true.
	 * 
	 * @return <code>true<code> if the _distinct property is true
	 */
	public boolean getDistinct() {
		return _distinct;
	}

	/**
	 * Sets the value of the _distinct property.
	 * 
	 * @param aDistinct the new value of the _distinct property
	 */
	public void setDistinct(boolean aDistinct) {
		_distinct = aDistinct;
	}

	/**
	 * Access method for the _itemList property.
	 * 
	 * @return the current value of the _itemList property
	 */
	public List getItemList() {
		return _itemList;
	}

	public void setAssociatedSort(UnOpSort associatedSort) {
		_associatedSort = associatedSort;
	}

	public UnOpSort getAssociatedSort() {
		return _associatedSort;
	}

	/**
	 * Sets the value of the _itemList property.
	 * 
	 * @param aItemList the new value of the _itemList property
	 */
	public void setItemList(List itemList) {
		_itemList = null;
		_itemNames = null;

		if (null == itemList) {
			return;
		}
		addItemList(itemList);
	}

	public void addItemList(List itemList) {
		if (null == itemList) {
			return;
		}

		Iterator iter = itemList.iterator();
		while (iter.hasNext()) {
			Expression item = (Expression) iter.next();
			addItem(item);
		}
	}

	public Expression addItem(Expression itemExpr) {
		if (null != getItemList()) {
			return insertItem(getItemList().size(), itemExpr);
		} else {
			return insertItem(0, itemExpr);
		}
	}

	public Expression setItem(int index, Expression itemExpr) {
		_itemList.remove(index);
		_itemNames.remove(index);
		return insertItem(index, itemExpr);
	}

	public Expression insertItem(int index, Expression itemExpr) {
		Expression retVal = null;

		if (null == _itemList) {
			_itemList = new ArrayList();
			_itemNames = new ArrayList();
		}

		if (itemExpr instanceof Rowid) {
			itemExpr = new RenameItem(itemExpr, "OracleRowid");
		}

		if (itemExpr instanceof TupleExpression) {
			addItemTupleExpression((TupleExpression) itemExpr);
		} else if (itemExpr instanceof Relation) {
			retVal = new RenameItem(itemExpr, _attIdProvider);
			retVal.setOrginalXExpr(itemExpr.getOrginalXExpr());
			_itemList.add(index, retVal);
			_itemNames.add(index, retVal.getName());
		} else {
			String name = itemExpr.getName();
			if (null == name) {
				// when the new added item does not have a name, give it a anonymity
				retVal = new RenameItem(itemExpr, _attIdProvider);
				retVal.setOrginalXExpr(itemExpr.getOrginalXExpr());
				_itemNames.add(index, retVal.getName());
				_itemList.add(index, retVal);
			} else {
				// when the new added item havs a name, check if it falls in conflict with the existing names
				if (!hasNameConflict(itemExpr)) {
					//no nameing conflict
					_itemNames.add(index, name);
					_itemList.add(index, itemExpr);
					retVal = itemExpr;
				} else {
					//when there is a nameing conflict, rename the new added item
					_itemNames.add(index, name); // QUESTION : why stored before rename ?
					if (itemExpr instanceof RenameItem) {
						((RenameItem) itemExpr).setName(name, _attIdProvider);
						retVal = itemExpr;
					} else {
						retVal = new RenameItem(itemExpr, name, _attIdProvider);
						retVal.setOrginalXExpr(itemExpr.getOrginalXExpr());
					}
					_itemList.add(index, retVal);
				}
			}
		}

		if (null != retVal) {
			retVal.setFather(this);
		}

		return retVal;
	}

	public void insertItemWithoutRename(int index, FunAggregate funAggregate) {
		insertItem(index, funAggregate);
	}

	private boolean hasNameConflict(Expression itemExpr) {
		return hasNameConflict(itemExpr, _itemList);
	}

	public static boolean hasNameConflict(Expression itemExpr, List itemList) {
		boolean retVal = false;

		String newName = itemExpr.getName();
		if (null != itemList && null != itemExpr.getName()) {
			Expression item = null;
			for (int i = 0; i < itemList.size(); i++) {
				item = (Expression) itemList.get(i);

				if (item.getName().equalsIgnoreCase(itemExpr.getName())) {
					retVal = true;
					break;
				}
			}
		}
		return retVal;
	}

	public boolean removeItem(Expression itemExpr) {
		boolean retVal = false;

		int index = _itemList.indexOf(itemExpr);
		if (-1 != index) {
			_itemList.remove(index);
			_itemNames.remove(index);
			retVal = true;
		}
		return retVal;
	}

	private List addItemTupleExpression(TupleExpression tupleExpr) {
		List attrList = tupleExpr.getItemList();
		List retVal = new ArrayList();

		Iterator iter = attrList.iterator();
		Expression item = null;
		while (iter.hasNext()) {
			item = (Expression) iter.next();
			retVal.add(addItem(item));
		}
		tupleExpr.setItemList(retVal);
		/* after above line , tupleExpr is the father of all these items
		* so ...*/
		for (int i = 0; i < retVal.size(); i++) {
			item = (Expression) retVal.get(i);
			item.setFather(this);
		}
		return retVal;
	}

	public List getKeys() {
		return _keys;
	}

	public void setKeys(List keys) {
		_keys = keys;
	}

	public Expression SearchItem(Expression expr) {
		Expression retVal = null;
		Expression item;
		for (int i = 0; i < _itemList.size(); i++) {
			item = (Expression) _itemList.get(i);
			if (item.equals(expr)) {
				retVal = item;
				break;
			} else {
				item = getRealExpression(item);
				if (item.equals(expr)) {
					retVal = item;
					break;
				}
			}
		}
		return retVal;
	}

	private Expression getRealExpression(Expression expr) {
		if (expr instanceof RenameItem) {
			return ((RenameItem) expr).getOperand();
		} else {
			return expr;
		}
	}

	public void arithmeticOperation(int operater, Expression leftOperand, Expression rightOperand) {
		Iterator iter = _itemList.iterator();
		Expression item;
		List newItemList = new ArrayList();
		Expression arOp;
		while (iter.hasNext()) {
			item = getRealExpression((Expression) iter.next());
			if (null != leftOperand) {
				arOp = new BinOpArithmetic(operater, leftOperand, item);
			} else {
				arOp = new BinOpArithmetic(operater, item, rightOperand);
			}
			newItemList.add(arOp);
		}
		setItemList(newItemList);
	}

	public void addNameTest(String name) {
		List itemList = ((Relation) _operand).nameTest(name);
		Expression item = null;

		AttributeExpression attrExpr = null;
		for (int i = 0; i < itemList.size(); i++) {
			item = (Expression) itemList.get(0);
			/* remove heading and tailing spaces */
			if (item instanceof Attribute) {
				if (((Attribute) item).getPrimitiveType().isString()) {
					item = new AfTrim(item);
				}
			}
			addItem(item);
		}
	}

	public List nameTest(String name) {
		List retVal = null;

		if (name.equals("*")) {
			retVal = _itemList;
			// in this case , there will be a  problem when there are duplicated names
		} else if (null == name) {
			retVal = new ArrayList();
			for (int i = 0; i < _itemNames.size(); i++) {
				if (null == _itemNames.get(i)) {
					retVal.add(_itemList.get(i));
				}
			}
			if (0 == retVal.size()) {
				retVal = null;
			}

		} else {
			// Name is not a wild card
			retVal = new ArrayList();
			for (int i = 0; i < _itemNames.size(); i++) {
				if (name.equalsIgnoreCase((String) _itemNames.get(i))) {
					retVal.add(_itemList.get(i));
				}
			}
			if (0 == retVal.size()) {
				retVal = null;
			}
		}

		return retVal;
	}

	public boolean predicateUp(Expression predicate) {
		return false;
	}

	public Mapper getMapper() {
		/* default implementation*/
		return _mapper;
	}

	public boolean replaceChild(Expression oldChild, Expression newChild) {
		boolean retVal = false;
		List list = getItemList();

		Expression item = null;
		String itemName = null;
		RenameItem renameItem = null;

		for (int i = 0; i < list.size(); i++) {
			item = (Expression) list.get(i);
			if (item.equals(oldChild)) {
				itemName = item.getName();
				Debug.assertTrue(null != itemName, "null != itemName");
				if (!itemName.equals(newChild.getName())) {
					/* keep the old name of this item so that the mapper can still works*/
					renameItem = new RenameItem(newChild, itemName);
					renameItem.setOrginalXExpr(item.getOrginalXExpr());
					list.set(i, renameItem);
					renameItem.setFather(this);
				} else {
					list.set(i, newChild);
					newChild.setFather(this);
				}
				retVal = true;
				break;
			}
		}

		if (getOperand().equals(oldChild)) {
			setOperand(newChild);
			retVal = true;
		}

		return retVal;
	}

	public Set providedTableInstances() {
		return _providedTableInstances;
	}

	public AttributeExpression findNonNullAttribute() {
		Iterator it = _itemList.iterator();
		Expression expr = null;
		AttributeExpression ret = null;

		while (it.hasNext() && ret == null) {
			expr = (Expression) it.next();
			if (expr instanceof AttributeExpression) {
				ret = (AttributeExpression) expr;
				Attribute att = (Attribute) ret.getUnderlyinExpr();
				if (!att.getAttribute().getNullable()) {
					ret = new AttributeExpression(ret.getTableInstance(), ret.getName());
					ret.setUnderlyingExpr(att);
				}
			} else if (expr instanceof RenameItem) {
				ret = new AttributeExpression(null, ((RenameItem) expr).getName());
			}
		}
		return ret;
	}

	public List getParameterList() {
		return getItemList();
	}

	static public Expression identicalExpr(List existingExprList, AttributeExpression newExpr) {
		Expression retVal = null;
		Expression expr = null;
		for (int i = 0; i < existingExprList.size(); i++) {
			expr = (Expression) existingExprList.get(i);
			if (isIdentical(expr, newExpr)) {
				retVal = expr;
				break;
			}
		}
		return retVal;
	}

	static public boolean isIdentical(Expression existingExpr, AttributeExpression newExpr) {
		boolean retVal = true;
		if (existingExpr instanceof AttributeExpression) {
			AttributeExpression attrExpr = (AttributeExpression) existingExpr;
			retVal = newExpr.getAttribute().equals(attrExpr.getAttribute());
			if (retVal && null != newExpr.getTableInstance()) {
				retVal = newExpr.getTableInstance().equals(attrExpr.getTableInstance());
			} else if (retVal && null == newExpr.getTableInstance()) {
				retVal = (null == attrExpr.getTableInstance());
			}
		} else if (existingExpr instanceof RenameItem) {
			RenameItem renameItem = (RenameItem) existingExpr;
			retVal = isIdentical(renameItem.getOperand(), newExpr);
		} else {
			retVal = false;
		}
		return retVal;
	}

	public SqlExpression accept(GenSqlVisitor visitor) throws SqlWrapperException {
		return visitor.visit(this);
	}

	public void accept(AlgebraVisitor visitor) throws SqlWrapperException {
		visitor.visit(this);
	}

	/**
	 * @see Expression#deepEquals(Object)
	 */
	public boolean deepEquals(Object o) {
		if (o instanceof UnOpProject) {
			UnOpProject cast = (UnOpProject) o;
			return super.deepEquals(o) && _distinct == cast.getDistinct() && _associatedSort != null && _associatedSort.deepEquals(cast.getAssociatedSort()) && AlgebraTools.areExprListEquivalent(_itemList, cast.getItemList());
		}
		return false;
	}
}
