/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.extractor.metadata;

import java.io.*;
import java.util.*;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;

import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.InputSource;
import org.xml.sax.XMLReader;
import org.xquark.extractor.algebra.CollectionName;
import org.xquark.extractor.runtime.Selection;
import org.xquark.jdbc.typing.DBMSInfo;
import org.xquark.schema.validation.SchemaValidationContext;
import org.xquark.schema.validation.ValidatingSchemaHandler;
import org.xquark.serialize.XMLSerializer;
import org.xquark.util.NamespaceContextStack;
import org.xquark.xml.xdbc.XMLDBCException;
import org.xquark.xquery.metadata.*;
import org.xquark.xquery.parser.*;
import org.xquark.xquery.parser.util.Constants;

/**
 * Object that encapsulates relational metadata information for a data source ( a relational database. This object is also used, basing on this information, to build medatata used by XQuery parser and the XDBC API (@see org.xquark.xquery.metadata.MetaDataImpl}).
 */
public class MetaDataManager {
    private static final String RCSRevision = "$Revision: 1.23 $";

    private static final String RCSName = "$Name: XQ1_1_1 $";

    /**
     * root NamedNode of the tree repressenting the relational metadata (site, catalogs, schema, tables, views. A site is another term for data source. It is loaded by
     * 
     * @link Loader basing on
     * @link Selection.
     */
    private Site _site = null;

    private Properties _tableMap = null;

    private DBMSInfo dbmsInfo = null;

    private List _schemaList = null;

    private List _moduleList = null;

    public MetaDataManager(DBMSInfo dbmsInfo) {
        this.dbmsInfo = dbmsInfo;
    }

    public void setSite(Site site) {
        _site = site;
    }

    public void setSchemaList(List schemaList) {
        _schemaList = schemaList;
    }

    public void setModuleList(List moduleList) {
        _moduleList = moduleList;
    }

    public List findTables(CollectionName colName) {
        //Trace.enter(this, "findTables(CollectionName colName)", colName);

        String catalogName = colName.getCatalog();
        String schemaName = colName.getSchema();
        String tableName = colName.getTable();
        List retVal = new ArrayList();
        Catalog catalog = null;

        if (null == catalogName || catalogName.equals("*")) {
            List nodeList = _site.getChildren();
            for (int i = 0; i < nodeList.size(); i++) {
                catalog = (Catalog) nodeList.get(i);
                retVal.addAll(findTables(catalog, schemaName, tableName));
            }
        } else {
            catalog = (Catalog) _site.findChild(catalogName);
            if (null != catalog) {
                retVal.addAll(findTables(catalog, schemaName, tableName));
            }
        }

        //Trace.exit(this, "findTables(String catalog String schema,String table)");
        return retVal;
    }

    private List findTables(Catalog catalog, String schemaName, String tableName) {
        //Trace.enter(this, "findTables(Catalog catalog, String schemaName, String tableName)", catalog.getName());

        List retVal = new ArrayList();

        Schema schema = null;
        if (null == schemaName || schemaName.equals("*")) {
            List nodeList = catalog.getChildren();
            for (int i = 0; i < nodeList.size(); i++) {
                schema = (Schema) nodeList.get(i);
                List tables = findTables(schema, tableName);
                if (null != tables) {
                    retVal.addAll(findTables(schema, tableName));
                }
            }
        } else {
            schema = (Schema) catalog.findChild(schemaName);
            if (null != schema) {
                retVal.addAll(findTables(schema, tableName));
            }
        }

        //Trace.exit(this, "findTables(Catalog catalog, String schemaName, String tableName)", catalog.getName());
        return retVal;
    }

    private List findTables(Schema schema, String tableName) {
        //Trace.enter(this, "findTables(Schema schema, String tableName)", schema.getName());

        List retVal = null;

        if (null == tableName || tableName.equals("*")) {
            retVal = schema.getChildren();
        } else {
            Table table = (Table) schema.findChildByAlias(tableName);
            if (null != table) {
                retVal = new ArrayList(1);
                retVal.add(table);
            }
        }

        //Trace.exit(this, "findTables(Schema schema, String tableName)", schema.getName());
        return retVal;
    }

    public void setTableMap(Properties tableMap) {
        _tableMap = tableMap;
    }

    public String originalNameOf(String fullEncodedName) {
        return _tableMap.getProperty(fullEncodedName);
    }

    public String pprint() {
        String retVal = null;
        return retVal;
    }

    /**
     * Builds metadata used by XQuery parser and for XDBC API.
     * 
     * @param name
     *            datasource symbolic name read in the configuration file or a default one.
     * @return @throws
     *         XMLDBCException
     */
    public MetaDataImpl buildMetaDataForParser(String name) throws XMLDBCException {

        if (_site == null)
            return null;

        boolean doModules = (_moduleList != null && !_moduleList.isEmpty());
        StringBuffer bufSchemas = new StringBuffer();
        StringBuffer bufModules = new StringBuffer();
        // only for compatibility purpose
        StringBuffer bufCollections = new StringBuffer();
        
        List catalogs = this._site.getChildren();
        String nameSpace = null;
        // create database associated metadata
        if (catalogs != null)
            for (int cat = 0; cat < catalogs.size(); cat++) {
                Catalog catalog = (Catalog) catalogs.get(cat);

                List schemas = catalog.getChildren();

                if (schemas != null) {
                    for (int i = 0; i < schemas.size(); i++) {
                        Schema schema = (Schema) schemas.get(i);
                        String targetNamespace = schema.getTargetNamespace();
                        int elementFormDefault = schema.getElementFormDefault();

                        bufSchemas.append("\n\t\t<schema\n\t\t\txmlns=\"http://www.w3.org/2001/XMLSchema\"\n\t\t\telementFormDefault=\"");
                        bufSchemas.append(Selection.ELEMENT_FORM_DEFAULT[elementFormDefault]);
                        bufSchemas.append("\"");
                        if (targetNamespace == null) {
                            bufSchemas.append(">");
                        } else {
                            bufSchemas.append(" targetNamespace=\"");
                            bufSchemas.append(targetNamespace);
                            bufSchemas.append("\">");
                        }

                        bufModules.append("\n\t\t<module");
                        String moduleTargetNamespace = targetNamespace;
                        if (moduleTargetNamespace == null) {
                            moduleTargetNamespace = "http://database";
                            if (catalog.getName() != null)
                                moduleTargetNamespace += "/" + catalog.getName();
                            moduleTargetNamespace += "/" + schema.getName();
                        }
                        bufModules.append(" xmlns:db=\"");
                        bufModules.append(moduleTargetNamespace);
                        bufModules.append("\" targetNamespace=\"");
                        bufModules.append(moduleTargetNamespace);
                        bufModules.append("\"");
                        if (catalog.getName() != null) {
                            bufModules.append(" catalogName=\"");
                            bufModules.append(catalog.getName());
                            bufModules.append("\"");
                        }
                        bufModules.append(" schemaName=\"");
                        bufModules.append(schema.getName());
                        bufModules.append("\">");
                        List tables = schema.getChildren();
                        if (tables != null) {
                            for (int j = 0; j < tables.size(); j++) {
                                Table table = (Table) tables.get(j);

                                List attributes = table.getChildren();
                                if (attributes == null)
                                    continue;

                                String collectionName = table.getAlias();
                                bufSchemas.append("\n\t\t\t<element name=\"");
                                bufSchemas.append(collectionName);
                                bufSchemas.append("\">\n\t\t\t\t<complexType>\n\t\t\t\t\t<sequence>");

                                // only for compatibility purpose
                                String fullCollectionName = table.getAlias();
								if (schema.getName() != null && schema.getName().length() != 0)
									fullCollectionName = schema.getName() + "." + fullCollectionName;
								if (catalog.getName() != null && catalog.getName().length() != 0)
									fullCollectionName = catalog.getName() + "." + fullCollectionName;
								bufCollections.append("\n\t\t<collection name=\"");
								bufCollections.append(fullCollectionName);
								//bufCollections.append(collectionName);
								bufCollections.append("\">");
								if (targetNamespace == null) {
									bufCollections.append("\n\t\t\t<step name=\"");
									bufCollections.append(collectionName);
									bufCollections.append("\">");
								} else {
									bufCollections.append("\n\t\t\t<step ns=\"");
									bufCollections.append(targetNamespace);
									bufCollections.append("\" name=\"");
									bufCollections.append(collectionName);
									bufCollections.append("\">");
								}

								for (int k = 0; k < attributes.size(); k++) {
                                    Attribute attribute = (Attribute) attributes.get(k);
                                    ExtractorMappingInfo mappingInfo = attribute.getMappingInfo();

                                    bufSchemas.append("\n\t\t\t\t\t\t<element name=\"");
                                    bufSchemas.append(attribute.getAlias());
                                    if (mappingInfo.getColumnMetaData().isNullable()) {
                                        bufSchemas.append("\" minOccurs=\"0");
                                    }
                                    if (mappingInfo.isBuiltInXMLType()) {
                                        bufSchemas.append("\" type=\"");
                                        bufSchemas.append(mappingInfo.getStringXMLType());
                                        bufSchemas.append("\"/>");
                                    } else {
                                        bufSchemas.append("\">\n\t\t\t\t\t\t\t");
                                        bufSchemas.append(mappingInfo.getStringXMLType());
                                        bufSchemas.append("\n\t\t\t\t\t\t</element>");
                                    }

                                    // only for compatibility purpose
									if (Selection.QUALIFIED != elementFormDefault || targetNamespace == null) {
										bufCollections.append("\n\t\t\t\t<step name=\"");
										bufCollections.append(attribute.getAlias());
										bufCollections.append("\"/>");
									} else {
										bufCollections.append("\n\t\t\t\t<step ns=\"");
										bufCollections.append(targetNamespace);
										bufCollections.append("\"  name=\"");
										bufCollections.append(attribute.getAlias());
										bufCollections.append("\"/>");
									}

								}

                                bufSchemas.append("\n\t\t\t\t\t</sequence>\n\t\t\t\t</complexType>\n\t\t\t</element>");

                                // only for compatibility purpose
								bufCollections.append("\n\t\t\t</step>");
								bufCollections.append("\n\t\t</collection>");

								// add variable and function to module
                                bufModules.append("\n\t\t\t<variable name=\"db:");
                                bufModules.append(collectionName);
                                bufModules.append("\">\n\t\t\t\t<xquerytype>element(");
                                if (targetNamespace != null)
                                    bufModules.append("db:");
                                bufModules.append(collectionName);
                                bufModules.append(")*</xquerytype>");
                                bufModules.append("\n\t\t\t</variable>");
                                bufModules.append("\n\t\t\t<function name=\"db:");
                                bufModules.append(collectionName);
                                bufModules.append("\">\n\t\t\t\t<xquerytype>element(");
                                if (targetNamespace != null)
                                    bufModules.append("db:");
                                bufModules.append(collectionName);
                                bufModules.append(")*</xquerytype>");
                                bufModules.append("\n\t\t\t</function>");
                            }
                        }
                        bufSchemas.append("\n\t\t</schema>");
                        bufModules.append("\n\t\t</module>");
                    }
                }
            }

        StringBuffer buf = new StringBuffer();
        buf.append("<?xml version='1.0'?>\n<xm:metadata\n\txmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"\n\txmlns:xm=\"http://www.xquark.org/XMLDBC/Metadata\"\n\tsource=\"");
        buf.append(name);
        buf.append("\">\n\t<schemas>");
        // add all others schemas
        if (_schemaList != null) {
            for (int i = 0; i < _schemaList.size(); i++) {
                bufSchemas.append("\n\t\t");
                try {
                    readSchemaFile((String) _schemaList.get(i), bufSchemas);
                } catch (Exception e) {
                    throw new XMLDBCException("Build metadata failed : " + e.getMessage(), e);
                }
                bufSchemas.append("\n");
            }
        }
        buf.append(bufSchemas.toString());
        buf.append("\n\t</schemas>");
        buf.append("\n\t<modules>\n");
        buf.append(bufModules.toString());
        buf.append("\n\t</modules>");
		buf.append("\n\t<collections>");
		buf.append(bufCollections.toString());
		buf.append("\n\t</collections>");
        buf.append("\n</xm:metadata>");

        //Trace.log(this, "buildMetaDataForParser(String name)", "metadata schema :");
        //Trace.log(this, "buildMetaDataForParser(String name)", buf.toString());
        //System.out.println(buf.toString());

        MetaDataImpl metadata = new MetaDataImpl(null);
        try {
            MetaParser metaparser = new MetaParser(metadata);
            MetaWrapper metawrapper = metaparser.createWrapper(new InputSource(new StringReader(buf.toString())));
            metadata.addMetaWrapper(metawrapper);
        } catch (MetadataException e) {
            throw new XMLDBCException("build metadata failed " + e.getMessage(), e);
        }

        if (!doModules) {
            //System.out.println(buf.toString());
            return metadata;
        }

        // add modules
        XQueryParser parser = null;
        for (int i = 0; i < _moduleList.size(); i++) {
            String moduleFileName = (String) _moduleList.get(i);
            try {
                bufModules.append("\n\t\t<module declared=\"true\">\n\t\t\t<declaration><![CDATA[");
                readFromFile(moduleFileName,bufModules);
                bufModules.append("]]></declaration>\n\t\t</module>");
            } catch (IOException e) {
                throw new XMLDBCException("Could not parse view definition", e);
            }
        }
        //		// create metadata string
        buf.setLength(0);
        buf.append("<?xml version='1.0'?>\n<xm:metadata\n\txmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"\n\txmlns:xm=\"http://www.xquark.org/XMLDBC/Metadata\"\n\tsource=\"");
        buf.append(name);
        buf.append("\">\n\t<schemas>");
        buf.append(bufSchemas.toString());
        buf.append("\n\t</schemas>");
        buf.append("\n\t<modules>\n");
        buf.append(bufModules.toString());
        buf.append("\n\t</modules>");
		buf.append("\n\t<collections>");
		buf.append(bufCollections.toString());
		buf.append("\n\t</collections>");
        buf.append("\n</xm:metadata>");

        //Trace.log(this, "buildMetaDataForParser(String name)", "metadata schema :");
        //Trace.log(this, "buildMetaDataForParser(String name)", buf.toString());
        //System.out.println(buf.toString());

        metadata = new MetaDataImpl(null);
        try {
            MetaParser metaparser = new MetaParser(metadata);
            MetaWrapper metawrapper = metaparser.createWrapper(new InputSource(new StringReader(buf.toString())));
            metadata.addMetaWrapper(metawrapper);
        } catch (MetadataException e) {
            throw new XMLDBCException("build metadata failed " + e.getMessage(), e);
        }

        return metadata;
    }

    public DBMSInfo getDbmsInfo() {
        return dbmsInfo;
    }

    private void readFromFile(String fileName, StringBuffer buf) throws IOException {
        FileReader reader = new FileReader(fileName);
        char[] charbuf = new char[1024];
        int len;
        while ((len = reader.read(charbuf)) != -1)
            buf.append(charbuf,0,len);
    }
    
    private void readSchemaFile(String filename, StringBuffer buf) throws Exception {
        StringWriter bufWriter = new StringWriter();
        XMLSerializer serializer = new XMLSerializer(bufWriter);
        serializer.setGenerateXMLDeclaration(false);
        ValidatingSchemaHandler vsh = new ValidatingSchemaHandler(new SchemaValidationContext());
        vsh.setContentHandler(serializer);
        vsh.setLexicalHandler(serializer);
        SAXParserFactory factory = SAXParserFactory.newInstance();
        SAXParser saxParser = factory.newSAXParser();
        XMLReader reader = saxParser.getXMLReader();
        reader.setContentHandler(vsh);
        reader.setErrorHandler(vsh);
        reader.parse(new InputSource(new FileReader(filename)));
        buf.append(bufWriter.toString());
    }

}