/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.extractor.runtime;

import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Properties;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.*;
import org.xml.sax.helpers.DefaultHandler;
import org.xquark.extractor.common.Constants;
import org.xquark.extractor.common.MessageLibrary;
import org.xquark.schema.SchemaManager;
import org.xquark.schema.validation.SchemaValidationContext;
import org.xquark.schema.validation.ValidatingSchemaFilter;
import org.xquark.xml.xdbc.XMLDBCException;

/**
 * SAX2 content handler implementation that performs configuration file loading.
 */
public class ExtractorConfiguration extends DefaultHandler implements Constants {
	
	private static final String RCSRevision = "$Revision: 1.3 $";
	private static final String RCSName = "$Name: XQ1_1_1 $";

	private static SAXParserFactory spf = SAXParserFactory.newInstance();
	static {
		spf.setNamespaceAware(true);
	}

	public static final String DATASOURCE_XSD = "/org/xquark/extractor/resources/Datasource.xsd";
	private static SchemaManager loadingManager = null;
	static {
		try {
			SchemaValidationContext context = new SchemaValidationContext();
			loadingManager = context.getSchemaManager();

			URL url = ExtractorConfiguration.class.getResource(DATASOURCE_XSD);
			InputSource source = new InputSource(url.toString());
			loadingManager.loadSchema(context, source);
		} catch (SAXException saxe) {
		}
	}

	/* copied for defaultAccessorhandler */
	private Properties xmlproperties;

	private Selection current_site;
	private Selection current_catalog;
	private Selection current_schema;
	private Selection current_table;
	private Selection current_column;
	private StringBuffer _currentChars = new StringBuffer();
	private String _dataSourceName = null;
	private Properties _substitutions = null;
	private Integer _nameCase = new Integer(CASE_MIXED);
	private byte includeType = Selection.NONE;
	private ArrayList schemas = null;
	private ArrayList modules = null;


	public ExtractorConfiguration() {
		xmlproperties = new Properties();
	}

	public static Properties loadConfiguration(String configURI) throws MalformedURLException, XMLDBCException {
		URL url = new URL(configURI);
		try {
			XMLReader reader = spf.newSAXParser().getXMLReader();
			ExtractorConfiguration handler = new ExtractorConfiguration();
			ValidatingSchemaFilter filter = new ValidatingSchemaFilter(reader, new SchemaValidationContext(loadingManager));
			filter.setContentHandler(handler);
			filter.setErrorHandler(handler);
			filter.parse(url.toString());
			return handler.getProperties();
		} catch (SAXParseException ex) {
			String msg = "Error in parsing file " + ex.getSystemId() + " at line " + Integer.toString(ex.getLineNumber()) +
						 " : " + ex.getMessage();
			throw new XMLDBCException(msg, ex);
		} catch (SAXException ex) {
			throw new XMLDBCException("Could not parse configuration file", ex);
		} catch (IOException ex) {
			throw new XMLDBCException("Could not read configuration file", ex);
		} catch (ParserConfigurationException ex) {
			throw new XMLDBCException("Could not allocate XML parser", ex);
		}
		
	}
	
	public static Properties getDefaultProperties(String url, String user) {
		Properties properties = new Properties();
		properties.put(LABEL_SPEC_SUBST_NAMECASE, new Integer(CASE_MIXED));
		properties.put(LABEL_SPEC_URL, url);
		properties.put(LABEL_SPEC_USER, user);
		Properties substitutions = new Properties();
		substitutions.setProperty("$", "_");
		substitutions.setProperty("#", "_");
		properties.put(LABEL_SPEC_SUBST, substitutions);
		return properties;
	}

	public Properties getProperties() {
		return xmlproperties;
	}

	public void startElement(String namespaceURI, String localName, String qualifiedName, Attributes atts) throws SAXException {
		_currentChars.setLength(0);
		if (localName.equalsIgnoreCase(LABEL_DATASOURCE)) {
			String name = atts.getValue(LABEL_SPEC_NAME);
			_dataSourceName = name;
			setProperty(LABEL_SPEC_NAME, name);
			current_site = new Selection(name, null, Selection.NONE, null);
			xmlproperties.put(LABEL_SPEC_SEL, current_site);
			includeType = Selection.INCLUDES;
		} else if (localName.equalsIgnoreCase(LABEL_SPEC_SUBST)) {
			_substitutions = new Properties();
			xmlproperties.put(LABEL_SPEC_SUBST, _substitutions);
		} else if (localName.equalsIgnoreCase(LABEL_SPEC_SUBST_CHAR)) {
			_substitutions.setProperty(atts.getValue(LABEL_SPEC_SUBST_VALUE), atts.getValue(LABEL_SPEC_SUBST_SUBST));
		} else if (localName.equalsIgnoreCase(LABEL_SPEC_CATALOG)) {
			String name = atts.getValue(LABEL_SPEC_NAME);
			current_catalog = new Selection(name, null, includeType, null);
			current_site.addToMap(name, current_catalog);
			includeType = Selection.INCLUDES;
		} else if (localName.equalsIgnoreCase(LABEL_SPEC_SCHEMA)) {
			String name = atts.getValue(LABEL_SPEC_NAME);
			String targetNamespace = atts.getValue(LABEL_SPEC_TARGETNAMESPACE);
			current_schema = new Selection(name, targetNamespace, includeType, null);
			String efd = atts.getValue(LABEL_SPEC_ELEMENTFORMDEFAULT);
			int elementFormDefault;
			if (LABEL_SPEC_QUALIFIED.equalsIgnoreCase(efd)) {
				elementFormDefault = Selection.QUALIFIED;
			} else {
				elementFormDefault = Selection.UNQUALIFIED;
			}
			current_schema.setElementFormDefault(elementFormDefault);

			current_catalog.addToMap(name, current_schema);
		} else if (localName.equalsIgnoreCase(LABEL_SPEC_TABLE)) {
			String name = atts.getValue(LABEL_SPEC_NAME);
			if (null != name) {
				current_table = new Selection(name, null, includeType, null);
				String alias = atts.getValue(LABEL_SPEC_ALIAS);
				if (null != alias && !Selection.verifyName(alias)) {
					throw new SAXException(MessageLibrary.getMessage("C_INV_NCNAME", alias));
				}
				current_table.setAlias(alias);
				current_schema.addToMap(name, current_table);
			} else {
				name = atts.getValue(LABEL_SPEC_REGEX);
				if (null != name) {
					current_table = new Selection(name, true, null, includeType, null);
					current_schema.addToMap(name, current_table);
				} else {
					//                    throw new Exception("invalid table element in configuration file.");
				}
			}
		} else if (localName.equalsIgnoreCase(LABEL_SPEC_COLUMN)) {
			String name = atts.getValue(LABEL_SPEC_NAME);

			if (null != name) {
				current_column = new Selection(name, null, includeType, null);
				String alias = atts.getValue(LABEL_SPEC_ALIAS);
				if (null != alias && !Selection.verifyName(alias)) {
					throw new SAXException(MessageLibrary.getMessage("C_INV_NCNAME", alias));
				}
				current_column.setAlias(alias);
				current_table.addToMap(name, current_column);
			} else {
				name = atts.getValue(LABEL_SPEC_REGEX);
				if (null != name) {
					current_column = new Selection(name, true, null, includeType, null);
					current_table.addToMap(name, current_table);
				} else {
					//                    throw new Exception("invalid table element in configuration file.");
				}
			}

			String isPK = atts.getValue(LABEL_SPEC_PK_COLUMN);

			if (null != isPK && isPK.equalsIgnoreCase("true")) {
				current_column.setPKColumn(true);
			}

		} else if (localName.equalsIgnoreCase(LABEL_SPEC_INCLUDES)) {
			includeType = Selection.INCLUDES;
		} else if (localName.equalsIgnoreCase(LABEL_SPEC_EXCLUDES)) {
			includeType = Selection.EXCLUDES;
						
		} else if (localName.equalsIgnoreCase(LABEL_SPEC_IMPORT_SCHEMA)) {
			String name = atts.getValue(LABEL_SPEC_FILENAME);
			if (schemas == null) {
			    schemas = new ArrayList();
				xmlproperties.put(LABEL_SPEC_SCHEMAS,schemas);
			}
			schemas.add(name);
		} else if (localName.equalsIgnoreCase(LABEL_SPEC_IMPORT_MODULE)) {
			String name = atts.getValue(LABEL_SPEC_FILENAME);
			if (modules == null) {
			    modules = new ArrayList();
				xmlproperties.put(LABEL_SPEC_MODULES,modules);
			}
			modules.add(name);
		}

	}

	public void characters(char ch[], int start, int length) throws SAXException {
		_currentChars.append(ch, start, length);
	}

	public void endElement(String namespaceURI, String localName, String qualifiedName) throws SAXException {
		if (localName.equalsIgnoreCase(LABEL_SPEC_DESCRIPTION)) {
			setProperty(LABEL_SPEC_DESCRIPTION, _currentChars.toString().trim());
		} else if (localName.equalsIgnoreCase(LABEL_SPEC_CASE)) {
			setProperty(LABEL_SPEC_CASE, _currentChars.toString().trim());
		} else if (localName.equalsIgnoreCase(LABEL_SPEC_SUBST_NAMECASE)) {
			String cur_car = _currentChars.toString().trim();
			if (cur_car.equalsIgnoreCase(CASE_LOWER_TOKEN)) {
				_nameCase = new Integer(CASE_LOWER);
			} else if (cur_car.equalsIgnoreCase(CASE_UPPER_TOKEN)) {
				_nameCase = new Integer(CASE_UPPER);
			} else if (cur_car.equalsIgnoreCase(CASE_MIXED_TOKEN)) {
				_nameCase = new Integer(CASE_MIXED);
			}
			xmlproperties.put(LABEL_SPEC_SUBST_NAMECASE, _nameCase);
		} else if (localName.equalsIgnoreCase(LABEL_SPEC_DRIVER)) {
			setProperty(LABEL_SPEC_DRIVER, _currentChars.toString().trim());
		} else if (localName.equalsIgnoreCase(LABEL_SPEC_URL)) {
			setProperty(LABEL_SPEC_URL, _currentChars.toString().trim());
		} else if (localName.equalsIgnoreCase(LABEL_SPEC_USER)) {
			setProperty(LABEL_SPEC_USER, _currentChars.toString().trim());
		} else if (localName.equalsIgnoreCase(LABEL_SPEC_PASSWORD)) {
			setProperty(LABEL_SPEC_PASSWORD, _currentChars.toString().trim());
		} else if (localName.equalsIgnoreCase(LABEL_SPEC_JNDI)) {
			setProperty(LABEL_SPEC_JNDI, _currentChars.toString().trim());
		} 
		_currentChars.setLength(0);
	}

	/**
	 * Set a property of the accessor.
	 *
	 * @param a property of the accessor.
	 */
	protected void setProperty(String key, String value) {
		if (key == null) {
			System.err.println("Key " + key + " is null.");
			return;
		}
		if (value == null) {
			System.err.println("Value " + value + " of " + key + " is null.");
			return;
		}
		xmlproperties.setProperty(key, value);
	}

	/**
	 * Return  a property of the accessor.
	 *
	 * @return a property of the accessor.
	 */
	protected String getProperty(String key) {
		return (String) xmlproperties.getProperty(key);
	}

	public void error(SAXParseException spe) throws SAXException {
		throw spe;
	}
}
