/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.extractor.runtime;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.xquark.extractor.ExtractorConnection;
import org.xquark.extractor.common.MessageLibrary;
import org.xquark.extractor.metadata.MetaDataManager;
import org.xquark.extractor.sql.StatementInfo;
import org.xquark.xml.xdbc.*;
import org.xquark.xquery.parser.XQueryException;
import org.xquark.xquery.parser.XQueryModule;
import org.xquark.xquery.xdbc.XDBCResultSetInterface;
import org.xquark.xquery.xdbc.XResultSetImpl;

public class ExtractorStatement implements XMLStatementInternal {
	final static String KEY_NAME = "key_name";

	private static final String RCSRevision = "$Revision: 1.9 $";
	private static final String RCSName = "$Name: XQ1_1_1 $";

	private static Log log = LogFactory.getLog(ExtractorStatement.class);

	protected ExtractorConnection _connection;
	protected MetaDataManager _metadataManager = null;
	protected HashMap selections = null;
	protected XMLResultSet _currentResultSet = null;
    protected Statement _stmt = null;
	protected boolean _hasResultSet = false;
	protected boolean _isClosed = false;
	protected int _id;
	protected QueryFactory _queryFactory;

	public ExtractorStatement(ExtractorConnection connection,
	        int stmtId,
	        QueryFactory queryFactory,
	        MetaDataManager metadataManager
	        )
	throws XMLDBCException {
		_connection = connection;
		_metadataManager = metadataManager;
		_id = stmtId;
		_queryFactory = queryFactory;
		_queryFactory.setStatement(this);
	}

	/**
	 * Access method for the _connection property.
	 * @return   the current value of the _connection property
	 */
	public XMLConnection getConnection() {
		return _connection;
	}

	public Statement createJdbcStatement() throws SQLException {
		Statement retVal = null;
		if (_connection.getExtractor().useScrollableResultSets())
			retVal = _connection.getJdbcConnection().createStatement(ResultSet.TYPE_SCROLL_INSENSITIVE, ResultSet.CONCUR_READ_ONLY);
		else
			retVal = _connection.getJdbcConnection().createStatement(ResultSet.TYPE_FORWARD_ONLY, ResultSet.CONCUR_READ_ONLY);

		return retVal;
	}

    /*********************************************************************/
    /*                       XDBC API IMPLEMENTATION                     */
    /*********************************************************************/
	
	public boolean execute(String query) throws XMLDBCException {
		if (_currentResultSet != null)
			_currentResultSet.close();
		_currentResultSet = executeXQuery(query);
		return true;
	}

	public boolean execute(String query, int queryType) throws XMLDBCException {
		return execute(query);
	}

	public XMLResultSet executeQuery(String query) throws XMLDBCException {
		execute(query);
		return _currentResultSet;
	}

	public XMLResultSet executeQuery(String query, int queryType) throws XMLDBCException {
		return executeQuery(query);
	}

	public XMLDocumentSet executeDocumentQuery(String query) throws XMLDBCException {
		throw new XMLDBCNotSupportedException("executeDocumentQuery not yet supported");
	}

	// Utility for LS
    public XQueryModule parseXQuery(String xquery) throws XMLDBCException {
        return _queryFactory.parseXQuery(xquery);
    }
    
	public XMLResultSet getResultSet() throws XMLDBCException {
		return _currentResultSet;
	}

	public XMLDocumentSet getDocumentSet() throws XMLDBCException {
		throw new XMLDBCNotSupportedException("executeDocumentQuery not yet supported");
	}

	public void setBaseURI(String baseURI) { _queryFactory.setBaseURI(baseURI);}

	public String getBaseURI() { return _queryFactory.getBaseURI();}
	
	public int getStatementId() { return _id;}

	public void close() throws XMLDBCException {
		if (!_isClosed) {
			_metadataManager = null;
			selections = null;
			if (null != _currentResultSet) {
				_currentResultSet.close();
				_currentResultSet = null;
			}
			if (_stmt != null) {
                try {
                    _stmt.close();
                } catch (SQLException e) { /* no op */ }
			}
			_connection.statementClosed(this);
			_isClosed = true;
		}
	}

	public boolean isClosed() throws XMLDBCException {
		return _isClosed;
	}

    /*********************************************************************/
    /*                       INTERNAL QUERY EXECUTION                    */
    /*********************************************************************/
	
	/**
	 * Execute the Xquery request and return an XML document as result.
	 *
	 * @param xquery the query to execute.
	 * @return the XML result document.
	 */
	protected XMLResultSet executeXQuery(String xquery) throws XMLDBCException {
    	if (isClosed())
    		throw new XMLDBCException(MessageLibrary.getMessage("ST_CLOSED"));
    
		return runQuery(_queryFactory.createCompiledQuery(xquery).getCompiledQuery());
	}
	
	protected XMLResultSet runQuery(List compiledQuery) throws XMLDBCException {

		XMLResultSet retVal = null;

		if (compiledQuery.size() == 1)
			retVal = runQuery((Query.CompiledExpression) compiledQuery.get(0));
		else if (compiledQuery.size()> 1)
			retVal = new XResultSetUnion(compiledQuery, this);

		return retVal;
	}

	public XMLResultSet runQuery(Query.CompiledExpression cExpr) throws XMLDBCException {
	    try {
            return new XResultSetImpl(
                    cExpr.sqlInfoTree == null ? null : runQueryTree(cExpr.sqlInfoTree),
                    cExpr.expr.getQMEM(),
                    cExpr.expr.getIdVarList(),
                    this
                    );
        }
	    catch (XQueryException e) {
            throw new XMLDBCException(e.getMessage(), e);
        }
	}
	
	private XDBCResultSetInterface runQueryTree(Query.SQLExecutionInfo jdbcInfo)
	throws XMLDBCException, XQueryException {
	    XDBCResultSetInterface rs = null;
	    
        List childResultSetList = new ArrayList();
        if (jdbcInfo.childrenQueries != null)
	        for (int i = 0; i < jdbcInfo.childrenQueries.size(); i++) {
	            childResultSetList.add(runQueryTree((Query.SQLExecutionInfo) jdbcInfo.childrenQueries.get(i)));
	        }
        
	    if (jdbcInfo.isUnion())
	        rs = new UnionResultSet(childResultSetList);
	    else {
            rs = executeSQL(jdbcInfo);
	        SynchronizedResultSet srs = (SynchronizedResultSet) rs;
	         srs.setSlaveList(childResultSetList);
	        if (jdbcInfo.idCount >= 0)
	            srs.setReconstructionIdCount(jdbcInfo.idCount);
	        
	        if (jdbcInfo.idPosition != null)
	            srs.setIdPosition(jdbcInfo.idPosition);
	    }
	    return rs;
	}

	protected SynchronizedResultSet executeSQL(Query.SQLExecutionInfo sqlInfo)
	throws XQueryException, XMLDBCException {

	    List requestList = sqlInfo.sql.getRequestList();
	    
		if (log.isDebugEnabled()) {
			log.debug("=================< SQL >================="); 
			for (int i = 0; i < requestList.size(); i++)
				log.debug(((StatementInfo)requestList.get(i)).sStmt);
			log.debug("=================< MAPPER >================="); 
			log.debug(sqlInfo.mapper.pprint());
		}
		
	    // create a statement for every Qdb (one current ResultSet per statement in JDBC) 
	    SynchronizedResultSet retVal = null;
		try {
			retVal = new SynchronizedResultSet(
			        				requestList,
			        				sqlInfo.mapper,
			        				createJdbcStatement(),
			        				_metadataManager.getDbmsInfo()
			        				);
		}
		catch (XMLDBCException ex) {
		    throw ex;
		}
		catch (SQLException ex) {
		    throw new XMLDBCException(ex.getMessage(), ex);
		}
		finally {
			try {
			    if (_stmt == null)
			        _stmt = createJdbcStatement();
				sqlInfo.sql.cleanAll(_stmt);
			}
			catch (SQLException ex) {}
		}

		return retVal;
	}

	protected void finalize() throws Exception {
		close();
	}
}
