/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.extractor.runtime;

import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.util.HashMap;
import java.util.Map;

import org.xquark.extractor.algebra.Mapper;
import org.xquark.extractor.common.Debug;
import org.xquark.extractor.common.MessageLibrary;
import org.xquark.extractor.metadata.ExtractorMappingInfo;
import org.xquark.jdbc.typing.DBMSInfo;
import org.xquark.schema.validation.ValidationContextProvider;
import org.xquark.xquery.parser.XQueryException;
import org.xquark.xquery.typing.QAtomicSerializer;

/**
 * A look ahead JDBC result set implementation used by {@link org.xquark.extractor.runtime.SynchronizedResultSet}.
 * The look ahead is needed to be able to detect changes of skolem IDs.
 */
class LookAheadJdbcResultSet {
	private static final String RCSRevision = "$Revision: 1.4 $";
	private static final String RCSName = "$Name: XQ1_1_1 $";

	private static final int BOR = 0; /* begin of resultset*/
	private static final int BUFFER = 1;
	private static final int RESULTSET = 2;

	private ResultSet _resultset;
	/** Buffer for an entire row **/
	private String[] _buffer;
	private Map _attributeMap = new HashMap();
	private ColumnInfo[] _attributes; // because some drivers (SQLServer) don't like fetching out of the regular order....

	private boolean _bufferNotEmpty;
	private boolean _resultSetNotEmpty;
	private int _cursor;
	static private ValidationContextProvider _validationContext = new ValidationContextProvider() {
		public String getNamespaceURI(String prefix) {
			return null;
		}
		public String getPrefix(String uri) {
			return null;
		}
		public String getDocumentBase() {
			return null;
		}
		public Map getNotationDeclarations() {
			return null;
		}
	};
	private QAtomicSerializer qSerializer = new QAtomicSerializer(_validationContext);

	public LookAheadJdbcResultSet(ResultSet resultset, Mapper mapper, DBMSInfo info) throws SQLException, XQueryException {
		_resultset = resultset;

		ResultSetMetaData rsm = _resultset.getMetaData();

		int columnCount = rsm.getColumnCount();
		_buffer = new String[columnCount];
		_attributes = new ColumnInfo[columnCount];
		String columnName = null;
		ColumnInfo ci = null;
		for (int i = 0; i < columnCount; i++) {
			columnName = rsm.getColumnName(i + 1);
			ci = new ColumnInfo(i + 1, mapper.findMapItem(columnName).getMappingInfo(info.getTypeMap()));
			_attributeMap.put(columnName, ci);
			_attributes[i] = ci;
		}

		_bufferNotEmpty = _resultset.next();
		if (_bufferNotEmpty) {
			refreshBuffer();
			_resultSetNotEmpty = _resultset.next();
		}
		_cursor = BOR;
	}

	public boolean next() throws SQLException, XQueryException {
		boolean retVal;
		if (BOR == _cursor) {
			_cursor = BUFFER;
		} else {
			Debug.assertTrue(BUFFER == _cursor, "BUFFER == _cursor");
			if (_resultSetNotEmpty) {
				refreshBuffer();
				_resultSetNotEmpty = _resultset.next();
			} else {
				_resultset.close();
				_resultset = null;
				_bufferNotEmpty = false;
			}
		}
		return _bufferNotEmpty;
	}

	public boolean hasNext() {
		boolean retVal = false;

		if (BUFFER == _cursor) {
			retVal = _resultSetNotEmpty;
		} else if (BOR == _cursor) {
			retVal = _bufferNotEmpty;
		} else {
			Debug.assertTrue(false, MessageLibrary.getMessage("IE_ERR"));
		}

		return retVal;
	}

	public boolean lookAhead() {
		boolean retVal = false;
		switch (_cursor) {
			case BOR :
				retVal = _bufferNotEmpty;
				if (retVal) {
					_cursor = BUFFER;
				}
				break;
			case BUFFER :
				retVal = _resultSetNotEmpty;
				if (retVal) {
					_cursor = RESULTSET;
				}
				break;
			default :
				org.xquark.extractor.common.Debug.assertTrue(false, "Logic error");
		}

		return retVal;
	}

	public boolean goBack() {
		boolean retVal = false;
		switch (_cursor) {
			case BUFFER :
				_cursor = BOR;
				retVal = true;
				break;
			case RESULTSET :
				_cursor = BUFFER;
				retVal = true;
				break;
			case BOR :
				retVal = false;
				break;
			default :
				org.xquark.extractor.common.Debug.assertTrue(false, "Logic error");
		}

		return retVal;
	}

	public String getString(String name) throws SQLException {
		String retVal = null;
		ColumnInfo columnInfo = (ColumnInfo) _attributeMap.get(name);
		if (BUFFER == _cursor) {
			retVal = getStringFromBuffer(columnInfo.getIndex());
		} else {
			try {
				ExtractorMappingInfo mappingInfo = columnInfo.getMappingInfo();
				if (mappingInfo == null) // pseudo-id
					retVal = _resultset.getString(columnInfo.getIndex());
				else
					retVal = mappingInfo.getParameter(_resultset, columnInfo.getIndex(), qSerializer, _validationContext);
			} catch (XQueryException e) {
			}
		}
		return retVal;
	}

	private void refreshBuffer() throws SQLException, XQueryException {
		ExtractorMappingInfo mappingInfo = null;

		for (int i = 0; i < _attributes.length; i++) {
			mappingInfo = _attributes[i].getMappingInfo();
			if (mappingInfo == null) { // pseudo-id
				_buffer[i] = _resultset.getString(_attributes[i].getIndex());
			} else {
				_buffer[i] = mappingInfo.getParameter(_resultset, _attributes[i].getIndex(), qSerializer, _validationContext);
			}
		}
	}

	private String getStringFromBuffer(int index) {
		String retVal = null;
		if (index > 0)
			retVal = _buffer[index - 1];
		return retVal;
	}

	/** @todo this function is to be removed after debug  */
	public ResultSet getSqlResultSet() {
		return _resultset;
	}

	public void close() {
		if (_resultset != null) {
			try {
				_resultset.close();
				_resultset = null;
			} catch (SQLException ex) {
			}
		}
	}

	private class ColumnInfo {
		private int index = 0;
		private ExtractorMappingInfo mappingInfo = null;

		public ColumnInfo(int index, ExtractorMappingInfo mappingInfo) {
			super();
			this.index = index;
			this.mappingInfo = mappingInfo;
		}
		public int getIndex() {
			return index;
		}

		public ExtractorMappingInfo getMappingInfo() {
			return mappingInfo;
		}

	}
}
