/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 XQuark Group.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.extractor.runtime;

import java.io.StringReader;
import java.util.ArrayList;
import java.util.List;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.xquark.extractor.algebra.*;
import org.xquark.extractor.common.*;
import org.xquark.extractor.metadata.MetaDataManager;
import org.xquark.extractor.runtime.Query.SQLExecutionInfo;
import org.xquark.extractor.sql.AbstractGenSqlVisitor;
import org.xquark.extractor.sql.SqlExpression;
import org.xquark.xml.xdbc.XMLDBCException;
import org.xquark.xml.xdbc.XMLDBCNotSupportedException;
import org.xquark.xquery.metadata.VarCounter;
import org.xquark.xquery.metadata.resolver.MetadataAccess;
import org.xquark.xquery.normalize.DependNode;
import org.xquark.xquery.normalize.TransformXQueryExpression;
import org.xquark.xquery.parser.*;
import org.xquark.xquery.typing.*;

/**
 * Produces DBMS-specific queries and embed specific compilation algorithm.
 */
public abstract class QueryFactory {

	private static Log log = LogFactory.getLog(QueryFactory.class);
	
	protected XQueryParser _parser = null;
	protected GenAlgebraVisitor _genAlgebraVisitor = null;
	protected AlgebraTypeVisitor _algebraTypeVisitor = null;
	protected RemoveIntermediateProjectVisitor _removeProjectVisitor = null;
	protected BuildReferredAttributes _brav = null;
	protected AnalyzePredicateVisitor _apv = null;
	protected AbstractGenSqlVisitor _genSqlVisitor = null;
	protected String _baseURI = null;
	protected MetaDataManager _metadataManager = null;
	protected MetadataAccess _metadata = null;
	protected ExtractorStatement _stmt = null;

	/* counters : only one per kind because statement is mono-threaded */
	protected AliasIDGenerator _attIDProvider = new AliasIDGenerator();
	protected AliasIDGenerator _relIDProvider = new AliasIDGenerator();

	public QueryFactory(MetaDataManager metadataManager, MetadataAccess metadata) throws XMLDBCException {
		try {
			_metadataManager = metadataManager;
			_metadata = metadata;
			createParser();
			createVisitors();
		} catch (ParseException ex) {
			throw new XMLDBCException("Can not create statement", ex);
		} catch (TypeException ex) {
			throw new XMLDBCException("Can not create statement", ex);
		}
	}

//	public Query createQuery(String query) throws XMLDBCException {
//		return new Query(query, this);
//	}

	public Query createCompiledQuery(String query) throws XMLDBCException {
		XQueryModule module = parseXQuery(query);
		return new Query(query, this, compileQuery(module), module);
	}

	public void setStatement(ExtractorStatement stmt) {
		_stmt = stmt;
	}

	public void setBaseURI(String baseURI) {
		_baseURI = baseURI;
	}

	public String getBaseURI() {
		return _baseURI;
	}

	protected void createParser() throws ParseException, TypeException {
		_parser = new XQueryParser(new StringReader("dummy"));
		_parser.setFactory("org.xquark.xquery.parser.ParserFactory");
	}

	abstract protected void createVisitors();

	public List compileQuery(XQueryModule module) throws XMLDBCException {
		List exprList = module.getExpressions();
		List cExpressions = new ArrayList();

		if (exprList != null)
			for (int i = 0; i < exprList.size(); i++) {
				cExpressions.add(compileQuery((XQueryExpression) exprList.get(i)));
			}
		return cExpressions;
	}

	protected Query.CompiledExpression compileQuery(XQueryExpression xExpr) throws XMLDBCException {
		Query.CompiledExpression retVal = null;

		try {
			/* transforms XQueryExpression */

			TransformXQueryExpression transformExpr = xExpr.getParentModule().canonize(xExpr, _parser.getTypeVisitor());

			// NOTES:
			// 1. normalization (no typing considered) AW,
			// 2. typing (for the case it is lost...),
			// 3. restructuration (=normalization considering types) LS,
			// 4. canonization (separates Qmem & Qdb) LS,
			// 5. typing Qdb (will be removed by LS ?),
			// 6. get Bound vars on QMem,
			// 7. patches on Qdb for specific needs of end modules (no agreement could be found).

			// Outcoming Qdbs: Flat, without tags but with let (correlation).
			// Multiple unrelated queries are not really processed (1 expression for 1 unit for 1 file).
			// Anyway, has changed in the new version of specs.

			ArrayList idVarList = transformExpr.getIdVarList();
			List Qdb = transformExpr.getQDB();

			if (log.isDebugEnabled()) {
				log.debug("=================< QMEM >================="); 
				log.debug(transformExpr.getQMEM().toString(null, true, false, true));
				log.debug("=================< QDBs >================="); 
				for (int i = 0; i < Qdb.size(); i++)
					log.debug(((Variable) Qdb.get(i)).getExpression().toString(null, true, false, true));
			}

			Query.SQLExecutionInfo jdbcExecInfo = null;
			ArrayList root = transformExpr.getDependMap();

			if (Qdb.size() > 1) {
				if (root != null) { // Multiple expressions imbricated in return with dependant variables.
					if (root.size() == 1)
						jdbcExecInfo = compileSOU((DependNode) root.get(0), idVarList);
					else {
						// Count the number of id variables for Qdb
						// Assuming id vars are sorted in the Qdb order...
						// TODO: put it in the front-end.
						FLWRExpression xQuery = null;
						ArrayList idVarLists = new ArrayList(root.size());

						int rootIndex = -1;
						int idIndex = 0;
						ArrayList tmpList = new ArrayList();
						for (int i = 0; i < Qdb.size() || idIndex < idVarList.size(); i++) {
							xQuery = (FLWRExpression) ((Variable) Qdb.get(i)).getExpression();
							if (rootIndex + 1 < root.size() && ((DependNode) root.get(rootIndex + 1)).getExpression() == xQuery) {
								rootIndex++;
								tmpList = new ArrayList();
								idVarLists.add(tmpList);
							}
							while (idIndex < idVarList.size() && xQuery.getVariables().contains(idVarList.get(idIndex)) && !tmpList.contains(idVarList.get(idIndex))) {
								tmpList.add(idVarList.get(idIndex));
								idIndex++;
							}
						}

						jdbcExecInfo = new Query.SQLExecutionInfo();
						Query.SQLExecutionInfo subQuery = null;
						for (int m = 0; m < root.size(); m++) {
							xQuery = (FLWRExpression) ((Variable) Qdb.get(m)).getExpression();
							subQuery = compileSOU((DependNode) root.get(m), (ArrayList) idVarLists.get(m));
							subQuery.idCount = ((ArrayList) idVarLists.get(m)).size();
							jdbcExecInfo.addChild(subQuery);
						}
					}
				} else { // For instance, 2 FLWORs side-by-side in an element constructor.

					// Count the number of id variables for Qdb
					// Assuming id vars are sorted in the Qdb order...
					// TODO: put it in the front-end.
					FLWRExpression xQuery = null;
					int[] idCounts = new int[Qdb.size()];
					int idCount = idVarList.size();
					int wQdb = 0;

					for (int i = 0; i < idCount; i++) {
						xQuery = (FLWRExpression) ((Variable) Qdb.get(wQdb)).getExpression();
						if (xQuery.getVariables().contains(idVarList.get(i)))
							idCounts[wQdb]++;
						else {
							wQdb++;
							i--;
						}
					}

					jdbcExecInfo = new Query.SQLExecutionInfo();
					Query.SQLExecutionInfo subQuery = null;
					for (int m = 0; m < Qdb.size(); m++) {
						xQuery = (FLWRExpression) ((Variable) Qdb.get(m)).getExpression();
						subQuery = compileSingleQDB(xQuery, true);
						subQuery.idCount = idCounts[m];
						jdbcExecInfo.addChild(subQuery);
					}
				}
			} else if (Qdb.size() == 1) { // Base single query : expression with potential let (correlation).
				jdbcExecInfo = compileSingleQDB(((Variable) Qdb.get(0)).getExpression(), false);
			}
			retVal = new Query.CompiledExpression(jdbcExecInfo, transformExpr);
		} catch (NotSupport ex) {
			throw new XMLDBCNotSupportedException(ex.getMessage(), ex);
		} catch (SqlWrapperException ex) {
			throw new XMLDBCException(ex.getMessage(), ex);
		} catch (AssertionException ex) {
			throw new XMLDBCException(MessageLibrary.getMessage("G_INT_LOG_ERR", ex.getMessage()), ex);
		} catch (ParseException ex) {
			throw new XMLDBCException(ex.getMessage(), ex);
		} catch (TypeException ex) {
			throw new XMLDBCException(ex.getMessage(), ex);
		} catch (XQueryException ex) {
			throw new XMLDBCException(ex.getMessage(), ex);
		}
		return retVal;
	}

	/**
	 * Performs recursively Qdb (adding the suitable sorts for SOU) following the
	 * dependency tree and build an homothetic result set tree with columns making
	 * row ids positionned in order for reconstruction to synchronize the SOU
	 * execution.
	 * @param node
	 * @param idVarList
	 * @return
	 * @throws Exception
	 */
	protected Query.SQLExecutionInfo compileSOU(DependNode node, List idVarList) throws XQueryException, XMLDBCException {

		FLWRExpression xQuery = node.getExpression();

		/* rebuild orderby list */
		ArrayList newSortExprList = new ArrayList();
		List varLevels = node.varLevels();
		List varsOfCurrentLevel = null;
		List sortExprLevels = node.sortExprLevels();
		List sortExprsOfCurrentLevel = null;
		int levelNumber = varLevels.size();
		for (int i = 0; i < levelNumber; i++) {
			sortExprsOfCurrentLevel = (List) sortExprLevels.get(i);
			if (sortExprsOfCurrentLevel != null)
				newSortExprList.addAll(sortExprsOfCurrentLevel);
			varsOfCurrentLevel = (List) varLevels.get(i);
			newSortExprList.add(new PlaceHolder(varsOfCurrentLevel));
		}
		xQuery.setOrderBy(newSortExprList);

		Query.SQLExecutionInfo jdbcExecInfo = compileSingleQDB(xQuery, true);

		/* calculate the position put the ids */
		jdbcExecInfo.idPosition = calculateIdPosition(node.getExpression().getVariables(), idVarList);

		/* Recursive call for dependent queries */
		List childQueryList = node.getSons();
		if (null != childQueryList && !childQueryList.isEmpty()) {
			for (int i = 0; i < childQueryList.size(); i++) {
				jdbcExecInfo.addChild(compileSOU((DependNode) childQueryList.get(i), idVarList));
			}
		}

		return jdbcExecInfo;
	}

	protected Query.SQLExecutionInfo compileSingleQDB(XQueryExpression xQuery, boolean keyNeeded) throws XQueryException, XMLDBCException {

		Expression algebraTree = null;

		/* Algebra tree building */
		_genAlgebraVisitor.reinit();

		/* Reset counters */
		_attIDProvider.reset();
		_relIDProvider.reset();

		if (keyNeeded) { // Used when Sorted Outer Union is needed to identify rows
			List souArgList = new ArrayList();
			souArgList.add(xQuery);
			_genAlgebraVisitor.noticSortedOuterUnion(souArgList);
		}
		xQuery.accept(_genAlgebraVisitor);
		algebraTree = _genAlgebraVisitor.getAlgebraTree();

		// INFO : YS types of visitors : GenAlgebraVisitor (a ParserVisitor impl),
		// AlgebraVisitor, GenSQLVisitor
		_removeProjectVisitor.reinit();
		algebraTree.accept(_removeProjectVisitor);

		algebraTree.accept(_brav);

		/* resolve invalid predicates */
		_apv.reinit();
		algebraTree.accept(_apv);

		_algebraTypeVisitor.reinit();
		algebraTree.accept(_algebraTypeVisitor);

		_genSqlVisitor.reinit();
		SqlExpression sql = algebraTree.accept(_genSqlVisitor);

		/* Generate string to be ready for prepared statements execution */
		sql.generateRequestList();
		return new SQLExecutionInfo(sql, algebraTree.getMapper());
	}

	public XQueryModule parseXQuery(String xquery) throws XMLDBCException {

		_parser.ReInit(new StringReader(xquery));
		_parser.setBaseURI(_baseURI);

		XQueryModule result = null;
		try {
			// parsing request
			result = _parser.Start(_metadata, _metadata.getSchemaManager(), _metadata.getModuleManager(), new VarCounter());
			result.normalize(_parser.getTypeVisitor());
		} catch (ParseException ex) {
			throw new XMLDBCException(ex.getMessage(), ex);
		} catch (XQueryException ex) {
			throw new XMLDBCException(ex.getMessage(), ex);
		}
		//_parser.ReInit(new StringReader(xquery));

		return result;
	}

	/**
	 * Retrieves id position in result set by var name.
	 * @param varList
	 * @param idVarList
	 * @return
	 */
	private int[] calculateIdPosition(List varList, List idVarList) {
		int[] retVal = new int[idVarList.size()];
		int idNum = -1;
		int idPosition = 0;
		int idVarIndex = -1;

		Object var = null;
		for (int i = 0; i < varList.size(); i++) {
			var = varList.get(i);
			idPosition = (int) idVarList.indexOf(var);
			if (-1 != idPosition) {
				/* there is a identifier on this */
				idNum++;
				retVal[idNum] = idPosition;
			}
		}

		return retVal;
	}

	protected class AliasIDGenerator implements IDProvider {
		private int counter = 0;

		public String getID() {
			return String.valueOf(counter++);
		}

		public void reset() {
			counter = 0;
		}
	}

	protected class TempTableIDGenerator implements IDProvider {
		private int counter = 0;

		public TempTableIDGenerator() {
		}

		public String getID() {
			return String.valueOf(_stmt.getStatementId()) + '_' + String.valueOf(counter++);
		}

		public void reset() {
			counter = 0;
		}
	}
}
