/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.extractor.sql;

import java.sql.Types;
import java.util.*;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.xquark.extractor.algebra.*;
import org.xquark.extractor.common.Debug;
import org.xquark.extractor.common.MessageLibrary;
import org.xquark.extractor.common.SqlWrapperException;
import org.xquark.extractor.metadata.MetaDataManager;
import org.xquark.extractor.runtime.IDProvider;
import org.xquark.extractor.xfunctions.*;
import org.xquark.jdbc.typing.DbType;

public class AbstractGenSqlVisitor  extends DefaultGenSqlVisitor  implements Constants
{
	private static final String RCSRevision = "$Revision: 1.19 $";
	private static final String RCSName     = "$Name: XQ1_1_1 $";
	
	private static Log log = LogFactory.getLog(AbstractGenSqlVisitor.class);

	public static final String TMP_TABLE_PREFIX = "XTMP";

	protected SqlFactory _factory = null;

	protected MetaDataManager _metadataManager= null;
    protected IDProvider _relIDProvider = null;

	// QUESTION: what's the use of this. peek() is never used and stack is only used by Sybase
	// to know if there are nested selects... A counter would be enough eventually ?
	protected Stack _nestStack = new Stack();

	protected AbstractGenSqlVisitor(SqlFactory aSqlFactory, IDProvider relIDProvider)
	{
		setSqlFactory(aSqlFactory);
		_relIDProvider = relIDProvider;
	}

	/**
	 * Sets the value of the _sqlFactory property.
	 * @param aSqlFactory the new value of the _sqlFactory property
	 */
	public void setSqlFactory(SqlFactory aSqlFactory)
	{
	   Debug.assertTrue(null != aSqlFactory,"null != aSqlFactory");
		_factory = aSqlFactory;
	}

	public void setMetadataManager(MetaDataManager metadataManager) {
		_metadataManager = metadataManager;
	}

	public void reinit ()
	{
		_nestStack.clear();
	}

	public SqlExpression visit(Expression arg)
	{
		throw new SqlWrapperException(MessageLibrary.getMessage("T_N_SUP_EXPR",arg.pprint()),null);
	}

	public SqlExpression visit(AttributeExpression arg)
	{
		String tableInstanceName = null;
		RenameRelation  rr = arg.getTableInstance();
		if (rr != null)
			tableInstanceName = rr.getUniqueName();

		return _factory.createAttributeExpression (tableInstanceName, arg.getName());
	}

	public SqlExpression visit(BinOpArithmetic arg)
	{
		return _factory.createBinOpArithmetic(arg.getOperator(),
		        arg.getLeftOperand().accept(this),
		        arg.getRightOperand().accept(this)
		        );
	}

	public SqlExpression visit(BinOpBoolean arg)
	{
		return _factory.createBinOpBoolean(arg.getOperator(),
		        arg.getLeftOperand().accept(this),
		        arg.getRightOperand().accept(this)
		        );
	}

	public SqlExpression visit(BinOpCompare arg)
	{
		SqlExpression  retVal = null;
		Expression      aLOprnd = arg.getLeftOperand();
		Expression      aROprnd = arg.getRightOperand();

		if (aLOprnd.getType().isString() && aROprnd.getType().isString()) {
			retVal = visitCompareString(arg);
		}
		else {
			SqlExpression   sLOprnd = aLOprnd.accept(this);
			SqlExpression   sROprnd = aROprnd.accept(this);
			int oprtr = arg.getOperator();
			retVal = _factory.createBinOpCompare(oprtr,sLOprnd,sROprnd);
		}

		return retVal;
	}

	protected SqlExpression visitCompareString(BinOpCompare arg)
	{
		SqlExpression  retVal = null;

		Expression      aLOprnd = arg.getLeftOperand();
		SqlExpression   sLOprnd = aLOprnd.accept(this);

		Expression      aROprnd = arg.getRightOperand();
		SqlExpression   sROprnd = aROprnd.accept(this);

		int oprtr = arg.getOperator();
		DbType leftType = ((SqlTypeAtom)aLOprnd.getType()).getType();
		DbType rightType = ((SqlTypeAtom)aROprnd.getType()).getType();
		boolean leftIsChar = leftType.getPrimitiveTypeCode() == java.sql.Types.CHAR;
		boolean rightIsChar = rightType.getPrimitiveTypeCode() == java.sql.Types.CHAR;

		if (leftIsChar || rightIsChar) {
		    /* when one operand is of CHAR (including string constant), suffix
			 space should be considered. that is why the size of the column is added */
		    
			SqlExpression comp = null;
			SqlExpression eq = null;
			SqlExpression lenLOprnd = null;
			SqlExpression lenROprnd = null;
			SqlExpression lenL_lenR = null;
			SqlExpression and =  null;
			SqlExpression or =  null;
			
			if (leftIsChar)
				lenLOprnd = _factory.createLitInteger(new Integer((int)leftType.getLength()));
			else
			    lenLOprnd = _factory.createSfStringLength(sLOprnd);

			if (rightIsChar)
			    lenROprnd = _factory.createLitInteger(new Integer((int)rightType.getLength()));
			else
			    lenROprnd = _factory.createSfStringLength(sROprnd);

			lenL_lenR = _factory.createBinOpCompare(oprtr,lenLOprnd,lenROprnd);
			
			// TODO: if two operands are char or litString, the length can either
			// make the length or the whole predicate useless (according the
			// fact that length are different or not) 

			switch (oprtr) {
				case LT_COMPOP:
					/* lOprnd < rOprnd ==> lOprnd< rOprnd or ( lOprnd = rOprnd and length(lOprnd) < length(rOprnd) ) */
				case GT_COMPOP:
					/* lOprnd > rOprnd ==> lOprnd > rOprnd or ( lOprnd = rOprnd and length(lOprnd) > length(rOprnd) ) */
					comp = _factory.createBinOpCompare(oprtr,sLOprnd,sROprnd);
					eq = _factory.createBinOpCompare(EQ_COMPOP,sLOprnd,sROprnd);
					and =  _factory.createBinOpBoolean(AND,eq,lenL_lenR);
					or =  _factory.createBinOpBoolean(OR, comp, and);
					retVal = or;
					break;
				case GEQ_COMPOP:
				case LEQ_COMPOP:
					retVal = _factory.createBinOpCompare(oprtr,sLOprnd,sROprnd);
					break;
				case NEQ_COMPOP:
					/* lOprnd != rOprnd ==> lOprnd != rOprnd or length(lOprnd) != length(rOprnd) */
					SqlExpression neq = _factory.createBinOpCompare(oprtr,sLOprnd,sROprnd);
					or = _factory.createBinOpBoolean(OR,neq,lenL_lenR);
					retVal = or;
					break;
				case EQ_COMPOP:
					/* lOprnd = rOprnd ==> lOprnd = rOprnd and length(lOprnd) = length(rOprnd) ) */
					eq = _factory.createBinOpCompare(oprtr,sLOprnd,sROprnd);
					and = _factory.createBinOpBoolean(AND,eq,lenL_lenR);
					retVal = and;
					break;
				default:
					throw new SqlWrapperException(MessageLibrary.getMessage("T_N_SUP_OPTR",COMPOPSSTRINGS[oprtr]));
			}
		}
		else	// Perform a simple comparison for VARCHARs
			retVal = _factory.createBinOpCompare(oprtr,sLOprnd,sROprnd);

		return retVal;
	}

	public SqlExpression visit(BinOpCompareAny arg)
	{
		return _factory.createBinOpCompareAny(arg.getOperator(),
		        arg.getLeftOperand().accept(this),
		        arg.getRightOperand().accept(this)
		        );
	}

	public SqlExpression visit(BinOpDifference arg)
	{
		throw new SqlWrapperException(MessageLibrary.getMessage("T_N_SUP_EXPR",arg),null);

	}

	public SqlExpression visit(BinOpIntersect arg)
	{
		throw new SqlWrapperException(MessageLibrary.getMessage("T_N_SUP_EXPR",arg),null);
	}

	public SqlExpression visit(BinOpMerge arg)
	{
		return _factory.createBinOpUnion (UNION_ALL,
		        arg.getLeftOperand().accept(this),
		        arg.getRightOperand().accept(this)
		        );
	}


	public SqlExpression visit(BinOpUnion arg)
	{
		return _factory.createBinOpUnion(UNION,
		        arg.getLeftOperand().accept(this),
		        arg.getRightOperand().accept(this)
		        );
	}

	public SqlExpression visit(UnOpExists arg)
	{
		return _factory.createUnOpExists(arg.getOperand().accept(this), arg.getNot());
	}

	public SqlExpression visit(UnOpIsNull arg) throws SqlWrapperException
	{
		return _factory.createUnOpIsNull(arg.getOperand().accept(this), arg.getNot());
	}

    public SqlExpression visit(ExternalVariable arg) throws SqlWrapperException {
        return _factory.createPlaceHolder(arg);
    }

	public SqlExpression visit(FunAggregate arg)
	{
		return _factory.createFunAggregate(
		        arg.getOperator(),
		        arg.getOperand().accept(this),
		        arg.getDistinct()
		        );
	}


	public SqlExpression visit(IfThenElse arg)
	{
		return _factory.createIfThenElse(
		        arg.getIf().accept(this),
		        arg.getThen().accept(this),
		        arg.getElse().accept(this)
		        );
	}

	// NOTE: we suppose no join would be heading an algebra tree.
	public SqlExpression visit(Join arg)
	{
		List relations = arg.getOperands();
		if (null == relations) {
			throw new SqlWrapperException(MessageLibrary.getMessage("INT_LOG_ERR"),null);
		}

		Expression      aRelation;
		int len = relations.size();
		
		/*  add join predicate to WHERE clause */
		List aPredicateList = arg.getPredicateList();
		if (null !=  aPredicateList) {
			Iterator iter = aPredicateList.iterator();
			SqlSelect select = (SqlSelect) _nestStack.peek();
			while (iter.hasNext()) {
				Expression aPredicate = (Expression)iter.next();
				SqlExpression sPredicate = aPredicate.accept(this);
				SqlExpression newWhereClause = whereClauseAddPredicate(select.getWhereClause(), sPredicate);
				select.setWhereClause(newWhereClause);
			}
		}
		else {
			if (log.isDebugEnabled())
				log.warn("join operation without predicate");
		}

		/* Create join clause */
		SqlJoin sJoin = _factory.createJoin();

		for (int i = 0; i < len; i++) {
			/* generate FROM clause for SELECT */
			aRelation = (Expression) relations.get(i);
			// ADD   06/07/2003
			if (aRelation instanceof UnOpSort)
				aRelation = ((UnOpSort)aRelation).getOperand();	
			if (aRelation instanceof UnOpRestrict)
				aRelation = ((UnOpRestrict)aRelation).getOperand();	
			// end ADD  	
			
			selectAddFromClause(sJoin, (RenameRelation)aRelation);
		}
		
		return sJoin;
	}

	public SqlExpression visit(LitBoolean arg)
	{
		return _factory.createLitBoolean (arg.getValue());
	}

	public SqlExpression visit(LitDate arg)
	{
		return _factory.createLitDate (arg.getValue());
	}


	public SqlExpression visit(LitDecimal arg)
	{
		return _factory.createLitDecimal(arg.getValue());
	}

	public SqlExpression visit(LitDouble arg)
	{
		return _factory.createLitDouble (arg.getValue());
	}

	public SqlExpression visit(LitFloat arg)
	{
		return _factory.createLitFloat (arg.getValue());
	}

	public SqlExpression visit(LitInteger arg)
	{
		return _factory.createLitInteger (arg.getValue());
	}

	public SqlExpression visit(LitList arg) {
		SqlLitList    retVal = null;
		List sExprList = new ArrayList(arg.getLitList().size());
		SqlExpression sExpr = null;
		for (int i = 0; i < arg.getLitList().size(); i++) {
			sExpr = ((Expression)arg.getLitList().get(i)).accept(this);
			sExprList.add(sExpr);
		}

		retVal = _factory.createLitList(sExprList);
		return retVal;
	}

	public SqlExpression visit(LitNull arg)
	{
		return _factory.createLitNull();
	}

	public SqlExpression visit(LitString arg)
	{
		return _factory.createLitString (arg.getValue());
	}

	public SqlExpression visit(RenameItem arg)
	{
		SqlRenameItem   retVal = null;
		SqlExpression   sExpr = arg.getOperand().accept(this);
		String          name  = arg.getName();
		retVal = _factory.createRenameItem (sExpr,name);

		return retVal;
	}

	public SqlExpression visit(RenameRelation arg)
	{
		SqlRenameRelation   retVal = null;
		SqlExpression   sExpr = arg.getOperand().accept(this);
		String          name  = arg.getUniqueName();
		retVal = _factory.createRenameRelation (sExpr,name);

		return retVal;
	}

	public SqlExpression visit(SortSpecification arg)
	{
		Expression sortExpr = arg.getSortExpression();
		SqlExpression   sqlSortExpr = sortExpr.accept(this);

		return _factory.createSortSpecification(sqlSortExpr,arg.getSortDirection());
	}


	public SqlExpression visit(Table arg)
	{
		return _factory.createTable(arg.getCatalogName(), arg.getSchemaName(), arg.getName());
	}

	public SqlExpression visit(ConstantTempTable arg)
	{
		SqlExpression retVal = null;
		List values = arg.getValues();
		Debug.assertTrue(null != values  && !values.isEmpty(), " TempTable is empty");

		SqlSelect   select = null;
		List        tuple = null;
		List        selectList = null;
		Expression  aAttr = null;
		SqlExpression sAttr = null;
		String      attrName = null;

		for (int i = 0; i < values.size(); i++) {
			select = _factory.createSelect(arg);

			tuple = (List)values.get(i);
			selectList = new ArrayList();
			for (int j = 0; j < tuple.size(); j++) {
				aAttr = (Expression)tuple.get(j);
				sAttr = aAttr.accept(this);
				if (0 ==i) {
					attrName = ((Attribute)arg.getAttributes().get(j)).getName();
					sAttr = _factory.createRenameItem(sAttr, attrName);
				}
				selectList.add(sAttr);
			}

			select.setSelectList(selectList);
			DummyTable dummyTable = new DummyTable();
			selectAddNextClause(select, dummyTable);

			if (null == retVal) {
				/* this select is the first select */
				retVal = select;
			}
			else {
				retVal = _factory.createBinOpUnion(Constants.UNION_ALL, retVal, select);
			}
		}

		return retVal;
	}

	public SqlExpression visit(TempTable arg)
	{
		SqlExpression retVal = null;
		
		// TODO: create a select into (create table ... as) statement

		return retVal;
	}
	
	public SqlExpression visit(TempValue arg)
	{
		return arg.getValue() == null ? arg.getOperand().accept(this) : arg.getValue().accept(this);
	}

	public SqlExpression visit(UnOpAggregate arg)
	{
		SqlSelect   retVal = _factory.createSelect(arg);
		_nestStack.push(retVal);
		retVal = selectAddSelectClause(retVal,arg);
		_nestStack.pop();

		return retVal;
	}

//    public SqlExpression visit(UnOpGroup arg)
//    {
//        Trace.enter(this,"visit(UnOpGroup arg)");
//
//        SqlSelect   retVal = _factory.createSelect();
//        retVal = selectAddGroupbyClause(retVal,arg);
//
//        Trace.exit(this,"visit(UnOpGroup arg)");
//        return retVal;
//    }

	public SqlExpression visit(UnOpMinus arg)
	{
		return _factory.createUnOpMinus(arg.getOperand().accept(this));
	}

	public SqlExpression visit(UnOpNot arg)
	{
		return _factory.createUnOpNot(arg.getOperand().accept(this));
	}

	public SqlExpression visit(UnOpProject arg)
	{
		SqlSelect   retVal = _factory.createSelect(arg);
		_nestStack.push(retVal);
		retVal = selectAddSelectClause(retVal,arg);
		_nestStack.pop();

		return retVal;
	}

	public SqlExpression visit(UnOpRestrict arg)
	{
		SqlSelect   retVal = _factory.createSelect(arg);
		_nestStack.push(retVal);
		if (arg.isPostResult())
			retVal = selectAddHavingClause(retVal, arg);
		else
			retVal = selectAddWhereClause(retVal, arg);
		_nestStack.pop();

		return retVal;
	}

	public SqlExpression visit(UnOpSort arg)
	{
		SqlSelect   retVal = _factory.createSelect(arg);
		_nestStack.push(retVal);
		selectAddSortClause(retVal,arg);
		_nestStack.pop();

		return retVal;
	}

	/**
	 * Main switch for select statement building.
     * @param select
     * @param operand
     * @return
     */
    protected SqlSelect selectAddNextClause(SqlSelect select, Expression operand)
	{
		if (operand instanceof UnOpProject) {
			if (select.getSelectDistinct() == false && null == select.getSelectList())
				return selectAddSelectClause(select,(UnOpProject)operand);
			else
				return selectAddFromClause(select, new RenameRelation(operand, _relIDProvider));
			/* NOTE:
			 * Some DBMS (like SQL Server) do not support select in from without
			 * relation alias. Doing it during algebra generation would prevent
			 * current RemoveProjectVisitor to collapse correctly piped projections.
			 */
			
		}
        if (operand instanceof UnOpRestrict) {
            UnOpRestrict restrict = (UnOpRestrict) operand;
            if (restrict.isPostResult())
                return selectAddHavingClause(select, (UnOpRestrict) operand);
            else
                return selectAddWhereClause(select, (UnOpRestrict) operand);
        }
        // TODO: Create a visitor to avoid these ugly "instanceof"
		if (operand instanceof UnOpAggregate)   return selectAddSelectClause(select,(UnOpAggregate)operand);
		if (operand instanceof UnOpGroup)       return selectAddGroupbyClause(select,(UnOpGroup)operand);
		if (operand instanceof UnOpSort)        return selectAddSortClause(select,(UnOpSort)operand);
		if (operand instanceof Join)            return selectAddFromClause(select,(Join)operand);
		if (operand instanceof Table)           return selectAddFromClause(select,(Table)operand);
		if (operand instanceof DummyTable)      return selectAddFromClause(select,(DummyTable)operand);
		if (operand instanceof BinOpOuterJoin)  return selectAddFromClause(select,(BinOpOuterJoin)operand);
		if (operand instanceof BinaryAlgebra)   return selectAddFromClause(select,(BinaryAlgebra)operand);
		if (operand instanceof RenameRelation)  return selectAddFromClause(select,(RenameRelation)operand);

		Debug.assertTrue(false,"NYI!"+operand);
		return null;
	}

	protected SqlSelect selectAddSortClause(SqlSelect select, UnOpSort sort)
	{
		List aSortList  = sort.getSortSpecificationList();
		if (null!=aSortList && null == select.getOrderbyClause()) {
			List sSortList  = new ArrayList();
			Iterator iter = aSortList.iterator();
			while (iter.hasNext()) {
				SortSpecification aSortSpecification = (SortSpecification)iter.next();
				SqlSortSpecification sSortSpecification = (SqlSortSpecification)aSortSpecification.accept(this);
				sSortList.add(sSortSpecification);
			}
			select.setOrderbyClause(sSortList);
		}
		select = selectAddNextClause(select,sort.getOperand());
		return select;
	}


	/**
	 * Sets the attributes of the select clause and calls the method to add the
	 * from clause corresponding to the operand. 
     * @param select
     * @param project
     * @return
     */
    protected SqlSelect selectAddSelectClause(SqlSelect select, UnOpProject project)
	{
		List aExprList = project.getItemList();
		ArrayList sExprList = new ArrayList();

		if (null != aExprList)
		{
			// translate item list in UnOpProject into SqlExpression
			Iterator  itr = aExprList.iterator () ;
			while ( itr.hasNext())
			{
				SqlExpression sExpr = ((Expression)itr.next()).accept (this);
				sExprList.add(sExpr) ;
			}
			select.setSelectClause ( project.getDistinct(),sExprList);
		}

		select = selectAddNextClause(select,project.getOperand());
		return select;
	}


	protected SqlJoin selectAddFromClause(SqlJoin join, RenameRelation arg)
	{
		join.addOperand(processPipedRelation(arg));
		return join;
	}
	
	protected SqlSelect selectAddFromClause(SqlSelect select, RenameRelation arg)
	{
		select.addFromClause(processPipedRelation(arg));
		return select;
	}

	protected SqlRenameRelation processPipedRelation(RenameRelation arg)
	{
		SqlRenameRelation sRR = _factory.createRenameRelation (
		        arg.getOperand().accept(this), arg.getUniqueName());

		return sRR;
	}


	protected SqlSelect selectAddFromClause (SqlSelect select, Join join)
	{
		/* Add it to the select from clause */
		select.addFromClause(join.accept(this));
		return select;
	}

	// TODO: this default implementation is for Oracle (7 because outer joins are
	// used). Move if mysql and progress use the same standard syntax ?
	protected SqlSelect selectAddFromClause (SqlSelect select, BinOpOuterJoin arg)
	{
		Expression [] relations = {arg.getLeftOperand(), arg.getRightOperand()};

		RenameRelation aRR;
		Expression      aRelation;
		SqlExpression   sRelation;
		SqlRenameRelation    sInnerTable = null;

		for (int i = 0; i < relations.length; i++) {
			/* generate FROM clause for SELECT */
			aRelation = relations[i];
			if (aRelation instanceof RenameRelation) {
				aRR = (RenameRelation)aRelation;
				aRelation = aRR.getOperand();
				
				// NOTE: a list of if (instanceof) used to check algebra integrity have
				// been removed because these tests make visitor automatic type switch worthless...
				sRelation = aRelation.accept(this);
				SqlRenameRelation sRR = _factory.createRenameRelation (sRelation, aRR.getUniqueName());
				select.addFromClause(sRR);
				
				if (LEFT_OUTER_JOIN == arg.getJoinType() && i == 1){
					sInnerTable = sRR;
				}
				else if (RIGHT_OUTER_JOIN == arg.getJoinType() && i == 0){
					sInnerTable = sRR;
				}
			}
			else {
				selectAddNextClause(select ,aRelation);
			}
		}

		/*  add join predicate to WHERE clause */
		List aPredicateList = arg.getPredicateList();
		if  (null !=  aPredicateList)
		{
			Iterator iter = aPredicateList.iterator();
			while (iter.hasNext()) {
				Expression aPredicate = (Expression)iter.next();
				SqlExpression sPredicate = aPredicate.accept(this);
				sPredicate = new SqlOuterJoinPredicate(sPredicate ,sInnerTable);
				SqlExpression newWhereClause = whereClauseAddPredicate(select.getWhereClause(), sPredicate);
				select.setWhereClause(newWhereClause);
			}
		}
		else
		{
			if (log.isDebugEnabled())
				log.warn("outer join operation without predicate");
		}

		return select;
	}


	protected SqlSelect selectAddWhereClause (SqlSelect select, UnOpRestrict restrict)
	{
		List aPredicates = restrict.getPredicateList();
		SqlExpression sPredicate ;

		if ( null != aPredicates ) {
			Iterator iter = aPredicates.iterator();
			while (iter.hasNext()) {
				Expression aPredicate = (Expression)iter.next();
				sPredicate = aPredicate.accept(this) ;
				SqlExpression newWhereClause = whereClauseAddPredicate(select.getWhereClause(), sPredicate);
				select.setWhereClause(newWhereClause);
			}
		}

		select = selectAddNextClause(select, restrict.getOperand());
		return select;
	}

	protected SqlSelect selectAddHavingClause (SqlSelect select, UnOpRestrict restrict)
	{
		List aPredicates = restrict.getPredicateList();
		SqlExpression sPredicate ;

		if ( null != aPredicates ) {
			Iterator iter = aPredicates.iterator();
			while (iter.hasNext()) {
				Expression aPredicate = (Expression)iter.next();
				sPredicate = aPredicate.accept(this) ;
				SqlExpression newWhereClause = whereClauseAddPredicate(select.getHavingClause(), sPredicate);
				select.setHavingClause(newWhereClause);
			}
		}

		select = selectAddNextClause(select, restrict.getOperand());
		return select;
	}

	public SqlSelect selectAddGroupbyClause (SqlSelect select, UnOpGroup group)
	{
		List aExprList = group.getGroupExpressionList();
		ArrayList sExprList = new ArrayList();
		if (null != aExprList)
		{
			// translate group expression list in UnOpGroup into SqlExpression
			Iterator  itr = aExprList.iterator () ;
			while ( itr.hasNext())
			{
				SqlExpression sExpr = ((Expression)itr.next()).accept (this);
				sExprList.add(sExpr) ;
			}
			select.setGroupbyClause (sExprList);
			select = selectAddNextClause(select,group.getOperand());
		}

		return select;
	}

	protected SqlSelect selectAddFromClause (SqlSelect select, Table operand)
	{
		SqlTable newFrom = _factory.createTable (operand.getCatalogName(), operand.getSchemaName(), operand.getName());
		select.addFromClause(newFrom);
		return select;
	}

	protected SqlSelect selectAddFromClause (SqlSelect select, DummyTable operand)
	{
		return select;
	}

	protected SqlSelect selectAddFromClause (SqlSelect select, BinaryAlgebra operand)
	{
		SqlExpression sRelation = operand.accept(this);
		SqlRenameRelation sRR = _factory.createRenameRelation ( sRelation,null);

		select.addFromClause(sRR);
		return select;
	}

	public SqlExpression visit(AfTrim arg) throws SqlWrapperException
	{
		return _factory.createSfTrim(((Expression)arg.getArgument(0)).accept(this));
	}

	public SqlExpression visit(XfCast arg) throws SqlWrapperException
	{
		Expression expr = (Expression)arg.getExpression();
		SqlExpression sqlExpr = expr.accept(this);
		DbType type0= ((SqlTypeAtom)expr.getType()).getType();
		DbType type1= (DbType)arg.getTargetType();
		SqlExpression retVal = _factory.createConvert(sqlExpr, (DbType)type0, type1);
		return retVal;
	}

	public SqlExpression visit(AfCast arg) throws SqlWrapperException
	{
		SqlExpression retVal = null;
		Expression expr = (Expression)arg.getExpression();

		DbType type0= ((SqlTypeAtom)expr.getType()).getType();
		DbType type1= arg.getTargetType();

		String name = expr.getName();
		if (null != name){
			if (expr instanceof RenameItem) {
				expr = ((RenameItem)expr).getOperand();;
			}
			SqlExpression sqlExpr = expr.accept(this);
			retVal = _factory.createConvert(sqlExpr, (DbType)type0, type1);
			retVal = _factory.createRenameItem(retVal,name);
		}
		else {
			SqlExpression sqlExpr = expr.accept(this);
			retVal = _factory.createConvert(sqlExpr, (DbType)type0, type1);
		}

		return retVal;
	}

	public SqlExpression visit(XfAbs arg) throws SqlWrapperException
	{
		return _factory.createSfAbs(((Expression)arg.getArgument(0)).accept(this));
	}

	public SqlExpression visit(XfCeiling arg) throws SqlWrapperException
	{
		return _factory.createSfCeiling(((Expression)arg.getArgument(0)).accept(this));
	}

	public SqlExpression visit(XfConcat arg)
	{
		SfConcat  retVal = _factory.createSfConcat();

		List  sqlArguments = new ArrayList();

		List  algArguments = arg.getArguments();
		Iterator iter = algArguments.iterator();
		while (iter.hasNext()) {
			  Expression item = (Expression)iter.next();
			  sqlArguments.add(item.accept(this));
		}

		retVal.setArguments (sqlArguments);
		return retVal;
	}


	public SqlExpression visit(XfContains arg) throws SqlWrapperException
	{
		Debug.assertTrue(2 == arg.getArgumentNumber(),"2 == arg.getArgumentNumber()");
		SqlBinOpLike retVal = _factory.createBinOpLike();

		Expression aArg0 = (Expression)arg.getArgument(0);
		SqlExpression sArg0 = aArg0.accept(this);

		Expression aArg1 = (Expression)arg.getArgument(1);
		SqlExpression sArg1 = aArg1.accept(this);

		retVal.setLeftOperand(sArg0);

		SfConcat pattern = _factory.createSfConcat();
		pattern.setArgument(0, _factory.createLitString("%"));
		pattern.setArgument(1,sArg1);
		pattern.setArgument(2, _factory.createLitString("%"));

		retVal.setRightOperand(pattern);

		return retVal;
	}

	public SqlExpression visit(XfCurrentDateTime arg) throws SqlWrapperException
	{
		return _factory.createSfCurrentDateTime();
	}

	public SqlExpression visit(XfDate arg) throws SqlWrapperException
	{
		Expression arg0 = (Expression)arg.getArgument(0);
		DbType arg0Type = (DbType)((SqlTypeAtom)arg0.getType()).getType();

		SqlExpression aArg = arg0.accept(this);
		return _factory.createConvert(aArg, arg0Type, new DbType(Types.DATE));
	}

	public SqlExpression visit(XfDateTime arg) throws SqlWrapperException
	{
		Expression arg0 = (Expression)arg.getArgument(0);
		DbType arg0Type = (DbType)((SqlTypeAtom)arg0.getType()).getType();

		SqlExpression aArg = arg0.accept(this);
		return _factory.createConvert(aArg, arg0Type, new DbType(Types.TIMESTAMP));
	}

	public SqlExpression visit(XfEndsWith arg) throws SqlWrapperException
	{
		Debug.assertTrue(2 == arg.getArgumentNumber(),"2 == arg.getArgumentNumber()");
		SqlBinOpLike retVal = _factory.createBinOpLike();

		Expression aArg0 = ((Expression)arg.getArgument(0));
		SqlExpression sArg0 = aArg0.accept(this);

		Expression aArg1 = (Expression)arg.getArgument(1);
		SqlExpression sArg1 = aArg1.accept(this);

		retVal.setLeftOperand(sArg0);

		SfConcat pattern = _factory.createSfConcat();
		pattern.setArgument(0, _factory.createLitString("%"));
		pattern.setArgument(1,sArg1);

		retVal.setRightOperand(pattern);

		return retVal;
	}


	public SqlExpression visit(XfFloor arg) throws SqlWrapperException
	{
		return _factory.createSfFloor(((Expression)arg.getArgument(0)).accept(this));
	}


	public SqlExpression visit(XfLowerCase arg) throws SqlWrapperException
	{
		Debug.assertTrue(1 == arg.getArgumentNumber(),"1 == arg.getArgumentNumber()");
		SfLowerCase retVal = _factory.createSfLowerCase();

		Expression aArg = ((Expression)arg.getArgument(0));
		SqlExpression sArg = aArg.accept(this);
		retVal.setArgument(0,sArg);

		return retVal;
	}


	public SqlExpression visit(XfRound arg) throws SqlWrapperException
	{
		Expression arg0 = (Expression)arg.getArgument(0);
		SqlExpression retVal = _factory.createBinOpArithmetic(PLUS_ARITHMETICS,
			arg0.accept(this), _factory.createLitFloat(new Float(0.5)));
		_factory.createSfFloor(retVal);
		return retVal;
	}


	public SqlExpression visit(XfStartsWith arg) throws SqlWrapperException
	{
		Debug.assertTrue(2 == arg.getArgumentNumber(),"2 == arg.getArgumentNumber()");
		SqlBinOpLike retVal = _factory.createBinOpLike();

		Expression aArg0 = ((Expression)arg.getArgument(0));
		SqlExpression sArg0 = aArg0.accept(this);

		Expression aArg1 = (Expression)arg.getArgument(1);
		SqlExpression sArg1 = aArg1.accept(this);

		retVal.setLeftOperand(sArg0);

		SfConcat pattern = _factory.createSfConcat();
		pattern.setArgument(0,sArg1);
		pattern.setArgument(1, _factory.createLitString("%"));

		retVal.setRightOperand(pattern);

		return retVal;
	}


	public SqlExpression visit(XfStringLength arg) throws SqlWrapperException
	{
		Debug.assertTrue(1 == arg.getArgumentNumber(),"1 == arg.getArgumentNumber()");
		SfStringLength retVal = _factory.createSfStringLength();

		Expression aArg = ((Expression)arg.getArgument(0));
		SqlExpression sArg = aArg.accept(this);
		retVal.setArgument(0,sArg);

		return retVal;
	}


	public SqlExpression visit(XfSubString arg) throws SqlWrapperException
	{
		SfSubString retVal = _factory.createSfSubString();

		Expression aArg = ((Expression)arg.getArgument(0));
		SqlExpression sArg = aArg.accept(this);
		retVal.setArgument(0,sArg);

		aArg = (Expression)arg.getArgument(1);
		sArg = aArg.accept(this);
		retVal.setArgument(1,sArg);

		if (2 < arg.getArgumentNumber()) {
			aArg = (Expression)arg.getArgument(2);
			sArg = aArg.accept(this);
			retVal.setArgument(2,sArg);
		}

		return retVal;
	}

	public SqlExpression visit(XfTime arg) throws SqlWrapperException
	{
		Expression arg0 = (Expression)arg.getArgument(0);
		DbType arg0Type = (DbType)((SqlTypeAtom)arg0.getType()).getType();

		SqlExpression aArg = arg0.accept(this);
		SqlConvert retVal = _factory.createConvert(aArg, arg0Type, new DbType(Types.TIME));

		return retVal;
	}

	public SqlExpression visit(XfUpperCase arg) throws SqlWrapperException
	{
		Debug.assertTrue(1 == arg.getArgumentNumber(),"1 == arg.getArgumentNumber()");
		SfUpperCase retVal = _factory.createSfUpperCase();

		Expression aArg = ((Expression)arg.getArgument(0));
		SqlExpression sArg = aArg.accept(this);
		retVal.setArgument(0,sArg);

		return retVal;
	}

	protected SqlExpression whereClauseAddPredicate(SqlExpression whereClause, SqlExpression predicate)
	{
		return whereClause == null ? predicate : _factory.createBinOpBoolean(AND, whereClause, predicate);
	}


    public org.xquark.extractor.sql.SqlExpression visit(BinOpOuterJoin arg)
    {
    	SqlSelect   retVal = _factory.createSelect(arg);
    	_nestStack.push(retVal);
    	selectAddFromClause(retVal,arg);
    	_nestStack.pop();
    
    	return retVal;
    }
}
