/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Odonata.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.jdbc.datasource;

import java.sql.Driver;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;

import org.apache.commons.dbcp.SQLNestedException;

public class JDBCDataSourceFactory {
	
	/** List of supported JDBC drivers */
	
	private static final String[] DRIVER_LIST = {
		"com.mysql.jdbc.Driver",
		"com.microsoft.jdbc.sqlserver.SQLServerDriver",
		"com.sybase.jdbc2.jdbc.SybDriver",
		"oracle.jdbc.driver.OracleDriver"
	};

	private ClassLoader classLoader;
	private List drivers;

    private JDBCDataSourceFactory(ClassLoader cl) {
    	if (cl != null) classLoader = cl;
    	else classLoader = getClass().getClassLoader();
		drivers = new ArrayList();
		for (int i = 0; i < DRIVER_LIST.length; i++) {
			try {
				Class c = Class.forName(DRIVER_LIST[i], true, classLoader);
				drivers.add(c.newInstance());
			} catch (ClassNotFoundException e) {
			} catch (IllegalAccessException e) {
			} catch (InstantiationException e) {
			}
		}
    }
    
    public static JDBCDataSourceFactory newInstance() {
    	return new JDBCDataSourceFactory(null);
    }
    
	public static JDBCDataSourceFactory newInstance(ClassLoader loader) {
		return new JDBCDataSourceFactory(loader);
	}
    
	public JDBCDataSource newDataSource(String url, String user, String password) 
	throws SQLException 
	{
		return newDataSource(null, url, user, password);
	}

    public JDBCDataSource newDataSource(String driverName, String url, String user, String password) 
    throws SQLException 
    {
    	Properties jdbcProperties = new Properties();
    	jdbcProperties.put("user", user);
    	jdbcProperties.put("password", password);
    	return createDataSource(driverName, url, jdbcProperties);
    }

	public JDBCDataSource newDataSource(String driverName, String url, Properties jdbcProperties) 
	throws SQLException 
	{
		if (jdbcProperties.get("user") == null)
			throw new SQLException("Missing user information in JDBC properties");
		if (jdbcProperties.get("password") == null)
			throw new SQLException("Missing password information in JDBC properties");
		return createDataSource(driverName, url, jdbcProperties);
	}
	
	private JDBCDataSource createDataSource(String driverName, String url, Properties jdbcProperties) 
	throws SQLException 
	{
		Driver driver = null;
		if (driverName != null) {	
            try {
                Class c = Class.forName(driverName, true, classLoader);
				Driver d = (Driver) c.newInstance();
				if (d.acceptsURL(url)) {
					driver = d;
				}
            } catch (ClassNotFoundException e) {
            	throw new SQLNestedException("Could not load driver "+driverName, e);
            } catch (InstantiationException e) {
				throw new SQLNestedException("Could not create driver "+driverName, e);
            } catch (IllegalAccessException e) {
				throw new SQLNestedException("Could not create driver "+driverName, e);
            }
		} else {
				for (int i = 0; driver == null && i < drivers.size(); i++) {
					Driver d = (Driver) drivers.get(i);
					if (d.acceptsURL(url)) {
						driver = d;
					}
				}
		}
		if (driver == null)
			throw new SQLException("No suitable driver found for "+url);
		return new JDBCDataSource(driver, url, jdbcProperties);
	}

}
