/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Odonata.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.jdbc.datasource;

import java.io.PrintWriter;
import java.sql.Connection;
import java.sql.SQLException;

import javax.sql.DataSource;

import org.apache.commons.dbcp.ConnectionFactory;
import org.apache.commons.dbcp.PoolableConnectionFactory;
import org.apache.commons.dbcp.PoolingDataSource;
import org.apache.commons.pool.impl.GenericObjectPool;


public class SingleConnectionDataSource implements DataSource {

	private PoolingDataSource dataSource;
	private Connection singleConnection;

	public SingleConnectionDataSource(Connection connection) throws SQLException {
		singleConnection = connection;

		GenericObjectPool connectionPool = new GenericObjectPool();
		connectionPool.setMaxActive(1);
		connectionPool.setMaxIdle(1);
		connectionPool.setMinIdle(1);
		connectionPool.setWhenExhaustedAction(GenericObjectPool.WHEN_EXHAUSTED_FAIL);

		ConnectionFactory jdbcConnectionFactory = new ConnectionFactory() {

            public Connection createConnection() throws SQLException {
                return singleConnection;
            }

		};

		PoolableConnectionFactory connectionFactory =
			new PoolableConnectionFactory(jdbcConnectionFactory,
										  connectionPool,
										  null,
										  null,
										  false,
										  true);
		if (connectionFactory == null) {
			throw new SQLException("Cannot create PoolableConnectionFactory");
		}
		dataSource = new PoolingDataSource(connectionPool);
	}
	
	/**
	 * Create (if necessary) and return a connection to the database.
	 *
	 * @exception SQLException if a database access error occurs
	 */
	public Connection getConnection() throws SQLException {
		return dataSource.getConnection();
	}

	/**
	 * Create (if necessary) and return a connection to the database.
	 *
	 * @param username Database user on whose behalf the Connection
	 *   is being made
	 * @param password The database user's password
	 *
	 * @exception SQLException if a database access error occurs
	 */
	public Connection getConnection(String username, String password) throws SQLException {
		return dataSource.getConnection(username, password);
	}

	/**
	 * Return the login timeout (in seconds) for connecting to the database.
	 *
	 * @exception SQLException if a database access error occurs
	 */
	public int getLoginTimeout() throws SQLException {
		return dataSource.getLoginTimeout();
	}

	/**
	 * Return the log writer being used by this data source.
	 *
	 * @exception SQLException if a database access error occurs
	 */
	public PrintWriter getLogWriter() throws SQLException {
		return dataSource.getLogWriter();
	}

	/**
	 * Set the login timeout (in seconds) for connecting to the database.
	 *
	 * @param loginTimeout The new login timeout, or zero for no timeout
	 *
	 * @exception SQLException if a database access error occurs
	 */
	public void setLoginTimeout(int loginTimeout) throws SQLException {
		dataSource.setLoginTimeout(loginTimeout);
	}

	/**
	 * Set the log writer being used by this data source.
	 *
	 * @param logWriter The new log writer
	 *
	 * @exception SQLException if a database access error occurs
	 */
	public void setLogWriter(PrintWriter logWriter) throws SQLException {
		dataSource.setLogWriter(logWriter);
	}

}
