/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * Copyright (C) 2003 Odonata.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.jdbc.typing;

import java.net.URL;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.SQLException;
import java.util.HashMap;
import java.util.Map;


/**
 * AbstractConnection factory.
 *
 * <p>This object holds database vendor specificities routing.</p>
 */
public class DBMSInfoMap implements DBMSConstants
{
    private static final String RCSRevision = "$Revision: 1.2 $";
    private static final String RCSName     = "$Name: XQ1_1_1 $";

    ///////////////////////////////////////////////////////////////
    //                      Constants
    ///////////////////////////////////////////////////////////////
	private static String RESOURCES_FOLDER = "resources/";
	
    ///////////////////////////////////////////////////////////////
    //                      Data
    ///////////////////////////////////////////////////////////////
    private static DBMSInfoLoader dbInfoLoader;
    private static Map dbmsInfos;
    
    static
    {
    	dbInfoLoader = new DBMSInfoLoader();
    	dbmsInfos = new HashMap();
    }
    
    private DBMSInfoMap() {}
    
    ///////////////////////////////////////////////////////////////
    //                      Public Methods
    ///////////////////////////////////////////////////////////////
	public static synchronized DBMSInfo getDBMSInfo(short DBMS)
	{
		Short key = new Short(DBMS);
		DBMSInfo ret = (DBMSInfo) dbmsInfos.get(key);
		if (ret == null)
		{
			String filename;
			switch (DBMS)
			{
				case ORACLE8I : case ORACLE8 : case ORACLE :
					filename = ORACLE_FILE;
					break;
				case TIMESTEN40 : case TIMESTEN :
				    filename = TIMESTEN_FILE;
					break;
				case SYBASE_ASE_11_9 : case SYBASE :
				    filename = SYBASE_FILE;
					break;
				case SQL_SERVER_2000 : case SQL_SERVER :
				    filename = SQL_SERVER_FILE;
					break;
				case MYSQL : case MYSQL323 :
					filename = MYSQL_FILE;
					break;
				default :
					throw new RuntimeException("Your RDBMS is not supported in the current version of XQuark products (basing on your JDBC driver informations).");
			}
			URL specsURL = DBMSInfoMap.class.getResource(RESOURCES_FOLDER + filename);
			ret = (DBMSInfo) dbInfoLoader.fromXML(specsURL.toString());
			ret.setDbms(DBMS);
			dbmsInfos.put(key, ret);
		}
		return ret;
	}

	public static synchronized DBMSInfo getDBMSInfo(Connection conn) throws SQLException
	{
	    return getDBMSInfo(getDatabaseVendor(conn));
	}
	
    /** Set the database type variable according to JDBC metadata
     */
    public static short getDatabaseVendor(Connection conn) throws SQLException
    {
        short databaseVendor = -1;
        DatabaseMetaData dbmeta = conn.getMetaData();
        String productName = dbmeta.getDatabaseProductName();
        String version = dbmeta.getDatabaseProductVersion();
        
        if (productName.equalsIgnoreCase("TimesTen"))   // TIMESTEN
        {
            if (version.startsWith("04.00"))
                databaseVendor = TIMESTEN40;
            else
                databaseVendor = TIMESTEN;
        }
        else if (productName.equalsIgnoreCase("Oracle"))   // ORACLE
        {
            if (version.startsWith("Oracle8i") || version.startsWith("Personal Oracle8i"))
                databaseVendor = ORACLE8I;
            else if (version.startsWith("Oracle8"))
                databaseVendor = ORACLE8;
            else
                databaseVendor = ORACLE;
        }
    	else if (productName.equalsIgnoreCase("MYSQL"))   // ORACLE
    	{
    		if (version.startsWith("3.23"))
    			databaseVendor = MYSQL323;
    		else
    			databaseVendor = MYSQL;
    	}
        else if (productName.equalsIgnoreCase("Adaptive Server Enterprise"))   // Sybase
        {
            if (version.startsWith("Adaptive Server Enterprise/11.9"))
                databaseVendor = SYBASE_ASE_11_9;
            else
                databaseVendor = SYBASE;
        }
        else if (productName.equalsIgnoreCase("Microsoft SQL Server"))
        {
            if (version.indexOf("2000") >= 0)
                databaseVendor = SQL_SERVER_2000;
            else
                databaseVendor = SQL_SERVER;
        }
        else if (isJDBC2(conn))
            databaseVendor = JDBC2;
        else
            databaseVendor = JDBC1;
        return databaseVendor;
    }

    private static boolean isJDBC2(Connection conn) throws SQLException
    {
        try {
            conn.getTypeMap();
        }
        catch (AbstractMethodError e) {
            return false;
        }
        return true;
    }
}
