/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * Copyright (C) 2003 Odonata.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.jdbc.typing;

import java.math.BigDecimal;
import java.sql.*;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.sql.Types;

import org.xquark.schema.SchemaException;
import org.xquark.schema.SimpleType;
import org.xquark.schema.datatypes.*;
import org.xquark.schema.datatypes.PrimitiveType;
import org.xquark.schema.datatypes.QName;
import org.xquark.schema.datatypes.URI;

/** This class describes features of the database corresponding type
 * for a datum to map, basing on the XML schema type.
 *
 * <p>Performs checks between XML type and SQL type.</p>
 * 
 * <p>The type info build is the natural mapping between schema types
 * and JDBC. It is not built from the database metadata but will be
 * checked against it.</p>
 */
public class MappingInfo extends JavaTypeInfo implements DBMSConstants
{
    private static final String RCSRevision = "$Revision: 1.2 $";
    private static final String RCSName     = "$Name: XQ1_1_1 $";
    
	/** mapping between schema type and JDBC base (abstract) types */
	public static int ABSTRACT_SCHEMA_JDBC_MAPPING[] = new int[33];
	
	static
	{
		// mapping between schema type and base JDBC base types
		ABSTRACT_SCHEMA_JDBC_MAPPING[PrimitiveType.STRING] = Types.VARCHAR;
		ABSTRACT_SCHEMA_JDBC_MAPPING[PrimitiveType.BOOLEAN] = Types.DECIMAL;
		ABSTRACT_SCHEMA_JDBC_MAPPING[PrimitiveType.DECIMAL] = Types.DECIMAL;
		ABSTRACT_SCHEMA_JDBC_MAPPING[PrimitiveType.FLOAT] = Types.FLOAT;
		ABSTRACT_SCHEMA_JDBC_MAPPING[PrimitiveType.DOUBLE] = Types.DOUBLE;
		ABSTRACT_SCHEMA_JDBC_MAPPING[PrimitiveType.QNAME] = Types.VARCHAR;
		ABSTRACT_SCHEMA_JDBC_MAPPING[PrimitiveType.NOTATION] = Types.VARCHAR;
		ABSTRACT_SCHEMA_JDBC_MAPPING[PrimitiveType.DURATION] = Types.VARCHAR;
		ABSTRACT_SCHEMA_JDBC_MAPPING[PrimitiveType.DATE_TIME] = Types.TIMESTAMP;
		ABSTRACT_SCHEMA_JDBC_MAPPING[PrimitiveType.TIME] = Types.TIMESTAMP;
		ABSTRACT_SCHEMA_JDBC_MAPPING[PrimitiveType.DATE] = Types.TIMESTAMP;
		ABSTRACT_SCHEMA_JDBC_MAPPING[PrimitiveType.GYEAR_MONTH] = Types.TIMESTAMP;
		ABSTRACT_SCHEMA_JDBC_MAPPING[PrimitiveType.GYEAR] = Types.TIMESTAMP;
		ABSTRACT_SCHEMA_JDBC_MAPPING[PrimitiveType.GMONTH_DAY] = Types.TIMESTAMP;
		ABSTRACT_SCHEMA_JDBC_MAPPING[PrimitiveType.GDAY] = Types.TIMESTAMP;
		ABSTRACT_SCHEMA_JDBC_MAPPING[PrimitiveType.GMONTH] = Types.TIMESTAMP;
		ABSTRACT_SCHEMA_JDBC_MAPPING[PrimitiveType.HEX_BINARY] = Types.VARBINARY;
		ABSTRACT_SCHEMA_JDBC_MAPPING[PrimitiveType.BASE64_BINARY] = Types.VARBINARY;
		ABSTRACT_SCHEMA_JDBC_MAPPING[PrimitiveType.ANY_URI] = Types.VARCHAR;
		ABSTRACT_SCHEMA_JDBC_MAPPING[PrimitiveType.UNION] = Types.VARCHAR; // For generation, a VARCHAR column, may accept almost anything...
		ABSTRACT_SCHEMA_JDBC_MAPPING[PrimitiveType.LIST] = Types.VARCHAR;
		ABSTRACT_SCHEMA_JDBC_MAPPING[PrimitiveType.ANYSIMPLETYPE] = Types.VARCHAR;
	}
	
    /* setString() and getString() can be used */
    protected boolean isJDBCStringIOAllowed = true; // for default mapping
    
    /* setCharacterStream() and set() must be used */
    protected boolean useStreams = false; // for default mapping
    
    /**
     * No conversion to performs since the schema output fits direcly the
     * column JDBC type.
     */
    protected static final short NONE = 0;
    /** JDBC performed */
    protected static final short IMPLICIT = 1;
    /**
     * Performed by mapper specific code. See
     * {@link #storeConvertedToSQL(PreparedStatement, int, Object)}.
     */
    protected static final short AUTOMATIC = 2;
    
    /** User-defined (NYI) */
    protected static final short EXPLICIT = 3;
    
    /** Forbidden */
    protected static final short IMPOSSIBLE = 4;
    
    /** Flag used to determine if a  conversion is needed between the schema
     * output and JDBC
     */
    protected short conversion = IMPOSSIBLE; // for default mapping
    
    protected DbType dbType = null;
    protected boolean targetIsString;
    /* use string delimitor if target is String */
    protected boolean useStringDelimitor = false;
    
    //////////////////////////////////////////////////////////////////////////
    //                 CONSTRUCTORS & ACCESSORS
    //////////////////////////////////////////////////////////////////////////
	protected MappingInfo() {}

	/** Constructor used for user generators
     */
    public MappingInfo(String XMLType)
    {
        super(XMLType);
    }
    
    /* Used for value generators */
    public MappingInfo(SimpleType XMLType)
    {
        super(XMLType);
    }
    
    public MappingInfo(SimpleType XMLType, DbType cmeta)
    {
        super(XMLType);
        setDBType(cmeta);
    }
    
    protected void setDBType(DbType cmeta) {
		// keep columnMetadata
		this.dbType = cmeta;
		if (cmeta.isLongType())
			useStreams = true;
        
		int targetJDBCType = cmeta.getJDBCType();
		targetIsString = (targetJDBCType == Types.CHAR)
						|| (targetJDBCType == Types.VARCHAR)
						|| (targetJDBCType == Types.LONGVARCHAR);
        
        
		// set the isJDBCStringIOAllowed flag
		isJDBCStringIOAllowed = targetIsString || (javaType == JAVA_STRING);
    }
    
    public DbType getDBType()
    {
        return dbType;
    }

    /**
     * true if target is a SQL String Type or the schema convert output is a
     * Java String. In other words, generator returns a value that is of type
     * String because the actual value is a String or because the normalized or
     * canonical value is used.
     */ 
    public boolean useJDBCStringMethods()
    {
        return isJDBCStringIOAllowed;
    }
    
	/**
	 * true if target is a SQL String Type and the schema convert output is not
	 * a Java String.
	 */ 
    public boolean storeCanonicalString()
    {
        return isJDBCStringIOAllowed && (javaType != JAVA_STRING);
    }
    
    public void setParameter(PreparedStatement pStmt, int i, String value)
    throws SQLException, SchemaException
    {
        Object oValue = sType.convert(value);
        switch(javaType) {
        	case JAVA_STRING :
        	    pStmt.setString(i, (String)oValue);
        	    break;
        	case SCHEMA_QNAME :
        	    pStmt.setString(i, ((QName)oValue).toString());
        	    break;
        	case SCHEMA_URI :
        	    pStmt.setString(i, ((URI)oValue).toString());
        	    break;
        	case SCHEMA_DATETIME :
        	    pStmt.setTimestamp(i, new Timestamp(((DateTime)oValue).getTime()));	// TODO: consider TZ
        	    break;
        	case SCHEMA_DURATION :
        	    pStmt.setString(i, ((Duration)oValue).toString());
        	    break;
        	case JAVA_LONG :
        	    pStmt.setLong(i, ((Long)oValue).longValue());
        	    break;
        	case JAVA_BIG_DECIMAL :
        	    if (oValue instanceof Long)
        	        pStmt.setLong(i, ((Long)oValue).longValue());
        	    else
        	        pStmt.setBigDecimal(i, (BigDecimal)oValue);
        	    break;
        	case JAVA_FLOAT :
        	    pStmt.setFloat(i, ((Float)oValue).floatValue());
        	    break;
        	case JAVA_DOUBLE :
        	    pStmt.setDouble(i, ((Double)oValue).doubleValue());
        	    break;
        	case JAVA_BOOLEAN :
        	    pStmt.setBoolean(i, ((Boolean)oValue).booleanValue());
        	    break;
        	case SCHEMA_BYTE_ARRAY :
        	    pStmt.setBytes(i, ((ByteArray) oValue).getData());
        	    break;
        	default:
        }
    }

    public String toString()
    {
        return "[javaType=" + javaType + ",string=" + isJDBCStringIOAllowed + ",targetJDBC=" + dbType.getJDBCType() + "]";
    }
}
