/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * Copyright (C) 2003 Odonata.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.jdbc.typing;

import java.sql.*;
import java.util.*;

/**
 * Class containing information on JDBC to Native SQL types mapping.
 *
 */
public class TypeMap extends HashMap
{
    private static final String RCSRevision = "$Revision: 1.3 $";
    private static final String RCSName     = "$Name: XQ1_1_1 $";
    
    private SQLType OTHER_Type = new SQLType();
    
    /**
     * Map between JDBC types and natives type.
     */
    private Map canonicalTypesMap = new HashMap();
    
	public TypeMap() {
	    OTHER_Type.addJDBCType(Types.OTHER, null, true, false);
	}

    public void load(Connection connection) throws SQLException
    {
        ResultSet rs = connection.getMetaData().getTypeInfo();
        while (rs.next())
            addType(new SQLType(rs));
        rs.close();
    }
    
	public SQLType createType() {
		return new SQLType();
	}
	
	public void addType(SQLType nativeType)
	{
	    String key = nativeType.getNativeTypeName().toLowerCase();
	    if (get(key) != null)
	        throw new RuntimeException("Internal Error: duplicate entry in type XML data file.");
	    put(key, nativeType);
	}
	
	public SQLType getType(String nativeType)
	{
	    SQLType ret = (SQLType)get(nativeType.toLowerCase());
	    if (ret == null)
	        ret = OTHER_Type; // No fatal error : caller may need to ignore.
	    return ret;
	}
	
    public SQLType getType(int JDBCType)
    {
        SQLType targetType = (SQLType)canonicalTypesMap.get(new Integer(JDBCType));
        if (targetType == null)
            throw new RuntimeException("The JDBC type " + JDBCType + " is not supported by the database JDBC driver.");
        return targetType;
    }
    
    public String getNativeTypeName(int JDBCType)
    {
        return getType(JDBCType).getNativeTypeName();
    }
    
    public long getSize(int JDBCType)
    {
        return getType(JDBCType).getPrecision();
    }
    
	public boolean isUnsigned(int JDBCType)
	{
		return getType(JDBCType).isUnsigned();
	}
    
	public boolean useGetString(int JDBCType)
	{
		return getType(JDBCType).useGetString();
	}
    
    public String getTypeParam(int JDBCType)
    {
        return getType(JDBCType).getTypeParam();
    }
    
    ///////////////////////////////////////////////////////////////////////////
    //                  INNER CLASSES
    ///////////////////////////////////////////////////////////////////////////
	private class JDBCType implements Cloneable
    {
		private int JDBCType;
		private String createParam;   // String to concatenate to nativeTypeName when creating table
		private boolean useGetString;
									  // (containing precision for instance)
        JDBCType(int type, String createParam, boolean useGetString) {
            JDBCType = type;
            this.createParam = createParam;
            this.useGetString = useGetString;
        }
        /**
         * @return Returns the createParam.
         */
        String getCreateParam() {
            return createParam;
        }

        /**
         * @return Returns the jDBCType.
         */
        int getJDBCType() {
            return JDBCType;
        }

        public boolean useGetString() {
            return useGetString;
        }
        public void setUseGetString(boolean useGetString) {
            this.useGetString = useGetString;
        }

	}
    public class SQLType implements Cloneable
    {
        private String nativeTypeName;
        private List JDBCTypes = new ArrayList();   // note : int in java.sql.Types but short in rs
        private long precision;   // note : int in java.sql.Types but short in rs
		private boolean unsigned;
		private boolean indexable = true;
		private boolean extraction = true;	// supports extraction
		private boolean storage = true;	// supports insertion
		private boolean autoFill = false;
		/**
		 * Default constructor for TargetNativeType.
		 */
		public SQLType() {};

        SQLType(ResultSet rs) throws SQLException
        {
            nativeTypeName = rs.getString(1);
            addJDBCType(rs.getShort(2), rs.getString(6), false, false);
            precision = rs.getLong(3);  // getInt -> getLong because of Oracle JDBC 8.1.7
            unsigned = rs.getBoolean(10);
        }
        
        String getNativeTypeName()
        {
            return nativeTypeName;
        }
        String getTypeParam()
        {
            return ((JDBCType)JDBCTypes.get(0)).getCreateParam();
        }
		int getJDBCType()
		{
			return ((JDBCType)JDBCTypes.get(0)).getJDBCType();
		}
		boolean useGetString()
		{
			return ((JDBCType)JDBCTypes.get(0)).useGetString();
		}
        long getPrecision()
        {
            return precision;
        }
        boolean isUnsigned()
        {
            return unsigned;
        }
        public void setNativeTypeName(String typeName)
        {
            nativeTypeName = typeName;
        }
        public void setPrecision(long prec)
        {
            precision = prec;
        }
        public void setUnsigned(boolean unsigned)
        {
            this.unsigned = unsigned;
        }
        public String toString()
        {
            return "[jdbc=" + JDBCTypes + ", sql=" + nativeTypeName + ", prec=" + precision
            + ", unsigned=" + unsigned + "]";
        }
        public Object clone()
        {
            try {
                return super.clone();
            }
            catch (CloneNotSupportedException e) {
                return null;
            }
        }
		/**
		 * Sets the jDBCType.
		 * @param jDBCType The jDBCType to set
		 */
		public void addJDBCType(int jDBCType, String createParams, boolean principal, boolean useGetString)
		{
		    Integer key = new Integer(jDBCType);
		    JDBCType typeInfo = new JDBCType(jDBCType, createParams, useGetString);
		    if (principal) {
		        JDBCTypes.add(0, typeInfo);
				canonicalTypesMap.put(key, this);
		    }
		    else {
				JDBCTypes.add(typeInfo);
		        if (canonicalTypesMap.get(key) == null)
		            canonicalTypesMap.put(key, this);
		    }
		}
		
		public void addJDBCType(String jDBCType, String createParams, boolean principal, boolean useGetString)
		{
			addJDBCType(parseJDBCType(jDBCType), createParams, principal, useGetString);
		}

        public boolean isAutoFill() { return autoFill;}

        public void setAutoFill(boolean autoFill) { this.autoFill = autoFill;}

		/**
		 * @return Returns the extraction.
		 */
		public boolean isExtraction() {
			return extraction;
		}
		/**
		 * @param extraction The extraction to set.
		 */
		public void setExtraction(boolean extraction) {
			this.extraction = extraction;
		}
		/**
		 * @return Returns the indexable.
		 */
		public boolean isIndexable() {
			return indexable;
		}
		/**
		 * @param indexable The indexable to set.
		 */
		public void setIndexable(boolean indexable) {
			this.indexable = indexable;
		}
		/**
		 * @return Returns the storage.
		 */
		public boolean isStorage() {
			return storage;
		}
		/**
		 * @param storage The storage to set.
		 */
		public void setStorage(boolean storage) {
			this.storage = storage;
		}
    }
    
	public static String getJDBCTypeName(int type)
	{
		switch (type)
		{
			case Types.ARRAY :
				return "ARRAY";
			case Types.BIGINT :
				return "BIGINT";
			case Types.BINARY :
				return "BINARY";
			case Types.BIT :
				return "BIT";
			case Types.BLOB :
				return "BLOB";
			case DbType.BOOLEAN :
				return "BOOLEAN";
			case Types.CHAR :
				return "CHAR";
			case Types.CLOB :
				return "CLOB";
			case Types.DATE :
				return "DATE";
			case Types.DECIMAL :
				return "DECIMAL";
			case Types.DISTINCT :
				return "DISTINCT";
			case Types.DOUBLE :
				return "DOUBLE";
			case Types.FLOAT :
				return "FLOAT";
			case Types.INTEGER :
				return "INTEGER";
			case Types.JAVA_OBJECT :
				return "JAVA_OBJECT";
			case Types.LONGVARBINARY :
				return "LONGVARBINARY";
			case Types.LONGVARCHAR :
				return "LONGVARCHAR";
			case Types.NULL :
				return "NULL";
			case Types.NUMERIC :
				return "NUMERIC";
			case Types.OTHER :
				return "OTHER";
			case Types.REAL :
				return "REAL";
			case Types.REF :
				return "REF";
			case Types.SMALLINT :
				return "SMALLINT";
			case Types.STRUCT :
				return "STRUCT";
			case Types.TIME :
				return "TIME";
			case Types.TIMESTAMP :
				return "TIMESTAMP";
			case Types.TINYINT :
				return "TINYINT";
			case Types.VARBINARY :
				return "VARBINARY";
			case Types.VARCHAR :
				return "VARCHAR";
			case DbType.ORACLE_ROWID :
				return "ROWID";
			default :
				return "";
		}

	}
	private static final int[] dataTypes = {
			Types.ARRAY,
			Types.BIGINT,
			Types.BINARY,
			Types.BIT,
			Types.BLOB,
			DbType.BOOLEAN,
			Types.CHAR,
			Types.CLOB,
			Types.DATE,
			Types.DECIMAL,
			Types.DISTINCT,
			Types.DOUBLE,
			Types.FLOAT,
			Types.INTEGER,
			Types.JAVA_OBJECT,
			Types.LONGVARBINARY,
			Types.LONGVARCHAR,
			Types.NULL,
			Types.NUMERIC,
			Types.OTHER,
			Types.REAL,
			Types.REF,
			DbType.ORACLE_ROWID,
			Types.SMALLINT,
			Types.STRUCT,
			Types.TIME,
			Types.TIMESTAMP,
			Types.TINYINT,
			Types.VARBINARY,
			Types.VARCHAR
	};
	
	/* Alphabetically sorted */
	private static final String[] dataTypeStrings =
		{
			"ARRAY",
			"BIGINT",
			"BINARY",
			"BIT",
			"BLOB",
			"BOOLEAN",        
			"CHAR",
			"CLOB",
			"DATE",
			"DECIMAL",
			"DISTINCT",
			"DOUBLE",
			"FLOAT",
			"INTEGER",
			"JAVA_OBJECT",
			"LONGVARBINARY",
			"LONGVARCHAR",
			"NULL",
			"NUMERIC",
			"OTHER",
			"REAL",
			"REF",
			"ROWID",
			"SMALLINT",
			"STRUCT",
			"TIME",
			"TIMESTAMP",
			"TINYINT",
			"VARBINARY",
			"VARCHAR"
	};

	public static int parseJDBCType(String type)
	{
		int index = Arrays.binarySearch(dataTypeStrings, type);
		if (index < 0)
			throw new RuntimeException("The " + type + " JDBC type used in tableSpec.xml is unknown.");
		return dataTypes[index];
	}
}
