/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.mapper;

import java.util.List;

import org.xquark.xml.xdbc.XMLDBCException;

/** Provide functions for manipulating XML collections and performing XQueries
 * on the XML repository to which it is connected.
 * <p>In case of multithread use, a RepositoryConnection must be created per
 * thread.</p>
 * @see org.xquark.mapper.AccessManager
 * @see org.xquark.mapper.XMLCollection
 */
public interface RepositoryConnection
extends org.xquark.xml.xdbc.XMLConnection
{
    ///////////////////////////////////////////////////////////////////////////
    //                          API Constants
    ///////////////////////////////////////////////////////////////////////////
    
    public static final String TEXT_LENGTH_PROPERTY
        = "http://www.xquark.org/repository/collection/properties/maxTextDataLength";
    public static final String EXTRA_DATA_LENGTH_PROPERTY
        = "http://www.xquark.org/repository/collection/properties/maxExtraDataLength";
    public static final String MAPPING_ID_PROPERTY
        = "http://www.xquark.org/repository/collection/properties/mappingId";
    public static final String DESC_PROPERTY
        = "http://www.xquark.org/repository/collection/properties/description";
    public static final String DOC_OID_SIZE_PROPERTY
        = "http://www.xquark.org/repository/collection/properties/docOIDSizeInBits";
    public static final String DOC_OID_PREALLOCATION_SIZE
        = "http://www.xquark.org/repository/collection/properties/docOIDPreallocationSize";
    
    /////////////////////////
    //      FEATURES
    /////////////////////////
    public static final String USE_SCHEMA_PREFIXES_FEATURE
        = "http://www.xquark.org/repository/collection/features/storePrefixes";

    //
    // SYSTEM COLLECTIONS
    //
    /** The name of the system repository used to store XML Schemas */
    public static final String SCHEMA_COLLECTION = "SCHEMA$";
    /** The name of the system repository used to store mapping files */
    public static final  String MAPPING_COLLECTION = "MAPPING$";

    ///////////////////////////////////////////////////////////////////////////
    //                      CONNECTION FUNCTIONS
    ///////////////////////////////////////////////////////////////////////////
    
    /** Remove <b>XMLCollection</b> configuration from the underlying database,
     * but not the collection tables. All collections must have been previously
     * destroyed before this function is able to work.
     * <B>Warning !</B> {@link #close() close()} is automatically called and
     * another connection is to be opened.
     * @throws RepositoryException API exception.
     * @see AccessManager
     */
    public void deleteConfiguration() throws XMLDBCException;
    
    /** Remove the whole repository from the underlying database, including the
     * the XML collections.
     * <B>Warning !</B> {@link #close() close()} is automatically called and
     * another connection is to be opened.
     * @throws RepositoryException API exception.
     * @see AccessManager
     */
    public void deleteRepository() throws XMLDBCException;
    
    /** Remove the whole repository from the underlying database, including the
     * the XML collections but a new one is created so that the connection is
     * still alive.
     * @throws RepositoryException API exception.
     * @see AccessManager
     */
    public void resetRepository() throws XMLDBCException;
    
    /** Open the system collection where XML schemas are stored.
     * <B>You cannot use the DOM methods to store schemas with includes</B>
     * @return an {@link XMLCollection XMLCollection} object.
     * @throws RepositoryException API exception.
     * @see RepositoryCollection
     */
    public RepositoryCollection getSchemaCollection() throws XMLDBCException;
    
    /** Open the system collection where mapping files are stored.
     * @param name the string identifier of the collection.
     * @return an {@link XMLCollection XMLCollection} object.
     * @throws RepositoryException API exception.
     * @see RepositoryCollection
     */    
    public RepositoryCollection getMappingCollection() throws XMLDBCException;
    
    /** Open an existing user collection. Compared to the XDBC
     * {@link XMLConnection#getCollection(String) getCollection()} method, this 
     * method avoids the cast operation
     * @return a {@link RepositoryCollection XMLCollection} object.
     * @throws RepositoryException API exception.
     */    
    public RepositoryCollection getRepositoryCollection(String name) throws XMLDBCException;
    
    /** Set the interactive mode for queries executed with this object. The
     * interactive mode supplies more performance on selective queries. When
     * the interactive mode is off, the queries are executed in batch mode with
     * is faster for queries retrieving many results.
     * @param interactive if true, the interactive mode is enabled and used for
     * the following query executions.
     */
    public void setInteractiveMode(boolean interactive);
    
    /** Get the current interactive mode used for queries executed with this
     * object.
     * @return true if the interactive mode is on. False if the batch is on.
     */
    public boolean getInteractiveMode();
    
    ///////////////////////////////////////////////////////////////////////////
    //                          DEPENDENCES
    ///////////////////////////////////////////////////////////////////////////
    
    /** Get the list of XML collection Ithat are currently using a given
     * mapping.
     * @param mappingID the String mapping ID
     * @return a List of String collection IDs.
     */
    public List getCollectionsUsingMapping(String mappingID)
    throws XMLDBCException;
    
    /** Get the list of XML collection that are currently using a given
     * namespace.
     * @param namespace the namespace URI
     * @return a List of String collection IDs.
     */
    public List getCollectionsUsingNamespace(String namespace)
    throws XMLDBCException;

    /** Get the list of mappings that are currently using a given
     * namespace.
     * @param namespace the namespace URI
     * @return a List of String mapping IDs.
     */
    public List getMappingsUsingNamespace(String namespace)
    throws XMLDBCException;
}


