/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.mapper;

import java.io.PrintWriter;
import java.sql.SQLException;
import java.util.HashMap;

import javax.sql.DataSource;

import org.xquark.mapper.dbms.AbstractConnection;
import org.xquark.mapper.dbms.AbstractConnectionFactory;
import org.xquark.mapper.dbms.RepositoryConnectionImpl;
import org.xquark.mapper.metadata.Repository;
import org.xquark.mapper.util.CounterMap;
import org.xquark.mapper.util.DestructionToken;
import org.xquark.xml.xdbc.XMLConnection;
import org.xquark.xml.xdbc.XMLDBCException;
import org.xquark.xml.xdbc.XMLDataSource;

/**
 * XMLDBC data source implementation
 */
public final class RepositoryDataSource implements XMLDataSource
{
    private static final String RCSRevision = "$Revision: 1.2 $";
    private static final String RCSName     = "$Name: XQ1_1_1 $";
    
    private static final String ORACLE_URL_PREFIX = "jdbc:oracle";
    private static final String TIMESTEN_URL_PREFIX = "jdbc:timesten";
    private static final String SYBASE_URL_PREFIX = "jdbc:sybase";
    private static final String SQLSERVER_URL_PREFIX = "jdbc:microsoft:sqlserver";
    
    private static final String ORACLE_DRIVER_CLASS = "oracle.jdbc.driver.OracleDriver";
    private static final String TIMESTEN_DRIVER_CLASS = "com.timesten.jdbc.TimesTenDriver";
    private static final String SYBASE_DRIVER_CLASS = "com.sybase.jdbc2.jdbc.SybDriver";
    private static final String SQLSERVER_DRIVER_CLASS = "com.microsoft.jdbc.sqlserver.SQLServerDriver";

	private static final String DATA_SOURCE_FAKE_NAME = "DS";
    
    private CounterMap repositories = new CounterMap();
    private HashMap standByRepositories = new HashMap();
    
    private AbstractConnectionFactory connectionFactory;
    private DataSource JDBCDs;
    
    private PrintWriter logger;
    private int logTimeout;
    
    /** Constructor to use for later calls to getConnection(user, password)
     * methods.
     * @param url a valid JDBC url to a relational database.
     * @throws XMLDBCException if URL provided is not reachable.
     */    
    public RepositoryDataSource(String url)
    throws XMLDBCException
    {
        acceptsJDBCURL(url);
        connectionFactory = new AbstractConnectionFactory(url);
    }
    
    /** Constructor to use for later calls to getConnection()
     * methods. This method allows using JDBC DataSources using a pool
     * of connections.
     * @param ds A JDBC {@link javax.sql.DataSource} object.
     */    
    public RepositoryDataSource(DataSource ds)
    {
        JDBCDs = ds;
    }
    
    private boolean acceptsJDBCURL(String url) throws XMLDBCException
    {
        try {
            // automatic load for know drivers
            if (url.startsWith(ORACLE_URL_PREFIX))
                Class.forName(ORACLE_DRIVER_CLASS);
            else if (url.startsWith(TIMESTEN_URL_PREFIX))
                Class.forName(TIMESTEN_DRIVER_CLASS);
            else if (url.startsWith(SYBASE_URL_PREFIX))
                Class.forName(SYBASE_DRIVER_CLASS);
            else if (url.startsWith(SQLSERVER_URL_PREFIX))
                Class.forName(SQLSERVER_DRIVER_CLASS);
            // else driver is supposed to be loaded
        }
        catch (Exception e)
        {
            throw new XMLDBCException("JDBC driver could not be found. Check your classpath or load it yourself if it is not supported.");
        }
             
        return true;
    }
    
    /** Retrieves the log writer currently used by the XDataSource object to print messages.
     * @return the java.io.PrintWriter object used to log data source messages, or null if none is defined
     */
    public java.io.PrintWriter getLogWriter()
    {
        return logger;
    }
    
    /** Sets the logging/tracing java.io.PrintWriter object that is used by the XDataSource object.
     * @param writer the new logging/tracing java.io.PrintWriter object, or null to disable logging and tracing
     */
    public void setLogWriter(PrintWriter writer)
    {
        logger = writer;
    }
    
    /** Sets the maximum time in seconds that the object will wait when attempting to log in to a data source. Setting the timeout to 0 will disable it.
     * @param seconds the login time limit in seconds
     */
    public void setLoginTimeout(int seconds)
    {
        logTimeout = seconds;
    }
    
    /** Gets the maximum time in seconds that the object can wait when attempting to log in to a data source
     * @return the data source login timeout in seconds, 0 if no timeout is defined
     */
    public int getLoginTimeout()
    {
        return logTimeout;
    }
    
    /** Attempts to establish a connection with the Repository specified by
     * the URL given to the constructor and the user parameter.
     * @param user the user owning the repository
     * @param password the user's password
     * @throws XMLDBCException If a data source access error occurs
     * @return an XMLConnection object that represents a connection to the
     * repository data source
     */
    public XMLConnection getConnection(String user, String password) throws XMLDBCException
    {
        return getRepositoryConnection(user, password);
    }
    
    /** The same as {@link #getConnection(String,String)} method but
     * returning a RepositoryConnection object, giving access to the
     * Repository extended API.
     * @param user the user owning the repository
     * @param password the user's password
     * @throws XMLDBCException If a data source access error occurs
     * @return an RepositoryConnection object that represents a connection to
     * the repository.
     */    
    public RepositoryConnection getRepositoryConnection(String user, String password)
    throws XMLDBCException
    {
        String key = user.toUpperCase();

        Repository rep = (Repository)repositories.checkOut(key);
        
        try {
            if (rep == null)
            {
                // look in the "stand bys"
                rep = (Repository)standByRepositories.get(key);
                if (rep == null)
                    rep = new Repository(getAbstractConnection(user, password));
                else
                    rep.restoreConnection(getAbstractConnection(user, password));
                repositories.put(key, rep);
            }
            
            return new RepositoryConnectionImpl(
                                            rep,
                                            getAbstractConnection(user, password),
                                            new ConnectionDestructor()
                                            );
        }
        catch (SQLException e) {
            throw new RepositoryException(RepositoryException.DB_ERROR,
            "JDBC error while creating a RepositoryConnection.", e);
        }

    }
    
    /** Attempts to establish a connection with a repository using the JDBC data
     * source specified in the constructor.
     * @throws XMLDBCException If a data source access error occurs
     * @return an XMLConnection object that represents a connection to the
     * repository data source
     */
    public XMLConnection getConnection() throws XMLDBCException
    {
        return getRepositoryConnection();
    }
    
    /** The same as {@link #getConnection()} method but returning a
     * RepositoryConnection object, giving access to the Repository extended API.
     * @throws XMLDBCException If a data source access error occurs
     * @return an RepositoryConnection object that represents a connection to
     * the repository.
     */
    public RepositoryConnection getRepositoryConnection() throws XMLDBCException
    {
        Repository rep = (Repository)repositories.checkOut(DATA_SOURCE_FAKE_NAME);
        
        try
        {
            if (rep == null)
            {
                // look in the "stand bys"
                rep = (Repository)standByRepositories.get(DATA_SOURCE_FAKE_NAME);
                if (rep == null)
                    rep = new Repository(getAbstractConnection());
                else
                    rep.restoreConnection(getAbstractConnection());
                repositories.put(DATA_SOURCE_FAKE_NAME, rep);
            }
            
            return new RepositoryConnectionImpl(
            rep,
            getAbstractConnection(),
            new DSConnectionDestructor()
            );
        }
        catch (SQLException e)
        {
            throw new RepositoryException(RepositoryException.DB_ERROR,
            "JDBC error while creating a RepositoryConnection.", e);
        }
    }
    
    //////////////////////////////////////////////////////////////
    //                      PRIVATE
    //////////////////////////////////////////////////////////////
    private AbstractConnection getAbstractConnection() throws SQLException
    {
        if (JDBCDs == null)
            return connectionFactory.getAbstractConnection();
        else
            return AbstractConnectionFactory.newConnection(JDBCDs.getConnection());
    }
    
    private AbstractConnection getAbstractConnection(String user, String password)
    throws SQLException
    {
        if (JDBCDs == null)
            return connectionFactory.getAbstractConnection(user, password);
        else
            return AbstractConnectionFactory.newConnection(
                                            JDBCDs.getConnection(user, password)
                                            );
    }
    
    private void releaseConnection(String user)
    {
            Repository rep = (Repository)repositories.remove(user);
            
            try
            {
                if (rep != null)
                {
                    if (rep.isAlive())  // physical repository may have been destroyed
                    {
                        rep.withdrawConnection();   // put repository metadata in standby
                        standByRepositories.put(user, rep);
                    }
                    else
                        rep.close();
                }
            }
            catch (XMLDBCException e)
            {
                // ignore
            }
    }
    
    private class ConnectionDestructor implements DestructionToken
    {
        public void destruct(Object o)
        {
            try
            {
                releaseConnection(((RepositoryConnection)o).getUserName());
            }
            catch (XMLDBCException e)
            {
                // ignore
            }
        }
    }
    
    private class DSConnectionDestructor implements DestructionToken
    {
        public void destruct(Object o)
        {
            releaseConnection(DATA_SOURCE_FAKE_NAME);
        }
    }
}
