/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

/*
 * StorageException.java
 *
 * Created on 18 avril 2001, 13:05
 */

package org.xquark.mapper;

import org.xml.sax.Locator;

/**
 * This exception replaces RepositoryException at storage time when XML document
 * location information is available.
 *
 */
public class StorageException extends RepositoryException implements Locator
{
private static final String RCSRevision = "$Revision: 1.2 $"; 
private static final String RCSName     = "$Name: XQ1_1_1 $";
    private int line;
    private int column;
    private String systemID;
    private String publicID;
    
    /**
     * Constructor.
     * @param code the code featuring the kind of error
     * (see RepositoryException constants).
     * @param message A litteral message bringing precisions on the error.
     */
    public StorageException(int code, String message)
    {
        super(code, message);
    }
    
    /**
     * Constructor.
     * @param exception base RepositoryException.
     */
    public StorageException(RepositoryException exception)
    {
        super(exception.getCode(), exception.getMessage(), exception.getException());
    }
    
    /**
     * Constructor.
     * @param exception base RepositoryException.
     * @param locator SAX locator for identifying last parsing position when
     * exception occured.
     */
    public StorageException(RepositoryException exception, Locator locator)
    {
        this(exception);
        setLocator(locator);
    }
    
    /** Constructor for application exceptions encapsulating an underlying exception.
     * @param code the code featuring the kind of error
     * (see RepositoryException constants).
     * @param message A litteral message bringing precisions on the error.
     * @param underlyingException the original Exception that carried out this
     * exception.
     */
    public StorageException(int code, String message, Exception underlyingException)
    {
        super(code, message, underlyingException);
    }
    /** Constructor for application exceptions encapsulating an underlying exception.
     * @param code the code featuring the kind of error
     * (see RepositoryException constants).
     * @param message A litteral message bringing precisions on the error.
     * @param underlyingException the original Exception that carried out this
     * exception.
     * @param locator SAX locator for identifying last parsing position when
     * exception occured.
     */
    public StorageException(int code, String message, Exception underlyingException, Locator locator)
    {
        super(code, message, underlyingException);
        setLocator(locator);
    }
    
    /** Set parsing document locator for exception.
     * @param locator SAX locator for identifying last parsing position when
     * exception occured.
     */
    public void setLocator(Locator locator)
    {
        line = locator.getLineNumber();
        column = locator.getColumnNumber();
        systemID = locator.getSystemId();
        publicID = locator.getPublicId();
    }
    
    /**
     * Return the public identifier for the parsed document.
     *
     * <p>The return value is the public identifier of the document
     * entity or of the external parsed entity in which the markup
     * triggering the exception appears.</p>
     *
     * @return A string containing the public identifier, or
     *        null if none is available.
     * @see #getSystemId
     */
    public String getPublicId()
    {
        return publicID;
    }
    
    /**
     * Return the system identifier for the parsed document.
     *
     * <p>The return value is the system identifier of the document
     * entity or of the external parsed entity in which the markup
     * triggering the exception appears.</p>
     *
     * @return A string containing the system identifier, or null
     *        if none is available.
     * @see #getPublicId
     */
    public String getSystemId()
    {
        return systemID;
    }
    
    /**
     * Return the line number where the exception occurs.
     *
     * <p><strong>Warning:</strong> The return value from the method
     * is intended only as an approximation for the sake of error
     * reporting; it is not intended to provide sufficient information
     * to edit the character content of the original XML document.</p>
     *
     * <p>The return value is an approximation of the line number
     * in the document entity or external parsed entity where the
     * markup triggering the exception appears.</p>
     *
     * @return The line number, or -1 if none is available.
     * @see #getColumnNumber
     */
    public int getLineNumber()
    {
        return line;
    }
    
    /**
     * Return the column number where the exception occurs.
     *
     * <p><strong>Warning:</strong> The return value from the method
     * is intended only as an approximation for the sake of error
     * reporting; it is not intended to provide sufficient information
     * to edit the character content of the original XML document.</p>
     *
     * <p>The return value is an approximation of the column number
     * in the document entity or external parsed entity where the
     * markup triggering the event appears.</p>
     *
     * @return The column number, or -1 if none is available.
     * @see #getLineNumber
     */
    public int getColumnNumber()
    {
        return column;
    }
    
}
