/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

/*
 * AbstractConnection.java
 *
 * Created on 9 janvier 2001, 18:23
 */

package org.xquark.mapper.dbms;

import java.sql.*;
import java.util.Collection;
import java.util.List;
import java.util.Map;

import org.xquark.jdbc.typing.DbType;
import org.xquark.jdbc.typing.TypeMap;
import org.xquark.mapper.RepositoryException;
import org.xquark.mapper.metadata.Repository;

/**
 * Interface offering vendor-independant database methods.
 *
 * <p>Objects implementing this interface are supposed to own a JDBC
 * connection and use it for all accesses to dbms.</p>
 */
public interface AbstractConnection
{

    ///////////////////////////////////////////////////////////////
    //                      Accessors
    ///////////////////////////////////////////////////////////////
    /** returns the encapsulated JDBC connection */
    public Connection getConnection();

    /** returns the database type.
     * @return a constant defined in {@link AbstractConnectionFactory}
     */
    public short getDBMSType();
    
	/** returns the current schema name (when the connection was created).
	 * @return the currentschema name as returned by the dbms 
	 */
	public String getSchemaName();
	
   /** closes the encapsulated JDBC connection and perform ending tasks */
    public void close() throws SQLException;

    /** returns the type map for the connection dbms */
    public TypeMap getTypeMap();
    
	/** Says if table and column names must be enclosed in double quotes */
	public boolean useDoubleQuotes4DDLNames();
	
    ///////////////////////////////////////////////////////////////
    //                      Specific methods
    ///////////////////////////////////////////////////////////////
    /** This method is called when a repository is initialized to perform
     * dbms preparation.
     */
    public void onInitRepository(Repository rep) throws SQLException;
    
    /** This method is called when a repository is deleted to restore initial
     * dbms state.
     */
    public void onDeleteRepository(Repository rep) throws SQLException;
    
    /** This method is called when a repositoryConnection is opened to perform
     * thread specific operations.
     */
    public void onInitConnection() throws SQLException;
    
    ///////////////////////////////////////////////////////////////
    //                      Connection methods
    ///////////////////////////////////////////////////////////////
    /** This method should be called before processing an updating user request.
     * This method actually saves the user specified auto-commit mode and
     * turns the auto-commit mode of the current connection to off.
     * @throws RepositoryException Application exception
     * @see #commit
     */
    public void start() throws RepositoryException;
    
    /** calls  the encapsulated JDBC connection rollback method */
    public void rollback() throws SQLException;
    
    /** Before processing a user request that updates the database, the
     * auto-commit mode of the current connection is turned off to allow all
     * the changes to be committed or aborted together. The initial
     * auto-commit mode of the connection is backed up so it can be
     * restored after the user request has been processed.
     *
     * This method is called upon successful completion of the user request.
     * This method actually commits the database operations assuming the
     * connection mode is off and restores the user auto-commit mode.
     *
     * If the initial auto-commit mode is false, this method restores the
     * auto-commit mode and leaves the database operation uncommitted.
     * @see #start
     * @throws RepositoryException Application exception
     */
    public void commit() throws RepositoryException;
    
    ///////////////////////////////////////////////////////////////
    //                      Table creation methods
    ///////////////////////////////////////////////////////////////
    /** Create a temporary relational table which name is automatically
     * generated according to the single result query table specs.
     * @return the features of the table needed especially for destruction
     * @throws SQLException If a database error occurs while creation
     */
    public TableInfo createQueryTemporaryTable() throws SQLException;
    
    /** Create a temporary relational table which name is automatically
     * generated.
     * @param query a query that allow to specify and fill the table.
     * @return the name of the table needed especially for destruction
     * @throws SQLException If a database error occurs while creation
     */
    public String createTemporaryTableAs(String subQuery) throws SQLException;
    
    /** Generate a random temporary relational table name WHICH IS NOT
     * GUARANTEED TO BE UNIQUE.
     * @return the name of the table.
     */
    public String getTemporaryTableName();
    
    /** Generate a random temporary relational table name WHICH IS NOT
     * GUARANTEED TO BE UNIQUE.
     * @return an object containing features of the table.
     */
    public TableInfo getTemporaryTableInfo();

    /** Delete a temporary relational.
     * @param type The table name
     * @return the name of the table needed especially for destruction
     * @throws SQLException If a database error occurs while creation
     */
    public void dropTemporaryTable(String tableName) throws SQLException;

    /** Drop all temporary relational tables created by the
     * createTemporaryTable() methods.
     * @throws SQLException If a database error occurs while creation
     */
    public void dropTemporaryTables() throws SQLException;
    
    /** Delete a relational table with a particular name.
     * @param type The table name
     * @return the name of the table needed especially for destruction
     * @throws SQLException If a database error occurs while creation
     */
    public void dropTable(String tableName) throws SQLException;

    /** Retrieves org.xquark.mapper.mapping.ColumnMetadata from the
     * underlying database.
     * @param tableName The table name
     * @throws SQLException If a database error occurs while creation
     */
    public Map getTableMetadata(String tableName) throws SQLException;
    
    ///////////////////////////////////////////////////////////////
    //                      Statements methods
    ///////////////////////////////////////////////////////////////
    /** Add a SQL statement in a batch for delayed execution.
     * @param stmt the JDBC statement which the SQL statement is to be added to
     */    
    public void addBatch(PreparedStatement stmt) throws SQLException;
    
    /** Add a SQL statement in a batch for delayed execution.
     * @param stmt the JDBC statement which the SQL statement is to be added to
     * @param SQLStatement the SQL statement that is to be added
     */    
    public void addBatch(Statement stmt, String SQLStatement) throws SQLException;
    
    /** Execute a batch created with the addBatch method.
     * @param stmt the JDBC statement wich the SQL statement is to be added to
     * @return an array of update counts containing one element for each command
     * in the batch. The elements of the array are ordered according to the
     * order in which commands were added to the batch.
     */    
    public int[] executeBatch(Statement stmt) throws SQLException;

    /** Gives the JDBC driver a hint as to the number of rows that should be fetched
     * from the database when more rows are needed.
     * @param stmt the JDBC statement wich the SQL statement is to be added to
     * @param rows  the number of rows to fetch
     */    
    public void setFetchSize(Statement stmt, int rows) throws SQLException;

    /** Verify the table exists.
     * @param name table name to check
     * @return true if the table exists.
     */
    public boolean checkTable(String name) throws SQLException;
    
    /** Verify the table exists.
     * @param table table which existence is to check
     * @return true if the table exists.
     */
    public boolean checkTable(TableInfo table) throws SQLException;
    
    /** Delete rows in the reconstruction temporary table name.
     */
    public void emptyTable(String tableName) throws SQLException;
    ///////////////////////////////////////////////////////////////
    //           Complex SQL statements (recursive, subqueries...)
    ///////////////////////////////////////////////////////////////
    
    public String getDeleteUserTableRowsStatement(String deletionStmt, String documentRowsSelectionClause)
    throws SQLException, RepositoryException;
    
    public void executeRangeUpdate(String statement, long first, long last)
    throws SQLException, RepositoryException;
    
    ///////////////////////////////////////////////////////////////
    //                      DBMS specificities
    ///////////////////////////////////////////////////////////////
    /**
     * Return the column specification conversion string specific to the
     * rdbms corresponding to an XML Schema primitive type, especially for 
     * date conversions (to_date() function).
     * @param type the XML schema primitive type
     */
    public String getColumnStatement(int type) throws RepositoryException;
    ///////////////////////////////////////////////////////////////
    //             UNIVERSAL ERROR CODES
    ///////////////////////////////////////////////////////////////
    public int UNKNOW_ERROR_CODE = 0;
	public int OBJECT_ALREADY_EXISTS  = 1;
	public int OBJECT_DOES_NOT_EXISTS  = 2;
    
    public int getUniversalError(SQLException e);
    
    ///////////////////////////////////////////////////////////////
    //                             OTHERS
    ///////////////////////////////////////////////////////////////
    public void updateStatistics(String tableName) throws SQLException;
    
    public int getMaxDataLength();
    public int getSystemCollectionDataLength();
    public boolean useStringDelimitor();
    public boolean distinguishNullAndEmptyStrings();
    
	public String getTemporaryTableCreationStatement(TableInfo table);
    
	public String getIndexOrganizedTableCreationStatement(TableInfo table);
	
	public String getLockSelectStatement(String columnList, String tableName, String filterClause);
	public String getLockSelectStatement(TableInfo table, String filterClause);
	
    public String getBitmapIndexClause();
    
    public String getHashIndexClause();
    
    public String getBTreeIndexClause();
    
    public void executeUpdate(String stmt) throws SQLException;
    
    public void createSequence(TableInfo table, short step) throws SQLException;
    public void dropSequence(String seqName) throws SQLException;
    
    public long nextSequenceValue(TableInfo table, short step) throws RepositoryException;
    
    /**
     * Create table and indexes.
     *
     * @param Collection collection of TableInfo objects.
     * @return a collection with tables not created (sequences for instance).
     */
    public Collection createTables(Collection tables) throws SQLException;
    
    public void createIndexes(Collection tables) throws SQLException;
    
    public void dropIndexes(Collection tables) throws SQLException;
    
    /** Gets a list of table names for all tables (any type) of the connection
     * schema and default catalog.
     * @param tablePattern a JDBC/SQL style regexp.
     * @return a List of String objects.
     */
    public List getUserTableNames(String tablePattern) throws SQLException;
    public List getUserSequenceNames(String seqPattern) throws SQLException;

    public void setObject(PreparedStatement pStmt, int index, Object o, int javaType, DbType sqlType)
    throws SQLException;
}
