/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.mapper.dbms;

import java.sql.*;

import org.xquark.jdbc.typing.DBMSConstants;
import org.xquark.jdbc.typing.DBMSInfoMap;
import org.xquark.mapper.dbms.microsoft.SQS2000Connection;
import org.xquark.mapper.dbms.mysql.MySQL323Connection;
import org.xquark.mapper.dbms.oracle.O8iConnection;
import org.xquark.mapper.dbms.sybase.ASE119Connection;
import org.xquark.mapper.dbms.timesten.TT40Connection;


/**
 * AbstractConnection factory.
 *
 * <p>This object holds database vendor specificities routing.</p>
 */
public class AbstractConnectionFactory implements DBMSConstants
{
    private static final String RCSRevision = "$Revision: 1.3 $";
    private static final String RCSName     = "$Name: XQ1_1_1 $";

    ///////////////////////////////////////////////////////////////
    //                      Data
    ///////////////////////////////////////////////////////////////
    private String URL;
    private short databaseVendor = -1;


    ///////////////////////////////////////////////////////////////
    //                      Contructor
    ///////////////////////////////////////////////////////////////
    public AbstractConnectionFactory(String URL)
    {
        this.URL = URL;
    }
    
    ///////////////////////////////////////////////////////////////
    //                      Public Methods
    ///////////////////////////////////////////////////////////////
    public String getURL() { return URL;}
    
    /** Get the constant featuring the database vendor and version.
     **/
    public short getDatabaseType()
    {
        return databaseVendor;
    }

    public AbstractConnection getAbstractConnection(String user, String password)
    throws SQLException
    {
        return getAbstractConnection(newConnection(user, password), databaseVendor);
    }
    
    public AbstractConnection getAbstractConnection()
    throws SQLException
    {
        return getAbstractConnection(newConnection(null, null), databaseVendor);
    }
    
    public static AbstractConnection newConnection(Connection conn) throws SQLException
    {
    	return newConnection(conn, false);
    }
    
    public static AbstractConnection newConnection(Connection conn, boolean preserveCommit) throws SQLException
    {
        /* memorize database vendor */
        short databaseVendor = DBMSInfoMap.getDatabaseVendor(conn);
        
        /* Set Auto-commit mode */
        if (!preserveCommit)
        	conn.setAutoCommit(true);
        
        /* Set Transaction isolation */
        setTransactionIsolation(conn, databaseVendor);
        
        return getAbstractConnection(conn, databaseVendor);
    }

    public String toString() 
    {
        return URL;
    }

    ///////////////////////////////////////////////////////////////
    //                      Private utilities
    ///////////////////////////////////////////////////////////////
    private static AbstractConnection getAbstractConnection(Connection conn, short databaseVendor) throws SQLException
    {
        Connection tmp = null;
        
        String url = conn.getMetaData().getURL();
        
        /* Create AbstractConnection */
        AbstractConnection ret;
        switch (databaseVendor)
        {
            case ORACLE8I : case ORACLE8 : case ORACLE :
                ret = new O8iConnection(conn, ORACLE, DBMSInfoMap.getDBMSInfo(ORACLE));
                break;
			case MYSQL : case MYSQL323 :
				ret = new MySQL323Connection(conn, MYSQL, DBMSInfoMap.getDBMSInfo(MYSQL));
				break;
            case TIMESTEN40 : case TIMESTEN :
                ret = new TT40Connection(conn, TIMESTEN, DBMSInfoMap.getDBMSInfo(TIMESTEN));
                break;
            case SYBASE_ASE_11_9 : case SYBASE :
                ret = new ASE119Connection(conn, SYBASE, DBMSInfoMap.getDBMSInfo(SYBASE));
                break;
            case SQL_SERVER_2000 : case SQL_SERVER :
                ret = new SQS2000Connection(conn, SQL_SERVER, DBMSInfoMap.getDBMSInfo(SQL_SERVER));
                break;
            case JDBC1 :
                ret = new JDBC1Connection(conn, databaseVendor);
                break;
            case JDBC2 :
                ret = new JDBC2Connection(conn, databaseVendor);
                break;
            default :
                throw new SQLException("Your RDBMS is not supported in the current version of the repository (basing on your JDBC driver informations).");
        }
        ret.onInitConnection();
        return ret;
    }

    private boolean isConnectionOK(Connection conn)
    {
        Statement testStmt = null;
        try {
            if (!conn.isClosed())
            {  // Try to createStatement to see if it's really alive
                testStmt = conn.createStatement();
                testStmt.close();
            }
            else
                return false;
        }
        catch (SQLException e) {
            if (testStmt != null)
            {
                try {
                    testStmt.close();
                }
                catch (SQLException se) {
                    // no op
                }
            }
            return false;
        }
        return true;
    }

    private Connection newConnection(String user, String password)
    throws SQLException
    {
        Connection conn = null;
        //Debug.print(this, "Opening JDBC connection");
        if (user == null)
            conn = DriverManager.getConnection(URL);
        else
            conn = DriverManager.getConnection(URL, user, password);

        /* memorize database vendor */
        if (databaseVendor == -1)
            databaseVendor = DBMSInfoMap.getDatabaseVendor(conn);

        /* Set Auto-commit mode */
        conn.setAutoCommit(true);   // Normally useless : default

        /* Set Transaction isolation */
        setTransactionIsolation(conn, databaseVendor);

        return conn;
    }

    /** Set the connection transaction isolation level according to vendor
     */
    private static void setTransactionIsolation(Connection conn, short databaseVendor) throws SQLException
    {
        switch (databaseVendor)
        {
            case TIMESTEN40 :
                conn.setTransactionIsolation(4096); // com.timesten.sql.TimesTenConnection.TRANSACTION_CURSOR_STABILITY(TRANSACTION_READ_COMMITTED has important limitations with JDBC)
                break;
            default :
                if (conn.getMetaData().supportsTransactionIsolationLevel(Connection.TRANSACTION_READ_COMMITTED))
                    conn.setTransactionIsolation(Connection.TRANSACTION_READ_COMMITTED);
                else
                    throw new SQLException("Your RDBMS does not support the needed transaction isolation level (TRANSACTION_READ_COMMITTED).");
        }
    }
}
