/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

/*
 * ColumnSpec.java
 *
 * Created on 11 janvier 2001, 16:57
 */

package org.xquark.mapper.dbms;

import java.sql.Types;

import org.xquark.jdbc.typing.ColumnMetaDataImpl;

/**
 * Class containing information on collections table columns  constructed from
 * repository configuration (no JDBC Access).
 *
 */
public class ColumnInfo
{
    private static final String RCSRevision = "$Revision: 1.3 $";
    private static final String RCSName     = "$Name: XQ1_1_1 $";
    
    private ColumnSpec template = null;
    protected long size = -1;
    protected int prec = -1;
    protected int scale = -1;
    int realisticJDBCType;  // data type modified to fit dbms limitations in terms of size
    private StringBuffer columnClause = null;
    private StringBuffer constraintColumnClause = null;
    private StringBuffer compactColumnClause = null;
    
    //////////////////////////////////////////////////////////////////////////
    //                          CONSTRUCTORS
    //////////////////////////////////////////////////////////////////////////
    /** Creates new ColumnInfo 
     */
    public ColumnInfo(ColumnSpec template, long[] sizeParameters, AbstractConnection dbmsInfo)
    {
        this.template = template;
        size = template.getSize(sizeParameters);
        prec = template.getPrecision(sizeParameters);
        scale = template.getScale(sizeParameters);
        calculateIntegerPrecision();
        realisticJDBCType = ColumnMetaDataImpl.findRealisticType(getDataType(), size, dbmsInfo.getTypeMap());
        initializeSatements(dbmsInfo);
    }
    
    //////////////////////////////////////////////////////////////////////////
    //                              ACCESSORS
    //////////////////////////////////////////////////////////////////////////
    
    public long getSize()
    {
        return size;
    }
    
    public int getPrecision()
    {
        return prec;
    }
    
    public int getScale()
    {
        return scale;
    }
    
    public String getColumnClause()
    {
        return columnClause.toString();
    }
    public String getConstraintColumnClause()
    {
        return constraintColumnClause.toString();
    }
    public String getCompactColumnClause()
    {
        return compactColumnClause.toString();
    }
    
    public String getName()
    {
        return template.getName();
    }
    
    public int getDataType()
    {
        return template.getDataType();
    }
    
    /** Return true if the columns belongs to the primary key */
    public boolean isPrimaryKey()
    {
        return template.isPrimaryKey();
    }
    
    public boolean isNullable()
    {
        return template.isNullable();
    }
    
    //////////////////////////////////////////////////////////////////////////
    //                      PUBLIC METHODS
    //////////////////////////////////////////////////////////////////////////
    private void initializeSatements(AbstractConnection dbmsInfo)
    {
        columnClause = new StringBuffer();
        constraintColumnClause = new StringBuffer();
        columnClause = generateCreationString(columnClause, dbmsInfo);
        compactColumnClause = new StringBuffer(columnClause.toString());
        
        if (isPrimaryKey())
        {
            constraintColumnClause.append(getName());
            constraintColumnClause.append(",");
            columnClause.append(" NOT NULL");
            compactColumnClause.append(" NOT NULL PRIMARY KEY");
        }
        else if (isNullable())
        {
            columnClause.append(" NULL");
            compactColumnClause.append(" NULL");
        }
        else
        {
            columnClause.append(" NOT NULL");
            compactColumnClause.append(" NOT NULL");
        }
    }
    public boolean isIndexable(int type, long size, AbstractConnection dbmsInfo)
    {
        return realisticJDBCType != Types.LONGVARBINARY
                && realisticJDBCType != Types.LONGVARCHAR
                && realisticJDBCType != Types.BLOB
                && realisticJDBCType != Types.CLOB;
    }
    
    //////////////////////////////////////////////////////////////////////////
    //                              PRIVATE
    //////////////////////////////////////////////////////////////////////////
    /* Compact method (PRIMARY KEY INLINE) */
    private StringBuffer generateCreationString(StringBuffer stmt, AbstractConnection dbmsInfo)
    {
        /* Column name */
        stmt.append(getName());
        
        /* Column type */
        
        // First set the correct JDBC type according to size issue specificities
        stmt.append(" ");
        stmt.append(ColumnMetaDataImpl.getTypeCreationString(
                                        realisticJDBCType,
                                        dbmsInfo.getTypeMap(),
                                        size,
                                        prec,
                                        scale
                                        ));
        
        return stmt;
    }
    
    
        private void calculateIntegerPrecision()
        {
            switch (getDataType())
            {
                case Types.DECIMAL: case Types.NUMERIC:
                    if (scale <= 0) // assume that precision was in bits !
                    prec = (int)Math.ceil(Math.log(Math.pow(2,prec))/Math.log(10));
                    break;
                case Types.BIGINT:
                    prec = 19;
                    break;
                case Types.INTEGER:
                    prec = 10;
                    break;
                case Types.SMALLINT:
                    prec = 5;
                    break;
                case Types.TINYINT:
                    prec = 3;
                    break;
                case Types.BIT:
                    prec = 1;
                    break;
                default:
            }
        }
    
}
