/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

/*
 * ColumnSpec.java
 *
 * Created on 11 janvier 2001, 16:57
 */

package org.xquark.mapper.dbms;

import org.xquark.jdbc.typing.ColumnMetaDataImpl;
import org.xquark.jdbc.typing.TypeMap;

/**
 * Class containing information for relational column creation.
 *
 */
public class ColumnSpec
{
    private static final String RCSRevision = "$Revision: 1.4 $";
    private static final String RCSName     = "$Name: XQ1_1_1 $";
    
    /* Column name */
    protected String name;
    protected int JDBCType;
    protected long size = -1;
    protected int prec = -1;
    protected int scale = -1;
    protected short paramSize = -1;
    protected short paramPrec = -1;
    protected short paramScale = -1;
    
    /* Is part of the primary key */
    protected boolean key = false;
    protected boolean nullable = true;

    //////////////////////////////////////////////////////////////////////////
    //                          CONSTRUCTORS
    //////////////////////////////////////////////////////////////////////////
    public ColumnSpec() {}
    
    /** Creates new ColumnSpec 
     * @param name The column name.
     * @param key true if the column belongs to the primary key.
     */
    public ColumnSpec(String name, int JDBCType, long size)
    {
        set(name, JDBCType, size);
    }
    
    /** Creates new ColumnSpec 
     * @param name The column name.
     * @param key true if the column belongs to the primary key.
     */
    public ColumnSpec(String name, int JDBCType, int prec, int scale)
    {
        set(name, JDBCType, prec, scale);
    }
    
    /** Creates new ColumnSpec 
     * @param name The column name.
     * @param key true if the column belongs to the primary key.
     */
    public ColumnSpec(String name, int JDBCType, boolean key)
    {
        set(name, JDBCType, key);
    }
    
    //////////////////////////////////////////////////////////////////////////
    //                              ACCESSORS
    //////////////////////////////////////////////////////////////////////////
    
    public void setName(String name)
    {
        this.name = name;
    }
    
    public void set(String name, int JDBCType, long size)
    {
        this.name = name;
        this.JDBCType = JDBCType;
        this.size = size;
    }
    
    public void set(String name, int JDBCType, int prec, int scale)
    {
        set(name, JDBCType, size);
        this.prec = prec;
        this.scale = scale;
    }
    
    /** Accessor.
     * <p>Used for object recycling.</p>
     * @param name The column name.
     * @param key true if the column belongs to the primary key.
     */
    public void set(String name, int JDBCType, boolean key)
    {
        this.name = name;
        this.JDBCType = JDBCType;
        setKeyRole(key);
    }
    
    public void setDataType(String type)
    {
        JDBCType = TypeMap.parseJDBCType(type);
    }
    
    /** Accessor.
     * @param key true if the column belongs to the primary key.
     */
    public void setKeyRole(boolean key)
    {
        this.key = key;
    }
    
    public void setSize(long size, short param)
    {
        this.size = size;
        paramSize = param;
    }
    
    public void setPrecision(int prec, short param)
    {
        this.prec = prec;
        paramPrec = param;
    }
    
    public void setScale(int scale, short param)
    {
        this.scale = scale;
        paramScale = param;
    }
    
    public void setNullable(boolean nullable)
    {
        this.nullable = nullable;
    }
    
    /** Return the column name */
    public String getName()
    {
        return name;
    }
    
    public int getDataType()
    {
        return JDBCType;
    }
    
    /** Return true if the columns belongs to the primary key */
    public boolean isPrimaryKey()
    {
        return key;
    }
    
    public boolean isNullable()
    {
        return nullable;
    }
    
    public long getSize(long[] sizeParameters)
    {
        if (paramSize >= 0 && sizeParameters[paramSize] >= 0)
            return sizeParameters[paramSize];
        else
            return size;
    }
    
    public int getPrecision(long[] sizeParameters)
    {
        if (paramPrec >= 0 && sizeParameters[paramPrec] >= 0)
            return (int)sizeParameters[paramPrec];
        else
            return prec;
    }
    
    public int getScale(long[] sizeParameters)
    {
        if (paramScale >= 0 && sizeParameters[paramScale] >= 0)
            return (int)sizeParameters[paramScale];
        else
            return scale;
    }
    
    /**
     * Use default values (no parameter).
     */
    public String getCreationString(TypeMap typeMap)
    {
        return name + ' ' + ColumnMetaDataImpl.getTypeCreationString(JDBCType, typeMap, size, prec,scale);
    }
    
    /**
     * Use default values (no parameter).
     */
    public String getTypeCreationString(TypeMap typeMap)
    {
        return ColumnMetaDataImpl.getTypeCreationString(JDBCType, typeMap, size, prec,scale);
    }
    
}
