/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.mapper.dbms;

import java.sql.*;
import java.util.*;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.xquark.jdbc.typing.*;
import org.xquark.mapper.RepositoryException;
import org.xquark.mapper.metadata.Repository;
import org.xquark.mapper.metadata.RepositoryConstants;
import org.xquark.schema.datatypes.PrimitiveType;

/**
 * Default implementation of AbstractConnection.
 *
 * <p>This class is abstract because some methods like table creation
 * are inevitably specific.</p>
 */
public abstract class DefaultConnectionImpl implements AbstractConnection, RepositoryConstants {
	private static final String RCSRevision = "$Revision: 1.6 $";
	private static final String RCSName = "$Name: XQ1_1_1 $";

    private static Log log = LogFactory.getLog(DefaultConnectionImpl.class);

    protected Connection connection;
	protected String schemaName;
	protected short dbms;
	protected DBMSInfo dbmsInfo;

	protected DefaultConnectionImpl(Connection connection, short dbms) throws SQLException {
		this(connection, dbms, new DBMSInfo());
		// load the JDBC standard type map...
		dbmsInfo.setTypeMap(new TypeMap());
		dbmsInfo.getTypeMap().load(connection);
	}

	protected DefaultConnectionImpl(Connection connection, short dbms, DBMSInfo dbmsInfo) throws SQLException {
		this.connection = connection;
		this.dbms = dbms;
		this.dbmsInfo = dbmsInfo;
		setSchemaName();
	}

	///////////////////////////////////////////////////////////////
	//                      Accessors
	///////////////////////////////////////////////////////////////
	protected void setSchemaName() throws SQLException {
		schemaName = connection.getMetaData().getUserName();
	}

	public String getSchemaName() {
		return schemaName;
	}

	/** returns the encapsulated JDBC connection
	 * @return a JDBC connection
	 */
	public Connection getConnection() {
		return connection;
	}

	/** returns the database type.
	 * @return a constant defined in {@link AbstractConnectionFactory}
	 */
	public short getDBMSType() {
		return dbms;
	}

	public TypeMap getTypeMap() {
		return dbmsInfo.getTypeMap();
	}

	public int getMaxDataLength() {
		return dbmsInfo.getMaxUserCollectionsDataLength();
	}

	public int getSystemCollectionDataLength() {
		return dbmsInfo.getSystemCollectionsDataLength();
	}

	public String getRowID() {
		return dbmsInfo.getRowId();
	}

	public boolean useDoubleQuotes4DDLNames() {
		return dbmsInfo.useDoubleQuotes4DDLNames();
	}
	///////////////////////////////////////////////////////////////
	//                      Connection methods
	///////////////////////////////////////////////////////////////
	/** calls  the encapsulated JDBC connection rollback method
	 * @throws SQLException JDBC error
	 */
	public void rollback() throws SQLException {
		if (connection.getAutoCommit() == false) // Start may have been called
			{
			connection.rollback();
			connection.setAutoCommit(true);
		}
	}
	/** This method should be called before processing an updating user request.
	 * This method actually saves the user specified auto-commit mode and
	 * turns the auto-commit mode of the current connection to off.
	 * @throws RepositoryException Application exception
	 * @see #commit
	 */
	public void start() throws RepositoryException {
		try {
			connection.setAutoCommit(false);
		} catch (SQLException e) {
			throw new RepositoryException(RepositoryException.DB_ERROR, "Relational database error while setting autocommit mode to false.", e);
		}
	}

	/** Before processing a user request that updates the database, the
	 * auto-commit mode of the current connection is turned off to allow all
	 * the changes to be committed or aborted together. The initial
	 * auto-commit mode of the connection is backed up so it can be
	 * restored after the user request has been processed.
	 *
	 * This method is called upon successful completion of the user request.
	 * This method actually commits the database operations assuming the
	 * connection mode is off and restores the user auto-commit mode.
	 *
	 * If the initial auto-commit mode is false, this method restores the
	 * auto-commit mode and leaves the database operation uncommitted.
	 * @see #start
	 * @throws RepositoryException Application exception
	 */
	public void commit() throws RepositoryException {
		try {
			if (connection.getAutoCommit() == false) // Start may have been called
				{
				connection.commit();
				connection.setAutoCommit(true);
			}
		} catch (SQLException e) {
			throw new RepositoryException(RepositoryException.DB_ERROR, "Relational database error while restoring autocommit mode to initial value.", e);
		}
	}

	///////////////////////////////////////////////////////////////
	//                      Table methods
	///////////////////////////////////////////////////////////////
	public boolean checkTable(TableInfo table) throws SQLException {
		return checkTable(table.getName());
	}

	public boolean checkTable(String name) throws SQLException {
		ResultSet rs = null;
		try {
			rs = connection.getMetaData().getTables(connection.getCatalog(), schemaName, name, null);
			return rs.next();
		} finally {
			if (rs != null)
				rs.close();
		}
	}

	/** Retrieves org.xquark.mapper.mapping.ColumnMetadata from the
     * underlying database.
     * @param tableName The table name
     * @throws SQLException If a database error occurs while creation
     */
    public Map getTableMetadata(String tableName) throws SQLException {
        Iterator it = dbmsInfo.getTableMetadata(connection.getCatalog(),
                schemaName, tableName, connection,
                new JDBCProperties() {
					public boolean useResultSet4ColumnMetadata() {return false;}
					public boolean useStaticMappingFromNativeType() {return true;}
				}).iterator();
    	HashMap columns = new HashMap();
		ColumnMetaData cmeta;
		while (it.hasNext()) {
			cmeta = (ColumnMetaData)it.next();
			columns.put(cmeta.getColumnName(), cmeta);
		}
    	return columns;
    }

	///////////////////////////////////////////////////////////////
	//                      Table creation methods
	///////////////////////////////////////////////////////////////
	/** Create a relational table
	 * @param collection The collection name
	 * @param columns table specification loaded from the TableSpec.xml resource
	 * @throws SQLException If a database error occurs while creation
	 */
	protected void createTable(TableInfo table) throws SQLException {
		/* Create table */
		Statement stmt = connection.createStatement();
		try {
			stmt.executeUpdate(table.getCreationStatement());
		} finally {
			stmt.close();
		}
	}

	///////////////////////////////////////////////////////////////
	//           Complex SQL statements (recursive, subqueries...)
	///////////////////////////////////////////////////////////////

	public String getDeleteUserTableRowsStatement(String tableName, String documentRowsSelectionClause) throws SQLException, RepositoryException {
		StringBuffer deletionStmt = new StringBuffer();
		deletionStmt.append("DELETE FROM ");
		deletionStmt.append(tableName);
		deletionStmt.append(" WHERE ");
		deletionStmt.append(getRowID());
		deletionStmt.append(" IN (");
		deletionStmt.append("SELECT w.");
		deletionStmt.append(getRowID());
		deletionStmt.append(' ');
		deletionStmt.append(documentRowsSelectionClause);
		deletionStmt.append(')');

		return deletionStmt.toString();
	}

	public void executeRangeUpdate(String statement, long first, long last) throws SQLException, RepositoryException {
		PreparedStatement stmt = null;
		try {
			stmt = connection.prepareStatement(statement);
			stmt.setLong(1, first);
			stmt.setLong(2, last);
			stmt.executeUpdate();
		} finally {
			if (stmt != null)
				stmt.close();
		}
	}

	///////////////////////////////////////////////////////////////
	//                      Private utilities
	///////////////////////////////////////////////////////////////

	///////////////////////////////////////////////////////////////
	//                      DBMS specificities
	///////////////////////////////////////////////////////////////
	/**
	 * Return the column specification conversion string specific to the
	 * rdbms corresponding to an XML Schema primitive type, especially for
	 * date conversions (to_date() function).
	 * @param type the XML schema primitive type
	 */
	public String getColumnStatement(int type) throws RepositoryException {
		switch (type) {
			//                   to_date(?, 'YYYY-MM-DD'), to_date(?, 'HH24:MI:SS')
			case PrimitiveType.DURATION : // TO CHECK
			case PrimitiveType.DATE_TIME :
				// REMOVED for TT40                                return "to_date(?, 'SYYYY-MM-DD'T'HH24:MI:SS')";
				return "to_date(?, 'YYYY-MM-DD'T'HH24:MI:SS')";
			case PrimitiveType.TIME :
				return "to_date(?, 'HH24:MI:SS')";
			case PrimitiveType.DATE :
				// REMOVED for TT40                                return "to_date(?, 'SYYYY-MM-DD')";
				return "to_date(?, 'YYYY-MM-DD')";
			case PrimitiveType.GYEAR_MONTH :
				// REMOVED for TT40                                return "to_date(?, 'SYYYY-MM')";
				return "to_date(?, 'YYYY-MM')";
			case PrimitiveType.GYEAR :
				// REMOVED for TT40                                return "to_date(?, 'SYYYY')";
				return "to_date(?, 'YYYY')";
			case PrimitiveType.GMONTH_DAY :
				return "to_date(?, 'MM-DD')";
			case PrimitiveType.GMONTH :
				return "to_date(?, 'DD')";
			case PrimitiveType.DECIMAL :
				return "to_number(?, '999999999.999999999')"; // TO DO if convert is rejected : determine how much
			case PrimitiveType.DOUBLE :
			case PrimitiveType.FLOAT :
			default :
				return "?";
		}
	}

	/** This method is called when a repository is initialized to perform
	 * dbms preparation.
	 */
	public void onInitRepository(Repository rep) throws SQLException {}

	/** This method is called when a repository is deleted to restore initial
	 * dbms state.
	 */
	public void onDeleteRepository(Repository rep) throws SQLException {}

	/** This method is called when a repositoryConnection is opened to perform
	 * thread specific operations.
	 */
	public void onInitConnection() throws SQLException {
		dropTemporaryTables();
	}

	protected short getRandomCollectionID() {
		return (short) (System.currentTimeMillis() % Short.MAX_VALUE);
	}

	/** Delete rows in the reconstruction temporary table name.
	 */
	public void emptyTable(String tableName) throws SQLException {
		Statement stmt = null;
		try {
			stmt = connection.createStatement();
			stmt.execute("TRUNCATE TABLE " + tableName);
		} finally {
			stmt.close();
		}
	}

	public TableInfo createQueryTemporaryTable() throws SQLException {
		String randomTableName = null;
		boolean created = false;
		TableInfo table = null;
		while (!created) {
			try {
				table = getTemporaryTableInfo();
				executeUpdate(table.getCreationStatement());
				created = true;
			} catch (SQLException e) {
				if (getUniversalError(e) != OBJECT_ALREADY_EXISTS)
					throw e;
			}
		}
		return table;
	}

	public String createTemporaryTableAs(String subQuery) throws SQLException {
		String randomTable = null;
		boolean created = false;
		Statement stmt = null;
		while (!created) {
			try {
				randomTable = getTemporaryTableName();
				stmt = connection.createStatement();
				StringBuffer DDLStatement = new StringBuffer("CREATE");
				//            	DDLStatement.append(getTempExpression(1));
				DDLStatement.append(" GLOBAL TEMPORARY ");
				DDLStatement.append("TABLE ");
				DDLStatement.append(randomTable);
				//            	DDLStatement.append(getTempExpression(2));
				DDLStatement.append(" ON COMMIT PRESERVE ROWS");
				DDLStatement.append(" AS ");
				DDLStatement.append(subQuery);

				if (log.isDebugEnabled())
					log.debug("Temp table creation\n" + DDLStatement);

				stmt.executeUpdate(DDLStatement.toString());
				created = true;
			} catch (SQLException e) {
				if (getUniversalError(e) != OBJECT_ALREADY_EXISTS)
					throw e;
			} finally {
				if (stmt != null)
					stmt.close();
			}
		}
		return randomTable;
	}

	public void dropTable(String tableName) throws SQLException {
		executeUpdate("DROP TABLE " + tableName);
	}

	public void executeUpdate(String sqlStmt) throws SQLException {
		Statement stmt = connection.createStatement();
		try {
			stmt.executeUpdate(sqlStmt);
		} finally {
			stmt.close();
		}
	}

	public void dropTemporaryTable(String tableName) throws SQLException {
		dropTable(tableName);
	}

	public void dropTemporaryTables() throws SQLException {
		DatabaseMetaData metadata = connection.getMetaData();
		TableSpec tempSpecs = TableSpecLoader.getInstance().getTableSpec(TableSpec.CAT_QUERY_TMP, TableSpec.TYPE_QUERY_TMP);
		ResultSet rs = metadata.getTables(connection.getCatalog(), schemaName, tempSpecs.getAllCollectionsWildcardTableName(), null); //new String[] {"GLOBAL TEMPORARY", "LOCAL TEMPORARY"}
		try {
			while (rs.next()) {
				try {
					dropTable(rs.getString(3));
				} catch (SQLException e) {
					// ignore
				}
			}
		} finally {
			rs.close();
		}
	}

	public int getUniversalError(SQLException e) {
		return UNKNOW_ERROR_CODE;
	}

	/** closes the underlying JDBC connection  */
	public void close() throws SQLException {
		connection.close();
	}

	public void updateStatistics(String tableName) throws SQLException {}

	public boolean useStringDelimitor() {
		return false;
	}

	public boolean distinguishNullAndEmptyStrings() {
		return true;
	}

	/** Generate a random temporary relational table name WHICH IS NOT
	 * GUARANTEED TO BE UNIQUE.
	 * @return the name of the table.
	 */
	public String getTemporaryTableName() {
		return TableSpecLoader.getInstance().getTableSpec(TableSpec.CAT_QUERY_TMP, TableSpec.TYPE_QUERY_TMP).getTableName(getRandomCollectionID());
	}

	public TableInfo getTemporaryTableInfo() {
		return new TableInfo(TableSpecLoader.getInstance().getTableSpec(TableSpec.CAT_QUERY_TMP, TableSpec.TYPE_QUERY_TMP), getRandomCollectionID(), this);
	}

	public String getTemporaryTableCreationStatement(TableInfo table) {
		return dbmsInfo.getTemporaryTableStatement().getTableCreationStatement(table.getName(), table.getLogicalSpecification());
	}

	public String getIndexOrganizedTableCreationStatement(TableInfo table) {
		return dbmsInfo.getIndexOrganizedStatement().getTableCreationStatement(table.getName(), table.getLogicalSpecification());
	}

	public String getLockSelectStatement(TableInfo table, String filterClause) {
		return getLockSelectStatement(table.getColumnList(), table.getName(), filterClause);
	}

	public String getLockSelectStatement(String columnList, String tableName, String filterClause) {
		return dbmsInfo.getSelectLockStatement().getSelectStatement(columnList, tableName, filterClause);
	}

	public String getBitmapIndexClause() {
		return null;
	}

	public String getHashIndexClause() {
		return null;
	}

	public String getBTreeIndexClause() {
		return null;
	}

	public void createSequence(TableInfo table, int step) throws SQLException {
		executeUpdate(table.getCreationStatement());
		// initialize sequence emulation
		StringBuffer sql = new StringBuffer();
		sql.append("INSERT INTO ");
		sql.append(table.getName());
		sql.append('(');
		sql.append(table.getColumns()[0]);
		sql.append(") VALUES(");
		sql.append(RepositoryConstants.SEQUENCE_FIRST_VALUE);
		sql.append(')');
		executeUpdate(sql.toString());
	}

	public void dropSequence(String seqName) throws SQLException {
		dropTable(seqName);
	}

	public long nextSequenceValue(TableInfo table, short step) throws RepositoryException {
		start();
		Statement stmt = null;
		ResultSet rs = null;
		long next = 0;
		try {
			stmt = connection.createStatement();
			rs = stmt.executeQuery(getLockSelectStatement(table, ""));
			if (rs.next())
				next = rs.getLong(1);
			else
				throw new RepositoryException(RepositoryException.CONSISTENCY_ERROR, "Table " + table.getName() + " should contain one row.");
		} catch (SQLException e) {
			throw new RepositoryException(RepositoryException.DB_ERROR, "JDBC error while querying the " + table.getName() + " table.");
		} finally {
			if (rs != null) {
				try {
					rs.close();
					stmt.close();
				} catch (SQLException e) {
					// no op
				}
			}
		}
		StringBuffer sql = new StringBuffer();
		String column = table.getColumns()[0].getName(); // assume there is only one column
		sql.append("UPDATE ");
		sql.append(table.getName());
		sql.append(" SET ");
		sql.append(column);
		sql.append('=');
		sql.append(next + step);
		sql.append(" WHERE ");
		sql.append(column);
		sql.append('=');
		sql.append(next);
		try {
			executeUpdate(sql.toString());
		} catch (SQLException e) {
			throw new RepositoryException(RepositoryException.DB_ERROR, "JDBC error while inserting new value in the " + table.getName() + " table.");
		}
		commit();

		return next;
	}

	public Collection createTables(Collection tables) throws SQLException {
		Iterator it = tables.iterator();
		TableInfo wTable;
		Statement stmt = null;
		ArrayList notCreated = new ArrayList();
		try {
			stmt = connection.createStatement();
			while (it.hasNext()) {
				wTable = (TableInfo) it.next();
				if (wTable.isSequence())
					notCreated.add(wTable);
				else {
					//                    addBatch(stmt, wTable.getCreationStatement());
					stmt.executeUpdate(wTable.getCreationStatement());
					for (int i = 0; i < wTable.getIndexCreationStatements().length; i++)
						stmt.executeUpdate(wTable.getIndexCreationStatements()[i]);
					//                        addBatch(stmt, wTable.getIndexCreationStatements()[i]);
				}
			}
			//            executeBatch(stmt);
		} finally {
			if (stmt != null)
				stmt = null;
		}
		return notCreated;
	}

	public void createIndexes(Collection tables) throws SQLException {
		Iterator it = tables.iterator();
		TableInfo wTable;
		Statement stmt = null;
		try {
			stmt = connection.createStatement();
			while (it.hasNext()) {
				wTable = (TableInfo) it.next();
				if (!wTable.isSequence()) {
					for (int i = 0; i < wTable.getIndexCreationStatements().length; i++)
						addBatch(stmt, wTable.getIndexCreationStatements()[i]);
				}
			}
			executeBatch(stmt);
		} finally {
			if (stmt != null)
				stmt = null;
		}
	}

	public void dropIndexes(Collection tables) throws SQLException {
		Iterator it = tables.iterator();
		TableInfo wTable;
		Statement stmt = null;
		try {
			stmt = connection.createStatement();
			while (it.hasNext()) {
				wTable = (TableInfo) it.next();
				if (!wTable.isSequence()) {
					for (int i = 0; i < wTable.getIndexDropStatements().length; i++)
						addBatch(stmt, wTable.getIndexDropStatements()[i]);
				}
			}
			executeBatch(stmt);
		} finally {
			if (stmt != null)
				stmt = null;
		}
	}

	public void createSequence(TableInfo table, short step) throws SQLException {}

	public void setObject(PreparedStatement pStmt, int index, Object o, int javaType, DbType sqlType)
	throws SQLException {
		pStmt.setObject(index, o);
	}

	public List getUserTableNames(String tablePattern) throws SQLException {
		ResultSet rs = null;
		ArrayList tables = new ArrayList();

		try {
			rs = connection.getMetaData().getTables(connection.getCatalog(), schemaName, tablePattern, null);
			while (rs.next())
				tables.add(rs.getString(3));
		} finally {
			if (rs != null)
				rs.close();
		}
		return tables;
	}

	public List getUserSequenceNames(String seqPattern) throws SQLException {
		return Collections.EMPTY_LIST;
	}

}
