/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.mapper.dbms;

import java.util.Arrays;

/**
 * Class containing information for relational table index creation (shared object).
 *
 */
public class IndexSpec extends ConstraintSpec
{
    private static final String RCSRevision = "$Revision: 1.2 $";
    private static final String RCSName     = "$Name: XQ1_1_1 $";
    
    /* dbms indexes types */
    public static final byte INDEX_DEFAULT = 0;
    public static final byte INDEX_BITMAP = 1;
    public static final byte INDEX_HASH = 2;
    public static final byte INDEX_BTREE = 3;
    
    private static final byte[] indexTypes =
    {INDEX_BITMAP, INDEX_BTREE, INDEX_DEFAULT, INDEX_HASH};
     /* Alphabetically sorted */
     private static final String[] indexTypeStrings =
     {"bitmap", "btree", "default", "hash"};
    
    
    private TableSpec table = null;
    private String suffix = null;
    
    //////////////////////////////////////////////////////////////////////////
    //                          CONSTRUCTORS
    //////////////////////////////////////////////////////////////////////////
    public IndexSpec(TableSpec table, String type, String suffix)
    {
        this.type = parseType(type);
        this.table = table;
        this.suffix = suffix;
    }
    
    public String getIndexName()
    {
        return table.getTableName() + '_' + suffix;
    }
    
    public String getIndexName(short collectionID)
    {
        return table.getTableName(collectionID) + '_' + suffix;
    }
    
    public String getIndexName(short collectionID, short tableID)
    {
        return table.getTableName(collectionID, tableID) + '_' + suffix;
    }
    
    public String generateCreationStatement(AbstractConnection dbmsInfo)
    {
        return generateCreationStatement(getIndexName(),
            table.getTableName(), dbmsInfo);
    }
    
    public String generateCreationStatement(short collectionID, AbstractConnection dbmsInfo)
    {
        return generateCreationStatement(getIndexName(collectionID),
            table.getTableName(collectionID), dbmsInfo);
    }
    
    public String generateCreationStatement(short collectionID, short tableID, AbstractConnection dbmsInfo)
    {
        return generateCreationStatement(getIndexName(collectionID, tableID),
            table.getTableName(collectionID, tableID), dbmsInfo);
    }
    
    public String generateDropStatement()
    {
        return generateDropStatement(getIndexName());
    }
    
    public String generateDropStatement(short collectionID)
    {
        return generateDropStatement(getIndexName(collectionID));
    }
    
    public String generateDropStatement(short collectionID, short tableID)
    {
        return generateDropStatement(getIndexName(collectionID, tableID));
    }
    
    //////////////////////////////////////////////////////////////////////////
    //                          ACCESSORS
    //////////////////////////////////////////////////////////////////////////
    public byte getType()
    {
        return type;
    }
    
    private byte parseType(String type)
    {
        int index = Arrays.binarySearch(indexTypeStrings, type);
        if (index < 0)
            throw new RuntimeException("The DBMS index type " + type + " used in tableSpec.xml is unknown.");
        return indexTypes[index];
    }
    
    //////////////////////////////////////////////////////////////////////////
    //                          PRIVATE
    //////////////////////////////////////////////////////////////////////////
    private String generateCreationStatement(String indexName, String tableName,
                                                    AbstractConnection dbmsInfo)
    {
        StringBuffer stmt = new StringBuffer();
        stmt.append("CREATE ");
        String typeClause = null;
        switch (type)
        {
            case INDEX_BITMAP:
                typeClause = dbmsInfo.getBitmapIndexClause();
                break;
            case INDEX_BTREE:
                typeClause = dbmsInfo.getBTreeIndexClause();
                break;
            case INDEX_HASH:
                typeClause = dbmsInfo.getBitmapIndexClause();
                break;
            default:
        }
        if (typeClause != null)
        {
            stmt.append(typeClause);
            stmt.append(' ');
        }
        stmt.append("INDEX ");
        stmt.append(indexName);
        stmt.append(" ON ");
        stmt.append(tableName);
        stmt.append(generateColumnList());
        return stmt.toString();
    }
    
    private String generateDropStatement(String indexName)
    {
        StringBuffer stmt = new StringBuffer();
        stmt.append("DROP INDEX ");
        stmt.append(indexName);
        return stmt.toString();
    }
}
