/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

/*
 * JDBCBatcher.java
 *
 * Created on 30 mai 2001, 10:28
 */

package org.xquark.mapper.dbms;

import java.io.StringWriter;
import java.sql.BatchUpdateException;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.util.*;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.xquark.mapper.RepositoryException;
import org.xquark.mapper.util.Tabulator;

/** Manage JDBC 2.0 batch feature with referential integrity
 * and automatic flush handling.
 * <p><B>Warning :</B> dependency is handled only as chains, not trees.</p>
 *
 */
public class JDBCBatcher
{
    private static final String RCSRevision = "$Revision: 1.4 $";
    private static final String RCSName     = "$Name: XQ1_1_1 $";

    private static Log log = LogFactory.getLog("internal." + JDBCBatcher.class.getName());

    final static int DEFAULT_BATCH_SIZE = 20;
    
    /* key is the statement, value is StatementInfo */
    HashMap statements;
    
    /* lists of dependent statements, less dependent 1st*/
    List statementLists;
    
    AbstractConnection conn;
    
    protected boolean autoFlush = true;
    
    /** Creates new JDBCBatcher */
    public JDBCBatcher(AbstractConnection connection, boolean autoFlush)
    {
        statements = new HashMap();
        statementLists = new ArrayList();
        conn = connection;
        this.autoFlush = autoFlush;
    }
    
    /* Add an independent statement (a new list is created) with the default
     * batch size.
     * @return the index of the list created
     */
    public int addStatement(PreparedStatement stmt, String tableName)
    {
        return addStatement(stmt, tableName, DEFAULT_BATCH_SIZE);
    }
    
    /* Add an independent statement (a new list is created)
     *  @return the index of the list created
     */
    public int addStatement(PreparedStatement stmt, String tableName, int batchSize)
    {
        
        return addStatement(- 1, stmt, tableName, batchSize);
    }
    
    /* add a statement at the end of a specific statement list
     * @param listNumber
     */
    public int addStatement(int listIndex, PreparedStatement stmt, String tableName, int batchSize)
    {
        if (stmt == null)
            return -1;
        if (listIndex == -1)
        {
            listIndex = statementLists.size();
            statementLists.add(new ArrayList());
        }
        StatementInfo info = new StatementInfo(batchSize,
        listIndex,
        ((List)statementLists.get(listIndex)).size(),
        stmt, tableName);
        statements.put(stmt, info);
        ((List)statementLists.get(listIndex)).add(info);
        return listIndex;
    }
    
    /* returns true if an automatic flush has been performed */
    public boolean addBatch(PreparedStatement stmt, int line, int column) throws RepositoryException
    {
        StatementInfo info = (StatementInfo)statements.get(stmt);
		try
		{
			if (info.batchSize != 1)	// else batch desactivated for debug (statement executed in flush)
				conn.addBatch(stmt);
		}
		catch (SQLException e)
		{
			throw new RepositoryException(
				RepositoryException.DB_ERROR,
				"JDBC error while inserting tuple in batch.",
				e);
		}
		if (info.inc(line, column)) // needs automatic flush
		{
			if (log.isDebugEnabled())
				log.debug("JDBC batcher automatic flush triggered by\n" + info + '\n' + this);
			flush(info.getListIndex(), info.getIndex());
			return true;
		}
		else
		{
			if (log.isDebugEnabled())
				log.debug("Add row to batch\n" + info);
			return false;
		}
	}
    
    /* returns true if an automatic flush has been performed */
    public boolean addBatch(PreparedStatement stmt) throws RepositoryException
    {
        return addBatch(stmt, -1, -1);
    }
    
    public void setAutoFlush(boolean mode)
    {
        autoFlush = mode;
    }
    
    public void flush() throws RepositoryException
    {
		if (log.isDebugEnabled())
			log.debug("General JDBC batcher Forced Flush\n" + this);
        Iterator it = statementLists.iterator();
        while (it.hasNext())
        {
            flush((List)it.next());
        }
    }
    
    public void flush(int listIndex) throws RepositoryException
    {
        flush((List)statementLists.get(listIndex));
    }
    
    private void flush(List stmtList) throws RepositoryException
    {
        for(int i = 0; i < stmtList.size(); i++)
        {
            flush((StatementInfo)stmtList.get(i));
        }
    }
    
    public String toString()
    {
        StringWriter sw = new StringWriter();
        Tabulator tab = new Tabulator(sw, 79, new int[] {0, 15, 30, 45, 60});
        tab.addItem("Table");
        tab.addItem("Batch size");
        tab.addItem("Batch count");
        tab.addItem("List index");
        tab.addItem("Stmt index");
        
        Iterator it = statementLists.iterator();
        List list;
        StatementInfo si;
        
        while (it.hasNext())
        {
            list = (List)it.next();
            for(int i = 0; i < list.size(); i++)
            {
                si = (StatementInfo)list.get(i);
                tab.addItem(si.tableName);
                tab.addItem(si.batchSize);
                tab.addItem(si.batchCount);
                tab.addItem(si.listIndex);
                tab.addItem(si.indexStatement);
            }
            tab.newLine();
        }
        return sw.toString();
    }
    
    public void flush(int indexList, int index) throws RepositoryException
    {
        flush((List)statementLists.get(indexList), index);
    }
    
    private void flush(List stmtList, int index) throws RepositoryException
    {
        for(int i = 0; i <= index; i++)
        {
            flush((StatementInfo)stmtList.get(i));
        }
    }
    
    public void flush(PreparedStatement stmt) throws RepositoryException
    {
        StatementInfo info = (StatementInfo)statements.get(stmt);
        flush(info.getListIndex(), info.getIndex());
    }
    
    /* Elementary method : no cascading */
    private void flush(StatementInfo info) throws RepositoryException
    {
        int [] result = null;
        try
        {
            if (info.getCount() > 0)
			{
				if (info.batchSize == 1)
					info.getStatement().execute();	// in order to debug
				else
					result = conn.executeBatch(info.getStatement());
			}
		}
        catch (BatchUpdateException e)
        {
            result = e.getUpdateCounts();
            int errorIndex = -1;
            for (errorIndex = 0; (errorIndex < result.length) && (result[errorIndex] != -3); errorIndex++);
            
            String message = "JDBC error while executing batch on table " + info.getTableName();
            LocatorImpl location = null;
            
            if (errorIndex >= 0)
                location = info.getLocator(errorIndex);
            
            if (location == null)
                throw new RepositoryException(RepositoryException.DB_ERROR, message, e);
            else
                throw new RepositoryException(RepositoryException.DB_ERROR, message + " " + location, e);
        }
        catch (SQLException e)
        {
            throw new RepositoryException(RepositoryException.DB_ERROR, "JDBC error while executing batch on table " + info.getTableName(), e);
        }
        finally
        {
            info.reset();
        }
    }
    
    public void clear() throws SQLException
    {
		if (log.isDebugEnabled())
			log.debug("Clear batcher\n" + this);
        statements.clear();
        statementLists.clear();
    }
    
    public void reset() throws SQLException
    {
		if (log.isDebugEnabled())
			log.debug("Reset batcher\n" + this);
        int max = statementLists.size();
        for (int i= 0; i < max; i++)
            reset(i);
    }
    
    public void reset(int i) throws SQLException
    {
        Iterator it = ((List)statementLists.get(i)).iterator();
        StatementInfo info;
        while (it.hasNext())
        {
            info = (StatementInfo)it.next();
            info.getStatement().clearBatch();
            info.reset();
        }
    }
    
    public void remove(PreparedStatement stmt)
    {
        StatementInfo info = (StatementInfo)statements.get(stmt);
        ((List)statementLists.get(info.getListIndex())).remove(info.getIndex());
        statements.remove(stmt);
    }
    
    public void close() throws SQLException
    {
        if (statements != null)
        {
            clear();
            statements = null;
            statementLists = null;
            conn = null;
        }
    }
    
    private class StatementInfo
    {
        private int batchSize;
        private int batchCount = 0;
        private List locators = new ArrayList();
        private int listIndex;
        private int indexStatement;
        private PreparedStatement statement;
        private String tableName;
        StatementInfo(int batchSize, int listIndex, int indexStatement, PreparedStatement statement, String tableName)
        {
            this.batchSize = batchSize;
            this.listIndex = listIndex;
            this.indexStatement = indexStatement;
            this.statement = statement;
            this.tableName = tableName;
        }
        boolean inc(int line, int column)
        {
            if ((line >= 0) || (column >= 0))
                locators.add(new LocatorImpl(line, column));
            batchCount++;
            return (autoFlush && (batchCount == batchSize || log.isDebugEnabled())); // Setting batch size to 1 is needed for debug.
        }
        void reset()
        {
            batchCount = 0;
            locators.clear();
        }
        LocatorImpl getLocator(int index)
        {
            if (index >= locators.size())
                return null;
            else
                return (LocatorImpl)locators.get(index);
        }
        int getCount()
        {return batchCount;}
        int getListIndex()
        {return listIndex;}
        int getIndex()
        {return indexStatement;}
        String getTableName()
        {return tableName;}
        PreparedStatement getStatement()
        {return statement;}
        public String toString()
        {
            return tableName + "\t"+ batchSize + "\t" + batchCount + "\t" + listIndex + "\t" + indexStatement;
        }
    }
    
    private class LocatorImpl
    {
        int line;
        int column;
        
        LocatorImpl(int line, int column)
        {
            this.line = line;
            this.column = column;
        }
        
        public String toString()
        {
            return "(line " + line + ", column " + column + ")";
        }
    }
}
