/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.mapper.dbms;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.net.URL;
import java.sql.SQLException;
import java.util.*;

import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;
import org.xquark.mapper.*;
import org.xquark.mapper.mapping.MappingFactory;
import org.xquark.mapper.metadata.*;
import org.xquark.mapper.storage.*;
import org.xquark.mapper.util.DestructionToken;
import org.xquark.mapper.util.DisposableReaderFilter;
import org.xquark.mapper.util.RepositoryRuntimeException;
import org.xquark.schema.SchemaManager;
import org.xquark.xml.xdbc.*;

/**
 * @see org.xquark.mapper.AccessManager
 * @see org.xquark.mapper.XMLCollection
 */
public final class RepositoryConnectionImpl implements _RepositoryConnection {
    private static final String RCSRevision = "$Revision: 1.6 $";

    private static final String RCSName = "$Name: XQ1_1_1 $";

    // Static variables (to be independant from XQuery package)
    private static Constructor stmtConstructor = null;

    private AbstractConnection connection;

    private Repository rep;

    /* param for dynamic instantiation of statement */
    private Object[] params = new Object[2];

    /** Destruction token given by the AccessManager */
    private DestructionToken destructor;

    private ArrayList XMLCollections = new ArrayList(); // to close XMLCollections when closing & set modes

    private HashSet XMLStatements = new HashSet(); // to close XMLStatements

    private InputSource source;

    //  Options
    //
    private boolean autoCommit = true;

    private boolean readOnly = false;

    private boolean interactiveQueries = true;

    public RepositoryConnectionImpl(Repository rep, AbstractConnection connection, DestructionToken destructor) throws RepositoryException {
        this.rep = rep;
        this.destructor = destructor;
        this.connection = connection;
        // dynamic instantiation of XQuery statement (removed from static statement
        // because QA loads classes to check RCS IDs...)
        loadXQueryStatementClass();
        params[0] = this;
    }

    public _Repository getMetadata() {
        return rep;
    }

    public SchemaManager getSchemaManager() {
        return rep.getSchemaManager();
    }

    public void commit() throws XMLDBCException, XMLDBCNotSupportedException {
        flush();
        try {
            connection.getConnection().commit();
            broadcastCommit();
        } catch (SQLException e) {
            throw new RepositoryException(RepositoryException.DB_ERROR, "Could not commit the underlying JDBC connection.", e);
        }
    }

    public void rollback() throws XMLDBCException, XMLDBCNotSupportedException {
        try {
            connection.getConnection().rollback();
            broadcastRollback();
        } catch (SQLException e) {
            throw new RepositoryException(RepositoryException.DB_ERROR, "Could not rollback the underlying JDBC connection.", e);
        }
    }

    public void setAutoCommit(boolean mode) throws XMLDBCException {
        autoCommit = mode;
        try {
            connection.getConnection().setAutoCommit(mode);
        } catch (SQLException e) {
            throw new RepositoryException(RepositoryException.DB_ERROR, "Could not set the autoommit mode on the underlying JDBC connection.", e);
        }

    }

    ///////////////////////////////////////////////////////////////////////////
    //                     Wrapping of Repository
    ///////////////////////////////////////////////////////////////////////////

    public RepositoryReader getMappingReader() throws XMLDBCException {
        return ((_RepositoryCollection) getMappingCollection()).getReader();
    }

    public String getURL() {
        return RepositoryDriver.REPOSITORY_URL_PREFIX + rep.getURL();
    }

    public String toString() {
        return rep.toString();
    }

    public String getUserName() {
        return rep.getUserName();
    }

    public AbstractConnection getConnection() {
        return connection;
    }

    public boolean getAutoCommit() {
        return autoCommit;
    }

    ///////////////////////////////////////////////////////////////////////////
    //                     XMLConnection implementation
    ///////////////////////////////////////////////////////////////////////////
    public Configurable createCollectionConfig() throws XMLDBCException, XMLDBCNotSupportedException {
        return new CollectionInfo(false);
    }

    public XMLCollection createCollection(String name, String description, Configurable config) throws XMLDBCException, XMLDBCNotSupportedException {
        protectReadOnlyMethod();
        CollectionInfo colInfo = new CollectionInfo(config);
        colInfo.setDescription(description);
        MappingFactory mapFactory = rep.getMapping((String) config.getProperty(MAPPING_ID_PROPERTY));
        rep.createCollection(name, colInfo, mapFactory);
        return getCollection(name);
    }

    public XMLCollection getCollection(String name) throws XMLDBCException, XMLDBCNotSupportedException {
        return getRepositoryCollection(name);
    }

    public void renameCollection(String oldname, String newname) throws XMLDBCException, XMLDBCNotSupportedException {
        protectReadOnlyMethod();
        rep.renameCollection(oldname, newname);
    }

    public void deleteCollection(String name) throws XMLDBCException, XMLDBCNotSupportedException {
        protectReadOnlyMethod();
        name = name.toUpperCase();

        // close all opened XMLCollections
        Iterator it = ((List) XMLCollections.clone()).iterator(); // clone because of destructor
        RepositoryCollection collection;
        while (it.hasNext()) {
            collection = (RepositoryCollection) it.next();
            if (collection.getName().equals(name))
                collection.close(); // will remove from the original list
        }
        // Close metadata if possible (another connection could use it)
        rep.deleteCollection(name);
    }

    public int deleteAllCollections() throws XMLDBCException, XMLDBCNotSupportedException {
        protectReadOnlyMethod();
        return rep.deleteAllCollections();
    }

    public void setTransactionIsolation(short level) throws XMLDBCException, XMLDBCNotSupportedException {
        throw new XMLDBCNotSupportedException("Not Supported yet");
    }

    public short getTransactionIsolation() throws XMLDBCException, XMLDBCNotSupportedException {
        throw new XMLDBCNotSupportedException("Not Supported yet");
    }

    public void setReadOnly(boolean readOnlyMode) throws XMLDBCException, XMLDBCNotSupportedException {
        if (readOnlyMode)
            broadcastReadOnly();
        readOnly = readOnlyMode;
    }

    public boolean isReadOnly() throws XMLDBCException {
        return readOnly;
    }

    public XMLDataSourceMetaData getMetaData() throws XMLDBCException {
        return new DataSourceMetaData(this);
    }

    public XMLDataSourceMetaData getMetaData(boolean refresh) throws XMLDBCException {
        return new DataSourceMetaData(this, refresh);
    }

    public void close() throws XMLDBCException {
        if (!isClosed()) {
            _close();
            try {
                connection.close();
            } catch (SQLException e) {
                // TO SEE
            }
            destructor.destruct(this);
            destructor = null;
            XMLCollections = null;
            XMLStatements = null;
            rep = null;
            connection = null;
        }
    }

    public boolean isClosed() throws XMLDBCException {
        return (destructor == null);
    }

    public XMLStatement createStatement(short queryType) throws XMLDBCException {
        if (queryType == XQUERY_STRING_TYPE)
            return createStatement();
        else
            throw new UnsupportedOperationException("Query type is not Supported.");
    }

    public XMLStatement createStatement() throws XMLDBCException {
        params[1] = new XMLStatementDestructor();
        XMLStatement stmt = null;
        try {
            stmt = (XMLStatement) stmtConstructor.newInstance(params);
        } catch (InvocationTargetException e) {
            if (e.getTargetException() instanceof XMLDBCException)
                throw (XMLDBCException) e.getTargetException();
            else
                throw new RepositoryException(RepositoryException.INTERNAL_ERROR, "Exception while creating XQuery statement.", e.getTargetException());
        } catch (Exception e) {
            throw new RepositoryException(RepositoryException.INTERNAL_ERROR, "Repository XQuery module is missing in your CLASSPATH.", e);
        }
        XMLStatements.add(stmt);
        return stmt;
    }

    public PreparedXMLStatement prepareStatement(String query) throws XMLDBCException, XMLDBCNotSupportedException {
        throw new XMLDBCNotSupportedException("Not Supported yet");
    }

    public void setBaseURI(String baseURI) {
        //throw new XMLDBCNotSupportedException("Not Supported yet");
    }

    ///////////////////////////////////////////////////////////////////////////
    //                     RepositoryConnection implementation
    ///////////////////////////////////////////////////////////////////////////
    /**
     * Open an existing user collection. Compared to the XDBC {@link XMLConnection#getCollection(String) getCollection()}method, this method avoids the cast operation
     * 
     * @return a {@link RepositoryCollection XMLCollection}object.
     * @throws RepositoryException
     *             API exception.
     */
    public RepositoryCollection getRepositoryCollection(String name) throws XMLDBCException {
        if (name == null)
            throw new RepositoryException(RepositoryException.PERMISSION_DENIED, "A collection name must be provided.");

        // system collections opened by name
        _RepositoryCollection collection;

        if (name.equals(SCHEMA_COLLECTION)) {
            collection = new SchemaXMLCollection(rep.getSchemaCollection(), this, new XMLCollectionDestructor());
        } else if (name.equals(MAPPING_COLLECTION)) {
            collection = new MappingXMLCollection(rep.getMappingCollection(), this, new XMLCollectionDestructor());
        } else {
            collection = new XMLCollectionImpl(rep.openCollection(name), this, new XMLCollectionDestructor());
        }
        collection.setReadOnly(readOnly);
        XMLCollections.add(collection); // name is not taken because of uppercase conversion
        return collection;
    }

    public RepositoryCollection getMappingCollection() throws XMLDBCException {
        return getRepositoryCollection(MAPPING_COLLECTION);
    }

    public RepositoryCollection getSchemaCollection() throws XMLDBCException {
        return getRepositoryCollection(SCHEMA_COLLECTION);
    }

    public void resetRepository() throws XMLDBCException {
        protectReadOnlyMethod();
        _close();
        rep.resetRepository();
    }

    public void deleteRepository() throws XMLDBCException {
        protectReadOnlyMethod();
        _close(); // close collection before close() try it after Reepository deleted
        rep.deleteRepository();
        close();
    }

    public void deleteConfiguration() throws XMLDBCException {
        protectReadOnlyMethod();
        rep.deleteConfiguration();
        close();
    }

    public void setInteractiveMode(boolean interactive) {
        interactiveQueries = interactive;
    }

    public boolean getInteractiveMode() {
        return interactiveQueries;
    }

    public List getCollectionsUsingMapping(String mappingID) throws XMLDBCException {
        return rep.getCollectionsUsingMapping(mappingID);
    }

    public List getCollectionsUsingNamespace(String namespace) throws XMLDBCException {
        return rep.getCollectionsUsingNamespace(namespace);
    }

    public List getMappingsUsingNamespace(String namespace) throws XMLDBCException {
        return Collections.EMPTY_LIST; // TO IMPLEMENT using XQuery ?
    }

    ///////////////////////////////////////////////////////////////////////////
    //                     ShemaLocator implementation
    ///////////////////////////////////////////////////////////////////////////

    public void clearSchemaLocations() {
        // no op
    }

    public Iterator getSchemaLocations(String namespace) {
        return null;
    }

    public void addSchemaLocation(String namespace, String location) {
        // no op
    }

    public XMLReader getSchemaReader() throws SAXException {
        try {
            return new DisposableReaderFilter(((_RepositoryCollection) getSchemaCollection()).getReader());
        } catch (XMLDBCException e) {
            throw new SAXException("Could not catch reader for the schema system XML collection.", e);
        }
    }

    public InputSource resolveLocation(String namespace, String location) {
        if (namespace == null)
            namespace = RepositoryConstants.NO_NAMESPACE_SCHEMA_ID;
        if (source == null)
            source = new InputSource(namespace);
        else
            source.setSystemId(namespace);
        try {
            if (getSchemaCollection().containsDocument(namespace))
                return source;
            else
                return null;
        } catch (XMLDBCException e) {
            return null;
        }
    }

    public InputSource resolveLocation(String namespace, URL location) {
        return resolveLocation(namespace, (String) null);
    }

    public String resolveInclude(String location) {
        return location;
    }

    ///////////////////////////////////////////////////////////////////////////
    //                      INNER CLASSES
    ///////////////////////////////////////////////////////////////////////////
    private class XMLCollectionDestructor implements DestructionToken {
        public void destruct(Object o) {
            XMLCollections.remove(o);
            if (!(o instanceof SchemaXMLCollection) && !(o instanceof MappingXMLCollection)) {
                try {
                    rep.releaseCollectionMetadata(((_RepositoryCollection) o).getCollectionName());
                } catch (RepositoryException e) {
                    throw new RepositoryRuntimeException(e);
                }
            }
        }
    }

    private class XMLStatementDestructor implements DestructionToken {
        public void destruct(Object o) {
            XMLStatements.remove(o);
        }
    }

    ///////////////////////////////////////////////////////////////////////////
    //                              PRIVATE
    ///////////////////////////////////////////////////////////////////////////
    private void protectReadOnlyMethod() throws RepositoryException {
        if (readOnly)
            throw new RepositoryException(RepositoryException.NOT_ALLOWED, "Such operation is forbidden because connection is read-only.");
    }

    private void flush() throws XMLDBCException {
        Iterator it = XMLCollections.iterator();
        while (it.hasNext()) {
            ((_RepositoryCollection) it.next()).flushFilers();
        }
    }

    private void broadcastReadOnly() throws XMLDBCException {
        Iterator it = XMLCollections.iterator();
        while (it.hasNext()) {
            ((XMLCollection) it.next()).setReadOnly(true);
        }
    }

    private void broadcastCommit() throws XMLDBCException {
        Iterator it = XMLCollections.iterator();
        while (it.hasNext()) {
            ((_RepositoryCollection) it.next()).commitPerformed();
        }
    }

    private void broadcastRollback() throws XMLDBCException {
        Iterator it = XMLCollections.iterator();
        while (it.hasNext()) {
            ((_RepositoryCollection) it.next()).rollbackPerformed();
        }
    }

    private void _close() throws XMLDBCException {
        Iterator it;
        if (XMLCollections.size() > 0) {
            it = ((List) XMLCollections.clone()).iterator(); // because of destructor
            while (it.hasNext())
                ((_RepositoryCollection) it.next()).close();

            XMLCollections.clear(); // in case (should not be necessary)
        }

        if (XMLStatements.size() > 0) {
            it = ((HashSet) XMLStatements.clone()).iterator();
            while (it.hasNext()) {
                ((XMLStatement) it.next()).close();
                it.remove();
            }
            XMLStatements.clear();
        }
    }

    private static synchronized void loadXQueryStatementClass() {
        if (stmtConstructor == null) // not synchronized
        {
            Class stmtImpl = null;
            try {
                stmtImpl = Class.forName("org.xquark.mapper.xquery.XStatementImpl");
            } catch (ClassNotFoundException e) {
                throw new RuntimeException("Repository XQuery module is missing in your CLASSPATH: " + e.getMessage());
            }
            Class[] paramTypes = new Class[] { RepositoryConnection.class, DestructionToken.class };
            try {
                stmtConstructor = stmtImpl.getConstructor(paramTypes);
            } catch (NoSuchMethodException e) {
                throw new RuntimeException("Repository XQuery statement constructor not found: " + e.getMessage());
            }
        }
    }
}