/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

/*
 * SQLTableBuilder.java
 *
 * Created on 5 dcembre 2000, 17:22
 */

package org.xquark.mapper.dbms;

import java.io.*;
import java.sql.SQLException;

import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xquark.mapper.RepositoryException;
import org.xquark.mapper.mapping.*;
import org.xquark.mapper.metadata.RepositoryConstants;

/** This utility allows to generate table in the relational database basing on
 * a mapping file and its associated schemas.
 * <p>It is able to directly generate tables through JDBC or generate an SQL script file.</p>
 * <p>This tool must be used after the referenced schemas repository storage with the
 * {@link org.xquark.mapper.Repository#saveSchema saveSchema} method and
 * before mapping storage with {@link org.xquark.mapper.Repository#saveMapping saveMapping}
 * that will refuse execution until relational table exist.
 * @see org.xquark.mapper.Repository#saveSchema
 * @see org.xquark.mapper.Repository#saveMapping
 */
public class SQLTableBuilder implements RepositoryConstants
{
private static final String RCSRevision = "$Revision: 1.2 $"; 
private static final String RCSName     = "$Name: XQ1_1_1 $";
    static StringBuffer pkConstraintStmt = new StringBuffer();
    static StringBuffer fkConstraintStmt1 = new StringBuffer();
    static StringBuffer fkConstraintStmt2 = new StringBuffer();

    public static void generateScripts(_RepositoryConnection repConn, String mappingURL, File wPath)
    throws Exception
    {
        final String SQLCreateFile = "CreateTables.sql";
        final String SQLDropFile = "DropTables.sql";
        
        RepositoryMapping mapping = null;
        
        ///////////////////// Loading Schema(s) and Mapping //////////////////////
        
        System.out.println("Loading Schema and Mapping...");
        try {
            MappingFactory factory = new MappingFactory(
                repConn.getSchemaManager(),
                repConn
                );
            mapping = (RepositoryMapping)factory.createTree();
            SAXParserFactory parserFactory = SAXParserFactory.newInstance();
            parserFactory.setNamespaceAware(true);
            Loader loader = new Loader(parserFactory.newSAXParser().getXMLReader(),
                                        repConn.getConnection(), repConn.getSchemaManager(), true);
            InputSource source = new InputSource(mappingURL);
            mapping = loader.load(source, factory, false);
        }
        catch(SAXException e) {
            throw new RepositoryException(RepositoryException.SAX_OUTPUT_ERROR, e.getMessage(), e.getException());
        }
        catch(SQLException e) {
            throw new RepositoryException(RepositoryException.DB_ERROR, "JDBC error while loading mapping", e);
        }

        ///////////////////// Processing table generation //////////////////////
        
        /* Create output files */
        PrintStream createFile = null;
        PrintStream dropFile = null;
        try {
            createFile = new PrintStream(new FileOutputStream(new File(wPath, SQLCreateFile)), true);
            dropFile = new PrintStream(new FileOutputStream(new File(wPath, SQLDropFile)), true);
        }
        catch(FileNotFoundException e) {
            throw new RepositoryException(RepositoryException.SYSTEM_ERROR, "Could not open files to generate", e);
        }
            
        /* Generate output file */
        System.out.println("Generating SQL script...");
        generateSQL(repConn, createFile, dropFile, mapping);
        createFile.close();
        dropFile.close();
    }
            
    private static void generateSQL(_RepositoryConnection repConn,
            PrintStream outPrinter,
            PrintStream dropPrinter,
            RepositoryMapping mapping)
    throws RepositoryException
    {
        MappingSetIterator itt = mapping.getSortedTableMappingSetIterator(false);
        TableMapping tableMapping;
        
        /* Table loop for creation statements */
        while (itt.hasNext())
        {
            pkConstraintStmt.setLength(0);
            fkConstraintStmt1.setLength(0);
            fkConstraintStmt2.setLength(0);
            boolean firstColumn = true;
            tableMapping = itt.next().getTableMapping();

            /* Create statement */
            outPrinter.print("CREATE TABLE ");
            outPrinter.print(tableMapping.getTableName());
            outPrinter.println("(");
            
            /* Columns loop */
            ColumnMapping[] columns = tableMapping.getColumnMappings();
            for(int i = 0; i < columns.length; i++)
            {
                if (firstColumn)
                    firstColumn = false;
                else
                    outPrinter.println(",");    // REGLER LE PB DE LA VIRGULE
                generateSQLforColumn(repConn, outPrinter, columns[i], mapping);
            }
            /* Writing pk constraint */
            if (pkConstraintStmt.length() > 0)
            {
                pkConstraintStmt.setLength(pkConstraintStmt.length() - 1);
                outPrinter.println(",");
                outPrinter.print("PRIMARY KEY(" + pkConstraintStmt + ")");
            }
            /* Writing fk constraint */
            if (fkConstraintStmt1.length() > 0)
            {
                fkConstraintStmt1.setLength(fkConstraintStmt1.length() - 1);
                fkConstraintStmt2.setLength(fkConstraintStmt2.length() - 1);
                outPrinter.println(",");
                outPrinter.print("FOREIGN KEY(" + fkConstraintStmt1 + ")");
                outPrinter.print(" REFERENCES " + fkConstraintStmt2 + ")");
            }
            /* Ending Create statement */
            outPrinter.println();
            outPrinter.println(");");
            outPrinter.println();
        }
        
        itt = mapping.getSortedTableMappingSetIterator(true);
        
        /* Table loop for drop statements */
        while (itt.hasNext())
        {
            tableMapping = itt.next().getTableMapping();
            /* Create statement */
            dropPrinter.print("DROP TABLE ");
            dropPrinter.println(tableMapping.getTableName() + ";");
        }
    }
    
    private static void generateSQLforColumn(_RepositoryConnection repConn, 
                                            PrintStream outPrinter,
                                            ColumnMapping mapping,
                                            RepositoryMapping repMapping)
    throws RepositoryException
    {
        TableMapping refTableMapping = null;
        ColumnMapping refColumnMapping = null;
        StringBuffer columnStmt = new StringBuffer();
        
        /* writing column definition */
        columnStmt.append(mapping.getColumnName());
        
        /* Column type */
        columnStmt.append(" ");
        columnStmt.append(mapping.getTypeInfo().getTypeCreationString(repConn.getConnection()));
        
        /* Primary key (if not, nullable for simplification) */
        if (mapping.getKeyColumnIndex() != -1)
        {
            pkConstraintStmt.append(mapping.getColumnName());
            pkConstraintStmt.append(",");
            columnStmt.append(" NOT NULL");
        }
        outPrinter.print(columnStmt);
        
        /* Foreign key */
        if (mapping.getTableRefIndex() != -1)
        {
            refTableMapping =  repMapping.getTableMapping(mapping.getTableRefIndex());
            refColumnMapping = ((ColumnRefGenerator)mapping.getGenerator()).getColumnRef();
            fkConstraintStmt1.append(mapping.getColumnName());
            fkConstraintStmt1.append(",");
            
            if (fkConstraintStmt2.length() == 0)
            {
                fkConstraintStmt2.append(refTableMapping.getTableName());
                fkConstraintStmt2.append("(");
            }
            fkConstraintStmt2.append(refColumnMapping.getColumnName());
            fkConstraintStmt2.append(",");
        }
    }    
}
