/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

/*
 * Sequence.java
 *
 * Created on 11 septembre 2002, 09:58
 */

package org.xquark.mapper.dbms;

import java.sql.SQLException;

import org.xquark.mapper.RepositoryException;

/**
 * This class encapsulates a database unique integers generators usually knows
 * as sequences. This objects allows ID preallocation to limit access to the
 * underlying database. Note that if the preallocation size becomes 0, the
 * sequence becomes all-in-memory (useful for monouser).
 *
 */
public class Sequence
{
	private static final String RCSRevision = "$Revision: 1.2 $";
	private static final String RCSName = "$Name: XQ1_1_1 $";

    private TableInfo seq;
    private short preAllocation;
    private long lastDBValue;
    private long currentValue;
    private long maxValue;
    private boolean initialized = false;
    
    /** Creates a new instance of Sequence */
    public Sequence(TableInfo seq, short preAllocation)
    {
        this.seq = seq;
        this.preAllocation = preAllocation;
        maxValue = (long) (Math.pow(10, seq.getColumns()[0].getPrecision()) - 1f);
    }
    
    /** Get a new sequence value */
    public long nextValue(AbstractConnection connection) throws RepositoryException
    {
        if (currentValue == maxValue)
            throw new RepositoryException(RepositoryException.NOT_ALLOWED,
            "The sequence " + seq.getName() + " has reached its maximum value.");
        if (!initialized || (currentValue == (lastDBValue + preAllocation - 1)))
            currentValue = lastDBValue = connection.nextSequenceValue(seq, preAllocation);
        else
            currentValue++;
        initialized = true;
        return currentValue;
    }
    
    /** Get the current sequence value for the current object (not the absolute
     * last value allocated by any Repository instance.
     */
    public long currentValue() throws RepositoryException
    {
        if (initialized)
            return currentValue;
        else
            throw new RepositoryException(RepositoryException.NOT_ALLOWED,
            "The sequence was not initialized.");
    }
    
    /** Create.
     */
    public void create(AbstractConnection connection) throws RepositoryException
    {
        if (initialized)
            throw new RepositoryException(RepositoryException.NOT_ALLOWED,
            "The sequence is already initialized.");
        else
        {
            try
            {
                connection.createSequence(seq, preAllocation);
            }
            catch (SQLException e)
            {
                throw new RepositoryException(RepositoryException.DB_ERROR,
                "JDBC error while creating the "+ seq.getName() + " sequence.");
            }
            initialized = true;
        }
    }
    
    public TableInfo getTableInfo()
    {
        return seq;
    }
}
