/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

/*
 * ColumnSpec.java
 *
 * Created on 11 janvier 2001, 16:57
 */

package org.xquark.mapper.dbms;

import java.sql.SQLException;

import org.xquark.mapper.mapping.TableMetaData;
import org.xquark.mapper.mapping.TableMetaDataImpl;

/**
 * Class containing information on collections tables constructed from repository
 * configuration (no JDBC Access).
 *
 */
public class TableInfo
{
    private static final String RCSRevision = "$Revision: 1.2 $";
    private static final String RCSName     = "$Name: XQ1_1_1 $";

    public static short NOT_A_PARAM = 0;
    public static short PARAM_DATA_SIZE = 1;
    public static short PARAM_EXTRA_SIZE = 2;
    public static short PARAM_CID_SIZE = 3; // numerics are in bits if scale=0 !
    public static short PARAM_DID_SIZE = 4;
    public static short PARAM_BID_SIZE = 5;
	public static short PARAM_UDID_SIZE = 6;
    
    public static int PARAM_NUMBER = 7;
        
    private TableSpec template = null;
    private ColumnInfo[] columns;
    private String[] indexesCreationStatements = null;
    private String[] indexesDropStatements = null;
	private String selectAllStatement = null;
	private String columnList = null;
	private String logicalSpecification = null;
    private String insertStatement = null;
    private String name = null;
    private String creationStatement = null;
    private TableMetaData tableMetaData = null;

    //////////////////////////////////////////////////////////////////////////
    //                          CONSTRUCTORS
    //////////////////////////////////////////////////////////////////////////
    /**
     * For Repository tables.
     * @param template
     */    
    public TableInfo(TableSpec template, long[] sizeParameters, AbstractConnection dbmsInfo)
    {
        /* order matters */
        this.template = template;
        name = template.getTableName();
        initializeIndexes(dbmsInfo);
        init(template, sizeParameters, dbmsInfo);
    }
    
    /**
     * @param template
     * @param cid
     * @param sizeParameters an array with the parameter ID as index.
     */    
    public TableInfo(TableSpec template, short cid, long[] sizeParameters, AbstractConnection dbmsInfo)
    {
        /* order matters */
        this.template = template;
        name = template.getTableName(cid);
        initializeIndexes(cid, dbmsInfo);
        init(template, sizeParameters, dbmsInfo);
    }
    
    public TableInfo(TableSpec template, short cid, AbstractConnection dbmsInfo)
    {
        /* order matters */
        this.template = template;
        name = template.getTableName(cid);
        initializeIndexes(cid, dbmsInfo);
        init(template, new long[PARAM_NUMBER], dbmsInfo);
    }
    
    public TableInfo(TableSpec template, short cid, short tid, long[] sizeParameters, AbstractConnection dbmsInfo)
    {
        /* order matters */
        this.template = template;
        name = template.getTableName(cid, tid);
        initializeIndexes(cid, tid, dbmsInfo);
        init(template, sizeParameters, dbmsInfo);
    }
    
    private void init(TableSpec template, long[] sizeParameters, AbstractConnection dbmsInfo)
    {
        /* order matters */
        this.template = template;
        initializeColumns(sizeParameters, dbmsInfo);
        generateCreationStatement(dbmsInfo);
        initStatements();
    }
    
    //////////////////////////////////////////////////////////////////////////
    //                          ACCESSORS
    //////////////////////////////////////////////////////////////////////////
    public String getName()
    {
        return name;
    }
    
    public String getCreationStatement()
    {
        return creationStatement;
    }
    
	public String getSelectAllStatement()
	{
		return selectAllStatement;
	}

	public String getColumnList()
	{
		return columnList;
	}

	public String getLogicalSpecification()
	{
		return logicalSpecification;
	}

    public String getInsertStatement()
    {
        
        return insertStatement;
    }
    
    public byte getType()
    {
        return template.getType();
    }
    
    public TableSpec getTemplate()
    {
        return template;
    }
    
    public boolean isSequence()
    {
        return template.isSequence();
    }
    
    public ColumnInfo[] getColumns()
    {
        return columns;
    }
    
    public ConstraintSpec[] getConstraints()
    {
        return template.getConstraints();
    }
    
    public IndexSpec[] getIndexes()
    {
        return template.getIndexes();
    }
    
    public String[] getIndexCreationStatements()
    {
        return indexesCreationStatements;
    }
    
    public String[] getIndexDropStatements()
    {
        return indexesDropStatements;
    }
    
    public TableMetaData getJDBCMetaData(AbstractConnection conn)
    throws SQLException
    {
        if (tableMetaData == null)  // not all tableInfo will need this (almost data table)
            tableMetaData = new TableMetaDataImpl(name, conn);
        return tableMetaData;
    }
    
    //////////////////////////////////////////////////////////////////////////
    //                          PRIVATE
    //////////////////////////////////////////////////////////////////////////
    private void initializeColumns(long[] sizeParameters, AbstractConnection dbmsInfo)
    {
        ColumnSpec[] templateColumns = template.getColumns();
        columns = new ColumnInfo[templateColumns.length];
        
        for (int i = 0; i < templateColumns.length; i++)
            columns[i] = new ColumnInfo(templateColumns[i], sizeParameters, dbmsInfo);
    }
    
    private void initializeIndexes(short cid, short tid, AbstractConnection dbmsInfo)
    {
        IndexSpec[] indexes = template.getIndexes();
        indexesCreationStatements = new String[indexes.length];
        indexesDropStatements = new String[indexes.length];
        
        for (int i = 0; i < indexes.length; i++)
        {
            indexesCreationStatements[i] = indexes[i].generateCreationStatement(cid, tid, dbmsInfo);
            indexesDropStatements[i] = indexes[i].generateDropStatement(cid, tid);
        }
    }
    
    private void initializeIndexes(short cid, AbstractConnection dbmsInfo)
    {
        IndexSpec[] indexes = template.getIndexes();
        indexesCreationStatements = new String[indexes.length];
        indexesDropStatements = new String[indexes.length];
        
        for (int i = 0; i < indexes.length; i++)
        {
            indexesCreationStatements[i] = indexes[i].generateCreationStatement(cid, dbmsInfo);
            indexesDropStatements[i] = indexes[i].generateDropStatement(cid);
        }
    }
    
    private void initializeIndexes(AbstractConnection dbmsInfo)
    {
        IndexSpec[] indexes = template.getIndexes();
        indexesCreationStatements = new String[indexes.length];
        indexesDropStatements = new String[indexes.length];
        
        for (int i = 0; i < indexes.length; i++)
        {
            indexesCreationStatements[i] = indexes[i].generateCreationStatement(dbmsInfo);
            indexesDropStatements[i] = indexes[i].generateDropStatement();
        }
    }
    
    private void initStatements()
    {
        /* select all */
    	StringBuffer sql = new StringBuffer();
        for (int i = 0; i < columns.length; i++)
        {
        	sql.append(columns[i].getName());
        	sql.append(',');
        }
    	sql.setLength(sql.length() - 1);
    	columnList = sql.toString();
    	sql.setLength(0);
    	sql.append("SELECT ");
    	sql.append(columnList);
    	sql.append(" FROM ");
        sql.append(name);
        selectAllStatement = sql.toString();
        
        /* insert */
        sql.setLength(0);
        sql.append("INSERT INTO ");
        sql.append(name);
        sql.append(" VALUES(");
        for (int i = 0; i < columns.length; i++)
            sql.append("?,");
        sql.setCharAt(sql.length() - 1, ')');
        insertStatement = sql.toString();
    }
    
    
    private void generateCreationStatement(AbstractConnection dbmsInfo)
    {
    	StringBuffer statementBuffer = new StringBuffer();
    	StringBuffer pk = new StringBuffer();
    	/* Columns loop */
    	for (int i = 0; i < columns.length; i++)
    	{
    		statementBuffer.append(columns[i].getColumnClause());
    		statementBuffer.append(',');

    		if (columns[i].isPrimaryKey())
    		{
    			pk.append(columns[i].getName());
    			pk.append(',');
    		}
    	}
    	/* Writing pk constraint */
    	if (pk.length() > 0)
    	{
    		pk.setCharAt(pk.length() - 1, ')');
    		pk.append(',');

    		statementBuffer.append("PRIMARY KEY(");
    		statementBuffer.append(pk.toString());
    	}

    	/* Writing table constraints */
    	ConstraintSpec[] constraints = template.getConstraints();
    	for (int i = 0; i < constraints.length; i++)
    	{
    		statementBuffer.append(constraints[i].generateConstraintClause());
    		statementBuffer.append(',');
    	}

    	/* Ending Create statement removing last coma */
    	statementBuffer.setLength(statementBuffer.length() - 1);

		logicalSpecification = statementBuffer.toString();

        
        /* Writing physical clause */
        switch (template.getDBMSType())
        {
            case TableSpec.TABLE_TEMPORARY:
                creationStatement = dbmsInfo.getTemporaryTableCreationStatement(this);
                break;
            case TableSpec.TABLE_INDEX_ORGANIZED:
        		creationStatement = dbmsInfo.getIndexOrganizedTableCreationStatement(this);
                break;
            default:
        	    statementBuffer.setLength(0);
        		statementBuffer.append("CREATE TABLE ");
        		statementBuffer.append(getName());
        		statementBuffer.append('(');
        		statementBuffer.append(logicalSpecification);
        		statementBuffer.append(')');
        		creationStatement = statementBuffer.toString();
       }
    }
 }
