/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

/*
 * ColumnSpec.java
 *
 * Created on 11 janvier 2001, 16:57
 */

package org.xquark.mapper.dbms;

import java.util.Arrays;

import org.xquark.mapper.metadata.RepositoryConstants;

/**
 * Class containing information for relational table creation (shared object).
 *
 */
public class TableSpec
{
    private static final String RCSRevision = "$Revision: 1.2 $";
    private static final String RCSName     = "$Name: XQ1_1_1 $";

    /* table types (IDs) */
    public static final byte TYPE_REPOSITORY = 0;
    public static final byte TYPE_COLLECTIONS = 1;
    public static final byte TYPE_TABLES = 2;
    public static final byte TYPE_TABLE_SEQ = 3;
    public static final byte TYPE_PATHS = 4;
    public static final byte TYPE_DUAL = 5;
    public static final byte TYPE_PATH_SEQ = 6;
    public static final byte TYPE_DOCS = 7;
    public static final byte TYPE_DOC_SEQ = 8;
    public static final byte TYPE_FREE_DOC = 9;
    public static final byte TYPE_STRUCT = 10;
    public static final byte TYPE_DATA = 11;
    public static final byte TYPE_USER_DATA = 12;
    public static final byte TYPE_EXTRA = 13;
    public static final byte TYPE_QUERY_TMP = 14;
    
    /* Categories */
    public static final byte CAT_QUERY_TMP = 0;
    public static final byte CAT_REPOSITORY = 1;
    public static final byte CAT_COLLECTION = 2;
    public static final byte CAT_USER_TABLE = 3;
    
    public static final int CAT_COUNT = 4;
    
    /* dbms table types */
    public static final byte TABLE_DEFAULT = 0;
    public static final byte TABLE_SEQUENCE = 1;
    public static final byte TABLE_INDEX_ORGANIZED = 2;
    public static final byte TABLE_TEMPORARY = 3;
    
    private static final byte[] tableTypes =
    {TABLE_DEFAULT, TABLE_INDEX_ORGANIZED, TABLE_SEQUENCE, TABLE_TEMPORARY};
     /* Alphabetically sorted */
     private static final String[] tableTypeStrings =
     {"default", "index_organized", "sequence", "temporary"};
      
    /* dbms types */
    public static final short ANY = 0;
    // other values are defined in AbstractConnectionFactory
    
    private byte type;
    private byte dbmsType;
    private byte cat;
    
    private boolean sequence = false;
    private short dbms;
    private String suffix;
    private ColumnSpec[] columns;
    private ConstraintSpec[] constraints;
    private IndexSpec[] indexes;

    //////////////////////////////////////////////////////////////////////////
    //                          CONSTRUCTORS
    //////////////////////////////////////////////////////////////////////////
    public TableSpec() {}
    
    public TableSpec(byte cat, byte type)
    {
        this.cat = cat;
        this.type = type;
    }
    
    public TableSpec(byte cat, byte type, boolean sequence,
                                String suffix, short dbms, String dbmsType)
    {
        set(cat, type, sequence, suffix, dbms, dbmsType);
    }
    
    public String getTableName()
    {
        StringBuffer tableName = new StringBuffer();
        tableName.append(RepositoryConstants.TABLE_PREFIX);
        tableName.append(suffix);
        
        return tableName.toString();
    }
    
    public String getTableName(short collectionID)
    {
        StringBuffer tableName = new StringBuffer();
        tableName.append(RepositoryConstants.TABLE_PREFIX);
        tableName.append(collectionID);
        tableName.append('_');
        tableName.append(suffix);
        
        return tableName.toString();
    }
    
    public String getAllCollectionsWildcardTableName()
    {
        StringBuffer tableName = new StringBuffer();
        tableName.append(RepositoryConstants.TABLE_PREFIX);
        tableName.append("%_");
        tableName.append(suffix);
        
        return tableName.toString();
    }
    
    public String getTableName(short collectionID, short tableID)
    {
        StringBuffer tableName = new StringBuffer();
        tableName.append(RepositoryConstants.TABLE_PREFIX);
        tableName.append(collectionID);
        tableName.append('_');
        tableName.append(suffix);
        tableName.append(tableID);
        
        return tableName.toString();
    }
    
    //////////////////////////////////////////////////////////////////////////
    //                          ACCESSORS
    //////////////////////////////////////////////////////////////////////////
    public void set(byte cat, byte type, boolean sequence,
                                String suffix, short dbms, String dbmsType)
    {
        this.cat = cat;
        this.type = type;
        this.sequence = sequence;
        this.suffix = suffix;
        this.dbms = dbms;
        this.dbmsType = parseDBMSType(dbmsType);
    }
    
    public String getSuffix()
    {
        return suffix;
    }
    
    public byte getType()
    {
        return type;
    }
    
    public byte getDBMSType()
    {
        return dbmsType;
    }
    
    public short getDBMSVendor()
    {
        return dbms;
    }
    
    public boolean isSequence()
    {
        return sequence;
    }
    
    public ColumnSpec[] getColumns()
    {
        return columns;
    }
    
    public ConstraintSpec[] getConstraints()
    {
        return constraints;
    }
    
    public IndexSpec[] getIndexes()
    {
        return indexes;
    }
    
    public void setSuffix(String suffix)
    {
        this.suffix = suffix;
    }
    
    public void setColumns(ColumnSpec[] columns)
    {
        this.columns = columns;
    }

    public void setConstraints(ConstraintSpec[] constraints)
    {
        this.constraints = constraints;
    }

    public void setIndexes(IndexSpec[] indexes)
    {
        this.indexes = indexes;
    }

    private byte parseDBMSType(String type)
    {
        int index = Arrays.binarySearch(tableTypeStrings, type);
        if (index < 0)
            throw new RuntimeException("The DBMS table type" + type + " used in tableSpec.xml is unknown.");
        return tableTypes[index];
    }
}
