/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.mapper.dbms.oracle;

import java.sql.*;
import java.util.*;

import org.xquark.jdbc.typing.DBMSInfo;
import org.xquark.mapper.RepositoryException;
import org.xquark.mapper.dbms.*;
import org.xquark.mapper.metadata.Repository;
import org.xquark.mapper.metadata.RepositoryConstants;
import org.xquark.mapper.util.LongList;

/**
 * Implementation of AbstractConnection for Oracle 8 databases.
 *
 */
public class O8iConnection extends JDBC2Connection
{
    private static final String RCSRevision = "$Revision: 1.4 $"; 
    private static final String RCSName     = "$Name: XQ1_1_1 $";
    
    private static final String SEQ_QUERY = "select SEQUENCE_NAME from USER_SEQUENCES where SEQUENCE_NAME LIKE('";
    
   /** Creates new O8iConnection */
    public O8iConnection(Connection connection, short dbms, DBMSInfo dbmsInfo)
    throws SQLException
    {
        super(connection, dbms, dbmsInfo);
    }

    ///////////////////////////////////////////////////////////////
    //                      Table methods
    ///////////////////////////////////////////////////////////////
    public void dropTemporaryTable(String tableName) throws SQLException
    {
        Statement stmt = connection.createStatement();
        try
        {
            stmt.executeUpdate("TRUNCATE TABLE "+ tableName);
        }
        finally
        {
            stmt.close();
        }
        dropTable(tableName);
    }
    
    public void createSequence(TableInfo table, short step) throws SQLException
    {
        StringBuffer sql = new StringBuffer();
        sql.append("CREATE SEQUENCE ");
        sql.append(table.getName());
        sql.append(" START WITH ");
        sql.append(RepositoryConstants.SEQUENCE_FIRST_VALUE);
        sql.append(" INCREMENT BY ");
        sql.append(step);
        executeUpdate(sql.toString());
    }
    
    public void dropSequence(String seqName) throws SQLException
    {
        executeUpdate("DROP SEQUENCE " + seqName);
    }
    
    public long nextSequenceValue(TableInfo table, short step) throws RepositoryException
    {
        long next = 1;
        Statement stmt = null;
        ResultSet rs = null;
        StringBuffer sql = new StringBuffer();
        sql.append("SELECT ");
        sql.append(table.getName());
        sql.append(".NEXTVAL FROM DUAL");
        try
        {
            stmt = connection.createStatement();
            rs = stmt.executeQuery(sql.toString());
            if (rs.next())
                next = rs.getLong(1);
        }
        catch (SQLException e)
        {
            throw new RepositoryException(RepositoryException.DB_ERROR,
            "JDBC error while querying the " + table.getName() + " table.");
        }
        finally
        {
            try
            {
                if (rs != null)
                {
                    rs.close();
                    stmt.close();
                }
            }
            catch (SQLException e)
            {
                // no op
            }
        }
        return next;
    }
    
    public List getUserSequenceNames(String seqPattern) throws SQLException
    {
        Statement stmt = null;
        ResultSet rs = null;
        ArrayList seqs = new ArrayList();
        
        StringBuffer sql = new StringBuffer();
        sql.append(SEQ_QUERY);
        sql.append(seqPattern);
        sql.append("')");
        try
        {
            stmt = connection.createStatement();
            rs = stmt.executeQuery(sql.toString());
            while (rs.next())
                seqs.add(rs.getString(1));
        }
        finally
        {
            if (rs != null)
                rs.close();
            if (stmt != null)
                stmt.close();
        }
        return seqs;
    }
    
    ///////////////////////////////////////////////////////////////
    //                      Private utilities
    ///////////////////////////////////////////////////////////////
    /** Returns a hierachical (or an emulation) SQL statement used by the
     * longest path set to remove paths.
     * @return an SQL statement.
     */    
    private StringBuffer hierarchicalQueryString(LongList path,String select, String tableName)
    {
        StringBuffer sql = new StringBuffer();

        int len = path.size();
        sql.append("SELECT "+select);
        sql.append(" FROM "+ tableName);
        sql.append(" START WITH label = '"+path.getValue(0)+"' ");
        sql.append("CONNECT BY (PRIOR target = source) AND (");
        for(int i=1; i<len; i++) {
            sql.append(" (level="+(i+1)+" AND label='"+path.getValue(i)+"') OR ");
        }
        if (len>1) sql.setLength(sql.length()-4);
        else sql.setLength(sql.length()-8);
        sql.append(") ");
        // __// __debug.pop();
        return sql;
    }
   
    /** Returns a hierachical (or an emulation) SQL statement used by the
     * longest path set to remove paths.
     * @return an SQL statement.
     */    
    private StringBuffer reversedHierarchicalQueryString(LongList path, String select, String tableName)
    {
        StringBuffer sql = new StringBuffer();

        int len = path.size();
        sql.append("select "+select);
        sql.append(" from " + tableName);
        sql.append(" start with label = '"+path.getValue(len-1)+"' \n");
        sql.append("connect by (prior source = target) AND (\n");
        for(int i=len-2; i>=0; i--) {
            sql.append(" (level="+(len-i)+" and label='"+path.getValue(i)+"') or \n");
        }
        if(len>1) sql.setLength(sql.length()-4);
        else sql.setLength(sql.length()-8);
        sql.append(") \n");
        return sql;
    }
    
    public int getUniversalError(SQLException e)
    {
        switch (e.getErrorCode())
        {
			case 955:
				return OBJECT_ALREADY_EXISTS; 
			case 942:
				return OBJECT_DOES_NOT_EXISTS; 
            default:
                return super.getUniversalError(e);
        }
    }
    
    public void updateStatistics(String tableName) throws SQLException
    {
        executeUpdate("ANALYZE TABLE " + tableName + " ESTIMATE STATISTICS");
    }
    
    /** This method is called when a repository is initialized to perform
     * dbms preparation.
     */
    public void onInitRepository(Repository rep) throws SQLException
    {
        Statement stmt = null;
        try
        {
	    // create dual table
        String dualTableName = rep.getTableInfo(TableSpec.TYPE_DUAL).getName();
	    stmt = connection.createStatement();
	    stmt.executeUpdate("INSERT INTO " + dualTableName + " VALUES('X')");
	    stmt.executeUpdate("ANALYZE TABLE " + dualTableName + " COMPUTE STATISTICS");
        }
        finally
        {
            if (stmt != null)
                stmt.close();
        }
    }
    public void onDeleteRepository() throws SQLException
    {
    }
    
    public boolean distinguishNullAndEmptyStrings()
    {
        return false;
    }
    
    ///////////////////////////////////////////////////////////////
    //                   Overidding static accessors
    ///////////////////////////////////////////////////////////////
    public String getBitmapIndexClause()
    {
        return "BITMAP";
    }
}
