/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.mapper.mapping;

import java.util.*;

import org.xml.sax.SAXException;
import org.xquark.mapper.dbms.AbstractConnection;
import org.xquark.mapper.metadata.RepositoryConstants;
import org.xquark.schema.SchemaComponent;
import org.xquark.schema.SchemaConstants;

/**
 * Implementation of the TableMapping interface for 'unmapped' elements or attributes.
 *
 * Elements or attributes are mapped to the default XML data table.
 *
 */
public abstract class BaseTableMappingImpl extends MappingImpl
implements TableMapping, MappingConstants, RepositoryConstants, SchemaConstants
{
    private static final String RCSRevision = "$Revision: 1.3 $";
    private static final String RCSName     = "$Name: XQ1_1_1 $";
    
    // primary key or equivalent used for making join
    protected ColumnMapping[] primaryKeyColumnMappings;
    
    /** Number of columns that make up the primary key and are used for the join
     * between the user and the OID table.
     * Key generators during construction then PK if any and finally key
     * generators else.
     */
    protected int keyColumnCount = 0;
    
    /* List of ColumnMapping of this TableMapping */
	protected ColumnMapping[] columnMappings;
	
    /** List of ColumnMapping that must be initialized before reading the
     * XML data (startElement()).
     */
    protected List initialValues;
    /** List of ColumnMapping that must be initialized after reading the
     * XML data (endElement()) or whent the tuple is supposed to be completed (retrieve then the context...).
     */
    protected List finalValues;
    
    /* Name of the relational table */
    protected String name;
    /* Object encapsulating the TableMappings for a repository */
    protected RepositoryMapping colMapping;
    protected int index;
    
    /* Object containing table metadata, ie. properties */
    protected TableMetaData tableMetaData;
    
    protected Collection dependencies = null;
    
    /* String for JDBC PreparedStatements used to create or destroy the user table */
    /* String for JDBC PreparedStatements used to access the user table */
    protected String insertStatement;
    // make a natural join between OID & values tables
    protected String joinCondition = "";
    protected String columnList = "";
    protected String namedColumnList = "";
    protected String wildcards = "";
    protected String joinColumnList = "";
    protected String joinWildcards = "";
    protected String joinColumnTypes = "";
    
    /* Statement column lists */
    protected ColumnMapping[] updateColumnMappings;
    protected ColumnMapping[] fetchColumnMappings;
    protected ColumnMapping[] selectColumnMappings;
    protected ColumnMapping[] OIDColumnMappings;
    
    /** Constructor.
     * @param colMapping the RepositoryMapping object to which this object belongs.
     * @param comp the SchemaComponent associated with this "mapping component".
     * @param name relational table name.
     * @param conn a connection allowing access to the metadata for the table retrieved from JDBC
     * @param action the action to perform on the rows of the table (check, insert, update, ...)
     * @param generate If true, mapping is loaded even if relational tables
     * are not present : mapping is loaded for table generation.
     */
    protected BaseTableMappingImpl(RepositoryMapping colMapping, SchemaComponent comp, AbstractConnection conn) throws SAXException
    {
        super(comp);
        this.colMapping = colMapping;
        initialValues = new ArrayList(5);
        finalValues = new ArrayList(3);
    }
    
    /**
     * Accessors.
     */
    
    public String toString()
    {
        return name;
/*        String result = "#<TableMapping "+name+"\n";
        result+="  "+getSchemaComponent()+"\n";
        Iterator it = columnMappings.iterator();
        while (it.hasNext()) result+="  "+it.next()+"\n";
        //      if (dependencies != null) {
        //        it = dependencies.iterator();
        //        while (it.hasNext()) result+="  "+((TableMappingImpl)it.next()).getTableName()+"\n";
        //      }
        //it = mappedElements.iterator();
        //while (it.hasNext()) result+="  "+it.next()+"\n";
        return result+">";
 */
    }
    
    /** Overriden because default mapping uses views on real table mapping.
     * As query reconstruction use equality, we use what realy identifies the
     * table mapping: its index.
     */
    public boolean equals(Object o)
    {
        boolean ret = false;
        if (o instanceof TableMapping)
            ret = (((TableMapping)o).getTableIndex() == getTableIndex());
        return ret;
    }
    /** Overriden because default mapping uses views on real table mapping.
     * (As query reconstruction use a map).
     */
    public int hashCode()
    {
        return getTableName().hashCode();
    }
    
    
    //    public abstract int getBatchSize();
    
	public int getColumnCount()
	{
		return columnMappings.length;
	}

//	public ColumnMapping[] getColumnMappings();
	protected Iterator getColumnMappingIterator()
	{
		return new Iterator()
		{
			private int next = 0;
			public boolean hasNext()
			{
				while (next < columnMappings.length && columnMappings[next] == null)
					next++;
				return next < columnMappings.length && columnMappings[next] != null;
			}

			public Object next()
			{
				ColumnMapping ret = null;
				if (hasNext())
					ret = columnMappings[next++];
				return ret;
			}

			public void remove()
			{
				throw new UnsupportedOperationException();
			}
		};
	}

    public ColumnMapping getColumnMapping(String columnName)
    {
    	Iterator it = getColumnMappingIterator();
    	ColumnMappingImpl column;
    	while (it.hasNext())
    	{
    		column = (ColumnMappingImpl)it.next();
            if (column.getColumnName().equalsIgnoreCase(columnName)) return column; // SR 18/12/2000 Oracle metadata in uppercase : TO IMPROVE
        }
        return null;
    }
    
    public ColumnMapping getColumnMapping(int index)
    {
        return columnMappings[index];
    }
    
    public ColumnMapping[] getPrimaryKey()
    {
        return this.primaryKeyColumnMappings;
    }
    
    public int getKeyColumnCount()
    { return keyColumnCount;}
    
    public List initParameters()
    { return initialValues;}
    
    public List finalParameters()
    { return finalValues;}
    
    public String getTableName()
    {
        return name;
    }
    
    public RepositoryMapping getRepositoryMapping()
    {
        return colMapping;
    }
    
    public TableMetaData getMetaData()
    {
        return tableMetaData;
    }
    
    public int getTableIndex()
    {
        return index;
    }
    
    public void setIndex(int index)
    {
        this.index = index;
    }
    
    /**
     * Returns the name of the column containing the path OID (both in clustered
     * and not clustered case).
     */
    public String getPathIDColumnName()
    {
        return "path";  // TO IMPROVE : should not be hardcoded
    }
    
    /**
     * Returns the name of the column containing the node UOID (both in clustered
     * and not clustered case).
     */
    public String getUOIDColumnName()
    {
        return "uoid";  // TO IMPROVE : should not be hardcoded
    }
    
    public String getJoinCondition()
    {
        return joinCondition;
    }
    public String getColumnList()
    {
        return columnList;
    }
    public String getNamedColumnList()
    {
        return namedColumnList;
    }
    public String getWildcards()
    {
        return wildcards;
    }
    public String getJoinColumnList()
    {
        return joinColumnList;
    }
    public String getJoinWildcards()
    {
        return joinWildcards;
    }
    public String getJoinColumnTypes()
    {
        return joinColumnTypes;
    }

    public Collection getDependencies()
    {
        return dependencies;
    }
    
    public ColumnMapping[] getUpdateColumns()
    {
        return updateColumnMappings;
    }
    
    public ColumnMapping[] getFetchColumns()
    {
        return fetchColumnMappings;
    }

    public ColumnMapping[] getJoinColumns()
    {
        return OIDColumnMappings;
    }
    
    public ColumnMapping[] getSelectColumns()
    {
        return selectColumnMappings;
    }
    
	public boolean isShareable()
	{
		return tableMetaData.areStatementsShareable();
	}
	
	public boolean isTableShared()
	{
		return tableMetaData.isShared();
	}

	////////////////////////////////////////////////////////////////////////////
	//                      PACKAGE PRIVATE
	////////////////////////////////////////////////////////////////////////////
	/**
	 * Add a ColumnMapping object to the ones that must be initialized before
	 * reading the XML data.
	 * @param mapping the ColumnMapping object that can be initialized.
	 */
	void addToInitialValues(ColumnMapping mapping)
	{
		initialValues.add(mapping);
	}

	/**
	 * Add a ColumnMapping object to the ones that must be initialized after
	 * reading the XML data, just before leaving the element.
	 * @param mapping the ColumnMapping object that can be initialized.
	 */
	void addToFinalValues(ColumnMapping mapping)
	{
		finalValues.add(mapping);
	}

	void setStatementsNotShareable()
	{
		tableMetaData.setStatementsNotShareable();
	}

	void setTableShared()
	{
		tableMetaData.setShared();
	}

	////////////////////////////////////////////////////////////////////////////
	//                      MappingInfo IMPLEMENTATION
	////////////////////////////////////////////////////////////////////////////

    public String getInsertStatement()
    {
        return insertStatement;
    }
    
    public TableMapping getTableMapping()
    {
        return this;
    }
}
