/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.mapper.mapping;

import org.xquark.jdbc.typing.ColumnMetaData;
import org.xquark.mapping.Generator;

/**
 * Interface giving access to information for mapping XML simple elements
 * & attributes on rdbms table columns.
 *
 */
public interface ColumnMapping extends Mapping
{
	/**
	 * Accessor to user column name.
	 * @return the user column name.
	 */
    public String getColumnName();
    
	/**
	 * Accessor to the table mapping which this column mapping belongs.
	 * @return the owner table mapping.
	 */
    public TableMapping getTableMapping();
    
	/**
	 * Accessor to column relational metadata.
	 * @return a ColumnMetaData object
	 */
    public ColumnMetaData getMetaData();
    
	/**
	 * Returns the column index allocated during mapping file loading.
	 * @return an integer index beginning with 0 corresponding to the ordinal
	 * position of the column in JDBC user table metadata.
	 * @see TableMapping#getColumnMapping(int)
	 */
    public int getColumnIndex();
    
	/**
	 * Returns the column index in the "key" columns array returned by
	 * {@link TableMapping#getPrimaryKey()}.
	 * @return the PK array index or -1 if column does not belong to "key"
	 * columns.
	 */
    public int getKeyColumnIndex();
    
	/**
	 * Returns true if the column is involved in join between the OID table and
	 * the user table.
	 * @return false if the column is not involved in join or if there is no
	 * join (clustered).
	 */
    public boolean isInJoin();
    
    /**
     * Returns true if the column is to be filled with null when the element
     * or attribute is missing.
     * @return
     */
    public boolean updateColumnWhenMissing();
    
    /**
	 * Returns the column index in the "join" columns array returned by
	 * {@link TableMapping#getJoinColumns()}.
	 * @return the "join" array index or -1 if column does not belong to
	 * "join" columns.
	 */
   public int getJoinColumnIndex();

	/**
	 * Returns the column index in the "insert" columns array returned by {@link
	 * TableMapping#getColumnMappings()}.
	 * @return the "select" array index or -1 if column does not belong to
	 * "select" columns.
	 */
	public int getInsertColumnIndex();

	/**
	 * Returns the column index in the "select" columns array returned by
	 * {@link TableMapping#getSelectColumns()}.
	 * @return the "select" array index or -1 if column does not belong to
	 * "select" columns.
	 */
	public int getSelectColumnIndex();

	/**
	 * Returns the column index in the "fetch" columns array returned by
	 * {@link TableMapping#getFetchColumns()}.
	 * @return the "fetch" array index or -1 if column does not belong to
	 * "select" columns.
	 */
	public int getFetchColumnIndex();

	/**
	 * Returns the column index in the "update" columns array returned by
	 * {@link TableMapping#getUpdateColumns()}.
	 * @return the "update" array index or -1 if column does not belong to
	 * "update" columns.
	 */
    public int getUpdateColumnIndex();
    
	/**
	 * Accessor to the MappingTypeInfo object of the column mapping which
	 * encapsulates all informations about mapping XML type on SQL type using
	 * JDBC.
	 * @return the MappingTypeInfo object associated with the colum.
	 */
    public MappingTypeInfo getTypeInfo();
    
	/**
	 * Accessor to the generator object associated to this column mapping
	 * which is used to supply the data stored in the column.
	 * The generator may be a user generator or an internal generator
	 * @return an object implementing Generator 
	 */
    public Generator getGenerator();
    
	/**
	 * If the XML node associated to this column is optional or nullable,
	 * returns the generator that will supply the default value.
	 * @return null if no default value is needed for the column.
	 */
    public Generator getDefaultGenerator();
    
	/**
	 * Return the default generator if it exists, the generator elese.
	 * @return a least a generator to call when tuple must be initialized.
	 */
    public Generator getInitGenerator() ;
    
	/**
	 * Returns the index identifying the table mapping that this column mapping
	 * references (feature corresponding to rdbms foreign key).
	 * @return the reference table mapping index.
	 */
    public int getTableRefIndex();
    
	/**
	 * Returns the base SQL type corresponding to the column,
	 * i.e. getMetaData().getDataType(). The base SQL types is a subset of SQL
	 * data types defined in {@link java.sql.Types} which represent a family
	 * of types. Whithin a family mainly size limits differ.
	 * @return a constant defined in {@link java.sql.Types}.
	 * @see ColumnMetaData
	 */
    public int getColumnType();
}
