/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.mapper.mapping;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.xml.sax.SAXException;
import org.xquark.jdbc.typing.ColumnMetaData;
import org.xquark.mapping.Generator;
import org.xquark.mapping.UserGenerator;
import org.xquark.schema.*;
import org.xquark.xml.xdbc.XMLDBCException;
import org.xquark.xml.xdbc.XMLErrorHandler;

/**
 * Implementation of the ColumnMapping interface.
 *
 */
class ColumnMappingImpl extends MappingImpl
implements ColumnMapping, MappingConstants
{
    private static final String RCSRevision = "$Revision: 1.8 $";
    private static final String RCSName     = "$Name: XQ1_1_1 $";

    private static Log log = LogFactory.getLog(ColumnMappingImpl.class);

    private TableMappingImpl tableMapping;
    private String name;
    private int index;
    private int keyColumnIndex = -1;
    private int keyGenColumnIndex = -1;
    private int joinColumnIndex = -1;
    private int primaryKeyColumnIndex = -1;
	private int insertColumnIndex = -1;
	private int selectColumnIndex = -1;
	private int fetchColumnIndex = -1;
    private int updateColumnIndex = -1;
    private Generator generator;
    private MappingTypeInfo typeInfo;
    /** generator to call when initializing the tuple.
     * Contains a constant generator for columns which have a default value in
     * the database
     */
    private Generator defaultGenerator = null;
    private int tableRefIndex = -1;
    private int minOccurs;
    private boolean updateColumnWhenMissing = false;
    private XMLErrorHandler messageHandler = null;	// TODO: is it useful yet ?

    ColumnMappingImpl(TableMappingImpl tableMapping, SchemaComponent comp,
        int minOccurs, String name, int role, Generator generator,
        XMLErrorHandler messageHandler)
    throws SAXException
    {
        super(comp);
        this.name = name;
        this.tableMapping = tableMapping;
        this.messageHandler = messageHandler;
        index = tableMapping.register(this);
        this.minOccurs = minOccurs;
    	insertColumnIndex = tableMapping.incrementColumnMappingCount();
        ColumnMetaData cmeta = getMetaData();
        if (cmeta != null) {
            if (cmeta.isPrimaryKey()) {
                keyColumnIndex = tableMapping.incrementKeyColumnCount();
            }
        }
        if ((role & KEY_ROLE) != 0) // PB IS HERE for COUNTRY
        {
            if ((cmeta != null) && cmeta.isLongType())
                throw new SAXException("A column with a long datatype cannot be \"inKey\".");
            keyGenColumnIndex = tableMapping.incrementKeyGenColumnCount();
        }
        if ((role & SELECT_ROLE) != 0)
        {
            if ((cmeta != null) && cmeta.isLongType() && (tableMapping.getAction() != INSERT))
                throw new SAXException("A column with a long datatype cannot be \"inSelect\".");
            selectColumnIndex = tableMapping.incrementSelectColumnCount();
        }
        else {
        	fetchColumnIndex = tableMapping.incrementFetchColumnCount();
        	if ((role & KEY_ROLE) == 0)    // update all column no inSelect and not inKey
                updateColumnIndex = tableMapping.incrementUpdateColumnCount();
        }
        if (generator != null)
        	setGenerator(generator);
    }

    ColumnMappingImpl(TableMappingImpl tableMapping, SchemaComponent comp,
			int minOccurs, String name, int role, Generator generator,
			boolean updateColumnWhenMissing, XMLErrorHandler messageHandler)
			throws SAXException {
		this(tableMapping, comp, minOccurs, name, role, generator, messageHandler);
		this.updateColumnWhenMissing = updateColumnWhenMissing;
	}
    
    public String toString() {
        return getTableName()+"."+name;
    }

    public String getTableName() { return tableMapping.getTableName();}

    public int getTableIndex() { return tableMapping.getTableIndex();}

    public String getColumnName() { return name;}
    
    public boolean isInJoin() { return joinColumnIndex > -1;}
    
    public int getJoinColumnIndex() { return joinColumnIndex;}

    public ColumnMetaData getMetaData()
    {
        TableMetaData tmeta = tableMapping.getMetaData();
        if (tmeta == null) return null;
        return tmeta.getColumnMetaData(name);
    }

	public int getColumnIndex() { return index; }

	public int getInsertColumnIndex()
	{ return insertColumnIndex; }

    public int getKeyColumnIndex()
    {
        return keyColumnIndex;
    }

    public int getSelectColumnIndex() { return selectColumnIndex;}

    public int getFetchColumnIndex() { return fetchColumnIndex;}

    public int getUpdateColumnIndex() { return updateColumnIndex;}

    public Generator getGenerator() { return generator;}

    public MappingTypeInfo getTypeInfo() { return typeInfo;}
    
    public int getKeyGenColumnIndex() {return keyGenColumnIndex;}
        
    public Generator getDefaultGenerator() { return defaultGenerator;}
    
    public Generator getInitGenerator() {
    	return defaultGenerator == null ? generator : defaultGenerator;
    }

    public int getTableRefIndex() { return tableRefIndex; }

    public int getColumnType() { return getMetaData().getDataType();}

    int getMinOccurs() { return minOccurs;}
    
    void setKeyColumnIndex(int index) { keyColumnIndex = index;}
    
    public void setJoinColumnIndex(int index) {joinColumnIndex = index;}

    void setGenerator(Generator generator) throws SAXException
    {
		this.generator = generator;
		
		if (comp == null)   // no component : real generator (added to initial or generators)
			setTableMappingGenerator(generator);
		else // if default value in the database add the generator to init generators
			checkOccurrenceConstraints(minOccurs);	  // (will be overwriten if the element or attribute is encountered)
				  			
        if (generator instanceof UserGenerator)
            typeInfo = new MappingTypeInfo(((UserGenerator)generator).getXMLType());
        else if (generator instanceof ValueGenerator)
        {
            Type type;
            if (comp instanceof Declaration)
                type = ((Declaration)comp).getType();
            else
                type = (Type)comp;
            typeInfo = new MappingTypeInfo(type.getValueType());
            ((ValueGenerator)generator).setTypeInfo(typeInfo);
        }
        else // ref generator
        {
            ColumnMapping refColumn = null;
            if (generator instanceof ColumnRefGenerator)
                refColumn = ((ColumnRefGenerator)generator).getColumnRef();
            if (refColumn != null)
                typeInfo = refColumn.getTypeInfo();
            if (typeInfo == null)
                throw new SAXException("Cannot reference a ref generator in column "
                        + getMetaData()
                        +". Please directly reference the column referenced by the ref generator");
        }
        
        checkTypeConstraints();
    }

    private void setDefaultGenerator(Generator generator) {
        defaultGenerator = generator;
        tableMapping.addToInitialValues(this);
    }

	void setTableMappingGenerator(Generator generator)
	{
		if (generator instanceof ValueGenerator || generator instanceof ColumnRefGenerator)
			tableMapping.addToFinalValues(this);    // cos' value node is ready at the end of node
		else
			tableMapping.addToInitialValues(this);
	}
	
    private void checkOccurrenceConstraints(int minOccurs)
    throws SAXException
    {
        ColumnMetaData cmeta = getMetaData();
        if (cmeta != null) {
            SchemaComponent comp = getSchemaComponent();
            if (minOccurs == 0) {
                String defVal = cmeta.getDefaultValue();
                if (defVal != null)
                setDefaultGenerator(new ConstantGenerator(defVal));
                else if (cmeta.isNullable())
                setDefaultGenerator(new ConstantGenerator(null));
                else
                throw new SAXException("Optional element or attribute "+comp
                +" cannot be mapped onto " +cmeta
                +". The column should be nullable or have a default value");
            }
            if (comp instanceof ElementDeclaration
            && ((ElementDeclaration)comp).isNillable()
            && !cmeta.isNullable())
            throw new SAXException("Nullable element "+comp
            +"cannot be mapped onto the non-nullable "
            +cmeta);
        }
    }

    private void checkTypeConstraints()
    throws SAXException
    {
        ColumnMetaData cmeta = getMetaData();
        if (cmeta != null) {
            if (cmeta.getRefTable() == null) {
                if (generator instanceof ColumnRefGenerator)
                    // Store dependency
                    this.tableRefIndex = (((ColumnRefGenerator)generator).getColumnRef()).getTableIndex();
            }
            else {
                if (generator instanceof ColumnRefGenerator)
                {
                    ColumnMapping refCol = ((ColumnRefGenerator)generator).getColumnRef();
                    if (!refCol.getTableName().equals(cmeta.getRefTable())
                    || !refCol.getColumnName().equals(cmeta.getRefColumn()))
                        throw new SAXException("The reference mapping from "
                                + cmeta + " to "
                                + refCol.getMetaData() +
                                " is not consistent with existing foreign key constraints");
                    // Store dependency
                    this.tableRefIndex = refCol.getTableIndex();
                }
                else
                {
                    try
                    {
                        XMLDBCException ex = new XMLDBCException("The mapping for "
                                + cmeta + " is not consistent with existing foreign key constraints");
                        if (messageHandler != null)
                            messageHandler.warning(ex);
                        log.warn(ex.getMessage());
                    }
                    catch (XMLDBCException e)
                    {
                        throw new SAXException("User abort", e);
                    }
                }
            }
            typeInfo.checkTargetType(cmeta, messageHandler);
        } // Table generation mode : no metadata, no check but store dependencies
        else if (generator instanceof ColumnRefGenerator) {
            // Store dependency
            this.tableRefIndex = ((ColumnRefGenerator)generator).getColumnRef().getTableIndex();
        }
    }
    
    public TableMapping getTableMapping() { return tableMapping;}
    
	/**
	 * @return Returns the updateColumnWhenMissing.
	 */
	public boolean updateColumnWhenMissing() { return updateColumnWhenMissing;}
}
