/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.mapper.mapping;

import java.sql.SQLException;
import java.sql.Types;
import java.util.*;

import org.xml.sax.SAXException;
import org.xquark.jdbc.typing.ColumnMetaData;
import org.xquark.mapping.Generator;
import org.xquark.mapper.RepositoryException;
import org.xquark.mapper.dbms.*;
import org.xquark.mapper.metadata.CollectionMetadata;
import org.xquark.mapper.util.RepositoryProperties;
import org.xquark.schema.*;
import org.xquark.schema.datatypes.PrimitiveType;

/**
 * Implementation of the TableMapping interface for 'unmapped' elements or attributes.
 *
 * Elements or attributes are mapped to the default XML data table.
 *
 */
public class DefaultTableMapping extends BaseTableMappingImpl
{
    private static final String RCSRevision = "$Revision: 1.8 $";
    private static final String RCSName     = "$Name: XQ1_1_1 $";
    private TableInfo tableInfo;
    private ArrayList valueMappings = new ArrayList(3);
    
    /** Constructor.
     * @param colMapping the RepositoryMapping object to which this object belongs.
     * @param comp the SchemaComponent associated with this "mapping component".
     * @param name relational table name.
     * @param conn a connection allowing access to the metadata for the table retrieved from JDBC
     * @param action the action to perform on the rows of the table (check, insert, update, ...)
     * @param generate If true, mapping is loaded even if relational tables
     * are not present : mapping is loaded for table generation.
     */
    public DefaultTableMapping(CollectionMetadata colMeta, AbstractConnection conn)
    throws SAXException, SQLException, RepositoryException
    {
        super(colMeta.getMapping(), null, conn);
        tableInfo = colMeta.getTableInfo(TableSpec.TYPE_DATA);
        name = tableInfo.getName();
        this.tableMetaData = tableInfo.getJDBCMetaData(conn);
        
    	// create column mapping array
    	columnMappings = new ColumnMapping[tableMetaData.getColumnCount()];

        initialize();
        dependencies = Collections.EMPTY_LIST;
        initialValues = Collections.EMPTY_LIST;  // XQuark index parameters initialized when entering the element
    }
    
    /**
     * Accessors.
     */
    
	public ColumnMapping[] getColumnMappings() {
		return columnMappings;
	}

	public int getColumnMappingCount() {
		return getColumnCount();
	}

	public int getBatchSize() {
		return RepositoryProperties.getIntProperty(CONF_DATA_BATCHSIZE);
	}

	public int getOIDTableColumnCount() {
		return 0;
	}
	public int getSelectColumnCount() {
		return 0;
	}
	public int getFetchColumnCount() {
		return 0;
	}
	public int getUpdateColumnCount() {
		return 0;
	}
	public int getAction() {
		return INSERT;
	}
	public boolean isClustered() {
		return true;
	}
	public int getPathIDIndex() {
		return V_PATHID_INDEX;
	}
	public int getUOIDIndex() {
		return V_UOID_INDEX;
	}

	public boolean isShareable() {
		return true;
	}

	public boolean isTableShared() {
		return true;
	}

	public boolean isShared() {
		return true;
	}

    /*
	 * This is a trick because we dont want to create a TableMapping for every
	 * couple(type, default table) to limit the number of tuple factories and
	 * statements.
	 */
    public ColumnMapping createDefaultMappingView(SchemaComponent comp, boolean useStringDelimitor)
    throws RepositoryException {
		return new DefaultTableMappingView(comp, useStringDelimitor).getColumnView();
	}
    
    ////////////////////////////////////////////////////////////////////////////
    //                      MappingInfo IMPLEMENTATION
    ////////////////////////////////////////////////////////////////////////////
    /* It is a trick because user table are not created automatically whereas OIDs are */
    public String getSelectStatement() {
		return null;
	}

	public String getUpdateStatement() {
		return null;
	}

	public String getOIDInsertStatement() {
		return null;
	}

	public String getOIDTableName() {
		return null;
	}
    
    //////////////////////////////////////////////////////////////////////////
    //                                PRIVATE
    //////////////////////////////////////////////////////////////////////////
    private void initializeStatements()
    {
        insertStatement = tableInfo.getInsertStatement();
        
        StringBuffer columnList = new StringBuffer();
        StringBuffer namedColumnList = new StringBuffer();
        StringBuffer wildcards = new StringBuffer();
        
        ColumnInfo[] columns = tableInfo.getColumns();
        String columnName = null;
        for (int i = 0; i < columns.length; i++)
        {
            columnName = columns[i].getName();
            
            columnList.append(columnName);
            columnList.append(',');
            
            namedColumnList.append("w.");
            namedColumnList.append(columnName);
            namedColumnList.append(',');
            
            wildcards.append("?,");
        }
        columnList.setLength(columnList.length() - 1);
        namedColumnList.setLength(namedColumnList.length() - 1);
        wildcards.setLength(wildcards.length() - 1);
        
        this.columnList = columnList.toString();
        this.namedColumnList = namedColumnList.toString();
        this.wildcards = wildcards.toString();
    }
    
    private void initialize()
    throws RepositoryException, SQLException
    {
        ColumnInfo[] columns = tableInfo.getColumns();
        
        StringBuffer columnList = new StringBuffer();
        primaryKeyColumnMappings = new ColumnMapping[columns.length];
        ColumnMapping column = null;
        for (int i = 0; i < columns.length; i++)
        {
            columnList.append("w.");
            columnList.append(columns[i].getName());
            columnList.append(',');
            switch (i)
            {
                case V_UOID_INDEX :
                    column = new DefaultColumnMapping(columns[i],
                    SystemVariableGenerator.createGenerator(UOID_VALUE), i, true);
                    break;
                case V_PATHID_INDEX :
                    column = new DefaultColumnMapping(columns[i],
                    SystemVariableGenerator.createGenerator(PATHID_VALUE), i, true);
                    break;
                default :
                    column = new DefaultColumnMapping(columns[i],
                    new ValueGenerator(), i, false);
                    valueMappings.add(i-2, column);
                    break;
            }
            if (columns[i].isPrimaryKey())
                primaryKeyColumnMappings[keyColumnCount++] = column;
            columnMappings[i] = column;
        }
        columnList.setLength(columnList.length() - 1);
        namedColumnList = columnList.toString();
        initializeStatements();
    }
    
    public ColumnMapping getDefaultColumnMapping(SchemaComponent comp)
    {
        int mappingIndex = V_CHAR_VALUE_INDEX - 2;
        if (comp != null)
        {
            Type type = null;
            if (comp instanceof Type)
                type = (Type)comp;
            else
                type = ((Declaration)comp).getType();
            
            SimpleType sType = type.getValueType();
            if (sType != null)
            {
                int pType = sType.getPrimitive().getType();
                
                if (pType == PrimitiveType.UNION)
                {
                    // try to find a common target basic type
                    Iterator it = sType.getMemberTypes().iterator();
                    PrimitiveType wType = null;
                    
                    int previousMappingIndex = -1;
                    while (it.hasNext())
                    {
                        wType = ((SimpleType)it.next()).getPrimitive();
                        mappingIndex = getMappingIndex(MappingTypeInfo.ABSTRACT_SCHEMA_JDBC_MAPPING[wType.getType()]);
                        if ((previousMappingIndex != -1) && (mappingIndex != previousMappingIndex))
                        {
                            mappingIndex = V_CHAR_VALUE_INDEX - 2;
                            break;
                        }
                        previousMappingIndex = mappingIndex;
                    }
                }
                else
                    mappingIndex = getMappingIndex(MappingTypeInfo.ABSTRACT_SCHEMA_JDBC_MAPPING[pType]);
            }
        }
        return (ColumnMapping)valueMappings.get(mappingIndex);
    }
    
    private int getMappingIndex(int basicJDBCType)
    {
        int mappingIndex = V_CHAR_VALUE_INDEX - 2;
        switch (basicJDBCType)
        {
            case Types.VARCHAR: case Types.VARBINARY:
                mappingIndex = V_CHAR_VALUE_INDEX - 2;
                break;
            case Types.DECIMAL: case Types.FLOAT: case Types.DOUBLE:
                mappingIndex = V_NUM_VALUE_INDEX - 2;
                break;
            case Types.TIMESTAMP:
                mappingIndex = V_DATE_VALUE_INDEX - 2;
                break;
        }
        return mappingIndex;
    }
   
    //////////////////////////////////////////////////////////////////////////
    //                            INNER CLASSES
    //////////////////////////////////////////////////////////////////////////
    private class DefaultColumnMapping implements ColumnMapping
    {
        
        private Generator generator;
        private MappingTypeInfo typeInfo;
        private int index;
        private ColumnInfo columnInfo;
        
        DefaultColumnMapping(ColumnInfo column, Generator generator, int index, boolean finalGen)
        throws RepositoryException
        {
            columnInfo = column;
            this.generator = generator;
            // for OIDs create a typeInfo (for values a different will be created for each path)
            String type = null;
            if (generator instanceof SystemVariableGenerator)
                type = ((SystemVariableGenerator)generator).getXMLType();
            else    // set a type info for null setting
            {
                switch (getMetaData().getDataType())
                {
                    case Types.VARCHAR: case Types.VARBINARY:
                        type = "string";
                        break;
                    case Types.DECIMAL: case Types.FLOAT: case Types.DOUBLE:
                        type = "decimal";
                        break;
                    case Types.TIMESTAMP:
                        type = "dateTime";
                        break;
                }
            }
            try
            {
                typeInfo = new MappingTypeInfo(type, false);
                typeInfo.checkTargetType(getMetaData());    // no handler (system)
            }
            catch (SAXException e)
            {
                throw new RepositoryException(RepositoryException.INTERNAL_ERROR,
                "Error while setting type conversion flags on DATA table.", e);
            }
            if (finalGen)
                addToFinalValues(this);
            this.index = index;
        }
        
        public SchemaComponent getSchemaComponent() {
			return getSchemaType();
		}

		public Type getSchemaType() {
			Type ret = null;
			if (typeInfo != null)
				ret = typeInfo.getXMLType();
			return ret;
		}

		public String getColumnName() {
			return columnInfo.getName();
		}

		public ColumnMetaData getMetaData() {
			return DefaultTableMapping.this.getMetaData().getColumnMetaData(
					getColumnName());
		}

		public int getColumnIndex() {
			return index;
		}

		public int getInsertColumnIndex() {
			return getColumnIndex();
		}

		public int getKeyColumnIndex() {
			if (columnInfo.isPrimaryKey())
				return index;
			else
				return -1;
		}

		public int getSelectColumnIndex() {
			return -1;
		}

		public int getFetchColumnIndex() {
			return -1;
		}

		public int getUpdateColumnIndex() {
			return -1;
		}

		public Generator getGenerator() {
			return generator;
		}

		public Generator getDefaultGenerator() {
			return null;
		}

		public Generator getInitGenerator() {
			return getGenerator();
		}

		public int getTableRefIndex() {
			return -1;
		}

		public int getColumnType() {
			return columnInfo.getDataType();
		}

		public TableMapping getTableMapping() {
			return DefaultTableMapping.this;
		}

		public String getTableName() {
			return name;
		}

		public int getTableIndex() {
			return DefaultTableMapping.this.getTableIndex();
		}

		public String toString() {
			return getTableName() + "." + getColumnName();
		}

		public MappingTypeInfo getTypeInfo() {
			return typeInfo;
		}

		public int getJoinColumnIndex() {
			return -1;
		}

		public boolean isInJoin() {
			return false;
		}

		public boolean updateColumnWhenMissing() {
			return false;
		}
    }
    
    private class DefaultTableMappingView extends MappingImpl implements TableMapping
    {
        private ColumnMapping[] columns;
        private DefaultColumnMappingView columnView;
        
        DefaultTableMappingView(SchemaComponent comp, boolean useStringDelimitor)
        throws RepositoryException
        {
            super(comp);
//        	columns = new ColumnMapping[DefaultTableMapping.this.columnMappings.length];
        	columns = (ColumnMapping[])DefaultTableMapping.this.columnMappings.clone();
            columnView = new DefaultColumnMappingView(this, comp, useStringDelimitor);
            columns[columnView.getColumnIndex()] = columnView;
        }
        
        /**
		 * Overriden because default mapping uses views on real table mapping.
		 * As query reconstruction use equality, we use what realy identifies
		 * the table mapping: its index.
		 */
        public boolean equals(Object o)
        {
            boolean ret = false;
            if (o instanceof TableMapping)
                ret = o.equals(DefaultTableMapping.this);
            return ret;
        }
        /** Overriden because default mapping uses views on real table mapping.
         * (As query reconstruction use a map).
         */
        public int hashCode()
        {
            return DefaultTableMapping.this.hashCode();
        }
   
        public int getColumnCount()
        {
            return DefaultTableMapping.this.getColumnCount();
        }
        
    	public int getColumnMappingCount()
    	{
    		return getColumnCount();
    	}

    	public ColumnMapping[] getColumnMappings()
    	{
    		return DefaultTableMapping.this.columnMappings;
    	}

        public DefaultColumnMappingView getColumnView()
        {
            return columnView;
        }
        
        public ColumnMapping getColumnMapping(int index)
        {
            return columns[index];
        }
        
        public List finalParameters()
        {
            return DefaultTableMapping.this.finalParameters();
        }
        
        public int getAction()
        {
            return DefaultTableMapping.this.getAction();
        }
        
        public int getBatchSize()
        {
            return DefaultTableMapping.this.getBatchSize();
        }
        
        public String getColumnList()
        {
            return DefaultTableMapping.this.getColumnList();
        }
        
        public ColumnMapping getColumnMapping(String name)
        {
            return DefaultTableMapping.this.getColumnMapping(name);
        }
        
        public String getInsertStatement()
        {
            return DefaultTableMapping.this.getInsertStatement();
        }
        
        public String getJoinCondition()
        {
            return DefaultTableMapping.this.getJoinCondition();
        }
        
        public int getKeyColumnCount()
        {
            return DefaultTableMapping.this.getKeyColumnCount();
        }
        
        public String getJoinColumnList()
        {
            return DefaultTableMapping.this.getJoinColumnList();
        }
        
        public String getJoinColumnTypes()
        {
            return DefaultTableMapping.this.getJoinColumnTypes();
        }
        
        public String getJoinWildcards()
        {
            return DefaultTableMapping.this.getJoinWildcards();
        }
        
        public TableMetaData getMetaData()
        {
            return DefaultTableMapping.this.getMetaData();
        }
        
        public String getNamedColumnList()
        {
            return DefaultTableMapping.this.getNamedColumnList();
        }
        
        public ColumnMapping[] getJoinColumns()
        {
            return DefaultTableMapping.this.getJoinColumns();
        }
        
        public String getOIDInsertStatement()
        {
            return DefaultTableMapping.this.getOIDInsertStatement();
        }
        
        public int getOIDTableColumnCount()
        {
            return DefaultTableMapping.this.getOIDTableColumnCount();
        }
        
        public String getPathIDColumnName()
        {
            return DefaultTableMapping.this.getPathIDColumnName();
        }
        
        public int getPathIDIndex()
        {
            return DefaultTableMapping.this.getPathIDIndex();
        }
        
        public ColumnMapping[] getPrimaryKey()
        {
            return DefaultTableMapping.this.getPrimaryKey();
        }
        
        public int getSelectColumnCount()
        {
            return DefaultTableMapping.this.getSelectColumnCount();
        }
        
        public int getFetchColumnCount()
        {
            return DefaultTableMapping.this.getFetchColumnCount();
        }
        
        public ColumnMapping[] getSelectColumns()
        {
            return DefaultTableMapping.this.getSelectColumns();
        }
        
        public ColumnMapping[] getFetchColumns()
        {
            return DefaultTableMapping.this.getFetchColumns();
        }
        
        public String getSelectStatement()
        {
            return DefaultTableMapping.this.getSelectStatement();
        }
        
        public int getTableIndex()
        {
            return DefaultTableMapping.this.getTableIndex();
        }
        
        public String getTableName()
        {
            return DefaultTableMapping.this.getTableName();
        }
        
        public String getUOIDColumnName()
        {
            return DefaultTableMapping.this.getUOIDColumnName();
        }
        
        public int getUOIDIndex()
        {
            return DefaultTableMapping.this.getUOIDIndex();
        }
        
        public int getUpdateColumnCount()
        {
            return DefaultTableMapping.this.getUpdateColumnCount();
        }
        
        public ColumnMapping[] getUpdateColumns()
        {
            return DefaultTableMapping.this.getUpdateColumns();
        }
        
        public String getUpdateStatement()
        {
            return DefaultTableMapping.this.getUpdateStatement();
        }
        
        public String getWildcards()
        {
            return DefaultTableMapping.this.getWildcards();
        }
        
        public List initParameters()
        {
            return DefaultTableMapping.this.initParameters();
        }
        
        public boolean isClustered()
        {
            return DefaultTableMapping.this.isClustered();
        }
        
        public void setIndex(int index)
        {
            // no op
        }
        
        public String getOIDTableName()
        {
            return DefaultTableMapping.this.getOIDTableName();
        }
        
        public TableMapping getTableMapping()
        {
            return this;
        }
        
    	public boolean isShareable()
    	{
    		return DefaultTableMapping.this.isShareable();
    	}
    	
    	public boolean isTableShared()
    	{
    		return DefaultTableMapping.this.isTableShared();
    	}

    	public boolean isShared()
    	{
    		return DefaultTableMapping.this.isShared();
    	}
    }
    
    private class DefaultColumnMappingView extends MappingImpl implements ColumnMapping
    {
        
        private ValueGenerator generator;
        private MappingTypeInfo typeInfo;
        private ColumnMapping columnMapping;
        private DefaultTableMappingView tableMapping;
        
        DefaultColumnMappingView(
                                DefaultTableMappingView tableMapping,
                                SchemaComponent comp,
                                boolean useStringDelimitor
                                )
        throws RepositoryException
        {
            super(comp);
            this.columnMapping = getDefaultColumnMapping(comp);
            this.tableMapping = tableMapping;
            // clone the value generator to keep its copy of MappingTypeInfo
            generator = (ValueGenerator)((ValueGenerator)columnMapping.getGenerator()).clone();
            try
            {
                Type schemaType = getSchemaType();
                if (schemaType == null)
                    typeInfo = columnMapping.getTypeInfo(); // use the lax one corresponding to the target type
                else
                {
                    typeInfo = new MappingTypeInfo(schemaType.getValueType(), useStringDelimitor);
                    typeInfo.checkTargetType(getMetaData(), null);
                }
            }
            catch (SAXException e)
            {
                throw new RepositoryException(RepositoryException.NOT_ALLOWED,
                "XML type for " + comp + " does not fit the default mapping table.", e);
            }
            generator.setTypeInfo(typeInfo);
        }
        
        public String getColumnName() {
			return columnMapping.getColumnName();
		}

		public ColumnMetaData getMetaData() {
			return columnMapping.getMetaData();
		}

		public int getColumnIndex() {
			return columnMapping.getColumnIndex();
		}

		public int getInsertColumnIndex() {
			return getColumnIndex();
		}

		public int getKeyColumnIndex() {
			return columnMapping.getKeyColumnIndex();
		}

		public int getSelectColumnIndex() {
			return columnMapping.getSelectColumnIndex();
		}

		public int getFetchColumnIndex() {
			return columnMapping.getFetchColumnIndex();
		}

		public int getUpdateColumnIndex() {
			return columnMapping.getUpdateColumnIndex();
		}

		public Generator getGenerator() {
			return generator;
		}

		public Generator getDefaultGenerator() {
			return null;
		}

		public Generator getInitGenerator() {
			return getGenerator();
		}

		public int getTableRefIndex() {
			return -1;
		}

		public int getColumnType() {
			return columnMapping.getColumnType();
		}

		public TableMapping getTableMapping() {
			return tableMapping;
		}

		public String getTableName() {
			return name;
		}

		public int getTableIndex() {
			return DefaultTableMapping.this.getTableIndex();
		}

		public String toString() {
			return getTableName() + "." + getColumnName();
		}

		public MappingTypeInfo getTypeInfo() {
			return typeInfo;
		}

		public int getJoinColumnIndex() {
			return -1;
		}

		public boolean isInJoin() {
			return false;
		}

		public boolean updateColumnWhenMissing() {
			return false;
		}
    }
}
