/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.mapper.mapping;

import java.util.List;

import org.xquark.schema.ElementDeclaration;
import org.xquark.schema.SchemaContext;
import org.xquark.xpath.NodeKind;
import org.xquark.xpath.PathExpr;
import org.xquark.xpath.XNode;

/** This class allows to iterate through a repository mapping (and a schema)
 * while parsing a document.
 * <p>It is used for instance to provide table or column mappings while
 * storing or reading mapped XML documents. This object maintains a context
 * for the current Schema component. It contains features to manage the fact
 * that it follows both document (therefore schema) and mapping structure that
 * is a subset of the first.</p>
 * @deprecated use MappingIterator instead
 */
public class MappingContext extends SchemaContext
implements MappingLocator
{
private static final String RCSRevision = "$Revision: 1.2 $"; 
private static final String RCSName     = "$Name: XQ1_1_1 $";
    
    //
    //  DATA
    //
    
    /** The XMLCollection Mapping browsed
     */
    protected MappingFactory factory;
    protected MappingNode wNode;
    protected XNode pattern = new XNode();
    
    //
    //  CONSTRUCTORS
    //
    
    public MappingContext(MappingFactory factory)
    {
        super(factory.getSchemaManager());
        this.factory = factory;
        wNode = (MappingNode)factory.getTree().getRoot();
    }
    
    //
    //  MappingLocator IMPLEMENTATION
    //
    /** Get the table mappings for the current context
     */
    public List currentTableMappings()
    {
        return wNode.getTableMappings();
    }
    
    /** Get the column mappings for the current context
     */
    public List currentColumnMappings()
    {
        return wNode.getColumnMappings();
    }
    
    /** Get the table mappings for a particular subelement in the current context
     */
    public List getElementTableMappings(String namespace, String localName)
    {
        pattern.set(namespace, localName, NodeKind.ELEMENT);
        MappingNode node = (MappingNode)wNode.getChild(pattern);
        if (node == null)
            return null;
        else
            return node.getTableMappings();
    }
    
    /** Get the column mappings for a particular subelement in the current context
     */
    public List getElementColumnMappings(String namespace, String localName)
    {
        pattern.set(namespace, localName, NodeKind.ELEMENT);
        MappingNode node = (MappingNode)wNode.getChild(pattern);
        if (node == null)
            return null; // SET schema component ?
        else
            return node.getColumnMappings();
    }
    
    /** Get the table mappings for a particular attribute in the current context
     */
    public List getAttributeTableMappings(String namespace, String localName)
    {
        pattern.set(namespace, localName, NodeKind.ATTRIBUTE);
        MappingNode node = (MappingNode)wNode.getChild(pattern);
        if (node == null)
            return null; // SET schema component ?
        else
            return node.getTableMappings();
    }
    
    /** Get the column mappings for a particular attribute in the current context
     */
    public List getAttributeColumnMappings(String namespace, String localName)
    {
        pattern.set(namespace, localName, NodeKind.ATTRIBUTE);
        MappingNode node = (MappingNode)wNode.getChild(pattern);
        if (node == null)
            return null; // SET schema component ?
        else
            return node.getColumnMappings();
    }
    
    //
    //  SchemaContext OVERLOADING
    //
    
    public final void push(ElementDeclaration decl)
    {
        super.push(decl);
        pattern.set(decl.getNamespace(), decl.getName(), NodeKind.ELEMENT);
        wNode = (MappingNode)wNode.getChild(pattern);
    }
    
    public final ElementDeclaration pop()
    {
        ElementDeclaration ret = super.pop();
        wNode = (MappingNode)wNode.getParent();
        return ret;
    }
    
    
    /** Return the absolute position of the locator.
     * <p>i.e., !isDiscardable()</p>
     */
    public PathExpr getLocation()
    {
        return wNode.getLocation();
    }
    
}
