/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.mapper.mapping;

import org.xml.sax.SAXException;
import org.xquark.mapper.metadata.Repository;
import org.xquark.schema.*;
import org.xquark.xpath.*;

/**
 * Specific implementation of XTreeBuilder.
 *
 */
public class MappingFactory extends XTreeBuilder
{
    private static final String RCSRevision = "$Revision: 1.2 $";
    private static final String RCSName     = "$Name: XQ1_1_1 $";
    
    private SchemaManager schemaManager;
    private SchemaLocator schemaLocator;
    private Repository rep;
    
    public MappingFactory(SchemaManager manager, SchemaLocator locator)
    {
        set(manager, locator);
    }
    
    public MappingFactory(Repository rep, SchemaManager manager, SchemaLocator locator)
    {
        this(manager, locator);
        this.rep = rep;
    }
    
    public MappingFactory   (
                            RepositoryMapping mapping,
                            SchemaManager manager,
                            SchemaLocator locator
                            )
    {
        super(mapping);
        set(manager, locator);
    }
    
    private void set(SchemaManager manager, SchemaLocator locator)
    {
        this.schemaManager = manager;
        this.schemaLocator = locator;
    }
    
    public SchemaLocator getSchemaLocator()
    {
        return schemaLocator;
    }
    
    public SchemaManager getSchemaManager()
    {
        return schemaManager;
    }
    
    public XTree allocateTree()
    {
        return new RepositoryMapping(rep);  // may be null
    }
    
    //
    //  XModelFactory IMPLEMENTATION
    //
    
    public XTreeNode allocateNode(XTreeNode parent, String namespace, String localName, byte type)
    {
        MappingNode node = new MappingNode(tree, parent, namespace, localName, type);
        // get decl from parent or from Schema Manager
        node.setComponent(findComponent(parent == null ? null:((MappingNode)parent).getSchemaComponent(), node));
        return node;
    }
    
    /** Version used by the PathSet class for flying nodes.
     */
    public XTreeNode allocateNode(SchemaComponent comp, XTreeNode pathNode)
    {
        MappingNode node = new MappingNode(tree, null, pathNode.getNamespace(),
                                            pathNode.getLocalName(),
                                            pathNode.getType());
        // get decl from parent or from Schema Manager
        node.setComponent(findComponent(comp, node));
        return node;
    }
    
    private Declaration findComponent(SchemaComponent comp, MappingNode node)
    {
        Declaration decl = null;
        
        // use parent element declaration (parent is an element because type has no parent...)
        if (comp != null)
        {
            org.xquark.schema.Type pType;
            if (comp instanceof org.xquark.schema.Type)   // for a type mapping
                pType = (org.xquark.schema.Type)comp;
            else
                pType = ((ElementDeclaration)comp).getType();
            switch (node.getType())
            {
                case NodeKind.ATTRIBUTE:
                    decl = pType.getAttributeDeclaration(node.getNamespace(), node.getLocalName());
                    break;
                case NodeKind.ELEMENT:
                    decl = pType.getElementDeclaration(node.getNamespace(), node.getLocalName());
                    break;
                default:
            }
        }
        // use SchemaManager
        if ((decl == null) && (node.getType() != NodeKind.NONE))  // not found or no parent comp:
                    // may be a wildcard or a substitution group : jump to declaration
        {           //  IMPORTANT : responsability is left to document validation or mapping loader
                    // to verify that "jump" is allowed
            try
            {
                Schema schema = schemaManager.loadSchema(schemaLocator, node.getNamespace());
                if (schema != null)
                    decl = schema.getElementDeclaration(node.getLocalName());
            }
            catch (SAXException e)
            {
                // TO DO :  not so good !
            }
//            if ((decl == null) && !node.getNamespace().equals(MappingConstants.MAPPING_URI))
//                throw new RuntimeException("No declaration was found in the Repository for: " + node);
//  TO DO : add control to check when schema is added : doc already stored may be unvalid            
        }
        return decl;
    }
    
    public MappingNode createTypeNode(org.xquark.schema.Type type)
    {
        MappingNode node = (MappingNode)createNode(null, NodeKind.NONE);    // type ROOT is needed to avoid null pointer exception in getLocation()
        node.setComponent(type);
        ((RepositoryMapping)tree).addTypeNode(node);
        return node;	// return the created node, not the replaced returned by put (should not happen)
    }
    
    /** Add a new metaNode for an element.
     * @param type if ELEMENT normal processing. If NODE, supposed to be a shared
     * type mapping, the corresponding type node is cloned.
     */
    public MappingNode createElementNode(MappingNode parent, ElementDeclaration decl, boolean typeMapping)
    {
        MappingNode node;
        if (typeMapping)	// The corresponding type NodeEntry is cloned
        {
            MappingNode typeNode = ((RepositoryMapping)tree).getTypeNode(decl.getType());
            // duplicate the type branch
            node =(MappingNode)typeNode.clone();
            node.getHashKey().set(decl.getNamespace(), decl.getName(), NodeKind.ELEMENT);
            node.setComponent(decl);
            parent.addChild(node);
            tree.register(node);
            // add type mapping in scope to element hierachy
            for (MappingNode elt = node; elt != null; elt = (MappingNode)elt.getParent())
            {
                elt.addMappingsInScope(typeNode.getScopeMappings());
            }
        }
        else
            node = (MappingNode)createNamedNodeIfNotExist(parent, decl.getNamespace(), decl.getName(), NodeKind.ELEMENT);
        return node;
    }
    
    /* Used by mapping loader (ns must no be null) */
    public org.xquark.schema.Type findType(String namespace, String localName) throws SAXException
    {
        org.xquark.schema.Schema schema = schemaManager.loadSchema(schemaLocator, namespace);
        if (schema == null)
            throw new SAXException("The schema for \"" + namespace
            + "\" namespace is not available so the mapping file cannot be loaded.");
        return schema.getType(localName);
    }
    
    
}
