/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.mapper.mapping;

import org.xquark.mapper.RepositoryException;
import org.xquark.mapper.metadata.StoragePathMetadata;
import org.xquark.mapper.storage.PathIterator;
import org.xquark.xpath.NodeKind;
import org.xquark.xpath.PathExpr;
import org.xquark.xpath.XNode;

/** This class allows to iterate through mappings of RepositoryMapping object
 * using names (and not Schema declarations) reading only. Push made out of the
 * mapping will lead the getPathNode() method to return null.
 */
public class MappingIterator implements PathIterator 
{
    private static final String RCSRevision = "$Revision: 1.2 $"; 
    private static final String RCSName     = "$Name: XQ1_1_1 $";
    
    //
    //  DATA
    //
    
    /** The XMLCollection Mapping browsed
     */
    private RepositoryMapping mapping;
    private MappingNode wNode;
    private XNode pattern = new XNode();
    private int extraPop;
    
    //
    //  CONSTRUCTORS
    //
    
    public MappingIterator(RepositoryMapping mapping)
    {
        this.mapping = mapping;
        reset();
    }

    public void reset()
    {
        wNode = (MappingNode)mapping.getRoot();
        extraPop = 0;
    }
    
    /* Create a node if it does not exist without entering it */
    public StoragePathMetadata createNode(String namespace, String localName, byte type) throws RepositoryException
    {
        return getChild(namespace, localName, type);
    }

    /* Enter the specified node element or create it if does not exist */
    public StoragePathMetadata push(String namespace, String localName) throws RepositoryException
    {
        MappingNode node = null;
        if ((extraPop > 0) || ((node = getChild(namespace, localName, NodeKind.ELEMENT)) == null))
            extraPop++;
        else
            wNode = node;
        return node;
    }

    /** Get the column mappings for a particular attribute in the current context
     */
    private MappingNode getChild(String namespace, String localName, byte type)
    {
        pattern.set(namespace, localName, type);
        return (MappingNode)wNode.getChild(pattern);
    }

    /* Enter the specified node or create it if does not exist */
    public StoragePathMetadata pop()
    {
        MappingNode leftNode = wNode;
        if (extraPop == 0)
            wNode = (MappingNode)wNode.getParent();
        else
        {
            extraPop--;
            leftNode = null;
        }
        return leftNode;
    }
    
    /** Return the absolute position of the locator.
     * <p>i.e., !isDiscardable()</p>
     */
    public PathExpr getLocation()
    {
        return wNode.getLocation();
    }
    
    public StoragePathMetadata getStoragePathMetadata()
    {
        if (extraPop == 0)
            return wNode;
        else
            return null;
    }
}
