/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.mapper.mapping;

import java.util.*;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.AttributesImpl;
import org.xquark.mapper.metadata.StoragePathMetadata;
import org.xquark.schema.*;
import org.xquark.xpath.*;

/**
 * XTreeNode implementation for XQuark mapping file memory representation.
 *
 */
public class MappingNode extends TypedXTreeNode
implements StoragePathMetadata
{
	private static final String RCSRevision = "$Revision: 1.3 $"; 
	private static final String RCSName     = "$Name: XQ1_1_1 $";
	    
	private static Log log = LogFactory.getLog(MappingNode.class);

	/** The schema component related to the node, if any */
    protected SchemaComponent comp;
    /** Table mappings defined for the SchemaComponent, if any */
    protected List tableMappings;
    /** Column mappings defined for the SchemaComponent, if any */
    protected List columnMappings;
    /** Column mapping used for reconstruction */
    protected ColumnMapping refColumnMapping;
    
    /** Table mappings that are defined in the scope of the element,
     * The default mapping is excluded because they will be added
     * when adding the MappingNode to the PathNode.
     */
        // MappingNode mappingInScope+hasDefaultInScope but null if not buildable
        // add a List data
        // determine buildable flag : some column mappings in scope have a table mapping defined in an ancestor
        // to implement a flag set when registering column mappings in Repository mapping
        // TO DO : change ArrayList of mappingInScope in a HashSet
    protected HashSet scopeTableMappings;
    
    /** flag that says if the node can be built with the mapping in scope
     */
    private boolean isBuildable = true;
    
    /** the first ancestor that has all table mappings in scope for this path
     * construction
     */
    private int builderDepthDiff = 0;   // builder was originally a reference but it was hard
                                      // to manage when cloning for type mappings. Now
                                      // the relative depth is stored.
    
    
    public MappingNode(XTree tree, XTreeNode parent, String namespace, String localName, byte type)
    {
        super(tree, parent, namespace, localName, type);
    }
    
    //	ACCESSORS
    /** Set the schema component and the defaultAllowed internal flag */
    public void setComponent(SchemaComponent comp)
    {
         this.comp = comp;
    }
    
    /** Used by RepositoryMapping.register()
     */
    void setBuildable(boolean flag)
    {
        isBuildable = flag;
    }
    
    public boolean isBuildable()
    {
        return isBuildable;
    }
    
    /** Used by RepositoryMapping.register()
     */
    void setBuilderDiff(int diff)
    {
        builderDepthDiff = diff;
    }
    
    public MappingNode getBuilder()
    {
        MappingNode node = this;
        for (int i = 0; i < builderDepthDiff; i++)
        {
            node = (MappingNode)node.getParent();
        }
        return node;
    }
    
    /** True if the schema authorize a node to use default mapping when
     * no mapping is defined by the user (simple type or attribute).
     *  The default mapping is added in a second pass following the mapping 
     * file loading if this flag is true.
     */
    public boolean isDefaultAllowed()
    {
        boolean defaultAllowed = true;
        if ((comp != null) && (getType() == NodeKind.ELEMENT))   // attribute is always leaf
        {
            Type type;
            if (comp instanceof Type)
                type = (Type)comp;
            else    // Declaration (attribute or element)
                type = ((Declaration)comp).getType();
            if (type instanceof ComplexType)
            {
                if (((ComplexType)type).getContentType() != SchemaConstants.TEXT_ONLY)
                    defaultAllowed = false;
            }
            // else // (SimpleType) : default value
        }
        return defaultAllowed;
    }
        
    public boolean potentialNewPathsExist()
    {
        boolean potentialNewPathsExist = false;
        if (comp != null)
        {
            if ((getType() == NodeKind.ELEMENT)
                && (((ElementDeclaration)getDeclaration()).getType() instanceof ComplexType))
            {
                ComplexType type = (ComplexType)((ElementDeclaration)getDeclaration()).getType();
                potentialNewPathsExist = ((type.getAttributeDeclarations().size() + type.getElementDeclarations().size())
                > (getChildren() == null ? 0:getChildren().size()));
            }
        }
        else if (getType() == NodeKind.ELEMENT)
            potentialNewPathsExist = true;
        
        // recursive call
        if (!potentialNewPathsExist && (getChildren() != null))
        {
            Iterator it = getChildren().iterator();
            while (!potentialNewPathsExist && it.hasNext())
                potentialNewPathsExist |= ((MappingNode)it.next()).potentialNewPathsExist();
        }
        
        return potentialNewPathsExist;
    }
    
    public SchemaComponent getSchemaComponent()
    {
        return comp;
    }
    
    public Declaration getDeclaration()
    {
        return comp instanceof Declaration ? (Declaration)comp : null;
    }
    
    /** get table mappings for the schema component.
     * if component has tableMapping for its type, these mapping are concatenated
     */
    public List getTableMappings()
    {
        return tableMappings;
    }
    
    /** get column mappings for the schema component */
    public List getColumnMappings()
    {
        return columnMappings;
    }
    
    /** get column mapping used for reconstruction and queries */
    public ColumnMapping getReferenceColumnMapping()
    {
        return refColumnMapping;
    }
    
    /** get table mapping used for reconstruction and queries */
    public TableMapping getReferenceTableMapping()
    {
        if (tableMappings != null)
            return (TableMapping)tableMappings.get(0);
        else
            return null;
    }
    
    /** Get the table mappings in the scope element
     * @return null if notable mapping
     */
    public Collection getScopeMappings()
    {
        return scopeTableMappings;
    }
    
    /** Add a new table mapping in the element scope */
    boolean addMappingInScope(TableMapping table)
    {
        if (scopeTableMappings == null)
            scopeTableMappings = new HashSet();
        return scopeTableMappings.add(table);
    }
    
    /** Add a new set of table mapping in the element scope */
    boolean addMappingsInScope(Collection mappings)
    {
        if (scopeTableMappings == null)
            scopeTableMappings = new HashSet();
        return scopeTableMappings.addAll(mappings);
    }
    
    //	PUBLIC METHODS
    
    /** Add a table mapping to this entry and to the RepositoryMapping list */
    void register(Mapping mapping)
    {
        RepositoryMapping colMetaData = (RepositoryMapping)tree;
        if (mapping instanceof TableMapping)
        {
            if (tableMappings == null)
                tableMappings = new ArrayList(1);   // bet most of elements will have only one table mapping
            tableMappings.add(mapping);
            ((TableMapping)mapping).setIndex(colMetaData.getTableCount());
            colMetaData.addTableMapping((TableMapping)mapping);
        }
        else if (mapping instanceof ColumnMapping)
        {
            if (columnMappings == null)
                columnMappings = new ArrayList();
            columnMappings.add(mapping);
            if (log.isDebugEnabled())
            	log.debug("Registering [mapping] on [schema comp]=" + mapping + comp);
            if ((refColumnMapping == null)
                && (((ColumnMapping)mapping).getGenerator() != null)) // to discard the table-column mapping case
                refColumnMapping = (ColumnMapping)mapping;
                
        }
    }
    
	/*
	 * Moves a column mapping at the end of the list because ref generators assume
	 * that other generators were called before to perform the potential select
	 * in the db.
	 */
    void moveColumnMappingToTheEnd(ColumnMapping mapping) {
        if (columnMappings != null && columnMappings.contains(mapping)) {
            columnMappings.remove(mapping);
            columnMappings.add(mapping);
        }
    }
    
    public Object clone()
    {
        MappingNode node = null;
        node = (MappingNode)super.clone();
        if (tableMappings != null)
            node.tableMappings = new ArrayList(tableMappings);
        if (columnMappings != null)
            node.columnMappings = new ArrayList(columnMappings);
        return node;
    }
    

    public String toString()
    {
        String ret = super.toString() + "(" + comp + ")";
        if (scopeTableMappings != null)
            ret += " Scope" +scopeTableMappings;
        if (tableMappings != null)
            ret += " Tables" +tableMappings;
        if (columnMappings != null)
            ret += " Columns" + columnMappings ;
        ret += " b=" + isBuildable + " br=" + builderDepthDiff;
        return ret;
    }
    
    public void toXML(ContentHandler handler)
    throws SAXException
    {
        AttributesImpl atts = new AttributesImpl();
        String value;
        
        // schema component
        if (comp != null)
        {
            handler.startElement("", "schemaComponent", "", atts);
            value = comp.toString();
            handler.characters(value.toCharArray(), 0, value.length());
            handler.endElement("", "schemaComponent", "");
        }
        
        // Builder depth diff
        handler.startElement("", "builderDepthDiff", "", atts);
        value = String.valueOf(builderDepthDiff);
        handler.characters(value.toCharArray(), 0, value.length());
        handler.endElement("", "builderDepthDiff", "");
        
        // Optimization flags
        atts.addAttribute("", "buildable", "", "CDATA", String.valueOf(isBuildable));
        handler.startElement("", "flags", "", atts);
        atts.clear();
        handler.endElement("", "flags", "");
        
        // Reference column mappping
        handler.startElement("", "refColumn", "", atts);
        if (refColumnMapping != null)
        {
            value = refColumnMapping.getTableName() + "." + refColumnMapping.getColumnName();
            handler.characters(value.toCharArray(), 0, value.length());
        }
        handler.endElement("", "refColumn", "");
        
        
        // Table mapppings
        handler.startElement("", "tableMappings", "", atts);
        if (tableMappings != null)
        {
            Iterator it = tableMappings.iterator();
            while (it.hasNext())
            {
                atts.addAttribute("", "name", "", "CDATA", ((TableMapping)it.next()).getTableName());
                handler.startElement("", "table", "", atts);
                atts.clear();
                handler.endElement("", "table", "");
            }
        }
        handler.endElement("", "tableMappings", "");
        
        // Column mapppings
        handler.startElement("", "columnMappings", "", atts);
        if (columnMappings != null)
        {
            Iterator it = columnMappings.iterator();
            ColumnMapping column;
            while (it.hasNext())
            {
                column = (ColumnMapping)it.next();
                atts.addAttribute("", "name", "",  "CDATA", column.getTableName() + "." + column.getColumnName());
                handler.startElement("", "column", "", atts);
                atts.clear();
                handler.endElement("", "column", "");
            }
        }
        handler.endElement("", "columnMappings", "");
    
        // Table mapppings in scope
        handler.startElement("", "tableMappingsInScope", "", atts);
        if (scopeTableMappings != null)
        {
            Iterator it = scopeTableMappings.iterator();
            while (it.hasNext())
            {
                atts.addAttribute("", "name", "",  "CDATA", ((TableMapping)it.next()).getTableName());
                handler.startElement("", "table", "", atts);
                atts.clear();
                handler.endElement("", "table", "");
            }
        }
        handler.endElement("", "tableMappingsInScope", "");
    }
    
    ////////////////////////////////////////////////////////////////////
    //          FAKE IMPLEMENTATION of StoragePathMetadata interface.
    ////////////////////////////////////////////////////////////////////
    public short getPathID() { return -1;}
    public byte getStorageMode() { return -1;}
    public boolean isPositionEnforced() { return false;}
    public boolean isPersistent() { return true;}
    public boolean isNilImplicit() { return false;}
}
