/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Odonata.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.mapper.mapping;

import java.io.*;
import java.math.BigDecimal;
import java.sql.*;
import java.util.Arrays;
import java.util.Iterator;

import org.xml.sax.SAXException;
import org.xquark.jdbc.typing.*;
import org.xquark.mapper.RepositoryException;
import org.xquark.mapper.dbms.AbstractConnection;
import org.xquark.schema.Declaration;
import org.xquark.schema.SimpleType;
import org.xquark.schema.datatypes.ByteArray;
import org.xquark.schema.datatypes.DateTime;
import org.xquark.schema.datatypes.PrimitiveType;
import org.xquark.schema.validation.ValidationContextProvider;
import org.xquark.xml.xdbc.XMLDBCException;
import org.xquark.xml.xdbc.XMLErrorHandler;

/**
 * @author sr
 *
 * To change the template for this generated type comment go to
 * Window - Preferences - Java - Code Generation - Code and Comments
 */
public class MappingTypeInfo extends org.xquark.jdbc.typing.MappingInfo {

	/** Conversion authorized by setObject */
	public static int AUTHORIZED_CONVERSIONS[][] = new int[13][];
	
	static
	{
        // Conversion authorized by setObject ("getting Started" of JDBC user guide)
        AUTHORIZED_CONVERSIONS[JAVA_STRING] = new int[]
        {
            Types.TINYINT, Types.SMALLINT, Types.INTEGER, Types.BIGINT,
            Types.REAL, Types.FLOAT, Types.DOUBLE, Types.DECIMAL, Types.NUMERIC,
            Types.BIT, Types.CHAR, Types.VARCHAR, Types.LONGVARCHAR,
            Types.BINARY, Types.VARBINARY, Types.LONGVARBINARY, Types.DATE,
            Types.TIME, Types.TIMESTAMP
        };
        Arrays.sort(AUTHORIZED_CONVERSIONS[JAVA_STRING]); // for binary search
        // Note: this sort could be avoided but is safer if arrays are modified later
        // further more is precompiled-data
        
        AUTHORIZED_CONVERSIONS[JAVA_BIG_DECIMAL] =
        AUTHORIZED_CONVERSIONS[JAVA_BOOLEAN] =
        AUTHORIZED_CONVERSIONS[JAVA_FLOAT] =
        AUTHORIZED_CONVERSIONS[JAVA_DOUBLE] = new int[]
        {
            Types.TINYINT, Types.SMALLINT,
            Types.INTEGER, Types.BIGINT, Types.REAL, Types.FLOAT, Types.DOUBLE,
            Types.DECIMAL, Types.NUMERIC, Types.BIT, Types.CHAR, Types.VARCHAR,
            Types.LONGVARCHAR
        };
        Arrays.sort(AUTHORIZED_CONVERSIONS[JAVA_DOUBLE]); // for binary search
        
        AUTHORIZED_CONVERSIONS[SCHEMA_BYTE_ARRAY] = new int[]
        {
            Types.BINARY,
            Types.VARBINARY, Types.LONGVARBINARY
        };
        Arrays.sort(AUTHORIZED_CONVERSIONS[SCHEMA_BYTE_ARRAY]); // for binary search
        
        AUTHORIZED_CONVERSIONS[SCHEMA_DATETIME] = new int[]
        {
            Types.CHAR,
            Types.VARCHAR, Types.LONGVARCHAR, Types.DATE, Types.TIME,
            Types.TIMESTAMP
        };
        Arrays.sort(AUTHORIZED_CONVERSIONS[SCHEMA_DATETIME]); // for binary search
    }

    /**
     * @param XMLType
     */
    public MappingTypeInfo(SimpleType XMLType) {
        super(XMLType);
    }

    /**
     * @param XMLType
     * @throws SAXException
     */
    public MappingTypeInfo(String XMLType) {
        super(XMLType);
    }

    public MappingTypeInfo(String XMLType, boolean useStringDelimitor)
    {
        super(XMLType);
        this.useStringDelimitor = useStringDelimitor;
    }
    
    /* Used for value generators */
    public MappingTypeInfo(SimpleType XMLType, boolean useStringDelimitor)
    {
        super(XMLType);
        this.useStringDelimitor = useStringDelimitor;
    }
    
    void checkTargetType(ColumnMetaData cmeta) throws SAXException {
        checkTargetType(cmeta, null);
    }

    void checkTargetType(ColumnMetaData cmeta, XMLErrorHandler messageHandler)
    throws SAXException {
    	setDBType(cmeta.getType());
    	
		if (!cmeta.getSqlType().isStorage())
			throw new SAXException("Storage is not supported for the "
					+ cmeta.getTypeName() + " data type.");
    	
    
        if (sType.getPrimitive().getType() == PrimitiveType.UNION)   // UNION
        {
            Iterator it = sType.getMemberTypes().iterator();
            PrimitiveType wType = null;
            
            while (it.hasNext())
            {
                conversion = IMPOSSIBLE;    // other flags do not use default value
                wType = ((SimpleType)it.next()).getPrimitive();
                checkTargetType(wType, JAVA_TYPE_BINDINGS[wType.getType()],
                                    messageHandler);
                // TO IMPROVE : do global flags have to be consistent for every type of Union ? What if not ? The last are kept ?
            }
        }
        else
            checkTargetType(sType.getPrimitive(), javaType, messageHandler);
    }

    /** Schema conversion is supposed to be done
     * @param useEmptyStringAlias
     */
    public void setParameter(PreparedStatement pStmt, int i, Object value, AbstractConnection connection, boolean useEmptyStringAlias, XMLErrorHandler messageHandler) throws SQLException, XMLDBCException {
        if (value == null)   // for Oracle : because setObjet(i, null) leads to an exception
            pStmt.setNull(i, dbType.getJDBCType());
        else
        {
            if (conversion == AUTOMATIC)
                storeConvertedToSQL(pStmt, i , adaptSchemaObjectToJDBC(value));
            else if (useJDBCStringMethods())
            {
                String str = (String)value;
    
                if (targetIsString)
                {
                    if (useStringDelimitor)
                        str += STRING_DELIMITOR;
                
                    if (useEmptyStringAlias && (str.length() == 0))
                        pStmt.setString(i, STRING_DELIMITOR);
                    else if (str.length() > dbType.getLength())
                        throw new RepositoryException(RepositoryException.DATA_LOSS, "Value is too large for column.");
                    else
                    {
                        if (useStreams)
                            pStmt.setCharacterStream(i, new StringReader(str), str.length());
                        else
                            pStmt.setString(i, str);
                    }
                }
                else
                    pStmt.setString(i, str);
            }
            else    // object other than string (converted)
            {
                checkDataLoss(value, messageHandler);
                value = adaptSchemaObjectToJDBC(value);
                if (useStreams && (javaType == SCHEMA_BYTE_ARRAY))
                {
                    byte[] bytes = (byte[])value;
                    pStmt.setBinaryStream(i, new ByteArrayInputStream(bytes), bytes.length);
                }
                else
                    connection.setObject(pStmt, i, value, javaType, dbType);
            }
        }
    }

    private Object adaptSchemaObjectToJDBC(Object value) {
        
        switch (javaType) {
            case SCHEMA_DURATION:
    		case SCHEMA_URI:
    		case SCHEMA_QNAME:
                value = value.toString();
    			break;
    			
    		case SCHEMA_DATETIME:
       		    if (value instanceof java.util.Date)
       		        value = new Timestamp(((java.util.Date)value).getTime());
    			// NOTE: Use the base date type for schema API and JDBC
    			// in order to allow user generators to use the JDBC java.sql.Date type.
    			break;
    		
    		case SCHEMA_BYTE_ARRAY:
    		    if (value instanceof ByteArray)	// user generator could pass a byte array.
    		        value = ((ByteArray)value).getData();
    			break;
    			
    		case JAVA_LIST:
                value = sType.toXMLString(value);
    			break;
    
            default :
                break;
        }
        return value;
    }

    private void storeConvertedToSQL(PreparedStatement pStmt, int i, Object o) throws SQLException, XMLDBCException {
        
        // no check is performed on JDBC types because control has been performed
        // before: conversion is authorized and AUTOMATIC
        switch (javaType) {
    		case SCHEMA_DURATION:	// Java types that use strings for JDBC
    		case SCHEMA_URI:
    		case SCHEMA_QNAME:
    		case JAVA_STRING:
    		case JAVA_LIST:
                try {
                    pStmt.setBytes(i, ((String) o).getBytes("UTF-8"));
                    // The default XML one
                }
                catch (UnsupportedEncodingException e) {
                    throw new RuntimeException("UTF-8 encoding is not supported by the current JRE. Mapping from string to binary is impossible");
                }
                break;
    
            case SCHEMA_DATETIME :
    			DateTime dt = (DateTime)o;
                int n = 0;
                switch (sType.getPrimitive().getType()) {
                    case PrimitiveType.GYEAR :
                        n = dt.getYear();
                        break;
                    case PrimitiveType.GMONTH :
                        n = dt.getMonth() + 1;
                        break;
                    case PrimitiveType.GDAY :
                        n = dt.getDate();
                        break;
                    default :
                        }
                pStmt.setInt(i, n);
                break;
    
            default :
                throw new RepositoryException(
                    RepositoryException.INTERNAL_ERROR,
                    "This automatic conversion is not currently supported (javaType: "
                    + javaType + ", JDBCtype: " + dbType.getJDBCType() + ')');
        }
    }

    private void checkDataLoss(Object o, XMLErrorHandler messageHandler) throws XMLDBCException {
        switch(javaType)
        {
            case JAVA_BIG_DECIMAL:
                if (o instanceof BigDecimal && messageHandler != null && dbType.getLength() > 0)    // system generators (Long, Integer) are not controlled
                {
                    BigDecimal bd = (BigDecimal)o;
                    int precision = (int)Math.ceil(Math.log(bd.unscaledValue().doubleValue())/Math.log(10d));
                    if (precision > dbType.getLength() ||
                        (dbType.getScale() >= 0 && bd.scale() > dbType.getScale()))
                        messageHandler.warning(new RepositoryException(RepositoryException.DATA_LOSS, "Precision loss."));
                }
                break;
            case SCHEMA_BYTE_ARRAY:
                if (((ByteArray)o).getLength() > dbType.getLength())
                    throw new RepositoryException(RepositoryException.DATA_LOSS, "Value is too large for column.");
                break;
            default:
        }
    }

    public String getTypedNullValue() throws RepositoryException {
        switch (dbType.getJDBCType())
        {
            case Types.BIT: case Types.TINYINT: case Types.SMALLINT:
            case Types.INTEGER: case Types.BIGINT: case Types.FLOAT:
            case Types.REAL: case Types.DOUBLE: case Types.NUMERIC:
            case Types.DECIMAL:
                return "TO_NUMBER(null)";
            case Types.CHAR: case Types.VARCHAR: case Types.LONGVARCHAR:
            case Types.CLOB:
                return "''";
            case Types.BINARY: case Types.VARBINARY: case Types.LONGVARBINARY:
            case Types.BLOB:
                return "0";
            case Types.DATE: case Types.TIME: case Types.TIMESTAMP:
                return "TO_DATE(null)";
            default:
                throw new RepositoryException(RepositoryException.NOT_IMPLEMENTED,
                "Such query cannot be handled because of the column " + dbType.getJDBCType()
                + " SQL type");
        }
    }

	/* Should be set in the RDBMS portability layer in AbstractStatement (to ADD) */
	public String getParameter(ResultSet rs, int i, Declaration decl, ValidationContextProvider nsContext)
	throws SQLException
	{
		String ret = null;
		SimpleType type = null;
        
		if (decl == null)   // for Text nodes or generic mapping
			type = sType;
		else   // if no type, take the pseudo type from the table definition
			type = decl.getType().getValueType();
        
		if (useJDBCStringMethods() && (conversion != AUTOMATIC))
		{
			if (useStreams) // use getCharacterStream because Oracle does not tolerate getString for CLOBs
			{
				StringBuffer sb = new StringBuffer();
				Reader reader = rs.getCharacterStream(i);
				char[] buffer = new char[256];
				int len = 0;
				try
				{
					while ((len = reader.read(buffer)) > 0)
						sb.append(buffer, 0, len);
				}
				catch (IOException e)
				{
					throw new SQLException("IOException while using read() on a Reader object obtained by getCharacterStream().");
				}
				ret = sb.toString();
			}
			else
			{
				ret = rs.getString(i);
				if (targetIsString)
				{
					if (useStringDelimitor)
						ret = ret.substring(0, ret.length() - STRING_DELIMITOR_LENGTH);
					else if ((ret != null) && ret.equals(STRING_DELIMITOR))
						ret = "";
				}
				if (!storeCanonicalString()) // data is not stored converted : canonize now
					ret = type.toXMLString(ret, nsContext);
			}
		}
		else
		{
			Object o;
			if (conversion == AUTOMATIC)
				o = convertToXML(rs, i);
			else if (useStreams)    // BLOB or LONGVARBINARY
			{
				ByteArrayOutputStream out = new ByteArrayOutputStream(256);
				InputStream in = rs.getBinaryStream(i);
				byte[] buffer = new byte[256];
				int len = 0;
				try
				{
					while ((len = in.read(buffer)) > 0)
						out.write(buffer, 0, len);
				}
				catch (IOException e)
				{
					throw new SQLException("IOException while using read() on a Reader object obtained by getCharacterStream().");
				}
				o = out.toByteArray();
			}
			else
				o = rs.getObject(i);    // because first index starts at 0 and OID belongs to columns
            
			if (o != null)
				ret = type.toXMLString(o, nsContext);
		}

		return ret;
	}
    
	/* generated object type must be insensitive to a setObject-getObject U-turn */
	private Object convertToXML(ResultSet rs, int i)
	throws SQLException
	{
		switch (dbType.getJDBCType()) {
			case Types.VARBINARY :

				try {
					return new String(rs.getBytes(i), "UTF-8");
				}
				catch (UnsupportedEncodingException e) {
					throw new RuntimeException("UTF-8 encoding is not supported by the current JRE. Mapping from string to binary is impossible");
				}
			case Types.BIGINT :
			case Types.DECIMAL :
			case Types.DOUBLE :
			case Types.FLOAT :
			case Types.INTEGER :
			case Types.NUMERIC :
			case Types.REAL :
			case Types.SMALLINT :
			case Types.TINYINT :

				Timestamp ts = null;
				int n = rs.getInt(i);
				switch (sType.getPrimitive().getType()) {
					case PrimitiveType.GYEAR :
						ts = new Timestamp(n, 0, 0, 0, 0, 0, 0);
						break;
					case PrimitiveType.GMONTH :
						ts = new Timestamp(0, n, 0, 0, 0, 0, 0);
						break;
					case PrimitiveType.GDAY :
						ts = new Timestamp(0, 0, n, 0, 0, 0, 0);
						break;
					default :
						}
				return ts;

			default :
				throw new RuntimeException("This automatic conversion is not currently supported (javaType: "
						+ javaType
						+ ", JDBCtype: "
						+ dbType.getJDBCType()
						+ ')');
		}
	}
    
    public String getTypeCreationString(AbstractConnection connection) {
        PrimitiveType pType = sType.getPrimitive();
        int JDBCType = ABSTRACT_SCHEMA_JDBC_MAPPING[pType.getType()];
        
        int optimumJDBCType = optimize(
                JDBCType,
                connection.getTypeMap(),
                pType.getMaxLength(),
                pType.getTotalDigits(),
                pType.getFractionDigits(),
                (Number)pType.getMinValue(),
                (Number)pType.getMaxValue()
                );
        
        return ColumnMetaDataImpl.getTypeCreationString(
                optimumJDBCType,
                connection.getTypeMap(),
                pType.getMaxLength(),
                pType.getTotalDigits(),
                pType.getFractionDigits()
                );
    }

    private void checkTargetType(PrimitiveType pType, int localJavaType, XMLErrorHandler messageHandler)
    throws SAXException {
        int targetJDBCType = getAbstractJDBCType(dbType.getJDBCType());
        
        // First verify that setObject is authorized (redundant ?)
        if (localJavaType < JAVA_LIST
        		&& (Arrays.binarySearch(AUTHORIZED_CONVERSIONS[localJavaType], (int)targetJDBCType) < 0)    // Warning getDataType returns a short
                && !isJDBCStringIOAllowed)
            throw new SAXException(sType + " XML data type cannot be mapped on "
            + dbType + " with SQL type " + dbType.getNativeType() + ".");
        
        // Verify that the XML and SQL type fit
        
        // 1. Classify the XML type in a basic (JDBC) type within (VARCHAR,
        // VARBINARY, DECIMAL, FLOAT, DOUBLE, TIMESTAMP)
       int abstractBaseJDBCType = ABSTRACT_SCHEMA_JDBC_MAPPING[pType.getType()];
    
       // conversion type flag setting
       if (abstractBaseJDBCType == targetJDBCType)
           conversion = NONE;
       else
       {
           switch (targetJDBCType)
           {
               case Types.VARCHAR:
                   conversion = IMPLICIT;
                   break;
                   
               case Types.VARBINARY:
                   if (abstractBaseJDBCType == Types.VARCHAR)
                       conversion = AUTOMATIC;
                   break;
                   
               case Types.DECIMAL:
                   if (abstractBaseJDBCType == Types.REAL
                    || abstractBaseJDBCType == Types.FLOAT
                    || abstractBaseJDBCType == Types.DOUBLE)
                       conversion = IMPLICIT;
                   else if (pType.getType() == PrimitiveType.GYEAR
                    || pType.getType() == PrimitiveType.GMONTH
                    || pType.getType() == PrimitiveType.GDAY)
                       conversion = AUTOMATIC;
                   break;
                   
               case Types.REAL:
                   if (abstractBaseJDBCType == Types.DECIMAL
					|| abstractBaseJDBCType == Types.FLOAT
					|| abstractBaseJDBCType == Types.DOUBLE)
                       conversion = IMPLICIT;
                   break;
                   
               case Types.DOUBLE:
                   if (abstractBaseJDBCType == Types.DECIMAL
                    || abstractBaseJDBCType == Types.FLOAT
                    || abstractBaseJDBCType == Types.REAL)
                       conversion = IMPLICIT;
                   break;
                   
               default:                   
           }
           
           if (conversion == IMPOSSIBLE)
               throw new SAXException(sType + " XML data type cannot be mapped on column with SQL type "
                       + dbType.getNativeType() + ".");
       }
    
       // pseudo facet setting (could be performed only if the target type requires it)
       int  maxLen = pType.getMaxLength(),
            precision = pType.getTotalDigits(),
            scale = pType.getFractionDigits(); // if -1, no need to check
    
       switch (pType.getType())
       {
           case PrimitiveType.BOOLEAN:
               maxLen = 5;
               break;
           case PrimitiveType.DECIMAL:
               // maxlen calculation for writing in VARCHAR SQL type
               if (precision > 0) // use absolute limits if any
                   maxLen = precision + (scale > 0?2:1);   // sign and dot
               else if (scale >= 0) // if no scale and no precision no prediction on string size can be performed
               {
                   if ((pType.getMinValue() != null) && (pType.getMaxValue() != null))    // bounded
                   {
                       double maxAbs = Math.max(
                                        Math.abs(((Number)pType.getMinValue()).doubleValue()),
                                        Math.abs(((Number)pType.getMaxValue()).doubleValue())
                                        );
                       maxLen = (int)Math.ceil(Math.log(maxAbs)/Math.log(10d));
                       if (scale > 0)
                           maxLen += scale + 2;
                       else
                           maxLen += scale + 1;
                   }
               }
               break;
               
           case PrimitiveType.DATE_TIME:
               maxLen = 19; // this is the minimum value for a date string using schema syntax
               break;
           case PrimitiveType.TIME:
               maxLen = 8; // this is the minimum value for a date string using schema syntax
               break;
           case PrimitiveType.DATE:
               maxLen = 10; // this is the minimum value for a date string using schema syntax
               break;
           case PrimitiveType.GYEAR_MONTH:
               maxLen = 7; // this is the minimum value for a date string using schema syntax
               break;
           case PrimitiveType.GYEAR:
               maxLen = 4; // this is the minimum value for a date string using schema syntax
               precision = 4;
               scale = 0;
               break;
           case PrimitiveType.GMONTH_DAY:
               maxLen = 7; // this is the minimum value for a date string using schema syntax
               break;
           case PrimitiveType.GDAY:
               maxLen = 5; // this is the minimum value for a date string using schema syntax
               precision = 2;
               scale = 0;
               break;
           case PrimitiveType.GMONTH:
               maxLen = 6; // this is the minimum value for a date string using schema syntax
               precision = 2;
               scale = 0;
               break;
       }
    
       // check capacity fit
       try
       {
           switch (targetJDBCType)
           {
               case Types.VARCHAR:
               case Types.VARBINARY:
                   
                   if (maxLen < 0)
                   {
                       if (messageHandler != null)
                           messageHandler.warning(new XMLDBCException("XML data may overflow in " + dbType));
                   }
                   else if (maxLen > dbType.getLength())
                       throw new SAXException("XML data type length is larger than " + dbType);
                   break;
                   
               case Types.DECIMAL:
                   if (precision < 0)
                   {
                       if (messageHandler != null)
                           messageHandler.warning(new XMLDBCException("XML data may overflow in " + dbType));
                   }
                   else if (precision > dbType.getLength())
                       throw new SAXException("XML data type precision is larger than " + dbType);
                   else if ((scale >= 0) && (scale > dbType.getScale()))
                       throw new SAXException("XML data type scale is larger than " + dbType);
                   break;
                   
               default:
           }
       }
       catch (XMLDBCException e)
       {
           throw new SAXException("Operation aborted by user application.", e);
       }
    }

    /**
	 * <B>JDBC Types returned by the driver must be substituted to match one
	 * of the following basic types : VARCHAR, VARBINARY, DECIMAL, REAL,
	 * DOUBLE, TIMESTAMP</B>
	 * 
	 * @return
	 */
    private int getAbstractJDBCType(int targetType) throws SAXException {
		switch (targetType) {
			case Types.BIGINT :	case Types.BIT : case Types.TINYINT :
			case Types.SMALLINT : case Types.DECIMAL : case Types.NUMERIC :
			case Types.INTEGER : case DbType.BOOLEAN : case DbType.ORACLE_ROWID:
			    return Types.DECIMAL;
			    
			case Types.CHAR : case Types.CLOB :
			case Types.LONGVARCHAR : case Types.VARCHAR : 
			    return Types.VARCHAR;
			    
			case Types.DOUBLE :
		       return Types.DOUBLE;
			    
		    case Types.FLOAT : case Types.REAL :
		       return Types.REAL;
		        
			case Types.DATE : case Types.TIME : case Types.TIMESTAMP :
			    return Types.TIMESTAMP;
			    
			case Types.LONGVARBINARY :case Types.BLOB :
			case Types.VARBINARY : case Types.BINARY :
			    return Types.VARBINARY;

			case Types.NULL :
				throw new SAXException("The JDBC NULL type is not currently supported.");
			case Types.OTHER :
				throw new SAXException("The JDBC OTHER type is not currently supported.");
			case Types.REF :
				throw new SAXException("The JDBC REF type is not currently supported.");
			case Types.JAVA_OBJECT :
				throw new SAXException("The JDBC JAVA_OBJECT type is not currently supported.");
			case Types.STRUCT :
				throw new SAXException("The JDBC STRUCT type is not currently supported.");
			case Types.DISTINCT :
				throw new SAXException("The JDBC CLOB type is not currently supported.");
			case Types.ARRAY :
				throw new SAXException("The JDBC ARRAY type is not currently supported.");
			default:
			    return targetType;
		}
    }

    private int optimize(int dataType, TypeMap typeMap, long maxLength, int precision, int scale, Number min, Number max) {
        int optimum = dataType;
        
        switch (dataType) // switch according to "base" target JDBC type
        {   // and try to analyse facets(restrictions) to find a 'smaller one'
            case Types.VARCHAR :
                if (maxLength <= typeMap.getSize(Types.VARCHAR))
                    optimum = Types.VARCHAR;
                else
                    optimum = Types.LONGVARCHAR;
                break;
            case Types.DOUBLE : case Types.FLOAT : // 15 digits mantissa
                if (precision <= 7)
                    optimum = Types.REAL;
                break;
            case Types.NUMERIC :  case Types.DECIMAL : // 15 digits of precision and scale at least
                if (scale == 0) // integers
                {
                    int smallest = getSmallestInteger(precision,
                    min == null ? Double.MIN_VALUE : min.doubleValue(),
                    max == null ? Double.MAX_VALUE : max.doubleValue());
                    if (smallest != 0)
                        optimum = smallest;
                }
                // else use scale and precision coming from XML schema type
                break;
            case Types.VARBINARY :
                if (maxLength <= typeMap.getSize(Types.VARBINARY))
                    optimum = Types.VARBINARY;
                else
                    optimum = Types.LONGVARBINARY;
                break;
            default :   // Nothing to do with VARCHAR, TIMESTAMP, REAL
        }
        return optimum;
    }

    /** Calculate the most adapted JDBC integer type basing on the precision & max value given.
     * @return The calculated JDBC-SQL type from java.sql.Types
     * @param precision
     * @param min
     * @param max
     */
    private int getSmallestInteger(int precision, double min, double max) {
        int ret = 0;
        int prec = 0;
        
        /* downsizing with boundaries */
        prec = (int)Math.ceil(Math.log(Math.max(Math.abs(min), Math.abs(max)))/Math.log(10d));
        
        if (precision > 0)
            prec = Math.min(precision, prec);
        
        /* downsizing with precision */
        if (prec <= 3)
            ret = Types.TINYINT;
        else if (prec <= 5)
            ret = Types.SMALLINT;
        else if (prec <= 10)
            ret = Types.INTEGER;
        else if (prec <= 19)
            ret = Types.BIGINT;
        
        return ret;
    }

}
