/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.mapper.mapping;


import java.sql.SQLException;
import java.util.*;

import org.xml.sax.SAXException;
import org.xquark.mapper.RepositoryException;
import org.xquark.mapper.dbms.AbstractConnection;
import org.xquark.mapper.metadata.Repository;
import org.xquark.xpath.XTree;


/** Shared memory data structure representing an XQuark Mapping file.
 *
 */
public class RepositoryMapping extends XTree
implements MappingSet
{
    private static final String RCSRevision = "$Revision: 1.2 $";
    private static final String RCSName     = "$Name: XQ1_1_1 $";
    
    /**
	 *
	 */
	//
    //  DATA
    //
    
	/** List used used during construction (in fact construction order,
	 * i.e. table index order).
	 */
	private ArrayList tableMappingList = new ArrayList();

	/** Array used notably when need to get table column mappings in no
	 * particular order or for direct access (in fact construction order,
	 * i.e. table index order).
	 */
	private TableMapping[] tableMappings;

	/** List used notably when need to get table column mappings in the
	 * relational table reference order.
	 */
	private TableMapping[] mostDependantFirst;

	/** List used notably when need to get table column mappings sorted by
	 * name (ascending).
	 */
//	private List tableNameSorted;

	/** Metadata for user tables referenced by mapping */
	private HashMap tableMetadata = new HashMap();
	
	/** Metadata nodes for Types indexed by schema components */
	private HashMap typeMappings = null;
    private Repository rep;
    
    public RepositoryMapping(Repository rep)
    {
        this.rep = rep;
    }
    
    public Repository getRepository()
    {
        return rep;
    }
    
    public MappingNode addTypeNode(MappingNode node)
    {
        if (typeMappings == null)
            typeMappings = new HashMap();
        typeMappings.put(node.getSchemaComponent(), node);
        return node;	// return the created node, not the replaced returned by put (should not happen)
    }
    
    public MappingNode getTypeNode(org.xquark.schema.Type type)
    {
        if (typeMappings == null)
            return null;
        return (MappingNode)typeMappings.get(type);
    }
    
    public void addTableMapping(TableMapping tm)
    {
        tableMappingList.add(tm);
    }
    
    public TableMapping getTableMapping(int index)
    {
        return tableMappings[index];
    }
    
    public int getTableCount()
    {
    	if (tableMappings == null)
    		return tableMappingList.size();
    	else
        	return tableMappings.length;
    }
    
    /** Should overwrite the default registration
     */
    public void register(MappingNode wNode, Mapping mapping)
    {
        // update the table mapping list in scope for ascendants
        if (mapping instanceof TableMapping)
        {
            for (MappingNode elt = wNode;
            (elt != null) && elt.addMappingInScope((TableMapping)mapping);
            elt = (MappingNode)elt.getParent());
        }
        // set the buildable flag and builder node
        if (mapping instanceof ColumnMapping)
        {
            ColumnMapping column = (ColumnMapping)mapping;
            if (column.getGenerator() instanceof ValueGenerator)    // other are not needed by reconstruction
            {
                MappingNode elt;
                for ( elt = wNode;
                (elt.getTableMappings() == null) || !elt.getTableMappings().contains(column.getTableMapping());
                elt = (MappingNode)elt.getParent())
                {
                    elt.setBuildable(false);
                }
                int tableAncestorDepth = elt.getDepth();
                // Now table node has been located replace in the hierachy if the node
                // is "upper" in the tree than the current builder.
                for ( elt = wNode;
                elt.getDepth() > tableAncestorDepth; // diff 0 is default
                elt = (MappingNode)elt.getParent())
                {
                    if ((elt.getBuilder() == null) || (tableAncestorDepth < elt.getBuilder().getDepth()))
                        elt.setBuilderDiff(elt.getDepth() - tableAncestorDepth);
                }
            }
        }
        
        wNode.register(mapping);
    }
    
    public String toString()
    {
        String ret = super.toString();
        if (typeMappings != null)
             ret += "\nTable mappings list:\n" + tableMappings.toString();
        return ret;
    }
    

	/**
	 * Verify mapping does not define multiple table mapping with different
	 * join columns (used by repository not mapper).
	 * COULD BE PERFORMED ON-LINE (on register) ?
	 */
	public void checkRedundantTables() throws RepositoryException
	{
		// browse the array to process doubles
		MappingSetIterator it = new MappingSetIteratorImpl(tableMappings);
		while (it.hasNext())
		{
			BaseTableMappingImpl tm = (BaseTableMappingImpl)it.next().getTableMapping();
			
			// if table mapping on the same table, check it is equivalent or throw an exception
			if (!it.isTableFirstOccurence())
			{
				TableMapping tf = tableMappings[it.getFirstOccurenceTableIndex()];
				
				// set the shared flag
				tm.setTableShared();
				
				// check the join columns
				if (!areColumnArraysEquivalent(
					tf.getJoinColumns(),
					tm.getJoinColumns()
					))
					throw new RepositoryException(
						RepositoryException.NOT_ALLOWED,
							"Multiple mappings defined on the same table("
							+ tm.getTableName()
							+ ") must have the same key specification.");
							
				// check the statements can be shared
				if (tm.isShareable() &&  !(
					areColumnArraysEquivalent(tm.getColumnMappings(), tf.getColumnMappings())
				&& areColumnArraysEquivalent(tm.getSelectColumns(), tf.getSelectColumns())
				&& areColumnArraysEquivalent(tm.getUpdateColumns(), tf.getUpdateColumns())))
					tm.setStatementsNotShareable();
			}
		}
	}

	/**
	 * Return an iterator that browse table mappings with join tables
	 * (unclustered) and different user tables.
	 * @return Iterator
	 */
	public Iterator joinTableIterator()
	{
		return new TableIterator(true);
	}
	
    ////////////////////////////////////////////////////////////////////
    //                      MappingSet implementation
    ////////////////////////////////////////////////////////////////////
    public int getMappingStatementsCount()
    {
        return getTableCount();
    }
    
    public MappingInfo getMappingStatements(int index)
    {
        return getTableMapping(index);
    }
    
    public MappingSetIterator getSortedTableMappingSetIterator(boolean mostDependantFirst)
    {
        return new MappingSetIteratorImpl(this.mostDependantFirst, !mostDependantFirst);
    }
    
    public MappingSetIterator getMappingSetIterator()
    {
        return new MappingSetIteratorImpl(tableMappings);
    }
    
	////////////////////////////////////////////////////////////////////
	//                      	PACKAGE METHODS
	////////////////////////////////////////////////////////////////////
	void initialize(AbstractConnection conn) throws SAXException
	{
		initializeArrays(conn);
	}

	TableMetaData getTableMetaData(String tableName, AbstractConnection conn)
	throws SAXException
	{
		TableMetaData ret = (TableMetaData) tableMetadata.get(tableName);
		if (ret == null)
		{
			try
			{
				if (conn.checkTable(tableName))
				{
					ret = new TableMetaDataImpl(tableName, conn);
					tableMetadata.put(tableName, ret);
				}
			}
			catch (SQLException ex)
			{
				throw new SAXException(
					"Error while fetching table " + tableName + " metadata",
					ex);
			}
		}
		return ret;
	}
	////////////////////////////////////////////////////////////////////
	//                      		PRIVATE
	////////////////////////////////////////////////////////////////////
	private void initializeArrays(AbstractConnection conn) throws SAXException
	{
		// create table mapping array
		tableMappings = (TableMapping[])tableMappingList.toArray(new TableMapping[tableMappingList.size()]);
		tableMappingList = null;
		

		// initialize key dependencies
		TableMappingImpl tm = null;
		for (int i = 0; i < tableMappings.length; i++) {
		    tm = (TableMappingImpl)tableMappings[i];
			// NOTE: some mapping (type mapping) may not be used and thus are not initialized
		    // but removing it now would corrupt ref index...
			if (!tm.isInitialized())
			    tm.initialize(conn);
			tm.initializeDependencies();
		}
			
		// create ordered table list
		mostDependantFirst = new TableMapping[tableMappings.length];
		System.arraycopy(tableMappings, 0, mostDependantFirst, 0, tableMappings.length);
		Arrays.sort(mostDependantFirst, new Comparator()
		{
			public int compare(Object o1, Object o2)
			{
				BaseTableMappingImpl tm1 = (BaseTableMappingImpl)((MappingInfo)o1).getTableMapping();
				BaseTableMappingImpl tm2 = (BaseTableMappingImpl)((MappingInfo)o2).getTableMapping();
					return tm2.getDependencies().size()
						- tm1.getDependencies().size();
			}
		});
	}

	/**
	 * This iterator only returns table mappings with join table only
	 * one time by user table.
	 * 
	 */
	private class TableIterator implements Iterator
	{
		MappingSetIterator iterator;
		MappingInfo next = null;
		boolean unclusteredOnly;
		
		public TableIterator(boolean unclusteredOnly)
		{
			iterator = new MappingSetIteratorImpl(tableMappings);
			this.unclusteredOnly = unclusteredOnly;
		}

		public TableIterator()
		{
			this(false);
		}

		public boolean hasNext()
		{
			while (next == null && iterator.hasNext())
			{
				next = iterator.next();
				if (!keepCondition())
					next = null;
			}
			return next != null;
		}
		
		private boolean keepCondition()
		{
			return iterator.isTableFirstOccurence()
			&& (!unclusteredOnly || !next.getTableMapping().isClustered());
		}

		public Object next()
		{
			MappingInfo ret = null;
			if (hasNext())
			{
				ret = next;
				next = null;
			}
			return ret;
		}
		
		public void remove()
		{
			throw new UnsupportedOperationException();
		}

	}
	
	private boolean areColumnArraysEquivalent(
		ColumnMapping[] t1,
		ColumnMapping[] t2)
	{
		boolean equivalent = true;
		if (t1 != null && t2 != null)
		{
			if (t1.length != t2.length)
				equivalent = false;

			// loop looking for difference
			for (int j = 0; j < t2.length; j++)
			{
				if (!t1[j]
					.getColumnName()
					.equals(t2[j].getColumnName()))
					equivalent = false;
			}
		}
		else if (t1 != t2)
			equivalent = false;
		return equivalent;
	}

}
