/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.mapper.mapping;

import java.util.List;
;

/**
 * Interface giving access to information for mapping XML elements
 * on rdbms tables.
 *
 */
public interface TableMapping extends Mapping, MappingInfo
{



	/**
	 * Setter for the table mapping index (unique ID).
	 * @param index a positive integer.
	 */
	public void setIndex(int index);

	/**
	 * Accessor to user table relational metadata.
	 * @return a TableMetaData object.
	 */
	public TableMetaData getMetaData();

	/**
	 * Accessor to column mapping information for all the columns of the user
	 * table.
	 * @return an iterator on ColumnMapping objects.
	 */
//	public Iterator getColumnMappingIterator();

	/**
	 * Accessor to column mapping in a dense array where index does not match
	 * the column index. Unmapped columns are discarded. Implemented because
	 * query module uses arrays instead of iterators
	 * @return a ColumnMapping array.
	 */
	public ColumnMapping[] getColumnMappings();

	/**
	 * Accessor to column mapping information by name.
	 * @param name the name of the user table column.
	 * @return a ColumnMapping object.
	 */
	public ColumnMapping getColumnMapping(String name);

	/**
	 * Accessor to column mapping information by index.
	 * @param index the index of the user table column. This index depends on
	 * the order of the column in the user table.
	 * @return a ColumnMapping object or null if the user column has no mapping.
	 */
	public ColumnMapping getColumnMapping(int index);

	/**
	 * Returns the number of columns in the user table.
	 */
	public int getColumnCount();

	/**
	 * Returns the number of column mappings in the table mapping. It may be
	 * inferior to the actual number of columns in the table.
	 * @return the number of column mappings.
	 */
	public int getColumnMappingCount();

	/**
	 * Returns the column mappings of the relational user table PK or if no PK
	 * exists, the columns mapping which flag "inKey" is true.
	 * @return a ColumnMapping array.
	 */
	public ColumnMapping[] getPrimaryKey();

	/**
	* Returns the number of column mappings involved in the user table PK or
	* if no PK exists, the columns mapping which flag "inKey" is true.
	* @return a positive integer.
	*/
	public int getKeyColumnCount();

	/**
	 * Returns the number of columns of the OID table (external XQuark index
	 * table), including XQuark index columns. <b>Different from
	 * {@link #getJoinColumns()}.length</b>.
	 * @return a positive integer.
	 */
	public int getOIDTableColumnCount(); // includes docOID & OID (OID table)

	/**
	 * Returns the column mappings involved in natural join between XQuark index
	 * and user tables. That is not necessarily the relational user table PK.
	 * <b>XQuark index columns are not included</b>.
	 * @return a ColumnMapping array.
	 */
	public ColumnMapping[] getJoinColumns();

	/**
	 * Returns the columns of the relational user table that will be updated
	 * if the insert mode (action) is "update".
	 * @return a ColumnMapping array. May be null if action is not "update".
	 * @see #getAction()
	 */
	public ColumnMapping[] getUpdateColumns();

	/**
	 * Returns the number of columns of the relational user table that will be
	 * updated if the insert mode (action) is "update".
	 * @return a positive integer.
	 */
	public int getUpdateColumnCount();

	/**
	 * Returns the columns of the relational user table that will be used for
	 * retrieving the tuple to be selected or updated if the insert mode
	 * (action) is different from "insert".
	 * @return a ColumnMapping array. May be null if action is "insert".
	 * @see #getAction()
	 */
	public ColumnMapping[] getSelectColumns();

	/**
	 * Returns the number of columns of the relational user table that will be
	 * used for retrieving the tuple to be selected or updated if the insert
	 * mode (action) is different from "insert".
	 * @return a positive integer.
	 */
	public int getSelectColumnCount();

	/**
	 * Returns the columns of the relational user table that will be
	 * retrieved from the tuple to be selected or updated if the insert mode
	 * (action) is different from "insert".
	 * @return a ColumnMapping array. May be null if action is "insert".
	 * @see #getAction()
	 */
	public ColumnMapping[] getFetchColumns();

	/**
	 * Returns the number of columns of the relational user table that will be
	 * retrieved from the tuple to be selected or updated if the insert mode
	 * (action) is different from "insert".
	 * @return a positive integer.
	 */
	public int getFetchColumnCount();

	/**
	 * Returns column mapping corresponding to generators or to optional or
	 * nillable elements which column are nullable or have default value. For
	 * these columns, a default value is generated at the beginning of the
	 * element of the table mapping so that it may be overwritten if the
	 * element eventually exists.
	 * @return a list containing ColumnMapping objects.
	 */
	public List initParameters();

	/**
	 * Returns column mapping that are ready (and must be added to tuple)
	 * when the table mapping is complete. Mainly used to store an element
	 * value when a table mapping has been defined on it for instance to store
	 * attributes.
	 * @return a list containing ColumnMapping objects.
	 */
	public List finalParameters();

	/**
	 * Returns the insertion mode of the table mapping.
	 * @return an integer within the values INSERT, CHECK, UPDATE & SELECT
	 * defined in {@link MappingConstants}.
	 */
	public int getAction();

	/**
	 * Returns the batch size used for storage in the mapping's table. May be
	 * the default one, the user set one (in the configuration or mapping file)
	 * or a value imposed by the data types (for instance LOBs enforce batch
	 * size to 1.
	 * @return a positive non-null integer.
	 */
	public int getBatchSize();

	/**
	 * Returns the cluster mode of the table mapping. "Clustered" means XQuark
	 * index columns are incorporated in user tables in the form of generators.
	 * This is only possible if the insertion mode is "INSERT".
	 * @return true if the OID table is clustered
	 */
	public boolean isClustered();

	/**
	 * Returns the XQuark index path OID column index either in the user table,
	 * if clustered, either in the OID table if not.
	 * @return an integer index beginning with 0.
	 */
	public int getPathIDIndex();

	/**
	 * Returns the name of the column containing the path OID (both in
	 * clustered and not clustered case).
	 * @return the path OID column name.
	 */
	public String getPathIDColumnName();

	/**
	 * Returns the XQuark index node UOID column index either in the user table,
	 * if clustered, either in the OID table if not.
	 * @return an integer index beginning with 0.
	 */
	public int getUOIDIndex();

	/**
	 * Returns the name of the column containing the node UOID (both in
	 * clustered and not clustered case).
	 * @return the node UOID column name.
	 */
	public String getUOIDColumnName();

	/**
	 * Returns the SQL statement fragment corresponding to the columns of the
	 * user table to put in a SELECT clause.
	 * @return a comma-separated list of column names.
	 */
	public String getColumnList();
	
	/**
	 * Returns the SQL statement fragment corresponding to the columns of the
	 * user table to put in a SELECT clause.
	 * @return a comma-separated list of column names using table aliases.
	 */
	public String getNamedColumnList();
	
	/**
	 * Returns the SQL statement fragment corresponding to the columns of the
	 * user table to put in a SELECT clause of a prepared statement.
	 * @return a comma-separated list of question marks.
	 */
	public String getWildcards();
	
	/**
	 * Returns the SQL statement fragment corresponding to the columns of the
	 * user table that participate in the join with the OID table.
	 * @return a comma-separated list of column names.
	 */
	public String getJoinColumnList();

	/**
	 * Returns the SQL statement fragment corresponding to the columns of the
	 * user table that participate in the join with the OID table of a prepared
	 * statement.
	 * @return a comma-separated list of question marks.
	 */
	public String getJoinWildcards();
	
	/**
	 * Returns the DDL statement fragment corresponding to the columns of the
	 * user table that participate in the join with the OID table of a prepared
	 * statement. Used for OID table creation.
	 * @return a comma-separated list DDL column specifications.
	 */
	public String getJoinColumnTypes();

	/**
	 * Returns the SQL statement fragment to put in a WHERE clause to perform
	 * natural join between the OID table and the user table.
	 * @return null if clustered.
	 */
	public String getJoinCondition();
	
	/**
	 * Says if all statements can be shared among the table mappings defined on
	 * the same user table.
	 * @return boolean
	 */
	public boolean isShareable();

	/**
	 * Says if this mapping is shared by multiple paths. This is the case for
	 * default mapping an most of the time type mappings.
	 * @return boolean
	 */
	public boolean isShared();

	/**
	 * True if this table mapping shares its user table with other table
	 * mappings or if this mapping is shared itself, i.e., multiple path are
	 * mapped on the same user table.
	 * @return boolean
	 */
	public boolean isTableShared();
}
