/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.mapper.mapping;

import java.util.*;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.xml.sax.SAXException;
import org.xquark.jdbc.typing.ColumnMetaData;
import org.xquark.mapper.dbms.AbstractConnection;
import org.xquark.schema.ElementDeclaration;
import org.xquark.schema.SchemaComponent;
import org.xquark.schema.Type;

/**
 * Implementation of the TableMapping interface.
 *
 * Every TableMapping object owns a set of ColumnMapping object corresponding
 * to its table's columns.
 *
 */
class TableMappingImpl extends BaseTableMappingImpl
{
    private static final String RCSRevision = "$Revision: 1.9 $";
    private static final String RCSName     = "$Name: XQ1_1_1 $";
    
    private static final short USE_PK = 0;
    private static final short USE_KEYGEN = 1;
    private static final short USE_ALL = 2;
    
    private static Log log = LogFactory.getLog(TableMappingImpl.class);
    
    /* Set of TableMapping object which this TableMapping is dependent on */
    
    /* USE ? */
    private HashSet mappedElements = new HashSet();
    /* USE ? */
    private HashSet endingElements = new HashSet();
    /* String for JDBC PreparedStatements used to access the OID */
    protected String selectStatement;
    protected String updateStatement;
    
	private short joinMode = -1;
        
    /* Number of columns that are used in the where clause to retrieve rows */
    private int keyGenColumnCount = 0;
    private int selectColumnCount = 0;
    private int fetchColumnCount = 0;
    private int updateColumnCount = 0;
	private int columnMappingCount = 0;	// different from columnCount
	private ColumnMappingImpl[] denseColumnMappingArray = null; // i.e., insert column mapping

    private int action;
    private boolean clustered = false; // if true, OID table is clusterd to the value table
    private ColumnMappingImpl uoidColumn = null; // column corresponding to the universal OID if clustering
    private ColumnMappingImpl pathColumn = null; // column corresponding to the path ID if clustering
    
    private int batchSize; // JDBC batch size
    /* This flag is used to desactivate OID tables in mapping for mapper */
	private boolean userTablesOnly = false;
	private boolean shared = false;
    
    /** Constructor.
     * @param colMapping the RepositoryMapping object to which this object belongs.
     * @param comp the SchemaComponent associated with this "mapping component".
     * @param name relational table name.
     * @param conn a connection allowing access to the metadata for the table retrieved from JDBC
     * @param action the action to perform on the rows of the table (check, insert, update, ...)
     * @param generate If true, mapping is loaded even if relational tables
     * are not present : mapping is loaded for table generation.
     */
    TableMappingImpl(RepositoryMapping colMapping, SchemaComponent comp,
    String tableName, AbstractConnection conn, int action, boolean generate,
    int batchSize, boolean userTablesOnly) throws SAXException
    {
        super(colMapping, comp, conn);
        this.name = tableName;
        this.userTablesOnly = userTablesOnly;
    	tableMetaData = colMapping.getTableMetaData(name, conn);
        shared = comp instanceof Type;
        
        // Check table existence
    	if (!generate && tableMetaData == null)
    		throw new SAXException("Table " + name + " does not exist in database.");
    		
        if (tableMetaData == null)
        	// create column mapping array using abitrary size (no metadata)
        	columnMappings = new ColumnMapping[20];
        else	// create column mapping array
	    	columnMappings = new ColumnMapping[tableMetaData.getColumnCount()];

        this.action = action;
        this.batchSize = batchSize;
    }
    
    public int getBatchSize() { return batchSize;}
    
    public int getOIDTableColumnCount() { return OIDColumnMappings.length + 2;}
    
    public int getSelectColumnCount() { return selectColumnCount;}
    
    public int getFetchColumnCount() { return fetchColumnCount;}
    
    public int getKeyGenColumnCount() { return keyGenColumnCount;}
    
    public int getUpdateColumnCount() { return updateColumnCount;}
    
	public int getColumnMappingCount() { return columnMappingCount;}
	
	public ColumnMapping[] getColumnMappings() { return denseColumnMappingArray;}
	
    public int getAction() { return action;}
    
    /**
     * Returns true if the OID table is clustered.
     */
    public boolean isClustered() { return clustered;}
    
	/**
	 * @see org.xquark.mapper.mapping.TableMapping#isShared()
	 */
	public boolean isShared() { return shared;}

    /**
     * if clustered, return the OID column index.
     */
    public int getPathIDIndex()
    {
        if (pathColumn != null)
            return pathColumn.getColumnIndex();
        else
            return 1;
    }
    
    /**
     * if clustered, return the root OID column index.
     */
    public int getUOIDIndex()
    {
        if (uoidColumn != null)
            return uoidColumn.getColumnIndex();
        else
            return 0;
    }
    
    /**
     * Returns the name of the column containing the path OID (both in clustered
     * and not clustered case).
     */
    public String getPathIDColumnName()
    {
        if (pathColumn != null)
            return pathColumn.getColumnName();
        else
            return super.getPathIDColumnName();
    }
    
    /**
     * Returns the name of the column containing the node UOID (both in clustered
     * and not clustered case).
     */
    public String getUOIDColumnName()
    {
        if (uoidColumn != null)
            return uoidColumn.getColumnName();
        else
            return super.getUOIDColumnName();
    }
    /**
     * USE ? Never updated !
     */
    public boolean isTerminatedBy(ElementDeclaration decl)
    {
        return endingElements.contains(decl);
    }
    
    ////////////////////////////////////////////////////////////////////////////
    //                      MappingInfo IMPLEMENTATION
    ////////////////////////////////////////////////////////////////////////////
    // performed by CollectionMappingInfo in the Repository case
    public String getOIDTableName() { return null;}
    
    // performed by CollectionMappingInfo in the Repository case
    public String getOIDInsertStatement() { return null;}
    
    public String getSelectStatement() { return selectStatement;}
    
    public String getUpdateStatement() { return updateStatement;}
    
    ////////////////////////////////////////////////////////////////////////////
    //                          PACKAGE PRIVATE
    ////////////////////////////////////////////////////////////////////////////
    /**
     * Used by the mapping loader when building element mapping.
     * @param decl a schema ElementDeclaration.
     */
    void addMappedChild(ElementDeclaration decl)
    {
        if (!mappedElements.contains(decl))
            mappedElements.add(decl);
    }
    
    /**
     * Used by the mapping loader to perform TableMapping initialization
     * including column mapping check and statement building.
     * Called after ColumnMapping registration.
     */
    void initialize(AbstractConnection conn) throws SAXException
    {
        initializeColumnMappings();
        checkMapping();
        initializeCluster();
        initializeStatements(conn);
        initializeEndingElements();
    }
    
    boolean isInitialized() { return primaryKeyColumnMappings != null;}
    
    ////////////////////////////////////////////////////////////////////////////
    //                              PRIVATE
    ////////////////////////////////////////////////////////////////////////////
    /**
     * Checks against column's metadata info got from JDBC that if a
     * column is not registered is authorized because it's nullable
     */
    private void initializeColumnMappings() throws SAXException
    {
        if (getColumnMappingCount() == 0)
            throw new SAXException("Table mapping must specify at least one column mapping");
        
        /* Set the Key columns used for user table join */
        int uoidColumnCount = 0;
        
        if (keyColumnCount > 0) // A PK exists use it
        {
            joinMode = USE_PK;
            uoidColumnCount = keyColumnCount;
        }
        else if (keyGenColumnCount > 0) // use key generators
        {
            joinMode = USE_KEYGEN;
            uoidColumnCount = keyColumnCount = keyGenColumnCount;
        }
        else    // use all columns
        {
            joinMode = USE_ALL;
            uoidColumnCount = getColumnCount();
        }
        
        primaryKeyColumnMappings = new ColumnMapping[keyColumnCount];
        OIDColumnMappings = new ColumnMapping[uoidColumnCount];
        updateColumnMappings = new ColumnMapping[updateColumnCount];
        selectColumnMappings = new ColumnMapping[selectColumnCount];
        fetchColumnMappings = new ColumnMapping[fetchColumnCount];
		denseColumnMappingArray = new ColumnMappingImpl[columnMappingCount];
        
        int joinColumnCount = 2;    // Start with 2 because of UOID and path
        int OIDcount = 0;
        
        Iterator it = getColumnMappingIterator();
        while (it.hasNext())
        {
            ColumnMappingImpl mapping = (ColumnMappingImpl)it.next();
            
        	denseColumnMappingArray[mapping.getInsertColumnIndex()] = mapping;
        	
            switch (joinMode)
            {
                case USE_PK:
                    if (mapping.getKeyColumnIndex() != -1)
                    {
                        primaryKeyColumnMappings[mapping.getKeyColumnIndex()] = mapping;
                        mapping.setJoinColumnIndex(joinColumnCount++);
                        OIDColumnMappings[mapping.getKeyColumnIndex()] = mapping;
                    }
                    break;
                case USE_KEYGEN:
                    if (mapping.getKeyGenColumnIndex() != -1)
                    {
                        primaryKeyColumnMappings[mapping.getKeyGenColumnIndex()] = mapping;
                        mapping.setJoinColumnIndex(joinColumnCount++);
                        OIDColumnMappings[mapping.getKeyGenColumnIndex()] = mapping;
                    }
                    break;
                case USE_ALL:
                    mapping.setJoinColumnIndex(joinColumnCount++);
                    OIDColumnMappings[OIDcount++] = mapping;
                    break;
            }
            if (mapping.getSelectColumnIndex() != -1)
                selectColumnMappings[mapping.getSelectColumnIndex()] = mapping;
            
            if (mapping.getFetchColumnIndex() != -1)
                fetchColumnMappings[mapping.getFetchColumnIndex()] = mapping;

            if (mapping.getUpdateColumnIndex() != -1)    // should be mutually exclusive with select
                updateColumnMappings[mapping.getUpdateColumnIndex()] = mapping;
        }
    }

    /**
     * Checks against column's metadata info got from JDBC that if a
     * column is not registered is authorized because it's nullable
     */
    private void checkMapping() throws SAXException
    {
        // for update, check or select, a select column must exist at least
        if (action != INSERT && selectColumnCount == 0)
            throw new SAXException("When using the SELECT, CHECK or UPDATE mode, there must be at least one 'select' column.");
        
        // for update, an update column must exist at least
        if (action == UPDATE && updateColumnCount == 0)
            throw new SAXException("When using the UPDATE mode, there must be at least one column to update. Else use the CHECK mode.");
        
        // The following tests are not performed when generating table scripts
        if (getMetaData() != null)
        {
            // for update or select : mapping can be missing for nullable columns
            if ((action == INSERT) || (action == CHECK))
            {
                Iterator it = tableMetaData.getColumnsMetaData().iterator();
                while (it.hasNext())
                {
                    ColumnMetaData cmeta = (ColumnMetaData)it.next();
                    ColumnMapping mapping = getColumnMapping(cmeta.getColumnName());

                    // check mapping completeness
                    if (mapping == null && !cmeta.isOptional())
                        log.warn("Missing mapping specification for " + cmeta);
                }
            }
            
            // REPOSITORY ONLY
            if (!userTablesOnly)
            {
                if (action == INSERT)
                {
                    if (joinMode != USE_PK) // PK is unique...
                    {
                        // Check that the "key" generators are unique
                        Set keys = new HashSet();
                        ColumnMappingImpl column;
                        
                        for (int i = 0; i < denseColumnMappingArray.length; i++)
                        {
                            column = denseColumnMappingArray[i];
                            if (column.isInJoin())
                                keys.add(column.getMetaData());
                        }
                        
                        // compare this set to every constraint (equals is performed on
                        // ColumnMetadata objects and not column names for performance reasons)
                    	Iterator it = getMetaData().getUniqueConstraints().iterator();
                        boolean matches = false;
                        
                        while (it.hasNext() && !matches)
                            matches |= keys.containsAll((Collection)it.next());
                        
                        // TO IMPROVE : do not force to have a unique constraint when generator UOID is inKey
                        if (!matches)
                            throw new SAXException("When using the INSERT mode, a relational UNIQUE constraint (or Primary Key) must be defined on key columns.");
                    }
                }
                else if (action != SELECT)  // update or check : key columns must be included in select columns
                {
                    HashSet inter = new HashSet(Arrays.asList(OIDColumnMappings));
                    inter.retainAll(Arrays.asList(updateColumnMappings));
                    // not clustered
                    if (!inter.isEmpty())
                        throw new SAXException("When using the CHECK or UPDATE mode, Primary Key columns cannot be updated. Check that attribute 'inSelect' is true for columns " + inter);
                    // Note: inKey generators are automatically excluded from update.
                }
            }
        }
    }
    
    /**
     * Checks if the OID table is to be clustered to the values
     */
    private void initializeCluster() throws SAXException
    {
        ColumnMappingImpl column;
        boolean oid = false, root = false;
        SystemVariableGenerator sysGen;
        for (int i = 0; i < denseColumnMappingArray.length; i++)
        {
            column = denseColumnMappingArray[i];
            if (column.getGenerator() instanceof SystemVariableGenerator)
            {
                sysGen = (SystemVariableGenerator)column.getGenerator();
                if (sysGen.getType() == PATHID_CODE)    // not mandatory
                    pathColumn = column;
                else if (sysGen.getType() == UOID_CODE)
                    uoidColumn = column;
            }
        }
        if (userTablesOnly || ((action == INSERT)
        && (uoidColumn != null) && (uoidColumn.getKeyGenColumnIndex() >= 0)
        && (pathColumn != null) && (pathColumn.getKeyGenColumnIndex() >= 0)))
            clustered = true;
    }
    
    /**
     * Build SQL strings for use in JDBC PreparedStatements/.
     */
    void initializeStatements(AbstractConnection conn)
    {
        if (tableMetaData == null)
            return;
        StringBuffer lColumnList = new StringBuffer(),
			lNamedColumnList = new StringBuffer(),
			lWildcards = new StringBuffer();
        
        // Basic patterns construction
        for (int i = 0; i < denseColumnMappingArray.length; i++)
        {
            ColumnMappingImpl mapping = denseColumnMappingArray[i];
            String columnName = mapping.getColumnName();
            if (conn.useDoubleQuotes4DDLNames())
				columnName = '"' + columnName + '"';
            if (lColumnList.length() > 0)
            {
                lColumnList.append(", ");
                lNamedColumnList.append(", ");
                lWildcards.append( ", ");
            }
            lColumnList.append(columnName);
            lNamedColumnList.append("w.");
            lNamedColumnList.append(columnName);
            lWildcards.append("?");
            if (mapping.isInJoin())
            {
                /* getting OID table columns creation info from database metadata */
                ColumnMetaData cmeta = mapping.getMetaData();
                String columnType = null;
                if (cmeta.getColumnSize() > 0)
                    columnType = cmeta.getTypeCreationString();
                if (joinColumnList.length() > 0)
                {
                    joinColumnList += ", ";
                    joinWildcards += ", ";
                    joinColumnTypes += ", ";
                    joinCondition += " AND ";
                }
                joinColumnList += columnName;
                joinWildcards += "?";
                joinColumnTypes += columnName + " " + columnType;
                if (mapping.getMinOccurs() > 0)	// SR 14/05/01 was 1 ! =multivaluated and not mandatory...
                    joinCondition += "v."+columnName+"=w."+columnName;
                else
                    joinCondition += "((v."+columnName+"=w."+columnName+") OR (v."+columnName+" IS NULL AND w."+columnName+" IS NULL))";
            }
        }
        columnList = lColumnList.toString();
        namedColumnList = lNamedColumnList.toString();
        wildcards = lWildcards.toString();
        
        lColumnList.setLength(0);
        for (int i = 0; i < selectColumnMappings.length; i++)
        {
            if (i > 0)
            	lColumnList.append(" AND ");
            if (conn.useDoubleQuotes4DDLNames())
            	lColumnList.append('"');
            lColumnList.append(selectColumnMappings[i].getColumnName());
            if (conn.useDoubleQuotes4DDLNames())
            	lColumnList.append('"');
            lColumnList.append(" = ?");
        }
        String selectCondition = lColumnList.toString();
        
        lColumnList.setLength(0);
        for (int i = 0; i < updateColumnMappings.length; i++)
        {
            if (i > 0)
            	lColumnList.append(", ");
            if (conn.useDoubleQuotes4DDLNames())
            	lColumnList.append('"');
            lColumnList.append(updateColumnMappings[i].getColumnName());
            if (conn.useDoubleQuotes4DDLNames())
            	lColumnList.append('"');
            lColumnList.append(" = ?");
        }
        String updateColumnList = lColumnList.toString();
        
        lColumnList.setLength(0);
        for (int i = 0; i < fetchColumnMappings.length; i++)
        {
            if (i > 0)
            	lColumnList.append(", ");
            if (conn.useDoubleQuotes4DDLNames())
            	lColumnList.append('"');
            lColumnList.append(fetchColumnMappings[i].getColumnName());
            if (conn.useDoubleQuotes4DDLNames())
            	lColumnList.append('"');
        }
        String fetchColumnList = lColumnList.toString();
        
        // Statements construction
        String tableName = getTableName();  // collection name is fake here : for default mapping
		if (conn.useDoubleQuotes4DDLNames())
			tableName = '"' + tableName + '"';
        StringBuffer sql = new StringBuffer();
        sql.append("INSERT INTO ");
        sql.append(tableName);
        sql.append('(');
        sql.append(columnList);
        sql.append(") VALUES(");
        sql.append(wildcards);
        sql.append(')');
        insertStatement = sql.toString();
        if (selectCondition.length() > 0)
        {
            sql.setLength(0);
            sql.append("SELECT ");
            if (fetchColumnCount == 0)
                sql.append('1');
            else
            	sql.append(fetchColumnList);
            sql.append(" FROM ");
            sql.append(tableName);
            sql.append(" WHERE ");
            sql.append(selectCondition);
            selectStatement = sql.toString();
            if (updateColumnList.length() > 0)
            {
                sql.setLength(0);
                sql.append("UPDATE ");
                sql.append(tableName);
                sql.append(" SET ");
                sql.append(updateColumnList);
                sql.append(" WHERE ");
                sql.append(selectCondition);
                updateStatement = sql.toString();
            }
        }
    }
    
    /* Not implemented : USE ? */
    void initializeEndingElements() {}
    
    /**
     * Called by the RepositoryMapping object once TableMappings have
     * been constructed to build dependance link between TableMappings
     * due to references between ColumnMapping objects.
     */
    Collection initializeDependencies() throws SAXException
    {
        if (dependencies == null)
        {
            dependencies = new HashSet();
        	for (int i = 0; i < denseColumnMappingArray.length; i++)
        	{
                int tableIndex = denseColumnMappingArray[i].getTableRefIndex();
                if (tableIndex != -1)
                {
                    TableMappingImpl tm = (TableMappingImpl)colMapping.getTableMapping(tableIndex);
                    dependencies.add(tm);
                    Collection dep = tm.getDependencies();
                    if (dep == null)  // SR : Bug 14/11/2000 : was if (tm == null)
                        dep = tm.initializeDependencies();
                    dependencies.addAll(dep);
                }
            }
            if (dependencies.contains(this))
                throw new SAXException("Circular integrity constraints between tables are not supported");
        }
        return dependencies;
    }
    
    /**
     * Used by the mapping loader to attach the ColumnMapping objects it is
     * building to this TableMapping.
     */
    int register(ColumnMapping mapping) throws SAXException
    {
        // check the column hasn't got already a mapping
        if (!userTablesOnly) // only for the repository (not the mapper)
        {
        	Iterator it = getColumnMappingIterator();
        	ColumnMapping current = null;
        	while (it.hasNext())
        	{
        		current = (ColumnMappingImpl)it.next();
                if (current.getColumnName().equalsIgnoreCase(mapping.getColumnName()))
                    throw new SAXException("A column mapping has already been defined on column '"
                    + current.getColumnName() + "' for " + current.getSchemaComponent() + ".");
            }
        }
        
        int index;
        
        // table DDL generation
        if (tableMetaData == null)	// used as a flag for detecting the table DDL generation
        {
        	// resize array if necessary
        	if (getColumnMappingCount() == columnMappings.length)
        	{
        		ColumnMapping[] newArray = new ColumnMapping[columnMappings.length * 2];
        		System.arraycopy(columnMappings, 0, newArray, 0, columnMappings.length);
        		columnMappings = newArray;
        	}
        	index = getColumnMappingCount();
        }
        else
        {
	        // Desactivate batch for LOB types (batcher should not use batch API when size is 1)
	        ColumnMetaData cMeta = mapping.getMetaData();
	        if (cMeta.isLongType())
	            batchSize = 1;
        	index = cMeta.getOrdinalPosition() - 1;
        }
    	columnMappings[index] = mapping;
        return index;	// because JDBC parameters starts with 1
    }
    
	int incrementColumnMappingCount()
	{
		return columnMappingCount++;
	}

	/**
	 * Increment the counter for the columns containing key generators.
	 * May be replaced by PK if any.
	 * Called at initialization of this object or columns
	 * that are registered to it.
	 */
	int incrementKeyColumnCount()
	{
		return keyColumnCount++;
	}

    /**
     * Increment the counter for the columns that constitute the where
     * clause for selection. Called at initialization of columns that
     * are registered to it.
     */
    int incrementSelectColumnCount()
    {
        return selectColumnCount++;
    }
    
    /**
     * Increment the counter for the columns that constitute the select
     * clause for selection. Called at initialization of columns that
     * are registered to it.
     */
    int incrementFetchColumnCount()
    {
        return fetchColumnCount++;
    }
    
    /**
     * Increment the counter for the columns that are likely to be updated
     * in a row. Should match with the complement of select columns.
     */
    int incrementUpdateColumnCount()
    {
        return updateColumnCount++;
    }
    
    int incrementKeyGenColumnCount()
    {
        return keyGenColumnCount++;
    }
    
}
