/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.mapper.mapping;

import java.sql.DatabaseMetaData;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.*;

import org.xquark.jdbc.typing.ColumnMetaData;
import org.xquark.jdbc.typing.ColumnMetaDataImpl;
import org.xquark.mapper.dbms.AbstractConnection;

/**
 * Table properties gained from the JDBC API.
 *
 * Owns a set of metadata objects for table's columns.
 *
 */
public class TableMetaDataImpl implements TableMetaData
{
	private static final String RCSRevision = "$Revision: 1.4 $";
	private static final String RCSName = "$Name: XQ1_1_1 $";

	/* Table name */
	private String name;
	/* List of colum metadata */
	private Map columnsMetaData;
	/* Map {index name, {column metadatas}} (In fact unique indexes, JDBC does not provide constraints) */
	private Map uniqueConstraints = new HashMap();
	
	private boolean statementsAreShareable = true;
	private boolean shared = false;

	/**
	 * Constructor.
	 * This constructor uses native information for column information
	 *
	 * Performs metadata construction, included for table's columns.
	 */
	public TableMetaDataImpl(String name, AbstractConnection connection)
		throws SQLException
	{
		DatabaseMetaData metadata = connection.getConnection().getMetaData();
		this.name = name;
		String catalog = connection.getConnection().getCatalog();
		String schemaName = connection.getSchemaName();
		ResultSet rs = null;
		try
		{
			// Use database specific access for retrieving column info (Oracle
			// JDBC implementation is incorrect
			columnsMetaData = connection.getTableMetadata(name);

			// PK
			rs = metadata.getPrimaryKeys(catalog, schemaName, name);
			while (rs.next())
			{
				String cname = rs.getString(4);
				short seq = rs.getShort(5);
				ColumnMetaDataImpl cmeta =
					(ColumnMetaDataImpl) columnsMetaData.get(cname);
				cmeta.setPrimaryKeySeq(seq);
			}
			rs.close();

			// FK
			rs = metadata.getImportedKeys(catalog, schemaName, name);
			while (rs.next())
			{
				String pkTableName = rs.getString(3);
				String pkColumnName = rs.getString(4);
				String cname = rs.getString(8);
				ColumnMetaDataImpl cmeta =
					(ColumnMetaDataImpl) columnsMetaData.get(cname);
				cmeta.setRefTable(pkTableName);
				cmeta.setRefColumn(pkColumnName);
			}
			rs.close();

			// Unique constraints
			// Warning : JDBC does not provide access to UNIQUE constraints but
			// only to Index which on Oracle match since Oracle create an index
			// for these constraints.
			try {
				rs = metadata.getIndexInfo(catalog, schemaName, name, true, true);
				while (rs.next())
				{
					String indexName = rs.getString(6);
					String indexColumnName = rs.getString(9);
	
					ArrayList columns =
						(ArrayList) uniqueConstraints.get(indexName);
					if (columns == null)
					{
						columns = new ArrayList();
						uniqueConstraints.put(indexName, columns);
					}
					// put the column name in the set
					columns.add(columnsMetaData.get(indexColumnName));
				}
	
				rs.close();
			}
			catch (SQLException e) {	// TODO: check if it is a JDBC oracle bug to automatically convert
				// for this method the table name to uppercase. Either change driver or perform a native call.
				if (connection.getUniversalError(e) != AbstractConnection.OBJECT_DOES_NOT_EXISTS)
					throw e;
			}
		}
		finally
		{
			if (rs != null)
				rs.close();
		}
	}

	//////////////////////////////////////////////////////////////////////////
	//                        TableMetaData IMPLEMENTATION
	//////////////////////////////////////////////////////////////////////////
	public boolean areStatementsShareable()
	{
		return shared && statementsAreShareable;
	}

	public void setStatementsNotShareable()
	{
		statementsAreShareable = false;
	}

	public String getTableName()
	{
		return name;
	}

	public ColumnMetaData getColumnMetaData(String cname)
	{
		ColumnMetaData ret = (ColumnMetaData) columnsMetaData.get(cname);
//		if (ret == null)
//			// SR 18/12/2000 Oracle is case-sensitive for metadata !
//			ret = (ColumnMetaData) columnsMetaData.get(cname.toUpperCase());
//		if (ret == null)
//			// SR 18/12/2000 Oracle is case-sensitive for metadata !
//			ret = (ColumnMetaData) columnsMetaData.get(cname.toLowerCase());
		return ret;
	}

	public Collection getColumnsMetaData()
	{
		return columnsMetaData.values();
	}

	public int getColumnCount()
	{
		return columnsMetaData.size();
	}

	public Collection getUniqueConstraints()
	{
		return uniqueConstraints.values();
	}

	public boolean isShared()
	{
		return shared;
	}

	public void setShared()
	{
		shared = true;
	}

}
