/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

/*
 * CollectionConfig.java
 *
 * Created on 20 juillet 2000, 14:24
 */
 
package org.xquark.mapper.metadata;

import java.io.StringWriter;

import org.xquark.mapper.RepositoryConnection;
import org.xquark.mapper.util.Tabulator;
import org.xquark.xml.xdbc.*;

/** Data structure that is used to set features when creating an XMLCollection.
 * <p>This object can be recycled because every XMLCollection keeps a copy
 * of its configuration.</p>
 */

public class CollectionInfo implements Cloneable, Configurable
{
    private static final String RCSRevision = "$Revision: 1.2 $";
    private static final String RCSName     = "$Name: XQ1_1_1 $";
	
    /////////////////////////
    //      PROPERTIES
    /////////////////////////
    public static final String TYPE_PROPERTY
        = "http://www.xquark.org/repository/collection/properties/type";
    public static final String BUCKET_ID_SIZE_PROPERTY
        = "http://www.xquark.org/repository/collection/properties/bucketIDSizeInBits";
    
    /* "public" properties */
    private static final String[] SUPPORTED_PROPERTIES
        = new String[] {
            RepositoryConnection.TEXT_LENGTH_PROPERTY,
            RepositoryConnection.EXTRA_DATA_LENGTH_PROPERTY,
            RepositoryConnection.DESC_PROPERTY,
            RepositoryConnection.DOC_OID_SIZE_PROPERTY,
            RepositoryConnection.DOC_OID_PREALLOCATION_SIZE,
            RepositoryConnection.MAPPING_ID_PROPERTY};
        
    // TO DO : define which properties are read-only (all for the moment)
    
    /////////////////////////
    //      FEATURES
    /////////////////////////

    /* "public" features */
    private static final String[] SUPPORTED_FEATURES
        = new String[] {RepositoryConnection.USE_SCHEMA_PREFIXES_FEATURE};
    
    /** System repository using Generic storage. These repositories are
     * processed in special way by the repository. For instance, they
     * are deleted at last. They are used to store, for instance,
     * schemas or mappings.
     *<p><B>This constants has been set here to be hidden from API users.</B></p>
     */
    public static final byte SYSTEM = 0x00;
   
    /** The default collection type
     */
    public static final byte USER = 0x01;
    
    /////////////////////////
    //      DATA
    /////////////////////////

    /** Storage method type. By extension, type of the collection itself. At the present
     * time, only one kind of collection is available : USER.
     */
    protected byte collectionType = USER;

    /** A user-free string.
     */
    protected String description;
    
    protected byte docOIDSize =  RepositoryConstants.DID_SIZE_DEFAULT_VALUE;
    protected short docOIDPrealloc = 1;
//    protected byte bucketIDSize = 24;
    
    /** Maximum length of text values. Default is 255 bytes */
    protected int maxTextLength = RepositoryConstants.DEFAULT_DATA_LENGTH;

    protected int maxExtraDataLength = RepositoryConstants.DEFAULT_DATA_LENGTH;

    /** The public ID for mapping file used for identification in the system collection.
     * A null value means that the default mapping method must be used.
     */
    protected String mappingURI = "";
    
    /** Option used to discard prefix mapping information storage for documents.
     * The prefix mappings of the schemas associated are used at reconstruction
     * time. If false, document info is discarded and schema one is used.
     */
    protected boolean useSchemaPrefixes = false;
    
    private boolean writeProtected = true;
    
    /** Creates new CollectionConfig with default features.
     * <p>Default is 512 for text length.</p>
     */
    public CollectionInfo() {}
    
    public CollectionInfo(byte type, int textLength)
    {
        collectionType = type;
        maxTextLength = maxExtraDataLength = textLength;
        
    }

    public CollectionInfo(boolean writeProtected)
    {
        setWriteProtected(writeProtected);
    }

    /** "Clone" constructor.
     * @param config base CollectionConfig object.
     */
    public CollectionInfo(Configurable config)
    {
        try {
            collectionType = Byte.parseByte((String)config.getProperty(TYPE_PROPERTY));
            maxTextLength = Integer.parseInt((String)config.getProperty(RepositoryConnection.TEXT_LENGTH_PROPERTY));
            maxExtraDataLength = Integer.parseInt((String)config.getProperty(RepositoryConnection.EXTRA_DATA_LENGTH_PROPERTY));
            mappingURI = (String)config.getProperty(RepositoryConnection.MAPPING_ID_PROPERTY);
            description = (String)config.getProperty(RepositoryConnection.DESC_PROPERTY);
            setDocOIDSize(Byte.parseByte((String)config.getProperty(RepositoryConnection.DOC_OID_SIZE_PROPERTY)));
            docOIDPrealloc = Short.parseShort((String)config.getProperty(RepositoryConnection.DOC_OID_PREALLOCATION_SIZE));
//            bucketIDSize = Byte.parseByte((String)config.getProperty(BUCKET_ID_SIZE_PROPERTY));
            useSchemaPrefixes = config.getFeature(RepositoryConnection.USE_SCHEMA_PREFIXES_FEATURE);
        }
        catch (XMLDBCException e) {
            // no op : we know what is supported !!!
        }
    }
    
    public void setWriteProtected(boolean mode)
    {
        writeProtected = mode;
    }
    
    /** Builds a String in order to displays the information of this configuration.
     * @return the string representing the repository features.
     */
    public String toString()
    {
        StringWriter out = new StringWriter();
        Tabulator tabulator = new Tabulator
                                    (
                                    out,
                                    79,
                                    new int[] {0, 40}
                                    );
                                    
        if (description != null && description.length() > 0)
        {
            tabulator.addItem("Description:");
            tabulator.addItem(description);
        }
        if (mappingURI.length() != 0)
        {
            tabulator.addItem("Mapping file:");
            tabulator.addItem(mappingURI);
            tabulator.addItem("Use Schema prefixes :");
            tabulator.addItem(String.valueOf(useSchemaPrefixes));
        }
        tabulator.addItem("Text length:");
        tabulator.addItem(String.valueOf(maxTextLength));
        tabulator.addItem("Extra data length:");
        tabulator.addItem(String.valueOf(maxExtraDataLength));
        tabulator.addItem("Document ID size in bits:");
        tabulator.addItem(String.valueOf(docOIDSize));
        tabulator.addItem("Document OID preallocation size:");
        tabulator.addItem(String.valueOf(docOIDSize));
//        tabulator.addItem("\nBucket ID size in bits:\t");
//        tabulator.addItem(bucketIDSize);
        
        return out.toString();
    }
    
    /** Builds a copy of this object.
     * @return a CollectionConfig object.
     */
    public Object clone()
    {
        Object copy = null;
        try {
            copy = super.clone();
        }
        catch(CloneNotSupportedException e) {
        }
        return copy;
    }
    
    public void setProperty(String propertyId, Object value) throws XMLDBCNotRecognizedException, XMLDBCNotSupportedException
    {
        // TO REFINE FOR EVERY PARAMETER
        if (writeProtected)
            throw new XMLDBCNotSupportedException("This object is read-only.");
        
        if (propertyId.equals(TYPE_PROPERTY))
        {
            if (value instanceof String)
                collectionType = Byte.parseByte((String)value);
            else if (value instanceof Number)
                collectionType = ((Number)value).byteValue();
            else
                throw new XMLDBCNotSupportedException("Property value type is not supported.");
        }
        else if (propertyId.equals(RepositoryConnection.TEXT_LENGTH_PROPERTY))
        {
            if (value instanceof String)
                maxTextLength = Integer.parseInt((String)value);
            else if (value instanceof Number)
                maxTextLength = ((Number)value).intValue();
            else
                throw new XMLDBCNotSupportedException("Property value type is not supported.");
        }
        else if (propertyId.equals(RepositoryConnection.EXTRA_DATA_LENGTH_PROPERTY))
        {
            if (value instanceof String)
                maxExtraDataLength = Integer.parseInt((String)value);
            else if (value instanceof Number)
                maxExtraDataLength = ((Number)value).intValue();
            else
                throw new XMLDBCNotSupportedException("Property value type is not supported.");
        }
        else if (propertyId.equals(RepositoryConnection.DOC_OID_SIZE_PROPERTY))
        {
            if (value instanceof String)
                setDocOIDSize(Byte.parseByte((String)value));
            else if (value instanceof Number)
                setDocOIDSize(((Number)value).byteValue());
            else
                throw new XMLDBCNotSupportedException("Property value type is not supported.");
        }
        else if (propertyId.equals(RepositoryConnection.DOC_OID_PREALLOCATION_SIZE))
        {
            if (value instanceof String)
                docOIDPrealloc = Short.parseShort((String)value);
            else if (value instanceof Number)
                docOIDPrealloc = ((Number)value).shortValue();
            else
                throw new XMLDBCNotSupportedException("Property value type is not supported.");
        }
//        else if (propertyId.equals(BUCKET_ID_SIZE_PROPERTY))
//        {
//            if (value instanceof String)
//                bucketIDSize = Byte.parseByte((String)value);
//            else if (value instanceof Number)
//                bucketIDSize = ((Number)value).byteValue();
//            else
//                throw new XMLDBCNotSupportedException("Property value type is not supported.");
//        }
        else if (propertyId.equals(RepositoryConnection.MAPPING_ID_PROPERTY))
        {
            if (value instanceof String)
                mappingURI = (String)value;
            else
                throw new XMLDBCNotSupportedException("Property value type is not supported.");
        }
         else if (propertyId.equals(RepositoryConnection.DESC_PROPERTY))
        {
            if (value instanceof String)
                description = (String)value;
            else
                throw new XMLDBCNotSupportedException("Property value type is not supported.");
        }
        else
            throw new XMLDBCNotRecognizedException("Unkown property " + propertyId);
    }
    
    public Object getProperty(String propertyId) throws XMLDBCNotRecognizedException
    {
        if (propertyId.equals(TYPE_PROPERTY))
            return Byte.toString(collectionType);
        else if (propertyId.equals(RepositoryConnection.TEXT_LENGTH_PROPERTY))
            return Integer.toString(maxTextLength);
        else if (propertyId.equals(RepositoryConnection.EXTRA_DATA_LENGTH_PROPERTY))
            return Integer.toString(maxExtraDataLength);
        else if (propertyId.equals(RepositoryConnection.DOC_OID_SIZE_PROPERTY))
            return Byte.toString(docOIDSize);
        else if (propertyId.equals(RepositoryConnection.DOC_OID_PREALLOCATION_SIZE))
            return Integer.toString(docOIDPrealloc);
//        else if (propertyId.equals(BUCKET_ID_SIZE_PROPERTY))
//            return Byte.toString(bucketIDSize);
        else if (propertyId.equals(RepositoryConnection.MAPPING_ID_PROPERTY))
            return mappingURI;
        else if (propertyId.equals(RepositoryConnection.DESC_PROPERTY))
            return description;
        else
            throw new XMLDBCNotRecognizedException("Unkown property " + propertyId);
    }
    
    public boolean getFeature(String name) throws XMLDBCNotRecognizedException
    {
        if (name.equals(RepositoryConnection.USE_SCHEMA_PREFIXES_FEATURE))
            return useSchemaPrefixes;
        else
            throw new XMLDBCNotRecognizedException("Unkown feature " + name);
    }
    
    public void setFeature(String name, boolean state) throws XMLDBCNotRecognizedException, XMLDBCNotSupportedException
    {
        // TO REFINE FOR EVERY PARAMETER
        if (writeProtected)
            throw new XMLDBCNotSupportedException("This object is read-only.");
        
        if (name.equals(RepositoryConnection.USE_SCHEMA_PREFIXES_FEATURE))
            useSchemaPrefixes = state;
        else
            throw new XMLDBCNotRecognizedException("Unkown feature " + name);
    }
    
    public String[] getFeatureList()
    {
        return SUPPORTED_FEATURES;
    }
    
    public String[] getPropertyList()
    {
        return SUPPORTED_PROPERTIES;
    }
    
    public String getMappingID()
    {
        return mappingURI;
    }
    
    public String getDescription()
    {
        return description;
    }
    
    public int getMaxTextLength()
    {
        return maxTextLength;
    }
    
    public int getMaxExtraDataLength()
    {
        return maxExtraDataLength;
    }
    
    public byte getType()
    {
        return collectionType;
    }
    
    public byte getDocOIDSize()
    {
        return docOIDSize;
    }
    
    public short getDocOIDPreallocationSize()
    {
        return docOIDPrealloc;
    }
    
//    public byte getBucketIDSize()
//    {
//        return bucketIDSize;
//    }
    
    public void setMappingID(String uri)
    {
        mappingURI = uri;
    }
    
    public void setDescription(String description)
    {
        this.description = description;
    }
    
    public void setMaxTextLength(int maxLength)
    {
        maxTextLength = maxLength;
    }
    
    public void setMaxExtraDataLength(int maxLength)
    {
        maxExtraDataLength = maxLength;
    }
    
    public void setType(byte type)
    {
        collectionType = type;
    }
    
    public void setDocOIDSize(byte docSize)
    {
        if (docSize < RepositoryConstants.DID_SIZE_MIN_VALUE)
            docOIDSize = RepositoryConstants.DID_SIZE_MIN_VALUE;
        else if (docSize > RepositoryConstants.DID_SIZE_MAX_VALUE)
            docOIDSize = RepositoryConstants.DID_SIZE_MAX_VALUE;
        else
            docOIDSize = docSize;
    }
    
    public void setDocOIDPreallocationSize(short size)
    {
        docOIDPrealloc = size;
    }
    
//    public void setBucketIDSize(byte docSize)
//    {
//        bucketIDSize = docSize;
//    }
}
