/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

/*
 * TableMappingStatements.java
 *
 * Created on 3 septembre 2002, 09:34
 */

package org.xquark.mapper.metadata;

import org.xquark.mapper.dbms.*;
import org.xquark.mapper.mapping.MappingInfo;
import org.xquark.mapper.mapping.TableMapping;

/**
 *
 */
public class CollectionMappingInfo implements MappingInfo  
{
    private static final String RCSRevision = "$Revision: 1.2 $";
    private static final String RCSName     = "$Name: XQ1_1_1 $";
    
    private final static String DATA_INDEX_SUFFIX = "_D";
    
    protected String deleteStatement;
    protected String insertStatement;
    protected String[] oidIndexCreateStatement;
    protected String[] oidIndexDropStatement;
    protected String oidTableCreateStatement;
    protected String oidTableDropStatement;
    private String oidDeleteStatement;
    private String oidInsertStatement;
    private String selectValuesStatement;
    private String getRowsClause;
    private String oidTableName;
    private TableMapping table;
    
    private static TableSpec OIDTableSpecs;
    
    static
    {
        OIDTableSpecs = (TableSpec)TableSpecLoader.getInstance().getTableSpec(
                            TableSpec.CAT_USER_TABLE, TableSpec.TYPE_USER_DATA);
    }
    
    /** To use at collection creation or deletion (static).
     */
    public CollectionMappingInfo(TableMapping table, short cid, short tid, AbstractConnection conn)
    {
        this.table = table;
        initializeCreationStatements(cid, tid, conn);
    }
    
	/** To use during collection lifetime.
	 */
	public CollectionMappingInfo(TableMapping table)
	{
		this.table = table;
	}

	/** Build SQL strings for use in JDBC PreparedStatements/.
	 */
	public void initializeStatements(CollectionMetadata collection, short tid, short pid)
	{
		initializeStatements(
			collection.getTableInfo(TableSpec.TYPE_DOCS),
			collection.getUOIDManager(),
			collection.getCollectionID(),
			tid,
			pid
			);
	}
	
    public String[] getOIDIndexDropStatements()
    {
        return oidIndexDropStatement;
    }
    
    public String[] getOIDIndexCreateStatements()
    {
        return oidIndexCreateStatement;
    }
    public String getOIDTableCreateStatement()
    {
        return oidTableCreateStatement;
    }
    public String getOIDDeleteStatement()
    {
        return oidDeleteStatement;
    }
    public String getOIDTableDropStatement()
    {
        return oidTableDropStatement;
    }
    public String getDocumentRowsStatement()
    {
        return getRowsClause;
    }
    public String getDeleteStatement()
    {
        return deleteStatement;
    }
    public String getSelectValuesStatement()
    {
        return selectValuesStatement;
    }
    
    ////////////////////////////////////////////////////////////////////////////
    //                      MappingInfo IMPLEMENTATION
    ////////////////////////////////////////////////////////////////////////////
    
    public String getInsertStatement()
    {
        return insertStatement;
    }
    
    public String getOIDInsertStatement()
    {
        return oidInsertStatement;
    }
    
    public String getOIDTableName()
    {
       return oidTableName;
     }
    
    public String getSelectStatement()
    {
        return table.getSelectStatement();
    }
    
    public String getUpdateStatement()
    {
        return table.getUpdateStatement();
    }
    
    public TableMapping getTableMapping()
    {
        return table;
    }
    
	/**
	 * Says if all statements can be shared among the table mappings defined on
	 * the same user table.
	 * @return boolean
	 */
	public boolean isShareable()
	{
		return table.isShareable();
	}
	
	/**
	 * True if filtering on the path is needed for document reconstruction
	 * because multiple table mapping use the same table but there is as many
	 * table mapping as paths, i.e., isTableShared() && ! isShared()
	 * @return boolean
	 */
	public boolean isPathFilteringNeeded()
	{
		return table.isTableShared() && !table.isShared();
	}

    ////////////////////////////////////////////////////////////////////////////
    //                              PRIVATE
    ////////////////////////////////////////////////////////////////////////////
    
    /** Build SQL strings for use in JDBC PreparedStatements/.
     */
    private void initializeStatements(TableInfo documentTable ,UOIDManager manager, short cid, short tid, short pid)
    {
    	String uoidColumnName = table.getUOIDColumnName();
    	String pidColumnName = table.getPathIDColumnName();
        String docTableName = documentTable.getName();
        String userTableName = table.getTableName();  // collection name is fake here : for default mapping
        
        // Statements construction
        insertStatement =  table.getInsertStatement();
        
        StringBuffer sql = new StringBuffer();
        if (table.isClustered())  // no OID table
        {
            // TO DO : use systematically prefixes to avoid column name collision !!!
            sql.setLength(0);
            sql.append("SELECT ");
            sql.append(table.getNamedColumnList());
            sql.append(" FROM ");
            sql.append(docTableName);
            sql.append(" d,");
            sql.append(userTableName);
            sql.append(" w WHERE d.NAME=? and w.");
            sql.append(uoidColumnName);
            sql.append(">=(d.UDID*");
            sql.append(manager.getDocMultiplier());
            // assume first local node OID is 0
            sql.append(") AND w.");
            sql.append(uoidColumnName);
            sql.append("<=(d.UDID*");
            sql.append(manager.getDocMultiplier());
            sql.append('+');
            sql.append(manager.getDocMaxConstant());
        	sql.append(")");
        	if (isPathFilteringNeeded())
        	{
        		sql.append(" AND w.");
        		sql.append(pidColumnName);
        		sql.append("=");
        		sql.append(pid);
        	}
        	sql.append(" ORDER BY w.");
            sql.append(uoidColumnName);
            selectValuesStatement = sql.toString();
            
            sql.setLength(0);
            sql.append("DELETE FROM ");
            sql.append(userTableName);
            sql.append(" WHERE ");
            sql.append(uoidColumnName);
            sql.append(">=? AND ");
            sql.append(uoidColumnName);
            sql.append("<=?");
            deleteStatement = sql.toString(); // check because of portable two-times deletion
            
            getRowsClause = null;
            
            /* OID table statements */
            oidTableCreateStatement = null;
            oidTableDropStatement = null;
            oidIndexCreateStatement = null;
            oidIndexDropStatement = null;
            oidInsertStatement = null;
            oidDeleteStatement = null;
        }
        else
        {
            oidTableName = OIDTableSpecs.getTableName(cid, tid);
            
            sql.setLength(0);
            sql.append("SELECT v.UOID,v.PATH, ");
            sql.append(table.getNamedColumnList());
            sql.append(" FROM ");
            sql.append(docTableName);
            sql.append(" d,");
            sql.append(oidTableName);
            sql.append(" v, ");
            sql.append(userTableName);
            sql.append(" w WHERE d.NAME=? and v.UOID>=(d.UDID*");
            sql.append(manager.getDocMultiplier());
            // assume first local node OID is 0
            sql.append(") AND v.UOID<=(d.UDID*");
            sql.append(manager.getDocMultiplier());
            sql.append('+');
            sql.append(manager.getDocMaxConstant());
        	sql.append(")");
        	if (isPathFilteringNeeded())
        	{
        		sql.append(" AND v.PATH=");
        		sql.append(pid);
        	}
        	sql.append(" AND ");
            sql.append(table.getJoinCondition());
            sql.append(" ORDER BY v.UOID");
            selectValuesStatement = sql.toString();
            
            deleteStatement = null;
            
            sql.setLength(0);
            sql.append("FROM ");
            sql.append(oidTableName);
            sql.append(" v, ");
            sql.append(userTableName);
            sql.append(" w WHERE v.UOID>=? AND v.UOID<=? AND ");
            sql.append(table.getJoinCondition());
            getRowsClause = sql.toString();
            
            sql.setLength(0);
            sql.append("INSERT INTO ");
            sql.append(oidTableName);
            sql.append("(UOID,PATH, ");
            sql.append(table.getJoinColumnList());
            sql.append(") VALUES(?,?,");
            sql.append(table.getJoinWildcards());
            sql.append(")");
            oidInsertStatement = sql.toString();
            
            sql.setLength(0);
            sql.append("DELETE FROM ");
            sql.append(oidTableName);
            sql.append(" WHERE UOID>=? AND UOID<=?");
            oidDeleteStatement = sql.toString();
        }
    }
    
    /** Build SQL strings for use in JDBC PreparedStatements/.
     */
    private void initializeCreationStatements(short cid, short tid, AbstractConnection conn)
    {
        String uoidColumnName = table.getUOIDColumnName();
        
        // Statements construction
        StringBuffer sql = new StringBuffer();
        
        if (!table.isClustered())  // OID table
        {
            oidTableName = OIDTableSpecs.getTableName(cid, tid);
            
            /* OID table statements */
            ColumnSpec[] columns = OIDTableSpecs.getColumns();
            sql.setLength(0);
            sql.append("CREATE TABLE ");
            sql.append(oidTableName);
            sql.append('(');
            sql.append(columns[0].getCreationString(conn.getTypeMap()));
            sql.append(',');
            sql.append(columns[1].getCreationString(conn.getTypeMap()));
            sql.append(',');
            sql.append(table.getJoinColumnTypes());
            sql.append(')');
            oidTableCreateStatement = sql.toString();
            
            sql.setLength(0);
            sql.append("DROP TABLE ");
            sql.append(oidTableName);
            oidTableDropStatement = sql.toString();
            
            IndexSpec[] indexes = OIDTableSpecs.getIndexes();
            oidIndexCreateStatement = new String[indexes.length + 1];
            oidIndexDropStatement =  new String[indexes.length + 1];
            
            int i = 0;
            for (; i < indexes.length; i++)
            {
                oidIndexCreateStatement[i] = indexes[i].generateCreationStatement(cid, tid, conn);
                oidIndexDropStatement[i] = "DROP INDEX " + indexes[i].getIndexName(cid, tid);
            }
            // add to XML file ?
            sql.setLength(0);
            sql.append("CREATE INDEX ");
            sql.append(oidTableName);
            sql.append(DATA_INDEX_SUFFIX);
            sql.append(" ON ");
            sql.append(oidTableName);
            sql.append('(');
            sql.append(table.getJoinColumnList());
            sql.append(')');
            oidIndexCreateStatement[i] = sql.toString();
            
            sql.setLength(0);
            sql.append("DROP INDEX ");
            sql.append(oidTableName);
            sql.append(DATA_INDEX_SUFFIX);
            oidIndexDropStatement[i] = sql.toString();
        }
    }
}
