/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

/*
 * CollectionMetadata.java
 *
 * Created on 30 mai 2001, 16:38
 */

package org.xquark.mapper.metadata;

import java.sql.*;
import java.util.*;

import org.xml.sax.SAXException;
import org.xquark.mapper.RepositoryException;
import org.xquark.mapper.dbms.*;
import org.xquark.mapper.mapping.*;
import org.xquark.mapper.storage.*;
import org.xquark.schema.SchemaComponent;
import org.xquark.xml.xdbc.XMLDBCException;
import org.xquark.xpath.XTreeRuntimeException;

/**
 *  Metadata object for a Repository Collection. Unique for a given collection
 * in the executing instance.
 *
 */
public class CollectionMetadata implements RepositoryConstants, _Collection
{
    private static final String RCSRevision = "$Revision: 1.5 $";
    private static final String RCSName     = "$Name: XQ1_1_1 $";


    //
    // Data
    //
    protected String name;	// the name of the collection
    protected short cid;
    protected CollectionInfo config;  // the same object must be kept during collection life (used by object like UOIDManager)
    protected Repository rep;

    /** The last allocated OID for the nodes in the repository.
     */
    protected UOIDManager uoidManager;
    private PathSetFactory pathSetFactory;

    private PathSet pathSet;
    private RepositoryMapping mapping;
    /** Mapping returned when no mapping is available : is put in the mapping list at position 0 */
    private DefaultTableMapping defaultMapping;
    private List defaultMappingCollection;
    /** List used notably when need to get table column mappings in a particular order */
    private CollectionMappingInfo[] mappingStatements;
    
    private int defaultMappingIndex;
    
    ////////////////////////////////
    // Prepared string statements
    ////////////////////////////////
    private String selectDocIDStmt;
    private String renameDocStmt;
    private String countDocStmt;
    private String containsDocStmt;
    private String docListStmt;
    private String deleteDocIDStmt;
    
    /** A map containing TableInfo for repository tables.
     */
    private Map collectionTables;
    
	/**
	 *	Index for accessing path node from a table mapping
	 */
	private short[] tableIndex;

    /** Creates new CollectionMetadata */
    public CollectionMetadata(
                                String name,
                                short cid,
                                CollectionInfo config,
                                Repository rep, 
                                MappingFactory mappingFactory
                                ) 
    throws SQLException, XMLDBCException
    {
        // order is fundamental
        this.name = name;
        this.cid = cid;
        this.rep = rep;
    	this.config = new CollectionInfo(config);
    	this.mapping = (RepositoryMapping)mappingFactory.getTree();
        collectionTables = loadTableInfo(cid, rep.getInfo().getColIDSize(), config, getMetadataConnection());
        initStatements(); // after loading table info...
        uoidManager = new UOIDManager(this);
        
    	initMapping(mapping, getMetadataConnection()); // after the preceeding initializations
        pathSetFactory = new PathSetFactory(this, mappingFactory);
        
        AbstractConnection metaConn = getMetadataConnection();
        synchronized(metaConn)
        {
            metaConn.start();  // because rows may have been inserted when mapping is duplicated
            try {
                this.pathSet = (PathSet)pathSetFactory.createTree(); // factory is not kept since it is "thread local"
            }
            catch (XTreeRuntimeException e) {  // because interface does not define exceptions
                metaConn.rollback();
                if (e.getException() instanceof RepositoryException)
                    throw (RepositoryException)e.getException();
                else if (e.getException() instanceof SQLException)
                    throw (SQLException)e.getException();
                else
                    throw e;
            }
            metaConn.commit();  // because rows may have been inserted when mapping is duplicated
        }
    	initMappingStatements(); // after the preceeding initializations
    }
    
    private void initMapping(RepositoryMapping mapping, AbstractConnection metadataConn)
    throws SQLException, XMLDBCException
    {
        try
        {
//            int textLength;
//            try
//            {
//                textLength = Integer.parseInt((String)config.getProperty(RepositoryConnection.TEXT_LENGTH_PROPERTY));
//            }
//            catch (XMLDBCException e)
//            {
//                // no op : internal use
//                e.printStackTrace();
//                throw new RuntimeException("Internal error: unknown feature or property.");
//            }
//            
            defaultMapping = new DefaultTableMapping(this, metadataConn);
            defaultMappingIndex = mapping.getTableCount();
            defaultMapping.setIndex(defaultMappingIndex);
		}
		catch (SAXException e)
		{
			throw new RuntimeException(
				"Error while creating default table mapping:" + e.getMessage());
		}

    	int mappingCount = mapping.getTableCount();
    	mappingStatements = new CollectionMappingInfo[mappingCount + 1];
    	tableIndex = new short[mappingCount + 1];
        for (int i = 0; i < mappingCount; i++)
            mappingStatements[i] = new CollectionMappingInfo(mapping.getTableMapping(i));

        // add after user table mappings to respect indexes
        mappingStatements[defaultMappingIndex] = new CollectionMappingInfo(defaultMapping);
        defaultMappingCollection = Collections.singletonList(defaultMapping);
    }
    
	private void initMappingStatements() throws XMLDBCException
	{
		TableMapping wTable;
		for (int i = 0; i < mappingStatements.length - 1; i++)
		{
			wTable = mappingStatements[i].getTableMapping();
			mappingStatements[i].initializeStatements(
				this,
				rep.getTableManager().getTableID(wTable.getTableName()),
				tableIndex[i]);
		}
		mappingStatements[defaultMappingIndex].initializeStatements(
			this,
			(short)-1,
			(short)0);
	}
		
    private void initStatements()
    {
        // Documents table statements
        TableInfo docsTableTable = getTableInfo(TableSpec.TYPE_DOCS);
        ColumnInfo[] columns = docsTableTable.getColumns();
        
        StringBuffer sql = new StringBuffer();
        sql.append("SELECT ");
        sql.append(columns[1].getName());
        sql.append(" FROM ");
        sql.append(docsTableTable.getName());
        sql.append(" WHERE ");
        sql.append(columns[0].getName());
        sql.append("='");
        selectDocIDStmt = sql.toString();
        
        sql.setLength(0);
        sql.append("UPDATE ");
        sql.append(docsTableTable.getName());
        sql.append(" SET ");
        sql.append(columns[0].getName());
        sql.append("=? WHERE ");
        sql.append(columns[0].getName());
        sql.append("=?");
        renameDocStmt = sql.toString();

        sql.setLength(0);
        sql.append("SELECT COUNT(*) FROM ");
        sql.append(docsTableTable.getName());
        countDocStmt = sql.toString();
        
        sql.setLength(0);
        sql.append(countDocStmt);
        sql.append(" WHERE ");
        sql.append(columns[0].getName());
        sql.append("='");
        containsDocStmt = sql.toString();
        
        sql.setLength(0);
        sql.append("SELECT ");
        sql.append(columns[0].getName());
        sql.append(" FROM ");
        sql.append(docsTableTable.getName());
        docListStmt = sql.toString();
        
        sql.setLength(0);
        sql.append("DELETE FROM ");
        sql.append(docsTableTable.getName());
        sql.append(" WHERE ");
        sql.append(columns[0].getName());
        sql.append("=?");
        deleteDocIDStmt = sql.toString();
    }
    
	public void registerMapping(TableMapping tm, PathNode node)
	{
		tableIndex[tm.getTableIndex()] = node.getPathID();
	}

	public short getMappingPID(int tableMappingIndex)
	{
		return tableIndex[tableMappingIndex];
	}

    ////////////////////////////////////////////////////////////////////
    //                          ACCESSORS
    ////////////////////////////////////////////////////////////////////
    public AbstractConnection getMetadataConnection()
    {
        return rep.getMetadataConnection();
    }
    
    PathSetFactory getPathSetFactory()
    {
        return pathSetFactory;
    }
    
    public Map getTableInfoMap()
    {
        return collectionTables;
    }
    
    public String getSelectDocIDStmt()
    {
        return selectDocIDStmt;
    }
    
    public String getRenameDocStmt()
    {
        return renameDocStmt;
    }
    
    public String getCountDocStmt()
    {
        return countDocStmt;
    }
    
    public String getContainsDocStmt()
    {
        return containsDocStmt;
    }
    
    public String getDocListStmt()
    {
        return docListStmt;
    }
    
    public String getDeleteDocIDStmt()
    {
        return deleteDocIDStmt;
    }
    
    public ColumnMapping createDefaultMappingView(SchemaComponent comp)
    throws RepositoryException
    {
        return defaultMapping.createDefaultMappingView(
                            comp, getMetadataConnection().useStringDelimitor());
    }
    
    public int getDefaultMappingIndex() { return defaultMappingIndex;}
    
    /** Iterators are responsible for setting the correct SchemaComponent
     * when the default mapping has to be used out of the mapping area.
     * This special table mapping has index 0
     */
    public List getDefaultMappingList()
    {
        return defaultMappingCollection;
    }
    
    public DefaultTableMapping getDefaultMapping()
    {
        return defaultMapping;
    }
    
    /** Set the name of the collection.
     * <B>Must be used when changing the name of the repository, but does
     * not perform the renaming of database configuration, including table
     * names, which must be done elsewhere.</B>
     * @param newName the new value of the name
     */
    public void setName(String newName) { this.name = newName;}
    
    /** Returns the string name of this repository.
     * @return the string name of the repository used in particular
     * for naming storing tables.
     */
    public String getCollectionName() { return name;}
    
    public short getCollectionID() { return cid;}

    /** Returns the Repository object used for creation of this repository.
     * @return a Repository object.
     */
    public Repository getRepository() { return rep;}
    
    /** Returns the configuration used at the creation time of this repository.
     * @return a Configurable object.
     */
    public CollectionInfo getInfo() { return config;}

    public UOIDManager getUOIDManager() { return uoidManager;}

    /** Free all resources. Repository is warned that repository is closed.
     * <B>Any subsequent call to a function of the object will raise an
     * exception.</B>
     */
    public void close() throws RepositoryException
    {
        uoidManager.close();
        uoidManager = null;
        pathSet.close();
        pathSet = null;
        name = null;
        rep = null;
    }
    
    /** Returns the pathSet information of this repository.
     * <I>Not part of the repository interface.</I>
     * @return a XSSchema object
     * @see org.xquark.mapper.metadata.XSSchema
     */
    public PathSet getPathSet() { return pathSet;}
    
    public RepositoryMapping getMapping()
    {
        return mapping;
    }
    
     /**
     * Get the OID table names for internal use.
     *
     * @param tableMappingIndex index of table pathSet.
     * @return the OID table name.
     */
    public String getOIDTableName(int tableMappingIndex) throws XMLDBCException
    {
        return mappingStatements[tableMappingIndex].getOIDTableName();
    }
    
    private static Map loadTableInfo(short cid, short cidSize, CollectionInfo info, AbstractConnection conn)
    throws RepositoryException
    {
        HashMap map = new HashMap();

        /* loading tableInfo */
        List tables = TableSpecLoader.getInstance().getTableSpecs(TableSpec.CAT_COLLECTION);
        Iterator it = tables.iterator();
        TableSpec wSpec = null;
        
        long[] params = new long[TableInfo.PARAM_NUMBER];
        params[TableInfo.PARAM_DATA_SIZE] = info.getMaxTextLength();
        params[TableInfo.PARAM_EXTRA_SIZE] = info.getMaxExtraDataLength();
    	params[TableInfo.PARAM_DID_SIZE] = info.getDocOIDSize(); // DID
    	params[TableInfo.PARAM_UDID_SIZE] = cidSize + info.getDocOIDSize(); // UDID
    	params[TableInfo.PARAM_BID_SIZE] = params[TableInfo.PARAM_UDID_SIZE]; // bucketOID = DocOID
    
        while (it.hasNext())
        {
            wSpec = (TableSpec)it.next();
            map.put(new Byte(wSpec.getType()), new TableInfo(wSpec, cid, params, conn));
        }
        return map;
    }

    public TableInfo getTableInfo(byte type)
    {
        return (TableInfo)collectionTables.get(new Byte(type));
    }
    
    ////////////////////////////////////////////////////////////////////
    //                          PUBLIC
    ////////////////////////////////////////////////////////////////////
    public synchronized DocumentInfo getDocumentInfo(String docID)
    throws RepositoryException
    {
        PersistentDocumentInfo docRow = null;
        try
        {
            docRow = new PersistentDocumentInfo(
                getTableInfo(TableSpec.TYPE_DOCS),
                getMetadataConnection()
                );
            docRow.fetchRow(docID);
            return docRow.getDocumentInfo();
        }
        catch (SQLException e)
        {
            throw new RepositoryException(RepositoryException.DB_ERROR,
                "SQL exception while getting document information.", e);
        }
        finally
        {
            if (docRow != null)
                docRow.close();
        }
    }
    
    public synchronized List getDocumentInfoList(XMLCollectionImpl collection) throws RepositoryException
    {
        PersistentDocumentInfo docRow = null;
        ArrayList docList = new ArrayList();
        try
        {
            docRow = new PersistentDocumentInfo(
                getTableInfo(TableSpec.TYPE_DOCS),
                getMetadataConnection()
                );
            docRow.retrieveDataSet();
            while (docRow.fetchNextRow())
                docList.add(new RepositoryDocumentImpl(docRow.getDocumentInfo(), collection));
        }
        catch (SQLException e)
        {
            throw new RepositoryException(RepositoryException.DB_ERROR,
                "SQL exception while getting document information.", e);
        }
        finally
        {
            if (docRow != null)
                docRow.close();
        }
        return docList;
    }
    
    public synchronized void createIndexes() throws RepositoryException
    {
        createIndexes(mapping, config, cid, collectionTables.values(), rep, getMetadataConnection());
    }
    
    public synchronized void dropIndexes() throws RepositoryException
    {
        dropIndexes(mapping, config, cid, collectionTables.values(), rep, getMetadataConnection());
    }
    
    /* Should synchronize ? */
    public synchronized void updateStatistics() throws XMLDBCException
    {
        DatabaseMetaData DBMeta = null;
        ResultSet rs = null;
        
        try
        {
            // XQuark tables
            DBMeta = getMetadataConnection().getConnection().getMetaData();
            rs = DBMeta.getTables(
                            null,
        					getMetadataConnection().getSchemaName(),
                            TABLE_PREFIX + name + "%",
                            null
                            );
            while (rs.next())
            {
                getMetadataConnection().updateStatistics(rs.getString(3));
            }
            
            // user tables
            int tableCount = mapping.getTableCount();
            for (int i = 0; i <tableCount; i++)
            {
                getMetadataConnection().updateStatistics(mapping.getTableMapping(i).getTableName());
            }
        }
        catch (SQLException e) {
            throw new RepositoryException(RepositoryException.DB_ERROR,
                "SQL exception while trying to update database statistics", e);
        }
        finally
        {
            try
            {
                if (rs != null)
                    rs.close();
            }
            catch (SQLException e)
            {
                // no op
            }
            
        }
        
    }
    
    /** Synchronizes metadata with database.
     */
    public synchronized void refresh() throws XMLDBCException
    {
        pathSet.refresh();
    }
    
    public void updateDescription(String desc) throws XMLDBCException
    {
        try
        {
        rep.updateCollectionDescription(cid, desc);
        }
        catch (SQLException e)
        {
            throw new RepositoryException(RepositoryException.DB_ERROR,
                "Could not update collection description.", e);
        }
        config.setDescription(desc);
    }
        
    ////////////////////////////////////////////////////////////////////
    //                          STATIC
    ////////////////////////////////////////////////////////////////////
    static void initializeCollection(
                    RepositoryMapping mapping,
                    PersistentCollectionInfo configRow,
                    Repository rep,
                    AbstractConnection connection
                    )
                    throws RepositoryException
    {
        try {
            /* First, add (reserve) an entry in the collections table (create a CID) */
            try
            {
                configRow.createRow();
            }
            catch (SQLException e)
            {
                throw new RepositoryException(RepositoryException.DB_ERROR,
                "Failed to insert collection in configuration table.", e);
            }
            
            /* load tableInfo for collection tables and create it */
            Map tableInfo = loadTableInfo
            (
            configRow.getCollectionID(),
            rep.getInfo().getColIDSize(),
            configRow.getCollectionInfo(),
            connection
            );
            
            /* table creation */
            Iterator it = connection.createTables(tableInfo.values()).iterator();
            TableInfo wTable;
            
            /* sequence creation (step must be specified) */
            while (it.hasNext())
            {
                wTable = (TableInfo)it.next();
                short step = 1;
                if (wTable.getType() == TableSpec.TYPE_DOC_SEQ)
                    step = configRow.getCollectionInfo().getDocOIDPreallocationSize();
                new Sequence(wTable, step).create(connection);
            }
        }
        catch (SQLException e)
        {
            dropCollectionTables(configRow, connection); // because in general transactions don't work with DDL
            // and sometimes (Sybase) carry out an exception
            throw new RepositoryException(RepositoryException.DB_ERROR,
                                "Failed to create tables for collection.", e);
        }
        
        // Create OID tables
        Statement stmt = null;
        try
        {
            Iterator it = mapping.joinTableIterator();
            TableMapping table;
            String SQLString;
            String[] indexes;
            
            stmt = connection.getConnection().createStatement();
			while (it.hasNext())
			{
				table = (TableMapping) it.next();
				// create a CollectionMappingInfo object
				CollectionMappingInfo generator =
					new CollectionMappingInfo(
						table,
						configRow.getCollectionID(),
						rep.getTableManager().getTableID(table.getTableName()),
						connection);
				connection.addBatch(
					stmt,
					generator.getOIDTableCreateStatement());
				indexes = generator.getOIDIndexCreateStatements();

				for (int i = 0; i < indexes.length; i++)
					connection.addBatch(stmt, indexes[i]);
			}
            connection.executeBatch(stmt);
            
        }
        catch (SQLException e)
        {
            dropCollectionTables(configRow, connection); // because in general transactions don't work with DDL
            // and sometimes (Sybase) carry out an exception

            // free colection table entry
            try
            {
                configRow.deleteRow();
            }
            catch (SQLException ex)
            {
                throw new RepositoryException(RepositoryException.DB_ERROR,
                        "Failed to create system tables for collection and could no remove entry in collection table.", ex);
            }
 
            throw new RepositoryException(RepositoryException.DB_ERROR,
                        "Failed to create system tables for collection.", e);
        }
    }
    
    static void dropCollectionTables(PersistentCollectionInfo configRow, short cidSize, AbstractConnection connection)
    throws RepositoryException
    {
        /* load tableInfo for collection tables */
        Map tableInfo = loadTableInfo
                                    (
                                    configRow.getCollectionID(),
                                    cidSize,
                                    configRow.getCollectionInfo(),
                                    connection
                                    );
        dropCollectionTables(configRow, connection);
    }
    
    private static void dropCollectionTables(
        PersistentCollectionInfo configRow,
        AbstractConnection connection
        )
    throws RepositoryException
    {
        // delete all tables with XQuark prefixes
        String pattern = TABLE_PREFIX + configRow.getCollectionID() + "%";
        try
        {
            Iterator it = connection.getUserTableNames(pattern).iterator();
            while (it.hasNext())
            {
                try
                {
                    connection.dropTable((String)it.next());
                }
                catch (SQLException e)
                {
                    // ignore
                }
            }
        }
        catch (SQLException e)
        {
            throw new RepositoryException(RepositoryException.DB_ERROR,
            "Failed to drop collection tables.", e);
        }
        
        // drop sequences
        try
        {
            Iterator it = connection.getUserSequenceNames(pattern).iterator();
            while (it.hasNext())
            {
                try
                {
                    connection.dropSequence((String)it.next());
                }
                catch (SQLException e)
                {
                    // ignore
                }
            }
        }
        catch (SQLException e)
        {
            throw new RepositoryException(RepositoryException.DB_ERROR,
            "Failed to drop collection sequences.", e);
        }
    }
    
    /** Creates indexes on the storage tables.
     * To use in order to improve performance during storage.
     * @throws RepositoryException Application exception.
     */
    public static void createIndexes(
                                    RepositoryMapping mapping,
                                    CollectionInfo config,
                                    short cid,
                                    Repository rep,
                                    AbstractConnection connection
                                    ) 
                                    throws RepositoryException
    {
        createIndexes (
                    mapping,
                    config,
                    cid,
                    loadTableInfo(cid, rep.getInfo().getColIDSize(), config, connection).values(),
                    rep,
                    connection
                    );
        
    }
    
    /** Drops indexes on the storage tables.
     * To use in order to improve performance during storage.
     * @throws RepositoryException Application exception.
     */
    public static void dropIndexes(
                                    RepositoryMapping mapping,
                                    CollectionInfo config,
                                    short cid,
                                    Repository rep,
                                    AbstractConnection connection
                                    ) 
                                    throws RepositoryException
    {
        dropIndexes (
                    mapping,
                    config,
                    cid,
                    loadTableInfo(cid, rep.getInfo().getColIDSize(), config, connection).values(),
                    rep,
                    connection
                    );
        
    }
    
	////////////////////////////////////////////////////////////////////
	//                          PACKAGE PRIVATE
	////////////////////////////////////////////////////////////////////
	void releaseJDBCRessources()
	{
		pathSet.releaseJDBCRessources();
	}
	
	////////////////////////////////////////////////////////////////////
	//                          PRIVATE
	////////////////////////////////////////////////////////////////////

    private static void createIndexes(
                                        RepositoryMapping mapping,
                                        CollectionInfo config,
                                        short cid,
                                        Collection tableInfos,
                                        Repository rep,
                                        AbstractConnection connection
                                        ) 
                                        throws RepositoryException
    {
        /* load tableInfo for collection tables and create indexes */
        try {
            connection.createIndexes(tableInfos);
        }
        catch(SQLException e)
        {
            throw new RepositoryException(RepositoryException.DB_ERROR,
                "JDBC error while creating indexes on collection tables.", e);
        }
        
        Statement stmt = null;
        try {
            // Create path OID tables indexes (default mapping excluded cos' does not exist yet)
            Iterator it = mapping.joinTableIterator();
            String[] SQLStmts;
            TableMapping table = null;
            
            stmt = connection.getConnection().createStatement();
            while (it.hasNext())
            {
				table = (TableMapping) it.next();
				// create a CollectionMappingInfo object
				CollectionMappingInfo generator =
					new CollectionMappingInfo(
						table,
						cid,
						rep.getTableManager().getTableID(table.getTableName()),
						connection);
				SQLStmts = generator.getOIDIndexCreateStatements();
				for (int i = 0; i < SQLStmts.length; i++)
					connection.addBatch(stmt, SQLStmts[i]);
            }
            connection.executeBatch(stmt);
        }
        catch(SQLException e)
        {
            throw new RepositoryException(RepositoryException.DB_ERROR,
                "JDBC error while creating indexes on data index tables.", e);
        }
        finally
        {
            try
            {
                if (stmt != null)
                    stmt.close();
            }
            catch(SQLException ex)
            {
                // no op
            }
        }
    }
    
    private  static void dropIndexes(
                                    RepositoryMapping mapping,
                                    CollectionInfo config,
                                    short cid,
                                    Collection tableInfos,
                                    Repository rep,
                                    AbstractConnection connection
                                    ) 
                                    throws RepositoryException
    {
        /* load tableInfo for collection tables and destoy indexes */
        try {
            connection.dropIndexes(tableInfos);
        }
        catch(SQLException e)
        {
            throw new RepositoryException(RepositoryException.DB_ERROR,
                "JDBC error while dropping indexes on collection tables.", e);
        }
        
        /* destroys indexes on value tables */
        Statement stmt = null;
        try
        {
            // Create path OID tables indexes (default mapping excluded cos' does not exist yet)
            Iterator it = mapping.joinTableIterator();
            String[] SQLStmts;
            TableMapping table = null;

			stmt = connection.getConnection().createStatement();
			while (it.hasNext())
			{
				table = (TableMapping) it.next();
				// create a CollectionMappingInfo object
				CollectionMappingInfo generator =
					new CollectionMappingInfo(
						table,
						cid,
						rep.getTableManager().getTableID(table.getTableName()),
						connection);
				SQLStmts = generator.getOIDIndexDropStatements();
				for (int i = 0; i < SQLStmts.length; i++)
					connection.addBatch(stmt, SQLStmts[i]);
			}
            connection.executeBatch(stmt);
        }
        catch(SQLException e)
        {
            throw new RepositoryException(RepositoryException.DB_ERROR, "JDBC error while creating indexes on OID tables.", e);
        }
        finally
        {
            try
            {
                if (stmt != null)
                    stmt.close();
            }
            catch(SQLException e)
            {
                throw new RepositoryException(RepositoryException.DB_ERROR, "JDBC error while closing statement.", e);
            }
        }
    }
    
    ////////////////////////////////////////////////////////////////////
    //                      MappingSet implementation
    ////////////////////////////////////////////////////////////////////
    public MappingSetIterator getMappingSetIterator()
    {
        return new MappingSetIteratorImpl(mappingStatements);
    }
    
    public MappingInfo getMappingStatements(int index)
    {
        return mappingStatements[index];
    }
    
    public int getMappingStatementsCount()
    {
        return mappingStatements.length;
    }
    
    public MappingSetIterator getSortedTableMappingSetIterator(final boolean mostDependantFirst)
    {
        return new MappingSetIterator()
        {
        	private int next = 0;
        	private MappingSetIterator mappingIterator
        		= mapping.getSortedTableMappingSetIterator(mostDependantFirst);
        		
        	public boolean hasNext()
        	{
        		return (next == 0 && !mostDependantFirst)
        				|| mappingIterator.hasNext()
        				|| (next == (mappingStatements.length - 1) && mostDependantFirst);
        	}
        	
        	public MappingInfo next()
        	{
        		MappingInfo ret = null;
        		next++;
        		if (next == 1 && !mostDependantFirst)
        			ret = mappingStatements[defaultMappingIndex];
        		else if (next == mappingStatements.length && mostDependantFirst)
        			ret = mappingStatements[defaultMappingIndex];
        		else if (mappingIterator.hasNext())
        			ret = mappingStatements[mappingIterator.next().getTableMapping().getTableIndex()];
        		return ret;
        	}
        	
        	public boolean isTableFirstOccurence()
        	{
        		return (((next == 1) && !mostDependantFirst)
        			|| ((next == mappingStatements.length) && mostDependantFirst))
        			|| mappingIterator.isTableFirstOccurence();
        	}

        	public int getFirstOccurenceTableIndex()
        	{
        		if ((next == 1 && !mostDependantFirst)
        		|| (next == mappingStatements.length && mostDependantFirst))
        			return defaultMappingIndex;
        		else
					return mappingIterator.getFirstOccurenceTableIndex();
        	}

        };
    }
}
