/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.mapper.metadata;

import java.util.List;

import org.xquark.mapper.RepositoryException;
import org.xquark.mapper.storage.PathIterator;
import org.xquark.schema.ElementDeclaration;
import org.xquark.xpath.*;

/** This class allows to iterate through mappings of a collection RepositoryMapping 
 * object using names (and not Schema declarations).
 */
public class MetadataIterator implements PathIterator
{
    private static final String RCSRevision = "$Revision: 1.2 $";
    private static final String RCSName     = "$Name: XQ1_1_1 $";
    
    //
    //  DATA
    //
    
    /** The XMLCollection Mapping browsed
     */
    protected PathSetFactory factory;
    protected PathSet pathSet;
    protected PathNode wNode;
    protected XNode pattern = new XNode();
    
    //
    //  CONSTRUCTORS
    //
    
    public MetadataIterator(PathSetFactory factory)
    {
        this(factory, (PathNode)factory.getTree().getRoot());
    }
    public MetadataIterator(PathSetFactory factory, PathNode root)
    {
        this.factory = factory;
        pathSet = (PathSet)factory.getTree();
        wNode = root;
    }
    
    
    public void reset()
    {
        wNode = (PathNode)factory.getTree().getRoot();
    }
    
    /** Get the default table mapping
     */
    public List getDefaultMapping()
    {
        return pathSet.getCollection().getDefaultMappingList();
    }
    
    /** Get the table mappings for a particular subelement in the current context
     */
    public List getElementTableMappings(String namespace, String localName)
    {
        PathNode node = getChild(namespace, localName, NodeKind.ELEMENT);
        if (node == null)
            return null;
        else
            return node.getMapping().getTableMappings();
    }
    
    /** Get the column mappings for a particular subelement in the current context
     */
    public List getElementColumnMappings(String namespace, String localName)
    {
        PathNode node = getChild(namespace, localName, NodeKind.ELEMENT);
        if (node == null)
            return null; // SET schema component ?
        else
            return node.getMapping().getColumnMappings();
    }
    
    /** Get the table mappings for a particular attribute in the current context
     */
    public List getAttributeTableMappings(String namespace, String localName)
    {
        PathNode node = getChild(namespace, localName, NodeKind.ATTRIBUTE);
        if (node == null)
            return null; // SET schema component ?
        else
            return node.getMapping().getTableMappings();
    }
    
    /** Get the column mappings for a particular attribute in the current context
     */
    public List getAttributeColumnMappings(String namespace, String localName)
    {
        PathNode node = getChild(namespace, localName, NodeKind.ATTRIBUTE);
        if (node == null)
            return null; // SET schema component ?
        else
            return node.getMapping().getColumnMappings();
    }
    
    /** Get the column mappings for a particular attribute in the current context
     */
    public PathNode getChild(String namespace, String localName, byte type)
    {
        pattern.set(namespace, localName, type);
        return (PathNode)wNode.getChild(pattern);
    }

    /** Get the column mappings for a particular attribute in the current context
     */
    public PathMetadata getPathMetadata(short pid)
    throws RepositoryException
    {
        return (PathNode)pathSet.get(pid);
    }
    
    public PathNode getPathNode(short pid)
    throws RepositoryException
    {
        return (PathNode)pathSet.get(pid);
    }

    public String getNameSpace()
    {
        return wNode.getNamespace();
    }
    
    public String getLocalName()
    {
        return wNode.getLocalName();
    }
    
    public ElementDeclaration getElementDeclaration()
    {
        return (ElementDeclaration)wNode.getDeclaration();
    }
    
    /* Create a node if it does not exist without entering it */
    public StoragePathMetadata createNode(String namespace, String localName, byte type)
    throws RepositoryException
    {
        PathNode node;
        try {
            node = (PathNode)factory.createNamedNodeIfNotExist(wNode, namespace, localName, type); // SET schema component ?
        }
        catch (XTreeRuntimeException e)
        {
            if (e.getException() instanceof RepositoryException)
                throw (RepositoryException)e.getException();
            XNode parser = new XNode(namespace, localName, type);
            throw new RepositoryException(RepositoryException.DATA_LOSS,
            "Error while creating new path " + wNode.getLocation() + "/" + parser.getAbbreviatedSyntax()
            + " in metadata.", e.getException());
        }
        
        return node; // SET schema component ?
    }

    /* Enter the specified node element or create it if does not exist */
    public StoragePathMetadata push(String namespace, String localName) throws RepositoryException
    {
        return wNode = (PathNode)createNode(namespace, localName, NodeKind.ELEMENT);
    }

    /* Enter the specified node */
    public PathMetadata push(short pid)
    throws RepositoryException
    {
        return wNode = (PathNode)pathSet.get(pid);
    }

    /* Enter the specified node */
    public PathMetadata push(PathNode node)
    {
        return wNode = node;
    }
    
    /* Enter the specified node or create it if does not exist */
    public StoragePathMetadata pop()
    {
        PathNode leftNode = wNode;
        wNode = (PathNode)wNode.getParent();
        if (wNode == null)
            wNode = leftNode;
//        Debug.print("MetadataIterator :: POP = ", leftNode.getAbbreviatedSyntax());
        return leftNode;
    }
    
    /** Return the absolute position of the locator.
     * <p>i.e., !isDiscardable()</p>
     */
    public PathExpr getLocation()
    {
        return wNode.getLocation();
    }
    
    public StoragePathMetadata getStoragePathMetadata()
    {
        return wNode;
    }
    
    public PathMetadata getPathMetadata()
    {
        return wNode;
    }
    
    public PathNode getPathNode()
    {
        return wNode;
    }
}
