/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

/*
 * PathMetadata.java
 *
 * Created on 18 septembre 2001, 18:11
 */
package org.xquark.mapper.metadata;

import java.util.Collection;
import java.util.List;

import org.xquark.mapper.mapping.ColumnMapping;

/**
 * Common interface for accessing an XML collection metadata for a particular
 * path. Should be used both by both storage and query modules.
 *
 */
public interface PathMetadata extends StoragePathMetadata
{
    /**
     * Get the column mapping corresponding to one of the table mappings
     * passed as a parameter.
     * @param tableMappingIndexes an array containing the mapping numerical IDs
     * (index) of table mappings
     * @return the reference one if no table or multiple mapping match. Null if
     * no column mappings available
     * @see org.xquark.mapper.mapping.TableMapping
     * @see org.xquark.mapper.mapping.ColumnMapping
     */
     ColumnMapping getColumnMapping(int[] tableMappingIndexes);
     
     /**
      * Get the MetadataNode on which the table mapping corresponding to
      * the 'read' column mapping is defined.
      * @return null if no column mapping is defined on the current path
      */
     PathMetadata getTableMappingPath();    // to perform at run-time (pre-compilation could be processed using mapping loader method)
     
     /**
      * Get the MetadataNode on which the table mapping corresponding to the
      * specified column mapping is defined.
      * @param column the column mapping which the table mapping path is searched
      * for
      * @return the path of the node on which the table mapping is defined
      * @see org.xquark.mapper.mapping.ColumnMapping
      */
     PathMetadata getTableMappingPath(ColumnMapping column);    // to perform a run-time (pre-compilation could be processed using mapping loader method)
     
     /**
      * Get the depth of the node in the metadata tree.
      * @return an int value where the tree root has 0 value.
      */
     int getDepth();
     
     /** Flag to indicate if the node can be removed from the storage
      * model (no OID allocated).
      * i.e., getStorageMode() == DISCARDED
      * @return true if node can be optimized
      */
    boolean isDiscardable();
    
    /** Indicate if the path may have extra-data (xsi and xmlns attributes)
     * regarding the documents stored in the database.
     * @return True if a some paths have extra-data (but not all.
     */
    boolean hasExtraData();
    
    /** Indicate if the position of the node children is enforced
     * by schema information.
     * @return True if positon of the node can be enforced by schema information.
     */
    boolean isChildrenPositionEnforced();
    
    /** Get the table mappings needed to build the element (default included
     * if necessary).
     * @return null if the node cannot be built from itself but using its
     * builder ancestor node.
     */
    Collection getBuildMappings();
    
     /**
      * Get the first ancestor having in scope all the tables to build itself
      * and the current node.
      * @return a PathMetadata node.
      */
     PathMetadata getBuildableAncestor();	// TO PRECOMPILE ?
     
     /**
      * Get the first ancestor having in scope all the tables to build the
      * current node (even if itself is not buildable).
      * @return a PathMetadata node.
      */
     PathMetadata getBuilderAncestor();	// TO PRECOMPILE ?
     
     /**
      * Get the first ancestor having a structure tuple.
      * <I>Potentially useless.</I>
      * @return a PathMetadata node.
      */
     PathMetadata getStructAncestor(); // USE ?
     
     /**
      * Get the first ancestor having an OID.
      * @return a PathMetadata node.
      */
     PathMetadata getModelAncestor();
     
     /**
      * Get the minimum & maximum number of occurences of this path
      * in an ancestor.
      * @param ancestor the ancestor PathMetadata node which the occurrence is
      * calculated in . If the parameter node is not
      * encountered as one of the current node ancestor,
      * calculation is performed relatively to the metadata
      * tree root.
      * @return a two int arrays where first if the minimum number of
      * occurence, and the second is the maximum.
      */
    int[] getOccurenceCount(PathMetadata ancestor);    // to perform at run-time (pre-compilation could be processed using mapping loader method)

    /** Return true if the element can only appear one time in its parent.
     * @return true if monovaluated
     */
    boolean isMonoValuated();
    
    /** Return true if the element can appear several times in its parent.
     * @return true if multivaluated
     */   
    boolean isMultiValuated();
    
    /** Return true if the element can be missing if its parent exists.
     * @return true if optional
     */    
    boolean isOptional();
    
    /** Return a list of the path nodes with an OID contained in the subtree.
     * @return null if the path is of root type.
     */
    public List getModelSubPaths();
    
    /** Return a list of the path nodes with or without an OID contained in
     * the subtree.
     * @return null if the path is of root type.
     */
    public List getSubPaths();
    
    /** get column mapping used for reconstruction and queries (that may be
     * different than the user-defined ones). Data may or may not exist.
     * This method returns a value only if node type is TEXT_ONLY. For MIXED,
     * use the {@link getTextColumnMapping()} method.
     */
    public ColumnMapping getReadColumnMapping();
    
    /** get Table mapping used for reconstruction and queries (that may be
     * different than the user-defined ones). Data may or may not exist.
     * This method returns a value only if node type is TEXT_ONLY. For MIXED,
     * use the {@link getTextTableMapping()} method.
     */
    public CollectionMappingInfo getReadTableMapping();

    /** Shortcut to the schema content type.
     * @return true if elemnt is mixed.
     */
    boolean isMixed();
    
    /** get column mapping used for text nodes. This method returns a value only
     * if {@link isMixed()} is true
     */
    public ColumnMapping getTextColumnMapping();
    
    /** get table mapping used for text nodes. This method returns a value only
     * if {@link isMixed()} is true
     */
    public CollectionMappingInfo getTextTableMapping();

}

