/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.mapper.metadata;


import java.sql.SQLException;
import java.util.*;

import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.AttributesImpl;
import org.xquark.mapper.RepositoryException;
import org.xquark.mapper.mapping.*;
import org.xquark.schema.*;
import org.xquark.xpath.*;

/**
 * Data structure that corresponds to XML structure paths and allows
 * direct access to mapping.
 *
 */
public class PathNode extends TypedXTreeNode implements Cloneable, PathMetadata
{
    private static final String RCSRevision = "$Revision: 1.2 $"; 
    private static final String RCSName     = "$Name: XQ1_1_1 $";
    private static final short POTENTIAL  = POTENTIAL_STRUCT|DATA_AND_STRUCT;
    private static final short STRUCT     = DATA_AND_STRUCT|STRUCT_ONLY|POTENTIAL_STRUCT;
    
    private short pid = RepositoryConstants.ROOT_PATH_ID;
    private MappingNode mapping;
    
    private CollectionMappingInfo readTableMapping;
    private ColumnMapping readColumnMapping;
    
    private CollectionMappingInfo textTableMapping;
    private ColumnMapping textColumnMapping;
    
    private byte storageMode = UNKNOWN;    // for compatibility with ROOT OR
    private boolean childrenOrderEnforced = false;  // false when no schema info available
    private boolean mixed = false;
    private boolean defaultAllowed = false;
    /** the first ancestor that allows this node construction
     */
    private PathNode builder = null;
    /** Number min and max of occurrence of this node in its parent node */
    private int [] occurrenceCount;
    
    /** Path node list of decendant-or-self. Cached when asked */
    private List modelSubPaths = null;  // for struct & data
    private List subPaths = null;   // for extra data
    
    private boolean hasDefaultMappingInScope = false;   // TO IMPROVE : for typed default mapping, could be a bitmap for all the tables for basic types
    // TO CHECK HOW DEFAULT MAPPING IS HANDLED for schemaless and Mixed with optional sub elements
    private List buildMappings = null; // allocated only if necessary
    
    public PathNode(PathSet tree, PathNode parent, String namespace, String localName, byte type)
    {
        super(tree, parent, namespace, localName, type);
    }
    
    public synchronized short getPathID()
    {
        return pid;
    }
    
    public synchronized void setPathID(short pathId)
    {
        pid = pathId;
    }
    
    /* to use when a mapping node already exist. Called by RepositoryMapping duplication */
    private void setMapping(MappingNode node)
    throws RepositoryException
    {
        PathSet pathSet = (PathSet)getTree();
        
        // register mapping to mapping to path index
        if (node.getTableMappings() != null)
        {
	        Iterator it = node.getTableMappings().iterator();
	        while (it.hasNext())
	        	pathSet.registerMapping((TableMapping)it.next(), this);
        }
       
        mapping = node;
        setOccurenceCount();    // used by setStorageFlags
        if ((mapping.getColumnMappings() == null)
                && mapping.isDefaultAllowed())
        {
            defaultAllowed = true;
            setDefaultMappingInScope();
        }
        else if (mapping.potentialNewPathsExist())
            setDefaultMappingInScope();
        setStorageFlags(!pathSet.isLoading());
        setReadMappings(pathSet.getCollection());
        
        if (pathSet.isNew())
            saveIfNoLoading(pathSet);
    }
    
    private void saveIfNoLoading(PathSet pathSet)
    throws RepositoryException
    {
        if ((pid != RepositoryConstants.ROOT_PATH_ID) && !pathSet.isLoading())
        {
            try
            {
                pathSet.store(this);
            }
            catch (SQLException e)
            {
                // collision (hope it is)
                // remove node from the tree and from the path set index
                remove();
                pathSet.freeID(this);
                
                // eventually, refresh to fetch the new node
                pathSet.refresh();
            }
        }
    }
    
    /* to use when a mapping node does not exist */
    void setDefaultMapping(MappingFactory mappingFactory, byte type)
    throws RepositoryException
    {
        PathSet pathSet = (PathSet)getTree();
        // Alocate a new "empty" (no default mapping in it til storage) standalone mapping node
        mapping = (MappingNode)mappingFactory.allocateNode(
                    ((PathNode)getParent()).getMapping().getSchemaComponent(),
                    this
                    );
        setOccurenceCount();    // used by setStorageFlags
        if (mapping.isDefaultAllowed())
            defaultAllowed = true;
        setDefaultMappingInScope();
        
        setStorageFlags(!pathSet.isLoading());
        setReadMappings(pathSet.getCollection());
        
        saveIfNoLoading(pathSet);
    }
    
    private void setReadMappings(CollectionMetadata collection)
    throws RepositoryException
    {
        ColumnMapping ref = getReferenceColumnMapping();
        CollectionMappingInfo defMappingInfo =
            (CollectionMappingInfo)collection.getMappingStatements(collection.getDefaultMappingIndex());
        TableMapping refTm = getReferenceTableMapping();
        // read mappings
        if (ref == null)
        {
            if(isDefaultAllowed())// || (storageMode == POTENTIAL_STRUCT))
            {
                readColumnMapping = collection.createDefaultMappingView(getDeclaration());
                readTableMapping = defMappingInfo;
            }
            else if (refTm != null)
                readTableMapping = (CollectionMappingInfo)collection.getMappingStatements(refTm.getTableIndex());
        }
        else
        {
            if (ref.getTableMapping() == refTm)
                readTableMapping = (CollectionMappingInfo)collection.getMappingStatements(refTm.getTableIndex());
            readColumnMapping = ref;
        }
        // text mappings
        if (isMixed())
        {
            textColumnMapping = collection.getDefaultMapping().getDefaultColumnMapping(null);
            textTableMapping = defMappingInfo;
        }
    }
    
    private void setOccurenceCount()
    {
        Declaration parentDecl, decl = getDeclaration();
        if (decl == null)
            occurrenceCount = new int[] {0, Integer.MAX_VALUE}; // Integer.MAX_VALUE is used as unbounded by schema API
        else
        {
            switch (getType())
            {
                case NodeKind.NONE:
                    occurrenceCount = new int[] {1, 1};
                    return;
                    
                case NodeKind.ATTRIBUTE:
                    switch (((AttributeDeclaration)mapping.getSchemaComponent()).getUse())
                    {
                        case SchemaConstants.REQUIRED:
                            occurrenceCount = new int[] {1, 1};
                            break;
                        case SchemaConstants.OPTIONAL:
                            occurrenceCount = new int[] {0, 1};
                            break;
                        case SchemaConstants.PROHIBITED:
                            occurrenceCount = new int[] {0, 0};
                            break;
                    }
                    break;
                case NodeKind.ELEMENT:
                    if ((parentDecl = ((PathNode)getParent()).getDeclaration()) == null)
                    {
                        occurrenceCount = new int[] {1, 1};
                        return;
                    }
                    occurrenceCount = ((ComplexType)parentDecl.getType()).getDeclarationOccurrence
                                    (
                                    decl.getNamespace(),
                                    decl.getName()
                                    );
                    break;
            }
        }
    }
    
    private int[] getOccurrenceCount(int[] occurrence)
    {
        long mult;
        
        for (int i = 0; i <= 1; i++)
        {
            mult = (long)occurrence[i] * (long)occurrenceCount[i];
            occurrence[i] = (mult <= Integer.MAX_VALUE ? (int)mult : Integer.MAX_VALUE);
        }
        return occurrence;
    }
    
    public MappingNode getMapping()
    {
        return mapping;
    }
    
    private void setDefaultMappingInScope()
    {
        // TO CHECK : setting the flag for node without mapping node ?
        for (
            PathNode elt = this;
            !((elt == null) || elt.hasDefaultMappingInScope);
            elt = (PathNode)elt.getParent()
            )
        {
            elt.hasDefaultMappingInScope = true;
        }
    }

    void setStorageModeFlag(byte storageMode)
    {
        this.storageMode = storageMode;
    }
    
    private void setStorageFlags(boolean calculateStorageFlag)
    {
        
        //
        //      BUILDER LINK
        //
        
        PathExpr builderLocation = mapping.getBuilder().getLocation();
        if (builderLocation == null)    // <> ROOT or generic mapping (flying node)
            builder = this;
        else
            builder = (PathNode)getTree().getNode(builderLocation);
        
        //
        //      mixed FLAG
        //

        if (getType() == NodeKind.ELEMENT)
        {
            if (getDeclaration() == null)
                mixed = true;
            else
            {
                org.xquark.schema.Type type = ((ElementDeclaration)getDeclaration()).getType();
                if ((type instanceof ComplexType) &&
                    (((ComplexType)type).getContentType() ==  SchemaConstants.MIXED))
                    mixed = true;
            }
        }
        if (mixed)
            setDefaultMappingInScope();

        
        //
        //      STORAGE MODE FLAG
        //
        
        if (getType() == NodeKind.NONE)
            storageMode = DISCARDED;
        else if(calculateStorageFlag)
            calculateStorageModeFlag();
        // else will be read from database
        
        
        //
        //      CHILDREN ORDER FLAG
        //
        if (getDeclaration() != null)   // schema info available
        {
            if (((Declaration)getDeclaration()).getType() instanceof SimpleType)
                childrenOrderEnforced = true;    // meaningless in fact but true in a sense
            else
            {
                ComplexType type = (ComplexType)((Declaration)getDeclaration()).getType();
                switch (type.getContentType())
                {
                    case SchemaConstants.EMPTY: case SchemaConstants.TEXT_ONLY:
                        childrenOrderEnforced = true;    // meaningless in fact but true in a sense
                        break;
                    case SchemaConstants.MIXED:
                    case SchemaConstants.ELEMENT_ONLY:
                        childrenOrderEnforced = browseParticle((Particle)type.getContentModel().getModel(),
                        new ArrayList(type.getElementDeclarations()));
                        break;
                }
            }
        }
        // else, no schema: default = false
    }
    
    void calculateStorageModeFlag()
    {
        // POTENTIAL_STRUCT : schemaless case
        if (getDeclaration() == null)
        {
            if (getType() == NodeKind.ELEMENT)
                storageMode = POTENTIAL_STRUCT;
            else // attribute
                storageMode = DATA_ONLY;
        }
        else
        {
//            boolean potentialStruct = false;
            boolean potentialLeafElement = true; // leaf element, not model-leaf (attributes ignored)
            boolean potentialModelLeaf = true;
            boolean modelLeaf = false;
            boolean hasChildren = false;
            boolean hasAttributes = false;
            
            // Browsing schema info
            switch (getType())
            {
                case NodeKind.ELEMENT:
                    if (((ElementDeclaration)getDeclaration()).getType() instanceof ComplexType)
                    {
                        ComplexType type = (ComplexType)((ElementDeclaration)getDeclaration()).getType();
                        switch (type.getContentType())
                        {
                            case SchemaConstants.MIXED: case SchemaConstants.ELEMENT_ONLY:
                                // sub declarations browsing
                                Iterator it = type.getElementDeclarations().iterator();
                                ElementDeclaration decl;
                                while (it.hasNext())
                                {
                                    decl = (ElementDeclaration)it.next();
                                    int[] occurence = type.getDeclarationOccurrence(decl.getNamespace(), decl.getName());
                                    if (occurence[1] != 0)
                                        hasChildren = true;
                                    if (occurence[0] != 0)
                                        potentialLeafElement = false;
                                }
                            case SchemaConstants.TEXT_ONLY: case SchemaConstants.EMPTY:
                                potentialModelLeaf = potentialLeafElement;
                                if (type.getAttributeDeclarations() != null)
                                {
                                    it = type.getAttributeDeclarations().iterator();
                                    int use;
                                    while (it.hasNext())
                                    {
                                        use = ((AttributeDeclaration)it.next()).getUse();
                                        if (use != SchemaConstants.PROHIBITED)
                                            hasAttributes = true;
                                        if(use == SchemaConstants.REQUIRED)
                                            potentialModelLeaf = false;
                                    }
                                }
                                modelLeaf = !hasAttributes && !hasChildren;
//                                potentialStruct = potentialModelLeaf;
                                break;
                            default:
                        }
                    }
                    else
                        modelLeaf = true;
                    break;
                case NodeKind.ATTRIBUTE:
                    modelLeaf = true;
                    break;
                default:    // SchemaConstants.EMPTY, SchemaConstants.ELEMENT_ONLY:
            }
            
            // main algorithm (refer to specifications)
            byte algorithmicStorageMode = DATA_AND_STRUCT;
            storageMode = DISCARDED;    // important in order that getModelAncestor() do not return itself
            boolean structTuple = true, dataTuple = true;
            boolean defaultOrText = defaultAllowed || mixed;
            boolean noTableMapping = (getTableMappings() == null) && !defaultOrText;
//            if (noTableMapping || (potentialStruct && !potentialLeafElement))
            if (noTableMapping)
                dataTuple = false;
            if (getModelAncestor() == null) // not to discard both model root and root element
                structTuple = true;
            else
            {
                int[] occurrence = getOccurenceCount(getModelAncestor());
                if (noTableMapping)  // keep struct node only if an intermediate non mono-valuated
                    structTuple = (getColumnMappings() == null)
                    && ((occurrence[0] == 0) || (occurrence[1] > 1));
                else    // if not leaf, range necessary
                {
                    int tmCount = 0;
                    if (getTableMappings() != null)
                        tmCount = getTableMappings().size();
                    if (defaultOrText && (tmCount == 0) && !modelLeaf)
                        structTuple = true;
                    else
                    {
                        if (defaultOrText)
                            tmCount++;
                        
                        // builder has been set previously
                        structTuple = !(getBuilderAncestor().getBuildMappings().size() == tmCount);
                    }
                }
            }
            
            if (structTuple)
            {
                if (dataTuple)
                    algorithmicStorageMode = DATA_AND_STRUCT;
                else
                    algorithmicStorageMode = STRUCT_ONLY;
            }
            else if (dataTuple)
                algorithmicStorageMode = DATA_ONLY;
            else
                algorithmicStorageMode = DISCARDED;
            
            // setting the actual flag : prioritization of rules
            // -> set the potential struct only if an OID is allocated
            if (potentialModelLeaf && (algorithmicStorageMode  == DATA_AND_STRUCT))
                storageMode = POTENTIAL_STRUCT;
            else
                storageMode = algorithmicStorageMode;
        }
    }
    
    /* false if ALL model group encountered or if an element declaration is
     * found in multiple particules in the content model.
     */
    private boolean browseParticle(Particle particle, Collection elementDecls)
    {
        
        if (particle.getTerm() instanceof Wildcard)
            return false;
        else if (particle.getTerm() instanceof ElementDeclaration)
        {
            if (!elementDecls.remove(particle.getTerm()))
                return false;   // not the first time it is removed
        }
        else
        {
            ModelGroup modelGroup = (ModelGroup)particle.getTerm();
            if (modelGroup.getCompositor() == ModelGroup.ALL)
                return false;
            
            Iterator it = ((ModelGroup)particle.getTerm()).iterator();
            boolean ret = true;
            
            while (it.hasNext())
            {
                if (!browseParticle((Particle)it.next(), elementDecls))
                    return false;
            }
        }
        return true;
    }
    
    /* Add customization to the nodes created by duplicate */
    protected XTreeNode customizeDuplicate(XTreeNode original)
    {
        PathSet pathSet = (PathSet)getTree();
        try {
            setMapping((MappingNode)original);
        }
        catch (RepositoryException e) {
            throw new XTreeRuntimeException(e);
        }
        return this;
    }
    
    public String toString()
    {
        StringBuffer buf = new StringBuffer();
        buf.append(getExpandedName());
        buf.append(" [pid=");
        buf.append(getPathID());
        buf.append(",rm=");
        buf.append(getReferenceColumnMapping());
        buf.append(",sm=");
        buf.append(storageMode);
        buf.append(",co=");
        buf.append(childrenOrderEnforced);
        buf.append(",ni=");
        buf.append(isNilImplicit());
        buf.append(",dma=");
        buf.append(mapping.isDefaultAllowed());
        buf.append(",dmis=");
        buf.append(hasDefaultMappingInScope);
        buf.append("]");

        return buf.toString();
    }

    ////////////////////////////////////////////////////////////////////////////
    //                      PathMetadata implementation
    ////////////////////////////////////////////////////////////////////////////
    
     /**
      * Get the minimum & maximum number of occurences of this path
      * in an ancestor.
      * @param ancestor the ancestor PathMetadata node which the occurrence is
      * calculated in . If the parameter node is not
      * encountered as one of the current node ancestor,
      * calculation is performed relatively to the metadata
      * tree root.
      * @return a two int arrays where first if the minimum number of
      * occurrence, and the second is the maximum.
      */
    public int[] getOccurenceCount(PathMetadata ancestor)
    {
        int[] occurrence = {1,1};
        
        for(PathNode current = this;
            (current.getPathID() != ancestor.getPathID()) && (current.getPathID() != RepositoryConstants.ROOT_PATH_ID); 
            current = (PathNode)current.getParent())
        {
            current.getOccurrenceCount(occurrence);
        }
        return occurrence;
    }
    
    public boolean isDefaultAllowed()
    {
        return defaultAllowed;
    }
    
    /**
     * Get the MetadataNode on which the table mapping corresponding to
     * the 'Read' column mapping is defined.
     * @return null if no column mapping is defined on the current path
     */
    public PathMetadata getTableMappingPath()
    {
        ColumnMapping cm = getReferenceColumnMapping();
        if (cm == null)
            return this;
        else
            return getTableMappingPath(cm);
    }
    
    /**
     * Get the MetadataNode on which the table mapping corresponding to the
     * specified column mapping is defined.
     * @param column the column mapping which the table mapping path is searched
     * for
     * @return the path of the node on which the table mapping is defined
     * @see org.xquark.mapper.mapping.ColumnMapping
     */
    public PathMetadata getTableMappingPath(ColumnMapping column)
    {
        PathNode node = this;
        if (column == null)
            return null;
        while((node.getTableMappings() == null)
            || !node.getTableMappings().contains(column.getTableMapping()))
        {
            node = (PathNode)node.getParent();
        }
        return node;
    }
    
    public PathMetadata getBuildableAncestor()
    {
        PathNode node = this;
        while(!node.getMapping().isBuildable())
        {
            node = (PathNode)node.getParent();
        }
        return node;
    }
    
    /**
     * Get the first ancestor having in scope all the tables to build the
     * current node (even if itself is not buildable).
     * @return a PathMetadata node.
     */
    public PathMetadata getBuilderAncestor()
    {
        return builder;
    }
    
    /**
     * Get the first ancestor having an OID.
     * @return a PathMetadata node.
     */
    public PathMetadata getModelAncestor()
    {
        PathNode node = this;
        while((node.getType() != NodeKind.NONE) && (node.getStorageMode() == DISCARDED))
        {
            node = (PathNode)node.getParent();
        }
        if (node.getType() == NodeKind.NONE)
            return null;
        else
            return node;
    }
    
    /**
     * Get the first ancestor having a structure tuple.
     * <I>Potentially useless.</I>
     * @return a PathMetadata node.
     */
    public PathMetadata getStructAncestor()
    {
        // Process the POTENTIAL_STRUCT case
        if ((getStorageMode() == POTENTIAL_STRUCT) && !isLeaf())
            return this;
        
        PathNode node = this;
        while((node.getStorageMode() & STRUCT) == 0)
        {
            node = (PathNode)node.getParent();
        }
        return node;
    }
    
    /** Indicate if looking in the extra-data is necessary to determine the
     * missing/nil status for a node.
     * @return True if a check is needed.
     */
    public boolean isNilImplicit()
    {
        // schemaless: nil has no meaning and no optimization is performed
        if (getDeclaration() == null)
            return false;
        
        boolean nillable = false, optional = false;
        switch (getType())
        {
            case NodeKind.ATTRIBUTE :
                optional = (((AttributeDeclaration)getDeclaration()).getUse() == SchemaConstants.OPTIONAL);
                break;
            case NodeKind.ELEMENT :
                ElementDeclaration decl = (ElementDeclaration)getDeclaration();
                optional = (occurrenceCount[0] == 0);
                nillable = decl.isNillable();
                break;
        }
        
        return ((getStorageMode() == DISCARDED)
                && (getReferenceColumnMapping() != null)
                && (nillable && !optional));
        // TO IMPROVE : PRECOMPILE
    }
    
    public boolean hasExtraData()
    {
        return true;    // TO IMPROVE: in a first step, no dynamic flag is maintained in PathSet
    }
    
    /** Indicate the storage mode of the node, i.e., if the node is
     * stored and where it is stored.
     * @return a constant for all possible storage scheme
     * @see #DISCARDED
     * @see #POTENTIAL_STRUCT
     * @see #STRUCT_ONLY
     * @see #DATA_ONLY
     * @see #DATA_AND_STRUCT
     */
    public byte getStorageMode()
    {
        return storageMode;
    }
    
    /** Flag to indicate if the node can be removed from the storage
     * model (no OID allocated).
     * i.e., getStorageMode() == DISCARDED
     * @return true if node can be optimized
     */
    public boolean isDiscardable()
    {
        return getStorageMode() == DISCARDED;
    }
    
    /** Flag to indicate if the node has been stored and an OID allocated.
     * <p>i.e., getStorageMode() != DISCARDED</p>
     * @return true if an OID exists for the node
     */
    public boolean isPersistent()
    {
        return getStorageMode() != DISCARDED;
    }
    
    public ColumnMapping getReferenceColumnMapping()
    {
        return mapping.getReferenceColumnMapping();
    }
    
    public TableMapping getReferenceTableMapping()
    {
        return mapping.getReferenceTableMapping();
    }
    
     /**
      * Get the column mappings defined on this path.
      * @return a list of column mappings
      * @see org.xquark.mapper.mapping.ColumnMapping
      */
     public List getColumnMappings()
     {
         return mapping.getColumnMappings();
     }
     
    /**
     * Get the column mapping corresponding to one of the table mappings
     * passed as a parameter.
     * @param tableMappingIndexes an array containing the mapping numerical IDs
     * (index) of table mappings
     * @return the reference one if no table or multiple mapping match. Null if
     * no column mappings available
     * @see org.xquark.mapper.mapping.TableMapping
     * @see org.xquark.mapper.mapping.ColumnMapping
     */
    public ColumnMapping getColumnMapping(int[] tableMappingIndexes)
    {
        ColumnMapping ret = null, column;
        if (mapping.getColumnMappings() == null)
            return null;
        
        Iterator it = mapping.getColumnMappings().iterator();
        while (it.hasNext())
        {
            column = (ColumnMapping)it.next();
            if (Arrays.binarySearch(tableMappingIndexes, column.getTableMapping().getTableIndex()) < 0)
            {
                if (ret == null)
                    ret = column;
                else
                    return getReferenceColumnMapping();
            }
        }
        return ret;
    }
    
    /**
     * Get the table mappings defined on this path.
     * @return a list of table mappings
     * @see org.xquark.mapper.mapping.TableMapping
     */
    public List getTableMappings()
    {
        return mapping.getTableMappings();
    }
    
    /** Get the table mappings in the scope of the element (default included
     * if necessary) needed to build itself or a subpath.
     * <p>To be used after getBuilderAncestor() or getBuildableAncestor().</p>
     * @return a collection of TableMapping objects or null (if no mapping).
     */
    public Collection getBuildMappings()
    {
        if (hasDefaultMappingInScope)
        {
            if (buildMappings == null)
            {
                if (mapping.getScopeMappings() == null)
                    buildMappings = (((PathSet)getTree()).getCollection()).getDefaultMappingList();
                else
                {
                    buildMappings = new ArrayList(mapping.getScopeMappings());
                    buildMappings.add((((PathSet)getTree()).getCollection()).getDefaultMapping());
                }
            }
            return buildMappings;
        }
        else
            return mapping.getScopeMappings();
    }
    
    /** Returns the schema component for this node.
     * <p>Allows to access declarations info like nillable, minOccurs...</p>
     * @return The schema component corresponding to the path if any.
     * Null else.
     */
    public Declaration getDeclaration()
    {
        return mapping.getDeclaration();
    }
    
    /** Indicate if the position of the node iwithin its siblings is enforced
     * by schema information.
     * @return True if positon of the node can be enforced by schema information.
     */
    public boolean isPositionEnforced()
    {
        if (getType() == NodeKind.ATTRIBUTE)
            return true;    // order ins meaningless and not guaranteed
        else
        {
            PathNode parent = (PathNode)getParent();
            return parent.childrenOrderEnforced && !parent.mixed;
        }
    }
    
    public boolean isChildrenPositionEnforced()
    {
        return childrenOrderEnforced;
    }
    
    protected void toXML(ContentHandler handler)
    throws SAXException
    {
        AttributesImpl atts = new AttributesImpl();
        String value;
        
        handler.startElement("", "data", "", atts);
        
        // Path ID
        handler.startElement("", "pid", "", atts);
        value = Short.toString(pid);
        handler.characters(value.toCharArray(), 0, value.length());
        handler.endElement("", "pid", "");
        
        // Occurence count
        atts.addAttribute("", "min", "",  "CDATA", String.valueOf(occurrenceCount[0]));
        atts.addAttribute("", "max", "",  "CDATA", String.valueOf(occurrenceCount[1]));
        handler.startElement("", "occurrenceCount", "", atts);
        atts.clear();
        handler.endElement("", "occurrenceCount", "");
        
        // Builder path
        handler.startElement("", "builder", "", atts);
        value = Short.toString(builder.getPathID());
        handler.characters(value.toCharArray(), 0, value.length());
        handler.endElement("", "builder", "");
        
        // Model ancestor path (anchor)
        PathMetadata modelAncestor = getModelAncestor();
        if (modelAncestor != null)
        {
            handler.startElement("", "anchor", "", atts);
            value = Short.toString(modelAncestor.getPathID());
            handler.characters(value.toCharArray(), 0, value.length());
            handler.endElement("", "anchor", "");
        }
       
        // Optimization flags
        switch (storageMode)
        {
            case UNKNOWN:
                value = "unknown";
                break;
            case DISCARDED:
                value = "discarded";
                break;
            case POTENTIAL_STRUCT:
                value = "potential struct";
                break;
            case STRUCT_ONLY:
                value = "struct only";
                break;
            case DATA_ONLY:
                value = "data only";
                break;
            case DATA_AND_STRUCT:
                value = "data & struct";
                break;
        }
        atts.addAttribute("", "storageMode", "",  "CDATA", value);
        atts.addAttribute("", "childrenOrderEnforced", "",  "CDATA", String.valueOf(childrenOrderEnforced));
        atts.addAttribute("", "positionEnforced", "",  "CDATA", String.valueOf(isPositionEnforced()));
        atts.addAttribute("", "nilImplicit", "",  "CDATA", String.valueOf(isNilImplicit()));
        atts.addAttribute("", "hasDefaultMappingInScope", "",  "CDATA", String.valueOf(hasDefaultMappingInScope));
        atts.addAttribute("", "mixed", "",  "CDATA", String.valueOf(isMixed()));
        atts.addAttribute("", "defaultAllowed", "",  "CDATA", String.valueOf(isDefaultAllowed()));
        handler.startElement("", "flags", "", atts);
        atts.clear();
        handler.endElement("", "flags", "");
        
        handler.startElement("", "readTable", "", atts);
        if (readTableMapping != null)
        {
            value = readTableMapping.getTableMapping().getTableName();
            handler.characters(value.toCharArray(), 0, value.length());
        }
        handler.endElement("", "readTable", "");
        
        handler.startElement("", "readColumn", "", atts);
        if (readColumnMapping != null)
        {
            value = readColumnMapping.getTableName() + "." + readColumnMapping.getColumnName();
            handler.characters(value.toCharArray(), 0, value.length());
        }
        handler.endElement("", "readColumn", "");
    
        handler.startElement("", "mapping", "", atts);
        mapping.toXML(handler);
        handler.endElement("", "mapping", "");
        
        handler.endElement("", "data", "");
    }
    
    /** Return true if the element can be missing if its parent exists.
     * @return true if optional
     */
    public boolean isOptional()
    {
        return occurrenceCount[0] == 0;
    }
    
    /** Return true if the element can only appear one time in its parent.
     * @return true if monovaluated
     */
    public boolean isMonoValuated()
    {
        return occurrenceCount[1] == 1;
    }
    
    /** Return true if the element can appear several times in its parent.
     * @return true if multivaluated
     */
    public boolean isMultiValuated()
    {
        return occurrenceCount[1] > 1;
    }
    
    public void invalidateCache()
    {
        modelSubPaths = null;
        subPaths = null;
        if (parent != null)
            ((PathNode)parent).invalidateCache();
    }
    
    /** Return a list of the path nodes with an OID contained in the subtree.
     * @return null if the path is of root type.
     */
    public List getModelSubPaths()
    {
        // TO SEE : could precompile it and update when pathset is modified
        if (modelSubPaths == null)
            browseSubPaths();
        
        return modelSubPaths;
    }
    
    /** Return a list of the path nodes with or without an OID contained in
     * the subtree.
     * @return null if the path is of root type.
     */
    public List getSubPaths()
    {
        // TO SEE : could precompile it and update when pathset is modified
        if (subPaths == null)
            browseSubPaths();
        
        return subPaths;
    }
    
    private void browseSubPaths()
    {
        // TO SEE : could precompile it and update when pathset is modified. Better : use range for paths too
        if (getType() == NodeKind.NONE)
            return;
        else if (isLeaf())
        {
            subPaths = Collections.singletonList(this);
            if (isPersistent())
                modelSubPaths = subPaths;
            else
                modelSubPaths = Collections.EMPTY_LIST;
        }
        else
        {
            Iterator it = new XTreeIterator(this);
            PathNode node;
            modelSubPaths = new ArrayList();
            subPaths = new ArrayList();
            
            // process first itself
//            if (isPersistent())
//                modelSubPaths.add(this);
//            subPaths.add(this);
            
            while (it.hasNext())
            {
                node = (PathNode)it.next();
                if (node.isPersistent())
                    modelSubPaths.add(node);
                subPaths.add(node);
            }
        }
    }
    
    /** get column mapping used for reconstruction and queries (that may be
     * different than the user-defined ones  */
    public ColumnMapping getReadColumnMapping()
    {
        return readColumnMapping;
    }
    
    /** get Table mapping used for reconstruction and queries (that may be
     * different than the user-defined ones  */
    public CollectionMappingInfo getReadTableMapping()
    {
        return readTableMapping;
    }
    
    /** Shortcut to the schema content type.
     * @return true if elemnt is mixed.
     */
    public boolean isMixed()
    {
        return mixed;
    }
    
    /** get column mapping used for text nodes. This method returns a value only
     * if {@link isMixed()} is true
     */
    public ColumnMapping getTextColumnMapping()
    {
        return textColumnMapping;
    }
    
    /** get table mapping used for text nodes. This method returns a value only
     * if {@link isMixed()} is true
     */
    public CollectionMappingInfo getTextTableMapping()
    {
        return textTableMapping;
    }
    
    /** get table mapping information customized for this pathset collection.
     */
    public CollectionMappingInfo getTableMappingInfo(TableMapping mapping)
    {
        return (CollectionMappingInfo)((PathSet)tree).getCollection().getMappingStatements(mapping.getTableIndex());
    }
    
}
