/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.mapper.metadata;

import java.sql.*;
import java.util.ArrayList;

import org.xquark.mapper.RepositoryException;
import org.xquark.mapper.dbms.*;
import org.xquark.mapper.mapping.TableMapping;
import org.xquark.xpath.*;

/** Shared memory data structure used to store paths of an XML collection
 * and provide a direct access to mapping information.
 *
 * <p><B>WARNING !</B> Path are never removed till a reconstruction is made
 * from the documents stored in the collection. Thus, some paths in the path set
 * may not correspond to stored documents.</p>
 * <p><B>WARNING !</B> This object keeps a collection.getMetadataConnection() till it is closed.</p>
 * <p>It uses PathNode object for storing a path ID for every node of the
 * collection metadata.</p>
 */
public class PathSet extends XTree implements RepositoryConstants
{
    private static final String RCSRevision = "$Revision: 1.2 $";
    private static final String RCSName     = "$Name: XQ1_1_1 $";
    // TO DO : CHECK OVERFLOW
    
    /** The collection metadata ewning this pathset
     */
    private CollectionMetadata collection;
    
    private Sequence pathSeq;
    
    private String selectDiffStmt;
    private String insertStmt;
    private String checkCountStmt;
    private boolean seqInitialized = false;
    
    /* synchonization flag */
    private short mawLoadedPID = 0; // initialized to 0 in order to load all at the beginning
    
    /** Hashtable used to provide rapid access from ID to a path
     */
    protected ArrayList pathIndex = new ArrayList();    // path ID is index (0 is reserved for ROOT)
    // missing paths will be represented as nulls
    
    /** JDBC data used for storing rows in the database with batch style.
     * IMPORTANT  : when storing only one batch is used for all threads.
     */
    private PreparedStatement pInsertStmt = null;
    private PreparedStatement pSelectStmt = null;
    
    private String tableName;
    private boolean isNew = false;
	private boolean loading = false;
	
    /** Constructor.
     * <p>Path set is loaded from the underlying database.</p>
     * @param collection the repository using the dictionary and providing database access.
     * @exception java.sql.SQLException if a database-access error occurs.
     */
    public PathSet(CollectionMetadata collection)
    {
        // Nota : pathIndex is instantiated by super() that calls AllocateNode
        this.collection = collection;
        pathSeq = new Sequence(
                                collection.getTableInfo(TableSpec.TYPE_PATH_SEQ),
                                (short)1
                                );
        initStatements();
        
        try
        {
            if (isEmpty())   // in order to know if flags must be calculated or read from the database
                isNew = true;
        }
        catch (Exception e)
        {
            throw new RuntimeException("Error while loading the path set" + e.getMessage());
        }
    }
    
    public synchronized boolean isNew()
    {
        return isNew;
    }
    
    public void registerMapping(TableMapping tm, PathNode node)
    {
    	collection.registerMapping(tm, node);
    }
    ////////////////////////////////////////////////////////////////////
    //                          ACCESSORS
    ////////////////////////////////////////////////////////////////////

    public CollectionMetadata getCollection()
    {
        return collection;
    }
    
    public synchronized PathNode get(short id)
    throws RepositoryException
    {
        if ((id > mawLoadedPID) && !loading) // refresh for document reconstruction using paths created by other instances
            refresh();
            
        return (PathNode)pathIndex.get(id);
    }
    
    /**
     * Returns the <code>PathNode</code> object accessible by the specified
     * <code>LocationExpression</code>.
     *
     * @param e the access path
     * @return the correspoding PathNode
     *
     */
    public synchronized PathMetadata getPathNode(PathExpr e)
    {
        return (PathNode) super.getNode(e);
    }
    
    ////////////////////////////////////////////////////////////////////
    //                          PUBLIC
    ////////////////////////////////////////////////////////////////////

    /* Rebuild the actual collection path set from documents stored in the
     * database. Path that are not found are removed.
     */
    public synchronized void refresh()
    throws RepositoryException
    {
        ResultSet rs = null;
        try
        {
            if (!loading)
            {
                loading = true;
            	if (standsBy())
            		initPreparedStatements(collection.getMetadataConnection());
                pSelectStmt.setShort(1, mawLoadedPID);
                rs = pSelectStmt.executeQuery();
                while(rs.next())
                    loadNode(
                            get(rs.getShort(2)),
                            rs.getShort(1),
                            rs.getString(3),
                            rs.getString(4),
                            rs.getByte(5),
                            rs.getByte(6)
                            );
                mawLoadedPID = (short)(pathIndex.size() - 1);
            }
        }
        catch (SQLException e)
        {
            throw new RepositoryException(RepositoryException.DB_ERROR,
            "JDBC error while refreshing path set metadata cache.", e);
        }
        finally
        {
            loading = false;
            try
            {
                if (rs != null)
                    rs.close();
            }
            catch (SQLException e)
            {
                throw new RepositoryException(RepositoryException.DB_ERROR,
                "JDBC error while refreshing path set metadata cache.", e);
            }
        }
    }
    
    public MetadataIterator createIterator()
    {
        return new MetadataIterator(collection.getPathSetFactory());
    }
    
    public MetadataIterator createIterator(PathNode path)
    {
        return new MetadataIterator(collection.getPathSetFactory(), path);
    }
    
    public MetadataIterator createIterator(short pid)
    throws RepositoryException
    {
        return new MetadataIterator(collection.getPathSetFactory(), (PathNode)get(pid));
    }
    
    ////////////////////////////////////////////////////////////////////
    //                          PACKAGE PRIVATE
    ////////////////////////////////////////////////////////////////////
	void releaseJDBCRessources()
	{
		try
		{
			if (pInsertStmt != null)
				pInsertStmt.close();
			if (pSelectStmt != null)
				pSelectStmt.close();
		}
		catch (SQLException e)
		{
			// no op : to see...
		}
		finally
		{
			pInsertStmt = null;
			pSelectStmt = null;
		}
	}

    void initRoot()
    {
        pathIndex.add(ROOT_PATH_ID, getRoot());
    }
    
    /** free resources used by this object.
     */
    void close()
    {
        if (collection != null)
        {
            clear();
        	releaseJDBCRessources();
            pathIndex = null;
            collection = null;
        }
    }
    
    /** Empty the path set (in memory).
     */
    void clear()
    {
        pathIndex.clear();
    }
    
    /**
     * Called by PathSetFactory when new paths added. Synchronized because of "get" methods.
     * @return if no collision (if false: do not create node but refresh)
     */
    synchronized void allocateID(PathNode node)
    throws RepositoryException
    {
        if (!loading && (node.getType() != NodeKind.NONE))   // allocate one and try to store // already set by default
            registerPath((short)pathSeq.nextValue(collection.getMetadataConnection()), node);
    }
    
    synchronized void freeID(PathNode node)
    {
        pathIndex.remove(node.getPathID());
    }
    
    boolean isLoading()
    {
        return loading;
    }

    void store(PathNode p)
    throws java.sql.SQLException
    {
        // preparation of PreparedStatement for insertion
        try
        {
        	AbstractConnection metaConn = collection.getMetadataConnection();
            synchronized(metaConn) // SHOULD add it to any use of metadata performing commit (reading?)
            {
                if (standsBy())
                    initPreparedStatements(metaConn);
                pInsertStmt.setShort(1, collection.getCollectionID());
                pInsertStmt.setShort(2, p.getPathID());
                if (p.getParent() == null)
                    pInsertStmt.setShort(3, ROOT_PATH_ID);
                else
                    pInsertStmt.setShort(3, ((PathNode)p.getParent()).getPathID());
                pInsertStmt.setString(4, p.getNamespace());
                pInsertStmt.setString(5, p.getLocalName());
                pInsertStmt.setByte(6, p.getType());
                pInsertStmt.setByte(7, p.getStorageMode());
                pInsertStmt.executeUpdate();
            }
        }
        catch (SQLException e)
        {
            if (pInsertStmt != null)
                pInsertStmt.close();
            pInsertStmt = null;
            throw e;
        }
    }
    
    ////////////////////////////////////////////////////////////////////
    //                          PRIVATE
    ////////////////////////////////////////////////////////////////////
    private void initStatements()
    {
        // Documents table statements
        TableInfo table = collection.getRepository().getTableInfo(TableSpec.TYPE_PATHS);
        ColumnInfo[] columns = table.getColumns();
        
        StringBuffer sql = new StringBuffer();
        sql.append("SELECT COUNT(*) FROM ");
        sql.append(table.getName());
        sql.append(" WHERE CID=");
        sql.append(collection.getCollectionID());
        checkCountStmt = sql.toString();
        
        sql.setLength(0);
        sql.append("SELECT PID,PPID,NS,NAME,TYPE,SMF FROM ");
        sql.append(table.getName());
        sql.append(" WHERE CID=");
        sql.append(collection.getCollectionID());
        sql.append(" AND PID>? ORDER BY PID");
        selectDiffStmt = sql.toString();
        
        insertStmt = table.getInsertStatement();
    }
    
	private boolean standsBy()
	{
		return pInsertStmt == null;
	}

	private void initPreparedStatements(AbstractConnection metaConn) throws SQLException
	{
		pInsertStmt = metaConn.getConnection().prepareStatement(insertStmt);
		pSelectStmt = metaConn.getConnection().prepareStatement(selectDiffStmt);
	}

    private boolean isEmpty()
    throws java.sql.SQLException
    {
        Statement stmt = null;
        ResultSet rs = null;
        boolean isEmpty;
        try
        {
            stmt = collection.getMetadataConnection().getConnection().createStatement();
            rs = stmt.executeQuery(checkCountStmt);
            rs.next();
            isEmpty = (rs.getInt(1) == 0);
        }
        finally
        {
            if (rs != null)
                rs.close();
            if (stmt != null)
                stmt.close();
        }
        return isEmpty;
    }
    
    /** Used by loader */
    /** called by load(). Set the pathID for node created by the mapping
     * duplication.
     */
    private XTreeNode loadNode(PathNode parent, short pid,
    String namespace, String localName, byte type, byte storageMode)
    {
        // calls allocate node if the node was nos part of the mapping
        PathNode node = (PathNode)collection.getPathSetFactory().createNamedNodeIfNotExist(
                        parent,
                        namespace,
                        localName,
                        type
                        );
        node.setStorageModeFlag(storageMode);
        registerPath(pid, node);
        return node;
    }
    
    private void registerPath(short pid, PathNode node)
    {
        node.setPathID(pid);
        while (node.getPathID() > pathIndex.size())
            pathIndex.add(null);   // lost pids
        
        pathIndex.add(node.getPathID(), node);
    }
    

}
