/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.mapper.metadata;


import java.sql.*;
import java.util.*;

import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xquark.mapper.*;
import org.xquark.mapper.dbms.*;
import org.xquark.mapper.mapping.Loader;
import org.xquark.mapper.mapping.MappingFactory;
import org.xquark.mapper.mapping.RepositoryMapping;
import org.xquark.mapper.storage.PersistentCollectionInfo;
import org.xquark.mapper.storage.PersistentRepositoryInfo;
import org.xquark.mapper.storage.RepositoryReader;
import org.xquark.mapper.util.*;
import org.xquark.schema.SchemaManager;
import org.xquark.xml.xdbc.XMLDBCException;
import org.xquark.xml.xdbc.XMLDBCNotRecognizedException;
import org.xquark.xpath.XTree;
import org.xquark.xquery.ModuleManager;
import org.xquark.xquery.metadata.resolver.SourceMetadata;


/**
 * Provide functions for manipulating repositories such as creating,
 * opening, deleting...
 * Repository manages a list of repositories for each database user. At
 * the present time a database user can only access his own repositories.
 * @see org.xquark.mapper.AccessManager
  */
public final class Repository implements RepositoryConstants, _Repository
{
    private static final String RCSRevision = "$Revision: 1.5 $";
    private static final String RCSName     = "$Name: XQ1_1_1 $";

    /** The repsoitory configuration information read from the Repository table */
    private RepositoryInfo config;

    /** A collection for opened XML Collections */
    private CacheManager collectionManager = new CacheManager(RepositoryProperties.getIntProperty(CONF_COLLECTION_METADATA_CACHE));

    /** A collection for loaded mappings (in fact {mappingID, MappingFactory} */
    private CounterMap loadedMappings = new CounterMap();

    /** The class managing loaded schemas */
    private SchemaManager schemaManager;
    
    /** The system repositories used to store mapping files and schemas.
     *  Only opened when a repository is opened or created.
     */
    private CollectionMetadata schemaCollection, mappingCollection;

    /** JDBC connection reserved for metadata operation, not submitted
     * to storage rollback.
     */
    private AbstractConnection metadataConnection;
    private _RepositoryConnection metadataRepConnection;

    /** The name of the user connected to this rdbms */
    private String user;
    private String URL;

    /** A map containing TableInfo for repository tables.
     */
    private Map repositoryTables;
    
    /** Object containing user table Information for user tables.
     */
    private UserTableManager userTables;
    
    /** XQuery parser interfaces implementation */
    private Collection sourceCollection;
    private Map proxyCollections = null;
    
    /** SQL Statements. */
    private String checkSchemaStmt;
    private String checkMappingStmt;
    private String deletePathSetStmt;
    private String updateColDescStmt;
    
    /** Create a Repository object using the Connection object passed in parameter.
     * @param conn a java.sql.Connection object
     * @throws RepositoryException application exception.
     */
    public Repository(AbstractConnection connection) throws XMLDBCException
    {
        setConnection(connection);
        sourceCollection = Collections.singletonList(this);
        
        // TO DO : define a user API for Repository conf !!!
        /* boot */
        config = new RepositoryInfo(user, DEFAULT_CID_SIZE);
        init(true);    // with the current default configuration (needed for boot)
        
        try
        {
            if (metadataConnection.checkTable(LEGACY_REPOSITORY_CHECKER))
                throw new RepositoryException(RepositoryException.DB_CONSISTENCY_ERROR,
                "A legacy repository has been detected an no migration tool is available (beta stage).\nPlease remove the old repository before using this version of the runtime. This can be performed with the old runtime (with XMLManager 'del rep' command) or by dropping all dbms tables beginning with the 'XQUARK$' prefix.");
            
            if (!metadataConnection.checkTable(getTableInfo(TableSpec.TYPE_REPOSITORY)))   // check if initialization was performed
                initializeRepository();
            /* loading repository configuration info */
            PersistentRepositoryInfo configReader = new PersistentRepositoryInfo(
            getTableInfo(TableSpec.TYPE_REPOSITORY), metadataConnection);
            try
            {
                configReader.fetchRow();
            }
            catch (SQLException e)
            {
                throw new RepositoryException(RepositoryException.DB_ERROR,
                "JDBC error while reading repository configuration.", e);
            }
            config = configReader.getRepositoryInfo();
            
            /* Repository relational schema compatibility check */
            if (!config.getVersion().equals(TableSpecLoader.REPOSITORY_DATA_VERSION))
               throw new RepositoryException(RepositoryException.CONSISTENCY_ERROR,
                "This existing repository is not compatible with the current software version. A migration must be considered.");
            
            init(false); // init schemaManager, read repository config
        }
        catch (SQLException e) {
            throw new RepositoryException(RepositoryException.DB_ERROR,
            "JDBC error while performing repository initialization.", e);
        }
    }
    
    ////////////////////////////////////////////////////////////////////
    //                          ACCESSORS
    ////////////////////////////////////////////////////////////////////
    private void setConnection(AbstractConnection connection)
    throws RepositoryException
    {
        try
        {
            restoreConnection(connection);
            DatabaseMetaData dbMeta = metadataConnection.getConnection().getMetaData();
            user = dbMeta.getUserName().toUpperCase();
            URL = dbMeta.getURL();
        }
        catch (SQLException e)
        {
            throw new RepositoryException(RepositoryException.DB_ERROR,
            "JDBC error while initialization a connection for metadata.", e);
        }
        
    }
    
    public boolean standsBy()
    {
        return (metadataConnection == null);
    }
    
    public boolean isAlive()
    {
        return (schemaManager != null);
    }
    
    public TableInfo getTableInfo(byte type)
    {
        return (TableInfo)repositoryTables.get(new Byte(type));
    }
    
    public UserTableManager getTableManager()
    {
        return userTables;
    }
    
    public RepositoryInfo getInfo()
    {
        return config;
    }
    
    /** Open collection metadata without guarantee it is not released. Consequently,
     * metadata object collection may be performed at every call.
     */
    public synchronized _Collection getCollection(String name) throws XMLDBCException
    {
        if (name == null)
            throw new RepositoryException(RepositoryException.ILLEGAL_EXPRESSION, "All method parameters are mandatory and must not be null.");
        
        name=name.toUpperCase();
        
        // check if not already opened
        CollectionMetadata collection = (CollectionMetadata)collectionManager.get(name);
        if (collection == null)
        {
            collection = createCollectionMetadata(name);
            collectionManager.put(name, collection);
        }
        return collection;
    }
  
    public String getName()
    {
        return config.getName();
    }
    
    public String getVersion()
    {
        return config.getVersion();
    }
    
    /** Returns the dbms user name used by this object.
     * @return the user name supplied at opening time.
     */
    public String getUserName()
    {
        return user;
    }
    
    /** Returns the JDBC string used by this object.
     * @return the JDBC string supplied at opening time.
     */
    public String getURL()
    {
        return URL;
    }
    
    public String toString() 
    {
        return RepositoryDriver.REPOSITORY_URL_PREFIX + URL + ", " + user;
    }

    /**
     * Return the database repository list.
     *
     * @return repository names in a string List.
     * @throws RepositoryException application exception.
     */
    public List getCollectionList() throws XMLDBCException
    {
        ArrayList names = new ArrayList();
        AbstractConnection connection = null;
        PersistentCollectionInfo wConfig = null;
        try {
            connection = metadataConnection;
            wConfig = new PersistentCollectionInfo(
                    getTableInfo(TableSpec.TYPE_COLLECTIONS), connection);

            wConfig.retrieveDataSet(CollectionInfo.USER); // system repositories are not fetched
            while (wConfig.fetchNextRow())
            {
                names.add(wConfig.getCollectionName());
            }
        }
        catch (SQLException e) {
            throw new RepositoryException(RepositoryException.DB_ERROR, "JDBC error while getting repository list.", e);
        }
        finally {
            if (wConfig != null)
                wConfig.close();
        }

        return names;
    }

    /**
     * Get the system Collection object for this database object,
     * containing schema files. Actually, this repository is
     * opened the the first time this method is called
     *
     * @return a CollectionMetadata object.
     */
    public synchronized CollectionMetadata getSchemaCollection()
    throws XMLDBCException
    {
        return schemaCollection;
    }
    
    public SchemaManager getSchemaManager()
    {
        return schemaManager;
    }
    
	public ModuleManager getModuleManager()
	{
		//TODO to be done!!
		return null;
	}
    	
    public AbstractConnection getMetadataConnection()
    {
        return metadataConnection;
    }
    
    /**
     * Get the system Collection object for this database object,
     * containing mapping files. Actually, this repository is
     * opened the the first time this method is called
     *
     * @return a CollectionMetadata object.
     */
    public synchronized CollectionMetadata getMappingCollection()
    throws XMLDBCException
    {
        return mappingCollection;
    }

    public MappingFactory getMapping(String MappingID) throws RepositoryException
    {
        MappingFactory ret = (MappingFactory)loadedMappings.checkOut(MappingID);
        if (ret == null)    // if not already opened
        {
            // Load mapping
            ret = loadMapping(MappingID, false);
            // Put the mapping factory in the hastable of loaded mappings
            loadedMappings.put(MappingID, ret);
        }
        return ret;
    }
    
    ////////////////////////////////////////////////////////////////////
    //                          PUBLIC
    ////////////////////////////////////////////////////////////////////
    public synchronized void resetRepository() throws XMLDBCException
    {
        deleteRepository();
        initializeRepository();
        init(false);
    }
    
    /**
     * Free all resources. For instance, close JDBC connection and repositories.
     * <B>Any subsequent call to a function of the object will raise an
     * exception.</B>
     * @throws RepositoryException application exception.
     */
    public synchronized void close()
    throws XMLDBCException
    {
        // free collection metadata opened for queries is useless since we'return closing (waiting for LRU)
        withdrawConnection();
        reset();

        repositoryTables = null;
        collectionManager = null;
        loadedMappings = null;
    }

    public void withdrawConnection() throws XMLDBCException
    {
        metadataRepConnection.close();
        metadataRepConnection = null;
        metadataConnection = null;
        // release collection metadata JDBC Ressources
        schemaCollection.releaseJDBCRessources();
        mappingCollection.releaseJDBCRessources();
        
        Iterator it = collectionManager.iterator();
        while (it.hasNext())
        	((CollectionMetadata)it.next()).releaseJDBCRessources();
    }
    
    public void restoreConnection(AbstractConnection connection)
    throws SQLException, RepositoryException
    {
        metadataRepConnection = new RepositoryConnectionImpl(
        this,
        connection,
        new DestructionToken()
        {
            public void destruct(Object o)
            {};
        });
        
        metadataConnection = connection;
        metadataConnection.getConnection().setAutoCommit(true);
        // NOTE prepared statements will be automatically recreated from metadata connection
    }
    
    private void reset() throws RepositoryException
    {
        if (isAlive())
        {
            collectionManager.clear();
            loadedMappings.clear();
            schemaManager = null;
            userTables = null;
            schemaCollection = null;
            mappingCollection = null;
            proxyCollections = null;
        }
    }
    
    /** Remove <b>XQuark Repository</b> configuration from the underlying database,
     * but not the repository tables. All repositories must have been previously
     * removed from the configuration table before this function is able to work.
     *
     * @throws RepositoryException application exception.
     */
    public synchronized void deleteConfiguration() throws XMLDBCException
    {
        PersistentCollectionInfo wConfig = null;
        try {
            wConfig = new PersistentCollectionInfo(getTableInfo(TableSpec.TYPE_COLLECTIONS),
                metadataConnection);

            wConfig.retrieveDataSet(CollectionInfo.USER);
            if(wConfig.fetchNextRow())
                throw new RepositoryException(RepositoryException.NOT_ALLOWED, "All collections must be deleted before using this function.");
            deleteRepository();
        }
        catch (SQLException e) {
            throw new RepositoryException(RepositoryException.DB_ERROR, "JDBC error while deleting database.", e);
        }
        finally {
            wConfig.close();
        }
    }

    /** Remove <b>XQuark Repository</b> configuration from the underlying database
     * and all the repositories referenced. This method does not stops if a failure
     * occurs while deleting a repository.
     * <B>Don't use this method unless you are sure not to use XQuark Repository
     * any more in the underlying database, because all data will be lost.</B>
     * @throws RepositoryException application exception.
     * @return the number of XML collections deleted.
     */
    public synchronized int deleteAllCollections() throws XMLDBCException
    {
        List collections = getCollectionList();
        Iterator it = collections.iterator();
        
        while (it.hasNext())
            deleteCollection((String)it.next());
        
        return collections.size();
    }
    
    public synchronized void deleteRepository()
    throws XMLDBCException
    {
        // release system collections
        releaseCollectionMetadata(RepositoryConnection.MAPPING_COLLECTION);
        releaseCollectionMetadata(RepositoryConnection.SCHEMA_COLLECTION);
        
        // check no collection metadata is currently in use (lightweith
        // RepositoryConnection object should close them first)
        if (collectionManager.used())
            throw new RepositoryException(RepositoryException.NOT_ALLOWED,
                "All collections must be closed before repository can be deleted.");
        
            // Delete all repository tables (don't remove user data)
        try
        {
            dropRepositoryTables();
        }
        catch (SQLException e)
        {
            throw new RepositoryException(RepositoryException.DB_ERROR,
                "JDBC error while destroying repository tables.", e);
        }
        try
        {
            metadataConnection.onDeleteRepository(this);
        }
        catch (SQLException e)
        {
            throw new RepositoryException(RepositoryException.DB_ERROR,
                "JDBC error while performing dbms-specific cleaning tasks.", e);
        }
        // reinit schema manager and cache
        reset();
    }
    
    /** Open a repository owned by the connection user, identified by the
     * <code>name</code> parameter. If the repository has been opened
     * previously, the same object is returned. A counter is kept on the object
     * so it won't be destroyed before beeing released.
     * @param name the name of the repository.
     * @return the repository opened.
     * @throws RepositoryException Application exception. For instance if
     * the specified repository does not exist.
     */
    public synchronized CollectionMetadata openCollection(String name)
    throws XMLDBCException
    {
        if (name == null)
        throw new RepositoryException(RepositoryException.ILLEGAL_EXPRESSION, "All method parameters are mandatory and must not be null.");

        name=name.toUpperCase();

        // check if not already opened
        CollectionMetadata collection = (CollectionMetadata)collectionManager.checkOut(name);
        if (collection == null)
        {
            collection = createCollectionMetadata(name);
            // put it in the list
            collectionManager.use(name, collection);
        }
//        else	// increment mapping use counter
//        {
//        	String MappingID = collection.getInfo().getMappingID();
//        	MappingFactory ret = (MappingFactory)loadedMappings.checkOut(MappingID);
//        	if (ret == null)    // if not already opened
//        	{
//        		// Load mapping
//        		ret = collection.getMappingFactory();
//        		// Put the mapping factory in the hastable of loaded mappings
//        		loadedMappings.put(MappingID, ret);
//        	}
//
//        }
        return collection;
    }

    private CollectionMetadata createCollectionMetadata(String name)
    throws XMLDBCException
    {
        CollectionMetadata collection = null;
        
        PersistentCollectionInfo wConfig = new PersistentCollectionInfo(
            getTableInfo(TableSpec.TYPE_COLLECTIONS), metadataConnection);

        // searching for the collection in config table
        try {
            if (wConfig.fetchRow(name))
            {
                // Create a collection object
                switch(wConfig.getCollectionInfo().getType())
                {
                    case CollectionInfo.SYSTEM:
                    case CollectionInfo.USER:
                        collection = new CollectionMetadata(
                                                name,
                                                wConfig.getCollectionID(),
                                                wConfig.getCollectionInfo(),
                                                this,
                                                getMapping(wConfig.getCollectionInfo().getMappingID())
                                                );
                        break;
                    default:
                        throw new RepositoryException(RepositoryException.ILLEGAL_EXPRESSION, 
                                    "This type of collection (" + wConfig.getCollectionInfo().getType() + 
                                    ") stored in the XQUARK$_CONFIG table is unknown.");
                }
            }
            else
                throw new RepositoryException(RepositoryException.NOT_EXISTS, "XMLCollection information not found for " + name + " collection.");
        }
        catch (SQLException e) {
            throw new RepositoryException(RepositoryException.DB_ERROR, "Error while fetching information for " + name + " collection.", e);
        }
        
        return collection;
    }
    
    public MappingFactory loadDefaultMapping()
    throws RepositoryException
    {
        MappingFactory ret =  new MappingFactory(this, schemaManager, metadataRepConnection);
        ret.createTree();
        return ret;
    }
    
    public MappingFactory loadMapping(String MappingID, boolean generate)
    throws RepositoryException
    {
        RepositoryMapping mapping;
        MappingFactory ret =  new MappingFactory(this, schemaManager, metadataRepConnection);
        mapping = (RepositoryMapping)ret.createTree();
        // load the mapping from the system repository (and save it if no saved yet)
        if (MappingID!= null && MappingID.length() > 0)
        {
            RepositoryReader mappingReader = null;
            try
            {
                mappingReader = metadataRepConnection.getMappingReader();
                Loader loader = new Loader
                (
                mappingReader,
                metadataRepConnection.getConnection(),
                schemaManager,
                generate
                );
                InputSource source = new InputSource(MappingID);
                mapping = loader.load(source, ret, false);
            }
            catch (Exception e)
            {
                throw new RepositoryException(RepositoryException.NOT_EXISTS, "Error while loading mapping or schema.", e);
            }
            finally
            {
                if (mappingReader != null)
                    mappingReader.close();
            }
            // Verify (only for Repository) that multiple join tables
            // specification do not exist for a same user table
            mapping.checkRedundantTables();
        }
        return ret;
    }
    
    /**
     * @return a list of collection names
     */
    public synchronized List getCollectionsUsingMapping(String mappingID)
    throws XMLDBCException
    {
        // query execution
        ArrayList collections = new ArrayList();
        PreparedStatement pStmt = null;
        ResultSet rs = null;
        try
        {
            pStmt = metadataConnection.getConnection().prepareStatement(checkMappingStmt);
            pStmt.setString(1, mappingID);
            rs = pStmt.executeQuery();
            
            while (rs.next())
                collections.add(rs.getString(1));
        }
        catch (SQLException e)
        {
            throw new RepositoryException(RepositoryException.DB_ERROR,
            "JDBC error while collecting information on namespaces used in the repository.", e);
        }
        finally
        {
            try
            {
                if (rs != null)
                    rs.close();
                if (pStmt != null)
                    pStmt.close();
            }
            catch (SQLException e)
            {
                // no op
            }
        }
        
        return collections;
    }

    /**
     * @return a list of collection names
     */
    public synchronized List getCollectionsUsingNamespace(String targetNamespace)
    throws XMLDBCException
    {
        // query execution
        ArrayList collections = new ArrayList();
        PreparedStatement pStmt = null;
        ResultSet rs = null;
        try
        {
            pStmt = metadataConnection.getConnection().prepareStatement(checkSchemaStmt);
            pStmt.setString(1, targetNamespace);
            rs = pStmt.executeQuery();
            
            ListIterator it = collections.listIterator();
            
            while (rs.next())
                collections.add(rs.getString(1));
        }
        catch (SQLException e)
        {
            throw new RepositoryException(RepositoryException.DB_ERROR,
            "JDBC error while collecting information on namespaces used in the repository.", e);
        }
        finally
        {
            try
            {
                if (rs != null)
                    rs.close();
                if (pStmt != null)
                    pStmt.close();
            }
            catch (SQLException e)
            {
                // no op
            }
        }
        
        return collections;
    }
    
    /** Create a new repository which features are given by the CollectionInfo
     * object. If a repository with the same ID (name, owner) already exists,
     * the existing repository is returned without any warning.
     * Processes database object initialization for the repository.
     * @param name the name of the repository.
     * @param config a CollectionInfo object allowing to set the features of
     * the new repository.
     * @return the newly created repository
     * @throws RepositoryException application exception.
     */
    public synchronized void createCollection(String name, CollectionInfo config, MappingFactory mappingFactory)
    throws XMLDBCException
    {
        name = name.toUpperCase();
        
        // Check if a Collection of same name exists.
        if (getCollectionList().indexOf(name) >= 0)
            throw new RepositoryException(RepositoryException.ALREADY_EXISTS, "This collection seems to exist.");
        
        if ((name == null) || (config == null))
            throw new RepositoryException(RepositoryException.ILLEGAL_EXPRESSION, "All method parameters are mandatory and must not be null.");
        
        PersistentCollectionInfo configRow = new PersistentCollectionInfo
                                        (
                                        getTableInfo(TableSpec.TYPE_COLLECTIONS),
                                        name,
                                        config,
                                        getInfo().getColIDMaxValue(),
                                        metadataConnection
                                        );
        CollectionMetadata.initializeCollection
                                        (
                                        (RepositoryMapping)mappingFactory.getTree(),
                                        configRow,
                                        this,
                                        metadataConnection
                                        );
    }
    
    /** Remove completely a repository from the base.
     * <B>The repository is supposed not to be in use.</B>
     * <p>Attempts is made to delete repository tables even if the repository
     * is not present in the configuration table.</p>
     * @param name repository name
     * @throws RepositoryException application exception
     */
    public synchronized void deleteCollection(String name) 
    throws XMLDBCException
    {
        if (name == null)
        throw new RepositoryException(RepositoryException.ILLEGAL_EXPRESSION, "All method parameters are mandatory and must not be null.");

        if (collectionManager.used(name))
            throw new RepositoryException(RepositoryException.NOT_ALLOWED,
                "Collection must be closed before it can be deleted.");
            
        PersistentCollectionInfo wConfig = null;
        try {
            wConfig = new PersistentCollectionInfo(
                getTableInfo(TableSpec.TYPE_COLLECTIONS), metadataConnection);

            name=name.toUpperCase();

            // TO DO : CHECK REPOSITORYCONNECTIONS ARE WARNED OR PREVENT REMOVAL WHEN IN USE BY OTHER REPOSITORYCONNECTIONS
            
            // close collection if opened
            cleanCollection(name);

            // searching for the repository in config table
            if (wConfig.fetchRow(name))
            {
                // Create a repository object
                switch(wConfig.getCollectionInfo().getType())
                {
                    case CollectionInfo.SYSTEM:
                    case CollectionInfo.USER:
                        CollectionMetadata.dropCollectionTables(wConfig, config.getColIDSize(), metadataConnection);
                        break;
                    default:
                        throw new RepositoryException(RepositoryException.ILLEGAL_EXPRESSION, 
                                    "This type of repository (" + wConfig.getCollectionInfo().getType() + 
                                    ") stored in the XQUARK$_CONFIG table is unknown.");
                }
                
                // remove the path set
                metadataConnection.executeUpdate(deletePathSetStmt + wConfig.getCollectionID());
                
                // remove the entry from the config table
                try {
                    wConfig.deleteRow();
                }
                catch (SQLException e) {
                    throw new RepositoryException(RepositoryException.DB_ERROR, "Failed to delete repository from the configuration table.", e);
                }
            }
            else    // try to delete generic repository tables because table dropping may have previously failed
                CollectionMetadata.dropCollectionTables(wConfig, config.getColIDSize(), metadataConnection);
                
        }
        catch (SQLException e) {
            // NO ROLLBACK is attempted
            throw new RepositoryException(RepositoryException.DB_ERROR, "JDBC error while deleting repository.", e);
        }
        finally {
            if (wConfig != null)
                wConfig.close();
        }
    }
    /** Rename a repository.
     * <p>Proceed renaming of all repository-related database objects.</p>
     * <B>Result is unpredictable if the renaming occurs while repository
     * access is being done. It is recommended to stop all repository use.</B>
     * @param oldName current name of the repository
     * @param newName new name to set for the repository.
     * @throws RepositoryException application exception.
     */
    public synchronized void renameCollection(String oldName, String newName)
    throws XMLDBCException
    {
        if ((oldName == null) || (newName == null))
        throw new RepositoryException(RepositoryException.ILLEGAL_EXPRESSION, "All method parameters are mandatory and must not be null.");

        AbstractConnection connection = null;
        PersistentCollectionInfo wConfig = null;
        try {
            connection = metadataConnection;
            wConfig = new PersistentCollectionInfo(getTableInfo(TableSpec.TYPE_COLLECTIONS), connection);

            oldName=oldName.toUpperCase();
            newName=newName.toUpperCase();

            // searching for the repository in config table in order to change it
            if (wConfig.fetchRow(oldName))
            {
                // rename a repository object
                switch(wConfig.getCollectionInfo().getType())
                {
                    case CollectionInfo.SYSTEM:
                        throw new RepositoryException(RepositoryException.NOT_ALLOWED, 
                                    "Cannot rename a system XML collection.");
                    case CollectionInfo.USER:
                        wConfig.rename(oldName, newName);
                        break;
                    default:
                        throw new RepositoryException(RepositoryException.ILLEGAL_EXPRESSION, 
                                    "This type of repository (" + wConfig.getCollectionInfo().getType() + 
                                    ") stored in the XQUARK$_CONFIG table is unknown.");
                }
            }
            else
                throw new RepositoryException(RepositoryException.NOT_EXISTS, "XMLCollection information not found.");
            
            // change name in memory if  already opened
            CollectionMetadata colMeta = (CollectionMetadata)collectionManager.get(oldName);
            if (colMeta != null)
            {
                colMeta.setName(newName);
                collectionManager.setKey(oldName, newName);
            }

        }
        catch (SQLException e) {
            throw new RepositoryException(RepositoryException.DB_ERROR, "JDBC error while renaming repository.", e);
        }
        finally {
            if (wConfig != null)
                wConfig.close();
        }
    }

    public synchronized void releaseCollectionMetadata(String collectionName)
    throws RepositoryException
    {
        // remove from the list
        CollectionMetadata collection = (CollectionMetadata)collectionManager.release(collectionName.toUpperCase());
        
        // TO IMPROVE: this close should be performed when LRU discards it (using object should be warned it can be closed())
        // release mapping if any and if the collection is really removed (last use)
        if (collection != null)
        {
            try
            {
                if (collection.getMapping() != null)  // TO CHECK (un mapping quand mme ?)
                    loadedMappings.remove(collection.getInfo().getProperty(RepositoryConnection.MAPPING_ID_PROPERTY));
            }
            catch (XMLDBCNotRecognizedException e)
            {
                // no op : internal use
                throw new RuntimeException("Internal error: unknown feature or property.");
            }
//            collection.close();
            
            if (collectionName.equals(RepositoryConnection.MAPPING_COLLECTION))
                mappingCollection = null;
            else if (collectionName.equals(RepositoryConnection.SCHEMA_COLLECTION))
                schemaCollection = null;
        }
    }
    
    ///////////////////////////////////////////////////////////////////////////
    //                     MetadataAccess implementation
    ///////////////////////////////////////////////////////////////////////////
    
    /**
     *	@return a Map object {sourceName, Map {collectionName, CollectionMetadata}}
     */
    public Collection getSources()
    {
        if (proxyCollections == null)
            refreshCollectionList();
        return sourceCollection;
    }
    
    /** Provide access to a particular data source.
     * <B>Warning:</B> implementation may be case unsensitive.
     * @param sourceName the ID of the source
     * @return a SourceMetadata object
     */
    public SourceMetadata getSourceMetadata(String sourceName)
    {
        if (proxyCollections == null)
            refreshCollectionList();

        if (sourceName.equals(getSourceName()))
            return this;
        else
            return null;
    }
    
    private void refreshCollectionList()
    {
        String name;
        Iterator it;
        try
        {
            it = getCollectionList().iterator();
        }
        catch (XMLDBCException e)
        { // TO SEE: How to process it ?
            throw new RuntimeException("Error while getting collection list: " + e.getMessage());
        }
        
        // refresh list at every call
        if (proxyCollections == null)
            proxyCollections = new HashMap();
        else
            proxyCollections.clear();
        while (it.hasNext())
        {
            name = (String)it.next();
            proxyCollections.put(name, new ProxyCollection(name));
        }
        
    }
    
//    public List getCachedCollectionList()
//    {
//        return new ArrayList(proxyCollections.keySet());
//    }
//    
    ///////////////////////////////////////////////////////////////////////////
    //                     SourceMetadata implementation
    ///////////////////////////////////////////////////////////////////////////
    
    public boolean isCaseSensitive()
    {
        return false;
    }
    
	public boolean isStrict()
	{
		return false;
	}

    public String getSourceName()
    {
        return RepositoryConstants.REPOSITORY_SOURCE_NAME;
    }
    
    public Collection getCollectionsMetadata()
    {
        return proxyCollections.values();
    }
    
    public org.xquark.xquery.metadata.resolver.CollectionMetadata getCollectionMetadata(String collectionName)
    {
        return (org.xquark.xquery.metadata.resolver.CollectionMetadata)proxyCollections.get(collectionName.toUpperCase());
    }
    
    
	///////////////////////////////////////////////////////////////////////////
	//                     XPathResolver implementation
	///////////////////////////////////////////////////////////////////////////
//	public ArrayList resolveXPath(ArrayList xnodes, StepExpr step) {
//		return null;
//	}
//
//	public ArrayList resolveXPath(ArrayList xnodes, ArrayList steps)  {
//		return null;
//	}
//
//	public QType resolveQType(byte context, QType qtype, StepExpr step) {
//		return null;
//	}
//
//	public ArrayList getCollectionRootNodes(String collectionName) {
//		return null;
//	}
//
//	public XNode getDocumentRootNode(URI documentURI) {
//		return null;
//	}
//
//	public ArrayList getInputRootNodes() {
//		return null;
//	}

    ////////////////////////////////////////////////////////////////////
    //                          PACKAGE-PRIVATE
    ////////////////////////////////////////////////////////////////////
    void updateCollectionDescription(short cid, String desc)
    throws SQLException
    {
        PreparedStatement pStmt = null;
        try
        {
            pStmt = metadataConnection.getConnection().prepareStatement(updateColDescStmt);
            pStmt.setString(1, desc);
            pStmt.setShort(2, cid);
            pStmt.executeUpdate();
        }
        finally
        {
            if (pStmt != null)
                pStmt.close();
        }
    }
    
    ////////////////////////////////////////////////////////////////////
    //                          PRIVATE
    ////////////////////////////////////////////////////////////////////
    /** Initialize and configure the underlying database for managing XML
     * documents and other semi-structured data. This method needs to be
     * called only once per database. Subsequence raise an exception.
     * Config table, dictionary and system repositories are created.
     * @throws RepositoryException Application exception.
     */
    private void initializeRepository()
    throws XMLDBCException
    {
        try {
            /* create repository level tables and indexes */
            Iterator it = metadataConnection.createTables(repositoryTables.values()).iterator();
            
            /* sequence creation (step must be specified) */
            while (it.hasNext())
                new Sequence((TableInfo)it.next(), (short)1).create(metadataConnection);
           
            /* store repository config */
            PersistentRepositoryInfo infoSaver = new PersistentRepositoryInfo(
                config,
                getTableInfo(TableSpec.TYPE_REPOSITORY),
                metadataConnection
                );
            infoSaver.save();
            
            /* perform dbms-specific tasks */
            metadataConnection.onInitRepository(this);
        }
        catch (RepositoryException e) {
            try {
                dropRepositoryTables();
            }
            catch (SQLException ex) {
                // no op
            }
            throw e;
        }
        catch (SQLException e) {
            try {
                dropRepositoryTables();
            }
            catch (SQLException ex) {
            	throw new RepositoryException(RepositoryException.DB_ERROR, 
            	"JDBC error while repository initialization. Could not remove tables.", e);
            }
            throw new RepositoryException(RepositoryException.DB_ERROR, "JDBC error while repository initialization.", e);
        }
        
        /* Create "system" repository for storing mapping files and schemas */
        CollectionInfo config = new CollectionInfo(CollectionInfo.SYSTEM,
                            metadataConnection.getSystemCollectionDataLength());
        
        createCollection(RepositoryConnection.SCHEMA_COLLECTION, config, getMapping(config.getMappingID()));
        createCollection(RepositoryConnection.MAPPING_COLLECTION, config, getMapping(config.getMappingID()));
    }

    private void init(boolean boot) throws XMLDBCException
    {
        /* reloading tableInfo with configuration read and not boot one.*/
        repositoryTables = loadTableInfo(config);
        
        /* creating a new schema manager */
        schemaManager = new SchemaManager();
        String resource = RepositoryDataSource.class.getResource(
            MAPPING_PACKAGE + RESOURCES_FOLDER + MAPPING_SCHEMA_FILE).toString();
        try {
            schemaManager.loadSchema(new InputSource(resource));
        }
        catch (SAXException e) {
            throw new RepositoryException(RepositoryException.PARSER_ERROR,
            "Error while loading mapping language schema.", e);
        }
        
        /* initialize statements */
        initStatements();
        
        if (!boot)
        {
            /* loading UserTableManager */
            userTables = new UserTableManager(this);
            
            /* opening system collections */
            schemaCollection = openCollection(RepositoryConnection.SCHEMA_COLLECTION);   // Open the system repository (ignore exception
            mappingCollection = openCollection(RepositoryConnection.MAPPING_COLLECTION);   // Open the system repository (ignore exception
        }
   }

    private void initStatements()
    {
        StringBuffer sql = new StringBuffer();
        sql.append("select distinct c.NAME from ");
        sql.append(getTableInfo(TableSpec.TYPE_PATHS).getName());
        sql.append(" p,");
        sql.append(getTableInfo(TableSpec.TYPE_COLLECTIONS).getName());
        sql.append(" c where p.NS=? and p.CID=c.CID");
        checkSchemaStmt = sql.toString();
        
        sql.setLength(0);
        sql.append("delete from ");
        sql.append(getTableInfo(TableSpec.TYPE_PATHS).getName());
        sql.append(" where CID=");
        deletePathSetStmt = sql.toString();
        
        sql.setLength(0);
        sql.append("select NAME from ");
        sql.append(getTableInfo(TableSpec.TYPE_COLLECTIONS).getName());
        sql.append(" where MAPPING_ID=?");
        checkMappingStmt = sql.toString();
        
        sql.setLength(0);
        sql.append("update ");
        sql.append(getTableInfo(TableSpec.TYPE_COLLECTIONS).getName());
        sql.append(" set DESCRIPTION=? WHERE CID=?");
        updateColDescStmt = sql.toString();
    }
    
    private Map loadTableInfo(RepositoryInfo info) throws RepositoryException
    {
        HashMap map = new HashMap();
        
        /* loading tableInfo */
        List tables = TableSpecLoader.getInstance().getTableSpecs(TableSpec.CAT_REPOSITORY, metadataConnection.getDBMSType());
        Iterator it = tables.iterator();
        TableSpec wSpec = null;
        
        long[] params = new long[TableInfo.PARAM_NUMBER];
        params[TableInfo.PARAM_CID_SIZE] = info.getColIDSize();
        
        while (it.hasNext())
        {
            wSpec = (TableSpec)it.next();
            map.put(new Byte(wSpec.getType()), new TableInfo(wSpec, params, metadataConnection));
        }
        return map;
    }

    /** Close a collection identified by the <code>name</code> parameter
     * @param name the name of the collection.
     */
    private synchronized void cleanCollection(String name)
    throws RepositoryException
    {
        CollectionMetadata collection = (CollectionMetadata)collectionManager.remove(name.toUpperCase());
        if (collection != null)
        {
            try {
                if (collection.getMapping() != null)  // TO CHECK (un mapping quand mme ?)
                    loadedMappings.remove(collection.getInfo().getProperty(RepositoryConnection.MAPPING_ID_PROPERTY));
            }
            catch (XMLDBCNotRecognizedException e) {
                // no op : internal use
                throw new RuntimeException("Internal error: unknown feature or property.");
            }
            collection.close();
        }
    }
    
    private void dropRepositoryTables() throws SQLException, XMLDBCException
    {
        // delete all tables with XQuark prefixes
        String pattern = TABLE_PREFIX + "%";
        Iterator it = metadataConnection.getUserTableNames(pattern).iterator();
        while (it.hasNext())
        {
            try
            {
                metadataConnection.dropTable((String)it.next());
            }
            catch (SQLException e)
            {
                // ignore
            }
        }

        // delete all sequences with XQuark prefixes
        it = metadataConnection.getUserSequenceNames(pattern).iterator();
        while (it.hasNext())
        {
            try
            {
                metadataConnection.dropSequence((String)it.next());
            }
            catch (SQLException e)
            {
                // ignore
            }
        }
    }

    public synchronized void refresh() throws XMLDBCException
    {
        Iterator it = collectionManager.iterator();
        
        while(it.hasNext())
            ((_Collection)it.next()).refresh();
    }
    
    ///////////////////////////////////////////////////////////////////////////
    //                      INNER CLASSES
    ///////////////////////////////////////////////////////////////////////////
    private class ProxyCollection implements org.xquark.xquery.metadata.resolver.CollectionMetadata
    {
        String collectionName;
        XTree tree = null;
        
        public ProxyCollection(String collectionName)
        {
            this.collectionName = collectionName;
        }
        
        public XTree getXTree()
        {
            try
            {
                if (tree == null)
                    tree = getCollection(collectionName).getPathSet(); // won't be freed till LRU
                else
                    ((PathSet)tree).refresh();
            }
            catch (XMLDBCException e)
            { // TO SEE: How to process it ?
                throw new RuntimeException("Error while opening collection: " + e.getMessage());
            }
            return tree;
        }
        
        public String getCollectionName()
        {
            return collectionName;
        }
        
        public boolean isXTreeLoaded()
        {
            return tree != null;
        }

		public Collection getRootNodes() {
			return null;
		}

    }
    
}
