/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

/*
 * PathMetadata.java
 *
 * Created on 18 septembre 2001, 18:11
 */
package org.xquark.mapper.metadata;

import java.util.List;

import org.xquark.mapper.mapping.ColumnMapping;
import org.xquark.schema.Declaration;

/**
 * Common interface for XML storage to metadata for a particular
 * path. Should be used both by both Repository and Mapper.
 *
 */
public interface StoragePathMetadata
{

	/** Unknow mode */
	byte UNKNOWN = 0x0;
	/** no OID allocated (discarded from the storage model) */
	byte DISCARDED = 0x1;
	/** struct tuple may be present or not (Mixed with optional subelements or 
	 * no schema): TO DO mettre a jour dynamiquement pour chemins sans schema (assurer la persistence dans path set)
	 */
	byte POTENTIAL_STRUCT = 0x2;
	/** struct tuple present and no data tuple
	 */
	byte STRUCT_ONLY = 0x4;
	/** data tuple present and no range tuple
	 */
	byte DATA_ONLY = 0x8;
	/** struct and data tuple present
	 */
	byte DATA_AND_STRUCT = 0x10;

	/** Indicate the storage mode of the node, i.e., if the node is
	 * stored and where it is stored.
	 * @return a constant for all possible storage scheme
	 * @see #DISCARDED
	 * @see #POTENTIAL_STRUCT
	 * @see #STRUCT_ONLY
	 * @see #DATA_ONLY
	 * @see #DATA_AND_STRUCT
	 */
	byte getStorageMode();

	/** Indicate if an element can ne xsi:nil even if no extra-nil info is 
	 * available.
	 * @return if True, when there is a possible confusion beetween 
	 * missing and nil (in fact null column case), then the element is nil.
	 */
	boolean isNilImplicit();

	/**
	  * Get the column mapping defined on this path used for reconstruction
	  * or querying.
	  * @return the reference {@link ColumnMapping org.xquark.mapper.mapping.ColumnMapping} object
	  * @see org.xquark.mapper.mapping.ColumnMapping
	  */
	// TO DO : when multiple column mappings exist on a declaration add an attribute to
	// mapping language to let the user specify it
	ColumnMapping getReferenceColumnMapping();
	// take first in the list (in a 1st step)

	/**
	* Get the table mappings defined on this path.
	* @return a list of table mappings
	* @see org.xquark.mapper.mapping.TableMapping
	*/
	List getTableMappings();

	/**
	 * Get the column mappings defined on this path.
	 * @return a list of column mappings
	 * @see org.xquark.mapper.mapping.ColumnMapping
	 */
	List getColumnMappings();

	/** Get the unique identifier for a path in an XMLCollection.
	 * @return a short identifier <B>that must be used for node comparison</B>
	 */
	short getPathID();

	/** Flag to indicate if the node has been stored and an OID allocated.
	 * <p>i.e., getStorageMode() != DISCARDED</p>
	 * @return true if an OID exists for the node
	 */
	boolean isPersistent();

	/** Indicate if the position of the node iwithin its siblings is enforced
	 * by schema information.
	 * @return True if positon of the node can be enforced by schema information.
	 */
	boolean isPositionEnforced();

	/** Returns the XML schema component for this node.
	 * <p>Allows to access declarations info like nillable, minOccurs...</p>
	 * @return The schema component corresponding to the path if any.
	 * Null else.
	 */
	Declaration getDeclaration();

	/**
	 * Returns the XPath model node type.
	 * @return an integer constant defined in {@link org.xquark.xml.xpath.Type}
	 */
	public byte getType();
	
	/**
	 * Returns the namespace of the metadata node.
	 * @return the namespace URI
	 * @see org.xquark.xml.xpath.XTreeNode
	 */
	public String getNamespace();
	
	/**
	 * Returns the local name of the metadata node.
	 * @return the local name
	 * @see org.xquark.xml.xpath.XTreeNode
	 */
	public String getLocalName();
	
	/**
	 * Whether, according to the XML schema type of the metadata node, the
	 * default mapping (not planned by user using mapping language) is authorized
	 * for the path.
	 * @return boolean
	 */
	public boolean isDefaultAllowed();
}
