/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

/*
 * UserTableManager.java
 *
 * Created on 29 aot 2002, 14:55
 */

package org.xquark.mapper.metadata;

import java.sql.*;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.xquark.mapper.RepositoryException;
import org.xquark.mapper.dbms.*;

/**
 * Class that encapsulate UOID allocation and manipulation functions.
 * This object manages concurrency using a sequence an locking the recycling
 * table.
 *
 */
public class UOIDManager
{
    private static final String RCSRevision = "$Revision: 1.2 $";
    private static final String RCSName     = "$Name: XQ1_1_1 $";
    
    public static final String SEQ_YES = "Y";
    
    private CollectionMetadata collection;
    private String deleteStmt;
    private String selectStmt;
    
    private TableInfo freeDocOIDTable;
    private Sequence docOIDSeq;
    private long cid;
    private byte cid_size;
    private byte did_size;
    private int did_shift;
    private int cid_shift;
    private long did_shift_mult;
    private long nid_mask;
    private long did_udid_mask;
    private long did_uoid_mask;
    private long max_node_oid;
    private CollectionInfo collectionInfo;  // must not change during collection life
    private ArrayList freeDIDs;
    
    /** Creates a new instance of UserTableManager */
    public UOIDManager(CollectionMetadata collection)
    throws RepositoryException
    {
        this.collection = collection;
        cid = (long)collection.getCollectionID();
        cid_size = collection.getRepository().getInfo().getColIDSize();
        did_size = collection.getInfo().getDocOIDSize();
        collectionInfo = collection.getInfo(); //.getDocOIDPreallocationSize(),
        freeDocOIDTable = collection.getTableInfo(TableSpec.TYPE_FREE_DOC);
        docOIDSeq = new Sequence(collection.getTableInfo(TableSpec.TYPE_DOC_SEQ),
        collectionInfo.getDocOIDPreallocationSize()
        );
        freeDIDs = new ArrayList(collectionInfo.getDocOIDPreallocationSize());
        init();
    }
    
    /** Creates a new instance of UserTableManager */
    private void init()
    {
        did_shift = RepositoryConstants.UOID_SIZE - (cid_size + did_size);
        cid_shift = RepositoryConstants.UOID_SIZE - cid_size;
        nid_mask = ~(Long.MAX_VALUE << did_shift);
        did_shift_mult = 1L << did_shift;
        max_node_oid = did_shift_mult - 1L;
        did_udid_mask = (1L << did_size) - 1L;
//        did_uoid_mask = did_udid_mask << did_shift;
        initStatements();
    }
    
    private void initStatements()
    {
        StringBuffer sql = new StringBuffer();
        sql.append("DELETE FROM ");
        sql.append(freeDocOIDTable.getName());
        sql.append(" WHERE ");
        sql.append(freeDocOIDTable.getColumns()[0].getName());
        sql.append("<=");
        deleteStmt = sql.toString();
        
       selectStmt = collection.getMetadataConnection().getLockSelectStatement(
    		freeDocOIDTable, "ORDER BY FREE_DID");
    }
    
    public synchronized long getNewDocOID() throws RepositoryException
    {
        Statement stmt = null;
        ResultSet rs = null;
        
        if (freeDIDs.size() == 0)   // no more cached OIDs
        {
            AbstractConnection metaConn = collection.getMetadataConnection();
            synchronized(metaConn) // SHOULD add it to any use of metadata performing commit (reading?)
            {
                // First, try to recycle
                try
                {
                    metaConn.start();
                    stmt = metaConn.getConnection().createStatement();
                    stmt.setFetchSize(collectionInfo.getDocOIDPreallocationSize());
                    rs = stmt.executeQuery(selectStmt);
                    // NOTE: if updatable cursor not portable, use an "order by" instead followed by a "delete where FREE_DID > x"
                    // For Oracle: must not use "select *" but error on deleteRow() : missing parenthesis !
                    while (rs.next() && (freeDIDs.size() < collectionInfo.getDocOIDPreallocationSize()))
                    {
                        freeDIDs.add(new Long(rs.getLong(1)));
                    }
                    rs.close();
                    
                    // delete selected rows (last must be the biggest)
                    
                    if (freeDIDs.size() > 0)
                        stmt.executeUpdate(deleteStmt + freeDIDs.get(freeDIDs.size() - 1));
                }
                catch (SQLException e)
                {
                    try
                    {
                        metaConn.rollback();
                    }
                    catch (SQLException s)
                    {
                        // no op
                    }
                    freeDIDs.clear();  // were not recycled in fact
                    throw new RepositoryException(RepositoryException.DB_ERROR,
                    "JDBC error while recycling a document OIDs.", e);
                }
                finally
                {
                    metaConn.commit();  // release locks
                    try
                    {
                        if (stmt != null)
                            stmt.close();
                    }
                    catch (SQLException e)
                    {
                        // no op
                    }
                }
            }
            
            // Recycle was not profitable: allocate a new block of OIDs
            if (freeDIDs.size() == 0)   // no more cached OIDs
            {
                long newOID = docOIDSeq.nextValue(metaConn); // values are reserved from newOID to new OID + prealloc - 1
                for (long i = newOID + collectionInfo.getDocOIDPreallocationSize() - 1;
                i >= newOID; i--)
                    freeDIDs.add(new Long(i)); // add lowest OID at the end because the end is consumed first for performance reasons (ArrayList)
            }
        }
        
        return ((Long)freeDIDs.remove(freeDIDs.size() - 1)).longValue();
    }
    
    /** @param restoredDIDs List of Long objects containing DIDs (not UDIDs) */
    public synchronized void freeDocOIDs(List restoredDIDs) throws RepositoryException
    {
        PreparedStatement pStmt = null;
        try
        {
            pStmt = collection.getMetadataConnection().getConnection().prepareStatement(freeDocOIDTable.getInsertStatement());
            Iterator it = restoredDIDs.iterator();
            while (it.hasNext())
            {
                pStmt.setLong(1, ((Long)it.next()).longValue());
                pStmt.addBatch();
            }
            pStmt.executeBatch();
            restoredDIDs.clear();
        }
        catch (SQLException e)
        {
            throw new RepositoryException(RepositoryException.DB_ERROR,
            "JDBC error while recycling document OIDs.", e);
        }
        finally
        {
            try
            {
                if (pStmt != null)
                    pStmt.close();
            }
            catch (SQLException e)
            {
                // no op
            }
        }
    }
    
    /** Saves unused OIDS */
    private void clear() throws RepositoryException
    {
        if (freeDIDs.size() > 0)
            freeDocOIDs(freeDIDs);
    }
    
    public void close() throws RepositoryException
    {
        clear();
        freeDIDs = null;
    }
    
    public long buildUOID(long docOID, long nodeOID)
    {
        //        return (cid << cid_shift) + (docOID << did_shift) + nodeOID;
        return (cid << cid_shift) | (docOID << did_shift) | nodeOID;
    }
    
    public long buildUDID(long docOID)
    {
        //        return (cid << cid_shift) + (docOID << did_shift) + nodeOID;
        return (cid << did_size) | docOID;
    }
    
    public short getCID(long uoid)
    {
        return (short)(uoid >>> cid_shift);
    }
    
    public long getUDID(long uoid)
    {
        return uoid >>> did_shift;
    }
    
    public long getDID(long udid)
    {
        return udid & did_udid_mask;
    }
    
    public long getNodeOID(long uoid)
    {
        return uoid & nid_mask;
    }
    
    public long getDocMinUOID(long docOID)
    {
        return buildUOID(docOID, 0);
    }
    
    public long getDocMaxUOID(long docOID)
    {
        return buildUOID(docOID, max_node_oid);
    }
    
    public long getDocMaxConstant()
    {
        return max_node_oid;
    }
    
    public long getDocMultiplier()
    {
        return did_shift_mult;
    }
    
    public long getBucketDivisor()
    {
        return did_shift_mult;
    }
    
    public int getDIDShift()
    {
        return did_shift;
    }
}
