/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

/*
 * UserTableManager.java
 *
 * Created on 29 aot 2002, 14:55
 */

package org.xquark.mapper.metadata;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.HashMap;
import java.util.Map;

import org.xquark.mapper.RepositoryException;
import org.xquark.mapper.dbms.Sequence;
import org.xquark.mapper.dbms.TableInfo;
import org.xquark.mapper.dbms.TableSpec;

/**
 * User table dictionary.
 * This object manages concurrency problems using a lock-optimistic algorithm.
 *
 */
public class UserTableManager
{
    private static final String RCSRevision = "$Revision: 1.2 $";
    private static final String RCSName     = "$Name: XQ1_1_1 $";
    
    private Repository rep;
    private Sequence userTablesSeq;
    private Map tableIDs = new HashMap();
    private String selectStmt;
    private String insertStmt;
    
    /** Creates a new instance of UserTableManager */
    public UserTableManager(Repository rep)
    throws RepositoryException
    {
        this.rep = rep;
        userTablesSeq = new Sequence(rep.getTableInfo(TableSpec.TYPE_TABLE_SEQ), (short)1);
        initStatements(rep);
        refresh();
    }
    
    private void initStatements(Repository rep)
    {
        TableInfo userTablesTable = rep.getTableInfo(TableSpec.TYPE_TABLES);
        selectStmt = userTablesTable.getSelectAllStatement();
        insertStmt = "INSERT INTO " + userTablesTable.getName() + " VALUES(";
    }
    
    private synchronized void refresh() throws RepositoryException
    {
        tableIDs.clear();
        Statement stmt = null;
        ResultSet rs = null;
        try
        {
            stmt = rep.getMetadataConnection().getConnection().createStatement();
            rs = stmt.executeQuery(selectStmt);
            while (rs.next())
                tableIDs.put(rs.getString(1), new Short(rs.getShort(2)));
        }
        catch (SQLException e)
        {
            throw new RepositoryException(RepositoryException.DB_ERROR,
            "Could not read the user tables system table", e);
        }
        finally
        {
            try
            {
                if (rs != null)
                    rs.close();
                if (stmt != null)
                    stmt.close();
            }
            catch (SQLException e)
            {
                // no op
            }
        }
    }
    
    public synchronized short getTableID(String tableName)
    throws RepositoryException
    {
        short ret = -1;
        
        /* First look in the map */
        ret = getID(tableName);
        if (ret == -1)
        {
            // First refresh the cache (for the case another instance has already created table)
            refresh();
            // look again in the map
            ret = getID(tableName);
            if (ret == -1)
            {
                // allocate a new ID
                short newTableID = -1;
                newTableID = (short)userTablesSeq.nextValue(rep.getMetadataConnection());
                // save it
                try
                {
                    rep.getMetadataConnection().executeUpdate(insertStmt + "'" + tableName + "'," + newTableID + ')');
                    tableIDs.put(tableName, new Short(newTableID));
                    ret = newTableID;
                }
                catch (SQLException e)
                {
                    // collision
                    refresh();
                    ret = getID(tableName);
                    if (ret == -1)
                        throw new RepositoryException(RepositoryException.DB_CONSISTENCY_ERROR,
                        "Could not insert a row in the user tables system table", e);
                }
            }
        }
        
        return ret;
    }
    
    private short getID(String tableName)
    {
        short ret = -1;
        Short mapID = (Short)tableIDs.get(tableName);
        if (mapID != null)
            ret = mapID.shortValue();
        return ret;
    }
}
