/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

/*
 * QueryExplorer.java
 *
 * Created on 18 avril 2002, 17:15
 */

package org.xquark.mapper.storage;

import java.sql.SQLException;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.xquark.mapper.RepositoryException;
import org.xquark.mapper.dbms.*;
import org.xquark.mapper.mapping.TableMapping;
import org.xquark.mapper.metadata.CollectionMappingInfo;
import org.xquark.mapper.metadata.CollectionMetadata;
import org.xquark.xml.xdbc.XMLDBCException;

/**
 *
 */
class BatchQueryExplorer extends QueryExplorer
{
    private static final String RCSRevision = "$Revision: 1.3 $";
    private static final String RCSName     = "$Name: XQ1_1_1 $";
    
    private static Log log = LogFactory.getLog(BatchQueryExplorer.class);
    
    private VarInfoSet tempTable;
    
    private static TableSpec tmpTableSpecs;
    
    static
    {
        tmpTableSpecs = TableSpecLoader.getInstance().getTableSpec(
            TableSpec.CAT_QUERY_TMP,
            TableSpec.TYPE_QUERY_TMP
            );
    }
    
    /** Creates a new instance of QueryExplorer */
    BatchQueryExplorer(_RepositoryCollection collection) throws SQLException, XMLDBCException
    {
        super(collection);
        CollectionMetadata meta = collection.getMetadata();
        wNode = qNode = new BatchQueryStructExplorer(meta.getTableInfo(TableSpec.TYPE_STRUCT), meta.getUOIDManager());
        extraNode = qExtraNode = new BatchQueryExtraDataExplorer(meta.getTableInfo(TableSpec.TYPE_EXTRA));
    }
    
    private VarInfoSet prepareVarTable(VarInfoSet varInfoSet) throws SQLException
    {
        VarInfoSet tempTable = null;
        List collapseVars = varInfoSet.getCollapseColumns();
        if (collapseVars.size() > 0)
        {
            AbstractConnection connection = collection.getRepositoryConnection().getConnection();
            ColumnSpec[] columns = tmpTableSpecs.getColumns();
            StringBuffer subQuery = new StringBuffer("SELECT MIN(");
            subQuery.append(varInfoSet.getRankColumnName());
            subQuery.append(") as ");
            subQuery.append(columns[0].getName());
            subQuery.append(',');
            subQuery.append(varInfoSet.getFirstColumnName());
            subQuery.append(" as ");
            subQuery.append(columns[1].getName());
            subQuery.append(',');
            subQuery.append(varInfoSet.getLastColumnName());
            subQuery.append(" as ");
            subQuery.append(columns[2].getName());
            subQuery.append(',');
            subQuery.append(varInfoSet.getPathColumnName());
            subQuery.append(" as ");
            subQuery.append(columns[3].getName());
            subQuery.append(" FROM ");
            subQuery.append(varInfoSet.getTableName());
            subQuery.append(" GROUP BY ");
            
            for (int i = 0; i < collapseVars.size(); i++)
            {
                if (i != 0)
                    subQuery.append(',');
                if (collapseVars.get(i).equals(varInfoSet.getFirstColumnName()))
                {
                    subQuery.append(varInfoSet.getFirstColumnName());
                    subQuery.append(',');
                    subQuery.append(varInfoSet.getLastColumnName());
                    subQuery.append(',');
                    subQuery.append(varInfoSet.getPathColumnName());
                }
                else
                    subQuery.append((String)collapseVars.get(i));
                
            }
            String tableName = connection.createTemporaryTableAs(subQuery.toString());
            tempTable = new VarBuildInfoSet
            (
            tableName,
            columns[0].getName(),
            columns[1].getName(),
            columns[2].getName(),
            columns[3].getName()
            );
        }
        return tempTable;
    }
    
    /**
     */
    void close() throws SQLException, RepositoryException
    {
        super.close();
        
        if (tempTable != null)
        {
            collection.getRepositoryConnection().getConnection().dropTemporaryTable(tempTable.getTableName());
            tempTable = null;
        }
    }
    
    /** 
     */
    void reset() throws SQLException, RepositoryException
    {
        super.reset();
        
        if (tempTable != null)
        {
            collection.getRepositoryConnection().getConnection().dropTemporaryTable(tempTable.getTableName());
            tempTable = null;
        }
    }
    
    /** Perform Tuple allocation and therefore query execution
     * @param subPaths a  {PathNode subPath, PathNode variablePath} map giving
     * for an absolute subpath the variable root path
     * @param varPaths 0 index PathNode contains the variable root path other are
     * this path subpaths
     */
    void initExploration(VarInfoSet varInfoSet, AbstractConnection connection) throws SQLException, XMLDBCException
    {
        if (log.isDebugEnabled())
        	log.debug("QueryExplorer::initExploration [varInfoSet]=" + varInfoSet);
        
        // Intermediate table creation
        tempTable = prepareVarTable(varInfoSet);
        
        // Table partitioning
        Object[] ret = dispatchSubPaths(varInfoSet);
        Map tableMappingSubset = (Map)ret[0]; 
        List subPathNodes = (List)ret[1];
        
        // Table subset browsing for statement execution
        Iterator it = tableMappingSubset.entrySet().iterator();
        Map.Entry entry;
        CollectionMappingInfo tableStmts;
        int tableIndex;
        
        CollectionMetadata meta = collection.getMetadata();
        while (it.hasNext())
        {
            entry = (Map.Entry)it.next();
            tableIndex = ((TableMapping)entry.getKey()).getTableIndex();
            tableStmts = (CollectionMappingInfo)meta.getMappingStatements(tableIndex);
            if (queryTuples[tableIndex] == null)
                queryTuples[tableIndex] = new BatchQueryTuple(tableStmts, meta.getMappingPID(tableIndex));
            ((BatchQueryTuple)queryTuples[tableIndex]).execQuery((List)entry.getValue(),
            tempTable == null?varInfoSet:tempTable, meta, connection);
        }
        
        // initialize structure retrieval
        ((BatchQueryStructExplorer)qNode).getBranch(subPathNodes,
        tempTable == null?varInfoSet:tempTable, connection);
        ((BatchQueryExtraDataExplorer)qExtraNode).getBranch(subPathNodes,
        tempTable == null?varInfoSet:tempTable, connection);
    }    
    
    void setCurrentRank(int rank)
    {
        for (int i=0; i < queryTuples.length; i++)
        {
            if (queryTuples[i] != null)
                ((BatchQueryTuple)queryTuples[i]).setCurrentRank(rank);
        }
        ((BatchQueryStructExplorer)qNode).setCurrentRank(rank);
        ((BatchQueryExtraDataExplorer)qExtraNode).setCurrentRank(rank);
    }
    
}
