/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.mapper.storage;

import java.sql.SQLException;
import java.util.List;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.xquark.mapper.dbms.AbstractConnection;
import org.xquark.mapper.dbms.TableInfo;
import org.xquark.mapper.mapping.ColumnMapping;
import org.xquark.mapper.metadata.PathNode;
import org.xquark.mapper.metadata.UOIDManager;
import org.xquark.mapper.util.RepositoryProperties;

/**
 * Object used to perform XML structure node storage.
 *
 */
public class BatchQueryStructExplorer extends QueryStructExplorer
{
    private static final String RCSRevision = "$Revision: 1.3 $";
    private static final String RCSName     = "$Name: XQ1_1_1 $";
    
    private static Log log = LogFactory.getLog(BatchQueryStructExplorer.class);
    
    private int currentRank = -2;
    private int rank = -1;
    
    public BatchQueryStructExplorer(TableInfo tableInfo, UOIDManager manager)
    {
        super(tableInfo, manager);
    }
    
    /**
     * Launch the table entry retrieving for a particular bukect ID and a OID range.
     */
    public void getBranch(List paths, VarInfoSet tmpTable,
    AbstractConnection connection)
    throws SQLException
    {
        stmt = connection.getConnection().createStatement();
        try
        {
            connection.setFetchSize(stmt, RepositoryProperties.getIntProperty(CONF_TREE_FETCHSIZE));
            String stmtString = generateStatement(paths, tmpTable);
            if (stmtString != null)
            {
                rs = stmt.executeQuery(stmtString);
                notExhausted = true;
                currentResultExhausted = false;
            }
        }
        catch (SQLException e)
        {
            stmt.close();
            throw e;
        }
        // NOTE test ORDER BY source, rank to see if equivalent (would match any mapping)
    }
    
    private String generateStatement(List paths, VarInfoSet tmpTable)
    {
        StringBuffer SQLStmt = new StringBuffer();
        
        ColumnMapping refColumn;
        SubPathInfo path;
        
        for (int i = 0; i < paths.size(); i++)
        {
            path = (SubPathInfo)paths.get(i);
            List subpaths = path.subPath.getModelSubPaths();
            int subpathsCount = subpaths.size();
            if ((subpathsCount > 0) && !path.isText)
            {
                if (SQLStmt.length() > 0)   // i was not a correct flag because index 0 coul be discarded
                    SQLStmt.append("\nunion all\n");
                
                SQLStmt.append("select t.");
                SQLStmt.append(tmpTable.getRankColumnName());
                SQLStmt.append(',');
                SQLStmt.append(path.subPathIndex);
                SQLStmt.append(" as n,v.uoid,v.last,v.path\nfrom ");
                SQLStmt.append(tmpTable.getTableName());
                SQLStmt.append(" t,");
                SQLStmt.append(tableInfo.getName());
                SQLStmt.append(" v\nwhere v.bucket=floor(t.");
                SQLStmt.append(tmpTable.getFirstColumnName());
                SQLStmt.append('/');
                SQLStmt.append(manager.getDocMultiplier());
                SQLStmt.append(") and v.uoid between t.");
                SQLStmt.append(tmpTable.getFirstColumnName());
                SQLStmt.append(" and t.");
                SQLStmt.append(tmpTable.getLastColumnName());
                SQLStmt.append(" and t.");
                SQLStmt.append(tmpTable.getPathColumnName());
                SQLStmt.append('=');
                SQLStmt.append(path.varPath.getPathID());
                // path restriction
                // last struct ancestor is added in order for the builder to initiate range
                SQLStmt.append(" and v.path");
                
                if (subpathsCount == 1)
                {
                    SQLStmt.append("=");
                    SQLStmt.append(((PathNode)subpaths.get(0)).getPathID());
                }
                else
                {
                    SQLStmt.append(" in(");
                    
                    for (int j = 0; j < subpathsCount; j++)
                    {
                        SQLStmt.append(((PathNode)subpaths.get(j)).getPathID());
                        SQLStmt.append(',');
                    }
                    SQLStmt.setCharAt(SQLStmt.length()-1, ')');
                }
            }
        }
        
        if (SQLStmt.length() > 0)
        {
            SQLStmt.append("\norder by ");
            SQLStmt.append(tmpTable.getRankColumnName());
            SQLStmt.append(",n,uoid");
            
            if (log.isDebugEnabled())
            	log.debug("Value stmt for struct table:\n" + SQLStmt);
            
            return SQLStmt.toString();
        }
        else
            return null;
    }
    
    /**
     * get the next row from database and store it in the current object.
     */
    public boolean fetchNextRow() throws SQLException
    {
        try
        {
            if (resultActive())
            {
                notExhausted = rs.next();
                if (notExhausted)
                {
                    rank = rs.getInt(1);
                    subPathIndex = rs.getInt(2);
                    uoid = rs.getLong(3);
                    last = rs.getLong(4);
                    path = rs.getShort(5);
                    updateResultSegmentationFlag();
                }
                else
                    reset();
            }
            else
                return false;
            
            //                Debug.print("Tuple :: fetchNextRow for table " + mapping.getTableName() + " UOID=" + getUOID());
            return notExhausted;
        }
        catch (SQLException e)
        {
            close();
            throw e;
        }
    }
    
    void setCurrentRank(int rank)
    {
        currentSubPathIndex = 0;
        currentRank = rank;
        updateResultSegmentationFlag();
    }
    
    public void reset() throws SQLException
    {
        super.reset();
        currentRank = -2;
        rank = -1;
    }
    
    protected void updateResultSegmentationFlag()
    {
        currentResultExhausted =  (rank != currentRank) || (subPathIndex != currentSubPathIndex);
    }
}
