/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.mapper.storage;

import java.sql.SQLException;
import java.util.Collections;
import java.util.List;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.xquark.mapper.dbms.AbstractConnection;
import org.xquark.mapper.mapping.ColumnMapping;
import org.xquark.mapper.mapping.TableMapping;
import org.xquark.mapper.metadata.*;
import org.xquark.mapper.util.RepositoryProperties;
import org.xquark.xml.xdbc.XMLDBCException;

class BatchQueryTuple extends QueryTuple
{
    private static final String RCSRevision = "$Revision: 1.3 $";
    private static final String RCSName = "$Name: XQ1_1_1 $";
    
    private static Log log = LogFactory.getLog(BatchQueryTuple.class);
    
    protected int rank = -1;
    
    private int currentRank = -2;
    
    BatchQueryTuple(CollectionMappingInfo table, short pathID)
    {
        super(table, pathID);
        valueOffset = 3;
    }
    
    /** To be used by query reconstruction   */
    void execQuery(List paths, VarInfoSet tmpTable, CollectionMetadata collection,
    AbstractConnection connection)
    throws SQLException, XMLDBCException
    {
        try
        {
            selectStmt = connection.getConnection().createStatement(); // TO CHECK : could keep the statement ?
            connection.setFetchSize(selectStmt, RepositoryProperties.getIntProperty(RepositoryConstants.CONF_USER_FETCHSIZE));
            if (resultSet != null)
                resultSet.close();
            resultSet = selectStmt.executeQuery(generateStatement(paths, tmpTable, collection));
            notExhausted = true;
            currentResultExhausted = false;
        }
        catch (SQLException e)
        {
            selectStmt.close();
            throw e;
        }
    }
    
    private String generateStatement(List paths, VarInfoSet tmpTable, CollectionMetadata collection)
    throws XMLDBCException
    {
        StringBuffer SQLStmt = new StringBuffer();
        
        SubPathInfo path;
        ColumnMapping refColumn = null;
        TableMapping mapping = table.getTableMapping();
        String UOIDColumnName, pathColumnName;
        if (mapping.isClustered())
        {
            UOIDColumnName = mapping.getUOIDColumnName();
            pathColumnName = mapping.getPathIDColumnName();
        }
        else
        {
            UOIDColumnName = "uoid";
            pathColumnName = "path";
        }
        
        for (int i = 0; i < paths.size(); i++)
        {
            path = (SubPathInfo)paths.get(i);
            
            if (i > 0)
                SQLStmt.append("\nunion all\n");
            
            SQLStmt.append("select t.");
            SQLStmt.append(tmpTable.getRankColumnName());
            SQLStmt.append(',');
            SQLStmt.append(path.subPathIndex);
            SQLStmt.append(" as n");
            if (!mapping.isClustered())
            {
                SQLStmt.append(",v.");
                SQLStmt.append(UOIDColumnName);
                SQLStmt.append(",v.");
                SQLStmt.append(pathColumnName);
            }
            
            
            if (path.subPath.getPathID() != RepositoryConstants.ROOT_PATH_ID)  // because root has default mapping ? TO CHANGE
                refColumn = path.subPath.getReferenceColumnMapping();
            
            ColumnMapping[] columns = mapping.getColumnMappings();
            ColumnMapping column;
            for (int j = 0; j < columns.length; j++)
            {
                column = columns[j];
                if ((refColumn == null)
                || (column == refColumn)
                || (mapping.isClustered() && ((column.getColumnIndex() == mapping.getUOIDIndex())
                || (column.getColumnIndex() == mapping.getPathIDIndex()))))
                {
                    if (mapping.isClustered())
                        SQLStmt.append(",v.");
                    else
                        SQLStmt.append(",w.");
                    
                    SQLStmt.append(column.getColumnName());
                }
                else
                {
                    SQLStmt.append(',');
                    SQLStmt.append(column.getTypeInfo().getTypedNullValue()); // TO IMPROVE : all column mappings do not need ordering
                }
            }
            
            SQLStmt.append("\nfrom ");   // TO DO : optimize the query (join and condition order)
            SQLStmt.append(tmpTable.getTableName());
            SQLStmt.append(" t,");
            if (mapping.isClustered())
            {
                SQLStmt.append(mapping.getTableName());
                SQLStmt.append(" v");
            }
            else
            {
                SQLStmt.append(collection.getOIDTableName(mapping.getTableIndex()));
                SQLStmt.append(" v,");
                SQLStmt.append(mapping.getTableName());
                SQLStmt.append(" w");
            }
            SQLStmt.append("\nwhere v.");
            SQLStmt.append(UOIDColumnName);
            SQLStmt.append(" between t.");
            SQLStmt.append(tmpTable.getFirstColumnName());
            SQLStmt.append(" and t.");
            SQLStmt.append(tmpTable.getLastColumnName());
            SQLStmt.append(" and t.");
            SQLStmt.append(tmpTable.getPathColumnName());
            SQLStmt.append('=');
            SQLStmt.append(path.varPath.getPathID());
            if (!mapping.isClustered())
            {
                SQLStmt.append(" and ");
                SQLStmt.append(mapping.getJoinCondition());
            }
            // path restriction
            if (mapping.isTableShared())
            {
	            if (!mapping.isShared())  // no table mapping (optimization)
	            {
	                SQLStmt.append(" and abs(v.");
	                SQLStmt.append(pathColumnName);
	                SQLStmt.append(")=");
	                SQLStmt.append(tablePath);
	            }
	            else    // complex element mapping or generic
	            {
	                List subpaths = null;
	                if (path.isText)
	                    subpaths = Collections.singletonList(path.subPath.getBuilderAncestor());
	                else
	                    subpaths = path.subPath.getBuilderAncestor().getModelSubPaths();
	                int subpathsCount = subpaths.size();
	                if (subpathsCount > 0)
	                {
	                    SQLStmt.append(" and abs(v.");
	                    SQLStmt.append(pathColumnName);
	                    SQLStmt.append(')');
	                    
	                    if (subpathsCount == 1)
	                    {
	                        SQLStmt.append('=');
	                        SQLStmt.append(((PathNode)subpaths.get(0)).getPathID());
	                    }
	                    else
	                    {
	                        SQLStmt.append(" in(");
	                        
	                        for (int j = 0; j < subpathsCount; j++)
	                        {
	                            SQLStmt.append(((PathNode)subpaths.get(j)).getPathID());
	                            SQLStmt.append(',');
	                        }
	                        SQLStmt.setCharAt(SQLStmt.length()-1, ')');
	                    }
	                }
	            }
            }
        }
        
        SQLStmt.append("\norder by ");
        SQLStmt.append(tmpTable.getRankColumnName());
        SQLStmt.append(",n,");
        SQLStmt.append(UOIDColumnName);
        
        if (log.isDebugEnabled())
        	log.debug("Value recontruction statement stmt [table,query]\n"
        			+ mapping.getTableName() + '\n' + SQLStmt);
        return SQLStmt.toString();
    }
    
    boolean fetchNextRow() throws SQLException
    {
        super.fetchNextRow();
        try
        {
            if (resultActive())
            {
                notExhausted = resultSet.next();
                if (notExhausted)
                {
                    TableMapping mapping = table.getTableMapping();
                    rank = resultSet.getInt(1);
                    subPathIndex = resultSet.getInt(2);
                    uoid = resultSet.getLong(mapping.getUOIDIndex() + 3);   // num + resultset offset/ List
                    pathoid = resultSet.getShort(mapping.getPathIDIndex() + 3);
                    updateResultSegmentationFlag();
                }
                else
                    reset();
            }
            else
                return false;
            
            return notExhausted;
        }
        catch (SQLException e)
        {
            selectStmt.close();
            throw e;
        }
    }
    
    void reset() throws SQLException
    {
        super.reset();
        rank = -1;
        currentRank = -2;
    }
    
    void setCurrentRank(int rank)
    {
        currentSubPathIndex = 0;
        currentRank = rank;
        updateResultSegmentationFlag();
    }
    
    protected void updateResultSegmentationFlag()
    {
        currentResultExhausted =  (rank != currentRank) || (subPathIndex != currentSubPathIndex);
    }
}

