/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.mapper.storage;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

import org.xml.sax.SAXException;
import org.xquark.mapper.RepositoryException;
import org.xquark.mapper.metadata.RepositoryConstants;
import org.xquark.mapper.metadata.StoragePathMetadata;
import org.xquark.mapper.util.DestructionToken;
import org.xquark.mapper.util.Functions;
import org.xquark.schema.validation.SchemaValidationContext;
import org.xquark.xml.xdbc.*;
import org.xquark.xpath.NodeKind;

/** Class using SAX2 event handlers for building the optimized XML logical structure.
 *
 * <ul>
 * <li>{@link org.xml.sax.ContentHandler ContentHandler}</li>
 * </ul>
 *
 * @see org.xml.sax.ContentHandler
 */
public class CollectionFiler extends SAXHandler
{
    private static final String RCSRevision = "$Revision: 1.4 $";
    private static final String RCSName     = "$Name: XQ1_1_1 $";

    //
    //  XMLCollection (database) related data
    //
    protected _RepositoryCollection collection;
    protected String docID;
    private boolean IDSetByUser = false;
    private DestructionToken destructor;
    private boolean storePrefixes = true;   // if false, prefix storage is not performed.
    private List declaredPrefixes;

    
    /** Constructor
     */
    public CollectionFiler(_RepositoryCollection collection, SchemaValidationContext schemaContext, boolean storePrefixes, DestructionToken destructor) throws XMLDBCException
    {
        super(
            collection.getMetadata().getPathSet().createIterator(),
            collection.getRepositoryConnection().getConnection(),
            new ModelBuilder(collection, schemaContext)
            );
        this.storePrefixes = storePrefixes;
        if (storePrefixes)
            declaredPrefixes = new ArrayList(3);
        this.collection = collection;
        this.destructor = destructor;
        textData.ensureCapacity(collection.getInfo().getMaxTextLength());
    }
    
    public void reset() throws RepositoryException
    {
        super.reset();
        IDSetByUser = false;
        declaredPrefixes.clear();
    }
    
    public void close() throws XMLDBCException
    {
        super.close();
        if (destructor != null)
        {
            declaredPrefixes = null;
            destructor.destruct(this);
            destructor = null;
        }
    }
    
    protected void processFatalError(Exception e) throws SAXException
    {
        super.processFatalError(e);
        collection.restoreDID(builder.getDocumentOID());
    }
    
    //////////////////////////////////////////////////////
    //      Implementation of LexicalHandler interface.
    //////////////////////////////////////////////////////
    public void comment(char ch[], int start, int length) throws SAXException
    {
        //Debug.print("comment");
        
        // create a node in the model
        try
        {
            ((ModelBuilder)builder).extraNode(NodeKind.COMMENT, context.getStoragePathMetadata(), new String(ch, start, length), locator);
        }
        catch(XMLDBCException e)
        {
            processFatalError(e);
        }
		catch(SQLException e)
		{
			processFatalError(e);
		}
    }
    
    public void startCDATA() throws SAXException
    {
        //Debug.print("startCDATA");
        
        // create a node in the model
        try
        {
            ((ModelBuilder)builder).extraNode(RepositoryConstants.START_CDATA_SECTION,context.getStoragePathMetadata(), null, locator);
        }
        catch(XMLDBCException e)
        {
            processFatalError(e);
        }
		catch(SQLException e)
		{
			processFatalError(e);
		}
    }

    public void endCDATA() throws SAXException
    {
        //Debug.print("startCDATA");
        
        // create a node in the model
        try
        {
            ((ModelBuilder)builder).extraNode(RepositoryConstants.END_CDATA_SECTION,context.getStoragePathMetadata(), null, locator);
        }
		catch(XMLDBCException e)
		{
			processFatalError(e);
		}
		catch(SQLException e)
		{
			processFatalError(e);
		}
    }
    ////////////////////////////////////////////////////////////////////
    //    Implementation of ContentHandler interface.
    ////////////////////////////////////////////////////////////////////
    public void startPrefixMapping(String prefix, String uri)
    throws SAXException
    {
        //Debug.print("startPrefixMapping");
        // Store the prefix until the element is saved
        declaredPrefixes.add("{" + uri + "}" + prefix);
    }
    
    protected void processPrefixes()
    throws XMLDBCException, SQLException
    {
        if (storePrefixes)
        {
            int prefixCount = declaredPrefixes.size();
            String prefix;
            for (int i = 0; i < prefixCount; i++)
            {
                // create a model node for prefix definition
                ((ModelBuilder)builder).extraNode(
                NodeKind.NAMESPACE,
                context.getStoragePathMetadata(),
                (String)declaredPrefixes.get(i),
                locator
                );
            }
            declaredPrefixes.clear();
        }
    }
    
    public void processingInstruction(String target, String data) throws SAXException
    {
        //Debug.print("processingInstruction");
        
        // create a node in the model
        try
        {
            ((ModelBuilder)builder).extraNode(NodeKind.PI, context.getStoragePathMetadata(), target + " " + data, locator);
        }
		catch(XMLDBCException e)
		{
			processFatalError(e);
		}
		catch(SQLException e)
		{
			processFatalError(e);
		}
    }
    
    protected boolean textBetweenTransitions()
    {
        boolean ret = textBetweenTransitions;
        
        // removes (if option set) white space data passed by the parser because there is no DTD
        if (removeWhitespace)
        {
            int i, max = textData.length();
            for (i = 0; i < max; i++)
                if (!Functions.isWhitespace(textData.charAt(i)))
                    break;
            if (i == max)// Only found WS : text ignored
            {
                textData.setLength(0);
                ret = false;
            }
        }
        return ret;
    }
          
    ////////////////////////////////////////////////
    //              STORAGE METHODS
    ////////////////////////////////////////////////
    protected StoragePathMetadata push(String namespaceURI, String localName)
    throws SAXException, RepositoryException
    {
        StoragePathMetadata pathNode = super.push(namespaceURI, localName);
        // Updating current path and send it to path set
        if (pathNode == null)
            throw new SAXException("Path {" + namespaceURI + "}" + localName
            + " could not be written in the underlying database. Storage is cancelled to avoid metadata corruption.");
        return pathNode;
    }
    
    /** Saves an XML document in this repository from a SAX 2.0 source. The
     * external SAX parser is responsible for parsing and validating documents.
     * @param reader The SAX XMLReader used for parsing the document.
     * The parser is assumed to be SAX 2.0 compliant.
     * @param source the URL or stream where document can be found.
     * @param ID The internal name used to identify the document in
     * the repository. If null, the internal generated numeric handle is used.
     * @return the ID of the document passed as a parameter.
     * @throws RepositoryException Application exception.
     */
    protected void startParsing() throws RepositoryException, SQLException
    {
        super.startParsing();
        
        // remove old user value
        if (!IDSetByUser)
            docID = null;
        
        // transmits document ID to model builder and get the OID
        docID = ((ModelBuilder)builder).allocateDocOID(docID);
    }
    
    ///////////////////////////////////////////////////////////////////////////
    //                     XMLDocumentFiler-like implementation
    ///////////////////////////////////////////////////////////////////////////
    public void setDocumentId(String ID) throws XMLDBCException
    {
        if ((ID != null) && (ID.length() > ID_LENGTH))
            throw new RepositoryException(RepositoryException.ILLEGAL_EXPRESSION, "The maximum size authorized for document ID ("
            + ID_LENGTH + ") is exceeded.");
        
        if (ID != null)
        {
            docID = ID.trim();
            IDSetByUser = true;
        }
    }
    
    public XMLCollection getCollection()
    {
        return collection;
    }
    
    public String getDocumentId()
    {
        return docID;   // do not use the storage context value cos' it is reset at the end of parsing
    }
    
    public void setXMLProperty(String propertyId, Object value) throws XMLDBCNotRecognizedException, XMLDBCNotSupportedException
    {
        throw new XMLDBCNotSupportedException("Not supported yet");
        // TO ADD : batch size for instance
    }
    
    public Object getXMLProperty(String propertyId) throws XMLDBCNotRecognizedException
    {
        throw new UnsupportedOperationException("Not supported yet");
        // TO ADD : batch size for instance
    }
    
    public String[] getXMLFeatureList()
    {
        throw new UnsupportedOperationException("Not supported yet");
        // TO ADD : removeWS for instance
    }
    
    public void setXMLFeature(String featureId, boolean state) throws XMLDBCNotRecognizedException, XMLDBCNotSupportedException
    {
        throw new XMLDBCNotSupportedException("Not supported yet");
        // TO ADD : removeWS for instance
    }
    
    public boolean getXMLFeature(String featureId) throws XMLDBCNotRecognizedException
    {
        throw new UnsupportedOperationException("Not supported yet");
        // TO ADD : removeWS for instance
    }
    
    public String[] getXMLPropertyList()
    {
        throw new UnsupportedOperationException("Not supported yet");
        // TO ADD : batch size for instance
    }
    
}


