/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

/*
 * DocumentGenerator.java
 *
 * Created on 12 mars 2001, 18:01
 */
package org.xquark.mapper.storage;

import java.io.IOException;
import java.sql.SQLException;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;
import org.xml.sax.*;
import org.xml.sax.ext.LexicalHandler;
import org.xquark.mapper.RepositoryConnection;
import org.xquark.mapper.RepositoryException;
import org.xquark.mapper.metadata.PathNode;
import org.xquark.mapper.metadata.PathSet;
import org.xquark.util.DOMBuilder;
import org.xquark.xml.xdbc.XMLDBCException;

/** This object is used to generate an XML document stored in a repository
 * with an XML schema using the SAX2 interface.
 *
 * <p>Generation is guided by the schema and indecisions are removed by
 * document instance information (stored in the structure table).</p>
 * @see org.xquark.mapper.XMLCollection
 */
public class DocumentProducer extends BaseXMLReaderImpl
implements Locator
{
    private static final String RCSRevision = "$Revision: 1.3 $";
    private static final String RCSName     = "$Name: XQ1_1_1 $";
    
    public static DocumentBuilderFactory DOMFactory;
    static
    {
        DOMFactory = DocumentBuilderFactory.newInstance();
        DOMFactory.setNamespaceAware(true);
    }
    
    private final static String DEFAULT_PREFIX = "ns";

    private DocumentExplorer docTuplePool;
    
    private PathSet pathSet;
    private String Id;
    
    /** Creates new DocumentGenerator */
    public DocumentProducer(_RepositoryCollection collection) throws XMLDBCException
    {
        super(collection);
        try
        {
            /* Preparing database queries */
            tuplePool = docTuplePool = new DocumentExplorer(collection);
        }
        catch (SQLException e)
        {
            throw new RepositoryException(RepositoryException.DB_ERROR,
                                    "JDBC error while initializing reader.");
        }
        generateDBPrefix = !collection.getFeature(RepositoryConnection.USE_SCHEMA_PREFIXES_FEATURE);
        prefixProvider = new PrefixProvider();
        pathSet = collection.getMetadata().getPathSet();
        try {
            docTuplePool.initDocReading(collection.getRepositoryConnection().getConnection());
        }
        catch (SQLException e) {
            throw new RepositoryException(RepositoryException.DB_ERROR,
            "JDBC error while initializing prepared statements", e);
        }
    }
    
    ////////////////////////////////////////////////////////////////////////////////
    //                      XMLReader IMPLEMENTATION
    ////////////////////////////////////////////////////////////////////////////////
    
    /**
     * Parse an XML document from a system identifier (URI).
     *
     * @param systemId The system identifier (URI).
     * @exception org.xml.sax.SAXException Any SAX exception, possibly
     *            wrapping another exception.
     * @exception java.io.IOException An IO exception from the parser,
     *            possibly from a byte stream or character stream
     *            supplied by the application.
     * @see #parse(org.xml.sax.InputSource)
     */
    public void parse(String systemId)
    throws IOException, SAXException
    {
        if (systemId == null)
            throw new IOException("A string ID must be provided in order to retrieve the document.");
        
        // remove useless spaces
        Id = systemId.trim();
        try
        {
            try
            {
                docTuplePool.initDocumentExploration(Id);
                
                // initiate row lookup
                if (!docTuplePool.lookup())
                    throw new RepositoryException(RepositoryException.NOT_EXISTS,
                    "Document with ID " + Id + " not found in this collection");
                
                contentHandler.setDocumentLocator(this);
                
                ////////////////////////
                //  SAX startDocument
                ////////////////////////
                contentHandler.startDocument();
                
                browseBranch
                        (
                        0,
                        Long.MAX_VALUE,
                        -1,
                        -1,
                        pathSet.createIterator
                                (
                                (PathNode)pathSet.getRoot()
                                ),
                        0,
                        false
                        );
                
                ////////////////////////
                //  SAX endDocument
                ////////////////////////
                contentHandler.endDocument();
                reset();
            }
            catch(SQLException e)
            {
                throw new SAXException("JDBC error raised while generating SAX events.", e);
            }
            catch(XMLDBCException e)
            {
                throw new SAXException("Repository error raised while generating SAX events.", e);
            }
        }
        catch (SAXException e)
        {
            if (getErrorHandler() != null)
                getErrorHandler().fatalError(
                new SAXParseException(e.getMessage(), this, e.getException()));
            try {
                reset();
            }
			catch (SQLException ex) {
			}catch (RepositoryException ex) {
                // no op
            }
            throw e;
        }
    }
    
    /** Builds a DOM Level 2 document from an XML document stored in the
     * repository base on an empty document passed as a paramater.
     * @param ID The name used to identify the document in the repository
     * that can be, for instance, the document URL.
     * @return a DOM Level 2 document.
     * @throws RepositoryException Application exception.
     */
    public Document getDocument(String ID) throws RepositoryException
    {
        Document document = null;
        try {
            document = DOMFactory.newDocumentBuilder().newDocument();
        }
        catch (ParserConfigurationException e) {
            throw new RepositoryException(RepositoryException.PARSER_ERROR, "Could not create an empty DOM level 2 document using JAXP. Check a JAXP DOM implementation is in your classpath.");
        }
        
        // SAX to DOM converter instantiation
        DOMBuilder builder =  new DOMBuilder(document);
        
        // save handlers plugged by user
        ContentHandler initialContentHandler = getContentHandler();
        LexicalHandler initialLexicalHandler;
         try {
            initialLexicalHandler = (LexicalHandler)getProperty(SAX_LEXICAL_PROPERTY);
        }
        catch (SAXException e) {
            throw new RepositoryException(RepositoryException.INTERNAL_ERROR, "Couldn't access the lexical handler in XMLReaderImpl.",e);
        }
        
        setContentHandler(builder);
        try {
            setProperty(SAX_LEXICAL_PROPERTY, builder);
        }
        catch (SAXException e) {
            throw new RepositoryException(RepositoryException.INTERNAL_ERROR, "Couldn't access the lexical handler in XMLReaderImpl.",e);
        }
        
        try {
            parse(ID);
        }
        catch (IOException e) {
            throw new RepositoryException(RepositoryException.INTERNAL_ERROR, "Error while processing internal SAX parsing.",e);
        }
		catch (SAXException e) {
			throw new RepositoryException(RepositoryException.INTERNAL_ERROR, "Error while processing internal SAX parsing.",e);
		}
        
        // restore handlers plugged by user
        setContentHandler(initialContentHandler);
        try {
            setProperty(SAX_LEXICAL_PROPERTY, initialLexicalHandler);
        }
        catch (SAXException e) {
            throw new RepositoryException(RepositoryException.INTERNAL_ERROR, "Couldn't access the lexical handler in XMLReaderImpl.",e);
        }
        
        return document;
    }
    
    public void reset() throws SQLException, RepositoryException
    {
        Id = null;
        super.reset();
    }
    
    public java.lang.String getSystemId(){return Id;}
    public java.lang.String getPublicId(){return null;}
    public int getLineNumber() {return -1;}
    public int getColumnNumber() {return -1;}
    
}
