/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.mapper.storage;

import java.sql.SQLException;
import java.util.List;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.xquark.mapper.RepositoryException;
import org.xquark.mapper.dbms.AbstractConnection;
import org.xquark.mapper.dbms.TableInfo;
import org.xquark.mapper.mapping.ColumnMapping;
import org.xquark.mapper.metadata.PathNode;
import org.xquark.mapper.metadata.RepositoryConstants;
import org.xquark.mapper.util.RepositoryProperties;

/**
 * Object used to perform XML structure node storage.
 *
 */
public class InteractiveQueryExtraDataExplorer extends QueryExtraDataExplorer
{
    private static final String RCSRevision = "$Revision: 1.3 $";
    private static final String RCSName     = "$Name: XQ1_1_1 $";
    
    private static Log log = LogFactory.getLog(InteractiveQueryExtraDataExplorer.class);
    
    private int pathCount = 0;
    private boolean useStringDelimitor = false;
    
    public InteractiveQueryExtraDataExplorer(TableInfo tableInfo)
    {
        super(tableInfo);
    }
    
    void prepareStatement(List paths, AbstractConnection connection)
    throws SQLException, RepositoryException
    {
        useStringDelimitor = connection.useStringDelimitor();
        pStmt = connection.getConnection().prepareStatement(generateStatement(paths)); // TO CHECK : could keep the statement ?
        connection.setFetchSize(pStmt, RepositoryProperties.getIntProperty(RepositoryConstants.CONF_DATA_FETCHSIZE));
    }
    
    private String generateStatement(List paths) throws RepositoryException
    {
        StringBuffer SQLStmt = new StringBuffer();
        
        ColumnMapping refColumn;
        SubPathInfo path;
        
        pathCount = paths.size();
        for (int i = 0; i < pathCount; i++)
        {
            path = (SubPathInfo)paths.get(i);
            
            if (i > 0)
                SQLStmt.append("\nunion all\n");
            
            SQLStmt.append("select ");
            SQLStmt.append(path.subPathIndex);
            SQLStmt.append(" as n,v.uoid,v.path,v.type,v.pos,v.offset,v.data,v.num\nfrom ");
            SQLStmt.append(tableInfo.getName());
            SQLStmt.append(" v\nwhere v.uoid between ? and ? and ?=");
            SQLStmt.append(path.varPath.getPathID());
            
            // path restriction
            List subpaths = path.subPath.getSubPaths();
            if (subpaths != null)
            {
                // last struct ancestor is added in order for the builder to initiate range
                SQLStmt.append(" and v.path");
                
                int subpathsCount = subpaths.size();
                if (subpathsCount == 1)
                {
                    SQLStmt.append("=");
                    SQLStmt.append(((PathNode)subpaths.get(0)).getPathID());
                }
                else
                {
                    SQLStmt.append(" in(");
                    
                    for (int j = 0; j < subpathsCount; j++)
                    {
                        SQLStmt.append(((PathNode)subpaths.get(j)).getPathID());
                        SQLStmt.append(',');
                    }
                    SQLStmt.setCharAt(SQLStmt.length()-1, ')');
                }
            }
        }
        
        SQLStmt.append("\norder by n,num");
        
        if (log.isDebugEnabled())
        	log.debug("Value stmt for extra table:\n" + SQLStmt);
        
        return SQLStmt.toString();
    }
    /**
     * get the next row from database and store it in the current object.
     */
    public boolean fetchNextRow() throws SQLException
    {
        try
        {
            if (resultActive())
            {
                notExhausted = rs.next();
                if (notExhausted)
                {
                    subPathIndex = rs.getInt(1);
                    uoid = rs.getLong(2);
                    node.path = rs.getShort(3);
                    node.type = rs.getShort(4);
                    node.position = rs.getShort(5);
                    node.offset = rs.getInt(6);
                    node.data = rs.getString(7);
                    if (useStringDelimitor)
                        node.data = node.data.substring(0, node.data.length() - STRING_DELIMITOR_LENGTH);
                    node.rowNum = rs.getShort(8);
                    updateResultSegmentationFlag();
                }
                else
                    reset();
            }
            else
                return false;
            
            return notExhausted;
        }
        catch (SQLException e)
        {
            close();
            throw e;
        }
    }
    
    /** To be used by query reconstruction     */
    void execQuery(long first, long last, short path) throws SQLException, RepositoryException
    {
        try
        {
            if (rs != null)
                rs.close();
            for (int i = 0; i < pathCount; i++)
            {
                pStmt.setLong(3*i + 1, first);
                pStmt.setLong(3*i + 2, last);
                pStmt.setShort(3*i + 3, path);
            }
            rs = pStmt.executeQuery();
            currentResultExhausted = false;
            currentSubPathIndex = 0;
            notExhausted = true;
        }
        catch (SQLException e)
        {
            pStmt.close();
            throw e;
        }
    }
    
}
