/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.mapper.storage;

import java.sql.SQLException;
import java.util.List;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.xquark.mapper.RepositoryException;
import org.xquark.mapper.dbms.AbstractConnection;
import org.xquark.mapper.dbms.TableInfo;
import org.xquark.mapper.mapping.ColumnMapping;
import org.xquark.mapper.metadata.PathNode;
import org.xquark.mapper.metadata.RepositoryConstants;
import org.xquark.mapper.metadata.UOIDManager;
import org.xquark.mapper.util.RepositoryProperties;

/**
 * Object used to perform XML structure node storage.
 *
 */
public class InteractiveQueryStructExplorer extends QueryStructExplorer
{
    private static final String RCSRevision = "$Revision: 1.3 $";
    private static final String RCSName     = "$Name: XQ1_1_1 $";
    
    private static Log log = LogFactory.getLog(InteractiveQueryStructExplorer.class);
    
    private int pathCount = 0;
    
    public InteractiveQueryStructExplorer(TableInfo tableInfo, UOIDManager manager)
    {
        super(tableInfo, manager);
    }
    
    /**
     * get the next row from database and store it in the current object.
     */
    public boolean fetchNextRow() throws SQLException
    {
        try
        {
            if (resultActive())
            {
                notExhausted = rs.next();
                if (notExhausted)
                {
                    subPathIndex = rs.getInt(1);
                    uoid = rs.getLong(2);
                    last = rs.getLong(3);
                    path = rs.getShort(4);
                    updateResultSegmentationFlag();
                }
                else
                    reset();
            }
            else
                return false;
            
            return notExhausted;
        }
        catch (SQLException e)
        {
            close();
            throw e;
        }
    }
    
    /** To be used by query reconstruction    */
    void execQuery(long first, long last, short path) throws SQLException, RepositoryException
    {
        if (pStmt == null)
            return;
        try
        {
            if (rs != null)
                rs.close();
            for (int i = 0; i < pathCount; i++)
            {
                pStmt.setLong(4*i + 1, first >>> manager.getDIDShift());
                pStmt.setLong(4*i + 2, first);
                pStmt.setLong(4*i + 3, last);
                pStmt.setShort(4*i + 4, path);
            }
            rs = pStmt.executeQuery();
            currentResultExhausted = false;
            currentSubPathIndex = 0;
            notExhausted = true;
        }
        catch (SQLException e)
        {
            pStmt.close();
            throw e;
        }
    }
    
    void prepareStatement(List paths, AbstractConnection connection) throws SQLException, RepositoryException
    {
        String SQL = generateStatement(paths);
        if (SQL != null)
        {
            pStmt = connection.getConnection().prepareStatement(SQL); // TO CHECK : could keep the statement ?
            connection.setFetchSize(pStmt, RepositoryProperties.getIntProperty(RepositoryConstants.CONF_TREE_FETCHSIZE));
        }
        else
            pStmt = null;
    }
    
    private String generateStatement(List paths) throws RepositoryException
    {
        StringBuffer SQLStmt = new StringBuffer();
        
        ColumnMapping refColumn;
        SubPathInfo path;
        
        for (int i = 0; i < paths.size(); i++)
        {
            path = (SubPathInfo)paths.get(i);
            List subpaths = path.subPath.getModelSubPaths();
            int subpathsCount = subpaths.size();
            if ((subpathsCount > 0) && !path.isText)
            {
                pathCount++;
                if (SQLStmt.length() > 0)   // i was not a correct flag because index 0 coul be discarded
                    SQLStmt.append("\nunion all\n");
                
                SQLStmt.append("select ");
                SQLStmt.append(path.subPathIndex);
                SQLStmt.append(" as n,v.uoid,v.last,v.path\nfrom ");
                SQLStmt.append(tableInfo.getName());
                SQLStmt.append(" v\nwhere v.bucket=? and v.uoid between ? and ? and ?=");
                SQLStmt.append(path.varPath.getPathID());
                // path restriction
                // last struct ancestor is added in order for the builder to initiate range
                SQLStmt.append(" and v.path");
                
                if (subpathsCount == 1)
                {
                    SQLStmt.append("=");
                    SQLStmt.append(((PathNode)subpaths.get(0)).getPathID());
                }
                else
                {
                    SQLStmt.append(" in(");
                    
                    for (int j = 0; j < subpathsCount; j++)
                    {
                        SQLStmt.append(((PathNode)subpaths.get(j)).getPathID());
                        SQLStmt.append(',');
                    }
                    SQLStmt.setCharAt(SQLStmt.length()-1, ')');
                }
            }
        }
        
        if (SQLStmt.length() > 0)
        {
            SQLStmt.append("\norder by n,uoid");
            
            if (log.isDebugEnabled())
            	log.debug("Value stmt for struct table:\n" + SQLStmt);
            
            return SQLStmt.toString();
        }
        else
            return null;
    }
    
}
