/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

/*
 * LogicalStructureBuilder.java
 *
 * Created on 25 juillet 2001, 11:45
 */

package org.xquark.mapper.storage;

import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import org.xml.sax.Locator;
import org.xquark.mapper.RepositoryException;
import org.xquark.mapper.dbms.*;
import org.xquark.mapper.mapping.ColumnMapping;
import org.xquark.mapper.metadata.*;
import org.xquark.mapper.util.RecyclingStack;
import org.xquark.mapper.util.RepositoryProperties;
import org.xquark.schema.SchemaConstants;
import org.xquark.schema.validation.ElementPSVInfoset;
import org.xquark.schema.validation.PSVInfoset;
import org.xquark.schema.validation.SchemaValidationContext;
import org.xquark.xml.xdbc.XMLDBCException;
import org.xquark.xpath.NodeKind;

/**
 *
 */
class ModelBuilder extends AbstractModelBuilder
{
    private static final String RCSRevision = "$Revision: 1.2 $";
    private static final String RCSName     = "$Name: XQ1_1_1 $";
    
    private static final short NO_STRUCT = StoragePathMetadata.DISCARDED|StoragePathMetadata.DATA_ONLY;
    private static final short POTENTIAL_DATA = StoragePathMetadata.POTENTIAL_STRUCT|StoragePathMetadata.DATA_AND_STRUCT|StoragePathMetadata.DATA_ONLY;
    
    protected PreparedStatement idStmt;
    
    private long lastOID = -1;
    private long nodeCount = 0;
    private long docDataSize = 0;
    
    private long DID = -1;
    private long UDID = -1;
    
    private int extraDataMaxSize = -1;
    private String docID = null;
    private short rootPosition = 0;
    private short rowNum = 0;
    private ExtraNode wExtraNode = new ExtraNode();
    private UOIDManager manager;
    
    // XQuark index persistency
    StructSaver nodeRow;
    ExtraDataSaver extraNodeRow;
    
    // Mapping information variables
    List textTableMappingList = null;
    List textColumnMappingList = null;
    
    public ModelBuilder(_RepositoryCollection collection, SchemaValidationContext schemaContext)
    throws XMLDBCException
    {
        super(collection.getRepositoryConnection().getConnection(), schemaContext);
        try
        {
            initBatcher(batcher, collection.getMetadata());
        }
        catch (SQLException e)
        {
            throw new RepositoryException(RepositoryException.DB_ERROR, "JDBC error while creating statement for documents IDs.", e);
        }
        AbstractConnection conn = collection.getRepositoryConnection().getConnection();
        CollectionMetadata meta = collection.getMetadata();
        manager = meta.getUOIDManager();
        storageBuffer = new StorageBuffer
                                        (
                                        meta,
                                        meta.getCollectionName(),
                                        conn,
                                        batcher,
                                        this,
                                        !conn.distinguishNullAndEmptyStrings()
                                        );
        ColumnMapping textColumnMapping = meta.createDefaultMappingView(null);
        textColumnMappingList = Collections.singletonList(textColumnMapping);
        textTableMappingList = Collections.singletonList(textColumnMapping.getTableMapping());
        extraDataMaxSize = collection.getInfo().getMaxExtraDataLength();
    }
    
    private void initBatcher(JDBCBatcher batcher, CollectionMetadata collection) throws SQLException
    {
        // doc list table
        TableInfo table = collection.getTableInfo(TableSpec.TYPE_DOCS);
        idStmt = connection.getConnection().prepareStatement(table.getInsertStatement());
        batcher.addStatement(idStmt, table.getName(), RepositoryProperties.getIntProperty(CONF_DOCID_BATCHSIZE));
        
        // structure table
        table = collection.getTableInfo(TableSpec.TYPE_STRUCT);
        nodeRow = new StructSaver(table, collection.getUOIDManager());
        batcher.addStatement(nodeRow.getStatement(connection), table.getName(),
        RepositoryProperties.getIntProperty(CONF_TREE_BATCHSIZE));
        
        // extra-data table
        table = collection.getTableInfo(TableSpec.TYPE_EXTRA);
        extraNodeRow = new ExtraDataSaver(table);
        batcher.addStatement(extraNodeRow.getStatement(connection), table.getName(),
        RepositoryProperties.getIntProperty(CONF_TREE_BATCHSIZE));
    }
    
    public String allocateDocOID(String docID) throws RepositoryException
    {
        DID = manager.getNewDocOID();
        UDID = manager.buildUDID(DID);
        if (docID == null)
            docID = Long.toString(DID);
        this.docID = docID;
        nodeRow.set(DID);
        extraNodeRow.set(DID);
        return this.docID;
    }
    
    /**
     * For PI, comments and prefix definition.
     **/
    public void extraNode(byte type, StoragePathMetadata path, String data, Locator locator)
    throws XMLDBCException, SQLException
    {
        if ((data != null) && (data.length() > extraDataMaxSize))
        {
            String msg;
            switch (type)
            {
                case NodeKind.COMMENT:
                    msg = "Comment discarded";
                    break;
                case NodeKind.PI:
                    msg = "Processing instruction discarded";
                    break;
                default:
                    msg = "Extra data discarded";
            }
            if (getXMLErrorHandler() != null)
                getXMLErrorHandler().warning(new RepositoryException(RepositoryException.EXTRA_DATA_LOSS, msg));
        }
        else
        {
            if (path.getPathID() == ROOT_PATH_ID)
                wExtraNode.set(manager.buildUOID(DID, 0), ++rowNum, ROOT_PATH_ID, type, ++rootPosition, 0, data);
            else
                wExtraNode.set(getAnchor(), ++rowNum, path.getPathID(), type,
                type == NodeKind.NAMESPACE ? 0 : getCurrentModelNode().incModelPosition(),
                getCurrentModelNode().getCharOffset(), data);
                batcher.addBatch(extraNodeRow.setParameters(wExtraNode), locator.getLineNumber(), locator.getColumnNumber());
        }
        // update stats
        nodeCount++;
        if  (data != null)
            docDataSize += data.length();
    }
    
    protected AbstractModelNode getLogicalStructureNode(byte type, StoragePathMetadata path, String data, Locator locator)
    throws SQLException, XMLDBCException
    {
        List nodeTableMappings = null, nodeColumnMappings = null;
        short modelPosition = 0;
        short rank = 0;
        
        // process mappings for elements & attributes & increment parent child counters
        ModelNode parent = getCurrentModelNode();
        switch (type)
        {
            case NodeKind.TEXT:
                nodeTableMappings = textTableMappingList;
                modelPosition = parent.incModelPosition();    // save for model child ordering
                break;
                
            case NodeKind.ELEMENT:
                if (lastOID != -1)
                    getCurrentModelNode().flushExtraNode();  // because xsi attributes adding is achieved now
                // no need to wait endElement for non-leaf elements
                nodeTableMappings = path.getTableMappings();
                nodeColumnMappings = path.getColumnMappings();
                if (lastOID == -1)  // root element
                    rootPosition++;
                else
                {
                    rank = parent.incChildCount();
                    modelPosition = parent.incModelPosition();    // save for model child ordering
                }
                break;
                
            case NodeKind.ATTRIBUTE:
                nodeTableMappings = path.getTableMappings();
                nodeColumnMappings = path.getColumnMappings();
                parent.incAttCount();
            break;
                
            default :
        }
    
        // allocate or recycle a node
        ModelNode node = (ModelNode)stack.push();
        node.set(path, type, data, rank, modelPosition, locator);
        
        if (path.isPersistent() || (type == NodeKind.TEXT))   // OID allocation
        {
            // Now that parent has a children with UOID, if POTENTIAL_STRUCT, change its path sign
           
            if ((parent != null) && (parent.getPath().getStorageMode() == StoragePathMetadata.POTENTIAL_STRUCT))
                parent.schedulePathSignToggle();
            
            if (type == NodeKind.TEXT)
                node.schedulePathSignToggle();
            
            node.setOID(++lastOID);
            
            // Process table mappings (create tupleNode)
            // the MIXED/leaf stay without column mapping (and potentially no table mapping)
            // till we know if they are model leaf
            if (nodeTableMappings != null)
                node.setTupleNode(storageBuffer.bufferizeNode(nodeTableMappings, locator));
            
        }
        else if (!path.isPositionEnforced())
            node.addPositionExtraNode();
        
        // Process column mappings (set dependencies)
        if (nodeColumnMappings !=  null)
        {
            // Create dependance links between tupleNode (references : up to now we
            // consider that ref => (FK,PK) and thus a tuple must no be stored
            // before its master.
            Iterator it = nodeColumnMappings.iterator();
            ColumnMapping column = null;
            int refIndex;
            while (it.hasNext())
            {
                column = (ColumnMapping)it.next();
                refIndex = column.getTableRefIndex();
                if (refIndex != -1)
                    storageBuffer.getTupleFactory(column.getTableIndex()).addMasterTuple(storageBuffer.getTupleFactory(refIndex).getTuple());
            }
        }
        
        // update stats
        nodeCount++;
        
        return node;
    }
    
    protected AbstractModelNode finalizeNode(Locator locator)
    throws XMLDBCException, SQLException
    {
        ModelNode node = getCurrentModelNode();
        
        // switch path sign to negative if detected necessary in a child: made her because context is pertinent
        node.togglePathSignInData();
        
        // set last OID of the range
        node.last = lastOID;
        
        // process mappings for elements & attributes
        List nodeColumnMappings = null;
        StorageBuffer.BufferNode tupleNode = node.getTupleNode();
        PathMetadata path = (PathMetadata)node.getPath();
        
        switch (node.type)
        {
            case NodeKind.TEXT:
                nodeColumnMappings = textColumnMappingList;
                break;
                
            case NodeKind.ELEMENT:
                // by default
                nodeColumnMappings = node.getPath().getColumnMappings();
                
                if ((nodeColumnMappings == null) && node.isLeaf())
                {
                    // add default mapping but with potential attributes
                    if ((path.getStorageMode() & POTENTIAL_DATA) != 0)
                    {
                        if (path.getReadColumnMapping() != null)
                        {
                            // set the default mapping typed)
                            nodeColumnMappings = Collections.singletonList(path.getReadColumnMapping());
                            if (tupleNode == null)
                                tupleNode = storageBuffer.bufferizeDefaultNode(path.getReadTableMapping(), locator);
                            else
                                tupleNode.addDefaultTuple(path.getReadTableMapping(), locator);
                            break;
                        }
                        else if (path.isMixed())        // add default mapping for mixed nodes
                        {
                            // set the default mapping (untyped)
                            nodeColumnMappings = textColumnMappingList;
                            if (tupleNode == null)
                                tupleNode = storageBuffer.bufferizeDefaultNode(path.getTextTableMapping(), locator);
                            else
                                tupleNode.addDefaultTuple(path.getTextTableMapping(), locator);
                            break;
                        }
                    }
                }
                break;
                
            case NodeKind.ATTRIBUTE:
                nodeColumnMappings = node.getPath().getColumnMappings();
                if (nodeColumnMappings == null)
                {
                    nodeColumnMappings = Collections.singletonList(path.getReadColumnMapping());;
                    tupleNode = storageBuffer.bufferizeDefaultNode(path.getReadTableMapping(), locator);
                }
                break;
            default :
        }
        
        // process column mappings (simple type elements or attributes)
        if (nodeColumnMappings != null)
        {
            Iterator it = nodeColumnMappings.iterator();
            ColumnMapping column = null;
            
            while (it.hasNext())
            {
                column = (ColumnMapping)it.next();
                storageBuffer.getTupleFactory(column.getTableIndex()).getTuple().addColumnMappingData(column, this);
            }
        }
        
        // Complete table mappings
        if (tupleNode != null)
            tupleNode.complete();
        
        // save struct if any
        node.flush();
        
        // TO IMPROVE save data : could be done on mapping complete
        if (node.type == NodeKind.ELEMENT)
        {
            // set tuple to complete, only waiting for dependences
            // TO IMPLEMENT : check before endElement that no subelement is still waited for an element completion
            //storageBuffer.getTupleFactory(index).setCompleted();
            
            // try to flush completed tupleNode when an element is ended
            // TO DO : refine it : do it only when necessary
            storageBuffer.flush();
        }
        if (node.isLeaf())
            node.flushExtraNode();  // because xsi:nil attribute is sure now
        
        // update stats (now that is in the stack
        if  (node.getData() != null)
            docDataSize += node.getData().length();

        return (ModelNode)stack.pop();
    }
    
    /* return the last OID allocated */
    public void endModel() throws XMLDBCException, SQLException
    {
        super.endModel();
        
        idStmt.setString(1, docID);
        idStmt.setLong(2, UDID);
        idStmt.setTimestamp(3, new java.sql.Timestamp(System.currentTimeMillis()));
        idStmt.setLong(4, nodeCount);
        idStmt.setInt(5, (int)(docDataSize/nodeCount));
        batcher.addBatch(idStmt, 0, 0);
    }
    
    public void reset()
    throws RepositoryException, SQLException
    {
        super.reset();
        lastOID = -1;
        nodeCount = 0;
        docDataSize = 0;
        rootPosition = 0;
        rowNum = 0;
        DID = -1;
        UDID = -1;
        docID = null;
        wExtraNode.clear();
    }
    
    public void close() throws RepositoryException, SQLException
    {
        super.close();
        idStmt.close();
        idStmt = null;
        nodeRow.close();
        extraNodeRow.close();
    }
    
    public ModelNode getCurrentModelNode()
    {
        return (ModelNode)stack.top();
    }
    
    ////////////////////////////////////////////////////////////////////
    //          IMPLEMENTATION of StorageContext interface.
    ////////////////////////////////////////////////////////////////////
    public long getBucketOID() { return DID;}
    public short getPathOID()
    {
        ModelBuilder.ModelNode node = getCurrentModelNode();
        if (node.negativePath())
            return (short)-node.path;
        else
            return node.path;
    }
    public String getDocumentID()
    { return docID;}
    public long getDocumentOID()
    { return UDID;}
    public long getOID()
    {return getCurrentModelNode().oid;}
    public long getUOID()
    { return manager.buildUOID(DID, getCurrentModelNode().oid);}
    public RecyclingStack.StackObject newStackObject()
    {
        return new ModelNode();
    }
    
    //
    //                  PRIVATE UTILITIES
    //
    
    private long getAnchor()
    {
        ModelNode wNode;
        for (int i = stack.size() - 1; i >= 0 ; i--)
        {
            wNode = (ModelNode)stack.get(i);
            if (wNode.oid >= 0)
                return manager.buildUOID(DID, wNode.oid);
        }
        return -1;
    }
    
    ////////////////////////////////////////////////////////////////////
    //                  STORAGE MODEL NODE
    ////////////////////////////////////////////////////////////////////
    public class ModelNode extends AbstractModelNode
    implements RecyclingStack.StackObjectFactory
    {
        private static final String RCSRevision = "$Revision: 1.2 $";
        private static final String RCSName     = "$Name: XQ1_1_1 $";
        private int charOffset =0;
        private RecyclingStack extraNodes = null; // for positionning or nil
        protected short attCount = 0; // attributes count
        private short extraDataCount = 0; // number of extra-nodes (PI & comments + elements) in containing element
        private short modelPosition = 0;
        private boolean negativePath = false;
        private boolean scheduleToggle = false;
        int line = -1;
        int column = -1;
       
        ModelNode()
        {}
        
        void set(StoragePathMetadata path, byte type, String data, short rank, short modelPosition, Locator locator)
        {
            set(path, type, data, rank);
            setModelPosition(modelPosition);
            setLocation(locator);
        }
        
        public void clear()
        {
            super.clear();
            if (extraNodes != null)
                extraNodes.clear();
            charOffset = 0;
            attCount = 0;
            extraDataCount = 0;
            modelPosition = 0;
            negativePath = false;
            scheduleToggle = false;
            line = -1;
            column = -1;
        }

        public void setLocation(Locator locator)
        {
            if (locator != null)
            {
                line = locator.getLineNumber();
                column = locator.getColumnNumber();
            }
        }
        
        void setModelPosition(short pos)
        {
            modelPosition = pos;
        }

        boolean negativePath()
        {
            return negativePath;
        }

        void schedulePathSignToggle()
        {
            negativePath = true;
            if (tupleNode != null)  // if tuple has not been initialized yet, no need to toggle
                scheduleToggle = true;
        }
        
        void togglePathSignInData()
        {
            if (scheduleToggle && (tupleNode != null))
                tupleNode.changePathSign();
        }
        
        void addPositionExtraNode()
        {
            if (extraNodes == null)
                extraNodes = new RecyclingStack(this);
            ExtraNode extra = (ExtraNode)extraNodes.push();
            extra.set(getAnchor(), ++rowNum, pathMetadata.getPathID(), type, modelPosition);
        }
        void flushExtraNode() throws SQLException, RepositoryException
        {
            if (extraNodes != null)
            {
               ExtraNode extra;
               while((extra = (ExtraNode)extraNodes.pop()) != null)
               {
                   batcher.addBatch(extraNodeRow.setParameters(extra), line, column);
               }
           }
        }
        void flush() throws SQLException, RepositoryException
        {
            if ((getNodeType() != NodeKind.TEXT) && ((pathMetadata.getStorageMode() & NO_STRUCT) == 0)
            && !((pathMetadata.getStorageMode() == StoragePathMetadata.POTENTIAL_STRUCT) && isEmptyRange()))
                 batcher.addBatch(nodeRow.setParameters(this), line, column);
        }

        short incAttCount()
        {
            return ++attCount;
        }
        
        boolean isModelLeaf()
        {
            return ((childCount + attCount) == 0);
        }

        boolean isEmptyRange()
        {
            return (oid == last);
        }

        void setOID(long oid)
        {
            this.oid = oid;
        }
        
        public boolean setXSIinfo(String attName, String value)
        throws RepositoryException
        {
            boolean nil = super.setXSIinfo(attName, value);
            if (nil && pathMetadata.isNilImplicit())   // no extra node needed
                return nil;
            
            ExtraNode xsiNode;
            if ((extraNodes == null) || (extraNodes.size() == 0))
                addPositionExtraNode();
            xsiNode = (ExtraNode)extraNodes.get(0);
            if (xsiNode.type >= RepositoryConstants.XSI_NIL)   // position node cannot be reused
            {
                xsiNode = (ExtraNode)extraNodes.push();
                xsiNode.set(lastOID, pathMetadata.getPathID(), type, modelPosition);
            }
            // use schema to convert value
            ElementPSVInfoset infoSet = validationContext.getCurrentInfoset();
            PSVInfoset attInfoSet = infoSet.getAttributePSVInfoset(SchemaConstants.XSI_URI, attName);
            // set type & data
            if (nil)
            {
                xsiNode.type = RepositoryConstants.XSI_NIL;
                xsiNode.data = attInfoSet.getNormalizedValue().toString();
            }
            else if (attName.equals(SchemaConstants.XSI_TYPE_ATTR))
            {
                if ((pathMetadata.getReferenceColumnMapping() != null)
                || (
                    (pathMetadata.getTableMappings() != null)
                    && (pathMetadata.getTableMappings().size() > 0)
                    ))
                    throw new RepositoryException(RepositoryException.NOT_ALLOWED, 
                    "Dynamic XML schema type binding (xsi:type) is not supported for mapped elements or attributes.");
                xsiNode.type = RepositoryConstants.XSI_TYPE;
                xsiNode.data = attInfoSet.getActualValue().toString();
            }
            else if (attName.equals(SchemaConstants.XSI_SCHEMA_LOCATION_ATTR))
            {
                xsiNode.type = RepositoryConstants.XSI_SCHEMA_LOCATION;
                xsiNode.data = attInfoSet.getNormalizedValue().toString();
            }
            else if (attName.equals(SchemaConstants.XSI_NO_NAMESPACE_SCHEMA_LOCATION_ATTR))
            {
                xsiNode.type = RepositoryConstants.XSI_NO_NAMESPACE_SCHEMA_LOCATION;
                xsiNode.data = attInfoSet.getNormalizedValue().toString();
            }
            
//            xsiNode.data = value;
            return nil;
        }

        public void incCharOffset(int length)
        {
            charOffset += length;
        }
        
        public short incModelPosition()
        {
            return ++extraDataCount;
        }
        
        int getCharOffset()
        {
            return charOffset;
        }
        
        public RecyclingStack.StackObject newStackObject()
        {
            return new ExtraNode();
        }
    }
    
}
