/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.mapper.storage;

import java.sql.*;
import java.util.NoSuchElementException;
import java.util.StringTokenizer;

import org.xquark.mapper.RepositoryConnection;
import org.xquark.mapper.RepositoryException;
import org.xquark.mapper.dbms.AbstractConnection;
import org.xquark.mapper.dbms.TableInfo;
import org.xquark.mapper.metadata.CollectionInfo;
import org.xquark.mapper.metadata.RepositoryConstants;
import org.xquark.xml.xdbc.Configurable;
import org.xquark.xml.xdbc.XMLDBCException;
import org.xquark.xml.xdbc.XMLDBCNotRecognizedException;

/** Data structure that corresponds to the storage table for collection
 * configuration and is used when creating a collection with the
 * repository object.
 * <p>This object contains all its JDBC calls and owns a connection
 * in order to perform database access on his own</p>
 */

public class PersistentCollectionInfo implements RepositoryConstants
{
    private static final String RCSRevision = "$Revision: 1.2 $";
    private static final String RCSName     = "$Name: XQ1_1_1 $";
    
    /** XMLCollection IDs */
    private String collectionName = null;
    private short CID = -1;
    private short parentCID = 0;
    private short maxCID = -1;
    
    private CollectionInfo config;
    private TableInfo configTable;
    private String selectLockStmt;

    /////////////////////////////////////
    //  Storage JDBC variables
    /////////////////////////////////////

    /** Connection object used by this data object. */
    private AbstractConnection connection;

    /** java.sql objects used to fetch the whole data set from the database. */
    private Statement stmt = null;
    private ResultSet rs = null;

    /** Constructor. This constructor should be used for listing collectionss
     *  an reading configuration.
     * @param connection the AbstractConnection object used by this object.
     */
    public PersistentCollectionInfo(TableInfo configTable, AbstractConnection connection)
    {
        this.config = new CollectionInfo(false);
        this.connection = connection;
        this.configTable = configTable;
        initStatements();
    }

    /** Constructor used for collection creation.
     * @param name the name of the repository.
     * @param config a CollectionInfo object.
     * @param connection the JDBC Connection object used by this object.
     * @throws RepositoryException application exception.
     * @see org.xquark.mapper.CollectionInfo#repositoryType
     */
    public PersistentCollectionInfo(
                                    TableInfo configTable,
                                    String name,
                                    CollectionInfo config,
                                    short maxCID,
                                    AbstractConnection connection
                                    ) 
    {
        this.connection = connection;
        this.configTable = configTable;
        this.config = config;
        this.collectionName = name;
        this.maxCID = maxCID;
        initStatements();
    }
    
    public PersistentCollectionInfo(
                                    TableInfo configTable,
                                    String name,
                                    Configurable config,
                                    short maxCID,
                                    AbstractConnection connection
                                    ) 
    {
        this(configTable, name, new CollectionInfo(config), maxCID, connection);
    }

    private void initStatements()
    {
        selectLockStmt = connection.getLockSelectStatement("CID", configTable.getName(), "ORDER BY CID");
    }

    /** Returns the string name of the repository.
     * @return the repository name.
     */
    public String getCollectionName()
    {
        return this.collectionName;
    }

    public short getCollectionID()
    {
        return CID;
    }

    public CollectionInfo getCollectionInfo()
    {
        return config;
    }
    
    /** Sets the name of the repository.
     * @param name the collection name.
     */
    public void setCollectionName(String name)
    {
        collectionName = name;
    }

    public void setCollectionID(short id)
    {
        CID = id;
    }

    /** Builds a String in order to displays the information of this configuration.
     * @return the string representing the repository features.
     */
    public String toString()
    {
        StringBuffer buffer = new StringBuffer(super.toString());

        buffer.append("Name:\t\t");
        buffer.append(collectionName);
        buffer.append("\nId :\t\t");
        buffer.append(CID);
        buffer.append("\n");
        
        return buffer.toString();
    }

    /**
     * Save this object data into a config table row with batch style.
     * Automatic flush is proceeded when max size is reached
     * @throws SQLException if a database access error occurs.
     */
    public void createRow()
    throws SQLException, RepositoryException
    {
        PreparedStatement pStmt = null;
        Statement stmt = null;
        ResultSet rs =null;
        int wCID = RepositoryConstants.SEQUENCE_FIRST_VALUE;

        // create a transaction for keeping locks
        connection.start();
        
        try
        {
            // select & lock the whole collection table
            stmt = connection.getConnection().createStatement();
            rs = stmt.executeQuery(selectLockStmt);
            
            while (rs.next())
            {
                if (rs.getShort(1) != wCID)
                    break;
                wCID++;
            }
            rs.close();
            rs = null;
            stmt.close();
            stmt = null;
            
            if (wCID > maxCID)
                throw new RepositoryException(RepositoryException.OUT_OF_RANGE,
                    "Maximum number of collections is reached");
            else
                CID =(short)wCID;
            
            // preparation of PreparedStatement
            pStmt = connection.getConnection().prepareStatement(configTable.getInsertStatement());
            pStmt.setString(1, collectionName.toUpperCase());
            pStmt.setShort(2, CID);
            pStmt.setShort(3, parentCID);
            pStmt.setByte(4, config.getType());
            pStmt.setString(5, config.getDescription());
            pStmt.setString(6, config.getMappingID());
            pStmt.setByte(7, config.getDocOIDSize());
            pStmt.setByte(8, config.getDocOIDSize());    // bucketSize = docSize
            pStmt.setShort(9, config.getDocOIDPreallocationSize());    // bucketSize = docSize
            pStmt.setString(10, generateConfigString());
            pStmt.executeUpdate();
            
        }
        catch (SQLException e)
        {
            connection.rollback();
            throw new RepositoryException(RepositoryException.DB_ERROR,
                "JDBC error while creating a new entry in collection table", e);
        }
        finally
        {
            connection.commit();
            if (rs != null)
                rs.close();
            if (stmt != null)
                stmt.close();
            if (pStmt != null)
                pStmt.close();
        }
    }

    /** Free all resources (JDBC)
     * @throws RepositoryException application error
     */
    public void close()
    {
        // release statement
        try {
            if (rs != null)
                rs.close();
            if (stmt != null)
                stmt.close();
            connection = null;
        }
        catch(SQLException e) {
            // no op
        }
    }
    
    public void retrieveDataSet()
    throws SQLException, RepositoryException
    {
        retrieveDataSet((short)-1);
    }
    /**
     * Launch the full table entry retrieving.
     * @param all if true, even system repositories will be retrieved
     * @throws SQLException if a database access error occurs.
     */
    public void retrieveDataSet(short type)
    throws SQLException, RepositoryException
    {
        stmt = connection.getConnection().createStatement();
        String query = configTable.getSelectAllStatement();
        if (type >= 0)
            query += " WHERE TYPE=" + type;
        try {
            rs = stmt.executeQuery(query);
        }
        catch (SQLException e) {
            stmt.close();
            throw e;
        }
    }

    /** get the next row from database and store it in the current object.
     * @throws SQLException if a database access error occurs.
     * @return true, if a row was fetched, false if no more rows.
     */
    public boolean fetchNextRow()
    throws SQLException, XMLDBCException
    {
        if (stmt == null)
            return false;
        try {
            if (fetchNextResult(rs))
                return true;
            else
            {
                config.setWriteProtected(true);
                rs.close();
                stmt.close();
                stmt = null;    // prevent closing 2 times
                return false;   // If no more row found
            }
        }
        catch (SQLException e) {
            rs.close();
            stmt.close();
            throw e;
        }
    }

    /** get row from the database corresponding to the collection ID passed and store it in he current object.
     * @param name the name of the collection
     * @throws SQLException if a database access error occurs.
     * @return true, if a row was fetched, false if no more rows.
     */
    public boolean fetchRow(String name)
    throws SQLException, XMLDBCException
    {
        Statement stmt = connection.getConnection().createStatement();
        ResultSet rs = null;
        try {
            rs = stmt.executeQuery(configTable.getSelectAllStatement()
                                    + " WHERE NAME='" + name + "'");
            
            if (fetchNextResult(rs))
            {
                config.setWriteProtected(true);
                return true;    // Take first row encountered
            }
            else
                return false;   // If no row found
        }
        finally
        {
            if (rs != null)
                rs.close();
            stmt.close();
        }
    }

    /**
     * remove row from database corresponding to the collection ID passed.
     * @param name the name of the collection
     * @throws SQLException if a database access error occurs.
     */
    public void deleteRow(String name)
    throws SQLException
    {
        connection.executeUpdate("DELETE FROM " + configTable.getName()
            + " WHERE NAME='" + name + "'");
    }
    
    public void deleteRow()
    throws SQLException
    {
        deleteRow(collectionName);
    }
    
    public void rename(String oldName, String newName)
    throws SQLException
    {
        connection.executeUpdate("UPDATE " + configTable.getName()
            + " SET NAME='" + newName + "' WHERE NAME='" + oldName + "'");
    }
    
    /** Analyse a configuration string and fills object variables.
     * @param configString the configuration string to parse.
     */
    private void parseConfigString(String configString)
    throws XMLDBCException
    {
        String keyword;
        StringTokenizer st = new StringTokenizer(configString, "<");
        while (st.hasMoreTokens())
        {
            try {
                keyword = st.nextToken();
                if (keyword.startsWith("maxTextDataLength"))
                    config.setProperty(RepositoryConnection.TEXT_LENGTH_PROPERTY, 
                        keyword.substring(keyword.indexOf('>') + 1));
                else if (keyword.startsWith("maxExtraDataLength"))
                    config.setProperty(RepositoryConnection.EXTRA_DATA_LENGTH_PROPERTY, 
                        keyword.substring(keyword.indexOf('>') + 1));
                else if (keyword.startsWith("useSchemaPrefixes"))
                    config.setFeature(RepositoryConnection.USE_SCHEMA_PREFIXES_FEATURE,
                        Boolean.valueOf(keyword.substring(keyword.indexOf('>') + 1)).booleanValue());
            }
            catch (NoSuchElementException e) {
                // no op
            }
        }
    }
    
    /** Produces a configuration string ready to be stored in a text columm.
     * @return the configuration string.
     */
    private String generateConfigString()
    {
        StringBuffer buf = new StringBuffer();
        try
        {
            buf.append("<maxTextDataLength>");
            buf.append(config.getProperty(RepositoryConnection.TEXT_LENGTH_PROPERTY));
            buf.append("<maxExtraDataLength>");
            buf.append(config.getProperty(RepositoryConnection.EXTRA_DATA_LENGTH_PROPERTY));
            buf.append("<useSchemaPrefixes>");
            buf.append(config.getFeature(RepositoryConnection.USE_SCHEMA_PREFIXES_FEATURE));
        }
        catch (XMLDBCNotRecognizedException e)
        {
            // no op : internal use
            throw new RuntimeException("Internal error: unknown feature or property.");
        }
        
        return buf.toString();
    }
    
    private boolean fetchNextResult(ResultSet rs)
    throws SQLException, XMLDBCException
    {
        if(rs.next())
        {
            collectionName = rs.getString(1);
            CID = rs.getShort(2);
            parentCID = rs.getShort(3);
            config.setType(rs.getByte(4));
            config.setDescription(rs.getString(5));
            config.setMappingID(rs.getString(6));
            config.setDocOIDSize(rs.getByte(7));
            config.setDocOIDPreallocationSize(rs.getShort(9));
            parseConfigString(rs.getString(10));
            return true;
        }
        else
            return false;   // If no more row found
    }
}
