/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.mapper.storage;

import java.sql.*;

import org.xquark.mapper.dbms.AbstractConnection;
import org.xquark.mapper.dbms.TableInfo;
import org.xquark.mapper.metadata.DocumentInfo;

/** Data structure that corresponds to the storage table for XML document
 * information used by the CollectionMetadata object.
 * <p>This object contains all its JDBC calls and owns a connection
 * in order to perform database access on his own</p>
 */

public class PersistentDocumentInfo
{
    private static final String RCSRevision = "$Revision: 1.2 $";
    private static final String RCSName     = "$Name: XQ1_1_1 $";
    
    private DocumentInfo wDocInfo;
    private TableInfo docTable;
    private String docInfoStmt;

    /////////////////////////////////////
    //  Storage JDBC variables
    /////////////////////////////////////

    /** Connection object used by this data object. */
    private AbstractConnection connection;

    /** java.sql objects used to fetch the whole data set from the database. */
    private Statement stmt = null;
    private ResultSet rs = null;

    public PersistentDocumentInfo(TableInfo docTable, AbstractConnection connection)
    {
        this.connection = connection;
        this.docTable = docTable;
        initStatements();
    }

    private void initStatements()
    {
        StringBuffer sql = new StringBuffer();
        sql.append(docTable.getSelectAllStatement());
        sql.append(" WHERE ");
        sql.append(docTable.getColumns()[0].getName());
        sql.append("=?");
        docInfoStmt = sql.toString();
    }

    public DocumentInfo getDocumentInfo()
    {
        return wDocInfo;
    }

     /**
     * Save this object data into a config table row with batch style.
     * Automatic flush is proceeded when max size is reached
     * @throws SQLException if a database access error occurs.
     */
    public void save()
    throws SQLException
    {
        // preparation of PreparedStatement
        PreparedStatement pStmt = connection.getConnection().prepareStatement(docTable.getInsertStatement());
        try
        {
            pStmt.setString(1, wDocInfo.docID);
            pStmt.setLong(2, wDocInfo.udid);
            pStmt.setTimestamp(3, wDocInfo.creationDate);
            pStmt.setLong(4, wDocInfo.nodeCount);
            pStmt.setInt(5, wDocInfo.avgNodeLen);
            pStmt.executeUpdate();
        }
        finally
        {
            pStmt.close();
        }
    }

    /** Free all resources (JDBC)
     * @throws RepositoryException application error
     */
    public void close()
    {
        // release statement
        try {
            if (rs != null)
                rs.close();
            if (stmt != null)
                stmt.close();
            connection = null;
        }
        catch(SQLException e) {
            // no op
        }
    }
    
    public void retrieveDataSet()
    throws SQLException
    {
        stmt = connection.getConnection().createStatement();
        try {
            rs = stmt.executeQuery(docTable.getSelectAllStatement());
        }
        catch (SQLException e) {
            stmt.close();
            throw e;
        }
    }

    /** get the next row from database and store it in the current object.
     * @throws SQLException if a database access error occurs.
     * @return true, if a row was fetched, false if no more rows.
     */
    public boolean fetchNextRow()
    throws SQLException
    {
        if (stmt == null)
            return false;
        try {
            if (fetchNextResult(rs))
                return true;
            else
            {
                rs.close();
                stmt.close();
                stmt = null;    // prevent closing 2 times
                return false;   // If no more row found
            }
        }
        catch (SQLException e) {
            rs.close();
            stmt.close();
            throw e;
        }
    }

    public boolean fetchRow(String docID)
    throws SQLException
    {
        PreparedStatement stmt = null;
        ResultSet rs = null;
        try {
            stmt = connection.getConnection().prepareStatement(docInfoStmt);
            stmt.setString(1, docID);
            rs = stmt.executeQuery();
            
            if (fetchNextResult(rs))
                return true;    // Take first row encountered
            else
                return false;   // If no row found
        }
        finally
        {
            if (rs != null)
                rs.close();
            if (stmt != null)
                stmt.close();
        }
    }

    private boolean fetchNextResult(ResultSet rs)
    throws SQLException
    {
        if(rs.next())
        {
            wDocInfo = new DocumentInfo(
                                rs.getString(1), 
                                rs.getLong(2), 
                                rs.getTimestamp(3),
                                rs.getLong(4),
                                rs.getInt(5)
                                );
            return true;
        }
        else
            return false;   // If no more row found
    }
}
