/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

/*
 * DocumentGenerator.java
 *
 * Created on 12 mars 2001, 18:01
 */

package org.xquark.mapper.storage;

import java.sql.SQLException;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.xml.sax.ContentHandler;
import org.xml.sax.ErrorHandler;
import org.xml.sax.SAXException;
import org.xml.sax.ext.LexicalHandler;
import org.xquark.mapper.RepositoryException;
import org.xquark.mapper.mapping.TableMapping;
import org.xquark.mapper.metadata.MetadataIterator;
import org.xquark.mapper.metadata.PathMetadata;
import org.xquark.mapper.metadata.PathNode;
import org.xquark.xml.xdbc.XMLDBCException;
import org.xquark.xpath.NodeKind;

/** 
 * 
 *
 * <p>Generation is guided by the schema and indecisions are removed by
 * document instance information (stored in the structure table).</p>
 * @see org.xquark.mapper.XMLCollection
 */
public abstract class QueryBuilder extends XMLBuilder
{
    private static final String RCSRevision = "$Revision: 1.3 $";
    private static final String RCSName     = "$Name: XQ1_1_1 $";

    private static Log log = LogFactory.getLog(QueryBuilder.class);
    
    private VarInfoSet varInfo;
    protected long varFirst = -1, varLast = -1;
    protected int currentRank = -1;
    protected short currentPath = -1;
    
    private int subPathCount = 0;
    
    protected QueryExplorer qTuplePool;
    
    /** To be used by query reconstruction */
    protected QueryBuilder(_RepositoryCollection collection, VarInfoSet varInfo) throws XMLDBCException
    {
        super(collection);
        prefixProvider = new PrefixProvider(varInfo.getNamespaceContextStack());
        this.varInfo = varInfo;
    }
    
    public void setSAXHandlers
    ( 
                        ContentHandler contentHandler,
                        LexicalHandler lexicalHandler,
                        ErrorHandler errorHandler
                        )
    {
        this.contentHandler = contentHandler;
        this.lexicalHandler = lexicalHandler;
        this.errorHandler = errorHandler;
    }
    
    public void newVariableRange(int rank, long first, long last, short path)
    throws XMLDBCException
    {
        if (log.isDebugEnabled())
        	log.debug("newVariableRange [rank, first, last, path]="
        			+ rank + ',' + first + ',' + last + ',' + path);
        currentRank = rank;
        currentPath = path;
        varFirst = first;
        varLast = last;
        if (subPathCount == qTuplePool.getSubPathCount())
            subPathCount = 0;
        if (subPathCount == 0)
            initNewResult(rank);
    }
    
    protected abstract void initNewResult(int rank) throws XMLDBCException;
    
    public boolean generateSubPath(PathNode path) throws XMLDBCException, SQLException, SAXException
    {
    	if (log.isDebugEnabled())
    		log.debug("generateSubPath [path]=" + path);
        if (path.getType() == NodeKind.ATTRIBUTE)
        {
            String attributeValue = fetch(path);
            contentHandler.characters(attributeValue.toCharArray(), 0, attributeValue.length());
            return true;
        }
        
        subPathCount++;
        qTuplePool.freezeCurrentSubPath();
        
        // create metadata iterator
        MetadataIterator metaContext = pathSet.createIterator(path);
        
        Collection unconsumedTuples; 
        if (path.getReadColumnMapping() == null)
           unconsumedTuples = path.getBuilderAncestor().getBuildMappings();
        else
           unconsumedTuples = Collections.singletonList(path.getReadColumnMapping().getTableMapping());
        
        // Process multivaluated paths
        long uoid = NO_UOID_FOUND, last = NOT_FOUND, anchor = NO_UOID_FOUND;
        while (((uoid = scanBrother(varFirst, varLast, path)) > last) || (uoid == -1))    // else entering a new result
        {
            if (uoid == NO_UOID_FOUND)  // DISCARDED potentially multivalued: consume tuple
            {
                anchor =  getElementUOID(varFirst, varFirst, (PathNode)path.getModelAncestor());
                if (anchor <= NO_UOID_FOUND)
                    anchor = varFirst;
            }
            else
                anchor = uoid;
            
            browseBranch(varFirst, varLast, uoid, anchor, metaContext, 0, false);
            
            if ((uoid == NO_UOID_FOUND)  // DISCARDED potentially multivalued: consume tuple
                && !consumeTuple(unconsumedTuples))
                break;
                 
            // memorize uoid
            last = uoid;
        }
        // removing useless tuples basing on result segmentation
        // Note: table mappings out of subpath contain data common to all
        //       multivaluated subpaths. This is the reason why consuming is
        //       performed at the end without impact.
        consumeTuples(unconsumedTuples);
        return false;
    }
    
    /** Check if a node with the same path is present in results (multivaluated elements).
     * @return -1 if not found. UOID else.
     */
    protected long scanBrother(long first, long last, PathNode path)
    throws SQLException
    {
        long uoid = NOT_FOUND;
        short pid = path.getPathID();
        StructExplorer structNode = null;
        Tuple nextDataTuple = null;
        
        switch (path.getStorageMode())
        {
            case PathMetadata.STRUCT_ONLY:
            case PathMetadata.DATA_AND_STRUCT:
                // look in the struct table
                structNode = qTuplePool.getStructNode();
                if (structNode.isInRange(first, last, pid))
                    uoid = qTuplePool.getStructNode().getUOID();
                break;
            case PathMetadata.DATA_ONLY:
                nextDataTuple = qTuplePool.get(path.getReadTableMapping().getTableMapping().getTableIndex());
                if (nextDataTuple.isInRange(first, last, pid))
                    uoid = nextDataTuple.getUOID();
                break;
            case PathMetadata.POTENTIAL_STRUCT:
                long dataUOID = NO_UOID_FOUND, structUOID = NO_UOID_FOUND;
                // look in struct
                structNode = qTuplePool.getStructNode();
                if (structNode.isInRange(first, last, pid))
                    structUOID = qTuplePool.getStructNode().getUOID();
                // look in data
                if (path.isMixed())
                    nextDataTuple = qTuplePool.get(path.getTextTableMapping().getTableMapping().getTableIndex());
                else if (path.getReadTableMapping() != null)
                    nextDataTuple = qTuplePool.get(path.getReadTableMapping().getTableMapping().getTableIndex());
                if ((nextDataTuple != null) && nextDataTuple.isInRangeAbs(first, last, pid))
                    dataUOID = nextDataTuple.getUOID();
                // take the least
                if (dataUOID == NO_UOID_FOUND)
                {
                    if (structUOID >= 0)
                        uoid = structUOID;
                }
                else if (structUOID == NO_UOID_FOUND)
                    uoid = dataUOID;
                else
                    uoid = Math.min(structUOID, dataUOID);
                break;
            case PathMetadata.DISCARDED:
                uoid = NO_UOID_FOUND;
            default:
        }
        
        return uoid;
    }
   
    public String fetch(PathNode path)
    throws SQLException, XMLDBCException
    {
        if (log.isDebugEnabled()) log.debug("fetch [path]=" + path);
        
        subPathCount++;
        qTuplePool.freezeCurrentSubPath();
        
        // Process multivaluated paths
        long uoid = -1, last = -2;
        StringBuffer value = new StringBuffer();
        String fetchedValue;
        while ((uoid = scanBrother(varFirst, varLast, path)) > last)   // else entering a new result
        {
            fetchedValue = getMappedValue(path, varFirst, varLast, true, true, true);
            if (fetchedValue != null)
                value.append(fetchedValue);
            
            // memorize uoid
            last = uoid;
        }
        return value.toString();
    }
    
    private void consumeTuples(Collection tableMappings) throws SQLException
    {
        if (tableMappings != null)
        {
            Iterator it = tableMappings.iterator();
            Tuple tuple;
            while (it.hasNext())
            {
                tuple = qTuplePool.get(((TableMapping)it.next()).getTableIndex());
                // exhaust tuples
                while(tuple.fetchNextRow());
            }
        }
    }
    
    private boolean consumeTuple(Collection tableMappings) throws SQLException
    {
        if (tableMappings != null)
        {
            Iterator it = tableMappings.iterator();
            Tuple tuple;
            while (it.hasNext())
            {
                tuple = qTuplePool.get(((TableMapping)it.next()).getTableIndex());
                // exhaust tuples
                return tuple.fetchNextRow();
            }
        }
        return false;
    }
    
    public void reset() throws RepositoryException, SQLException
    {
        super.reset();
        varFirst = -1;
        varLast = -1;
        currentRank = -1;
        subPathCount = 0;
    }
}
