/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.mapper.storage;

import java.sql.SQLException;

import org.xml.sax.*;
import org.xml.sax.ext.LexicalHandler;
import org.xml.sax.helpers.DefaultHandler;
import org.xquark.mapper.RepositoryException;
import org.xquark.mapper.dbms.AbstractConnection;
import org.xquark.mapper.metadata.PathNode;
import org.xquark.mapper.metadata.RepositoryConstants;
import org.xquark.mapper.metadata.StoragePathMetadata;
import org.xquark.mapper.util.RepositoryProperties;
import org.xquark.schema.SchemaConstants;
import org.xquark.util.SAXConstants;
import org.xquark.xml.xdbc.XMLDBCException;
import org.xquark.xml.xdbc.XMLErrorHandler;
import org.xquark.xpath.NodeKind;

/** Class using SAX2 event handlers for building the optimized XML logical structure.
 *
 * <ul>
 * <li>{@link org.xml.sax.ContentHandler ContentHandler}</li>
 * </ul>
 *
 * @see org.xml.sax.ContentHandler
 */
public class SAXHandler extends DefaultHandler
implements RepositoryConstants, LexicalHandler
{
    private static final String RCSRevision = "$Revision: 1.2 $";
    private static final String RCSName     = "$Name: XQ1_1_1 $";

    ///////////////////////////////////////////
    //                  Data
    ///////////////////////////////////////////
    
    //
    //  storage data
    //
    
    /** JDBC related data for persistent storage. */
    protected AbstractConnection connection;
    protected boolean performCommit; // used instead of XMLConnection.getConnection() to work with Bridge
    
    //
    //  Options
    //
    protected boolean removeWhitespace = false;   // if true removes all character data containing only whitespaces
    
    /* Algorithm variables */
    public static final short UNKNOWN = 0;
    static final short START = 0;
    static final short END =1;
    protected short lastEvent = UNKNOWN; // last event was start or end element ?
    protected boolean textBetweenTransitions = false; // to determine if text is to be saved
    protected PathIterator context;  // Working node
    protected StringBuffer textData;
    protected AbstractModelBuilder builder;
    protected Locator locator;
    private boolean nsFeatureNotTested = true;    // feature must be true
    
    
    /** Constructor
     */
    public SAXHandler(PathIterator metadataContext, AbstractConnection connection, AbstractModelBuilder builder)
    throws XMLDBCException
    {
        context = metadataContext;
        this.connection = connection;
        this.builder = builder;
        textData = new StringBuffer(256);
    }
    
    /**
     * Set the removeWhitespace option which if true removes all character data
     * containing only whitespaces
     *
     * @param value new value of the option.
     */
    public void setRemoveWhitespace(boolean value)
    {
        this.removeWhitespace = value;
    }
    
    public void reset() throws RepositoryException
    {
        context.reset();
        try {
            builder.reset();
        }
        catch (SQLException e) {
            throw new RepositoryException(RepositoryException.DB_ERROR,
                "JDBC error while resetting dbms resources", e);
        }
        lastEvent = UNKNOWN;
        textData.setLength(0);
        textBetweenTransitions = false;
        nsFeatureNotTested = true;
    }
    
    private void clear() throws XMLDBCException
    {
        reset();
        builder.clearBuffer();
    }
    
    public void close() throws XMLDBCException
    {
         if (context != null)
        {
            context = null;
            try
            {
                builder.close();
            }
            catch (SQLException e)
            {
                throw new RepositoryException(RepositoryException.DB_ERROR, "JDBC error while closing storage buffer.", e);
            }
            builder = null;
        }
    }
    
    /**
     * Perform JDBC statement cleaning before the object is destroyed.
     *
     * <p> Cleans what is to be cleaned. Store the tag (tag) dictionary</p>
     *
     * @exception org.xml.sax.SQLException if a database access error occurs.
     * @see org.xml.sax.ContentHandler#endDocument
     */
    public void clean()
    throws SQLException
    {
        // Save document longest path set to repository longest path set and database ? MADE UP TO NOW BY REPOSITORY.
    }
    
    ////////////////////////////////////////////////////////////////////
    //    Implementation of ErrorHandler interface.
    ////////////////////////////////////////////////////////////////////
    public void warning(SAXParseException ex) throws SAXException
    {
        processFatalError(new SAXException("Warning at line " + ex.getLineNumber()
                          + " in " + ex.getSystemId() + ": " + ex.getMessage()));
    }
    
    public void error(SAXParseException ex) throws SAXException
    {
        processFatalError(new SAXException("Error at line " + ex.getLineNumber()
                          + " in " + ex.getSystemId() + ": " + ex.getMessage()));
    }
    
    public void fatalError(SAXParseException ex) throws SAXException
    {
        processFatalError(new SAXException("Fatal error at line " + ex.getLineNumber()
                          + " in " + ex.getSystemId() + ": " + ex.getMessage()));
    }

    //////////////////////////////////////////////////////
    //      Implementation of LexicalHandler interface.
    //////////////////////////////////////////////////////
    public void startDTD(String name, String publicId, String systemId) throws SAXException
    {
        //Debug.print("startDTD");
    }

    public void endDTD()throws SAXException
    {
        //Debug.print("endDTD");
    }

    public void startEntity(String name) throws SAXException
    {
        //Debug.print("startEntity");
    }
    
    public void endEntity(String name) throws SAXException
    {
        //Debug.print("endEntity");
    }
    
    public void startCDATA() throws SAXException
    {
        //Debug.print("startCDATA");
    }
    
    public void endCDATA() throws SAXException
    {
        //Debug.print("endCDATA");
    }
    
    public void comment(char ch[], int start, int length) throws SAXException
    {
        //Debug.print("comment");
    }
    
    ////////////////////////////////////////////////////////////////////
    //    Implementation of ContentHandler interface.
    ////////////////////////////////////////////////////////////////////
    public void startDocument() throws SAXException
    {
        //Debug.print("startDocument");
        try
        {
            performCommit = connection.getConnection().getAutoCommit();
            if (performCommit)
                connection.start();    // START TRANSACTION
            startParsing();
        }
        catch(Exception e)
        {
            processFatalError(e);
        }
        
    }
    
    public void endDocument() throws SAXException
    {
        //Debug.print("endDocument");
        // Store the root node and perform flush
        
        try
        {
            builder.endModel();
            endParsing();
            if (performCommit)
            {
                flushBuffer();
                connection.commit();
            }
            // resource re-initialization
            reset();
        }
        catch(Exception e)
        {
            processFatalError(e);
        }
    }
    
    public void startPrefixMapping(String prefix, String uri)
    throws SAXException
    {
        // no op
    }
    
    public void startElement(String namespaceURI,String localName,String qName,Attributes atts)
    throws SAXException
    {
        // test SAX feature for namespaces
        if (nsFeatureNotTested)
        {
            if (localName.length() == 0)
                throw new SAXException("Your XML parser does not provide namespaces support. If you are using JAXP, make sure that setNamespaceAware(true) has been called on your parser factory.");
            nsFeatureNotTested = false;
        }
    
    
        //Debug.print("startElement element : " + qName);
        PathNode tmpPath;
        
        try
        {
            if (textBetweenTransitions())
            {
                // create a model node for "mixed" text
                builder.leafNode(NodeKind.TEXT, context.getStoragePathMetadata(), textData.toString(), locator);
                textData.setLength(0);
                textBetweenTransitions = false;
            }
            
            // create a model node and a path set node for the element
            builder.newNode(context.push(namespaceURI, localName), locator); // do not put the root path in the path set (already in)
            
            if (context.getStoragePathMetadata() != null) // not out of storage
            {
                //////////////////////////////////////////
                //          Prefixes storage
                //////////////////////////////////////////
                processPrefixes();
                
                //////////////////////////////////////////
                //          Attribute storage
                //////////////////////////////////////////
                int nbAtt = atts.getLength();
                StoragePathMetadata attNode;
                String uri; // because atts may be implemented by DOM2SAXAttrs that is dynamic
                
                // For each attribute
                for(int i = 0; i < nbAtt; i++)
                {
                    uri = atts.getURI(i);
                    
                    // discard namespace attributes (should not be : tolerance)
                    if (uri.equals(SAXConstants.XMLNS_URI))
                        continue;
                    
                    // Test if it is the xsi:null attribute (do not save)
                    if (uri.equals(SchemaConstants.XSI_URI))
                        builder.setXSIinfo(atts.getLocalName(i), atts.getValue(i));    // set element to null
                    else // save attribute
                    {
                        // add the attribute path to path set
                        attNode = context.createNode(uri, atts.getLocalName(i), NodeKind.ATTRIBUTE);
                        
                        // create a model node for attribute definition
                        if (attNode != null)    // may be null for Bridge
                            builder.leafNode(NodeKind.ATTRIBUTE, attNode, atts.getValue(i), locator);
                    }
                }
                lastEvent = START;
            }
        }
        catch(Exception e)
        {
            processFatalError(e);
        }
    }

    /* Detached to provide a control when a path is not found.
     * With Mapper it is allowed since there is no path set.
     */
    protected StoragePathMetadata push(String namespaceURI, String localName)
    throws SAXException, RepositoryException
    {
        return context.push(namespaceURI, localName);
    }

    
    protected void processPrefixes()
    throws XMLDBCException, SQLException
    {
        // no op
    }
    
    public void endElement(String namespaceURI,String localName,String qName) throws SAXException
    {
        try
        {
            if (context.getStoragePathMetadata() != null)
            {
                if (textBetweenTransitions())
                {
                    switch (lastEvent)
                    {
                        case START:
                            builder.setData(textData.toString());
                            break;
                        case END:
                            // create a model node for "mixed" text
                            builder.leafNode(NodeKind.TEXT, context.getStoragePathMetadata(),
                            textData.toString(), locator);
                            break;
                        default:
                    }
                    
                    textData.setLength(0);
                    textBetweenTransitions = false;
                }
                
                //  Achieve element node construction
                builder.finalizeNode(locator);
            }
            context.pop();
            lastEvent = END;
        }
        catch(Exception e)
        {
            processFatalError(e);
        }
        
    }
    
    public void characters(char ch[], int start, int length)
    throws SAXException
    {
        if (context.getStoragePathMetadata() != null)
        {
            //Debug.print("Characters");
            
            // Store the characters in the StringBuffer because as PI or Comment
            // are ignored from a value point of view
            textData.append(ch, start, length);
            
            // Offset for PI & comments update
            textBetweenTransitions=true;
            builder.incCharOffset(length);
        }
    }
    
    protected boolean textBetweenTransitions()
    {
        return textBetweenTransitions;
    }
          
    public void processingInstruction(String target, String data) throws SAXException
    {
        // no op
    }
    
    public void setDocumentLocator(Locator locator)
    {
        //Debug.print("setDocumentLocator");
        this.locator = locator;
    }
    
    ////////////////////////////////////////////////
    //              STORAGE METHODS
    ////////////////////////////////////////////////
    /** Saves an XML document in this repository from a SAX 2.0 source. The
     * external SAX parser is responsible for parsing and validating documents.
     * @param reader The SAX XMLReader used for parsing the document.
     * The parser is assumed to be SAX 2.0 compliant.
     * @param source the URL or stream where document can be found.
     * @param ID The internal name used to identify the document in
     * the repository. If null, the internal generated numeric handle is used.
     * @return the ID of the document passed as a parameter.
     * @throws RepositoryException Application exception.
     */
    protected void startParsing() throws RepositoryException, SQLException
    {
        setRemoveWhitespace(RepositoryProperties.getBooleanProperty(CONF_TRIM_WS));
    }
    
    protected void endParsing() throws Exception
    {
    }
    
    protected void processFatalError(Exception e) throws SAXException
    {
        try
        {
            connection.rollback();
             // restore JDBC autocommit mode
            if (performCommit)
                connection.getConnection().setAutoCommit(performCommit);
        }
        catch (SQLException ex)
        {
            // TO DO : change JDBC connection !
        }
        finally
        {
            try
            {
                clear();   // internal clear()
            }
            catch (XMLDBCException exc)
            {
                // no op
            }
        }
        if (e instanceof SAXException)
            throw (SAXException)e;
        if (e instanceof RepositoryException)
            throw new SAXException("Application error during document storage.", e);
        else if (e instanceof SQLException)
            throw new SAXException("JDBC error during document storage.", e);
        else    // e.g. RuntimeException
            throw new SAXException("Runtime error during document storage.", e);
    }
    
    
    ///////////////////////////////////////////////////////////////////////////
    //                     XDBC implementation methods
    ///////////////////////////////////////////////////////////////////////////
    
    public void setErrorHandler(XMLErrorHandler handler)
    {
        if (handler == null)
            throw new NullPointerException("An error handler must be provided if you use the setErrorHandler() method.");
        builder.setErrorHandler(handler);
    }
    
    public void flushBuffer() throws XMLDBCException
    {
        builder.flushBuffer();
    }
    
    public void clearBuffer() throws XMLDBCException
    {
        builder.clearBuffer();
    }
    
    public void setAutoFlush(boolean mode) throws XMLDBCException
    {
        builder.setAutoFlush(mode);
    }
    
    public boolean getAutoFlush()
    {
        return builder.getAutoFlush();
    }
    
    public XMLErrorHandler getXMLErrorHandler()
    {
        return builder.getXMLErrorHandler();
    }
}


