/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.mapper.storage;

import java.io.*;
import java.net.MalformedURLException;
import java.net.URL;

import javax.xml.parsers.SAXParserFactory;

import org.w3c.dom.Document;
import org.xml.sax.*;
import org.xquark.mapper.RepositoryException;
import org.xquark.schema.SchemaConstants;
import org.xquark.util.HandlerDecorator;
import org.xquark.util.SAXConstants;
import org.xquark.xml.xdbc.*;

/**
 * An XMLFilter implementation used to perform schema storage in a system repository
 * processing storage of included schemas files.
 */
public class SchemaSavingFilter extends HandlerDecorator
implements SchemaConstants, SAXConstants, XMLDocumentFiler
{
    private static final String RCSRevision = "$Revision: 1.2 $";
    private static final String RCSName     = "$Name: XQ1_1_1 $";
    
    private XMLReader schemaReader = null;
    private boolean topMost;
    private boolean firstElement = true;
    private SchemaSavingFilter inlineFilter = null;
    private XMLReader inlineReader = null;
    private URL baseURL = null;
    private SAXParserFactory parserFactory;
    private XMLDocumentFiler handler;
    
    public SchemaSavingFilter(XMLDocumentFiler handler)
    {
        this(SAXParserFactory.newInstance(), handler, true, null);
    }
    
    private SchemaSavingFilter(SAXParserFactory parserFactory, XMLDocumentFiler handler, boolean topMost, URL baseURL)
    {
        super(handler, handler);
        this.handler = handler;
        this.parserFactory = parserFactory;
        parserFactory.setNamespaceAware(true);
        this.topMost = topMost;
        this.baseURL = baseURL;
    }

    public void startPrefixMapping(String str, String str1) throws SAXException
    {
       // for the moment discard the prefix mapping related to the discarded root element
       if (topMost || !firstElement)
            super.startPrefixMapping(str, str1);
    }
    
    public void endPrefixMapping(String str)  throws SAXException
    {
       // for the moment discard the prefix mapping related to the discarded root element
       if (topMost || !firstElement)
            super.endPrefixMapping(str);
    }
    
    public void startElement(String namespaceURI, String localName, String qName, Attributes atts)
    throws SAXException
    {
        // include processing
        if (namespaceURI.equals(XMLSCHEMA_URI) && localName.equals(INCLUDE_TAG))
        {
            String schemaLocation = atts.getValue("", SCHEMA_LOCATION_ATTR);
            if (schemaLocation == null)
                throw new SAXException("Incorrect include element in schema : no schemaLocation attribute found.");
            processInlineParsing(schemaLocation);
        }
        else    // not the XML schema include element
            if (topMost)
            { // remove schema root element for included schemas
                if (firstElement)
                {
                    if (!(namespaceURI.equals(XMLSCHEMA_URI) && localName.equals(SCHEMA_TAG)))
                        throw new SAXException("This document is not a valid XML Schema.");
                    firstElement = false;
                }
                super.startElement(namespaceURI, localName, qName, atts);
            }
            else if (!(namespaceURI.equals(XMLSCHEMA_URI) && localName.equals(SCHEMA_TAG)))
            { // remove schema root element for included schemas
                super.startElement(namespaceURI, localName, qName, atts);
            }
    }
    
    public void endElement (String namespaceURI, String localName, String qName)
    throws SAXException
    {
        if (!(namespaceURI.equals(XMLSCHEMA_URI) && localName.endsWith(INCLUDE_TAG)) // BUG XERCES 1.1.x
        && (topMost || !(namespaceURI.equals(XMLSCHEMA_URI) && localName.endsWith(SCHEMA_TAG))))
            super.endElement(namespaceURI, localName, qName);
    }

    public void startDocument() throws SAXException
    {
        if (topMost)
            super.startDocument();
    }
    
    public void endDocument() throws SAXException
    {
        if (topMost)
        {
            super.endDocument();
            reset();
        }
    }

    private void reset()
    {
        firstElement = true;
    }
    private void processInlineParsing(String schemaLocation) throws SAXException
    {
        InputSource source = null;
        URL fileURL =null;
        
        try {   // try to resolve include using the schemaLocation as an absolute or relative URL
            fileURL = new URL(baseURL, schemaLocation);
            fileURL.openConnection().connect();
        }
        catch(Exception e) {
            try {   // try to resolve include using the schemaLocation as a file name
                File file = new File(schemaLocation);
                fileURL = file.toURL();
                fileURL.openConnection().connect();
            }
            catch (Exception ex) {
                if (baseURL == null)
                    throw new SAXException("The following schema location couln't be resolved : " + schemaLocation +
                                            ". If you are using relative URLs, please use the SAX interface and provide a system ID that will be used as a context.");
                else
                    throw new SAXException("The following schema location couln't be resolved : " + schemaLocation);
            }
        }
        /* create the SAX inputSource */
        try {
            source = new InputSource(new BufferedReader(new InputStreamReader(fileURL.openStream())));
        }
        catch (IOException e) {
            throw new SAXException("The following schema include couln't be opened even if it was located : " + schemaLocation);
        }            
        
        if (inlineReader == null)
        {
            try {
                inlineReader = parserFactory.newSAXParser().getXMLReader();
            }
            catch(Exception e) {
                throw new SAXException("Couldn't instantiate a new parser for included schema files.", e);
            }
            try {
                configureXMLReader(inlineReader);
            }
            catch(Exception e) {
                // no op
            }
        }
        if (inlineFilter == null)
        {
            inlineFilter = new SchemaSavingFilter(parserFactory, handler, false, baseURL);
        }
        inlineReader.setContentHandler(inlineFilter);
        try {
            inlineReader.setProperty(SAX_LEXICAL_PROPERTY, inlineFilter);
            inlineReader.parse(source);
        }
        catch(IOException e) {
            throw new SAXException("IO error while parsing included schema files.", e);
        }
    }
    
    public void setDocumentLocator(Locator locator)
    {
        String url = locator.getSystemId();
        try {
            if (url != null)
                baseURL = new URL(baseURL, url);
        }
        catch (MalformedURLException e) {
            // no op : keep the base URL null
        }
        super.setDocumentLocator(locator);
    }
    
    ///////////////////////////////////////////////////////////////////////////
    //                     XMLDocumentFiler implementation
    ///////////////////////////////////////////////////////////////////////////
    
    public XMLDocument insertDocument(XMLReader parser, InputSource input)
    throws XMLDBCException, SAXException
    {
        return handler.insertDocument(parser, input);
    }
    
    public XMLDocument insertDocument(InputSource input)
    throws XMLDBCException, SAXException
    {
        return handler.insertDocument(input);
    }
    
    public XMLDocument insertDocument(InputSource input, String id)
    throws XMLDBCException, SAXException
    {
        return handler.insertDocument(input, id);
    }
    
    public XMLDocument insertDocument(XMLReader parser, InputSource input, String id)
    throws XMLDBCException, SAXException
    {
        return handler.insertDocument(parser, input, id);
    }
    
    public XMLDocument insertDocument(String doc)
    throws XMLDBCException, SAXException
    {
        return handler.insertDocument(doc);
    }
    
    public XMLDocument insertDocument(String doc, String id) throws XMLDBCException, org.xml.sax.SAXException
    {
        return handler.insertDocument(doc, id);
    }
    
    public XMLDocument insertDocument(Document doc, String id) throws XMLDBCException
    {
        return handler.insertDocument(doc, id);
    }
    
    public XMLDocument insertDocument(Document doc) throws XMLDBCException
    {
        return handler.insertDocument(doc);
    }
    
    public void setDocumentId(String id) throws XMLDBCException
    {
        handler.setDocumentId(id);
    }
    
    public XMLCollection getCollection()
    {
        return handler.getCollection();
    }
    
    public void flushBuffer() throws XMLDBCException
    {
        handler.flushBuffer();
    }
    
    public String getDocumentId()
    {
        return handler.getDocumentId();
    }
    
    public void setErrorHandler(XMLErrorHandler handler)
    {
        this.handler.setErrorHandler(handler);
    }
    
    public XMLErrorHandler getErrorHandler()
    {
        return handler.getErrorHandler();
    }
    
    public void clearBuffer() throws XMLDBCException
    {
        handler.clearBuffer();
    }
    
    public void close() throws XMLDBCException
    {
        handler.close();
    }
    
    public boolean getAutoFlush()
    {
        return handler.getAutoFlush();
    }
    
    public void setAutoFlush(boolean mode) throws XMLDBCException
    {
        handler.setAutoFlush(mode);
    }
    
    public String[] getFeatureList()
    {
        return handler.getFeatureList();
    }
    
    public String[] getPropertyList()
    {
        return handler.getPropertyList();
    }
    
    public boolean getFeature(String featureId) throws XMLDBCNotRecognizedException
    {
        return handler.getFeature(featureId);
    }
    
    public void setFeature(String featureId, boolean state) throws XMLDBCNotRecognizedException, XMLDBCNotSupportedException
    {
        handler.setFeature(featureId, state);
    }
    
    public Object getProperty(String propertyId) throws XMLDBCNotRecognizedException
    {
        return handler.getProperty(propertyId);
    }
    
    public void setProperty(String propertyId, Object value) throws XMLDBCNotRecognizedException, XMLDBCNotSupportedException
    {
        handler.setProperty(propertyId, value);
    }
    ///////////////////////////////////////////////////////////////////////////
    //                            PRIVATE UTILITIES
    ///////////////////////////////////////////////////////////////////////////
    private XMLReader configureXMLReader(XMLReader reader) throws RepositoryException
    {
        try
        {
            reader.setFeature(SAX_NAMESPACE_FEATURE, true);
            reader.setFeature(SAX_PREFIX_FEATURE, false);
        }
        catch(SAXException e)
        {
            throw new RepositoryException(RepositoryException.PARSER_ERROR, "The XML SAX Parser is not SAX 2.0 compliant");
        }
        return reader;
    }
    
}
