/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.mapper.storage;

import java.io.StringWriter;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.*;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.xml.sax.Locator;
import org.xquark.mapping.Generator;
import org.xquark.mapping.StorageContext;
import org.xquark.mapper.RepositoryException;
import org.xquark.mapper.StorageException;
import org.xquark.mapper.dbms.AbstractConnection;
import org.xquark.mapper.dbms.JDBCBatcher;
import org.xquark.mapper.mapping.*;
import org.xquark.mapper.metadata.CollectionMappingInfo;
import org.xquark.mapper.metadata.Node;
import org.xquark.mapper.metadata.RepositoryConstants;
import org.xquark.mapper.util.*;
import org.xquark.mapper.util.Stack;
import org.xquark.xml.xdbc.XMLDBCException;
import org.xquark.xml.xdbc.XMLErrorHandler;

/** Bufferization class performing storage of XML model to database managing
 * relational referential integrity issues.
 *
 * @see org.xml.sax.ContentHandler
 */
public class StorageBuffer implements RepositoryConstants
{
    private static final String RCSRevision = "$Revision: 1.7 $";
    private static final String RCSName     = "$Name: XQ1_1_1 $";

    private static Log log = LogFactory.getLog(StorageBuffer.class);
    
    ////////////////////////////////////////////////////////////////////
    //                      DATA
    ////////////////////////////////////////////////////////////////////
    /* Array for storing one tuple for every relational table
     * index is managed by the repository mapping
     */
    TupleFactory[] tupleFactories = null;
    
    /* Stack containing ElementInfo to memorize that an element is mapped
     * and maintain info to maintains the repository context.
     * This allows to know if the text buffer must be emptied by the
     * endElement event or by the endMapping that needs buffer content.
     */
    Stack elementMappingStatus = null;
    
    // StorageContext data (for mapping generators)
    Node contextLink = null;   // link used to maintain the repository context
    AbstractConnection connection;
    
    JDBCBatcher batcher;
    
    private XMLErrorHandler errorHandler;

    private LinkedList waitingNodes = null; // BufferNode list waiting to be stored
    // OPTIMIZATION : LinkedList ?
    private ArrayList bufferPool = null; // BufferNode list waiting to be recycled
    private ArrayList modelPool = null; // ModelNode list waiting to be recycled

    private MappingSet mappings;
    private String collectionName;
    private StorageContext context;
    private boolean overrideEmptyString;
    //
    //  XMLCollection (database) related data
    //

    //
    //  Options
    //

    /** Creates new StorageBuffer */
    public StorageBuffer(MappingSet mappings, String collectionName, AbstractConnection connection,
            JDBCBatcher batcher, StorageContext context, boolean overrideEmptyString)
    throws RepositoryException
    {
        this.mappings = mappings;
        this.collectionName = collectionName;
        this.overrideEmptyString = overrideEmptyString;
        waitingNodes = new LinkedList();
        bufferPool = new ArrayList();
        this.context = context;
        this.batcher = batcher;
        tupleFactories = new TupleFactory[mappings.getMappingStatementsCount()];
		// TODO: Bug. When tables have FK but no ref are declared in mapping a useless
		// order is enforced. More generally, if no dependency exist, an order still
		// exist (blocking tuples...)
		// 2 things to do:
		// 1. After dependencies declared in mapping, a dependency tree could be build
		//    basing on DBDMS metadata verifying there is no conflict between the 2.
		// 2. Building list of lists dependency structures instead of simple lists (like
		//    in JDBC batcher) 
        try {
            initBatcher(connection);
        }
        catch (SQLException e) {
            throw new RepositoryException(RepositoryException.DB_ERROR,
            "JDBC error while creating JDBC2 batch handler", e);
        }
        
        this.elementMappingStatus = new Stack();
        this.connection = connection;
    }
    
    public void setErrorHandler(XMLErrorHandler handler)
    {
        errorHandler = handler;
    }
    
    public XMLErrorHandler getXMLErrorHandler()
    {
        return errorHandler;
    }
    
    ////////////////////////////////////////////////////////////////////
    //              Tuple buffering general methods.
    ////////////////////////////////////////////////////////////////////
    
    void reset() throws RepositoryException, SQLException
    {
        for (int i = 0; i < tupleFactories.length; i++)
        {
            tupleFactories[i].clear();
        }
        waitingNodes.clear();
        bufferPool.clear();
    }
    
    public void close() throws RepositoryException, SQLException
    {
        batcher = null;
        contextLink = null;
        connection = null;
        for (int i = 0; i < tupleFactories.length; i++)
        {
            tupleFactories[i].close();
        }
        tupleFactories = null;
    }

    public TupleFactory getTupleFactory(int tmIndex)
    {
        return tupleFactories[tmIndex];
    }
    
    public AbstractConnection getConnection()
    {
        return connection;
    }

    public void flush() throws SQLException, XMLDBCException
    {
        BufferNode wNode;
        Iterator it = waitingNodes.iterator();
        while (it.hasNext())
        {
            wNode = (BufferNode)it.next();
            
            if (wNode.isReady())
                wNode.save();
            if (wNode.isDiscardable())  // for post-nodes that could have been saved by dependance
            {
                recycle(wNode);
                it.remove();
            }
        }
    }
    
    private void recycle(BufferNode recycled)
    {
        bufferPool.add(recycled);
        recycled.clear();
        // OPTIMIZATION : limit the number of recycled nodes
    }
    
    public BufferNode bufferizeNode(Collection tableMappings, Locator locator)
    throws RepositoryException
    {
        BufferNode newNode;
        if (bufferPool.size() > 0)
        {
            newNode = (BufferNode)bufferPool.remove(bufferPool.size()-1);
            newNode.addTuples(tableMappings, locator);
            newNode.setLocation(locator);
        }
        else
            newNode = new BufferNode(tableMappings, locator);
        
        waitingNodes.add(newNode);
        return newNode;
    }
    
    public BufferNode bufferizeDefaultNode(CollectionMappingInfo mappingInfo, Locator locator)
    throws RepositoryException
    {
        // substitute the current default mapping view
        TableMapping tm = mappingInfo.getTableMapping();
        tupleFactories[tm.getTableIndex()].setMapping(tm);
        return bufferizeNode(Collections.singletonList(tm), locator);
    }
    
    protected void initBatcher(AbstractConnection connection)
    throws RepositoryException, SQLException
    {
        // Instantiate tuple factories & Fill batcher with statements in the right order
        int stmtIndex = -1;
        TupleFactory factory;
        MappingInfo mappingStmt;
        MappingSetIterator it = mappings.getSortedTableMappingSetIterator(false);
		while (it.hasNext())
		{
			mappingStmt = it.next();
			if (it.isTableFirstOccurence() || !mappingStmt.isShareable())
			{
				factory =
					tupleFactories[mappingStmt
						.getTableMapping()
						.getTableIndex()] =
						new TupleFactory(mappingStmt, connection);
				if (factory.getOIDStmt() != null)
					batcher.addStatement(
						factory.getOIDStmt(),
						mappingStmt.getOIDTableName(),
						RepositoryProperties.getIntProperty(
							CONF_TREE_BATCHSIZE));
				if (factory.getInsertStmt() != null)
					stmtIndex =
						batcher.addStatement(
							stmtIndex,
							factory.getInsertStmt(),
							factory.getMapping().getTableName(),
							factory.getMapping().getBatchSize());
				// TO REFINE
				if (factory.getUpdateStmt() != null)
					batcher.addStatement(
						stmtIndex,
						factory.getUpdateStmt(),
						factory.getMapping().getTableName(),
						factory.getMapping().getBatchSize());
				// TO REFINE
			}
			else // share statement & tuple factory
				tupleFactories[mappingStmt.getTableMapping().getTableIndex()] =
					tupleFactories[it.getFirstOccurenceTableIndex()];
		}
    }
    
    /*******************************************************************
    **                      INNER CLASSES                             **
    *******************************************************************/
    
    ////////////////////////////////////////////////////////////////////
    //                          BUFFER NODE
    ////////////////////////////////////////////////////////////////////
    /**
     * Elementary data structure used to store links & values waiting to be stored
     */
    public class BufferNode
    {
        private static final String RCSRevision = "$Revision: 1.7 $";
        private static final String RCSName     = "$Name: XQ1_1_1 $";
        ArrayList tuples = null;   // for nodes corresponding to table mappings
        boolean saved = false;
        int line = -1;
        int column = -1;
        
        BufferNode() {}
        BufferNode(Collection tableMappings, Locator locator)
        throws RepositoryException
        {
            addTuples(tableMappings, locator);
            setLocation(locator);
        }

        public void clear()
        {
            tuples.clear();
            saved = false;
            line = -1;
            column = -1;
        }
        
        public boolean hasTuples()
        {
            return (tuples != null) && (tuples.size() != 0);
        }
        
        public void addDefaultTuple(CollectionMappingInfo mappingInfo, Locator locator)
        throws RepositoryException
        {
            // substitute the current default mapping view
            TableMapping tm = mappingInfo.getTableMapping();
            tupleFactories[tm.getTableIndex()].setMapping(tm);
            addTuples(Collections.singletonList(tm), locator);
        }
        
        public void addTuples(Collection tableMappings, Locator locator)
        {
            if (tableMappings !=  null)
            {
                // Allocate tuples for node
                if (tuples == null)
                    tuples = new ArrayList(tableMappings.size());
                
                Iterator it = tableMappings.iterator();
                TableMapping table = null;
                while (it.hasNext())
                {
                    table = (TableMapping)it.next();
                    tuples.add(getTupleFactory(table.getTableIndex()).newTuple(locator));
                }
            }
        }
            

        public void setLocation(Locator locator)
        {
            if (locator != null)
            {
                line = locator.getLineNumber();
                column = locator.getColumnNumber();
            }
        }
        
        public void save() throws XMLDBCException
        {
            /* save tuples */
            try {
                if (tuples != null)
                {
                    int len = tuples.size();
                    for (int i = 0; i < len; i++)
                    {
                        ((Tuple)tuples.get(i)).save();
                    }   
                }
            }
            catch (RepositoryException e) {
                throw new StorageException(e, new Locator() {
                        public String getPublicId() { return null;}
                        public String getSystemId() { return null;}
                        public int getLineNumber() { return line;}
                        public int getColumnNumber() { return column;}
                    });
            }
            /* mark for removal */
            saved = true;
        }
        
        public void discardEmpty()
        {
            if (tuples != null)
            {
                int len = tuples.size();
                Tuple tuple;
                for (int i = 0; i < len; i++)
                {
                    tuple = (Tuple)tuples.get(i);
                    if (tuple.isEmpty())
                        tuple.setDiscarded();
                }
            }
       }
        
        public void changePathSign()
        {
            if (tuples != null)
            {
                int len = tuples.size();
                Tuple tuple;
                for (int i = 0; i < len; i++)
                {
                    ((Tuple)tuples.get(i)).changePathSign();
                }
            }
        }
        
        public void restore()
        {
            if (tuples != null)
            {
                int len = tuples.size();
                Tuple tuple;
                for (int i = 0; i < len; i++)
                {
                    tuple = (Tuple)tuples.get(i);
                    tupleFactories[tuple.tableIndex].setTuple(tuple);
                }   
            }
        }
        
        void complete() throws XMLDBCException
        {
            if (tuples != null)
            {
                int len = tuples.size();
                for (int i = 0; i < len; i++)
                {
                    ((Tuple)tuples.get(i)).complete();
                }   
            }
        }
        
        /** Check that all tuples are ready (completed) and that all tuples
         * which this node tuples are dependant are stored.
         */
        public boolean isReady() throws XMLDBCException
        {
            boolean ready = true;
            
            if (tuples != null)
            {
                int len = tuples.size();
                for (int i = 0; ready && (i < len); i++)
                {
                    ready = (ready && ((Tuple)tuples.get(i)).isReady());
                }   
            }
            
            return ready;
        }
        
        public boolean isDiscardable() { return saved;}
    }
    
    ////////////////////////////////////////////////////////////////////
    //                      TUPLE FACTORY
    ////////////////////////////////////////////////////////////////////
    public class TupleFactory
    {
        private static final String RCSRevision = "$Revision: 1.7 $";
        private static final String RCSName     = "$Name: XQ1_1_1 $";
        TableMapping mapping = null; // table mapping corresponding to the element
        private Tuple currentTuple = null;  // tuple being built
// TO REMOVE ?        private int parametersToBeFilled;   // number of parameters still missing, 
                                    // equal to optionalParameters if ready to be saved
        private PreparedStatement OIDStmt = null; // JDBC statement for value OIDs
        private PreparedStatement insertStmt = null; // JDBC statement for values
        private PreparedStatement selectStmt = null; // JDBC statement for reading values
        private PreparedStatement updateStmt = null; // JDBC statement for reading values
		
        TupleFactory(MappingInfo  mapStmts, AbstractConnection connection) throws SQLException, RepositoryException
        {
            this.mapping = mapStmts.getTableMapping();
            
            /* Initialize JDBC statements & parameter SQL types */
            if (!mapping.isClustered())
                OIDStmt = connection.getConnection().prepareStatement(mapStmts.getOIDInsertStatement());
            
            insertStmt = connection.getConnection().prepareStatement(mapStmts.getInsertStatement());
            
            int action = mapping.getAction();
            if ((mapping.getSelectColumnCount() != 0) && ((action == MappingConstants.SELECT)
                || (action == MappingConstants.CHECK) || (action == MappingConstants.UPDATE)))
                selectStmt = connection.getConnection().prepareStatement(mapStmts.getSelectStatement());
            if ((mapping.getUpdateColumnCount() != 0) && (action == MappingConstants.UPDATE))
                updateStmt = connection.getConnection().prepareStatement(mapStmts.getUpdateStatement());
            
        }
        
        TableMapping getMapping()
        {
            return mapping;
        }
        PreparedStatement getOIDStmt()
        {
            return OIDStmt;
        }
        PreparedStatement getInsertStmt()
        {
            return insertStmt;
        }
        PreparedStatement getUpdateStmt()
        {
            return updateStmt;
        }
         
        void setMapping(TableMapping mapping)
        {
            this.mapping = mapping;
        }
         
        Tuple newTuple(Locator locator)
        {
            // Allocate new tuple
            currentTuple = new Tuple(mapping.getTableIndex(), mapping.getColumnMappingCount(), mapping.getOIDTableColumnCount(), mapping.getSelectColumnCount(), mapping.getUpdateColumnCount(), locator);
            
            // Initialize values
            Iterator it = mapping.initParameters().iterator();
            ColumnMapping column;
            Object data;
            int refIndex;
            
            while (it.hasNext())
            {
                column = (ColumnMapping)it.next();
                data = column.getInitGenerator().getValue(context);
                
                currentTuple.insertParamList[column.getInsertColumnIndex()] = data;
                if ((OIDStmt != null) && column.isInJoin())
                    currentTuple.OIDParamList[column.getJoinColumnIndex()] = data;
                if ((selectStmt != null) && (column.getSelectColumnIndex() !=-1))
                {
                    currentTuple.selectParamList[column.getSelectColumnIndex()] = data;
                    currentTuple.selectParamCount++;
                }
                if ((updateStmt != null) && (column.getUpdateColumnIndex() !=-1))
                    currentTuple.updateParamList[column.getUpdateColumnIndex()] = data;
                
            }
            
            ColumnMapping[] cms = mapping.getColumnMappings();
            for (int i = 0; i < cms.length; i++)
            {
                /* create depency within tables from ref generators */
                // NOTE : 
            	// 1. table-column mappings are real column mappings and they can have default generators
            	// because they are nullable or have a default value. But are not in init generators if
            	// they do not have default (they will be called when node will be encountered).
            	// 2. They won't reference the subelement as a master tuple (since the referenced tuple does
            	// not exist yet), but it does not really matter since the current tuple will be finished
            	// after the refernced tuple by construction.
                refIndex = cms[i].getTableRefIndex();
                if (refIndex != -1)
                    currentTuple.addMasterTuple(tupleFactories[refIndex].getTuple());
                	// tuple is null in the table-colum maping case and will be ignored
            }

            // Set DocOID & OID
            if  (OIDStmt != null)
            {
                currentTuple.OIDParamList[0] = new Long(context.getUOID());
                currentTuple.OIDParamList[1] = new Integer(context.getPathOID());
            }
            
            return currentTuple;   // for buffer storing
        }
        
        void saveTuple(Tuple tuple, int line, int column) throws SQLException, XMLDBCException
        {
        	if (log.isDebugEnabled())
        		log.debug("Saving tuple:\n" + tuple);
            try
            {
                // according to access mode  : update operations
                switch (mapping.getAction())
                {
                    case MappingConstants.UPDATE :
                        if (tuple.isInbase())
                            performUpdate(tuple, line, column);
                        else
                            performInsert(tuple, line, column);
                        break;
                        
                    case MappingConstants.CHECK :
                        if (!tuple.isInbase())
                            performInsert(tuple, line, column);
                        break;
                        
                    case MappingConstants.INSERT :
                        // Insert directly in the table
                        // Fills values statement with parameters values
                        performInsert(tuple, line, column);
                        break;
                        
                    default :
                }
                
                /* OID insertion */
                // Fills OID statement with parameters values
                if  (OIDStmt != null)
                {
                    int i = 0;
                    // filling UOID & PathID (that do not have column Mappings
                    try
                    {
                        for (; i < 2; i++)
                        {
                            OIDStmt.setObject(i+1, tuple.OIDParamList[i]);
                        }
                        // filling key columns
                        for (; i < tuple.OIDParamList.length; i++)
                        {
                            mapping.getJoinColumns()[i-2].getTypeInfo().setParameter(
                            OIDStmt,
                            i+1,
                            tuple.OIDParamList[i],
                            connection,
                            overrideEmptyString,
                            errorHandler
                            );
                        }
                    }
                    catch (Exception e)
                    {
                        throw new RepositoryException(RepositoryException.DB_ERROR, "Error while setting data in column " + (i+1) + " of :\n" + tuple, e);
                    }
                    // Execute statement
                    batcher.addBatch(OIDStmt, line, column);
                }
            }
            catch (SQLException e)
            {
                throw new RepositoryException(RepositoryException.DB_ERROR, "Error while saving tuple :\n" + tuple, e);
            }
        }
		
        private void performInsert(Tuple tuple, int line, int column) throws SQLException, RepositoryException
        {
			int i = 0;
			try
			{
				ColumnMapping[] columns = mapping.getColumnMappings();
				ColumnMapping c;
				for (i = 0; i < columns.length; i++)
				{
					c = columns[i];
					c.getTypeInfo().setParameter(
						insertStmt,
						c.getInsertColumnIndex() + 1,
						tuple.insertParamList[c.getInsertColumnIndex()],
						connection,
						overrideEmptyString,
						errorHandler);
				}
			}
			catch (Exception e)
			{
				throw new RepositoryException(
					RepositoryException.DB_ERROR,
					"Error while setting data in column "
						+ (i + 1)
						+ " of :\n"
						+ tuple,
					e);
			}
			// Execute statement
			batcher.addBatch(insertStmt, line, column);
		}
        
        private void performUpdate(Tuple tuple, int line, int column) throws SQLException, RepositoryException // TO RETROFIT TO 1.0 ?
        {
            if (updateStmt != null)
            {
                int i = 0;
                try
                {
                    // Fills values statement with parameters values
                    for (i = 0; i < tuple.updateParamList.length; i++)
                    {
                        mapping.getUpdateColumns()[i].getTypeInfo().setParameter(
                            updateStmt,
                            i+1,
                            tuple.updateParamList[i],
                            connection,
                            overrideEmptyString,
                            errorHandler
                            );
                    }
                }
                catch (Exception e)
                {
                    throw new RepositoryException(RepositoryException.DB_ERROR, "Error in column " + (i+1) + " while updating tuple :\n" + tuple, e);
                }
                int j = 0;
                try
                {
                    for (j = 0; j < tuple.selectParamList.length; j++)
                    {
                        mapping.getSelectColumns()[j].getTypeInfo().setParameter(
                            updateStmt,
                            i+j+1,
                            tuple.selectParamList[j],
                            connection,
                            overrideEmptyString,
                            errorHandler
                            );
                    }
                }
                catch (Exception e)
                {
                    throw new RepositoryException(RepositoryException.DB_ERROR, "Error while setting data in column " + (i+j+1) + " of :\n" + tuple, e);
                }
                // Execute statement
                batcher.addBatch(updateStmt, line, column);
            }
            else
                throw new RepositoryException(RepositoryException.MAPPING_CONSISTENCY_ERROR,
                "The update statement is missing for "
                + mapping.getTableName()
                + " table in mapping information.");
        }
        
        Object getParameter(int index)
        {
            if (currentTuple != null)  // meens valid from a mapping point of view
                return currentTuple.insertParamList[index];
            else
                return null;
        }
        
        void addMasterTuple(Tuple tuple) 
        {
            currentTuple.addMasterTuple(tuple);
        }
        
        Tuple getTuple() { return currentTuple;}
        
        void setTuple(Tuple tuple) { currentTuple = tuple;}
        
        /**
         * If action is CHECK, SELECT or UPDATE fills the tuple with the row in the database.
         */
        // TODO : do it when sending it to batcher but what to do with references ?
        void updateFromDatabase(Tuple currentTuple, int selectParamCount)
        throws XMLDBCException
        {
            if ((selectParamCount < mapping.getSelectColumnCount())
                || (mapping.getAction() == MappingConstants.INSERT) // All other actions need to query the db
                || currentTuple.isInbase()) // already retrieved
                return;		// the tuple is not filled enough for selection
            
            // Flush needed for the case of intra-document redundancy
            // TO IMPROVE : check when mutiple use of the same table mapping in a document
            // to perform a selective flush or keep a list of unflushed tuples...
            log.debug("Flush triggered for select");
            batcher.flush();
            
            ResultSet rs = null;
            
            try
            {
                // according to access mode  : selection operation
                if (selectStmt != null)
                {
                    // Fills select statement with parameters values
                    int i = 0;
                    try
                    {
                        for (; i < currentTuple.selectParamList.length; i++)
                        {
                            mapping.getSelectColumns()[i].getTypeInfo().setParameter(
                            selectStmt,
                            i+1,
                            currentTuple.selectParamList[i],
                            connection,
                            overrideEmptyString,
                            errorHandler
                            );
                        }
                    }
                    catch (Exception e)
                    {
                        throw new RepositoryException(RepositoryException.DB_ERROR,
                            "Error while setting data in column " + (i+1) + " of :\n" + currentTuple, e);
                    }
                    // Fetch results and put it in the insert parameter list
                    rs = selectStmt.executeQuery();
                }
                else
                    throw new RepositoryException(RepositoryException.MAPPING_CONSISTENCY_ERROR,
                    "The select statement is missing for "
                    + mapping.getTableName()
                    + " table in mapping information.");
                
                /* update the tuple */
                if (rs.next())  // Take the first
                {
                    currentTuple.foundInBase();	// Mark the tuple is present in the database
                    int joinIndex, updateIndex;
                    ColumnMapping cm = null;
                    Object o = null;
                    // update the insert and the OID statement (+ references) with row read in the database
                    for (int i = 0; i < mapping.getFetchColumnCount(); i++)
                    {
                        cm = mapping.getFetchColumns()[i];
                        joinIndex = cm.getJoinColumnIndex(); // because column index is also the index in table's mapping list of columns mappings
                        o = rs.getObject(i+1);
                        if (mapping.getAction() == MappingConstants.UPDATE) { // Need to update tuple (CHECK or SELECT)
                        	if (joinIndex != -1) // update only the key columns
                        		currentTuple.insertParamList[cm.getInsertColumnIndex()] = o;
                        	updateIndex = cm.getUpdateColumnIndex();
                        	if (!cm.updateColumnWhenMissing() && updateIndex >= 0 && currentTuple.updateParamList[updateIndex] == null)
                        		currentTuple.updateParamList[cm.getUpdateColumnIndex()] = o;
                        }
                        else
                            currentTuple.insertParamList[cm.getInsertColumnIndex()] = o;
                         
                        // Update the key in the OID table
                        if (joinIndex != -1)
                            currentTuple.OIDParamList[joinIndex] = currentTuple.insertParamList[cm.getInsertColumnIndex()];
                    }
                	if (log.isDebugEnabled()) {
	                    List list = Arrays.asList(currentTuple.insertParamList);
	                    log.debug("Tuple fetched from the database: " + list);
                	}
                }
                else if (mapping.getAction() == MappingConstants.SELECT)   // no row found
                    throw new RepositoryException(RepositoryException.DB_CONSISTENCY_ERROR, "Couldn't find the tuple corresponding to the SELECT table mapping\n" + currentTuple);
            }
            catch (SQLException e) {
                throw new RepositoryException(RepositoryException.DB_ERROR, "JDBC error while reading tuple from the database\n" + currentTuple, e);
            }
            finally
            {
                try
                {
                    if (rs != null)
                        rs.close();
                }
                catch (SQLException e)
                {
                    throw new RepositoryException(RepositoryException.DB_WARNING, "JDBC error while releasing ressources for table mapping.", e);
                }
            }
        }
        
        public void clear()
        {
            currentTuple = null;
        }
        
        public void close() throws RepositoryException
        {
            try {
                if (OIDStmt != null)
                {
                    OIDStmt.close();
                    OIDStmt = null;
                }
                if (insertStmt != null)
                {
                    insertStmt.close();
                    insertStmt = null;
                }
                if (selectStmt != null)
                {
                    selectStmt.close();
                    selectStmt = null;
                }
                if (updateStmt != null)
                {
                    updateStmt.close();
                    updateStmt = null;
                }
            }
            catch (SQLException e) {
                throw new RepositoryException(RepositoryException.DB_WARNING, "JDBC error while releasing ressources for table mapping.", e);
            }
        }
    }
    
    ////////////////////////////////////////////////////////////////////
    //                          TUPLE
    ////////////////////////////////////////////////////////////////////
    
    public class Tuple
    {
        private static final String RCSRevision = "$Revision: 1.7 $";
        private static final String RCSName     = "$Name: XQ1_1_1 $";
        Object[] insertParamList = null;  // list for parameters for the insert statement
        Object[] OIDParamList = null;  // list for parameters for the OID table statement
        Object[] selectParamList = null;  // list for parameters for the select statement
        Object[] updateParamList = null;  // list for parameters for the select statement
        ArrayList waitingFor = null;    // List of tuples which this node is dependent on
                                        // in order to be flushed to database
        int tableIndex;                 // to identify the factory for saving
        private int selectParamCount = 0;	// Number of parameters filled that are part of the selection key
        boolean complete = false;
        boolean discarded = false;
        boolean empty = true;
        boolean saved = false;
        private boolean inBase = false;	// The tuple was found in the database
        int line = -1;
        int column = -1;
        
        Tuple(int tableIndex, int insertSize, int OIDsize, int selectSize, int updateSize, Locator locator)
        {
            this.tableIndex = tableIndex;
            insertParamList = new Object[insertSize];
            if (OIDsize != 0)
                OIDParamList = new Object[OIDsize];
            if (selectSize != 0)
                selectParamList = new Object[selectSize];
            if (updateSize != 0)
                updateParamList = new Object[updateSize];
            setLocation(locator);
        }

        public void setLocation(Locator locator)
        {
            if (locator != null)
            {
                line = locator.getLineNumber();
                column = locator.getColumnNumber();
            }
        }
        
        void save() throws XMLDBCException
        {
            try {
                if (!saved && !discarded)
                {
                    tupleFactories[tableIndex].saveTuple(this, line, column);
                    saved = true;
                }
            }
            catch (SQLException e) {
                throw new RepositoryException(RepositoryException.DB_ERROR, "JDBC error while saving tuple " + this, e);
            }
        }
        
        void changePathSign()
        {
            // Reinitialize system genrators because may contain path generator
            Iterator it = tupleFactories[tableIndex].getMapping().initParameters().iterator();
            ColumnMapping column;
            Object data;
            int refIndex;
            
            while (it.hasNext())
            {
                column = (ColumnMapping)it.next();
                Generator gen = column.getDefaultGenerator();
                if (gen instanceof SystemVariableGenerator)
                {
                    data = gen.getValue(context);
                    
                    insertParamList[column.getInsertColumnIndex()] = data;
                    if ((OIDParamList != null) && column.isInJoin())
                        OIDParamList[column.getJoinColumnIndex()] = data;
                    if ((selectParamList != null) && (column.getSelectColumnIndex() !=-1))
                    {
                        selectParamList[column.getSelectColumnIndex()] = data;
                        selectParamCount++;
                    }
                    if ((updateParamList != null) && (column.getUpdateColumnIndex() !=-1))
                        updateParamList[column.getUpdateColumnIndex()] = data;
                }
           }
            
            // change path ID sign
            if (OIDParamList != null)
                OIDParamList[1] = new Integer(context.getPathOID());
        }

        
        /**
         * The tuple is built. Automatic update is performed from database if needed.
         */
        void addColumnMappingData(ColumnMapping column, StorageContext context)
        throws XMLDBCException
        {
            addParameter(column, context);
            tupleFactories[tableIndex].updateFromDatabase(this, selectParamCount);
        }
        
        private void addParameter(ColumnMapping column, StorageContext context)
        {
            empty = false;
            Object data = column.getGenerator().getValue(context);
            insertParamList[column.getInsertColumnIndex()] = data;
            if ((OIDParamList != null) && column.isInJoin())
                OIDParamList[column.getJoinColumnIndex()] = data;
            if ((selectParamList != null) && (column.getSelectColumnIndex() >= 0))
            {
                selectParamList[column.getSelectColumnIndex()] = data;
                selectParamCount++;
            }
            if ((updateParamList != null) && (column.getUpdateColumnIndex() >= 0))
                updateParamList[column.getUpdateColumnIndex()] = data;
        }
        
        /**
         * The tuple is filled with generators and passed to the buffer.
         * No subsequent update should be performed.
         */
        void complete() throws XMLDBCException
        {
            // process ending parameters (generators
            Iterator it = tupleFactories[tableIndex].mapping.finalParameters().iterator();
            while (it.hasNext())
            {
                addColumnMappingData((ColumnMapping)it.next(), context);
            }

            setCompleted();
        }
        
        void addMasterTuple(Tuple tuple) 
        {
        	if (tuple != null) {
	            if (waitingFor == null)
	                waitingFor = new ArrayList();
	            waitingFor.add(tuple);
        	}
        }
        /** check that the tuple is complete and that all tuples which it
         * depends on are ready (eventually tries to trigger their storage).
         **/
        public boolean isReady() throws XMLDBCException
        {
            if (discarded || saved)
                return true;
            if (complete)
            {
                if  ((waitingFor == null) || (waitingFor.size() == 0))
                    return true;
                else    // try to flush dependent tuples
                {
                    boolean ready = true;
                    Iterator it = waitingFor.iterator();
                    Tuple waitedTuple;
                    while (it.hasNext())
                    {
                        waitedTuple = (Tuple)it.next();
                        // if ready, save it
                        if (waitedTuple.isReady())
                        {
                            waitedTuple.save();
                            it.remove();
                        }
                        else 
                           ready = false;   // one is enough 
                    }
                    return ready;
                }
            }
            else
                return false;
        }

        public void setCompleted()
        {
            complete = true;
        }
        
        public void setDiscarded()
        {
            discarded = true;
        }
		
        public void foundInBase()
        {
            inBase = true;
        }
		
        public boolean isInbase()
        {
            return inBase;
        }
        public boolean isEmpty()
        {
            return empty;
        }
        public String toString()
        {
            StringWriter sw = new StringWriter();
            Tabulator tab = new Tabulator(sw, 160, new int[] {0, 15});

            /* table display */
            tab.addItem("Table:");
            tab.addItem(tupleFactories[tableIndex].mapping.getTableName());
            
            /* Location */
            tab.addItem("XML location:");
            tab.addItem("line " + line + ", column " + column);
            
            TableMapping tm = tupleFactories[tableIndex].mapping;
            
            /* Column display */
            tab.addItem("Columns:");
            List list = Arrays.asList(tm.getColumnMappings());
            tab.addItem(list);
			
            /* Values display */
            tab.addItem("Values:");
            list = Arrays.asList(insertParamList);
            tab.addItem(list);
            if (log.isDebugEnabled())
            {
                /* Java types display */
                tab.addItem("Java types:");
                tab.addItem(diplayListClasses(list));
                if (!tm.isClustered())
                {
                    /* OID values display */
                    tab.addItem("OID:");
                    list = Arrays.asList(OIDParamList);
                    tab.addItem(list);
                    /* Java types display */
                    tab.addItem("Java types:");
                    tab.addItem(diplayListClasses(list));
                }
            }
            return sw.toString(); 
        }

        private StringBuffer diplayListClasses(List list)
        {
            StringBuffer display = new StringBuffer();
            Object o;
            display.append("");
            for (int i = 0; i < list.size() ; i++)
            {
                if (i == 0)
                    display.append("[");
                else
                    display.append(", ");
                o = list.get(i);
                if (o == null)
                    display.append("null");
                else
                    display.append(o.getClass().getName());
            }
            display.append("]");
            return display;
        }
    }
}
