/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

/*
 * TableExplorer.java
 *
 */

package org.xquark.mapper.storage;

import java.sql.SQLException;
import java.util.Collection;
import java.util.Iterator;

import org.xquark.mapper.RepositoryException;
import org.xquark.mapper.mapping.ColumnMapping;
import org.xquark.mapper.mapping.RepositoryMapping;
import org.xquark.mapper.mapping.TableMapping;
import org.xquark.mapper.metadata.PathSet;
import org.xquark.schema.Declaration;
import org.xquark.schema.SchemaException;
import org.xquark.schema.validation.ValidationContextProvider;
import org.xquark.xml.xdbc.XMLDBCException;

/**
 *
 * @see org.xquark.mapper.XMLCollection
 */
public abstract class TableExplorer
{
    private static final String RCSRevision = "$Revision: 1.2 $";
    private static final String RCSName     = "$Name: XQ1_1_1 $";

    public final static int NOT_FOUND = -3;
    public final static int STRUCT_TUPLE = -2;
    public final static int EXTRA_TUPLE = -1;
    public final static int MIN_DATA = 0;
    
    protected PathSet pathSet;
    protected Tuple[] tuples;
    protected StructExplorer wNode;
    protected ExtraDataExplorer extraNode;
 //    private RepositoryMapping collectionMapping;
    protected _RepositoryCollection collection;
    
    TableExplorer(_RepositoryCollection collection)
    throws SQLException, XMLDBCException
    {
        this.collection = collection;
        pathSet = collection.getMetadata().getPathSet();
        RepositoryMapping collectionMapping = collection.getMetadata().getMapping();
    }
    
    /* only for document reading */
    /** Return the tuple at the given index
     */
    Tuple get(int index) throws SQLException
    {
        return tuples[index];
    }
    
    StructExplorer getStructNode() throws SQLException
    {
        return wNode;
    }
    
    ExtraDataExplorer getExtraNode() throws SQLException
    {
        return extraNode;
    }
    
    Tuple pollData(Collection tableMappings, long first, long last) throws SQLException
    {
        int minIndex = NOT_FOUND, i;
        long minOID = Long.MAX_VALUE;   // so that it is overwritten by the first...
        if (tableMappings == null)
            return null;
        Iterator it = tableMappings.iterator();
        
        while (it.hasNext())
        {
            i = ((TableMapping)it.next()).getTableIndex();
            if ((tuples[i] != null)
            && tuples[i].queryPending()
            && (tuples[i].getUOID() <= minOID)   // inferior or equal
            && (tuples[i].isInRange(first, last)))
            {
                minIndex= i;
                minOID = tuples[i].getUOID();
            }
        }
        if (minIndex == NOT_FOUND)
            return null;
        else
            return get(minIndex);
    }
    
    /* look for a struct node or a text node in the range */
    int poll(long first, long last)
    throws SQLException
    {
        int minIndex = NOT_FOUND;
        long minOID = Long.MAX_VALUE;   // so that it is overwritten by the first...
        
        if ((getStructNode().getUOID() < minOID)    // strictly inferior
        && getStructNode().isInRange(first, last))
        {
            minIndex= STRUCT_TUPLE;
            minOID = getStructNode().getUOID();
        }
        int defaultMappingIndex = collection.getMetadata().getDefaultMappingIndex();
        if ((tuples[defaultMappingIndex] != null)
            && tuples[defaultMappingIndex].queryPending()
            && (tuples[defaultMappingIndex].getUOID() < minOID)   // inferior
            && tuples[defaultMappingIndex].isInRange(first, last))
            minIndex= defaultMappingIndex;
        return minIndex;
    }

    /* Fetch the next row for all tuples */
    boolean lookup() throws SQLException
    {
        boolean rows = wNode.fetchNextRow();
        rows |= extraNode.fetchNextRow();
        for (int i = 0; i < tuples.length; i++)
        {
            if (tuples[i] != null)
            {
                if (tuples[i].queryPending())
                    tuples[i].fetchNextRow();
                rows |= tuples[i].queryPending();
            }
        }
        return rows;
    }
    
    void close() throws SQLException, RepositoryException
    {
        wNode.close();
        extraNode.close();
        
            /* Releasing JDBC resources used for value tables */
        for (int i = 0; i < tuples.length; i++)
        {
            if (tuples[i] != null)
                tuples[i].close();
        }
    }
    
    /** 
     */
    void reset() throws SQLException, RepositoryException
    {
        wNode.reset();
        extraNode.reset();
        
        // Releasing JDBC resources used for value tables
        for (int i = 0; i < tuples.length; i++)
        {
            if (tuples[i] != null)
                tuples[i].reset();
        }
    }
    
    String getValue(ColumnMapping column, Declaration decl, ValidationContextProvider nsContext)
    throws SQLException, SchemaException, XMLDBCException
    {
        return get(column.getTableIndex()).getValue(column, decl, nsContext);
    }
}
