/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.mapper.storage;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Arrays;

import org.xquark.mapper.mapping.ColumnMapping;
import org.xquark.mapper.mapping.TableMapping;
import org.xquark.mapper.metadata.CollectionMappingInfo;
import org.xquark.schema.Declaration;
import org.xquark.schema.validation.ValidationContextProvider;
import org.xquark.xml.xdbc.XMLDBCException;

public abstract class Tuple
{
    
    private static final String RCSRevision = "$Revision: 1.2 $";
    
    private static final String RCSName = "$Name: XQ1_1_1 $";
    
    protected Statement selectStmt = null;
    
    protected ResultSet resultSet = null;
    
    protected boolean notExhausted = false;
    
    protected CollectionMappingInfo table;
    
    protected long uoid = -1;
    
    protected String[] values = null; // buffered because getValue() is called several times
    protected boolean[] fetched = null;
    
    
    protected short pathoid = -2;
    
    Tuple(CollectionMappingInfo table)
    {
        this.table = table;
        values = new String[getMapping().getColumnMappingCount()];
        fetched = new boolean[getMapping().getColumnMappingCount()];
    }
    
    boolean queryPending()
    {
        return notExhausted;
    }
    
    /** To be overridden */
    boolean resultActive()
    {
        return notExhausted;
    }
    
    boolean isClosed()
    {
        return (resultSet == null);
    }
    
    public boolean isInRange(long first, long last, short path)
    {
        return resultActive() && (uoid >= first) && (uoid <= last) && (pathoid == path);
    }
    
    public boolean isInRangeAbs(long first, long last, short path)
    {
        return resultActive() && (uoid >= first) && (uoid <= last) && ((short)Math.abs(pathoid) == path);
    }
    
    public boolean isInRange(long first, long last)
    {
        return resultActive() && (uoid >= first) && (uoid <= last);
    }
    
    long getUOID()
    {
        return uoid;
    }
    
    TableMapping getMapping()
    {
        return table.getTableMapping();
    }
    
    short getPathID()
    {
        return (short)Math.abs(pathoid);
    }
    
    boolean fetchNextRow() throws SQLException
    {
        Arrays.fill(fetched, false);
        return true;
    }
    
    String getValue(ColumnMapping column, Declaration decl, ValidationContextProvider nsContext)
    throws SQLException, XMLDBCException
    {
        int index = column.getInsertColumnIndex();
        
        if (!fetched[index])
        {
            values[index] = fetchValue(column, decl, nsContext);
            fetched[index] = true;
        }
        
        return values[index];
    }
    
    protected abstract String fetchValue(ColumnMapping column, Declaration decl, ValidationContextProvider nsContext)
    throws SQLException, XMLDBCException;
    
    void close() throws SQLException
    {
        reset();
        if (selectStmt != null) // table mapping may not be used for queries
        {
            selectStmt.close();
            selectStmt = null;
        }
    }
    
    void reset() throws SQLException
    {
        if (!isClosed())    // rs isClosed !!!
        {
            resultSet.close();
            resultSet = null;
            uoid = -1;
            pathoid = -2;
            notExhausted = false;
        }
    }
    
}

