/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.mapper.storage;

import java.io.IOException;
import java.io.StringReader;

import javax.xml.parsers.SAXParserFactory;

import org.w3c.dom.Document;
import org.xml.sax.*;
import org.xml.sax.ext.LexicalHandler;
import org.xquark.mapper.RepositoryException;
import org.xquark.util.DOMReader;
import org.xquark.util.HandlerDecorator;
import org.xquark.util.SAXConstants;
import org.xquark.xml.xdbc.XMLDBCException;
import org.xquark.xml.xdbc.XMLErrorHandler;

/**
 * Class used to convert Mapper calls to basic internal calls.
 */
public class XDBCMapperAdapter extends HandlerDecorator
implements SAXConstants
{
	private static final String RCSRevision = "$Revision: 1.2 $";
	private static final String RCSName     = "$Name: XQ1_1_1 $";
    private static SAXParserFactory parserFactory;
    
    static
    {
        parserFactory = SAXParserFactory.newInstance();
        parserFactory.setNamespaceAware(true);
    }
    
    private SAXHandler handler;
    private DOMReader DOMReader;
    
    public XDBCMapperAdapter(ContentHandler contentHandler, LexicalHandler lexicalHandler,
                                ErrorHandler errorHandler, SAXHandler handler)
    {
        super(contentHandler, lexicalHandler);
        super.setErrorHandler(errorHandler);
        this.handler = handler;
    }
    
    public void flushBuffer() throws XMLDBCException
    {
        handler.flushBuffer();
    }
    
    public void setErrorHandler(XMLErrorHandler handler)
    {
        this.handler.setErrorHandler(handler);
    }
    
    public XMLErrorHandler getErrorHandler()
    {
        return handler.getXMLErrorHandler();
    }
    
    public void clearBuffer() throws XMLDBCException
    {
        handler.clearBuffer();
    }
    
    public void close() throws XMLDBCException
    {
        handler.close();
    }
    
    public boolean getAutoFlush()
    {
        return handler.getAutoFlush();
    }
    
    public void setAutoFlush(boolean mode) throws XMLDBCException
    {
        handler.setAutoFlush(mode);
    }
    
    ///////////////////////////////////////////////////////////////////////////
    //                      Mapper implementation
    ///////////////////////////////////////////////////////////////////////////
    public void insertXMLDocument(org.xml.sax.InputSource input)
    throws XMLDBCException, SAXException
    {
        insertXMLDocument(createXMLReader(), input);
    }
    
    public void insertXMLDocument(org.xml.sax.XMLReader parser, org.xml.sax.InputSource input)
    throws XMLDBCException, SAXException
    {
        configureXMLReader(parser);
        try
        {
            parser.parse(input);
        }
        catch (SAXException e)
        {
            if (e.getException() instanceof XMLDBCException)
                throw (XMLDBCException)e.getException();
            else
                throw e;
        }
        catch (IOException e)
        {
            throw new RepositoryException(RepositoryException.PARSER_ERROR,
                "IO error during SAX parsing.", e);
        }
    }
    
    public void insertXMLDocument(String doc) throws XMLDBCException, org.xml.sax.SAXException
    {
        // remove first blanks (facility)
        int index = doc.indexOf("<?xml");
        if (index > 0)
            doc = doc.substring(index);
        insertXMLDocument(new InputSource(new StringReader(doc)));
    }
    
    public void insertXMLDocument(Document document) throws XMLDBCException
    {
        if (DOMReader == null)
            DOMReader = new DOMReader(document);
        else
            DOMReader.setDocument(document);
        XMLReader myReader = configureXMLReader(DOMReader);
        try
        {
            myReader.parse("Fake");
        }
        catch (IOException e)
        {
            throw new RepositoryException(RepositoryException.PARSER_ERROR,
                "Error during internal SAX parsing.", e);
        }
        catch (SAXException e)
        {
            if (e.getException() instanceof XMLDBCException)
                throw (XMLDBCException)e.getException();
            else
                throw new RepositoryException(RepositoryException.PARSER_ERROR,
                        e.getMessage(), e.getException());
        }
    }
    
    ///////////////////////////////////////////////////////////////////////////
    //                            PRIVATE UTILITIES
    ///////////////////////////////////////////////////////////////////////////
    protected XMLReader createXMLReader() throws RepositoryException
    {
        try
        {
            return parserFactory.newSAXParser().getXMLReader();
        }
        catch (Exception e)
        {
            throw new RepositoryException(RepositoryException.XQUARK_CONF, "Could not instantiate an XMLReader using the property in Repository configuration.", e);
        }
    }
    
    protected XMLReader configureXMLReader() throws RepositoryException
    {
        return configureXMLReader(createXMLReader());
    }
    
    protected XMLReader configureXMLReader(XMLReader reader)
    throws RepositoryException
    {
        try
        {
            reader.setContentHandler(this);
            reader.setErrorHandler(this);
            reader.setProperty(SAX_LEXICAL_PROPERTY, this);
        }
        catch(SAXException e)
        {
            throw new RepositoryException(RepositoryException.PARSER_ERROR, "The XML SAX Parser is not SAX 2.0 compliant");
        }
        return reader;
    }
    
}
