/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

/*
 * XMLBuilder.java
 *
 * Created on 12 mars 2001, 18:01
 */

package org.xquark.mapper.storage;

import java.sql.SQLException;
import java.util.*;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.xml.sax.ContentHandler;
import org.xml.sax.ErrorHandler;
import org.xml.sax.SAXException;
import org.xml.sax.ext.LexicalHandler;
import org.xml.sax.helpers.AttributesImpl;
import org.xquark.mapper.RepositoryException;
import org.xquark.mapper.mapping.ColumnMapping;
import org.xquark.mapper.mapping.TableMapping;
import org.xquark.mapper.metadata.*;
import org.xquark.schema.*;
import org.xquark.schema.validation.ContentIterator;
import org.xquark.schema.validation.ValidationContextProvider;
import org.xquark.util.NamespaceContextStack;
import org.xquark.xml.xdbc.XMLDBCException;
import org.xquark.xpath.NodeKind;

/** 
 * 
 *
 * <p>Generation is guided by the schema and indecisions are removed by
 * document instance information (stored in the structure table).</p>
 * @see org.xquark.mapper.XMLCollection
 */
public abstract class XMLBuilder
implements RepositoryConstants
{
    private static final String RCSRevision = "$Revision: 1.3 $"; 
    private static final String RCSName     = "$Name: XQ1_1_1 $";

    private static Log log = LogFactory.getLog(XMLBuilder.class);
    		
    private final static String DEFAULT_PREFIX = "ns";
    private final static short UNKNOWN = 0;
    private final static short FOUND_ELEMENT = 1;
    private final static short NIL_ELEMENT = 2;
    private final static short MISSING_ELEMENT = 3;
    
     // @return -2 if not found, -1 if found without UOID. UOID else.
    protected final static long NOT_FOUND = -3;
    protected final static long TEXT_FOUND = -2;
    protected final static long NO_UOID_FOUND = -1;
    
    private final static int INDENT_ARRAY_TAB_COUNT = 10;
    private final static char[] INDENT_CHAR_ARRAY = {0xA,'\t','\t','\t','\t',
                                                '\t','\t','\t','\t','\t','\t'};
    
    protected PathSet pathSet;
    
    //
    // Generation variables
    //
    private AttributesImpl attributes = new AttributesImpl();
    
    protected ContentHandler contentHandler = null;
    protected LexicalHandler lexicalHandler = null;
    protected ErrorHandler errorHandler = null;
    protected boolean generateDBPrefix = false; // if true the algorithm generate unconditionnaly prefixes found uin the db
    protected PrefixProvider prefixProvider;
    
    // database objects
    protected TableExplorer tuplePool;
    
    protected _RepositoryCollection collection;
    private int defaultMappingIndex;
    private ColumnMapping defaultColumnMapping; // for text nodes
    private SchemaManager schemaManager; // for xsi types access
    
    /** To be used by document reconstruction */
    protected XMLBuilder(_RepositoryCollection collection)
    throws XMLDBCException
    {
        this.collection = collection;
        schemaManager = collection.getRepositoryConnection().getSchemaManager();
        CollectionMetadata metadata = collection.getMetadata();
        pathSet = metadata.getPathSet();
        defaultMappingIndex = metadata.getDefaultMappingIndex();
        defaultColumnMapping = metadata.getDefaultMapping().getDefaultColumnMapping(null);
    }
    
    /**
     * @uoid negative if the UOID could not be determined by the caller
     * @level element level stating to 1 for the root element
     * @return the last UOID generated by the method
     */
    protected long browseBranch(long first, long last, long uoid, long anchor,
                             MetadataIterator metaContext, int level, boolean mixedContext)
    throws XMLDBCException, SQLException, SAXException
    {
        PathNode meta = metaContext.getPathNode();
        short pid = meta.getPathID();
        ElementInfoSet eltInfo = null;
        boolean nil;
        boolean textOnly = true;
        long branchLast = (uoid < 0 ? first : uoid);
        
        if (log.isDebugEnabled())
        	log.debug("Entering browseBranch [path, first, last]=" + meta + ',' + first + ',' + last);
        
        //
        //      PROCESS the CURRENT PATH
        //
        
        // initialize variables for multivaluated iteration
        nil = false;
        eltInfo = new ElementInfoSet();
    	prefixProvider.push();
        
        // process struct (consume at least)
        StructExplorer structRow = tuplePool.getStructNode();
        switch (meta.getStorageMode())
        {
            case PathMetadata.POTENTIAL_STRUCT:
                if (structRow.isInRange(first, last, pid) // note: testing range is necessary for query result segmentation since potential
                && ((uoid < 0) || (structRow.getUOID() == uoid)))
                {
                    anchor = uoid = first = structRow.getUOID();
                    branchLast = last = structRow.getLast();
                    structRow.fetchNextRow(); // consume struct node raw
                }
                else if (uoid >= first) // model leaf
                    anchor = first = last = branchLast = uoid;
                else if (meta.isMixed())
                {
                    Tuple dataTuple = tuplePool.get(defaultMappingIndex);
                    if (dataTuple.getPathID() != pid)
                        return first; // path missing
                    first = last = branchLast = anchor = uoid = dataTuple.getUOID();
                }
                eltInfo.setStatus(FOUND_ELEMENT);
                break;
            case PathMetadata.STRUCT_ONLY:
            case PathMetadata.DATA_AND_STRUCT:
                if (!structRow.isInRange(first, last, pid)) // note: testing range is necessary for query result segmentation since potential
                    return first; // path missing
                eltInfo.setStatus(FOUND_ELEMENT);
                anchor = uoid = first = structRow.getUOID();
                last = branchLast = structRow.getLast();
                structRow.fetchNextRow(); // consume struct node raw
                break;
            case PathMetadata.DATA_ONLY:
                if (uoid < 0)
                {
                    Tuple dataTuple = tuplePool.get(meta.getReadTableMapping().getTableMapping().getTableIndex());
                    if (dataTuple.getPathID() != pid)
                        return first; // path missing
                    uoid = dataTuple.getUOID();
                }
                eltInfo.setStatus(FOUND_ELEMENT);
                first = last = branchLast = anchor = uoid;
                break;
            case PathMetadata.DISCARDED:
                break;
            default:
                throw new RepositoryException(RepositoryException.CONSISTENCY_ERROR,
                "Metadata is corrupted for the collection "
                + collection.getCollectionName() + "(Unknow storage flag "
                + meta.getStorageMode() + " for pid " + meta.getPathID() + ")");
        }
        
        processElement(metaContext, first, last, uoid, anchor, eltInfo);
        
        switch (eltInfo.getStatus())
        {
            case NIL_ELEMENT:
                eltInfo.generateStartElement(meta);
                break;
                
            case FOUND_ELEMENT:
            case UNKNOWN:   // because query generation may start with a discarded node (need to check children)
                //
                //      PROCESS the CHILDREN of CURRENT PATH
                //
                long childUOID = NO_UOID_FOUND;
                long lastChild = uoid;
                PathNode child = null;
                
                // use iterator if possible
                if (meta.isChildrenPositionEnforced())  // use schema iterator
                {
                    ContentIterator it;
                    // If type is superseded by an XSI declaration, use it for schema guided-iteration
                    if (eltInfo.getXSIType() == null)
                        it = metaContext.getElementDeclaration().getType().childIterator();
                    else
                        it = eltInfo.getXSIType().childIterator();
                    
                    if (it != null)
                    {
                        ElementDeclaration eltDecl = null;
                        long textUOID;
                        List nextValidElements = it.nextValidElements();    // because a new list is built at each call
                        while ((nextValidElements != null)  && (nextValidElements.size() != 0))
                        {
                            childUOID = Long.MAX_VALUE;
                            eltDecl = null;
                            textUOID = NOT_FOUND;
                            // scan text nodes
                            if (meta.isMixed())
                                textUOID = scanTextNode(first, last, pid);
                            
                            // child selection
                            switch (nextValidElements.size())
                            {
                                case 1:
                                    eltDecl = (ElementDeclaration)nextValidElements.get(0);
                                    if (eltDecl != null)
                                    {
                                        child = metaContext.getChild(eltDecl.getNamespace(), eltDecl.getName(), NodeKind.ELEMENT);
                                        childUOID = checkOccurrence(lastChild + 1, last, anchor, child, eltInfo.getChildCount() + 1);
                                    }
                                    break;
                                default:    // contains null, or optional, or multivaluated
                                    // browse possible elements and use db to decide
                                    Iterator walker = nextValidElements.iterator();
                                    long foundUOID;
                                    ElementDeclaration potentialDecl = null;
                                    while (walker.hasNext())
                                    {
                                        potentialDecl = (ElementDeclaration)walker.next();
                                        if (potentialDecl == null)    // ending element is possible
                                            continue;
                                        child = metaContext.getChild(potentialDecl.getNamespace(), potentialDecl.getName(), NodeKind.ELEMENT);
                                        if (child != null)
                                        {
                                            foundUOID = checkOccurrence(lastChild + 1, last, anchor, child, eltInfo.getChildCount() + 1);
                                            if((foundUOID != NOT_FOUND) && ((foundUOID < childUOID) || (childUOID == NO_UOID_FOUND)))
                                            {
                                                if (foundUOID == NO_UOID_FOUND) // UOID not found
                                                {
                                                    if (eltDecl == null)
                                                    {
                                                        childUOID = foundUOID;
                                                        eltDecl = potentialDecl;
                                                    }
                                                    break; // if a node is discarded, the iterator gives the order
                                                }
                                                else // a smaller UOID is found
                                                {
                                                    childUOID = foundUOID;
                                                    eltDecl = potentialDecl;
                                                }
                                            }
                                        }
                                    }
                                    break;
                            }
                            if ((textUOID != NOT_FOUND)
                            && ((textUOID < childUOID) || (childUOID == NOT_FOUND)))
                            {
                                eltInfo.generateStartElement(meta);
                                eltInfo.generateTextNode();
                                continue;  // start again iteration because text node could hide the lowest subelement
                            }
                            
                            if (eltDecl == null) // no declaration found
                                break;
                            
                            eltInfo.generateStartElement(meta);
                            
                            //start It, push iterator,
                            eltInfo.flushExtraData(!mixedContext, true);
                            if (!mixedContext)
                                indent(level);
                            textOnly = false;
                            metaContext.push(eltDecl.getNamespace(), eltDecl.getName());
                            try {
                                it.startElement(eltDecl.getNamespace(), eltDecl.getName());
                                lastChild = browseBranch(first, last, childUOID, anchor,
                                metaContext, level + 1, mixedContext || metaContext.getPathNode().isMixed());
                                it.endElement(eltDecl.getNamespace(), eltDecl.getName());
                                nextValidElements = it.nextValidElements();
                            }
                            catch (SchemaException e) {
                                throw new RuntimeException("Error while iterating through schema: " + e.getMessage());
                            }
                            metaContext.pop();
                            eltInfo.incChildCount();   // must be done before generateTextNode()
                            processExtraData(anchor, pid, eltInfo);
                        }
                    }
                }
                else   // use db info (OID and extra)
                {
                    // child selection
                    while (((childUOID = scanChild(lastChild + 1, last, anchor, metaContext, eltInfo.getChildCount() + 1)) > NOT_FOUND)
                            || !eltInfo.extraFlushed())
                    {
                        eltInfo.generateStartElement(meta);
                        if (childUOID == NOT_FOUND)
                            eltInfo.flushExtraData(!mixedContext, true);
                        else if  (childUOID == TEXT_FOUND)
                            eltInfo.generateTextNode();
                        else
                        {
                            eltInfo.flushExtraData(!mixedContext, true);
                            if ((pid != ROOT_PATH_ID) && !mixedContext)
                                indent(level);
                            textOnly = false;
                            lastChild = browseBranch(first, last, childUOID, anchor,
                            metaContext, level + 1, mixedContext || metaContext.getPathNode().isMixed());
                            metaContext.pop();
                            eltInfo.incChildCount();   // must be done before generateTextNode()
                        }
                        processExtraData(anchor, pid, eltInfo);
                    }
                    // Flush remaining extra-data that may could not be positioned because of a wrong offset (normalization)
                    if (eltInfo.hasExtraData())
                    {
                        eltInfo.generateStartElement(meta);
                    	eltInfo.flushExtraData(!mixedContext, false); // purge
                    }
                }
                break;
                
            default:    // missing
                return first;
        }
        
        // for the case no child node has been found but start element must be generated
        if (eltInfo.leafElementButNotMissing())
            eltInfo.generateStartElement(meta);
        
        
        // generate endElement
        if (eltInfo.elementGenerated())
        {
            eltInfo.flushExtraData(!mixedContext, true);
            if (!textOnly && !mixedContext)
                indent(level - 1);
            contentHandler.endElement(
            meta.getNamespace() == null ? "" : meta.getNamespace(),
            meta.getLocalName(),
            "");
            eltInfo.generateEndPrefix();
            
            // process fetchNextRow for elements
            if (meta.getTableMappings() != null)
            {
                Iterator it = meta.getTableMappings().iterator();
                TableMapping refTableMapping = null, tm = null;
                if (meta.getReadColumnMapping() != null)
                    refTableMapping = meta.getReadColumnMapping().getTableMapping();
                
                while (it.hasNext())
                {
                    tm = (TableMapping)it.next();
                    // Do not consume ref column mapping if already consumed
                    if (!tm.equals(refTableMapping))
                        tuplePool.get(tm.getTableIndex()).fetchNextRow();
                }
            }
        }
    	prefixProvider.pop();
        
        if (log.isDebugEnabled())
        	log.debug("Exiting browseBranch [path]=" + meta);
        return branchLast;
    }
   
    /** Retrieve child info from the database without consuming tuples.
     * Used by schemaless iteration.
     * @return -2 if not found, -1 if found without UOID. UOID else.
     */
    private long scanChild(long first, long last, long anchor,
                MetadataIterator metaContext, int pos)
                throws SQLException, RepositoryException
    {
        short structPID = -1;
        short dataPID = -1;
        long structUOID = Long.MAX_VALUE;
        long dataUOID = Long.MAX_VALUE;
        short wPID, pid = metaContext.getPathMetadata().getPathID();
        
        // try first to find extra-data attached to the current node
        // positioning info is retrieved first because of query ordering for
        // the extra table.
        if ((tuplePool.getExtraNode().getAnchor() == anchor)
            && (tuplePool.getExtraNode().getType() >= NodeKind.ELEMENT)
            && (tuplePool.getExtraNode().getPosition() == pos)) // TO CHECK: is it necessary to check the path found is a child ?
        {
            metaContext.push(tuplePool.getExtraNode().getPath());
            return NO_UOID_FOUND;
        }
        else    // search for the UOID in the struct and then in the data tables
        {
            // look in the struct table
            StructExplorer structNode = null;
            
            if (tuplePool.getStructNode().isInRange(first, last))
                structNode = tuplePool.getStructNode();
            
            // look in the data tables but take if inferior to struct
            Tuple nextDataTuple = tuplePool.pollData(
                                metaContext.getPathMetadata().getBuildMappings(),
                                first,
                                last
                                );
            
            if (structNode != null)
            {
                wPID = structNode.getPath();
                if (wPID >= 0)
                {
                    PathNode structChild = pathSet.get(wPID);
                    if (((PathNode)structChild.getParent()).getPathID() == pid)
                    {
                        structUOID = structNode.getUOID();
                        structPID = wPID;
                    }
                }
            }
            
            if (nextDataTuple != null)
            {
                // text node found
                if (metaContext.getPathMetadata().isMixed()
                    && (nextDataTuple.getMapping().getTableIndex() == defaultMappingIndex)
                    && (nextDataTuple.getPathID() == pid)) // for TEXT nodes minus path may be is used
                    return TEXT_FOUND;

                wPID = nextDataTuple.getPathID();
                if (wPID >= 0)
                {
                    PathNode dataChild = pathSet.get(wPID); // may return a negative value if rs exhausted
                    if (((PathNode)dataChild.getParent()).getPathID() == pid)
                    {
                        dataUOID = nextDataTuple.getUOID();
                        dataPID = wPID;
                    }
                 }
            }
        }
        // take the smallest of potential childs
        long childUOID = structUOID;
        short childPID = structPID;
        
        if (dataUOID < childUOID)
        {
            childUOID = dataUOID;
            childPID = dataPID;
        }
        
        // update context
        if (childUOID != Long.MAX_VALUE)
        {
             metaContext.push(childPID);
             return childUOID;
        }
        else
            return NOT_FOUND;
    }
    
    /** Check in the database if a node exists without consuming tuples.
     * Used by schema iteration.
     * @return -2 if not found, -1 if found without UOID. UOID else.
     */
    private long checkOccurrence(long first, long last, long anchor,
                    PathNode node, int pos)
                    throws SQLException, XMLDBCException
    {
        long nodeUOID = NOT_FOUND;
                
        // try first to find extra-data attached to the current node
        // positioning info is retrieved first because of query ordering for
        // the extra table.
        if ((tuplePool.getExtraNode().getAnchor() == anchor)
        && (tuplePool.getExtraNode().getPath() == node.getPathID())
        && (tuplePool.getExtraNode().getType() >=  NodeKind.ELEMENT))
        {
            if (tuplePool.getExtraNode().getPosition() == pos)
                nodeUOID = NO_UOID_FOUND;
            else
                nodeUOID = NOT_FOUND;
        }
        else
            nodeUOID = getElementUOID(first, last, node);
        return nodeUOID;
    }
    
    /** Retrieves the potential text node UOID.
     * Used by schema iteration.
     * @return -2 if not found, . UOID else.
     */
    private long scanTextNode(long first, long last, short parentPID)
                    throws SQLException
    {
        long textNodeUOID = NOT_FOUND;
        
        // text node found
        Tuple defaultTuple = tuplePool.get(defaultMappingIndex);
        if (defaultTuple.isInRangeAbs(first, last , parentPID)) // for TEXT nodes minus path may be used
            textNodeUOID = defaultTuple.getUOID();
        
        return textNodeUOID;
    }
    /** Retrieves the element UOID.
     * Used by schema iteration.
     * @return -2 if not found, -1 if found without UOID. UOID else.
     */
    protected long getElementUOID(long first, long last, PathNode node)
                    throws SQLException, XMLDBCException
    {
        long nodeUOID = NOT_FOUND;
        Tuple dataTuple;
        
        // try first to find extra-data attached to the current node
        // positioning info is retrieved first because of query ordering for
        // the extra table.
        switch (node.getStorageMode())
        {
            case PathMetadata.POTENTIAL_STRUCT:
                // check in struct
                if (tuplePool.getStructNode().isInRange(first, last, node.getPathID()))
                    nodeUOID = tuplePool.getStructNode().getUOID();
                
                // check in data and take if smaller
                dataTuple = tuplePool.get(defaultMappingIndex);
                if ((dataTuple != null) // may be null in query reconstruction
                && dataTuple.isInRangeAbs(first, last, node.getPathID())
                && ((nodeUOID < 0) || (dataTuple.getUOID() < nodeUOID)))
                    nodeUOID = dataTuple.getUOID();
                break;
                
            case PathMetadata.DATA_ONLY:
                // check in data (node must be associated to a tuple else would be discarded)
                dataTuple = tuplePool.get(node.getReadTableMapping().getTableMapping().getTableIndex());
                if (dataTuple.isInRangeAbs(first, last, node.getPathID()))
                    nodeUOID = dataTuple.getUOID();
                break;
                
            case PathMetadata.DATA_AND_STRUCT:
                // check in struct
                if (tuplePool.getStructNode().isInRange(first, last, node.getPathID()))
                    nodeUOID = tuplePool.getStructNode().getUOID();
                else if (node.getReadTableMapping() != null) // for instance if mixed with children
                {
                    // check in data (node must be associated to a tuple else would be discarded)
                    dataTuple = tuplePool.get(node.getReadTableMapping().getTableMapping().getTableIndex());
                    if (dataTuple.isInRangeAbs(first, last, node.getPathID()))
                        nodeUOID = dataTuple.getUOID();
                }
                break;
            case PathMetadata.STRUCT_ONLY:
                // check in struct
                if (tuplePool.getStructNode().isInRange(first, last, node.getPathID()))
                    nodeUOID = tuplePool.getStructNode().getUOID();
                break;
                
            case PathMetadata.DISCARDED:
                // the null column case or DISCARDED because not optional
                if (node.getReadColumnMapping() == null)
                    nodeUOID = NO_UOID_FOUND;   // discarded because intermediate mandatory node
                else if (isMappedValueNull(node))
                {
                    if (node.isNilImplicit())
                        nodeUOID = NO_UOID_FOUND;    // implicit nil (else an extra node would have been found
                }
                else
                    nodeUOID = NO_UOID_FOUND;
                break;
            default:
                throw new RepositoryException(RepositoryException.CONSISTENCY_ERROR,
                "Metadata is corrupted for the collection "
                + collection.getCollectionName() + "(Unknow storage flag "
                + node.getStorageMode() + " for pid " + node.getPathID() + ")");
        }
        return nodeUOID;
    }
    
    
    /** generate all the events attached to that specific element path. DO NOT PROCESS
     * SUBPATHS ! Note: struct is consumed (multivaluated process out of method by caller).
     * perform startElement and startPrefix but not end...
     * @param first UOID of deepest ancestor-or-self having a struct tuple
     * @param last range last of deepest ancestor-or-self having a struct tuple
     */
    private void processElement(MetadataIterator metaContext, long first, 
                                long last, long uoid, long anchor,
                                ElementInfoSet eltInfo)
    throws XMLDBCException, SQLException, SAXException
    {
        PathNode wPath = metaContext.getPathNode();
        short eltPID = wPath.getPathID();
        
        // process extra-data (ns prefixes, xsi atts, comments, PI...)
        if (processExtraData(anchor, eltPID, eltInfo))
            eltInfo.setStatus(NIL_ELEMENT);
        
        // generate prefix mapping if not taken from database and not declared at superior level
        String nsUri = wPath.getNamespace();
        if ((nsUri!= null) && !generateDBPrefix)
        {
            String pfx = prefixProvider.getPrefix(nsUri);
            if (pfx == null)
            {
                pfx = prefixProvider.getGeneratedPrefix(wPath);
                eltInfo.generateStartPrefixMapping(pfx, nsUri);
            }
        }
      
        if (eltPID == ROOT_PATH_ID)
        {
            // process root extra-nodes but not startElement
            eltInfo.flushExtraData(true, true);
            eltInfo.setStatus(FOUND_ELEMENT);
            return;
        }
        
        // attributes without tuples : browse path node attributes
        Collection children = metaContext.getPathNode().getChildren();
        if (children != null)
        {
            String value;
            Iterator it = children.iterator();
            
            while (it.hasNext())
            {
                wPath = (PathNode)it.next();
                if (wPath.getType() != NodeKind.ATTRIBUTE)
                    continue;
                
                if (wPath.getReadTableMapping() == null)   // tuple should be present
                {
                    value = getMappedValue(wPath, first, first, true, false, false);
                    if (value != null)
                        attributes.addAttribute(wPath.getNamespace() == null ? "" : wPath.getNamespace(),
                        wPath.getLocalName(), "", ATTRIBUTE_TYPES[7], value);
                }
            }
        }
        // Process the column mapping nil without extra-tuple (simple column mapping):
        // need to anticipate the column mapping processing to see if null
        wPath = metaContext.getPathNode();

        if ((eltInfo.getStatus() != NIL_ELEMENT)
            && ((wPath.getReadColumnMapping() != null) || wPath.isMixed()))
        {
            // store element character data
            eltInfo.setValue(getElementCharData(wPath, uoid, eltInfo));
            if (eltInfo.getValue() == null)
            {
                if (wPath.isNilImplicit())
                {
                    attributes.addAttribute(
                    SchemaConstants.XSI_URI,
                    SchemaConstants.XSI_NIL_ATTR,
                    "",
                    ATTRIBUTE_TYPES[7],
                    SchemaConstants.TRUE_VALUE
                    );
                    eltInfo.setStatus(NIL_ELEMENT);
                }
                else if (eltInfo.getStatus() == UNKNOWN)    // Discarded node
                    eltInfo.setStatus(MISSING_ELEMENT);
            }
            else
                eltInfo.setStatus(FOUND_ELEMENT);
        }
        
        // generate now attributes with table mapping (OID allocated)
        // do it after fetching element value !
        processAttributes(first, last, metaContext);
                
    }
    
    private boolean processExtraData(long anchor, short eltPID,
                                    ElementInfoSet eltInfo)
    throws SQLException, SAXException
    {
        boolean nil = false;
        ExtraDataExplorer extraNodeBrowser = tuplePool.getExtraNode();
        while ((extraNodeBrowser.getRowNum() > eltInfo.getLastRowNum())
                && (extraNodeBrowser.getPath() == eltPID)
                && (extraNodeBrowser.getUOID() == anchor)
                ||
                ((eltPID == ROOT_PATH_ID) && (extraNodeBrowser.getPath() == ROOT_PATH_ID))
                )
        {
            switch (extraNodeBrowser.getType())
            {
                case NodeKind.NAMESPACE:
                    String data = extraNodeBrowser.getData();   // "{" + uri + "}" + prefix
                    int i = data.indexOf('}');
                    String pfx = data.substring(i+1);
                    String uri = data.substring(1, i);
                    if (prefixProvider.declarePrefix(pfx, uri) || generateDBPrefix) // generate it in query reconstruction if no defined in query
                        eltInfo.generateStartPrefixMapping(pfx, uri);
                    break;
                case NodeKind.PI:
                case NodeKind.COMMENT:
                case START_CDATA_SECTION:
                case END_CDATA_SECTION:
                    eltInfo.addNode(extraNodeBrowser.copyNode());
                    break;
                case XSI_NIL:
                    attributes.addAttribute(
                        SchemaConstants.XSI_URI,
                        SchemaConstants.XSI_NIL_ATTR,
                        "",
                        ATTRIBUTE_TYPES[7],
                        extraNodeBrowser.getData()
                        );
                    nil = extraNodeBrowser.isNil();
                    break;
                case XSI_NO_NAMESPACE_SCHEMA_LOCATION:
                    attributes.addAttribute(
                        SchemaConstants.XSI_URI,
                        SchemaConstants.XSI_NO_NAMESPACE_SCHEMA_LOCATION_ATTR,
                        "",
                        ATTRIBUTE_TYPES[7],
                        extraNodeBrowser.getData()
                        );
                    break;
                case XSI_SCHEMA_LOCATION:
                    attributes.addAttribute(
                        SchemaConstants.XSI_URI,
                        SchemaConstants.XSI_SCHEMA_LOCATION_ATTR,
                        "",
                        ATTRIBUTE_TYPES[7],
                        extraNodeBrowser.getData()
                        );
                    break;
                case XSI_TYPE:
                    String typeString = extraNodeBrowser.getData();
                    String nsURI = null, localName = null;
                    int bracket1 = typeString.indexOf('{');
                    int bracket2 = typeString.indexOf('}');
                    nsURI = typeString.substring(bracket1 + 1, bracket2 - bracket1);
                    localName = typeString.substring(bracket2 + 1);
                    
                    eltInfo.setXSIType(schemaManager.getType(nsURI, localName));
                    attributes.addAttribute(
                        SchemaConstants.XSI_URI,
                        SchemaConstants.XSI_TYPE_ATTR,
                        "",
                        ATTRIBUTE_TYPES[7],
                        schemaManager.getQNameType().toXMLString(typeString, prefixProvider)
                        );
                    break;
                // change   --> comments
//                case NodeKind.ELEMENT:
//                    // position or occurence without xsi
//                    break;
                default:
                    // not expected : ignore
            }
            eltInfo.setExtraRowNum(extraNodeBrowser.getRowNum());
            extraNodeBrowser.fetchNextRow();
        }
        return nil;
    }

    /**
     * generate attributes with table mapping (OID allocated) for a particular
     * range.
     */
    
    private void processAttributes(long first, long last, MetadataIterator metaContext)
    throws XMLDBCException, SQLException
    {
        // WARNING ! : because of OR5 : attributes are stored only in data tables
        Tuple electedTuple = null;
        PathNode attNode;
        String value;
        Collection mappingSubset = metaContext.getPathMetadata().getBuildMappings();
        short eltPID = metaContext.getPathMetadata().getPathID();
        
        // loop till out of "range"
        while ((electedTuple = tuplePool.pollData(mappingSubset, first + 1, last)) != null)
        {
            attNode = metaContext.getPathNode(electedTuple.getPathID());
            // could check the attribute type and that it is attached to the right element
            if ((attNode.getType() != org.xquark.xpath.NodeKind.ATTRIBUTE)
            || (((PathNode)attNode.getParent()).getPathID() != eltPID))
                break;
            
            value = getMappedValue(attNode, first, last, false, true, false);
            if (value != null)    // not missing (optional)
            {
                attributes.addAttribute(attNode.getNamespace() == null ? "" : attNode.getNamespace(),
                attNode.getLocalName(), "", ATTRIBUTE_TYPES[7], value);
            }
        }
    }
    
    private boolean isMappedValueNull(PathMetadata node)
    throws SQLException, XMLDBCException
    {
        boolean  ret = false;
        ColumnMapping column = node.getReadColumnMapping();
        if (column != null) // no mapping
        {
            Tuple tuple = tuplePool.get(column.getTableIndex());
            if (tuple.getValue(column, node.getDeclaration(), prefixProvider) == null)
                ret = true;
        }
        return ret;
    }
    
    /**
     *
     */
    protected String getMappedValue(PathMetadata node, long first, long last,
                    boolean checkRange, boolean lookup, boolean systematic)
    throws SQLException, XMLDBCException
    {
        String ret =null;
        ColumnMapping column = node.getReadColumnMapping();
        if (column != null) // no mapping
        {
            Tuple tuple = tuplePool.get(column.getTableIndex());
            if (!checkRange || tuple.isInRange(first, last))
                ret = tuple.getValue(column, node.getDeclaration(), prefixProvider);
            if (lookup && (systematic || (tuple.getPathID() == node.getPathID())))
                tuple.fetchNextRow();
        }
        return ret;
    }
    
    /**
     * returns the value only if the table path OID fits the tuple one.
     * performs lookup if table path is node path
     * Used for simple mapping.
     */
    private String getElementCharData(PathMetadata node, long eltUOID,
    ElementInfoSet eltInfo)
    throws SQLException, XMLDBCException
    {
        String ret = null;
        ColumnMapping column = node.getReadColumnMapping();
        if (column == null)
            column = node.getTextColumnMapping();
        PathMetadata tableNode = node.getTableMappingPath();
        Tuple tuple = tuplePool.get(column.getTableIndex());
        if (((eltUOID < 0) || (tuple.getUOID() == eltUOID))
        && (tuple.getPathID() == (int)tableNode.getPathID()))
        {
            ret = tuple.getValue(column, node.getDeclaration(), prefixProvider);
            if (tableNode.getPathID() == node.getPathID())
                tuple.fetchNextRow();
        }
        return ret;
    }
    
    public void close() throws RepositoryException
    {
        if (tuplePool != null)
        {
            try
            {
                tuplePool.close();
                tuplePool = null;
                pathSet = null;
                attributes = null;
                prefixProvider = null;
                collection = null;
                contentHandler = null;
                lexicalHandler = null;
                errorHandler = null;
            }
            catch (SQLException e)
            {
                throw new RepositoryException(RepositoryException.DB_ERROR, "JDBC error while freeing ressources", e);
            }
        }
    }
    
    public void reset() throws RepositoryException, SQLException
    {
        // TO DO : Reset method metaContext;
        attributes.clear();
        tuplePool.reset();
        prefixProvider.clear();
    }
    
    /** Release handlers.
     * @throws RepositoryException an application exception
     */
    public void unPlug()
    {
        contentHandler = null;
        lexicalHandler = null;
        errorHandler = null;
    }
    
    
    private void indent(int indent) throws SAXException
    {
        int quotient = indent/INDENT_ARRAY_TAB_COUNT;
        
        if (quotient == 0)  // one array is enough (include eol)
            contentHandler.ignorableWhitespace(INDENT_CHAR_ARRAY, 0, indent + 1);
        else
        {
            // generate 1st array of tabs with initial eol
            contentHandler.ignorableWhitespace(INDENT_CHAR_ARRAY, 0, INDENT_ARRAY_TAB_COUNT + 1);
            for (int i = 2; i  <= quotient; i++)
            {
                contentHandler.ignorableWhitespace(INDENT_CHAR_ARRAY, 1, INDENT_ARRAY_TAB_COUNT);
            }
            int remainder = indent%INDENT_ARRAY_TAB_COUNT;
            if (remainder > 0)
                contentHandler.ignorableWhitespace(INDENT_CHAR_ARRAY, 1, remainder);
        }
    }
    
    ////////////////////////////////////////////////////////////////////////////////
    //                          INNER CLASSES
    ////////////////////////////////////////////////////////////////////////////////
    
    protected class ElementInfoSet
    {
        private ArrayList nodesToInsert = new ArrayList(2);
        private int childCount = 0;
        private int charCount = 0;
        private String value = null;
        private boolean notGenerated = true;
        private boolean extraFlushed = false;
        private short status = UNKNOWN;
        private short lastNum = -1;
        private List prefixes = new ArrayList(2);
        private org.xquark.schema.Type xsiType = null;

        void addNode(ExtraNode node)
        {
            nodesToInsert.add(node);
            extraFlushed = false;
        }
        void clear()
        {
            nodesToInsert.clear();
            charCount = 0;
            lastNum = -1;
            value = null;
            status = UNKNOWN;
            notGenerated = true;
            xsiType = null;
            extraFlushed = false;
        }
        
        void generateStartPrefixMapping(String pfx, String uri)
        throws SAXException
        {
            prefixes.add(pfx);
            contentHandler.startPrefixMapping(pfx, uri);
        }
        
        void generateEndPrefix()
        throws SAXException
        {
            int prefixCount = prefixes.size();
            String pfx;
            for (int i = 0; i < prefixCount; i++)
            {
                pfx = (String)prefixes.get(i);
                contentHandler.endPrefixMapping(pfx);
            }
        }
        
        void setValue(String value)
        {
            this.value = value;
        }
        
        void setXSIType(org.xquark.schema.Type type)
        {
            xsiType = type;
        }
        
        org.xquark.schema.Type getXSIType()
        {
            return xsiType;
        }
        
        String getValue()
        {
            return value;
        }
        
        boolean hasValue()
        {
            return value != null;
        }
        
        boolean elementGenerated()
        {
            return !notGenerated;
        }
        
        boolean extraFlushed()
        {
            return extraFlushed;
        }
        
        boolean hasExtraData()
        {
            return nodesToInsert.size() != 0;
        }
        
        boolean leafElementButNotMissing()
        {
            return  (childCount == 0) &&    (
                                            (status != MISSING_ELEMENT)
                                            || (value != null)
                                            || (attributes.getLength() != 0)
                                            );
        }
        
        void setStatus(short status)
        {
            this.status = status;
        }
        
        short getStatus()
        {
            return status;
        }
        
        void incChildCount()
        {
            childCount++;
        }
        
        int getChildCount()
        {
            return childCount;
        }
        
        void setExtraRowNum(short rowNum)
        {
            lastNum = rowNum;
        }
        
        short getLastRowNum()
        {
            return lastNum;
        }
        
        void generateStartElement(PathNode path)
        throws RepositoryException, SAXException
        {
            if (notGenerated &&  (path.getPathID() != ROOT_PATH_ID))
            {
                // start element (if OID IN DATA ? or NIL or extraTuple)
                contentHandler.startElement(
                path.getNamespace() == null ? "" : path.getNamespace(),
                path.getLocalName(),
                "",
                attributes);
                attributes.clear();
                
                if (value != null)  // extra-data will be flushed later
                    generateCharacters(value);
                notGenerated = false;
            }
      }
        
        /** For text nodes
         * if first = -1, means monovaluated
         */
        void generateTextNode(long first, long last, PathNode path)
        throws XMLDBCException, SQLException
        {
            Tuple defaultTuple = tuplePool.get(defaultMappingIndex);
            if (defaultTuple.isInRangeAbs(first, last , path.getPathID()))
                generateTextNode();
        }
        
        /** For text nodes
         * if first = -1, means monovaluated
         */
        void generateTextNode()
        throws RepositoryException, SQLException, XMLDBCException
        {
            Tuple defaultTuple = tuplePool.get(defaultMappingIndex);
            generateCharacters(defaultTuple.getValue(defaultColumnMapping, null, prefixProvider));
            defaultTuple.fetchNextRow();
            incChildCount();    // TEXT node
        }
        
        /** for element data */
        void generateCharacters(String textData) throws RepositoryException
        {
            /* interpolate character data with insertedNodes in the list and send SAX events */
            int lastOffset = 0;
            Iterator it = nodesToInsert.iterator();
            ExtraNode node;
            char[] data = null;
            
            if (textData == null)
            {
                if (nodesToInsert.size() == 0)
                    return;
                data = new char[0];
            }
            else
                data = textData.toCharArray();
            
            int maxGlobalOffset = charCount + data.length;
            while (it.hasNext())
            {
                node = (ExtraNode)it.next();
                if (node.offset > (lastOffset + charCount))
                {
                    ///////////////////////
                    //  SAX Characters
                    ///////////////////////
                    try
                    {
                        contentHandler.characters(data, lastOffset,
                        (node.offset < maxGlobalOffset ? node.offset : maxGlobalOffset)  - (lastOffset + charCount));
                    }
                    catch(SAXException e)
                    {
                        throw new RepositoryException(RepositoryException.SAX_OUTPUT_ERROR, "SAX Exception thrown during characters() with the following message : "+e.getMessage());
                    }
                    
                    lastOffset = node.offset - charCount;
                }
                
                if ((node.position == childCount + 1) && (node.offset <= maxGlobalOffset))
                {
                    try
                    {
                        switch (node.type)
                        {
                            case NodeKind.PI:
                                StringTokenizer st = new StringTokenizer(node.data);
                                contentHandler.processingInstruction(st.nextToken(), st.nextToken(""));
                                break;
                                
                            case NodeKind.COMMENT :
                                if (lexicalHandler != null)
                                    lexicalHandler.comment(node.data.toCharArray(), 0, node.data.length());
                                break;
                            case START_CDATA_SECTION:
                                if (lexicalHandler != null)
                                    lexicalHandler.startCDATA();
                                break;
                            case END_CDATA_SECTION:
                                if (lexicalHandler != null)
                                    lexicalHandler.endCDATA();
                                break;
                            default :
                        }
                    }
                    catch(SAXException e)
                    {
                        String methodName = null;
                        switch (node.type)
                        {
                            case NodeKind.PI:
                                methodName = "processingInstruction()";
                                break;
                            case NodeKind.COMMENT :
                                methodName = "comment()";
                                break;
                            case START_CDATA_SECTION:
                                methodName = "startCDATA()";
                                break;
                            case END_CDATA_SECTION:
                                methodName = "endCDATA()";
                                break;
                            default :
                        }
                        throw new RepositoryException(RepositoryException.SAX_OUTPUT_ERROR,
                        "SAX Exception thrown during " + methodName
                        + " with the following message : "+e.getMessage());
                    }
                    it.remove();
                    incChildCount();    // child count must only be incremented when the rigth place is found
                }
            }
            // process potential remaining character data
            if (lastOffset < data.length) // characters have to be saved
            {
                ///////////////////////
                //  SAX Characters
                ///////////////////////
                try
                {
                    contentHandler.characters(data, lastOffset, data.length - lastOffset);
                }
                catch(SAXException e)
                {
                    throw new RepositoryException(RepositoryException.SAX_OUTPUT_ERROR, "SAX Exception thrown during characters() with the following message : "+e.getMessage());
                }
            }
            
            // increment element counter
            charCount = maxGlobalOffset;
        }
        
        /** for extra node without character data */
        void flushExtraData(boolean carriageReturn, boolean checkPosition)
        throws RepositoryException, SAXException
        {
            extraFlushed = true;    // flag setting
            
            Iterator it = nodesToInsert.iterator();
            ExtraNode node;
            while (it.hasNext())
            {
                node = (ExtraNode)it.next();
                if ( !checkPosition || (node.position == childCount + 1)) // do not check offset : trim whitespaces may carry out "out of bounds"...
                {
                    if (carriageReturn)
                        contentHandler.ignorableWhitespace(INDENT_CHAR_ARRAY, 0, 1);
                   switch (node.type)
                    {
                        case NodeKind.PI:
                            StringTokenizer st = new StringTokenizer(node.data);
                            try
                            {
                                contentHandler.processingInstruction(st.nextToken(), st.nextToken(""));
                            }
                            catch(SAXException e)
                            {
                                throw new RepositoryException(RepositoryException.SAX_OUTPUT_ERROR, "SAX Exception thrown during processingInstruction() with the following message : "+e.getMessage());
                            }
                            break;
                            
                        case NodeKind.COMMENT :
                            if (lexicalHandler != null)
                            {
                                try
                                {
                                    lexicalHandler.comment(node.data.toCharArray(), 0, node.data.length());
                                }
                                catch(SAXException e)
                                {
                                    throw new RepositoryException(RepositoryException.SAX_OUTPUT_ERROR, "SAX Exception thrown during comment() with the following message : "+e.getMessage());
                                }
                            }
                            break;
                        default :
                    }
                    it.remove();
                    incChildCount();    // child count must only be incremented when the rigth place is found
                }
            }
        }
    }
    
    /* TO IMPROVE: This implementation ensures one prefix for one namespace but is
     * not contextual.
     */
    protected class PrefixProvider implements ValidationContextProvider
    {
        private NamespaceContextStack nsContextStack;
        private int generatedPrefixCount = 1;
        
		public PrefixProvider()
		{
			nsContextStack = new NamespaceContextStack();
		}

		public PrefixProvider(NamespaceContextStack nsContextStack)
		{
			this.nsContextStack = nsContextStack;
		}

        /** To override for queries
         */
        public String getPrefix(String uri)
        {
            return nsContextStack.getPrefix(uri);
        }
        
        String getGeneratedPrefix(PathNode path)
        {
            String newPfx = null;
            
            // look in the declaration first
            Declaration decl = path.getDeclaration();
            if (decl != null)
            	newPfx = decl.getSchema().getPrefix();
            
            // generate new prefix
            if ((newPfx == null) || (newPfx.length() == 0))
            {
                do
                {
                    newPfx = "ns" + generatedPrefixCount++;
                }
                while (nsContextStack.getNamespaceURI(newPfx) != null);
            }
           
            declarePrefix(newPfx, path.getNamespace());
            return newPfx;
        }
        
        /** @return true if the prefix is new
         */
        protected boolean declarePrefix(String prefix, String uri)
        {
            boolean newPfx = (getPrefix(uri) == null);
            if (newPfx)
        		nsContextStack.declarePrefix(prefix, uri);
            return newPfx;
        }
        
        public String getNamespaceURI(String prefix)
        {
            return nsContextStack.getNamespaceURI(prefix);
        }
        
        public String getDocumentBase()
        {
            return null;
        }
        
        void clear()
        {
            nsContextStack = new NamespaceContextStack();
            generatedPrefixCount = 1;
        }
        
        public java.util.Map getNotationDeclarations()
        {
            return null;
        }
        
        void push()
        {
        	nsContextStack.pushContext();
        }
        
    	void pop()
    	{
    		nsContextStack.popContext();
    	}
    }
}
