/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.mapper.storage;

import java.io.StringWriter;
import java.sql.*;
import java.util.*;

import org.xquark.mapper.RepositoryConnection;
import org.xquark.mapper.RepositoryDocument;
import org.xquark.mapper.RepositoryException;
import org.xquark.mapper.dbms.*;
import org.xquark.mapper.mapping.MappingConstants;
import org.xquark.mapper.mapping.MappingFactory;
import org.xquark.mapper.mapping.MappingSetIterator;
import org.xquark.mapper.metadata.*;
import org.xquark.mapper.util.*;
import org.xquark.schema.validation.SchemaValidationContext;
import org.xquark.util.SAXConstants;
import org.xquark.xml.xdbc.*;


/** Common part ot the {@link org.xquark.mapper.XMLCollection XMLCollection}
 * implementations.
 *
 * @see org.xquark.mapper.XMLCollection
 */

public class XMLCollectionImpl
implements _RepositoryCollection, RepositoryConstants, SAXConstants
{
    private static final String RCSRevision = "$Revision: 1.2 $";
    private static final String RCSName     = "$Name: XQ1_1_1 $";
    //
    // Data
    //
    protected _RepositoryConnection repConnection;
    protected AbstractConnection connection;
    private DestructionToken destructor;
    private CollectionMetadata metadata;
    
    /* data for removal and renaming */
    protected JDBCBatcher batcher;
    protected PreparedStatement removeStmt;
    protected PreparedStatement removeStructStmt;
    protected PreparedStatement removeExtraStmt;
    protected ArrayList removeValuesStmts;
    protected PreparedStatement renameStmt;
    
    protected HashSet filers = new HashSet(); // to close filers when closing & set modes
                                              // contains SAXHandlers & not XMLDocumentFilers !!!
    protected RepositoryReader cachedReader; // TO DO : instead of caching reader : put pstmt here !
    private ArrayList removedDocDIDs;
    
    private boolean readOnly = false;
    
    /** Constructor.
     * @param name The name of the repository.
     * @param config A CollectionConfig object.
     * and shared with other repositories of the same database.
     * @exception java.sql.SQLException Relational database exception.
     */
    
    public XMLCollectionImpl(CollectionMetadata metadata,
    _RepositoryConnection connection, DestructionToken destructor)
    {
        repConnection = connection;
        this.connection = connection.getConnection();
        this.destructor = destructor;
        this.metadata = metadata;
        MappingFactory factory = new MappingFactory(
            metadata.getMapping(),
            connection.getSchemaManager(),
            connection
            );
    }
    
    public CollectionMetadata getMetadata()
    {
        return metadata;
    }
    
    /** Get an XMLReader object in order to perform SAX2 parsing which is the
     * most performant method.
     * @return an XMLReader object.
     */
    public RepositoryReader getReader() throws XMLDBCException
    {
        RepositoryReader ret;
        if (cachedReader == null)
            ret = new NSPrefixRepositoryFilter(new DocumentProducer(this));
        else
        {
            ret = cachedReader;
            cachedReader = null;
        }
        return ret;
    }
    
    
    /** Get an XMLReader object in order to perform SAX2 parsing which is the
     * most performant method.
     */
    public void releaseReader(RepositoryReader reader)
    throws RepositoryException
    {
        if (cachedReader == null)
        {
            reader.unPlug();
            cachedReader = reader;
        }
        else
            reader.close();
    }
    
    /** Free all resources. Repository is warned that repository is closed.
     * <B>Any subsequent call to a function of the object will raise an
     * exception.</B>
     */
    public void close() throws XMLDBCException
    {
        // Must be done first because destruct call getCollectionName()
        if (destructor != null)
        {
            try
            {
                destructor.destruct(this);
            }
            catch (RepositoryRuntimeException e)
            {
                throw e.getRepositoryException();
            }
            destructor = null;
            
            repConnection = null;
            connection = null;
            
            metadata = null;
            
            closeFilers();
            filers = null;
            
            if (cachedReader != null)
                cachedReader.close();

            if (batcher != null) // does not close statements
                try
                {
                    batcher.close();
                }
                catch (SQLException e)
                {
                    throw new RepositoryException(RepositoryException.DB_ERROR,
                    "Could not close the JDBC Batcher");
                }
            try
            {
                if (removeStmt != null)
                {
                    removeStmt.close();
                    removeStmt = null;
                }
                if (removeStructStmt != null)
                {
                    removeStructStmt.close();
                    removeStructStmt = null;
                }
                if (removeExtraStmt != null)
                {
                    removeExtraStmt.close();
                    removeExtraStmt = null;
                }
                if (renameStmt != null)
                {
                    renameStmt.close();
                    renameStmt = null;
                }
                if (removeValuesStmts != null)
                {
                    int len = removeValuesStmts.size();
                    for (int i = 0; i < len; i++)
                        ((PreparedStatement)removeValuesStmts.get(i)).close();
                    removeValuesStmts = null;
                }
            }
            catch (SQLException e)
            {
                // no op
            }
        }
    }
    
    public _RepositoryConnection getRepositoryConnection()
    {
        return repConnection;
    }
    
    public class HandlerDestructor implements DestructionToken
    {
        public void destruct(Object o)
        {
            filers.remove(o);
        }
    }
    
    public long getDocumentID(String ID)
    throws SQLException, RepositoryException
    {
        Statement statement = null;
        long docOID = -1;
        try
        {
            statement = connection.getConnection().createStatement();
            ResultSet rs = statement.executeQuery(
            metadata.getSelectDocIDStmt() + ID + "'");
            if(rs.next())
                docOID = rs.getLong(1);
            rs.close();
        }
        finally
        {
            if (statement != null)
                statement.close();
        }
        return docOID;
    }
    
    public CollectionInfo getInfo()
    {
        return metadata.getInfo();
    }
    
    public String getCollectionName()
    {
        return metadata.getCollectionName();
    }
    
    public String toString()
    {
        StringWriter out = new StringWriter();
        Tabulator tabulator = new Tabulator
                                    (
                                    out,
                                    79,
                                    new int[] {0, 40}
                                    );
                                    
                                    
        tabulator.addItem("name");
        tabulator.addItem(getCollectionName());
        
        String[] properties = getPropertyList();
        for (int i = 0; i < properties.length; i++)
        {
            tabulator.addItem(properties[i].substring(properties[i].lastIndexOf('/') + 1));
            try
            {
                tabulator.addItem(getProperty(properties[i]));
            }
            catch (XMLDBCNotRecognizedException e)
            {
                // no op
            }
        }
        
        String[] features = getFeatureList();
        for (int i = 0; i < features.length; i++)
        {
            tabulator.addItem(features[i].substring(features[i].lastIndexOf('/') + 1));
            try
            {
                tabulator.addItem(String.valueOf(getFeature(features[i])));
            }
            catch (XMLDBCNotRecognizedException e)
            {
                // no op
            }
        }
        return out.toString();
    }
    
    ///////////////////////////////////////////////////////////////////////////
    //                     XMLCollection implementation
    ///////////////////////////////////////////////////////////////////////////
    
    /**
     * Creates a "filer" object that can be used to insert a new XML document
     * into the XML collection. A single filer can store several
     * documents sequentially, however it is an error to try to use a filer
     * before the current storage is finished.<BR>
     * @return an XMLDocumentFiler object.
     * @throws XMLDBCException if a database access error occurs.
     */
    public XMLDocumentFiler getDocumentFiler() throws XMLDBCException
    {
        protectReadOnlyMethod();
        SchemaValidationContext validationContext =
        new SchemaValidationContext(repConnection.getSchemaManager());
        ValidatingFilerAdapter validator = new ValidatingFilerAdapter(validationContext, repConnection);
        validator.setDocumentBase("http://www.fake.com/");
        CollectionFiler handler =  new CollectionFiler
        (
        this,
        validationContext,
        !getFeature(RepositoryConnection.USE_SCHEMA_PREFIXES_FEATURE),
        new XMLCollectionImpl.HandlerDestructor()
        );
        validator.setSAXHandler(handler);
        filers.add(handler);
        return new XDBCFilerAdapter(validator, handler, handler, handler);
    }
    
    
    public XMLConnection getConnection() throws XMLDBCException
    {
        return getRepositoryConnection();
    }
    
    public XMLDocument getDocument(String documentID) throws XMLDBCException
    {
        return getRepositoryDocument(documentID);
    }
    
    public boolean renameDocument(String oldID, String newID) throws XMLDBCException
    {
        return renameDocument(oldID, newID, true);
    }
    
    /** @param useTransaction used during schema storage */
    boolean renameDocument(String oldID, String newID, boolean useTransaction) throws XMLDBCException
    {
        protectReadOnlyMethod();
        boolean ret = false;
        // flush potential pending storages...
        flushFilers();
        // performs an update of the document list
        if (useTransaction)
            connection.start();
        PreparedStatement statement = null;
        try
        {
            statement = connection.getConnection().prepareStatement(metadata.getRenameDocStmt());
            statement.setString(1, newID);
            statement.setString(2, oldID);
            statement.executeUpdate();
        }
        catch(SQLException e)
        {
            throw new RepositoryException(RepositoryException.DB_ERROR, "Relational database error while renaming a document in the document table.", e);
        }
        finally
        {
            try
            {
                if (statement != null)
                    statement.close();
            }
            catch (SQLException e)
            {
                // no op
            }
        }
        if (useTransaction && getConnection().getAutoCommit())
        {
             connection.commit();
             commitPerformed();
        }
        return ret;
    }
    
    public boolean removeDocument(String documentID) throws XMLDBCException
    {
        protectReadOnlyMethod();
        boolean ret = false;
        // flush potential pending storages...
        flushFilers();
        ret = removeDocumentUsingBatch(documentID);
        batcher.flush();
        if (getConnection().getAutoCommit())
        {
             connection.commit();
             commitPerformed();
        }
        return ret;
    }
    
    public boolean containsDocument(String documentID) throws XMLDBCException
    {
        boolean contains = false;
        // check whether this document ID is present in the document table
        Statement statement = null;
        try
        {
            statement = connection.getConnection().createStatement();
            ResultSet rs = statement.executeQuery(metadata.getContainsDocStmt() + documentID + "'");
            
            if (rs.next()&& (rs.getInt(1) > 0))
                contains = true;
            
            rs.close();
        }
        catch(SQLException e)
        {
            throw new RepositoryException(RepositoryException.DB_ERROR, "Relational database error while checking for document existence.", e);
        }
        finally
        {
            try
            {
                if (statement != null)
                    statement.close();
            }
            catch (SQLException e)
            {
                // no op
            }
        }
        return contains;
    }
    
    
    public String getName() throws XMLDBCException
    {
        return metadata.getCollectionName();
    }
    
    public void setDescription(String description) throws XMLDBCException
    {
        metadata.updateDescription(description);
    }
    
    public String getDescription() throws XMLDBCException
    {
        return getInfo().getDescription();
    }
    
    public long getDocumentCount() throws XMLDBCException
    {
        // get the whole repository document table
        long numDocs = 0;
        Statement statement = null;
        
        try
        {
            statement = connection.getConnection().createStatement();
            ResultSet rs = statement.executeQuery(metadata.getCountDocStmt());
            
            if (rs.next())
                numDocs = rs.getLong(1);
            rs.close();
        }
        catch(SQLException e)
        {
            throw new RepositoryException(RepositoryException.DB_ERROR, "Relational database error while retrieving document list.", e);
        }
        finally
        {
            try
            {
                if (statement != null)
                    statement.close();
            }
            catch (SQLException e)
            {
                // no op
            }
        }
        return numDocs;
    }
    
    public List getIdentifierList() throws XMLDBCException
    {
        ArrayList list = new ArrayList();
        
        // get the whole repository document table
        Statement statement = null;
        try
        {
            statement = connection.getConnection().createStatement();
            ResultSet rs = statement.executeQuery(metadata.getDocListStmt());
            
            while (rs.next())
            {
                list.add(rs.getString(1));
            }
            rs.close();
        }
        catch(SQLException e)
        {
            throw new RepositoryException(RepositoryException.DB_ERROR, "Relational database error while retrieving document list.", e);
        }
        finally
        {
            try
            {
                if (statement != null)
                    statement.close();
            }
            catch (SQLException e)
            {
                // no op
            }
        }
        return list;
    }
    
    public int removeAllDocuments() throws XMLDBCException
    {
        protectReadOnlyMethod();
        // flush potential pending storages...
        flushFilers();
        Iterator it = getIdentifierList().iterator();
        int docCount = 0;
        
        while (it.hasNext())
        {
            removeDocumentUsingBatch((String)it.next());
            docCount++;
            if (getConnection().getAutoCommit())
            {
                batcher.flush();
                connection.commit();
                 commitPerformed();
            }
        }
        if (batcher != null)
            batcher.flush();
        return docCount;
    }
    
    public void setReadOnly(boolean readOnly) throws XMLDBCException
    {
        if (readOnly)
            closeFilers();
        this.readOnly = readOnly;
    }
    
    public boolean isReadOnly() throws XMLDBCException
    {
        return readOnly;
    }
    
    ///////////////////////////////////////////////////////////////////////////
    //                     RepositoryCollection implementation
    ///////////////////////////////////////////////////////////////////////////
    public List getRepositoryDocumentList() throws XMLDBCException
    {
        return metadata.getDocumentInfoList(this);
    }
    
    public RepositoryDocument getRepositoryDocument(String documentID) throws XMLDBCException
    {
        return new RepositoryDocumentImpl(documentID, this);
    }
    
    public void createIndexes() throws RepositoryException
    {
        protectReadOnlyMethod();
        metadata.createIndexes();
    }
    
    public void dropIndexes() throws RepositoryException
    {
        protectReadOnlyMethod();
        metadata.dropIndexes();
    }
    
    public void updateStatistics() throws XMLDBCException
    {
        protectReadOnlyMethod();
        metadata.updateStatistics();
    }
    
    ///////////////////////////////////////////////////////////////////////////
    //                     _RepositoryCollection implementation
    ///////////////////////////////////////////////////////////////////////////
    public void flushFilers() throws XMLDBCException
    {
        Iterator it = filers.iterator();
        while (it.hasNext())
        {
            ((SAXHandler)it.next()).flushBuffer();
        }
    }
    
    public void commitPerformed() throws XMLDBCException
    {
        if (removedDocDIDs != null)
            metadata.getUOIDManager().freeDocOIDs(removedDocDIDs);
    }
    
    public void rollbackPerformed() throws XMLDBCException
    {
        if (removedDocDIDs != null)
            removedDocDIDs.clear();
    }
    
    public void restoreDID(long did)
    {
        if (removedDocDIDs == null)
            removedDocDIDs = new ArrayList();
        removedDocDIDs.add(new Long(did));
    }
    
    ///////////////////////////////////////////////////////////////////////////
    //                     Configurable implementation
    ///////////////////////////////////////////////////////////////////////////
    
    public void setProperty(String propertyId, Object value) throws XMLDBCNotRecognizedException, XMLDBCNotSupportedException
    {
        getInfo().setProperty(propertyId, value);
    }
    
    public Object getProperty(String propertyId) throws XMLDBCNotRecognizedException
    {
        return getInfo().getProperty(propertyId);
    }
    
    public String[] getPropertyList()
    {
        return getInfo().getPropertyList();
    }
    
    public void setFeature(String featureId, boolean state) throws XMLDBCNotRecognizedException, XMLDBCNotSupportedException
    {
        getInfo().setFeature(featureId, state);
    }
    
    public boolean getFeature(String featureId) throws XMLDBCNotRecognizedException
    {
        return getInfo().getFeature(featureId);
    }
    
    public String[] getFeatureList()
    {
        return getInfo().getFeatureList();
    }
    
    ///////////////////////////////////////////////////////////////////////////
    //                          PRIVATE UTILITIES
    ///////////////////////////////////////////////////////////////////////////
    private boolean removeDocumentUsingBatch(String documentID) throws XMLDBCException
    {
        if (batcher == null)
            batcher = new JDBCBatcher(connection, true);
        
        long UDID;
        if (documentID == null)
            throw new RepositoryException(RepositoryException.ILLEGAL_EXPRESSION, "The ID is mandatory to remove document.");
        // remove useless spaces
        documentID = documentID.trim();
        
        try
        {
            UDID = getDocumentID(documentID);
        }
        catch(SQLException e)
        {
            throw new RepositoryException(RepositoryException.DB_ERROR, "Relational database error while retrieving the internal numeric ID.", e);
        }
        if (UDID < 0)
            return false;
        
        long first = metadata.getUOIDManager().getDocMinUOID(UDID);
        long last = metadata.getUOIDManager().getDocMaxUOID(UDID);
        
        // Delete first the document ID
        connection.start();
        try
        {
            if (removeStmt == null)
            {
                String tableName = metadata.getTableInfo(TableSpec.TYPE_DOCS).getName();
                removeStmt = connection.getConnection().prepareStatement(metadata.getDeleteDocIDStmt());
                batcher.addStatement(removeStmt, tableName, RepositoryProperties.getIntProperty(CONF_REMOVE_BATCHSIZE));
            }
            removeStmt.setString(1, documentID);
            batcher.addBatch(removeStmt);
            
            // Delete then document struct rows
            if (removeStructStmt == null)
            {
                TableInfo table = metadata.getTableInfo(TableSpec.TYPE_STRUCT);
                removeStructStmt = StructSaver.getDeleteStatement(connection, table);
                batcher.addStatement(removeStructStmt, table.getName(), RepositoryProperties.getIntProperty(CONF_REMOVE_BATCHSIZE));
            }
            removeStructStmt.setLong(1, first);
            removeStructStmt.setLong(2, last);
            batcher.addBatch(removeStructStmt);
            
            // Delete then document extra-data rows
            if (removeExtraStmt == null)
            {
                TableInfo table = metadata.getTableInfo(TableSpec.TYPE_EXTRA);
                removeExtraStmt = ExtraDataSaver.getDeleteStatement(connection, table);
                batcher.addStatement(removeExtraStmt, table.getName(), RepositoryProperties.getIntProperty(CONF_REMOVE_BATCHSIZE));
            }
            
            removeExtraStmt.setLong(1, first);
            removeExtraStmt.setLong(2, last);
            batcher.addBatch(removeExtraStmt);
            
            // Delete then document value rows
            removeValues(first, last);
            
            // recycle doc OID but only when committed because other instances
            // must not use it before it is effective in the database
            restoreDID(metadata.getUOIDManager().getDID(UDID));
            
        }
        catch(Exception e)
        {
            try
            {
                if (getConnection().getAutoCommit())
                    connection.rollback();
                rollbackPerformed();    // automatic
            }
            catch(SQLException sqle)
            {
                // no op : rdbms is supposed to have performed rollback
            }
            if (e instanceof RepositoryException)
                throw (RepositoryException)e;
            else
                throw new RepositoryException(RepositoryException.DB_ERROR, "Relational database error while deleting document rows.", e);
        }
        return true;
    }
    
    private void removeValues(long first, long last)
    throws SQLException, RepositoryException
    {
        PreparedStatement stmt = null;
        
        /* Put statements in batcher if not initialized */
        if (removeValuesStmts == null)
        {
            removeValuesStmts = new ArrayList();
        /* Delete the values in user tables for table of 'INSERT' type. That means
         * tables where tuples are not supposed to exist before document storage.
         * The collection given by the mapping is supposed to be ordered so there
         * is no reference constraint broken.
         */
            MappingSetIterator it = getMetadata().getSortedTableMappingSetIterator(true);
            CollectionMappingInfo table;
            
            while (it.hasNext())
            {
                table = (CollectionMappingInfo)it.next();
                String SQLString = null;
                
                if (it.isTableFirstOccurence())
                {
	                // user table : only if storage mode is INSERT else data is supposed to pre-exist
	                if (table.getTableMapping().getAction() == MappingConstants.INSERT)
	                {
	                    if (table.getTableMapping().isClustered())
	                        SQLString = table.getDeleteStatement();
	                    else
	                        SQLString = connection.getDeleteUserTableRowsStatement(
	                        table.getTableMapping().getTableName(),
	                        table.getDocumentRowsStatement()
	                        );
	                    stmt = connection.getConnection().prepareStatement(SQLString);
	                    removeValuesStmts.add(stmt);
	                    batcher.addStatement(stmt,table.getTableMapping().getTableName(),
	                    RepositoryProperties.getIntProperty(CONF_REMOVE_BATCHSIZE));
	                }
	                
	                // OID table (in fact user table if clustering...)
	                SQLString = table.getOIDDeleteStatement();
	                if (!table.getTableMapping().isClustered())
	                {
	                    stmt = connection.getConnection().prepareStatement(table.getOIDDeleteStatement());
	                    removeValuesStmts.add(stmt);
	                    batcher.addStatement(stmt,table.getTableMapping().getTableName(),
	                    RepositoryProperties.getIntProperty(CONF_REMOVE_BATCHSIZE));
	                }
	            }
            }
        }
        /* Put statements in the batcher */
        int len = removeValuesStmts.size();
        
        for(int i = 0; i < len; i++)
        {
            stmt = (PreparedStatement)removeValuesStmts.get(i);
            stmt.setLong(1, first);
            stmt.setLong(2, last);
            batcher.addBatch(stmt);
        }
    }
    
    protected void protectReadOnlyMethod() throws RepositoryException
    {
        if (readOnly)
            throw new RepositoryException(RepositoryException.NOT_ALLOWED,
            "Such operation is forbidden because collection is read-only.");
    }
    
    private void closeFilers() throws XMLDBCException
    {
        Iterator it = ((HashSet)filers.clone()).iterator();
        while(it.hasNext())
        {
            ((SAXHandler)it.next()).close();
            it.remove();
        }
    }
} // class
