/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.mapper.storage;

import java.sql.SQLException;

import org.xquark.mapper.RepositoryCollection;
import org.xquark.mapper.RepositoryException;
import org.xquark.mapper.dbms._RepositoryConnection;
import org.xquark.mapper.metadata.CollectionInfo;
import org.xquark.mapper.metadata.CollectionMetadata;
import org.xquark.xml.xdbc.XMLDBCException;

/**
 * Internal version of the XMLCollection interface used for hiding methods to API user.
 */
public interface _RepositoryCollection extends RepositoryCollection
{



	/** Checks whether an XML document exists in this repository and retrieve the
	 * corresponding document OID.
	 * @param ID The name ID of the document
	 * @return the document UDID.
	 * @throws SQLException database error.
	 */
	long getDocumentID(String ID) throws SQLException, RepositoryException;

	/**
	 * Accessor to the internal connection object that created this collection
	 * object.
	 * @return a repository connection with internal interface.
	 */
	_RepositoryConnection getRepositoryConnection();
	
	/**
	 * Accessor to the shared metadata object for collection.
	 * @return a memory shared metadata object.
	 */
	CollectionMetadata getMetadata();
	
	/**
	 * Accessor to collection features object.
	 * @return a Configurable object with primitive type methods for convenient
	 * access.
	 */
	CollectionInfo getInfo();
	
	/**
	 * Returns an {@link org.xml.sax.XMLReader} object allowing reading
	 * documents stored in the collection.
	 * @return a Repository XMLReader implementation.
	 * @throws XMLDBCException in case of application error.
	 */
	RepositoryReader getReader() throws XMLDBCException;
	
	/**
	 * Internal method used to release a RepositoryReader to its creator in
	 * order it can be closed or cached.
	 * @param reader an object gained by {@link #getReader()}.
	 * @throws RepositoryException in case of application error.
	 */
	void releaseReader(RepositoryReader reader) throws RepositoryException;
	
	/**
	 * Send message to all storage objects to flush storage buffers.
	 * @throws XMLDBCException in case of application error.
	 */
	void flushFilers() throws XMLDBCException;
	
	/**
	 * Accessor to collection name with no XDBC exception returned.
	 * @return the collection name
	 */
	public String getCollectionName();
	
	/**
	 * Message sent when a commit has been performed at the connection level.
	 * @throws XMLDBCException
	 */
	public void commitPerformed() throws XMLDBCException;
	
	/**
	 * Method rollbackPerformed.
	 * @throws XMLDBCException in case of application.
	 */
	public void rollbackPerformed() throws XMLDBCException;
	
	/**
	 * Restore a document OID that could not be used eventually.
	 * @param did the document OID (DID)
	 */
	public void restoreDID(long did);
}
