/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

/*
 * CacheManager.java
 *
 * Created on 17 septembre 2002, 16:15
 */

package org.xquark.mapper.util;

import java.util.Collection;
import java.util.Iterator;

/**
 *  This cache manager mixes use counters and LRU
 *
 */
public class CacheManager
{
	private static final String RCSRevision = "$Revision: 1.2 $";
	private static final String RCSName = "$Name: XQ1_1_1 $";

    private CounterMap usedObjects = new CounterMap();
    private LRU cache;
    private int cacheSize;
    
    /** Creates a new instance of CacheManager */
    public CacheManager(int cacheSize)
    {
        this.cacheSize = cacheSize;
        cache = new LRU(cacheSize);
    }
    
    public Object checkOut(Object key)
    {
        Object o = usedObjects.checkOut(key);
        
        // look if it's not in the LRU
        if (o == null)
        {
            o = cache.get(key);
            // if so, move it from the cache to the used objects
            if (o != null)
                use(key, o);
        }
        return o;
    }
    
    /* Object is kept in memory till released */
    public void use(Object key, Object o)
    {
        usedObjects.put(key, o);

        int lruSize = cacheSize - usedObjects.size();
        if (lruSize >= 0)
            cache.setMaxSize(lruSize);
    }
    
    /** Object is released and possibly sent to LRU is counter is 0
     * @return the object that was removed from the cache if no space is left.
     */
    public Object release(Object key)
    {
        Object removed = usedObjects.remove(key);
        
        // put it in LRU if no "strong" use
        if (removed != null)
        {
            cache.setMaxSize(cacheSize - usedObjects.size());
            removed = cache.put(key, removed);
        }
        return removed;
    }
    
    /* puts directly in the LRU (no release expected) */
    public void put(Object key, Object o)
    {
        cache.put(key, o);
    }
    
    /* look 1st in the used objects and then in the cache */
    public Object get(Object key)
    {
        Object o = usedObjects.get(key);
        
        if (o == null)
            o = cache.get(key);
        return o;
    }
    
    public boolean contains(Object key)
    {
        return get(key) != null;
    }
    
    public int size()
    {
        return usedObjects.size() + cache.size();
    }
    
    public int getCacheSize()
    {
        return cacheSize;
    }
    
    public void clear()
    {
        usedObjects.clear();
        cache.clear();
    }
    
    public boolean used()
    {
        return usedObjects.size() > 0;
    }
    
    public boolean used(Object key)
    {
        return usedObjects.containsKey(key);
    }
    
    public void setKey(Object oldKey, Object newKey)
    {
        usedObjects.setKey(oldKey, newKey);
        cache.setKey(oldKey, newKey);
    }
    
    public Object remove(Object key)
    {
        Object removed = cache.removeObject(key);
        
        removed = usedObjects.removeAll(key);
        if (removed != null)
            cache.setMaxSize(cacheSize - usedObjects.size());
        return removed;
    }
    
    public Iterator iterator()
    {
        Collection objects = usedObjects.values();
        objects.addAll(cache.values());
        return objects.iterator();
    }
}
