/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

/*
 * CacheManager.java
 *
 * Created on 17 septembre 2002, 16:15
 */

package org.xquark.mapper.util;

import java.util.*;

/**
 *  Implements a LRU collection manager
 *
 */
public class LRU extends LinkedList
{
	private static final String RCSRevision = "$Revision: 1.2 $";
	private static final String RCSName = "$Name: XQ1_1_1 $";

    private int maxSize;
    private HashMap index;
    
    /** Creates a new instance of CacheManager */
    public LRU(int maxSize)
    {
        setMaxSize(maxSize);
        index = new HashMap();
    }
    
    public void setMaxSize(int maxSize)
    {
        while (size() > maxSize)  // max size reached: discard oldest
        {
            Node node = (Node)remove(0);
            index.remove(node.key);
        }
        this.maxSize = maxSize;
    }
    
    public int getMaxSize()
    {
        return maxSize;
    }
    
    /**
     * @return the potential removed object
     */
    public Object put(Object key, Object o)
    {
        Object discarded = null;
        if (maxSize > 0)
        {
            Node node = new Node(key, o);
            super.add(node);
            index.put(key, node);
            if (size() > maxSize)  // max size reached: discard oldest
            {
                node = (Node)remove(0);
                index.remove(node.key);
                discarded = node.o;
            }
        }
        else
            discarded = o;
        
        return discarded;
    }
    
    public Object get(Object key)
    {
        Node node = (Node)index.get(key);
        // move object to the end of the list (LRU)
        if (node != null)
        {
            add(remove(indexOf(node)));
            return node.o;
        }
        else
            return null;
    }
    
    public void clear()
    {
        super.clear();
        index.clear();
    }

    public synchronized void setKey(Object oldKey, Object newKey)
    {
        Node node = (Node)index.remove(oldKey);
        if (node != null)
        {
            node.key = newKey;
            index.put(newKey, node);
        }
    }
    
    public Object removeObject(Object key)
    {
        Node removed = (Node)index.remove(key);
        if (removed != null)
            super.remove(removed);
        return removed;
    }
    
    public Collection values()
    {
        ArrayList list = new ArrayList();
        Iterator it = super.iterator();
        while (it.hasNext())
        {
            list.add(((Node)it.next()).o);
        }
        return list;
    }
    
    private class Node
    {
        Object key;
        Object o;
        
        Node(Object key, Object o)
        {
            this.key = key;
            this.o = o;
        }
    }
}
