/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Odonata.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.mapping;
import java.sql.*;

/**
 * This generator is a built-in default implementation intended as a device
 * for relational primary and foreign keys generation. It uses the compilation
 * context to get the needed parameters, especially for initialization.
 * <b>WARNING !!!</b>
 * Currently, this generator is limited: it uses the target column provided by
 * the compilation context to initialize a counter (performing a 'SELECT
 * MAX(...) FROM ...' which is totally managed in memory (using 1 increments).
 * This has the advantage to be extremely rapid (no useless database access,
 * JDBC batch usable...) but the major drawback of this approach is the fact that
 * <b>concurrency in insertion is not supported</b>...
 */
public class KeyGenerator implements UserGenerator {

    private long counter = -1;
    private String table = null;
	private String column = null;
	boolean useQuotes = false;
    
    public KeyGenerator(CompilationContext properties) {
    	table = properties.getProperty(CompilationContext.TABLE_NAME);
    	column = properties.getProperty(CompilationContext.COLUMN_NAME);
    	useQuotes = Boolean.valueOf(
    		properties.getProperty(CompilationContext.USE_QUOTES_4_DDL)
    		).booleanValue();
    }

    public String getXMLType() {
        return "long";
    }

    public synchronized Object getValue(StorageContext context) {
    	if (counter < 0)
    		initCounter(context.getConnection());
    	return new Long(++counter);
    }
    
    private void initCounter(Connection connection) {
		ResultSet rs = null;
		Statement statement = null;
		try {
			StringBuffer buf = new StringBuffer("select max(");
			if (useQuotes)
				buf.append('"');
			buf.append(column);
			if (useQuotes)
				buf.append('"');
			buf.append(") from ");
			if (useQuotes)
				buf.append('"');
			buf.append(table);
			if (useQuotes)
				buf.append('"');
			
			statement = connection.createStatement();
			rs = statement.executeQuery(buf.toString());
			while (rs.next()) {
				counter = rs.getLong(1);
			}
		}
		catch (Exception ex) {
			ex.printStackTrace();
		}
		finally {
			try {
				if (rs != null)
					rs.close();
				if (statement != null)
					statement.close();
			}
			catch (Exception ex) {
				// no op
			}
		}
    }
}
