/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.schema;

import java.util.ArrayList;
import java.util.Collection;

/**
 * class for Attribute Group Definition.
 *
 * <p>This is the implementation of the Attribute Group Definition.
 * The following properties are defined:</p>
 * <p> 1. name : optional, null or a NCName, defined in SchemaComponent.</p>
 * <p> 2. target namespace : null or a namespace name, defined in SchemaComponent.</p>
 * <p> 3. attribue declarations : a set of attribute declarations. </p>
 * <p> 4. attribute wildcard : optional, either a wildcard or null. </p>
 *
 * <p>See XML Schema Part 1: Structures 3.6
 * W3C Recommendation 2 May 2001</p>
 *
 * @see org.xquark.xml.schema.SchemaComponent
 */
public final class AttributeGroupDefinition extends SchemaComponent
{
private static final String RCSRevision = "$Revision: 1.2 $"; 
private static final String RCSName     = "$Name: XQ1_1_1 $";
  
  private ArrayList contents = new ArrayList();
  private Wildcard attributeWildcard = null;

  /**
   * Create a new AttributeGroupDefinition.
   *
   * @param schema The schema which contains this attribute group declaration.
   * @param name The attribute group name, a NCName.
   */
  public AttributeGroupDefinition(Schema schema, String name) {
    super(schema, name);
  }

  public void accept(SchemaVisitor visitor) throws SchemaException {
    visitor.visit(this);
  }

  /**
   * Look up the attribute wildcard, if any
   *
   * @return The ElementDeclaration, or null if there is none.
   */
  public Wildcard getAttributeWildcard() {
    return attributeWildcard;
  }

  /**
   * Deprecated. This methode has been replaced by getAttributeDeclarations()
   * Look up all local attribute declarations.
   *
   * @return The collection view of all attribute Declarations.
   */  
  public Collection getContents() {
    return contents;
  }
  
  /**
   * Look up all local attribute declarations.
   *
   * @return The collection view of all attribute Declarations.
   */  
  public Collection getAttributeDeclarations() {
    return contents;
  }

  public void add(Object declaration) {
    contents.add(declaration);
  }
  
  public boolean attributeFind(Object decl) {
    if ( contents.indexOf(decl) == -1 ) return false;
    else return true;
  }

  public void setAttributeWildcard(Wildcard wildcard) {
    attributeWildcard = wildcard;
  }
}
