/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.schema;

import java.util.*;

import org.xquark.schema.datatypes.PrimitiveType;
import org.xquark.schema.loader.Loader;
import org.xquark.schema.validation.*;

/**
 * class for Complex Type Definition.
 *
 * <p>This is the implementation of the Complex Type Definition.
 * The following properties are defined:</p>
 * <p> 1. name : optional, null or a NCName, defined in SchemaComponent.</p>
 * <p> 2. target namespace : null or a namespace name, defined in SchemaComponent.</p>
 * <p> 3. base type defintion : either a simple type defintion or a complex type defintion, defined in Type.</p>
 * <p> 4. derivation method : either extension or restriction. </p>
 * <p> 5. final : a subset of {extension, restriction}. </p>
 * <p> 6. abstract : a boolean, defined in Type.</p>
 * <p> 7. attribue declarations : a set of attribute declarations contained in the complex type. </p>
 * <p> 8. content model : a content model to define the content type and to contain the particle. </p>
 * <p> 9. element declarations : a set of element declarations contained in the complex type. </p>
 *
 * <p>See XML Schema Part 1: Structures 3.4
 * W3C Recommendation 2 May 2001</p>
 *
 * @see org.xquark.xml.schema.SchemaComponent
 * @see org.xquark.xml.schema.Type
 * @see org.xquark.xml.schema.SimpleType
 */
public final class ComplexType extends Type implements SchemaScope {
    private static final String RCSRevision = "$Revision: 1.2 $";
    private static final String RCSName = "$Name: XQ1_1_1 $";

    private boolean extension = false;
    private boolean abs = false;
    private int block = 0;
    private ContentModel contentModel = null;
    private HashMap elementDeclarations = new HashMap();
    private HashMap attributes = new HashMap();
    private ArrayList requiredAndDefaultAttributes = null;
    private Wildcard attributeWildcard = null;
    private AttributeDeclaration idAttr = null;
    private AttributeDeclaration notationAttr = null;

    /**
     * Create a new ComplexType.
     *
     * @param schema The schema which contains this attribute declaration.
     * @param name The complex type name, null or a NCName.
     * @param baseType The base type definition, null or a simple type.
     */
    public ComplexType(Schema schema, String name, Type baseType) {
        super(schema, name, baseType);
    }

    public void accept(SchemaVisitor visitor) throws SchemaException {
        visitor.visit(this);
    }

    /**
     * Retrun the complex type's derivation method.
     *
     * @return The extention, true, if derivation method is an extension; 
     * false, if derivation method is a restriction
     */
    public boolean isExtension() {
        return extension;
    }

    /**
     * Retrun the complex type's derivation method.
     *
     * @return The restriction, true, if derivation method is an restriction; 
     * false, if derivation method is a extension
     */
    public boolean isRestriction() {
        return !extension;
    }

    public int getDerivationMethod() {
        return extension ? EXTENSION : RESTRICTION;
    }

    /**
     * @return null if valid, a string containing the error code otherwise
     */

    public String checkTypeDerivationOK(Type ancestor, int exclusions, boolean useBlock) {
        if (this == ancestor)
            return null;
        if ((getDerivationMethod() & exclusions) != 0)
            return "cos-ct-derived-ok.1";
        Type parent = getBaseType();
        if (parent == null)
            return "cos-ct-derived-ok.2.3.1";
        if (useBlock)
            exclusions |= parent.getBlock();
        return parent.checkTypeDerivationOK(ancestor, exclusions, useBlock);
    }

    public int getBlock() {
        return block;
    }

    public String normalizedValue(String value) {
        switch (getContentType()) {
            case EMPTY :
            case ELEMENT_ONLY :
                return null;
            case MIXED :
                return value;
            case TEXT_ONLY :
            default :
                return getValueType().normalizedValue(value);
        }
    }

    public StringBuffer normalizedValue(StringBuffer value) {
        switch (getContentType()) {
            case EMPTY :
            case ELEMENT_ONLY :
                return null;
            case MIXED :
                return value;
            case TEXT_ONLY :
            default :
                return getValueType().normalizedValue(value);
        }
    }

    public StringBuffer normalizedValue(StringBuffer value, StringBuffer out) {
        switch (getContentType()) {
            case EMPTY :
            case ELEMENT_ONLY :
                return null;
            case MIXED :
                if (out == null)
                    out = new StringBuffer();
                out.append(value.toString());
                return out;
            case TEXT_ONLY :
            default :
                return getValueType().normalizedValue(value, out);
        }
    }

    public String toXMLString(Object value, ValidationContextProvider context) {
        switch (getContentType()) {
            case EMPTY :
            case ELEMENT_ONLY :
            case MIXED :
                return value.toString();
            case TEXT_ONLY :
            default :
                return getValueType().toXMLString(value, context);
        }
    }

    public ValidationInfo validate(String value, boolean normalize, ValidationContextProvider context)
        throws SchemaException {
        switch (getContentType()) {
            case EMPTY :
                throw new SchemaException(
                    "cvc-complex-type.2.1",
                    "Empty element cannot contain characters or children");
            case ELEMENT_ONLY :
                boolean bWhitespaceOnly = true;
                for (int i = 0; i < value.length(); i++) {
                    if (!PrimitiveType.isXMLWhitespace(value.charAt(i)))
                        bWhitespaceOnly = false;
                }
                if (!bWhitespaceOnly)
                    throw new SchemaException(
                        "cvc-complex-type.2.3",
                        "Illegal non-whitespace characters in element-only content type");
                else
                    return null;
            case MIXED :
                return null;
            case TEXT_ONLY :
            default :
                return getValueType().validate(value, normalize, context);
        }
    }

    public ValidationInfo validate(StringBuffer value, boolean normalize, ValidationContextProvider context)
        throws SchemaException {
        switch (getContentType()) {
            case EMPTY :
                throw new SchemaException(
                    "cvc-complex-type.2.1",
                    "Empty element cannot contain characters or children");
            case ELEMENT_ONLY :
                boolean bWhitespaceOnly = true;
                for (int i = 0; i < value.length(); i++) {
                    if (!PrimitiveType.isXMLWhitespace(value.charAt(i)))
                        bWhitespaceOnly = false;
                }
                if (!bWhitespaceOnly)
                    throw new SchemaException(
                        "cvc-complex-type.2.3",
                        "Illegal non-whitespace characters in element-only content type");
                else
                    return null;
            case MIXED :
                return null;
            case TEXT_ONLY :
            default :
                return getValueType().validate(value, normalize, context);
        }
    }

    public Object convert(String value, boolean normalize, ValidationContextProvider context) throws SchemaException {
        switch (getContentType()) {
            case EMPTY :
                throw new SchemaException(
                    "cvc-complex-type.2.1",
                    "Empty element cannot contain characters or children");
            case ELEMENT_ONLY :
                boolean bWhitespaceOnly = true;
                for (int i = 0; i < value.length(); i++) {
                    if (!PrimitiveType.isXMLWhitespace(value.charAt(i)))
                        bWhitespaceOnly = false;
                }
                if (bWhitespaceOnly)
                    throw new SchemaException(
                        "cvc-complex-type.2.3",
                        "Empty element cannot contain characters or children");
                else
                    return "";
            case MIXED :
                return value;
            case TEXT_ONLY :
            default :
                return getValueType().convert(value, normalize, context);
        }
    }

    public Object convert(StringBuffer value, boolean normalize, ValidationContextProvider context)
        throws SchemaException {
        switch (getContentType()) {
            case EMPTY :
                throw new SchemaException(
                    "cvc-complex-type.2.1",
                    "Empty element cannot contain characters or children");
            case ELEMENT_ONLY :
                boolean bWhitespaceOnly = true;
                for (int i = 0; i < value.length(); i++) {
                    if (!PrimitiveType.isXMLWhitespace(value.charAt(i)))
                        bWhitespaceOnly = false;
                }
                if (bWhitespaceOnly)
                    throw new SchemaException(
                        "cvc-complex-type.2.3",
                        "Empty element cannot contain characters or children");
                else
                    return "";
            case MIXED :
                return value;
            case TEXT_ONLY :
            default :
                return getValueType().convert(value, normalize, context);
        }
    }

    public String toCanonicalForm(String value, ValidationContextProvider context) throws SchemaException {
        switch (getContentType()) {
            case EMPTY :
            case ELEMENT_ONLY :
            case MIXED :
                return value;
            case TEXT_ONLY :
            default :
                return getValueType().toCanonicalForm(value, context);
        }
    }

    public String validateDefault(String value, ValidationContextProvider context) throws SchemaException {
        switch (getContentType()) {
            case EMPTY :
            case ELEMENT_ONLY :
                throw new SchemaException("cos-valid-default.2.1", this);
            case MIXED :
                if (getContentModel().getModel() != null && !((Particle) getContentModel().getModel()).isEmptiable()) {
                    throw new SchemaException("cos-valid-default.2.2.2", this);
                }
                return value;
            case TEXT_ONLY :
            default :
                try {
                    return getValueType().toCanonicalForm(value, context);
                } catch (SchemaException ex) {
                    throw new SchemaException("cos-valid-default.2.2.1", this, ex);
                }
        }
    }

    /**
     * Return the complex type's content model, if any
     * the model of a content model can be either a particle or a simple type definition
     *
     * @return The content model, or null if there is none
     * @see org.xquark.xml.schema.ContentModel
     */
    public ContentModel getContentModel() {
        return contentModel;
    }

    /**
     * Return the type's abstract, the notion 'abstract' is applied only to 
     * complex type definition
     *
     * @return The abstract, ture if type is abstract, false otherwise.
     */
    public boolean isAbstract() {
        return abs;
    }

    /**
     * Return the complex type's content type
     * the content type of a complex type can be : EMPTY, MIXED, TEXT_ONLY or ELEMENT_ONLY
     *
     * @return The content type
     * @see org.xquark.xml.schema.ContentModel
     */
    public int getContentType() {
        return contentModel.getContentType();
    }

    public SimpleType getValueType() {
        if (getContentType() == TEXT_ONLY)
            return (SimpleType) contentModel.getModel();
        else
            return null;
    }

    /**
     * Look up the attribute wildcard, if any
     *
     * @return The ElementDeclaration, or null if there is none.
     */
    public Wildcard getAttributeWildcard() {
        return attributeWildcard;
    }

    public AttributeDeclaration getIDAttribute() {
        return idAttr;
    }

    public AttributeDeclaration getNotationAttribute() {
        return notationAttr;
    }

    public ContentIterator childIterator() {
        switch (getContentType()) {
            case ELEMENT_ONLY :
            case MIXED :
                Particle particle = (Particle) contentModel.getModel();
                if (particle == null)
                    return null;
                return ContentIteratorFactory.createIterator(particle);
            default :
                return null;
        }
    }

    public ContentIterator contentIterator() {
        switch (getContentType()) {
            case ELEMENT_ONLY :
            case MIXED :
                return new ElementTreeContentIterator(this);
            default :
                return null;
        }
    }

    public boolean isGlobalScope() {
        return false;
    }

    /**
     * Giving an element declaration name, look up the local element declaration, if any
     *
     * @param name The element name.
     *
     * @return The ElementDeclaration, or null if there is none.
     */
    public ElementDeclaration getElementDeclaration(String namespace, String name) {
        if (namespace == null)
            return (ElementDeclaration) elementDeclarations.get(name);
        else
            return (ElementDeclaration) elementDeclarations.get("{" + namespace + "}" + name);
    }

    /**
     * Look up all local element declarations, if any
     *
     * @return The collection view of all Element Declarations, or null if there is none.
     */
    public Collection getElementDeclarations() {
        return elementDeclarations.values();
    }

    /**
     * Giving an attribute declaration name, look up the local attribute declaration.
     *
     * @param name The attribute name.
     *
     * @return The AttributeDeclaration.
     */
    public AttributeDeclaration getAttributeDeclaration(String namespace, String name) {
        if (namespace == null)
            return (AttributeDeclaration) attributes.get(name);
        else
            return (AttributeDeclaration) attributes.get("{" + namespace + "}" + name);
    }

    /**
     * Look up all local attribute declarations.
     *
     * @return The collection view of all attribute Declarations.
     */
    public Collection getAttributeDeclarations() {
        return attributes.values();
    }

    Set getAttributeEntries() {
        return attributes.entrySet();
    }

    public List getRequiredAndDefaultAttributes() {
        return requiredAndDefaultAttributes;
    }

    public void setExtension(boolean extension) {
        this.extension = extension;
    }

    public void setAbstract(boolean abs) {
        this.abs = abs;
    }

    public void setBlock(int block) {
        this.block = block;
    }

    public void setContentModel(ContentModel contentModel) {
        this.contentModel = contentModel;
    }

    void setIDAttribute(AttributeDeclaration decl) {
        idAttr = decl;
    }

    void setNotationAttribute(AttributeDeclaration decl) {
        notationAttr = decl;
    }

    public int[] getDeclarationOccurrence(String namespace, String name) {
        if (getContentType() == EMPTY || getContentType() == TEXT_ONLY || getContentModel().getModel() == null)
            return null;
        ComplexType.OccurrenceVisitor visitor = new ComplexType.OccurrenceVisitor();
        return visitor.getOccurrenceCount((Particle)getContentModel().getModel(), namespace, name);
    }

    public void register(SchemaComponent decl) throws SchemaException {
        if (decl instanceof ElementDeclaration)
            register((ElementDeclaration) decl);
        else if (decl instanceof AttributeDeclaration)
            register((AttributeDeclaration) decl);
        else {
            // AttributeGroups are stored with themselves as key
            // until initialization replaces them by their content
            attributes.put(decl, decl);
        }
    }

    private Object buildDeclarationKey(SchemaComponent comp) {
        if (comp.getNamespace() == null)
            return comp.getName();
        else
            return "{" + comp.getNamespace() + "}" + comp.getName();
    }

    public void register(AttributeDeclaration decl) throws SchemaException {
        Object nsName = buildDeclarationKey(decl);
        if (attributes.put(nsName, decl) != null) {
            //    "Two distinct attribute declarations should not have identical names : " + nsName;
            throw new SchemaException("ct-props-correct.4", decl);
        }
    }

    public void register(ElementDeclaration decl) throws SchemaException {
        Object nsName = buildDeclarationKey(decl);
        ElementDeclaration prev = (ElementDeclaration) elementDeclarations.put(nsName, decl);
        // The spec requires consistent declarations to have the same named type
        // This is not respected in the Schema for Schemas
        // if (prev != null && !decl.getType().hasName(prev.getType().getNamespace(), prev.getType().getName())) {
        if (prev != null && decl.getType() != prev.getType()) {
            //    "Element declaration '" + nsName + "' must have the same type as the existing element declaration";
            throw new SchemaException("cos-element-consistent.2", decl);
        }
        List substitutionList = decl.getSubstitutionElementList();
        if (substitutionList != null) {
            Iterator it = substitutionList.iterator();
            while (it.hasNext()) {
                ElementDeclaration d = (ElementDeclaration) it.next();
                ElementDeclaration existing = getElementDeclaration(d.getNamespace(), d.getName());
                // if (existing != null && !d.getType().hasName(existing.getType().getNamespace(), existing.getType().getName())) {
                if (existing != null && d.getType() != existing.getType()) {
                    //        "Element declaration '" + nsName + "' must have the same type as the existing element declaration";
                    throw new SchemaException("cos-element-consistent.2", d);
                }
            }
        }
    }

    public void setAttributeWildcard(Wildcard wildcard) {
        attributeWildcard = wildcard;
    }

    public void setRequiredAndDefaultAttributes(ArrayList list) {
        requiredAndDefaultAttributes = list;
    }

    // TYPING STUFF
    public boolean isSimpleType() {
        return false;
    }

    void checkUniqueParticleAttribution() throws SchemaException {
        if (this.getContentType() == EMPTY)
            return;
        if (this.getContentModel().getModel() instanceof SimpleType)
            return;
        Particle p = (Particle) this.getContentModel().getModel();
        if (p == null)
            return;
        Particle overLapParticle = p.uniqueParticleAttribution();
        if (overLapParticle == null)
            return;
        // Unique Particle Attribute Violation
        throw new SchemaException("cos-nonambig", overLapParticle);
    }

    void checkDerivationValidRestriction(Loader loader) throws SchemaException {
        Type baseType = this.getBaseType();
        if (this.isExtension() || baseType == null || baseType.isSimpleType())
            return; // it's not restriction
        ComplexType base = (ComplexType) baseType;

        if ((base.getFinal() & RESTRICTION) != 0) {
            //    "Base type final must not contain restriction";
            throw new SchemaException("derivation-ok-restriction.1", this);
        }

        Iterator it = this.getAttributeDeclarations().iterator();
        while (it.hasNext()) {
            AttributeDeclaration decl = (AttributeDeclaration) it.next();
            AttributeDeclaration baseDecl = base.getAttributeDeclaration(decl.getNamespace(), decl.getName());
            if (baseDecl != null) {
                if (!(baseDecl.getUse() != REQUIRED || decl.getUse() == REQUIRED)) {
                    //        "Attribute <use> isn't reqired, but attribute <use> in base type is required.";
                    throw new SchemaException("derivation-ok-restriction.2.1.1", decl);
                }
                String errCode02 =
                    decl.getType().checkTypeDerivationOK(baseDecl.getType(), baseDecl.getType().getFinal(), false);
                if (errCode02 != null) {
                    //        "Attribute's type definition must be validly derivded from that of the base type's attribute declaration.";
                    throw new SchemaException(
                        "derivation-ok-restriction.2.1.2",
                        decl,
                        new SchemaException(errCode02, decl));
                }
                if (!(baseDecl.getFixedValue() == null || baseDecl.getFixedValue().equals(decl.getFixedValue()))) {
                    //        "Attribute declaration's effective value constraint doesn't match with that of base type's attribute declaration.";
                    throw new SchemaException("derivation-ok-restriction.2.1.3", decl);
                }
            } else if (
                base.getAttributeWildcard() == null
                    || base.getAttributeWildcard().isAllowed(decl.getNamespace()) == false) {
                //      "Attribute declaration, or a valid attribute wildcard, must be defined in base type.";
                throw new SchemaException("derivation-ok-restriction.2.2", decl);
            }
        }

        // schema of schema doesn't respect the following rule
        it = base.getAttributeDeclarations().iterator();
        while (it.hasNext()) {
            AttributeDeclaration baseDecl = (AttributeDeclaration) it.next();
            if (baseDecl.getUse() == REQUIRED) {
                AttributeDeclaration decl = this.getAttributeDeclaration(baseDecl.getNamespace(), baseDecl.getName());
                if (decl == null || decl.getUse() != REQUIRED) {
                    //        "Attribute declaration's attribute <use> must be required.";
                    Object obj = decl;
                    if (obj == null)
                        obj = this;
                    throw new SchemaException("derivation-ok-restriction.3", obj);
                }
            }
        }

        Wildcard w = this.getAttributeWildcard();
        if (w != null) {
            Wildcard baseW = base.getAttributeWildcard();
            if (baseW == null) {
                //      "Base type must have an attribute wildcard.";
                throw new SchemaException("derivation-ok-restriction.4.1", this);
            } else {
                String errCode02 = w.validWildcardSubset(baseW);
                if (errCode02 != null) {
                    //        "Attribute wildcard must be a valid subset of base type's.";
                    throw new SchemaException("derivation-ok-restriction.4.2", this, new SchemaException(errCode02, w));
                }
            }
        }

        switch (this.getContentType()) {
            case TEXT_ONLY :
                switch (base.getContentType()) {
                    case EMPTY :
                    case ELEMENT_ONLY :
                        //          "Content type of base type must be mixed or a valid simple type";
                        throw new SchemaException("derivation-ok-restriction.5.1", this);
                    case TEXT_ONLY :
                        SimpleType st = (SimpleType) this.getContentModel().getModel();
                        SimpleType stBase = (SimpleType) base.getContentModel().getModel();
                        String errCode02 = st.checkTypeDerivationOK(stBase, stBase.getFinal(), false);
                        if (errCode02 != null) {
                            //            "Base {content type} must be a valid simple type.";
                            throw new SchemaException(
                                "derivation-ok-restriction.5.1.1",
                                this,
                                new SchemaException(errCode02, st));
                        }
                        break;
                    case MIXED :
                        Particle p = (Particle) base.getContentModel().getModel();
                        if (p == null || p.isEmptiable() == false) {
                            //            "Base type must have an emptiable particle.";
                            throw new SchemaException("derivation-ok-restriction.5.1.2", this);
                        }
                        break;
                }
                break;

            case EMPTY :
                switch (base.getContentType()) {
                    case EMPTY :
                        break;
                    case ELEMENT_ONLY :
                    case MIXED :
                        Particle p = (Particle) base.getContentModel().getModel();
                        if (p == null || p.isEmptiable() == false) {
                            //            "Base {content type} must be elementOnly or mixed, and must have an emptiable particle.";
                            throw new SchemaException("derivation-ok-restriction.5.2.2", this);
                        }
                        break;
                    case TEXT_ONLY :
                        //          "Base {content type} must not be textOnly";
                        throw new SchemaException("derivation-ok-restriction.5.2", this);
                }
                break;

            case ELEMENT_ONLY :
                switch (base.getContentType()) {
                    case EMPTY :
                    case TEXT_ONLY :
                        // "Base {content type} must be elementOnly or mixed, 
                        // and the particle must be a valid restrion of base type's particle";
                        throw new SchemaException("derivation-ok-restriction.5.3", this);
                    case ELEMENT_ONLY :
                    case MIXED :
                        Particle p = null;
                        try {
                            p = (Particle) this.getContentModel().getModel();
                            Particle pBase = (Particle) base.getContentModel().getModel();
                            if (p != null)
                                p.checkParticleValidRestriction(pBase);
                        } catch (SchemaException se) {
                            if (se.getInvalidObject() != null
                                && loader != null
                                && loader.getComponentLocaltor(se.getInvalidObject()) != null)
                                throw new SchemaException("derivation-ok-restriction.5.3", se.getInvalidObject(), se);
                            else
                                throw new SchemaException("derivation-ok-restriction.5.3", p, se);
                        }
                        break;
                }
                break;

            case MIXED :
                switch (base.getContentType()) {
                    case EMPTY :
                    case TEXT_ONLY :
                    case ELEMENT_ONLY :
                        //          "Base {content type} must be elementOnly or mixed";
                        throw new SchemaException("derivation-ok-restriction.5.3", this);
                    case MIXED :
                        Particle p = null;
                        try {
                            p = (Particle) this.getContentModel().getModel();
                            Particle pBase = (Particle) base.getContentModel().getModel();
                            if (p != null)
                                p.checkParticleValidRestriction(pBase);
                        } catch (SchemaException se) {
                            throw new SchemaException("derivation-ok-restriction.5.3", p, se);
                        }
                        break;
                }
                break;
        }
    }

    static class OccurrenceVisitor extends DefaultSchemaVisitor {
        private static final String RCSRevision = "$Revision: 1.2 $";
        private static final String RCSName = "$Name: XQ1_1_1 $";
        private int[] occurs = null;
        private String namespace = null;
        private String name = null;

        public int[] getOccurrenceCount(Particle particle, String namespace, String name) {
            occurs = new int[2];
            occurs[0] = occurs[1] = 0;
            this.namespace = namespace;
            this.name = name;
            try {
                particle.accept(this);
            } catch (SchemaException ex) {
                // Should not occur
            }
            return occurs;
        }

        private void setValues(long min, long max) {
            if (max > Integer.MAX_VALUE)
                max = Integer.MAX_VALUE;
            occurs[0] = (int) min;
            occurs[1] = (int) max;
        }

        public void visit(Particle particle) throws SchemaException {
            ((SchemaVisitable) particle.getTerm()).accept(this);
            long min = occurs[0] * particle.getMinOccurs();
            long max = occurs[1] * particle.getMaxOccurs();
            setValues(min, max);
        }

        public void visit(ModelGroup group) throws SchemaException {
            long min = 0;
            long max = 0;
            Iterator it = group.iterator();
            while (it.hasNext()) {
                occurs[0] = occurs[1] = 0;
                ((SchemaVisitable) it.next()).accept(this);
                min += occurs[0];
                max += occurs[1];
            }
            setValues(min, max);
        }

        public void visit(ChoiceModelGroup group) throws SchemaException {
            long min = Integer.MAX_VALUE;
            long max = 0;
            Iterator it = group.iterator();
            while (it.hasNext()) {
                occurs[0] = occurs[1] = 0;
                ((SchemaVisitable) it.next()).accept(this);
                if (occurs[0] < min)
                    min = occurs[0];
                if (occurs[1] > max)
                    max = occurs[1];
            }
            setValues(min, max);
        }

        public void visit(ElementDeclaration decl) throws SchemaException {
            if (decl.hasName(namespace, name) || decl.getSubstitutionElement(namespace, name) != null)
                occurs[0] = occurs[1] = 1;
        }

        public void visit(Wildcard wildcard) throws SchemaException {
            if (wildcard.isAllowed(namespace)) {
                occurs[0] = occurs[1] = 1;
            }
        }

    }

}
