/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.schema;

import java.util.*;

/**
 * class for Element Declaration.
 *
 * <p>This is the implementation of the Element Declaration.
 * The following properties are defined:</p>
 * <p> 1. name : a NCName, defined in SchemaComponent.</p>
 * <p> 2. target namespace : null or a namespace name, defined in SchemaComponent.</p>
 * <p> 3. type defintion : a simple or complex type defintion, defined in Declaration.</p>
 * <p> 4. scope : either a Schema (means gloabal) or a complex type definition, defined in Declaration. </p>
 * <p> 5. value constraint : optional. Either a defaultValue or a fixedValue, defined in Declaration. </p>
 * <p> 6. nillable : a boolean, true if the element declaration is nillable. </p>
 * <p> 7. identity-constraint definitions : optional. A set of constraint definitions. </p>
 * <p> 8. substitution group affiliation : optional. A top level element declaration. </p>
 *
 * <p>See XML Schema Part 1: Structures 3.3
 * W3C Recommendation 2 May 2001</p>
 *
 * @see org.xquark.xml.schema.SchemaComponent
 * @see org.xquark.xml.schema.Declaration
 * @see org.xquark.xml.schema.IdentityConstraint
 */
public class ElementDeclaration extends Declaration {
private static final String RCSRevision = "$Revision: 1.2 $"; 
private static final String RCSName     = "$Name: XQ1_1_1 $";
  
  private boolean nillable = false;
  private boolean abs = false;
  private int block = 0;
  private int fin = 0;
  private SchemaComponent substitutionGroup = null;
  private ArrayList substitutionElementList = null;
  private HashMap identityConstraints = null;
  
  /**
   * Create a new ElementDeclaration.
   *
   * @param schema The schema which contains this element declaration.
   * @param name The element declaration name, a NCName.
   * @param scope The scope, null if the element declaration belongs to a group
   * which contains this element declaration.
   */
  public ElementDeclaration(Schema schema, String name, SchemaScope scope) {
    super(schema, name, scope);
  }
  
  public void accept(SchemaVisitor visitor) throws SchemaException {
    visitor.visit(this);
  }

  public boolean isAbstract() {
    return abs;
  }
  
  /**
   * Return true if the element declaration is nillable, false, if not
   *
   * @return The nillable property
   */
  public boolean isNillable() { 
    return nillable;
  }

  public int getBlock() {
    return block;
  }

  public int getFinal() {
    return fin;
  }

  /**
   * Return substitution group affiliation, if any
   *
   * @return The substitution group affiliation property - a top level element
   * declaration, or null if there is none
   */
  public SchemaComponent getSubstitutionGroup() {
    return substitutionGroup;
  }
	
  public List getSubstitutionElementList() {
    //@@Scope 
//      if (scope.isGlobalScope()) return substitutionElementList;
//      else {
//        ElementDeclaration anElement = getSchema().getElementDeclaration(getName());
//        if (anElement == null) return null;
//        return anElement.getSubstitutionElementList();
//      }
    if (scope.isGlobalScope()) return substitutionElementList;
    else return null;
  }
	
  public void setAbstract(boolean abs) {
    this.abs = abs;
  }

  public void setNillable(boolean nillable) {
    this.nillable = nillable;
  }

  public void setBlock(int block) {
    this.block = block;
  }

  public void setFinal(int fin) {
    this.fin = fin;
  }

  public void setSubstitutionGroup(SchemaComponent substitutionGroup) {
    this.substitutionGroup = substitutionGroup;
  }

  void addSubstitutionMember(ElementDeclaration anElement) {
    if (substitutionElementList == null)
      substitutionElementList = new ArrayList();
    substitutionElementList.add(anElement); 
  }

  public ElementDeclaration getSubstitutionElement(String namespace, String localName)
  {
    List list = getSubstitutionElementList();
    if (list == null) return null;
    Iterator it = list.iterator();
    while (it.hasNext()) {
      ElementDeclaration anElement = (ElementDeclaration)it.next();
      // find localName in this element
      if (anElement.hasName(namespace, localName)) return anElement;
    }
		
    return null;
  }
  
  public void fillAllConcreteSubstitutionElements(List result) {
    List list = getSubstitutionElementList();
    if ( list == null ) return;
    Iterator it = list.iterator();
    while (it.hasNext()) {
      ElementDeclaration anElement = (ElementDeclaration)it.next();
      if (!anElement.isAbstract()) result.add(anElement);
    }	
  }
  
  // identity-constraint
  public void register(IdentityConstraint identityConstraint)  {
    if ( identityConstraints == null ) identityConstraints = new HashMap();
    identityConstraints.put(identityConstraint.getName(), identityConstraint);
  }
  
  /**
   * Look up all identity-constraint definitions.
   *
   * @return A collection view of all identity-constraint definitions
   */
  public Collection getIdentityConstraints() {
    if ( identityConstraints == null ) return null;
    return identityConstraints.values();
  }
  
  public IdentityConstraint getIdentityConstraint(String name) {
    if ( identityConstraints == null ) return null;
    if ( name.indexOf(':') != -1 ) {
      name = name.substring(name.indexOf(':') + 1);
    }
    IdentityConstraint result = (IdentityConstraint)identityConstraints.get(name);
    return result;
  }
  
  boolean equals(ElementDeclaration decl) {
    if ( this == decl ) return true;
    if ( decl == null ) return false;
    if ( this.getName().equals(decl.getName()) == false ) return false;
    if ( this.getNamespace() == null && decl.getNamespace() != null ) return false;
    if ( this.getNamespace() != null && !(this.getNamespace().equals(decl.getNamespace())) )
      return false;
    if ( this.abs != decl.abs ) return false;
    if ( this.nillable != decl.nillable ) return false;
    if ( this.fin != decl.fin ) return false;
    if ( this.block != decl.block ) return false;
    if ( this.defaultValue == null && decl.defaultValue != null ) return false;
    if ( this.defaultValue != null && !(this.defaultValue.equals(decl.defaultValue)) ) return false;
    if ( this.fixedValue == null && decl.fixedValue != null ) return false;
    if ( this.fixedValue != null && !(this.fixedValue.equals(decl.fixedValue)) ) return false;
    if ( this.getType() == decl.getType() ) return true;
    else return false;
  }
  
  boolean overlapUPAC(Object obj) {
    if ( obj == null ) return false;
    if ( obj instanceof ElementDeclaration )
      return overlapUPAC((ElementDeclaration)obj);
    else if ( obj instanceof Wildcard ) 
       return overlapUPAC((Wildcard)obj);
    else
      return false;
 }
  
  boolean overlapUPAC(ElementDeclaration decl) {
    if ( this == decl ) return true;
    if ( decl == null ) return false;
    
    if ( this.getName().equals(decl.getName()) &&
         ( ( this.getNamespace() == null && decl.getNamespace() == null ) ||
           ( this.getNamespace() != null && this.getNamespace().equals(decl.getNamespace())) ) )
      return true;
    
    // substitution group
    if ( this.substitutionGroup == decl || this == decl.substitutionGroup ) return true;
    
    return false;
  }
  
  boolean overlapUPAC(Wildcard wild) {
    if ( wild == null ) return false;

    if ( wild.isAllowed(this.getNamespace()) ) return true;
    
    // perhaps, there is a problem, here
    // overlapUPAC is called when ComplexType is initialized,
    // then ElementDeclaration isn't already initialized,
    // substitution group isn't initialized
    // Solution : 3rd pass of initialization is necessary ?
    // noted  Jan. 10, 2001
    if ( this.getSubstitutionElementList() != null ) {
      java.util.Iterator it = this.getSubstitutionElementList().iterator();
      while ( it.hasNext() ) {
        ElementDeclaration subElement = (ElementDeclaration)it.next();
        if ( wild.isAllowed(subElement.getNamespace()) ) return true;
      }
    }
     
    return false;
  }
  
}

