/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.schema;

import java.util.Map;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.SAXException;
import org.xquark.schema.validation.ContentIterator;
import org.xquark.schema.validation.ContentIteratorFactory;
import org.xquark.util.DOMBuilder;

public class ElementGenerator {
private static final String RCSRevision = "$Revision: 1.2 $"; 
private static final String RCSName     = "$Name: XQ1_1_1 $";
  
  private org.xml.sax.ContentHandler handler = null;
  private java.util.HashMap prefixMap;
  private SchemaManager manager;
  
  public ElementGenerator(SchemaManager manager, java.util.HashMap prefixMap) {
    this.manager = manager;
    if (prefixMap != null) {
      this.prefixMap = (java.util.HashMap)prefixMap.clone();
      if (prefixMap.get("xsi") == null)
	      this.prefixMap.put("xsi", SchemaConstants.XSI_URI);
    }
  }

  public void setContentHandler(org.xml.sax.ContentHandler handler) {
    this.handler = handler;
  }
  
  public org.xml.sax.ContentHandler getContentHandler() { 
    return handler;
  }
  
  public void generateDocument(String namespace, String localName, Particle particle) 
	throws SAXException {
    ElementDeclaration decl = 
      manager.getElementDeclaration(namespace, localName);
    if (decl == null) return;
    generateDocument(decl, particle);
  }
  
  public void generateDocument(ElementDeclaration decl, Particle particle) {
    if (handler == null) return;
    try {
      handler.startDocument();
      if (prefixMap != null) {
        java.util.Iterator it = prefixMap.entrySet().iterator();
        while (it.hasNext()) {
          Map.Entry entry = (Map.Entry)it.next();
          handler.startPrefixMapping((String)entry.getKey(),(String)entry.getValue());
        }
      }
      generate(decl, particle);
      if (prefixMap != null) {
        java.util.Iterator it = prefixMap.entrySet().iterator();
        while (it.hasNext()) {
          Map.Entry entry = (Map.Entry)it.next();
          handler.endPrefixMapping((String)entry.getKey());
        }
      }
      handler.endDocument();
    } catch (SAXException ex) { ex.printStackTrace(); }
  }

  public Document generateDocument(Document doc, String namespace, String localName)
	throws SAXException
  {
    handler = new DOMBuilder(doc);
    generateDocument(namespace, localName, null);
    return doc;
  }

  public Document generateDocument(Document doc, ElementDeclaration decl) {
    handler = new DOMBuilder(doc);
    generateDocument(decl, null);
    return doc;
  }

  public Element generateElement(Document doc, ElementDeclaration decl, boolean stripNamespaces)
  {
    handler = new DOMBuilder(doc);
    Element previous = doc.getDocumentElement();
    doc.removeChild(previous);
    generateDocument(decl, null);
    Element node = doc.getDocumentElement();
    doc.removeChild(node);
    doc.appendChild(previous);
    if (stripNamespaces) {
      org.w3c.dom.NamedNodeMap atts = node.getAttributes();
      for (int i = 0; i < atts.getLength(); i++) {
        org.w3c.dom.Attr attr = (org.w3c.dom.Attr)atts.item(i);
        if ("xmlns".equals(attr.getPrefix()) || "xsi".equals(attr.getPrefix())
            || "xmlns".equals(attr.getNodeName()))
          node.removeAttributeNode(attr);
      }
    }
    return node;
  }
  
  public org.w3c.dom.NodeList generateElement(Document doc, ElementDeclaration decl, Particle parentParticle)
  {    
    try {
      handler = new DOMBuilder(doc);
      Element previous = doc.getDocumentElement();
      doc.removeChild(previous);
      generateDocument(decl, parentParticle);
      Element node = doc.getDocumentElement();
      doc.removeChild(node);
      doc.appendChild(previous);
      return node.getChildNodes();
    }
    catch ( Exception e ) {
      e.printStackTrace();
    }
    return null;
  }
  
  private void generate(ElementDeclaration decl) {
    generate(decl.getNamespace(), decl.getName(), decl.getType(), decl.isNillable());
  }
  
  private void generate(String namespace, String localName, Type type) {
    generate(namespace, localName, type, false);
  }

  private void generate(String namespace, String localName, Type type, boolean nullable) 
  {
    try {
      if (namespace == null) namespace = "";
      org.xml.sax.helpers.AttributesImpl atts = new org.xml.sax.helpers.AttributesImpl();
      java.util.Collection attrDecls = type.getAttributeDeclarations();
      if (attrDecls != null) {
	java.util.Iterator it = attrDecls.iterator();
	while (it.hasNext()) {
	  AttributeDeclaration decl = (AttributeDeclaration)it.next();
          if (decl.getUse() == SchemaConstants.REQUIRED) {
            String attNamespace = decl.getNamespace();
            if (attNamespace == null) attNamespace = "";
            String value = decl.getDefaultValue();
            if (value == null) value = "";
            atts.addAttribute(attNamespace, decl.getName(), null, "CDATA", value);
          }
	}
      }
      
      ContentIterator iterator = type.childIterator();
      if (iterator == null) {
	if (nullable) {
	  atts.addAttribute(SchemaConstants.XSI_URI, "nil", null, 
			    "CDATA", "true");
	  handler.startElement(namespace, localName, null, atts);
	  handler.endElement(namespace, localName, null);
	} else {
	  handler.startElement(namespace, localName, null, atts);
	  handler.characters(new char[0], 0, 0);
	  handler.endElement(namespace, localName, null);
	}
      } else {
	handler.startElement(namespace, localName, null, atts);
	ElementDeclaration next = null;
	ElementDeclaration prev = null;
	int count = 0;
	while (true) {
	  java.util.List nextElements = iterator.nextValidElements();
          if (nextElements.size() > 0)
            next = (ElementDeclaration)nextElements.get(nextElements.size()-1);
          /*
	  Iterator it = nextElements.iterator();
          while (it.hasNext()) {
              next = (ElementDeclaration)it.next();
              if (next != null) {
                int[] occurrenceCount = 
                  ((ComplexType)type).getDeclarationOccurrence(next.getNamespace(), next.getName());
                if (occurrenceCount != null && occurrenceCount[0] > 0) break;
              }
             count++;
             next = null;
          }
	  if (prev != next) {
	    count++;
	    prev = next;
	  } else if (nextElements.size() > count) {
	    next = (ElementDeclaration)nextElements.get(count);
	    count++;
	  }
           */
	  if (next == null) break;
	  iterator.startElement(next.getNamespace(), next.getName());
	  generate(next);
	  iterator.endElement(next.getNamespace(), next.getName());
	}	
	handler.endElement(namespace, localName, null);
      }
    } catch (SchemaException ex) {
      ex.printStackTrace();
    } catch (SAXException ex) {
      ex.printStackTrace();
    }
  }
  
  private void generate(ElementDeclaration decl, Particle parentParticle) 
  {
    if ( parentParticle == null )
      generate(decl);
    else {
      try {     
        // create a root element
        handler.startElement("", "rootElement", null, new org.xml.sax.helpers.AttributesImpl());
        
        ContentIterator iterator = ContentIteratorFactory.createIterator(parentParticle);
	ElementDeclaration next = decl;
	while (true) {          
	  iterator.startElement(next.getNamespace(), next.getName());
	  generate(next);
	  iterator.endElement(next.getNamespace(), next.getName());
          
	  java.util.List nextElements = iterator.nextValidElements();
          if ( nextElements != null && nextElements.size() > 0)
              next = (ElementDeclaration)nextElements.get(nextElements.size()-1);
          else
            next = null;
          
	  if (next == null) break;
	}	
        
        handler.endElement("", "rootElement", null);
      } 
      catch ( Exception ex ) {
        ex.printStackTrace();
      }
    }
  }
  
}
  
  
