/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.schema;

import java.util.*;

import org.xquark.schema.datatypes.PrimitiveType;
import org.xquark.schema.loader.*;
import org.xquark.schema.validation.ValidationContextProvider;

public class InitializationVisitor extends DefaultSchemaVisitor implements SchemaConstants {
    private static final String RCSRevision = "$Revision: 1.6 $";
    private static final String RCSName = "$Name: XQ1_1_1 $";
    private ReferenceResolver refResolver = new ReferenceResolver();
    private HashSet initialized = new HashSet();
    private Schema schema;
    private ArrayList pendingTypes = null;
    private Map contextProviders = null;
    private Loader loader = null;

    public InitializationVisitor(Schema schema) {
        this.schema = schema;
    }

    public InitializationVisitor(Schema schema, Map contextProviders, Loader loader) {
        this.schema = schema;
        this.contextProviders = contextProviders;
        this.loader = loader;
    }

    public void visit(AttributeGroupDefinition attG) throws SchemaException {
        if (attG.getSchema() != schema || initialized.contains(attG))
            return;
        initialized.add(attG);

        Iterator it = attG.getAttributeDeclarations().iterator();
        while (it.hasNext()) {
            SchemaComponent comp = (SchemaComponent) it.next();
            refResolver.resolve(comp);
            comp.accept(this);
        }
    }

    public void visit(ElementDeclaration decl) throws SchemaException {
        if (initialized.contains(decl))
            return;
        //      if (decl.getSchema() != schema || initialized.contains(decl)) return;
        initialized.add(decl);

        //System.err.println("Init element : " + decl.getName());

        // if type isn't null, type is resolved by itself
        if (decl.getType() != null) {
            initType(decl);
            // it's an error for NOTATION to be used directly in a schema
            if (decl.getType().isSimpleType()
                && decl.getType().isDerivedFrom(decl.getManager().getType(XMLSCHEMA_URI, "NOTATION"))) {
                throw new SchemaException("enumeration-required-notation", decl);
            }
        }

        // handle substitutionGroup (inheriting its type if required)
        ElementDeclaration substitutionElement = null;
        if (decl.getSubstitutionGroup() != null) {
            substitutionElement = (ElementDeclaration) refResolver.resolve(decl.getSubstitutionGroup());
            substitutionElement.accept(this);
            if ((substitutionElement.getBlock() & SUBSTITUTION) != 0) {
                throw new SchemaException("cos-equiv-derived-ok-rec.1", decl);
            }
            if (decl.getType() == null) {
                decl.setType(substitutionElement.getType());
            } else {
                String errCode =
                    decl.getType().checkTypeDerivationOK(
                        substitutionElement.getType(),
                        substitutionElement.getFinal(),
                        false);
                if (errCode != null) {
                    //          "Type is not validly derived from substitution element type";
                    throw new SchemaException("e-props-correct.3", decl);
                }
            }
            String errCode =
                decl.getType().checkTypeDerivationOK(
                    substitutionElement.getType(),
                    substitutionElement.getBlock(),
                    true);
            if (errCode != null) {
                //        "Declaration is not substitutable to substitution element";
                throw new SchemaException("cos-equiv-derived-ok-rec.3", decl);
            }
            decl.setSubstitutionGroup(substitutionElement);
            substitutionElement.addSubstitutionMember(decl);
            propagateSubstitution(substitutionElement, decl);
        }
        // if type still null, use ur-type definition
        if (decl.getType() == null) {
            Schema w3cSchema = schema.getManager().getSchema(XMLSCHEMA_URI);
            if (w3cSchema == null)
                w3cSchema = schema;
            Type type = new TypeRef(w3cSchema, ANY_TYPE);
            decl.setType(type);
            initType(decl);
        }

        // validate default value if there is any
        if (decl.getValueConstraint() != null) {
            ValidationContextProvider vcp = null;
            if (contextProviders != null) {
                vcp = (ValidationContextProvider) contextProviders.get(decl);
            }
            decl.setValueConstraint(decl.getType().validateDefault(decl.getValueConstraint(), vcp));
        }

        // initialization identity-constraint
        Collection identityConstraints = decl.getIdentityConstraints();
        if (identityConstraints != null) {
            Iterator it = identityConstraints.iterator();
            while (it.hasNext()) {
                IdentityConstraint idConstraint = (IdentityConstraint) it.next();
                //initIdentityConstraint(idConstraint, decl);
                if (idConstraint.getCategory() == IdentityConstraint.CATEGORY_KEYREF) {
                    IdentityConstraint idRefer = idConstraint.getReferencedKey();
                    // idRefer is verificated by Schema of Schemas
                    idRefer = (IdentityConstraint) refResolver.resolve(idRefer);
                    idConstraint.setReferencedKey(idRefer);
                    // if the category is keyref, the cardinality of the fields must equal
                    // that of the fields of the referenced key
                    // c-props-correct.2
                    if (!idConstraint.isEqualToKeyRefFields()) {
                        throw new SchemaException("c-props-correct.2", idConstraint);
                    }
                }
            }
        }
    }

    public void visit(AttributeDeclaration decl) throws SchemaException {
        //      if (decl.getSchema() != schema || initialized.contains(decl)) return;
        if (initialized.contains(decl))
            return;
        initialized.add(decl);

        //      System.err.println("Init attribute : " + decl.getName());

        // if type  is null, it's simple ur-type definition
        if (decl.getType() != null) {
            initType(decl);
            if (!decl.getType().isSimpleType()) // attribute type must be simpleType
                throw new SchemaException("src-attribute.3.2", decl);

            // it's an error for NOTATION to be used directly in a schema
            if (decl.getType() == decl.getManager().getType(XMLSCHEMA_URI, "NOTATION")) {
                throw new SchemaException("enumeration-required-notation", decl);
            }
        } else {
            SimpleType type = new SimpleType(decl.getSchema(), SIMPLE_UR_TYPE, null);
            type.setPrimitive(PrimitiveType.createType(SIMPLE_UR_TYPE));
            decl.setType(type);
        }

        try {
            // validate default value if there is any
            if (decl.getValueConstraint() != null) {
                ValidationContextProvider vcp = null;
                if (contextProviders != null) {
                    vcp = (ValidationContextProvider) contextProviders.get(decl);
                }
                decl.setValueConstraint(decl.getType().validateDefault(decl.getValueConstraint(), vcp));
            }
        } catch (SchemaException e) {
            throw new SchemaException("a-props-correct.2", decl, e);
        }
    }

    public void visit(SimpleType type) throws SchemaException {
        if (type.getSchema() != schema || initialized.contains(type))
            return;
        initialized.add(type);

        //System.err.println("Init simpleType : " + type.getName());
        PrimitiveType primitive = null;
        switch (type.getVariety()) {
            // This is a hack. Atomic means that we have a restriction
            case SimpleType.VARIETY_ATOMIC :
                SimpleType base = (SimpleType) initBaseType(type);
                if (base != null) {
                    type.setVariety(base.getVariety());
                    type.setItemType(base.getItemType());
                    List memberTypes = base.getMemberTypes();
                    if (memberTypes != null)
                        type.addMemberTypes(memberTypes);
                    try {
                        primitive = (PrimitiveType) base.getPrimitive().clone();
                    } catch (CloneNotSupportedException e) {
                        // it's vide
                    }
                } else if (!XMLSCHEMA_URI.equals(type.getNamespace())) {
                    //        "Either it must have a base [attribute] or a <simpleType> among its [children], but not both";
                    throw new SchemaException("src-simple-type.2", type);
                }
                break;
            case SimpleType.VARIETY_LIST :
                SimpleType item = initItemType(type);
                if (item != null) {
                    if (item.getVariety() == SimpleType.VARIETY_LIST) {
                        //          "The {item type definition} must have a {variety} of atomic or union";
                        throw new SchemaException("cos-st-restricts.2.1", type);
                    }
                    primitive = PrimitiveType.createListType(item.getPrimitive());
                } else {
                    //        "Either it must have an itemType [attribute] or a <simpleType> among its [children], but not both";
                    throw new SchemaException("src-simple-type.3", type);
                }
                break;

            case SimpleType.VARIETY_UNION :
                primitive = initMemberTypes(type);
                break;
        }

        if (primitive != null) {
            type.setPrimitive(primitive);
            // Add new facets to primitive type
            Iterator it = type.getAllFacets().iterator();
            while (it.hasNext()) {
                Facet facet = (Facet) it.next();
                ContextLocatorImpl vcp = null;
                if (contextProviders != null) {
                    vcp = (ContextLocatorImpl) contextProviders.get(facet);
                    vcp.setNotationDeclarations(schema.getNotationDeclarationMap());
                }
                primitive.setFacet(facet, vcp);
            }
        }

        checkNotationType(type);
    }

    public void visit(ComplexType type) throws SchemaException {
        if (type.getSchema() != schema || initialized.contains(type)) {
            return;
        }
        initialized.add(type);

        // Complex types used in elements of this complex type are buffered
        // to avoid recursive graph traversal
        ArrayList localPendingTypes = null;
        if (pendingTypes == null) {
            localPendingTypes = new ArrayList();
            pendingTypes = localPendingTypes;
        }

        // System.err.println("Init complexType : " + type.getName());

        // If content model not initialized, set it to empty.
        if (type.getContentModel() == null) {
            type.setContentModel(new ContentModel());
        }

        Type base = initBaseType(type);

        if (type.isExtension()) {
            if (base == null) {
                // @@ should not occur
            } else if (base.isSimpleType()) {
                if (type.getContentModel().getContentType() == TEXT_ONLY) {
                    type.getContentModel().setModel((SimpleType) base);
                } else {
                    //          "Complex content complex type cannot derive from a simple type";
                    throw new SchemaException("src-ct.1", type);
                }
                initAttributes(type, null);
            } else {
                initAttributes(type, (ComplexType) base);
                if (type.getContentModel().getContentType() == TEXT_ONLY) {
                    if (((ComplexType) base).getContentModel().getContentType() != TEXT_ONLY) {
                        //            "Simple content complex type cannot derive from a complex type";
                        throw new SchemaException("src-ct.2", type);
                    }
                    type.getContentModel().setModel(((ComplexType) base).getContentModel().getModel());
                } else if (((ComplexType) base).getContentModel().getContentType() == EMPTY) {
                    initContent(type, null);
                } else if (type.getContentModel().getContentType() == EMPTY) {
                    type.getContentModel().setContentType(((ComplexType) base).getContentModel().getContentType());
                    initContent(type, (ComplexType) base);
                } else {
                    if (type.getContentModel().getContentType()
                        != ((ComplexType) base).getContentModel().getContentType()) {
                        //            "Both {content type}s must be mixed or both must be element-only";
                        throw new SchemaException("cos-ct-extends.1.4.2.2.2.1", type);
                    }
                    initContent(type, (ComplexType) base);
                }
            }
        } else {
            // Restriction
            if (base == null) {
                // Restriction from the ur-type
                initAttributes(type, null);
                initContent(type, null);
            } else if (base.isSimpleType()) {
                if (type.getContentModel().getContentType() == TEXT_ONLY) {
                    //          "Simple content complex type cannot restrict a simple base type";
                    throw new SchemaException("src-ct.2", type);
                } else {
                    //          "Complex content complex type cannot derive from a simple base type";
                    throw new SchemaException("src-ct.1", type);
                }
            } else {
                initAttributes(type, (ComplexType) base);
                if (type.getContentModel().getContentType() == TEXT_ONLY) {
                    if (type.getContentModel().getModel() == null)
                        type.getContentModel().setModel(((ComplexType) base).getContentModel().getModel());
                    else
                         ((SimpleType) type.getContentModel().getModel()).accept(this);
                } else {
                    initContent(type, null);
                }
                type.checkDerivationValidRestriction(loader);
            }
        }

        type.checkUniqueParticleAttribution();

        if (localPendingTypes != null) {
            // This will allow types to be initialized to store
            // their own pending type list
            pendingTypes = null;
            Iterator it = localPendingTypes.iterator();
            while (it.hasNext()) {
                ((SchemaVisitable) it.next()).accept(this);
            }
        }
    }

    private Type initType(Declaration decl) throws SchemaException {
        Type type = decl.getType();
        if (type != null) {
            type = (Type) refResolver.resolve(type);
            if (type.isSimpleType() || pendingTypes == null) {
                type.accept(this);
            } else {
                resolveDerivationHierarchy(type);
                pendingTypes.add(type);
            }
            decl.setType(type);
        }
        if (!type.isSimpleType() && type.getName() == null) {
            type.accept(this);
        }
        return type;
    }

    private void resolveDerivationHierarchy(Type type) throws SchemaException {
        Type previous = type;
        Type base = type.getBaseType();
        while (base != null) {
            base = (Type) refResolver.resolve(base);
            previous.setBaseType(base);
            if (base == type) {
                String errCode = null;
                if (type.isSimpleType())
                    errCode = "st-props-correct.2";
                else
                    errCode = "ct-props-correct.3";
                //      "circular definitions are disallowed";
                throw new SchemaException(errCode, type);
            }
            previous = base;
            base = base.getBaseType();
        }
    }

    private void propagateSubstitution(ElementDeclaration subst, ElementDeclaration decl) {
        for (subst = (ElementDeclaration) subst.getSubstitutionGroup();
            subst != null;
            subst = (ElementDeclaration) subst.getSubstitutionGroup()) {
            String errCode = decl.getType().checkTypeDerivationOK(subst.getType(), subst.getBlock(), true);
            if (errCode != null)
                return;
            List list = subst.getSubstitutionElementList();
            if (list.indexOf(decl) != -1)
                return;
            subst.addSubstitutionMember(decl);
        }
    }

    private Type initBaseType(Type type) throws SchemaException {
        Type base = type.getBaseType();
        // base is primitive
        if (base != null) {
            if (SIMPLE_UR_TYPE.equals(base.getName())) {
                if (XMLSCHEMA_URI.equals(schema.getNamespace())) {
                    base = null;
                    ((SimpleType) type).setPrimitive(PrimitiveType.createType(type.getName()));
                    type.setBaseType(base);
                } else {
                    //        "Base type defintion must be an atomic simple type or a built-in primitive datatype";
                    throw new SchemaException("cos-st-restricts.1.1", type);
                }
            } else {
                // base is SimpleType or ComplexType
                base = (Type) refResolver.resolve(base);
                base.accept(this);
                type.setBaseType(base);
            }
        }

        return base;
    }

    private Type initBaseType(SimpleType type) throws SchemaException {
        Type base = initBaseType((Type) type);
        if (base != null) {
        	if (!base.isSimpleType()) {
				throw new SchemaException("derivation-ok-restriction.5.1.1", base);
        	}
        	SimpleType simpleBase = (SimpleType) base;
            SimpleType current = simpleBase;
            while (current != null) {
                if (current == type) {
                    //        "circular definitions are disallowed";
                    throw new SchemaException("st-props-correct.2", type);
                }
                current = (SimpleType) current.getBaseType();
            }
            int exclusions = base.getFinal();
            if ((exclusions & RESTRICTION) != 0) {
                //      "Base type final must not contain restriction";
                throw new SchemaException("st-props-correct.3", type);
            } else if (simpleBase.getVariety() == SimpleType.VARIETY_LIST && (exclusions & LIST) != 0) {
                //      "Base type final must not contain list";
                throw new SchemaException("st-props-correct.4.2.1", type);
            } else if (simpleBase.getVariety() == SimpleType.VARIETY_LIST && (exclusions & LIST) != 0) {
                //      "Base type final must not contain union";
                throw new SchemaException("st-props-correct.4.2.2", type);
            }
        }
        return base;
    }

    private Type initBaseType(ComplexType type) throws SchemaException {
        Type base = initBaseType((Type) type);
        if (base != null) {
            Type current = base;
            while (current != null) {
                if (current == type) {
                    //        "circular definitions are disallowed";
                    throw new SchemaException("ct-props-correct.3", type);
                }
                current = current.getBaseType();
            }
            int exclusions = base.getFinal();
            if (type.isRestriction() && (exclusions & RESTRICTION) != 0) {
                //      "Base type final must not contain restriction";
                throw new SchemaException("derivation-ok-restriction.1", type);
            } else if (type.isExtension() && (exclusions & EXTENSION) != 0) {
                String errCode = null;
                if (base.isSimpleType())
                    errCode = "cos-ct-extends.2.2";
                else
                    errCode = "cos-ct-extends.1.1";
                //        "Base type final must not contain extension";
                throw new SchemaException(errCode, type);
            }
        }
        return base;
    }

    private SimpleType initItemType(SimpleType type) throws SchemaException {
        Type itemType = type.getItemType();
        if (itemType != null) {
            itemType = (Type) refResolver.resolve(itemType);
            itemType.accept(this);
            type.setItemType(itemType);
        }
        return (SimpleType) itemType;
    }

    private PrimitiveType initMemberTypes(Type type) throws SchemaException {
        ArrayList primitiveTypes = new ArrayList();
        String primitiveName = "union of ";
        ListIterator it = ((SimpleType) type).getMemberTypes().listIterator();
        while (it.hasNext()) {
            // init memberType
            // resolve TypeRef
            Type memberType = (Type) it.next();
            memberType = (Type) refResolver.resolve(memberType);
            memberType.accept(this);
            if (!memberType.isSimpleType()) {
                //      "Member type must have a variety of atomic or list";
                throw new SchemaException("cos-st-restricts.3.1", memberType);
            } else if (memberType == type) {
                //      "Circular union definition is disallowed";
                throw new SchemaException("src-simple-type.4", type);
            } else if (((SimpleType) memberType).getVariety() == SimpleType.VARIETY_UNION) {
                it.remove();
                Iterator it2 = ((SimpleType) memberType).getMemberTypes().iterator();
                while (it2.hasNext()) {
                    SimpleType m = (SimpleType) it2.next();
                    if (m == type) {
                        //          "Circular union definition is disallowed";
                        throw new SchemaException("src-simple-type.4", type);
                    }
                    it.add(m);
                    primitiveTypes.add(m.getPrimitive());
                    if (m.getName() != null)
                        primitiveName += memberType.getName() + " ";
                    else
                        primitiveName += m.getPrimitive().getName() + " ";
                }
            } else {
                it.set(memberType);
                PrimitiveType primitive = ((SimpleType) memberType).getPrimitive();
                primitiveTypes.add(primitive);
                if (memberType.getName() != null)
                    primitiveName += memberType.getName() + " ";
                else
                    primitiveName += primitive.getName() + " ";
            }
        }
        return PrimitiveType.createUnionType(primitiveTypes, primitiveName);
    }

    private void initAttributes(ComplexType type, ComplexType base) throws SchemaException {
        ArrayList requiredAndDefaultAttributes = new ArrayList();
        // groupContent is used for attributegroup
        ArrayList groupContent = new ArrayList();
        ArrayList prohibited = new ArrayList();

        Iterator it = type.getAttributeEntries().iterator();
        while (it.hasNext()) {
            Map.Entry entry = (Map.Entry) it.next();
            SchemaComponent comp = refResolver.resolve((SchemaComponent) entry.getValue());
            // check if it's an AttributeDeclaration
            if (comp instanceof AttributeDeclaration) {
                AttributeDeclaration decl = (AttributeDeclaration) comp;
                entry.setValue(decl);
                decl.accept(this);
                if (decl.getUse() == PROHIBITED) {
                    it.remove();
                    prohibited.add(decl);
                } else {
                    checkIDAttribute(type, decl);
                    checkNotationAttribute(type, decl);
                    if (decl.getUse() == REQUIRED || decl.getValueConstraint() != null)
                        requiredAndDefaultAttributes.add(decl);
                }
            }
            // it's an AttributeGroupDefinition
            else {
                //        comp.accept(this);  // AttributeGroupDefinition is visited below
                appendContent(type, (AttributeGroupDefinition) comp, groupContent, null);
                it.remove();
            }
        }

        // handle included AttributeGroupDefinition
        it = groupContent.iterator();
        while (it.hasNext()) {
            AttributeDeclaration decl = (AttributeDeclaration) it.next();
            decl = (AttributeDeclaration) importInScope(decl, type);
            decl.accept(this);
            if (decl.getUse() == PROHIBITED)
                prohibited.add(decl);
            else {
                type.register(decl);
                checkIDAttribute(type, decl);
                checkNotationAttribute(type, decl);
                if (decl.getUse() == REQUIRED || decl.getValueConstraint() != null)
                    requiredAndDefaultAttributes.add(decl);
            }
        }

        // if base != null, add base attributes to this
        if (base != null) {
            it = base.getAttributeDeclarations().iterator();
            while (it.hasNext()) {
                AttributeDeclaration decl = (AttributeDeclaration) it.next();
                AttributeDeclaration tDecl = type.getAttributeDeclaration(decl.getNamespace(), decl.getName());
                if (tDecl == null) {
                    boolean authorized = true;
                    Iterator it2 = prohibited.iterator();
                    while (authorized && it2.hasNext()) {
                        AttributeDeclaration p = (AttributeDeclaration) it2.next();
                        if (p.hasName(decl.getNamespace(), decl.getName()))
                            authorized = false;
                    }
                    if (authorized) {
                        // @@Scope
                        //decl = (AttributeDeclaration)importInScope(decl, type);
                        type.register((AttributeDeclaration) decl);
                        checkIDAttribute(type, decl);
                        checkNotationAttribute(type, decl);
                        if (decl.getUse() == REQUIRED || decl.getValueConstraint() != null)
                            requiredAndDefaultAttributes.add(decl);
                    }
                } else if (type.isExtension()) {
                    //        "Two distinct attribute declarations should not have identical names";
                    throw new SchemaException("ct-props-correct.4", tDecl);
                }
            }
            if (type.isExtension())
                type.setAttributeWildcard(wildcardUnion(type.getAttributeWildcard(), base.getAttributeWildcard()));
            else
                type.setAttributeWildcard(
                    wildcardIntersection(type.getAttributeWildcard(), base.getAttributeWildcard()));

        }
        type.setRequiredAndDefaultAttributes(requiredAndDefaultAttributes);
    }

    void checkIDAttribute(ComplexType type, AttributeDeclaration decl) throws SchemaException {
        Type attrType = decl.getType();
        if (attrType.isDerivedFrom(type.getManager().getType(XMLSCHEMA_URI, "ID"))) {
            if (type.getIDAttribute() != null) {
                //      "Two distinct attribute declarations must not have types which are or are derived from ID.";
                throw new SchemaException("ct-props-correct.5", type);
            }
            type.setIDAttribute(decl);
        }
    }

    void checkNotationAttribute(ComplexType type, AttributeDeclaration decl) {
        Type attrType = decl.getType();
        if (attrType.isDerivedFrom(type.getManager().getType(XMLSCHEMA_URI, "NOTATION"))) {
            type.setNotationAttribute(decl);
        }
    }

    void checkNotationType(SimpleType type) throws SchemaException {
        if (type.isDerivedFrom(type.getManager().getType(XMLSCHEMA_URI, "NOTATION"))) {
            Facet enumFacet = (Facet) type.getFacets().get(ENUMERATION_TAG);
            if (enumFacet == null) {
                // enumeration facet is mandatory
                throw new SchemaException("enumeration-required-notation", type);
            }
        }
    }

    private void appendContent(ComplexType type, AttributeGroupDefinition def, ArrayList content, ArrayList expanded)
        throws SchemaException {
        if (expanded == null)
            expanded = new ArrayList();
        if (expanded.indexOf(def) != -1) {
            //    "Circular attribute group definition";
            throw new SchemaException("src-attribute_group.3", def);
        } else {
            expanded.add(def);
            type.setAttributeWildcard(wildcardIntersection(type.getAttributeWildcard(), def.getAttributeWildcard()));
            Iterator it = def.getContents().iterator();
            while (it.hasNext()) {
                SchemaComponent comp = refResolver.resolve((SchemaComponent) it.next());
                if (comp instanceof AttributeDeclaration) {
                    content.add(comp);
                } else {
                    appendContent(type, (AttributeGroupDefinition) comp, content, expanded);
                }
            }
        }
    }

    private Wildcard wildcardIntersection(Wildcard wc1, Wildcard wc2) throws SchemaException {
        if (wc1 == null || wc1.isAny())
            return wc2;
        return wc1.wildcardIntersection(wc2);
    }

    private Wildcard wildcardUnion(Wildcard wc1, Wildcard wc2) throws SchemaException {
        if (wc1 == null)
            return wc2;
        return wc1.wildcardUnion(wc2);
    }

    private void initContent(ComplexType type, ComplexType base) throws SchemaException {
        // first, get and init particle
        Particle particle = (Particle) type.getContentModel().getModel();
        if (particle != null) {
            particle = initParticle(particle, type, false, null);
            type.getContentModel().setModel(particle);
        }

        // then, get and init base particle
        if (base != null) {
			Iterator decls = base.getElementDeclarations().iterator();
			while (decls.hasNext()) {
				ElementDeclaration decl = (ElementDeclaration) decls.next();
				type.register(decl);
			}
            ModelGroup group = null;

            // if the type particle is null, create a empty particle
            if (particle == null) {
                group = new SequenceModelGroup();
                particle = new Particle(group);
                type.getContentModel().setModel(particle);
            }
            // else, get particle's term and check if it's a sequence model
            else if (
                particle.getMinOccurs() == 1
                    && particle.getMaxOccurs() == 1
                    && particle.getTerm() instanceof SequenceModelGroup) {
                group = (ModelGroup) particle.getTerm();
            }
            // the particle's term isn't a sequence model, create a sequence model
            else {
                group = new SequenceModelGroup();
                group.add(particle);
                particle = new Particle(group);
                type.getContentModel().setModel(particle);
            }

            // init base particle
            // @@Scope
            //Particle basePart = initParticle((Particle)base.getContentModel().getModel(), type, true, null);
            Particle basePart = (Particle) base.getContentModel().getModel();
            if (basePart != null) {
                if (basePart.getTerm() instanceof ModelGroup) {
                    ModelGroup baseGroup = (ModelGroup) basePart.getTerm();
                    if (baseGroup.size() != 0) {
                        if (basePart.getMinOccurs() == 1
                            && basePart.getMaxOccurs() == 1
                            && baseGroup.getCompositor() == ModelGroup.SEQUENCE) {
                            Iterator it = baseGroup.iterator();
                            int i = 0;
                            while (it.hasNext())
                                group.add(i++, it.next());
                        } else {
                            group.add(0, basePart);
                        }
                    }
                } else {
                    group.add(0, basePart);
                }
            }

            if (particle == null) {
                type.getContentModel().setContentType(EMPTY);
            } else {
                particle.setTerm(group);
                new ValidationInfoInitializer(particle).initialize();
            }
        }
    }

    private Particle initParticle(Particle p, SchemaScope scope, boolean duplicate, ArrayList expanded)
        throws SchemaException {
        ModelGroup group = null;
        if (p.getMaxOccurs() == 0)
            return null;

        try {
            if (duplicate) {
                Particle newP = (Particle) p.clone();
                if (loader != null)
                    loader.addComp2ComponentContextLocators(newP, p);
                p = newP;
            }
        } catch (CloneNotSupportedException ex) {
            // should not happen
            throw new SchemaException(null, p);
        }

        if (p.getTerm() instanceof SchemaComponent) {
            SchemaComponent comp = refResolver.resolve((SchemaComponent) p.getTerm());
            if (comp instanceof ElementDeclaration) {
                comp = importInScope((ElementDeclaration) comp, scope);
                comp.accept(this);
                scope.register((ElementDeclaration) comp);
                p.setTerm(comp);
            } else if (comp instanceof ModelGroupDefinition) {
                group = (ModelGroup) ((ModelGroupDefinition) comp).getModelGroup();
                // @@@ This does not work when the ALL group is first in model
                // Should be checked afterwards instead
                if (group.getCompositor() == ModelGroup.ALL && expanded != null) {
                    //         "All group can only appear as main group in a content model";
                    //          throw new SchemaException("cos-all-limited.1", p);
                }
                if (expanded == null)
                    expanded = new ArrayList();
                else if (expanded.indexOf(comp) != -1) {
                    //        "Circular model group definition";
                    throw new SchemaException("mg-props-correct.2", p);
                }
                expanded.add(comp);
                duplicate = true;
            }
        } else if (p.getTerm() instanceof ModelGroup) {
            group = (ModelGroup) p.getTerm();
        }

        if (group != null) {
            if (expanded == null)
                expanded = new ArrayList();
            group = initGroup(group, scope, duplicate, expanded);
            if (group.size() == 0 && (group.getCompositor() != ModelGroup.CHOICE || p.getMinOccurs() == 0))
                return null;
            p.setTerm(group);
            if (group.size() == 1 && p.getMinOccurs() == 1 && p.getMaxOccurs() == 1) {
                p = (Particle) group.get(0);
            }
        }

        new ValidationInfoInitializer(p).initialize();
        return p;
    }

    private ModelGroup initGroup(ModelGroup group, SchemaScope scope, boolean duplicate, ArrayList expanded)
        throws SchemaException {
        if (duplicate)
            group = (ModelGroup) group.clone();
        //		System.err.println("group = " + group);
        ListIterator it = group.listIterator();
        while (it.hasNext()) {
            Particle p = initParticle((Particle) it.next(), scope, duplicate, expanded);
            it.set(p);
            if (p == null)
                it.remove();
            else if (p.getTerm() instanceof ModelGroup) {
                ModelGroup subgroup = (ModelGroup) p.getTerm();
                if (p.getMinOccurs() == 1
                    && p.getMaxOccurs() == 1
                    && subgroup.getCompositor() == group.getCompositor()) {
                    it.remove();
                    Iterator it2 = subgroup.iterator();
                    while (it2.hasNext())
                        it.add(it2.next());
                }
            }
        }

        return group;
    }

    private Declaration importInScope(Declaration decl, SchemaScope scope) throws SchemaException {
        if (decl.getScope() != null)
            return decl;
        // @@Scope: used to be
        // if (decl.getScope() == scope) return decl;
        Declaration result = null;
        try {
            result = (Declaration) decl.clone();
            if (loader != null)
                loader.addComp2ComponentContextLocators(result, decl);
        } catch (CloneNotSupportedException ex) {
            // should not happen
            throw new SchemaException(null, decl);
        }
        result.setScope(scope);
        return result;
    }

    class ReferenceResolver extends DefaultSchemaVisitor {
        private static final String RCSRevision = "$Revision: 1.6 $";
        private static final String RCSName = "$Name: XQ1_1_1 $";

        SchemaComponent result = null;

        SchemaComponent resolve(SchemaComponent ref) throws SchemaException {
            result = ref;
            ref.accept(this);
            if (result == null) {
                //      "Unresolved reference";
                throw new SchemaException("src-resolve.3", ref);
            }
            return result;
        }

        public void visit(SchemaComponentRef ref) {
            result = ref.getSchema().getSchemaComponent(ref.getName(), ref.getType());
        }

        public void visit(TypeRef ref) {
            result = ref.getSchema().getType(ref.getName());
        }

        public void visit(ElementDeclarationRef ref) throws SchemaException {
            result = ref.getSchema().getElementDeclaration(ref.getName());
        }

        public void visit(AttributeDeclarationRef ref) throws SchemaException {
            AttributeDeclaration decl = ref.getSchema().getAttributeDeclaration(ref.getName());
            if (decl != null) {
                try {
                    AttributeDeclaration newDecl = (AttributeDeclaration) decl.clone();
                    if (loader != null)
                        loader.addComp2ComponentContextLocators(newDecl, decl);
                    decl = newDecl;
                } catch (CloneNotSupportedException e) {
                    // should not happen
                    throw new SchemaException(null, decl);
                }
                decl.setUse(ref.getUse());
                String def = ref.getDefaultValue();
                String fix = ref.getFixedValue();
                String fixDecl = decl.getFixedValue();
                if (fixDecl != null) {
                    if (def != null || (fix != null && !fix.equals(fixDecl))) {
                        //          "Attribute ref must have same fixed value as original";
                        throw new SchemaException("au-props-correct.2", decl);
                    }
                } else if (fix != null) {
                    decl.setFixedValue(fix);
                    decl.setDefaultValue(null);
                } else if (def != null) {
                    decl.setDefaultValue(def);
                }
            }
            result = decl;
        }

        public void visit(SimpleBaseTypeRef ref) throws SchemaException {
            Type base = ref.getType().getBaseType();
            if (base.isSimpleType() == false) {
                ContentModel model = ((ComplexType) base).getContentModel();
                if (model.getContentType() == TEXT_ONLY) {
                    result = (SchemaComponent) model.getModel();
                    return;
                }
            }
            //    "Base content type must be a simple type";
            throw new SchemaException("derivation-ok-restriction.5.1.1", base);
        }

        public void visit(IdentityConstraintRef ref) throws SchemaException {
            result = ref.getSchema().getIdentityConstraint(ref.getName());
            if (result != null && ((IdentityConstraint) result).getCategory() == IdentityConstraint.CATEGORY_KEYREF)
                result = null;
        }
    }

    class ValidationInfoInitializer extends DefaultSchemaVisitor {
        private static final String RCSRevision = "$Revision: 1.6 $";
        private static final String RCSName = "$Name: XQ1_1_1 $";

        private Particle particle;

        ValidationInfoInitializer(Particle p) {
            particle = p;
        }

        void initialize() throws SchemaException {
            if (particle.getFirstElements().isEmpty() && particle.getFirstWildcards() == null) {
                ((SchemaVisitable) particle.getTerm()).accept(this);
            }
        }

        public void visit(ElementDeclaration decl) {
            particle.addFirstElement(decl);
            particle.setMinExtent(particle.getMinOccurs());
            particle.setMaxExtent(particle.getMaxOccurs());
        }

        public void visit(AllModelGroup group) {
            Iterator it = group.iterator();
            long minExtent = 0;
            long maxExtent = 0;
            while (it.hasNext()) {
                Particle p = (Particle) it.next();
                minExtent += p.getMinExtent();
                maxExtent += p.getMaxExtent();
                particle.addFirstElements(p.getFirstElements());
                particle.addFirstWildcards(p.getFirstWildcards());
            }
            if (maxExtent > Integer.MAX_VALUE)
                maxExtent = Integer.MAX_VALUE;
            minExtent *= particle.getMinOccurs();
            maxExtent *= particle.getMaxOccurs();
            if (maxExtent > Integer.MAX_VALUE)
                maxExtent = Integer.MAX_VALUE;
            particle.setMinExtent((int) minExtent);
            particle.setMaxExtent((int) maxExtent);
        }

        public void visit(ChoiceModelGroup group) {
            Iterator it = group.iterator();
            long minExtent = Integer.MAX_VALUE;
            long maxExtent = 0;
            while (it.hasNext()) {
                Particle p = (Particle) it.next();
                if (p.getMinExtent() < minExtent)
                    minExtent = p.getMinExtent();
                if (p.getMaxExtent() > maxExtent)
                    maxExtent = p.getMaxExtent();
                particle.addFirstElements(p.getFirstElements());
                particle.addFirstWildcards(p.getFirstWildcards());
            }
            minExtent *= particle.getMinOccurs();
            maxExtent *= particle.getMaxOccurs();
            if (maxExtent > Integer.MAX_VALUE)
                maxExtent = Integer.MAX_VALUE;
            particle.setMinExtent((int) minExtent);
            particle.setMaxExtent((int) maxExtent);
        }

        public void visit(SequenceModelGroup group) {
            Iterator it = group.iterator();
            long minExtent = 0;
            long maxExtent = 0;
            boolean optional = true;
            while (it.hasNext()) {
                Particle p = (Particle) it.next();
                minExtent += p.getMinExtent();
                maxExtent += p.getMaxExtent();
                if (optional) {
                    particle.addFirstElements(p.getFirstElements());
                    particle.addFirstWildcards(p.getFirstWildcards());
                }
                if (p.getMinExtent() > 0)
                    optional = false;
            }
            if (maxExtent > Integer.MAX_VALUE)
                maxExtent = Integer.MAX_VALUE;
            minExtent *= particle.getMinOccurs();
            maxExtent *= particle.getMaxOccurs();
            if (maxExtent > Integer.MAX_VALUE)
                maxExtent = Integer.MAX_VALUE;
            particle.setMinExtent((int) minExtent);
            particle.setMaxExtent((int) maxExtent);
        }

        public void visit(Wildcard wildcard) {
            particle.addFirstWildcard(wildcard);
            particle.setMinExtent(particle.getMinOccurs());
            particle.setMaxExtent(particle.getMaxOccurs());
        }
    }

}
