/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.schema;

import java.util.*;

/**
 * Basic class for Schema package user.
 *
 * <p>This is the implementation of The Schema Itself. At abstract level, 
 * the schema itself is just a container for its components.
 * The following properties are defined</p>
 * <p> 1. type definitions: a set of named simple and comple type definitions. </p>
 * <p> 2. attribute declarations: a set of named (top-level) attrubite declarations. </p>
 * <p> 3. element declarations: a set of named (top-level) element declarations. </p>
 * <p> 4. attribute group declarations: a set of named attribute group declarations. </p>
 * <p> 5. model group declarations: a set of named model group declarations. </p>
 *
 * <p>See XML Schema Part 1: Structures 3.15.1
 * W3C Recommendation 2 May 2001</p>
 *
 */
public class Schema implements SchemaConstants, SchemaVisitable, SchemaScope {
private static final String RCSRevision = "$Revision: 1.3 $"; 
private static final String RCSName     = "$Name: XQ1_1_1 $";

  public static final int DATATYPE = 0;
  public static final int ELEMENT_DECLARATION = 1;
  public static final int ATTRIBUTE_DECLARATION = 2;
  public static final int MODEL_GROUP_DEFINITION = 3;
  public static final int ATTRIBUTE_GROUP_DEFINITION = 4;
  public static final int NOTATION_DECLARATION = 5;

  private HashMap dataTypes = new HashMap();
  private HashMap modelGroupDefinitions = new HashMap();
  private HashMap attributeGroupDefinitions = new HashMap();
  private HashMap elementDeclarations = new HashMap();
  private HashMap attributeDeclarations = new HashMap();
  private HashMap notationDeclarations = new HashMap();

  private String targetNamespace;
  private String schemaLocation;
  private SchemaManager manager;
  private boolean loaded = false;
  private HashSet includedLocations = new HashSet();
  private HashSet importedNamespaces = new HashSet();
  private LinkedList defaults = new LinkedList();
  private HashMap prefixMap = null;
  private String prefix;
  private InitializationVisitor initializer = null;

  
  // identity constraints of this schema
  // this is only used to initialize keyref and to validate keyref
  private HashMap identityConstraints = null;

  /**
   * Create a new Schema.
   *
   * Schema location example 2: file:/d:/schema/example.xsd
   *
   * @param uri The target namesapce.
   * @param location The schema location.
   */
  public Schema(String uri, String location, SchemaManager manager) {
    targetNamespace = uri;
    schemaLocation = location;
    this.manager = manager;
    pushDefaults();
    addImportedNamespace(XMLSCHEMA_URI);
	initializer = new InitializationVisitor(this);
  }

  // SchemaVisitor is applied to initialize this schema
  public void accept(SchemaVisitor visitor) throws SchemaException {
    visitor.visit(this);
  }
  

  public void initialize() throws SchemaException {

	  java.util.Iterator it = getTypes().iterator();
      while (it.hasNext()) ((SchemaVisitable)it.next()).accept(initializer);

      it = getElementDeclarations().iterator();
      while (it.hasNext()) ((SchemaVisitable)it.next()).accept(initializer);

      it = getAttributeDeclarations().iterator();
      while (it.hasNext()) ((SchemaVisitable)it.next()).accept(initializer);
      
      it = getAttributeGroupDefinitions().iterator();
      while (it.hasNext()) ((SchemaVisitable)it.next()).accept(initializer);
  }

  /**
   * Return the schema's target namespace, if any.
   *
   * @return The target namespace, or null if there is none.
   */
  public String getNamespace() {
    return targetNamespace;
  }
  
  /**
   * Return the schema's target namespace, if any.
   *
   * @return The target namespace, or null if there is none.
   */
   public String toString() {
    return targetNamespace;
  }
  
  /**
   * Return true if schema has the same target namespace.
   *
   * @param schema The schema to compare.
   */
  public boolean equals(Schema schema) {
    return ( targetNamespace.equals(schema.getNamespace()) );
  }

  /**
   * Return schema location, if any.
   *
   * @return The schema location, or null if there is none.
   */
  public String getLocation() {
    return schemaLocation;
  }

  /**
   * Return the collection of schema document URIs
   * 
   * @return the collection of schema document URIs, as strings, or null if there are none
   */
  public Collection getSchemaDocuments() {
    ArrayList result = new ArrayList();
    if (schemaLocation != null) result.add(schemaLocation);
    result.addAll(includedLocations);
    return result;
  }
  
  /**
   * Return the set of imported schema namespaces.
   */
  public Set getImportedNamespaces() {
  	return Collections.unmodifiableSet(importedNamespaces);	
  }
  
  /**
   * Return the schema manager used to load the schema
   *
   * @return The schema manager associated to this schema
   */
  public SchemaManager getManager() {
    return manager;
  }

  /**
   * Giving a type name, look up the named type (simple type or complex type).
   *
   * @param name The type name.
   *
   * @return The Type, or null if there is none.
   */
  public Type getType(String name) {
    return (Type)getSchemaComponent(name, dataTypes);
  }

  /**
   * Giving an element declaration name, look up the top-level  element declaration.
   *
   * @param name The element declaration name.
   *
   * @return The ElementDeclaration, or null if there is none.
   */
  public ElementDeclaration getElementDeclaration(String name) {
    return (ElementDeclaration)getSchemaComponent(name, elementDeclarations);
  }
    
  /**
   * Giving an attribute declaration name, look up the top-level attribute declaration.
   *
   * @param name The attribute name.
   *
   * @return The AttributeDeclaration, or null if there is none.
   */
  public AttributeDeclaration getAttributeDeclaration(String name) {
    return (AttributeDeclaration)getSchemaComponent(name, attributeDeclarations);
  }
    
  /**
   * Giving a model group name, look up the named model group.
   *
   * @param name The model group name.
   *
   * @return The ModelGroupDefinition, or null if there is none.
   */
  public ModelGroupDefinition getModelGroupDefinition(String name) {
    return (ModelGroupDefinition)getSchemaComponent(name, modelGroupDefinitions);
  }

  /**
   * Giving an attribute group name, look up the named attribute group.
   *
   * @param name The attribute group name.
   *
   * @return The AttributeGroupDefinition, or null if there is none.
   */
  public AttributeGroupDefinition getAttributeGroupDefinition(String name) {
    return (AttributeGroupDefinition)getSchemaComponent(name, attributeGroupDefinitions);
  }

  /**
   * Giving a notation name, look up the named notation.
   *
   * @param name The notation name.
   *
   * @return The Notation, or null if there is none.
   */
  public NotationDeclaration getNotationDeclaration(String name) {
    return (NotationDeclaration)getSchemaComponent(name, notationDeclarations);
  }

  public SchemaComponent getSchemaComponent(String name, int type) {
    switch (type) {
    case DATATYPE: return getType(name);
    case ELEMENT_DECLARATION: return getElementDeclaration(name);
    case ATTRIBUTE_DECLARATION: return getAttributeDeclaration(name);
    case MODEL_GROUP_DEFINITION: return getModelGroupDefinition(name);
    case ATTRIBUTE_GROUP_DEFINITION: return getAttributeGroupDefinition(name);
    case NOTATION_DECLARATION: return getNotationDeclaration(name);
    default: return null;
    }
  }
  
  public boolean isGlobalScope() {
    return true;
  }
  
  public ElementDeclaration getElementDeclaration(String namespace, String name) {
    if (!targetNamespace.equals(namespace)) return null;
    return getElementDeclaration(name);
  }

  public AttributeDeclaration getAttributeDeclaration(String namespace, String name) {
    if (!targetNamespace.equals(namespace)) return null;
    return getAttributeDeclaration(name);
  }

  /**
   * Look up all named data types(all named simple types and complex types).
   *
   * @return A collection view of all named data types.
   */
  public Collection getTypes() {
    return dataTypes.values();
  }
  
  /**
   * Look up all top-level element declarations.
   *
   * @return A collection view of the top-level element declarations
   */
  public Collection getElementDeclarations() {
    return elementDeclarations.values();
  }

  /**
   * Look up all top-level attribute declarations.
   *
   * @return A collection view of the top-level attribute declarations
   */
  public Collection getAttributeDeclarations() {
    return attributeDeclarations.values();
  }
  
  /**
   * Look up all named model group definitions.
   *
   * @return A collection view of all named model group definitions
   */
  public Collection getModelGroupDefinitions() {
    return modelGroupDefinitions.values();
  }
  
  /**
   * Look up all named attribute group definitions.
   *
   * @return A collection view of all named attribute group definitions
   */
  public Collection getAttributeGroupDefinitions() {
    return attributeGroupDefinitions.values();
  }
  /**
   * Look up all notation declarations.
   *
   * @return A collection view of all notation declarations
   */
  public Collection getNotationDeclarations() {
    return notationDeclarations.values();
  }
  
  public HashMap getNotationDeclarationMap() {
    return notationDeclarations;
  }
  
  private SchemaComponent getSchemaComponent(String name, HashMap map) {
    SchemaComponent result = (SchemaComponent)map.get(name);
    return result;
  }

  boolean isLoaded() { 
    return loaded; 
  }
  
  void setLoaded(boolean loaded) {
    this.loaded = loaded;
  }

  public boolean isIncluded(String location) {
    return includedLocations.contains(location);
  }

  public void addIncludedLocation(String location) {
    includedLocations.add(location);
  }

  public boolean isImported(String namespace) {
    return importedNamespaces.contains(namespace);
  }

  public void addImportedNamespace(String namespace) {
    importedNamespaces.add(namespace);
  }

  public void register(Type entity) throws SchemaException {
    register(entity, dataTypes);
  }

  public void register(ElementDeclaration entity) throws SchemaException {
    register(entity, elementDeclarations);
  }

  public void register(AttributeDeclaration entity) throws SchemaException {
    register(entity, attributeDeclarations);
  }
  
  public void register(ModelGroupDefinition entity) throws SchemaException {
    register(entity, modelGroupDefinitions);
  }

 public void register(AttributeGroupDefinition entity) throws SchemaException {
    register(entity, attributeGroupDefinitions);
  }

  public void register(NotationDeclaration entity) throws SchemaException {
    register(entity, notationDeclarations);
  }
  
  private void register(SchemaComponent comp, HashMap map) throws SchemaException {
    String name = comp.getName();
    SchemaComponent existing = getSchemaComponent(name, map);
    if (existing != null && !existing.equals(comp)) 
      // Redeclaration of schema component
      throw new SchemaException("sch-props-correct.2", comp);
    map.put(name, comp);
  }
  
  public void removeRegisteredComponent(Type entity) {
    dataTypes.remove(entity.getName());
  }

  public void removeRegisteredComponent(ElementDeclaration entity) {
    elementDeclarations.remove(entity.getName());
  }

  public void removeRegisteredComponent(AttributeDeclaration entity) {
    attributeDeclarations.remove(entity.getName());
  }
  
  public void removeRegisteredComponent(ModelGroupDefinition entity) {
    modelGroupDefinitions.remove(entity.getName());
  }

  public void removeRegisteredComponent(AttributeGroupDefinition entity) {
    attributeGroupDefinitions.remove(entity.getName());
  }
    
  public void removeRegisteredComponent(NotationDeclaration entity) {
    notationDeclarations.remove(entity.getName());
  }
    
  public void pushDefaults() {
    defaults.addFirst(new HashMap());
  }

  public void popDefaults() {
    defaults.removeFirst();
  }

  public void setDefault(String key, String val) {
    ((HashMap)defaults.getFirst()).put(key, val);
  }

  public String getDefault(String key) {
    return (String)((HashMap)defaults.getFirst()).get(key);
  }

  public void setPrefixMap(HashMap map) {
    prefixMap = map;
    
    // get this schema's prefix
    Iterator it = ((Set)prefixMap.entrySet()). iterator();
    while (it.hasNext()) {
      Map.Entry entry = (Map.Entry)it.next();
      if ( getNamespace() == null ) {
        prefix = "";
        break;
      }
      if ( getNamespace().equals(entry.getValue()) ) {
        prefix = (String)entry.getKey();
        break;
      }
    }
  }

  public HashMap getPrefixMap() {
    return prefixMap;
  }
  
  public String getPrefix() {
    return prefix;
  }
  
  // identity-constraint
  public void register(IdentityConstraint entity) throws SchemaException  {
    if ( identityConstraints == null ) identityConstraints = new HashMap();    
    register(entity, identityConstraints);
  }
  
  public Collection getIdentityConstraints() {
    if ( identityConstraints == null ) return null;
    return identityConstraints.values();
  }
  
  // name is IdentityConstraint's name
  public IdentityConstraint getIdentityConstraint(String name) {
    if ( identityConstraints == null ) return null;
    if ( name.indexOf(':') != -1 ) {
      name = name.substring(name.indexOf(':') + 1);
    }
    IdentityConstraint result = (IdentityConstraint)identityConstraints.get(name);
    return result;
  }
  
}
