/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.schema.datatypes;

import java.text.FieldPosition;
import java.text.ParsePosition;
import java.text.SimpleDateFormat;
import java.util.Date;

import org.apache.xerces.impl.xpath.regex.RegularExpression;
import org.xquark.schema.SchemaException;
import org.xquark.schema.validation.ValidationContextProvider;


abstract class AbstractDateTimeType extends ComparableType {
    private static final String RCSRevision = "$Revision: 1.3 $";
    private static final String RCSName = "$Name: XQ1_1_1 $";
	private static int[] POWERSOFTEN = { 1, 10, 100, 1000 };
	protected static final String TIMEZONE_PATTERN = "((((\\+)|(\\-))[0-9]{2}:[0-9]{2})|Z)?";

	private SimpleDateFormat timezoneFormat;
	private FieldPosition emptyFieldPosition = new FieldPosition(0);
		
    AbstractDateTimeType(String name, int nType) {
        super(name, nType);
		timezoneFormat = new SimpleDateFormat("HH:mm");
		timezoneFormat.setLenient(false);
    }

	protected Object toValidType(Object data) {
		return (Date) data;
	}
	
    protected abstract SimpleDateFormat getDateTimeFormat();
    
    protected abstract RegularExpression getRegularExpression();

	protected abstract boolean isTimeFormat();
	
	protected abstract boolean isDateFormat();
	
	protected abstract DateTime createDateTime(long millis, long tz, boolean tzSpecified);
	
    protected Comparable toComparable(String value) throws SchemaException {
		RegularExpression regex = getRegularExpression();
		if (!regex.matches(value)) 
			super.invalidValue(value);
		int len = value.length();
		SimpleDateFormat format = getDateTimeFormat();
		ParsePosition pos = new ParsePosition(0);
		Date date = null;
		if (len > 1 && value.charAt(0) == '-' && value.charAt(1) != '-') {
			// First check that the date is valid without the leading minus sign
			format.setLenient(false);
			pos.setIndex(1);
			date = format.parse(value, pos);
			if (date == null) super.invalidValue(value);
			// Change to lenient to accept leading minus sign and compute real value
			format.setLenient(true);
			pos.setIndex(0);
			date = format.parse(value, pos);
		} else {
			format.setLenient(false);
			date = format.parse(value, pos);
			if (date == null) super.invalidValue(value);
		}
       	long tz = 0;
       	long millis = 0;
       	boolean tzSpecified = false;
       	int index = pos.getIndex();
		if (isTimeFormat() && index < len && value.charAt(index) == '.') {
			index++;
			int count = 0;
			while (index < len) {
				char c = value.charAt(index++);
				if (!Character.isDigit(c)) break;
				millis *= 10;
				millis += ((int) c - (int) '0');
				count++;
			}
			if (count > 3) {
				// truncate milliseconds digits after the third
				millis = (long)(millis/Math.pow(10, count-3));
			} else {
				millis *= POWERSOFTEN[3 - count];
			}
		}
       	if (index < len) {
       		tzSpecified = true;
       		char c = value.charAt(index);
            // check if value has a time zone
            if (c == 'Z') {
            	tz = 0;
            	index++;
            } else if (c == '+') {
            	pos.setIndex(index+1);
            	Date tzDate = timezoneFormat.parse(value, pos);
            	if (tzDate == null) super.invalidValue(value);
            	tz = tzDate.getTime();
            	if (tz > DateTime.MAX_TIMEZONE_OFFSET) super.invalidValue(value);
            	index = pos.getIndex();
			} else if (c == '-') {
				pos.setIndex(index+1);
				Date tzDate = timezoneFormat.parse(value, pos);
				if (tzDate == null) super.invalidValue(value);
				tz = -tzDate.getTime();
				if (tz < -DateTime.MAX_TIMEZONE_OFFSET) super.invalidValue(value);
				index = pos.getIndex();
            } else {
            	super.invalidValue(value);
            }
       	}
       	if (index != len)
			super.invalidValue(value);
		return createDateTime(date.getTime()+millis, tz, tzSpecified);
    }

	public String toXMLString(Object data, ValidationContextProvider context) {
		if (!isTimeFormat()) return stringValue(data);
		if (data == null) return null;
		Date date = (Date) data;
		StringBuffer result = new StringBuffer(20);
		SimpleDateFormat format = getDateTimeFormat();
		format.format(date, result, emptyFieldPosition);
		long millis = date.getTime() % 1000;
		if (millis > 0) {
			result.append('.');
			if (millis < 10) {
				result.append("00");
				result.append(millis);
			} else if (millis < 100) {
				result.append("0");
				result.append(millis);
			} else {
				result.append(millis);
			}
			int len = result.length();
			while (result.charAt(len-1) == '0') {
				result.setLength(len-1);
				len--;
			}
		}
		if (date instanceof DateTime && ((DateTime) date).hasTimeZone()) {
			result.append('Z');
		}
		return result.toString();
	}
	
    protected String stringValue(Object data) {
        if (data == null) return null;
        Date date = (Date) data;
        StringBuffer result = new StringBuffer(20);
		SimpleDateFormat format = getDateTimeFormat();
		long tz = 0;
		boolean tzSpecified = false;
        if (date instanceof DateTime && ((DateTime) date).hasTimeZone()) {
        	tz = ((DateTime) date).getTimeZone();
        	tzSpecified = true;
        }
        long time = date.getTime();
        long millis = time % 1000;
		format.format(new Date(time+tz), result, emptyFieldPosition);
		if (millis > 0) {
			result.append('.');
			if (millis < 10) {
				result.append("00");
				result.append(millis);
			} else if (millis < 100) {
				result.append("0");
				result.append(millis);
			} else {
				result.append(millis);
			}
			int len = result.length();
			while (result.charAt(len-1) == '0') {
				result.setLength(len-1);
				len--;
			}
		}
		if (tzSpecified) {
        	if (tz == 0) {
        		result.append('Z');
        	} else if (tz > 0) {
				result.append('+');
				timezoneFormat.format(new Date(tz), result, emptyFieldPosition);
			} else {
				result.append('-');
				timezoneFormat.format(new Date(-tz), result, emptyFieldPosition);
        	}
        }
        return result.toString();
    }

}
