/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.schema.datatypes;

import java.util.*;

import org.apache.xerces.impl.xpath.regex.RegularExpression;
import org.xquark.schema.Facet;
import org.xquark.schema.SchemaConstants;
import org.xquark.schema.SchemaException;
import org.xquark.schema.validation.ValidationContextProvider;

public abstract class PrimitiveType implements Cloneable, SchemaConstants {
    private static final String RCSRevision = "$Revision: 1.4 $";
    private static final String RCSName = "$Name: XQ1_1_1 $";
    public static final int VOID = 0;
    public static final int STRING = 1;
    public static final int BOOLEAN = 2;
    public static final int DECIMAL = 3;
    public static final int DOUBLE = 4;
    public static final int FLOAT = 5;
    public static final int DURATION = 6;
    public static final int DATE_TIME = 7;
    public static final int TIME = 8;
    public static final int DATE = 9;
    public static final int GYEAR_MONTH = 10;
    public static final int GYEAR = 11;
    public static final int GMONTH_DAY = 12;
    public static final int GDAY = 13;
    public static final int GMONTH = 14;
    public static final int HEX_BINARY = 15;
    public static final int BASE64_BINARY = 16;
    public static final int ANY_URI = 17;
    public static final int QNAME = 18;
    public static final int NOTATION = 19;

    public static final int ID = 20;
    public static final int IDREF = 21;

    public static final int LIST = 30;
    public static final int UNION = 31;
    public static final int ANYSIMPLETYPE = 32;

    // constants used only by whiteSpace treating
    public static final int PRESERVE = 0;
    public static final int REPLACE = 1;
    public static final int COLLAPSE = 2;

    // Facet constants
    public static final int LENGTH_FACET = 0;
    public static final int MIN_LENGTH_FACET = 1;
    public static final int MAX_LENGTH_FACET = 2;
    public static final int PATTERN_FACET = 3;
    public static final int ENUMERATION_FACET = 4;
    public static final int WHITESPACE_FACET = 5;
    public static final int MIN_EXCLUSIVE_FACET = 6;
    public static final int MIN_INCLUSIVE_FACET = 7;
    public static final int MAX_EXCLUSIVE_FACET = 8;
    public static final int MAX_INCLUSIVE_FACET = 9;
    public static final int TOTALDIGITS_FACET = 10;
    public static final int FRACTIONDIGITS_FACET = 11;

    private String name;
    private int type;
    protected int nWhiteSpace = COLLAPSE; // default value
    protected int fixedBitSet = 1 << WHITESPACE_FACET;

    // parterns
    private ArrayList patterns = null;

    PrimitiveType(String name, int type) {
        this.name = name;
        this.type = type;
    }

    public Object clone() throws CloneNotSupportedException {
        PrimitiveType result = (PrimitiveType) super.clone();
        if (patterns != null)
            result.patterns = (ArrayList) patterns.clone();
        return result;
    }

    public String getName() {
        return name;
    }

    public int getType() {
        return type;
    }

    public int getLength() {
        return -1;
    }

    public int getMinLength() {
        return -1;
    }

    public int getMaxLength() {
        return -1;
    }

    public int getTotalDigits() {
        return -1;
    }

    public int getFractionDigits() {
        return -1;
    }

    public Comparable getMinValue() {
        return null;
    }

    public Comparable getMaxValue() {
        return null;
    }

    public boolean isMinInclusive() {
        return false;
    }

    public boolean isMaxInclusive() {
        return false;
    }

    public List getValues() {
        return null;
    }

    public String toString() {
        return "Datatype " + name;
    }

    public String normalizedValue(String value) {
        return normalize(value, 0, value.length(), nWhiteSpace);
    }

    public StringBuffer normalizedValue(StringBuffer value, StringBuffer out) {
        if (out == null)
            out = new StringBuffer();
        return normalize(value, 0, value.length(), nWhiteSpace, out);
    }

    /**
     * This method modifies its input.
     */
    public StringBuffer normalizedValue(StringBuffer value) {
        return normalize(value, 0, value.length(), nWhiteSpace, value);
    }

    public String toXMLString(Object data, ValidationContextProvider context) {
        if (data == null) return null;
        else return data.toString();
    }

    public Object convert(String string, boolean normalize, ValidationContextProvider context) throws SchemaException {
    	String value = string;
    	if (normalize) value = normalize(string, 0, string.length(), nWhiteSpace);
        return convert(value, context);
    }
    
    public Object convertObject(Object data) throws SchemaException {
    	if (data == null) return null;
    	Object value = null;
    	try {
			value = toValidType(data);
    	} catch (ClassCastException e) {
    		throw new SchemaException(e);
    	}
		checkFacets(value);
		return value;
    }

    protected abstract Object toValidType(Object data);

	/**
	 * This method modifies its input.
	 * @param string
	 * @param normalize
	 * @param context
	 * @return
	 * @throws SchemaException
	 */
    public Object convert(StringBuffer string, boolean normalize, ValidationContextProvider context)
        throws SchemaException {
        if (normalize) normalize(string, 0, string.length(), nWhiteSpace, string);
        return convert(string.toString(), context);
    }

    public Object convert(String string, boolean normalize) throws SchemaException {
        return convert(string, normalize, null);
    }

    public Object convert(StringBuffer string, boolean normalize) throws SchemaException {
        return convert(string, normalize, null);
    }

    protected void setPattern(String value) throws SchemaException {
        try {
            RegularExpression pattern = new RegularExpression(value, "X");
            if (patterns == null)
                patterns = new ArrayList();
            patterns.add(pattern);
        } catch (Exception e) {
            //    "Illegal pattern";
            throw new SchemaException("unknown-error-code.2");
        }
    }

    public void checkPattern(String value) throws SchemaException {
        if (patterns == null)
            return;
        Iterator it = patterns.iterator();
        while (it.hasNext()) {
            RegularExpression pattern = (RegularExpression) it.next();
            if (!pattern.matches(value))
                invalidFacet("cvc-pattern-valid.1", pattern.getPattern(), value);
        }
    }

    public abstract void checkFacets(Object valueSpace) throws SchemaException;
    abstract protected Object toValueSpace(String value, ValidationContextProvider context) throws SchemaException;

    protected Object convert(String value, ValidationContextProvider context) throws SchemaException {
        checkPattern(value);
        Object result = toValueSpace(value, context);
        checkFacets(result);
        return result;
    }

    static public PrimitiveType createType(String name) {
        if (name.equals("anySimpleType"))
            return new AnySimpleType();
        else if (name.equals("string"))
            return new StringType();
        else if (name.equals("boolean"))
            return new BooleanType();
        else if (name.equals("decimal"))
            return new DecimalType();
        else if (name.equals("float"))
            return new FloatType();
        else if (name.equals("double"))
            return new DoubleType();
        else if (name.equals("duration"))
            return new DurationType();
        else if (name.equals("dateTime"))
            return new DateTimeType();
        else if (name.equals("time"))
            return new TimeType();
        else if (name.equals("date"))
            return new DateType();
        else if (name.equals("gYearMonth"))
            return new GYearMonthType();
        else if (name.equals("gYear"))
            return new GYearType();
        else if (name.equals("gMonthDay"))
            return new GMonthDayType();
        else if (name.equals("gDay"))
            return new GDayType();
        else if (name.equals("gMonth"))
            return new GMonthType();
        else if (name.equals("hexBinary"))
            return new HexBinaryType();
        else if (name.equals("base64Binary"))
            return new Base64BinaryType();
        else if (name.equals("anyURI"))
            return new AnyURIType();
        else if (name.equals("QName"))
            return new QNameType();
        else if (name.equals("NOTATION"))
            return new NOTATIONType();

        else
            return null;
    }

    static public PrimitiveType createListType(PrimitiveType itemPrimitive) {
        return new PrimitiveListType(itemPrimitive);
    }

    static public PrimitiveType createUnionType(ArrayList primitives, String unionName) {
        return new PrimitiveUnionType(primitives, unionName);
    }

    public void setFacet(Facet aFacet, ValidationContextProvider vcProvider) throws SchemaException {
        String facetName = aFacet.getName();
        String value = aFacet.getValue();
        boolean fixed = aFacet.isFixed();

        try {
            if (facetName.equals(LENGTH_TAG))
                setLength(value, fixed);
            else if (facetName.equals(MIN_LENGTH_TAG))
                setMinLength(value, fixed);
            else if (facetName.equals(MAX_LENGTH_TAG))
                setMaxLength(value, fixed);
            else if (facetName.equals(PATTERN_TAG))
                setPattern(value);
            else if (facetName.equals(ENUMERATION_TAG))
                setEnumeration(aFacet.getEnumFacets(), vcProvider);
            else if (facetName.equals(WHITESPACE_TAG))
                setWhiteSpace(value, fixed);
            else if (facetName.equals(MIN_EXCLUSIVE_TAG))
                setMinExclusive(value, fixed);
            else if (facetName.equals(MIN_INCLUSIVE_TAG))
                setMinInclusive(value, fixed);
            else if (facetName.equals(MAX_EXCLUSIVE_TAG))
                setMaxExclusive(value, fixed);
            else if (facetName.equals(MAX_INCLUSIVE_TAG))
                setMaxInclusive(value, fixed);
            else if (facetName.equals(TOTALDIGITS_TAG))
                setTotalDigits(value, fixed);
            else if (facetName.equals(FRACTIONDIGITS_TAG))
                setFractionDigits(value, fixed);
        } catch (SchemaException se) {
            throw new SchemaException(se.getErrorCode(), aFacet, se);
        }
    }

    protected void setLength(String value, boolean fixed) throws SchemaException {
        //  "Facet length not supported in " + this;
        throw new SchemaException("cos-applicable-facets");
    }

    protected void setMinLength(String value, boolean fixed) throws SchemaException {
        //  "Facet length not supported in " + this;
        throw new SchemaException("cos-applicable-facets");
    }

    protected void setMaxLength(String value, boolean fixed) throws SchemaException {
        //  "Facet length not supported in " + this;
        throw new SchemaException("cos-applicable-facets");
    }

    protected void setEnumeration(HashSet value, ValidationContextProvider vcProvider) throws SchemaException {
        //  "Facet length not supported in " + this;
        throw new SchemaException("cos-applicable-facets");
    }

    protected void setMinExclusive(String value, boolean fixed) throws SchemaException {
        //  "Facet length not supported in " + this;
        throw new SchemaException("cos-applicable-facets");
    }

    protected void setMinInclusive(String value, boolean fixed) throws SchemaException {
        //  "Facet length not supported in " + this;
        throw new SchemaException("cos-applicable-facets");
    }

    protected void setMaxExclusive(String value, boolean fixed) throws SchemaException {
        //  "Facet length not supported in " + this;
        throw new SchemaException("cos-applicable-facets");
    }

    protected void setMaxInclusive(String value, boolean fixed) throws SchemaException {
        //  "Facet length not supported in " + this;
        throw new SchemaException("cos-applicable-facets");
    }

    protected void setTotalDigits(String value, boolean fixed) throws SchemaException {
        //  "Facet length not supported in " + this;
        throw new SchemaException("cos-applicable-facets");
    }

    protected void setFractionDigits(String value, boolean fixed) throws SchemaException {
        //  "Facet length not supported in " + this;
        throw new SchemaException("cos-applicable-facets");
    }

    protected void setWhiteSpace(String value, boolean fixed) throws SchemaException {
        int ws = -1;
        if ("preserve".equals(value))
            ws = PRESERVE;
        else if ("replace".equals(value))
            ws = REPLACE;
        else if ("collapse".equals(value))
            ws = COLLAPSE;

        // check that value is legal (ws != -1) and  more constrained
        if ((isFixed(WHITESPACE_FACET) && ws != nWhiteSpace)) {
            //    "facet value is fixed";
            throw new SchemaException("unknown-error-code.1");
        }

        if ((ws == PRESERVE || ws == REPLACE) && nWhiteSpace == COLLAPSE) {
            //    "whiteSpace value can't be <replace> or <preserve>";
            throw new SchemaException("whiteSpace-valid-restriction.1");
        }

        if (ws == PRESERVE && nWhiteSpace == REPLACE) {
            //    "whiteSpace value can't be <preserve>";
            throw new SchemaException("whiteSpace-valid-restriction.2");
        }

        nWhiteSpace = ws;
        setFixed(WHITESPACE_FACET, fixed);
    }

    public int getWhiteSpace() {
        return nWhiteSpace;
    }

    protected boolean isFixed(int facet) {
        int mask = 1 << facet;
        return (mask & fixedBitSet) != 0;
    }

    protected void setFixed(int facet, boolean fixed) {
        if (fixed) {
            int mask = 1 << facet;
            fixedBitSet |= mask;
        }
    }

    protected void invalidFacet(String errCode, Object facetVal, String value) throws SchemaException {
        Object invalidFacet = "Constraint: " + facetVal + ", value: " + value;
        SchemaException se = new SchemaException(errCode, invalidFacet);
        if ("cvc-pattern-valid".equals(errCode))
            throw new SchemaException("cvc-datatype-valid.1.1", this, se);
        else
            throw new SchemaException("cvc-datatype-valid.2", this, se);
    }

    protected void invalidValue(String value) throws SchemaException {
        throw new SchemaException("cvc-datatype-valid.1.2.1", "value = " + value);
    }

    public static boolean isXMLWhitespace(char c) {
        switch (c) {
            case 0x20 :
            case 0x9 :
            case 0xA :
            case 0xD :
                return true;
            default :
                return false;
        }
    }

    private static String normalize(String value, int offset, int length, int whitespace) {
        StringBuffer result;
        switch (whitespace) {
            case PRESERVE :
                if (offset == 0 && length == value.length())
                    return value;
                else
                    return value.substring(offset, length);
            case REPLACE :
                result = new StringBuffer();
                for (int i = offset; i < offset + length; i++) {
                    if (isXMLWhitespace(value.charAt(i)))
                        result.append((char) 0x20);
                    else
                        result.append(value.charAt(i));
                }
                return result.toString();
            case COLLAPSE :
                result = new StringBuffer();
                boolean start = true;
                int nIndex = 0;
                for (int i = offset; i < offset + length; i++) {
                    if (isXMLWhitespace(value.charAt(i))) {
                        // if leading WS, do nothing
                        if (!start) {
                            // collapse continuous sequences of 0x20 to a single 0x20
                            if (result.charAt(nIndex - 1) != 0x20) {
                                result.append((char) 0x20);
                                nIndex++;
                            }
                        }
                    } else {
                        result.append(value.charAt(i));
                        nIndex++;
                        start = false;
                    }
                }
                // remove trailing 0x20
                if (nIndex > 0 && result.charAt(nIndex - 1) == 0x20)
                    result.setLength(nIndex - 1);
                return result.toString();
			default :
				throw new IllegalArgumentException("Illegal value for whitespace handling");
        }
    }

    private static StringBuffer normalize(
        StringBuffer value,
        int offset,
        int length,
        int whitespace,
        StringBuffer result) {
        int nIndex = offset;
        switch (whitespace) {
            case PRESERVE :
                if (result != value) {
                    if (offset == 0 && length == value.length()) {
                        result.append(value);
                    } else {
                        result.append(value.substring(offset, length));
                    }
                }
                return result;
            case REPLACE :
                if (result != value) {
                    nIndex = result.length();
                    result.setLength(nIndex + length);
                }
                for (int i = offset; i < offset + length; i++) {
                    if (isXMLWhitespace(value.charAt(i)))
                        result.setCharAt(nIndex, (char) 0x20);
                    else
                        result.setCharAt(nIndex, value.charAt(i));
                    nIndex++;
                }
                return result;
            case COLLAPSE :
                if (result != value) {
                    nIndex = result.length();
                    result.setLength(nIndex + length);
                }
                boolean start = true;
                for (int i = offset; i < offset + length; i++) {
                    if (isXMLWhitespace(value.charAt(i))) {
                        // if leading WS, do nothing
                        if (!start) {
                            // collapse continuous sequences of 0x20 to a single 0x20
                            if (result.charAt(nIndex - 1) != 0x20) {
                                result.setCharAt(nIndex, (char) 0x20);
                                nIndex++;
                            }
                        }
                    } else {
                        result.setCharAt(nIndex, value.charAt(i));
                        nIndex++;
                        start = false;
                    }
                }
                // remove trailing 0x20
                if (!start && result.charAt(nIndex - 1) == 0x20)
                    nIndex--;
                if (result != value)
                    result.setLength(nIndex);
                else
                    result.delete(nIndex, offset + length);
                return result;
			default :
				throw new IllegalArgumentException("Illegal value for whitespace handling");
        }
    }

}
