/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.schema.datatypes;

import java.io.Serializable;

/**********************************************************************
* This class represents an absolute or relative URI, as specified in
* IETF RFC 2396. 
**********************************************************************/

public class URI implements Serializable {
    private static final String RCSRevision = "$Revision: 1.2 $";
    private static final String RCSName = "$Name: XQ1_1_1 $";

    /*******************************************************************
    * MalformedURIExceptions are checked exceptions thrown when the
    * URI string does not conform to the specification
    *
    ********************************************************************/
    public static class MalformedURIException extends Exception {
        private static final String RCSRevision = "$Revision: 1.2 $";
        private static final String RCSName = "$Name: XQ1_1_1 $";

        /******************************************************************
         * Constructs a MalformedURIException with no specified message.
         ******************************************************************/
        public MalformedURIException() {
            super();
        }

        /*****************************************************************
        * Constructs a MalformedURIException with the specified message. 
        *
        * @param msg the message.
        ******************************************************************/
        public MalformedURIException(String msg) {
            super(msg);
        }
    }

    /** reserved characters */
    private static final String RESERVED_CHARACTERS = ";/?:@&=+$,";

    /** URI punctuation mark characters. Unreserved characters are 
     * the alphanumerics and these characters */
    private static final String MARK_CHARACTERS = "-_.!~*'()";

    /** scheme must start with an alphabetic and can be composed of 
     * alphanumerics and these characters */
    private static final String SCHEME_CHARACTERS = "+-.";

    private String scheme = null;
    private String authority = null;
    private String path = "";
    private String query = null;
    private String fragment = null;

    /**
     * Construct a new URI from a URI string. Three types of URI are recognized: 
     * 1. Absolute URIs with hierarchical part (generic URIs): those URIs will
     *    be decomposed into all possible components (scheme, authority, path, query).
     * 2. Absolute URIs with opaque part: only the scheme and path (holding the
     *    opaque part) components are defined.
     * 3. Relative URIs, which feature only a subset of the generic URIs components
     * 
     * In addition, a fragment component can appear after all types of URI.
     * 
     * @param uri the URI specification string 
     * @exception MalformedURIException if uri is not a valid URI specification
     */
    public URI(String uri) throws MalformedURIException {
        if (uri == null)
            throw new MalformedURIException("uri parameter cannot be empty.");
        uri = uri.trim();
        if (uri.length() == 0)
            throw new MalformedURIException("uri parameter cannot be empty.");
        parseURI(uri);
    }

    /**
     * Construct a new URI from a base URI and a URI string.
     * The URI string may be a relative URI. 
     *
     * @param base the base URI 
     * @param uri the URI specification string
     * @exception MalformedURIException if uri is not a valid URI specification
     */
    public URI(URI base, String uri) throws MalformedURIException {
        if (uri == null || uri.trim().length() == 0) {
            if (base != null) {
                copy(base);
                return;
            } else {
                throw new MalformedURIException("base and uri parameters cannot be both empty.");
            }
        }
        parseURI(uri);
        if (base != null)
            resolve(base);
    }

    /**
     * Copy constructor
     * 
     * @param orig the URI to copy
     */
    public URI(URI orig) {
        copy(orig);
    }

    /**
     * Copy all fields of another URI into this URI.
     *
     * @param orig the URI to copy
     */
    private void copy(URI orig) {
        scheme = orig.getScheme();
        authority = orig.getAuthority();
        path = orig.getPath();
        query = orig.getQuery();
        fragment = orig.getFragment();
    }

    /**
     * Initializes this URI from a base URI and a URI string.
     * 
     * @param uri the URI string which may be an absolute or
     *                  relative URI
     *
     * @exception MalformedURIException if uri is not a valid URI specification 
     */
    private void parseURI(String uri) throws MalformedURIException {
        int len = uri.length();
        int index = 0;
        int colonIndex = uri.indexOf(':', index);
        int slashIndex = uri.indexOf('/', index);
        if (colonIndex != -1 && (slashIndex == -1 || colonIndex < slashIndex)) {
            // absolute URI
            // get the scheme
            index = parseScheme(uri);
            if (index < len && uri.charAt(index) != '/') {
                // opaque URI: get the path component
                index = parseOpaque(uri, index);
            } else {
                // hierarchical URI
                if (index + 1 < len && uri.charAt(index + 1) == '/') {
                    // authority: get the authority component
                    index = parseAuthority(uri, index + 2);
                }
                // path: get the path component
                index = parseAbsolutePath(uri, index);
                // get the query component of the generic URI
                index = parseQuery(uri, index);
            }
        } else {
            // relative URI (must be hierarchical)
            if (index < len && uri.charAt(index) != '/') {
                // relative segment: get the first path component
                index = parseRelativeSegment(uri, index);
            } else if (index + 1 < len && uri.charAt(index + 1) == '/') {
                // authority: get the authority component
                index = parseAuthority(uri, index + 2);
            }
            // path: get the path component
            index = parseAbsolutePath(uri, index);
            // get the query component of the hierarchical URI
            index = parseQuery(uri, index);
        }
        index = parseFragment(uri, index);
    }

    /**
     * Initialize the scheme for this URI from a URI string spec.
     *
     * @param uri the URI specification
     *  
     * @return the start index for the next part
     * 
     * @exception MalformedURIException if URI does not have a conformant scheme
     */
    private int parseScheme(String uri) throws MalformedURIException {
        int len = uri.length();
        int index = 0;
        if (index >= len)
            throw new MalformedURIException("Scheme cannot be empty in: " + uri + ".");
        if (!isAlpha(uri.charAt(0)))
            throw new MalformedURIException("Illegal character in scheme part of " + uri + " at position 0.");
        index++;
        while (index < len) {
            char schemeChar = uri.charAt(index);
            if (":".indexOf(schemeChar) != -1)
                break;
            if (!isAlphanum(schemeChar) && SCHEME_CHARACTERS.indexOf(schemeChar) == -1)
                throw new MalformedURIException(
                    "Illegal character in scheme part of " + uri + " at position " + index + ".");
            index++;
        }
        scheme = uri.substring(0, index).toLowerCase();
        return index + 1;
    }

    /**
     * Initialize the authority for this URI from a URI string spec.
     * Subdivision of authority is not taken care of, as it is irrelevant for
     * XML Schema
     *
     * @param uri the URI specification 
     * @param start the start index for this part
     *
     * @return the start index for the next part
     * 
     * @exception MalformedURIException if uri is not a valid URI specification 
     */
    private int parseAuthority(String uri, int start) throws MalformedURIException {
        int len = uri.length();
        if (start >= len)
            return start;
        StringBuffer buffer = new StringBuffer(len-start);
        int result = processURIPart(uri, start, "/?#", buffer);
        authority = buffer.toString();
        return result;
    }

    /**
     * Initialize the opaque part for this URI from a URI string spec.
     * The opaque part is stored as the path component
     *
     * @param uri the URI specification
     * @param start the start index for this part
     *
     * @return the start index for the next part
     * 
     * @exception MalformedURIException if uri is not a valid URI specification
     */
    private int parseOpaque(String uri, int start) throws MalformedURIException {
		int len = uri.length();
		StringBuffer buffer = new StringBuffer(len-start);
		int result = processURIPart(uri, start, "#", buffer);
		path = buffer.toString();
		return result;
    }

    /**
     * Initialize the path for this URI from a URI string spec.
     *
     * @param uri the URI specification
     * @param start the start index for this part
     *
     * @return the start index for the next part
     * 
     * @exception MalformedURIException if uri is not a valid URI specification
     */
    private int parseAbsolutePath(String uri, int start) throws MalformedURIException {
		int len = uri.length();
		if (start >= len) {
			return start;
		}
		StringBuffer buffer = new StringBuffer(len-start);
		int result = processURIPart(uri, start, "?#", buffer);
		path += buffer.toString();
		return result;
    }

    /**
     * Initialize the first path segment for this relative URI from a URI string spec.
     *
     * @param uri the URI specification
     * @param start the start index for this part
     *
     * @return the start index for the next part
     * 
     * @exception MalformedURIException if uri is not a valid URI specification
     */
    private int parseRelativeSegment(String uri, int start) throws MalformedURIException {
        int len = uri.length();
        if (start >= len)
            throw new MalformedURIException("Relative path cannot be empty in: " + uri + ".");
		StringBuffer buffer = new StringBuffer(len-start);
		int result = processURIPart(uri, start, "/?#", buffer);
		path = buffer.toString();
		return result;
    }

    /**
     * Initialize the query part for this URI from a URI string spec.
     *
     * @param uri the URI specification
     * @param start the start index for this part
     *
     * @return the start index for the next part
     * 
     * @exception MalformedURIException if uri is not a valid URI specification
     */
    private int parseQuery(String uri, int start) throws MalformedURIException {
        int len = uri.length();
        if (start >= len)
            return start;
        if (uri.charAt(start) != '?')
            return start;
		StringBuffer buffer = new StringBuffer(len-start-1);
		int result = processURIPart(uri, start+1, "#", buffer);
		query = buffer.toString();
		return result;
    }

    /**
     * Initialize the fragment part for this URI from a URI string spec.
     *
     * @param uri the URI specification
     * @param start the start index for this part
     *
     * @return the start index for the next part
     * 
     * @exception MalformedURIException if uri is not a valid URI specification
     */
    private int parseFragment(String uri, int start) throws MalformedURIException {
		int len = uri.length();
		if (start >= len)
			return start;
		if (uri.charAt(start) != '#')
			return start;
		StringBuffer buffer = new StringBuffer(len-start-1);
		int result = processURIPart(uri, start+1, "", buffer);
		query = buffer.toString();
		return result;
    }
    
    /**
     * Fill the result StringBuffer by processing the specified URI part.
     * @param uri the URI specification string
     * @param start the start index in the specification string
     * @param stopChars the set of characters which denote the end of the processed part
     * @param result the buffer containing the part characters, after unescaping
     * @return the start index for the next part
     */
    private int processURIPart(String uri, int start, String stopChars, StringBuffer result) throws MalformedURIException {
		int len = uri.length();
		int index = start;
		if (index >= len)
			return index;
		while (index < len) {
			char c = uri.charAt(index);
			if (stopChars.indexOf(c) != -1)
				break;
			if (!isUnreservedCharacter(c) && !isReservedCharacter(c)) {
				if (c == '%' && index + 2 < len && isHex(uri.charAt(index + 1)) && isHex(uri.charAt(index + 2))) {
					result.append((char)Integer.parseInt(uri.substring(index+1, index+3), 16));
					index += 3;
				} else {
					throw new MalformedURIException("Illegal character in " + uri + " at position " + index + ".");
				}
			} else {
				result.append(c);
				index++;
			}
		}
		return index;
    }

	/**
	 * Append the provided URI part into the result buffer, escaping the characters when required 
	 * @param part the URI part
	 * @param result the buffer containing the escaped characters
	 */
	private void escapePart(String part, StringBuffer result) {
		for (int i = 0; i < part.length(); i++) {
			char c = part.charAt(i);
			if (!isUnreservedCharacter(c) && !isReservedCharacter(c)) {
				result.append('%');
				result.append(Integer.toHexString((int)c));
			} else {
				result.append(c);
			}
		}
	}
	
    /**
     * Transform this relative URI into an absolute URI, 
     * following the RFC 2396 spec.
     * @param base the base URI to resolve to
     */
    private void resolve(URI base) throws MalformedURIException {
        if (!base.isHierarchical())
            throw new MalformedURIException("Cannot resolve a relative URI to a non-hierarchical base");
        if (base.isRelative())
            throw new MalformedURIException("Cannot resolve a relative URI to a relative base");
        if (scheme != null)
            return;
        setScheme(base.getScheme());
        if (authority != null)
            return;
        setAuthority(base.getAuthority());
        if (path.length() == 0) {
            setPath(base.getPath());
        } else if (path.charAt(0) == '/') {
            return;
        } else {
            StringBuffer absPath = new StringBuffer(base.getPath());
            int lastSlash = base.getPath().lastIndexOf('/');
            if (lastSlash != -1) {
                absPath.setLength(lastSlash + 1);
            }
            absPath.append(path);
            String rawPath = absPath.toString();
            int index = -1;
            int delta = 0;
            while ((index = rawPath.indexOf("/./", index + 1)) != -1) {
                absPath.delete(index + 1 - delta, index + 3 - delta);
                delta += 2;
            }
            if (rawPath.endsWith("/.")) {
                absPath.setLength(absPath.length() - 1);
                delta += 1;
            }
            if (delta > 0) {
                rawPath = absPath.toString();
            }
            while ((index = rawPath.indexOf("/../")) > 0) {
                int prevIndex = rawPath.lastIndexOf('/', index - 1);
                if (prevIndex != -1) {
                    rawPath = rawPath.substring(0, prevIndex+1).concat(rawPath.substring(index+4));
	           }
            }
            if (rawPath.endsWith("/..")) {
                int prevIndex = rawPath.lastIndexOf('/', rawPath.length() - 4);
                if (prevIndex != -1 && !"..".equals(rawPath.substring(prevIndex + 1, prevIndex + 3))) {
                    rawPath = rawPath.substring(0, prevIndex+1);
                }
            }
            path = rawPath;
        }
    }

    /**
     * Get the scheme for this URI.
     *
     * @return the scheme for this URI
     */
    public String getScheme() {
        return scheme;
    }

    /**
    * Get the authority for this URI.
    *
    * @return the authority for this URI (null if not specified).
    */
    public String getAuthority() {
        return authority;
    }

    /**
     * Get the path for this URI. Note that the value returned is the path
     * only and does not include the query string or fragment.
     *
     * @return the path for this URI.
     */
    public String getPath() {
        return path;
    }

    /**
     * Get the query string for this URI.
     *
     * @return the query string for this URI. Null is returned if there
     *         was no "?" in the URI spec, empty string if there was a 
     *         "?" but no query string following it.
     */
    public String getQuery() {
        return query;
    }

    /**
     * Get the fragment for this URI.
     *
     * @return the fragment for this URI. Null is returned if there
     *         was no "#" in the URI spec, empty string if there was a 
     *         "#" but no fragment following it.
     */
    public String getFragment() {
        return fragment;
    }

    /**
     * Set the scheme for this URI.
     * 
     * @param scheme the new scheme value 
     */
    private void setScheme(String scheme) {
        this.scheme = scheme;
    }

    /**
    * Set the authority for this URI. 
    *
    * @param auth the new authority value
    */
    private void setAuthority(String auth) {
        this.authority = auth;
    }

    /**
     * Set the path for this URI. 
     * 
     * @param path the new path value
     */
    private void setPath(String path) {
        this.path = path;
    }

    /**
     * Set the query part for this URI. 
     *
     * @param query the new query value
     *
     */
    public void setQuery(String query) {
        this.query = query;
    }

    /**
     * Set the fragment for this URI. 
     *
     * @param fragment the new fragment value
     *
     */
    public void setFragment(String fragment) throws MalformedURIException {
        this.fragment = fragment;
    }

    /**
     * Compare an object to this URI.
     *
     * @param obj the Object to test for equality.
     *
     * @return true if obj is a URI with all values equal to this
     *         URI, false otherwise
     */
    public boolean equals(Object obj) {
    	if (this == obj) return true;
        if (obj instanceof URI) {
            URI testURI = (URI) obj;
            if (((scheme == null && testURI.scheme == null)
                || (scheme != null && testURI.scheme != null && scheme.equals(testURI.scheme)))
                && ((authority == null && testURI.authority == null)
                    || (authority != null && testURI.authority != null && authority.equals(testURI.authority)))
                && ((path == null && testURI.path == null)
                    || (path != null && testURI.path != null && path.equals(testURI.path)))
                && ((query == null && testURI.query == null)
                    || (query != null && testURI.query != null && query.equals(testURI.query)))
                && ((fragment == null && testURI.fragment == null)
                    || (fragment != null && testURI.fragment != null && fragment.equals(testURI.fragment)))) {
                return true;
            }
        }
        return false;
    }

    /**
     * Computes the hashcode value for this URI, based on the hashcode of its components
     *
     * @return the hashcode value
     */
    public int hashCode() {
        return (scheme == null ? 0 : scheme.hashCode())
            ^ (authority == null ? 0 : authority.hashCode())
            ^ (path == null ? 0 : path.hashCode())
            ^ (query == null ? 0 : query.hashCode())
            ^ (fragment == null ? 0 : fragment.hashCode());
    }

    /**
     * Get the URI as a string specification. See RFC 2396 Section 5.2.
     *
     * @return the URI string specification
     */
    public String toString() {
        StringBuffer uriSpecString = new StringBuffer();
        if (scheme != null) {
            uriSpecString.append(scheme);
            uriSpecString.append(':');
        }
        if (authority != null) {
        	uriSpecString.append("//");
        	escapePart(authority, uriSpecString);
        }
        escapePart(path, uriSpecString);
        if (query != null) {
        	uriSpecString.append('?');
			escapePart(query, uriSpecString);
        }
		if (fragment != null) {
			uriSpecString.append('#');
			escapePart(fragment, uriSpecString);
		}
        return uriSpecString.toString();
    }

	/**
	 * Get the URI length.
	 *
	 * @return the URI length
	 */
	public int getLength() {
		int len = 0;
		if (scheme != null) {
			len += scheme.length()+1;
		}
		if (authority != null) {
			len += authority.length()+2;
		}
		len += path.length();
		if (query != null) {
			len += query.length()+1;
		}
		if (fragment != null) {
			len += fragment.length()+1;
		}
		return len;
	}

    /**
     * Indicate whether this URI uses the generic hierarchical syntax
     * 
     * @return true if this URI uses the hierarchical URI syntax
     */
    public boolean isHierarchical() {
        return (scheme == null || path.charAt(0) == '/');
    }

    /**
     * Indicate whether this URI is relative, i.e. has no specified scheme
     * 
     * @return true if this URI is relative
     */
    public boolean isRelative() {
        return (scheme == null);
    }

    /**
     * Determine whether a char is a digit.
     *
     * @return true if the char is betweeen '0' and '9', false otherwise
     */
    private static boolean isDigit(char p_char) {
        return p_char >= '0' && p_char <= '9';
    }

    /**
     * Determine whether a character is a hexadecimal character.
     *
     * @return true if the char is betweeen '0' and '9', 'a' and 'f'
     *         or 'A' and 'F', false otherwise
     */
    private static boolean isHex(char p_char) {
        return (isDigit(p_char) || (p_char >= 'a' && p_char <= 'f') || (p_char >= 'A' && p_char <= 'F'));
    }

    /**
     * Determine whether a char is an alphabetic character: a-z or A-Z
     *
     * @return true if the char is alphabetic, false otherwise
     */
    private static boolean isAlpha(char p_char) {
        return ((p_char >= 'a' && p_char <= 'z') || (p_char >= 'A' && p_char <= 'Z'));
    }

    /**
     * Determine whether a char is an alphanumeric: 0-9, a-z or A-Z
     *
     * @return true if the char is alphanumeric, false otherwise
     */
    private static boolean isAlphanum(char p_char) {
        return (isAlpha(p_char) || isDigit(p_char));
    }

    /**
     * Determine whether a character is a reserved character:
     * ';', '/', '?', ':', '@', '&', '=', '+', '$' or ',' 
     *
     * @return true if the string contains any reserved characters
     */
    private static boolean isReservedCharacter(char p_char) {
        return RESERVED_CHARACTERS.indexOf(p_char) != -1;
    }

    /**
     * Determine whether a char is an unreserved character.
     *
     * @return true if the char is unreserved, false otherwise
     */
    private static boolean isUnreservedCharacter(char p_char) {
        return (isAlphanum(p_char) || MARK_CHARACTERS.indexOf(p_char) != -1);
    }

}
