/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.schema.loader;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xquark.schema.*;
import org.xquark.util.DefaultElementHandler;
import org.xquark.util.ElementHandler;

class AttributeGroupHandler extends DefaultElementHandler implements SchemaConstants
{
    private static final String RCSRevision = "$Revision: 1.2 $"; 
    private static final String RCSName     = "$Name: XQ1_1_1 $";
    
    private Loader loader = null;
    private AttributeGroupDefinition def;
    private boolean redefined;
    private AttributeGroupDefinition oldGroup;

	AttributeGroupHandler(Loader loader, AttributeGroupDefinition def) {
		this.loader = loader;
        this.def = def;
		this.redefined = false;
	}
    
	AttributeGroupHandler(Loader loader, AttributeGroupDefinition def, AttributeGroupDefinition oldGroup) {
        this(loader, def);
        this.redefined = true;
        this.oldGroup = oldGroup;
    }

	public ElementHandler startElement(String namespaceURI, String localName, Attributes atts) 
    throws SAXException 
    {
		if ( namespaceURI.equals(XMLSCHEMA_URI) ) {
        if (localName.equals(ATTRIBUTE_GROUP_TAG)) {
          // redefined attribute group definition : special treating
          String refName = atts.getValue("", REF_ATTR);
          // refName == null is verificated by Schema of Schemas
          if ( redefined && refName.equals(def.getName()) ) {
            // self reference
              def.add(oldGroup);
              redefined = false;
          }
          else {
            SchemaComponent ref = null;
            try {
               ref = loader.getRef(refName, Schema.ATTRIBUTE_GROUP_DEFINITION);
            } 
            catch ( SchemaException se ) {
              String errMsg = "Error while processing localName -> " + localName; 
              loader.reportLoadingError(errMsg, se);
            }
            def.add(ref);
          }
          return this;
        } 

        else if (localName.equals(ATTRIBUTE_TAG)) {
          AttributeDeclaration decl = null;
          try {
            decl = loader.buildAttributeDeclaration(atts, null);
          } 
          catch ( SchemaException se ) {
            String errMsg = "Error while processing localName -> " + localName; 
            loader.reportLoadingError(errMsg, se);
          }
          if ( redefined ) {
              try {
                checkAttributeValidRestriction(decl);
              }
              catch ( SchemaException e) {
                String errMsg = "Error while processing localName -> " + localName; 
                loader.reportLoadingError(errMsg, e);
              }
          }
          if ( def.attributeFind(decl) ) {
            String errCode = "ag-props-correct.2";
            String errMsg = "Error while processing localName -> " + localName; 
            loader.reportLoadingError(errMsg, new SchemaException(errCode));
          }
          def.add(decl);
          return new AttributeTypeHandler(loader, decl);
        } 

        else if (localName.equals(ANY_ATTRIBUTE_TAG)) {
          Wildcard wildcard = null;
          try {
            wildcard = loader.buildWildcard(atts);
          } 
          catch ( SchemaException se ) {
            String errMsg = "Error while processing localName -> " + localName; 
            loader.reportLoadingError(errMsg, se);
          }
          if ( redefined ) {
              try {
                checkAttributeValidRestriction(wildcard);
              }
              catch ( SchemaException e) {
                String errMsg = "Error while processing localName -> " + localName; 
                loader.reportLoadingError(errMsg, e);
              }
          }
          def.setAttributeWildcard(wildcard);
          return this;
        } 

        else if (localName.equals(ANNOTATION_TAG)) {
          return new AnnotationHandler();
        }
		}
		
		loader.notifyUnknownElement(namespaceURI, localName);
		return this;
	}
    
    private void checkAttributeValidRestriction(AttributeDeclaration decl) 
    throws SchemaException {
        java.util.Iterator it = this.oldGroup.getAttributeDeclarations().iterator();
        AttributeDeclaration baseDecl = null;
        while ( it.hasNext() ) {
            AttributeDeclaration attr = (AttributeDeclaration)it.next();
            if ( decl.getName().equals(attr.getName()) &&
                 ( ( decl.getNamespace() == null && attr.getNamespace() == null ) ||
                   ( decl.getNamespace().equals(attr.getNamespace()) ) ) ) {
               baseDecl = attr;
               break;
            }
        }
        
        if ( baseDecl != null ) {
            if ( !(baseDecl.getUse() != REQUIRED || decl.getUse() == REQUIRED) ) {
            //        "Attribute <use> isn't reqired, but attribute <use> in base type is required.";
                throw new SchemaException("derivation-ok-restriction.2.1.1", decl);
            }
            String errCode02 = decl.getType().checkTypeDerivationOK(baseDecl.getType(), baseDecl.getType().getFinal(), false);
            if ( errCode02 != null ) {
            //        "Attribute's type definition must be validly derivded from that of the base type's attribute declaration.";
                throw new SchemaException("derivation-ok-restriction.2.1.2", decl, new SchemaException(errCode02, decl));
            }
            if ( !( baseDecl.getFixedValue() == null || baseDecl.getFixedValue().equals(decl.getFixedValue()) ) ) {
            //        "Attribute declaration's effective value constraint doesn't match with that of base type's attribute declaration.";
                throw new SchemaException("derivation-ok-restriction.2.1.3", decl);
            }
        }
        else if ( this.oldGroup.getAttributeWildcard() == null ||
                  this.oldGroup.getAttributeWildcard().isAllowed(decl.getNamespace()) == false ) 
        {
        //      "Attribute declaration, or a valid attribute wildcard, must be defined in base type.";
            throw new SchemaException("derivation-ok-restriction.2.2", decl);
        }       
    }
    
    private void checkAttributeValidRestriction(Wildcard wild)
    throws SchemaException {
      Wildcard baseW = this.oldGroup.getAttributeWildcard();
      if ( baseW == null ) {
//      "Base type must have an attribute wildcard.";
        throw new SchemaException("derivation-ok-restriction.4.1", this);
      }
      else {
        String errCode02 = wild.validWildcardSubset(baseW);
        if ( errCode02 != null ) {
//        "Attribute wildcard must be a valid subset of base type's.";
          throw new SchemaException("derivation-ok-restriction.4.2", this, new SchemaException(errCode02, wild));
        }
      }
    }
}
	
