/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.schema.loader;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xquark.schema.*;
import org.xquark.util.DefaultElementHandler;
import org.xquark.util.ElementHandler;

class ComplexTypeHandler extends DefaultElementHandler implements SchemaConstants
{
private static final String RCSRevision = "$Revision: 1.2 $"; 
private static final String RCSName     = "$Name: XQ1_1_1 $";
	private Loader loader = null;
	private ComplexType type = null;
  private boolean complexContent;
  private boolean redefine;

	ComplexTypeHandler(Loader loader, ComplexType type, boolean complexContent, boolean redefine) 
    throws SAXException 
  {
			this.loader = loader;
			this.type = type;
      this.complexContent = complexContent;
      this.redefine = redefine;
	}
	
	public ElementHandler startElement(String namespaceURI, String localName, Attributes atts) 
	throws SAXException 
	{    
		if ( namespaceURI.equals(XMLSCHEMA_URI) ) {
        // in Redefinition, complexContent is always false
        if ( complexContent ) {         
          if ( localName.equals(RESTRICTION_TAG) || localName.equals(EXTENSION_TAG) ) {
                Type baseType = null;
                try {
                  baseType = loader.getType(atts.getValue("", BASE_ATTR));
                } 
                catch ( SchemaException se ) {
                  String errMsg = "Error while processing localName -> " + localName; 
                  loader.reportLoadingError(errMsg, se);
                }
                if ( redefine ) {
                  redefine = false;
                  if ( type.getName() == null || baseType == null || 
                       !type.getName().equals(baseType.getName()) ) {
                      String errCode = "src-redefine.5";
                      String errMsg = "Error while processing localName -> " + localName; 
                      loader.reportLoadingError(errMsg, new SchemaException(errCode));
                  }
                }
                else
                    type.setBaseType(baseType);
              
            // extention or restriction ?
            if ( localName.equals(RESTRICTION_TAG) ) type.setExtension(false);
            else type.setExtension(true);          

            return this;
          }
        }
        else {
          // case 1 : simpleContent
          if ( localName.equals(SIMPLECONTENT_TAG) ) {
            if ( !redefine ) {
              type.setContentModel(new ContentModel(TEXT_ONLY));
            }
            return new SimpleTypeHandler(loader, type, true, redefine);
          }

          // case 2 : complexContent
          if ( localName.equals(COMPLEXCONTENT_TAG) ) { 
            String mixed = atts.getValue("", MIXED_VALUE);
            if ( TRUE_VALUE.equals(mixed)) {
              type.setContentModel(new ContentModel(MIXED));
            }
            return new ComplexTypeHandler(loader, type, true, redefine);
          }
        }

        // redefined complexType can't have the following tags
        if ( redefine ) {
          String errCode = "src-redefine.5";
          String errMsg = "Error while processing localName -> " + localName; 
          loader.reportLoadingError(errMsg, new SchemaException(errCode));
        }

        // case 3 : sequence | choice | all | group
        if (localName.equals(GROUP_TAG) || localName.equals(SEQUENCE_TAG) 
            || localName.equals(CHOICE_TAG) || localName.equals(ALL_TAG)) {
          if ( !complexContent ) type.setExtension(false);
          Particle particle = null;
          try {
            particle = loader.buildParticle(localName, type, atts);
          } 
          catch ( SchemaException se ) {
            String errMsg = "Error while processing localName -> " + localName; 
            loader.reportLoadingError(errMsg, se);
          }
          if (type.getContentModel() == null) {
            type.setContentModel(new ContentModel(particle));
          } 
          else if (type.getContentModel().getModel() == null) {
            type.getContentModel().setModel(particle);
          } 
          Object term = particle.getTerm();
          if (term instanceof ModelGroup) {
            return new ContentModelHandler(loader, (ModelGroup)term, type);
          }
          else return this;
        }

        // case 4 : attribute
        if ( localName.equals(ATTRIBUTE_TAG) ) {
          if ( !complexContent ) type.setExtension(false);
          AttributeDeclaration decl = null;
          try {
            decl = loader.buildAttributeDeclaration(atts, type);
            type.register(decl);
          } catch ( SchemaException e ) {
            String errMsg = "Error while processing localName -> " + localName; 
            loader.reportLoadingError(errMsg, e);
          }
          return new AttributeTypeHandler(loader, decl);
        } 

        // case 5 : attribute group
        if ( localName.equals(ATTRIBUTE_GROUP_TAG) ) {
          if ( !complexContent ) type.setExtension(false);
          String refName = atts.getValue("", REF_ATTR); // refName is verified by Schema of Schemas
          SchemaComponent ref = null;
          try {
            ref = loader.getRef(refName, Schema.ATTRIBUTE_GROUP_DEFINITION);
            type.register(ref);
          } catch ( SchemaException e ) {
            String errMsg = "Error while processing localName -> " + localName; 
            loader.reportLoadingError(errMsg, e);
          }
          return this;
        } 

        // case 6 : anyAttribute
        if (localName.equals(ANY_ATTRIBUTE_TAG)) {
          if ( !complexContent ) type.setExtension(false);
          Wildcard wildcard = null;
          try {
            wildcard = loader.buildWildcard(atts);
          } catch ( SchemaException e ) {
            String errMsg = "Error while processing localName -> " + localName; 
            loader.reportLoadingError(errMsg, e);
          }
          type.setAttributeWildcard(wildcard);
          return this;
        } 

        // case 7 : annotation
        if ( localName.equals(ANNOTATION_TAG) ) return new AnnotationHandler();			
		}			
		loader.notifyUnknownElement(namespaceURI, localName);
		return this;
	}
}
