/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.util;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.xml.sax.Attributes;
import org.xml.sax.ContentHandler;
import org.xml.sax.ext.LexicalHandler;
import org.xml.sax.helpers.DefaultHandler;

/**
 * Define a SAX ContentHandler or/and LexicalHandler writing a DOM-tree inside a given document.
 * Such writing doesn't erase previous content but append to it.
 * So, if the document is not empty, this handler will try to add nodes anyway, and
 * probably cause a hierarchy error due to try to add root node or doc type in addition
 * of a previous one... <BR>
 * <BR>
 * WARNING: if this class is used for two differents types of handler to build a DOMTree from
 * the same parsing, the parser must reference the SAME DOMBuilder object for the two types of
 * handlers unless the DOM-Tree will not correspond to the waited DOM-Tree.
 *
 * IMPORTANT: SAX2 XMLReader
 * <code>http://xml.org/sax/features/namespace-prefixes</code> and
 * <code>http://xml.org/sax/features/namespace-prefixes</code> features must be
 * set to true.
 *
 * @see org.xml.sax.ContentHandler, org.xml.sax.LexicalHandler, org.w3c.dom.Document
 */
public class BasicDOMBuilder extends DefaultHandler
implements ContentHandler, LexicalHandler
{
    private static final String RCSRevision = "$Revision: 1.2 $";
    private static final String RCSName     = "$Name: XQ1_1_1 $";

    // ***********************************************************************
    // * Fields
    // ***********************************************************************
    
    /**
     * The document to build
     */
    private Document doc;
    
    /**
     * The most recent not closed element
     */
    private Node nodeactu;
    
    /**
     * Indicates if the characters must be considered as CDATA sections;
     */
    private boolean cs;
    
    // ***********************************************************************
    // * Constructors
    // ***********************************************************************
    
    /**
     * Contruct a DOMBuilder object.
     *
     * @param doc the document to fill
     *
     * @exception java.lang.NullPointerException the needed document was not provided.
     */
    public BasicDOMBuilder(Document doc) throws NullPointerException
    {
        setDocument(doc);
    }
    
    /**
     * Contruct a DOMBuilder object.
     *
     * @param element the element on which the DOM fragment will be built.
     *
     * @exception java.lang.NullPointerException the needed element was not provided.
     */
    public BasicDOMBuilder(Element element) throws NullPointerException
    {
        setElement(element);
    }
    
    // ***********************************************************************
    // * ContentHandler methods
    // ***********************************************************************
    
    public void characters(char[] ch, int start, int length)
    {
        if (nodeactu != doc)    // not allowed at document root
        {
            if (cs)
                nodeactu.appendChild(doc.createCDATASection(new String(ch,start,length)));
            else
                nodeactu.appendChild(doc.createTextNode(new String(ch,start,length)));
        }
    }
    
    public void endElement(String namespaceUI, String localName, String qName)
    {
        nodeactu=nodeactu.getParentNode();
    }
    
    public void ignorableWhitespace(char[] ch, int start, int length)
    {
        if (nodeactu != doc)    // not allowed at document root
            nodeactu.appendChild(doc.createTextNode(new String(ch,start,length)));
    }
    
    public void processingInstruction(String target, String data)
    {
        nodeactu.appendChild(doc.createProcessingInstruction(target,data));
    }
    
    public void skippedEntity(String name)
    {
        nodeactu.appendChild(doc.createEntityReference(name));
    }
    
    public void startElement
    (String namespaceURI,
    String localName,
    String qName,
    Attributes atts)
    {
        if (namespaceURI.equals(""))
            namespaceURI = null;
        
        Element element=doc.createElementNS(namespaceURI,qName);
        int nbattr=atts.getLength();
        for (int i=0; i<nbattr; i++)
        {
            namespaceURI=atts.getURI(i);
            if (namespaceURI.equals(""))
                namespaceURI = null;
            element.setAttributeNS(namespaceURI,atts.getQName(i),atts.getValue(i));
        }
        nodeactu.appendChild(element);
        nodeactu=element;
    }
    
    // ***********************************************************************
    // * LexicalHandler methods
    // ***********************************************************************
    
    public void comment(char[] ch, int start, int length)
    {
        nodeactu.appendChild
        (doc.createComment
        (new String(ch,start,length)));
    }
    
    public void endCDATA()
    {
        cs=false;
    }
    
    public void endDTD()
    {
    }
    
    public void endEntity(String name)
    {
    }
    
    public void startCDATA()
    {
        cs=true;
    }
    
    public void startDTD(String name, String publicId, String systemId)
    {
    }
    
    public void startEntity(String name)
    {
    }
    
    // ***********************************************************************
    // * Public methods
    // ***********************************************************************
    
    public void setDocument(Document doc)
    {
        if (doc==null)
            throw new NullPointerException("Couldn't create such an object from a null reference");
        this.doc=doc;
        nodeactu=doc;
        cs=false;
    }
    
    public void setElement(Element root)
    {
        if (root==null)
            throw new NullPointerException("Couldn't create such an object from a null reference");
        doc=root.getOwnerDocument();
        nodeactu=root;
        cs=false;
    }
    
}
