/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.util;

import org.xml.sax.*;
import org.xml.sax.ext.LexicalHandler;

/**
 * This class is a kind of SAX filter to plug on a contentHandler.
 */
public class HandlerDecorator implements LexicalHandler, ContentHandler, ErrorHandler
{
    private static final String RCSRevision = "$Revision: 1.2 $";
    private static final String RCSName     = "$Name: XQ1_1_1 $";
    
    protected LexicalHandler lexicalHandler = null;
    protected ContentHandler contentHandler = null;
    protected ErrorHandler errorHandler = null;
    
    /** Constructor
     */
    public HandlerDecorator()
    {
        this(null, null);
    }
    
    /** Constructor
     */
    public HandlerDecorator(ContentHandler contentHandler, LexicalHandler lexicalHandler)
    {
        this.contentHandler = contentHandler;
        this.lexicalHandler = lexicalHandler;
    }
    
    public void setContentHandler(ContentHandler handler)
    {
        contentHandler = handler;
    }
    
    public ContentHandler getContentHandler()
    {
        return contentHandler;
    }
    
    public void setLexicalHandler(LexicalHandler handler)
    {
        lexicalHandler = handler;
    }
    
    public LexicalHandler getLexicalHandler()
    {
        return lexicalHandler;
    }
    
    public void setErrorHandler(ErrorHandler handler)
    {
        errorHandler = handler;
    }
    
    public ErrorHandler getPluggedErrorHandler()
    {
        return errorHandler;
    }
    
    //////////////////////////////////////////////////////
    //      Implementation of ContentHandler interface.
    //////////////////////////////////////////////////////
    public void startDocument() throws SAXException
    {
        if (contentHandler != null)
            contentHandler.startDocument();
    }
    
    public void startElement(String str, String str1, String str2, Attributes attributes)
    throws SAXException
    {
        if (contentHandler != null)
            contentHandler.startElement(str, str1, str2, attributes);
    }
    
    public void setDocumentLocator(Locator locator)
    {
        if (contentHandler != null)
            contentHandler.setDocumentLocator(locator);
    }
    
    public void endDocument() throws SAXException
    {
        if (contentHandler != null)
            contentHandler.endDocument();
    }
    
    public void characters(char[] values, int param, int param2)
    throws SAXException
    {
        if (contentHandler != null)
            contentHandler.characters(values, param, param2);
    }
    
    public void processingInstruction(String str, String str1)
    throws SAXException
    {
        if (contentHandler != null)
            contentHandler.processingInstruction(str, str1);
    }
    
    public void endElement(String str, String str1, String str2)
    throws SAXException
    {
        if (contentHandler != null)
            contentHandler.endElement(str, str1, str2);
    }
    
    public void startPrefixMapping(String str, String str1) throws SAXException
    {
        if (contentHandler != null)
            contentHandler.startPrefixMapping(str, str1);
    }
    
    public void ignorableWhitespace(char[] values, int param, int param2)
    throws SAXException
    {
        if (contentHandler != null)
            contentHandler.ignorableWhitespace(values, param, param2);
    }
    
    public void endPrefixMapping(String str)  throws SAXException
    {
        if (contentHandler != null)
            contentHandler.endPrefixMapping(str);
    }
    
    public void skippedEntity(String str) throws SAXException
    {
        if (contentHandler != null)
            contentHandler.skippedEntity(str);
    }
    
    //////////////////////////////////////////////////////
    //      Implementation of LexicalHandler interface.
    //////////////////////////////////////////////////////
    public void startDTD(String name, String publicId,  String systemId)
    throws SAXException
    {
        if (lexicalHandler != null)
            lexicalHandler.startDTD(name, publicId, systemId);
    }
    
    public void endDTD() throws SAXException
    {
        if (lexicalHandler != null)
            lexicalHandler.endDTD();
    }
    
    public void startEntity(String name) throws SAXException
    {
        if (lexicalHandler != null)
            lexicalHandler.startEntity(name);
    }
    
    public void endEntity(String name) throws SAXException
    {
        if (lexicalHandler != null)
            lexicalHandler.endEntity(name);
    }
    
    public void startCDATA() throws SAXException
    {
        if (lexicalHandler != null)
            lexicalHandler.startCDATA();
    }
    
    public void endCDATA() throws SAXException
    {
        if (lexicalHandler != null)
            lexicalHandler.endCDATA();
    }
    
    public void comment(char ch[], int start, int length) throws SAXException
    {
        if (lexicalHandler != null)
            lexicalHandler.comment(ch, start, length);
    }
    
    //////////////////////////////////////////////////////
    //      Implementation of EventHandler interface.
    //////////////////////////////////////////////////////
    /**
     * Filter a warning event.
     *
     * @param e The warning as an exception.
     * @exception org.xml.sax.SAXException The client may throw
     *            an exception during processing.
     * @see org.xml.sax.ErrorHandler#warning
     */
    public void warning(SAXParseException e)
    throws SAXException
    {
        if (errorHandler != null)
        {
            errorHandler.warning(e);
        }
    }
    
    
    /**
     * Filter an error event.
     *
     * @param e The error as an exception.
     * @exception org.xml.sax.SAXException The client may throw
     *            an exception during processing.
     * @see org.xml.sax.ErrorHandler#error
     */
    public void error(SAXParseException e)
    throws SAXException
    {
        if (errorHandler != null)
        {
            errorHandler.error(e);
        }
    }
    
    
    /**
     * Filter a fatal error event.
     *
     * @param e The error as an exception.
     * @exception org.xml.sax.SAXException The client may throw
     *            an exception during processing.
     * @see org.xml.sax.ErrorHandler#fatalError
     */
    public void fatalError(SAXParseException e)
    throws SAXException
    {
        if (errorHandler != null)
        {
            errorHandler.fatalError(e);
        }
    }
    
}




