/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.util;

import java.util.Hashtable;
import java.util.Stack;

import org.xquark.schema.validation.ElementPSVInfoset;
import org.xquark.schema.validation.PSVInfoset;

/**
 * This class belongs to a library that allow to hide SAX specificities when
 * constructing java objects from an XML document with an XML schema.
 * This class is the base node used in a stack to keep track of the current tag.
 * This class must be extended in order to perform specific action for a tag.
 *
 * @see org.xquark.xml.util.SaxMapper
 */
public class TagTracker
{
	private static final String RCSRevision = "$Revision: 1.3 $";
	private static final String RCSName = "$Name: XQ1_1_1 $";

	// Table of tag trackers.
	// This table contains an entry for
	// every tag name that this TagTracker
	// has been configured to follow.
	// This is a single-level parent-child relation.
	//
	private Hashtable trackers = new Hashtable();

	// Useful for skipping tag names that are not
	// being tracked.
	private static SkippingTagTracker skip = new SkippingTagTracker();

	// default constructor
	public TagTracker()
	{}

	/**
	 * Configuration method for setting up a network of tag trackers...
	 * Each parent tag name should be configured (call this method) for each
	 * child tag name that it will track.
	 * @param tagName an absolute or relative path using the canonical XPath
	 * syntax (/{nsuri1}n1/{nsuri2}n2)
	 * @param tracker child 
	 */
	public void track(String tagName, TagTracker tracker)
	{
		trackAny(tagName, tracker);
	}
	/**
	 * Configuration method for setting up a network of tag trackers...
	 * Each parent tag name should be configured (call this method) for each
	 * child tag name that it will track.
	 * @param attName an absolute or relative path using the canonical XPath
	 * syntax (/{nsuri1}n1/@{nsuri2}n2)
	 * @param attHandler attribute actual value handler
	 */
	public void trackAttribute(String attName, AttributeTracker attHandler)
	{
		trackAny(attName, attHandler);
	}

	private void trackAny(String tagName, Object o)
	{

		int slashOffset = tagName.indexOf("/");

		if (slashOffset < 0)
		{
			// if it is a simple tag name ( no "/" seperators )
			// simply add it.
			trackers.put(tagName, o);

		}
		else if (slashOffset == 0)
		{
			// Oooops leading slash, remove it and 
			// try again recursivley.
			trackAny(tagName.substring(1), o);

		}
		else
		{
			// if it is not a simple tag name
			// recursively add the tag.
			String topTagName = tagName.substring(0, slashOffset);
			String remainderOfTagName = tagName.substring(slashOffset + 1);
			TagTracker child = (TagTracker) trackers.get(topTagName);
			if (child == null)
			{
				// Not currently tracking this 
				// tag. Add new tracker.
				child = new TagTracker();
				trackers.put(topTagName, child);
			}
			child.trackAny(remainderOfTagName, o);

		}

	}

	// Tag trackers work together on a stack.
	// The tag tracker at the top of the stack
	// is the "active" tag tracker and is responsible
	// for delegating the tracking to a child tag
	// tracker or putting a skipping place marker on the
	// stack.
	//
	public void startElement(ElementPSVInfoset infoSet, Stack tagStack)
	{
		// Lookup up the tag name in the tracker table.
		// Note, this implementation does not address 
		// using XML name space support that is now available
		// with SAX2.
		// We are simply using the localName as a key
		// to find a possible tracker.
		TagTracker tracker = (TagTracker) trackers.get(
		getExpandedName(infoSet.getNamespaceURI(), infoSet.getLocalName(), false));

		//
		// Are we tracking this tag name?
		//
		if (tracker == null)
		{
			// Not tracking this
			// tag name.  Skip the
			// entire branch.
			tagStack.push(skip);
		}
		else
		{

			// Found a tracker for this
			// tag name.  Make it the
			// new top of stack tag
			// tracker

			// Send the deactivate event to this tracker.
			onDeactivate();

			// Send the on start to the new active 
			// tracker.
			tracker.onStart(infoSet.getNamespaceURI(),
							infoSet.getLocalName());
				
			// Fire events to registered attribute trackers			
			int nbAtts = infoSet.getAttributeCount();
			AttributeTracker attHandler;
			PSVInfoset attInfoset;
			
			for (int i = 0; i < nbAtts; i++)
			{
				attInfoset = infoSet.getAttributePSVInfoset(i);
				attHandler =
					(AttributeTracker) tracker.trackers.get(
						getExpandedName(
							attInfoset.getNamespaceURI(),
							attInfoset.getLocalName(),
							true));
				if (attHandler != null)
					attHandler.onAttributeEncountered(
						attInfoset.getNamespaceURI(),
						attInfoset.getLocalName(),
						attInfoset.getActualValue(),
				        attInfoset.getSpecified() == PSVInfoset.SCHEMA ? true : false);
			}
			
			tagStack.push(tracker);

		}

	}

	// Tag trackers work together on a stack.
	// The tag tracker at the top of the stack
	// is the "active" tag tracker and is responsible
	// for reestablishing it's parent tag tracker
	// ( next to top of stack ) when it has
	// been notified of the closing tag.
	//
	public void endElement(
		ElementPSVInfoset infoSet,
		Stack tagStack)
	{

		// Send the end event.
		onEnd(infoSet.getNamespaceURI(),
				infoSet.getLocalName(),
				infoSet.getActualValue());

		// Clean up the stack...
		tagStack.pop();

		// Send the reactivate event.
		TagTracker activeTracker = (TagTracker) tagStack.peek();
		if (activeTracker != null)
		{
			activeTracker.onReactivate();
		}

	}

	// Methods for collecting content. These methods 
	// are intended to be overridden with specific
	// actions for nodes in the tag tracking network 
	// that require 

	public void onStart(
		String namespaceURI,
		String localName)
	{

		// default is no action...
	}

	public void onDeactivate()
	{

		// default is no action...
	}

	public void onEnd(
		String namespaceURI,
		String localName,
		Object actualValue)
	{

		// default is no action...
	}

	public void onReactivate()
	{

		// default is no action...
	}


	private String getExpandedName(
		String namespace,
		String localName,
		boolean att)
	{
		StringBuffer sb = new StringBuffer();

		if (att)
			sb.append('@');
		if (namespace != null && !namespace.equals(""))
		{
			sb.append('{');
			sb.append(namespace);
			sb.append('}');
		}
		sb.append(localName);
		return sb.toString();
	}

}
class SkippingTagTracker extends TagTracker
{
	private static final String RCSRevision = "$Revision: 1.3 $";
	private static final String RCSName = "$Name: XQ1_1_1 $";

	// Tag trackers work together on a stack.
	// The tag tracker at the top of the stack
	// is the "active" tag tracker.
	//
	// This class represents a skipping place
	// marker on the stack.  When a real tag
	// tracker places a skipping tag tracker on
	// the stack, that is an indication that
	// all tag names found during the skip are
	// of no intrest to the tag tracking network.
	//
	// This means that if the skipping tag tracker
	// is notified of a new tag name, this new
	// tag name should also be skipped.
	//
	// Since this class never varies it's behavior,
	// it is ok for it to skip new tag names by
	// placing itself on the stack again.
	public void startElement(ElementPSVInfoset infoSet, Stack tagStack)
	{

		//
		// If the current tag name is being
		// skipped, all children should be
		// skipped.
		//
		tagStack.push(this);

	}

	//
	// The skipping tag tracker has
	// nothing special to do when
	// a closing tag is found other
	// than to remove itself from
	// the stack, which as a side
	// effect replaces it with it's
	// parent as the "active", top
	// of stack tag tracker.
	//
	public void endElement(
		ElementPSVInfoset infoSet,
		Stack tagStack)
	{

		// Clean up the stack...
		tagStack.pop();

	}
}
