/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.xml.xdbc;

import java.io.InputStream;
import java.math.BigDecimal;
import java.util.Date;
import java.util.List;

import javax.xml.namespace.QName;


/**
 * This prepared statement interface is used for executing parameterized XQuery
 * or XPath prepared statements, setting external variables and obtaining the
 * results (update statements are not supported yet).<BR>
 * Prepared statements allow queries precompilation by the driver.<BR>
 * <p>In order to binds external variables, two kinds of methods are available:</p>
 * <ul>
 * <li>{@link #setExternalVariable(String, String, String)} assumes that the
 * value be parsed and validated.</li>
 * <li>setXXX() methods assumes that values passed match the variable primitive
 * type. These methods are supposed to be more efficient since validation only
 * consist in checking facets avoiding object allocation.</li>
 * </ul>
 */
public interface PreparedXMLStatement extends XMLStatement {

    
    /*********************************************************************/
    /*************************** Parameters management *******************/
    /*********************************************************************/

    /**
     * Binds the specified external variable to the current query of the
     * statement. Variable value will be parsed and validated.
     * @param ns The namespace of the external variable. Possibly an empty
     * string if the variable has no namespace.
     * @param varName local name of the external variable. Must not be null.
     * @param value a string that is the value to set for the variable.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void setExternalVariable(String ns, String varName, String value)
    throws XMLDBCException;

    /**
     * Binds the specified external variable to the current query of the
     * statement. Variable value will be parsed and validated.
     * @param varName local name of the external variable. Must not be null.
     * @param value a string that is the value to set for the variable.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void setExternalVariable(String varName, String value)
    throws XMLDBCException;

    /**
     * To set the specified variable with a value inside an object. This method
     * has to be used for instance when you own a java object instead of a
     * corresponding primitive type (e.g. Boolean instead of boolean, Long
     * instead of long, etc.).
     * @param ns The namespace of the external variable. Possibly an empty
     * string if the variable has no namespace.
     * @param varName local name of the external variable. Must not be null.
     * @param value an object that is the value to set for the parameter.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void setObject(String ns, String varName, Object value)
    throws XMLDBCException;

    /**
     * To set the specified variable with a value inside an object. This method
     * has to be used for instance when you own a java object instead of a
     * corresponding primitive type (e.g. Boolean instead of boolean, Long
     * instead of long, etc.).
     * @param varName local name of the external variable. Must not be null.
     * @param value an object that is the value to set for the parameter.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void setObject(String varName, Object value)
    throws XMLDBCException;

    /**
     * To set the specified variable with a @link java.math.BigDecimal value.
     * To be used with XML types derived from <code>decimal</code>.
     * @param ns The namespace of the external variable. Possibly an empty
     * string if the variable has no namespace.
     * @param varName local name of the external variable. Must not be null.
     * @param value a @link java.math.BigDecimal that is the value to set for
     * the variable.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void setBigDecimal(String ns, String varName, BigDecimal value)
    throws XMLDBCException;

    /**
     * To set the specified variable with a @link java.math.BigDecimal value.
     * To be used with XML types derived from <code>decimal</code>.
     * @param varName local name of the external variable. Must not be null.
     * @param value a @link java.math.BigDecimal that is the value to set for
     * the variable.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void setBigDecimal(String varName, BigDecimal value)
    throws XMLDBCException;

    /**
     * To set the specified variable with a @link java.io.InputStream value.
     * To be used with XML types derived from <code>Base64Binary</code> or 
     * <code>HexBinary</code>.
     * @param ns The namespace of the external variable. Possibly an empty
     * string if the variable has no namespace.
     * @param varName local name of the external variable. Must not be null.
     * @param value an input stream.
     * @param length the length of given input stream.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void setInputStream(String ns, String varName, InputStream value, int length)
    throws XMLDBCException;

    /**
     * To set the specified variable with a @link java.io.InputStream value.
     * To be used with XML types derived from <code>Base64Binary</code> or 
     * <code>HexBinary</code>.
     * @param varName local name of the external variable. Must not be null.
     * @param value an input stream.
     * @param length the length of given input stream.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void setInputStream(String varName, InputStream value, int length)
    throws XMLDBCException;

	/**
	 * To set the specified variable with a byte array value. To be used with
	 * XML types derived from <code>Base64Binary</code> or <code>HexBinary</code>.
     * @param ns The namespace of the external variable. Possibly an empty
     * string if the variable has no namespace.
	 * @param varName local name of the external variable. Must not be null.
	 * @param value a byte array that is the value to set for the parameter.
	 * @throws XMLDBCException if a data source access error occurs.
	 */
	public void setBytes(String ns, String varName, byte[] value)
	throws XMLDBCException;

	/**
	 * To set the specified variable with a byte array value. To be used with
	 * XML types derived from <code>Base64Binary</code> or <code>HexBinary</code>.
	 * @param varName local name of the external variable. Must not be null.
	 * @param value a byte array that is the value to set for the parameter.
	 * @throws XMLDBCException if a data source access error occurs.
	 */
	public void setBytes(String varName, byte[] value)
	throws XMLDBCException;

    /**
     * To set the specified variable with a boolean value. To be used with XML
     * types derived from <code>boolean</code>.
     * @param ns The namespace of the external variable. Possibly an empty
     * string if the variable has no namespace.
     * @param varName local name of the external variable. Must not be null.
     * @param value a boolean that is the value to set for the parameter.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void setBoolean(String ns, String varName, boolean value)
    throws XMLDBCException;

    /**
     * To set the specified variable with a boolean value. To be used with XML
     * types derived from <code>boolean</code>.
     * @param varName local name of the external variable. Must not be null.
     * @param value a boolean that is the value to set for the parameter.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void setBoolean(String varName, boolean value)
    throws XMLDBCException;

    /**
     * To set the specified variable with a @link java.util.Date value. To be
     * used with XML types derived from all built-in time types.
     * @param ns The namespace of the external variable. Possibly an empty
     * string if the variable has no namespace.
     * @param varName local name of the external variable. Must not be null.
     * @param value a java.sql.Date that is the value to set for the parameter.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void setDate(String ns, String varName, Date value)
    throws XMLDBCException;

    /**
     * To set the specified variable with a @link java.util.Date value. To be
     * used with XML types derived from all built-in time types.
     * @param varName local name of the external variable. Must not be null.
     * @param value a java.sql.Date that is the value to set for the parameter.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void setDate(String varName, Date value)
    throws XMLDBCException;

    /**
     * To set the specified variable with a double value. To be used with XML
     * types derived from <code>double</code>.
     * @param ns The namespace of the external variable. Possibly an empty
     * string if the variable has no namespace.
     * @param varName local name of the external variable. Must not be null.
     * @param value a double that is the value to set for the parameter.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void setDouble(String ns, String varName, double value)
    throws XMLDBCException;

    /**
     * To set the specified variable with a double value. To be used with XML
     * types derived from <code>double</code>.
     * @param varName local name of the external variable. Must not be null.
     * @param value a double that is the value to set for the parameter.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void setDouble(String varName, double value)
    throws XMLDBCException;

    /**
     * To set the specified variable with a float value. To be used with XML
     * types derived from <code>float</code>.
     * @param ns The namespace of the external variable. Possibly an empty
     * string if the variable has no namespace.
     * @param varName local name of the external variable. Must not be null.
     * @param value a float that is the value to set for the parameter.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void setFloat(String ns, String varName, float value)
    throws XMLDBCException;

    /**
     * To set the specified variable with a float value. To be used with XML
     * types derived from <code>float</code>.
     * @param varName local name of the external variable. Must not be null.
     * @param value a float that is the value to set for the parameter.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void setFloat(String varName, float value)
    throws XMLDBCException;

    /**
     * To set the specified variable with a long value. To be used with XML
     * types derived from <code>integer</code>.
     * @param ns The namespace of the external variable. Possibly an empty
     * string if the variable has no namespace.
     * @param varName local name of the external variable. Must not be null.
     * @param value a long that is the value to set for the parameter.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void setLong(String ns, String varName, long value)
    throws XMLDBCException;

    /**
     * To set the specified variable with a long value. To be used with XML
     * types derived from <code>integer</code>.
     * @param varName local name of the external variable. Must not be null.
     * @param value a long that is the value to set for the parameter.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void setLong(String varName, long value)
    throws XMLDBCException;

    /**
     * To set the specified variable with a String value. To be used with XML
     * types derived from <code>string</code>.
     * @param ns The namespace of the external variable. Possibly an empty
     * string if the variable has no namespace.
     * @param varName local name of the external variable. Must not be null.
     * @param value a string that is the value to set for the parameter.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void setString(String ns, String varName, String value)
    throws XMLDBCException;

    /**
     * To set the specified variable with a String value. To be used with XML
     * types derived from <code>string</code>.
     * @param varName local name of the external variable. Must not be null.
     * @param value a string that is the value to set for the parameter.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void setString(String varName, String value)
    throws XMLDBCException;

    /**
     * To set the specified variable with a Sequence value. 
     * @param ns The namespace of the external variable. Possibly an empty
     * string if the variable has no namespace.
     * @param varName local name of the external variable. Must not be null.
     * @param value a Sequence that is the value to set for the parameter.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void setSequence(String ns, String varName, List value)
    throws XMLDBCException;

    /**
     * To set the specified variable with a Sequence value.
     * @param varName local name of the external variable. Must not be null.
     * @param value a Sequence that is the value to set for the parameter.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void setSequence(String varName, List value)
    throws XMLDBCException;

     /**
     * Clears the current parameters values immediately.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void clearVariables() throws XMLDBCException;
    
    /**
     * Executes the query statement specified when this object was built with
     * the prepareStatement method (see XMLConnection interface) that may
     * return results.
     * If the query generates results, use the <I>getResultSet()</I> or
     * <I>getDocumentSet()</I> methods to get the set of results.
     * @return true if there are available resuts, else false.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public boolean execute() throws XMLDBCException;
    
    /**
     * Executes the query statement specified when this object was built with
     * the prepareStatement method (see XMLConnection interface) that returns
     * results.
     * @return an XMLResultSet object that contains the data produced by the
     * given query.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public XMLResultSet executeQuery() throws XMLDBCException;
    
    /**
     * Gets the array of parameter names of the statement
     */
    public QName[] getParameterNames();
	/**
	 * Gets the schema type of a particular parameter given it's name
	 */
	public QName getParameterType(QName varName);

}
