/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.xml.xdbc;

import java.util.List;

/**
 * This interface describes a XML collection of XML documents.<BR><BR>
 * <p><B>Note:</B></p>
 * <p>This object is likely to use resources that can be released using the
 * {@link #close()} method.</p>
 */
public interface XMLCollection extends Configurable
{

    /****************************************************************/
    /*************************** Constants declaration **************/
    /****************************************************************/
 


    /****************************************************************/
    /*************************** Collection management **************/
    /****************************************************************/

    /**
     * Closes this XMLCollection. Once a collection is closed, it should not 
     * be accessed any more.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void close() throws XMLDBCException;

    /****************************************************************/
    /*************************** Fields management ******************/
    /****************************************************************/

    /**
     * Retrieves the unique name of the XML collection.
     * @return the unique name of the XML collection.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public String getName() throws XMLDBCException;

    /**
     * Retrieves the description of the XML collection.
     * @return the description of the XML collection.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public String getDescription() throws XMLDBCException;

    /**
     * Sets the description of the XML collection.
     * @param description the description of the XML collection.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void setDescription(String description) throws XMLDBCException;

    /**
     * Retrieves the XML collection size (number of documents).
     * @return the XML collection size.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public long getDocumentCount() throws XMLDBCException;

    /**
     * Returns true if the XML collection is read-only mode.
     * @return true if read-only, false otherwise.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public boolean isReadOnly() throws XMLDBCException;

    /**
     * Changes the read-only mode for this XML collection.
     * @param readOnly true to change mode to read-only, false otherwise.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void setReadOnly(boolean readOnly) throws XMLDBCException;

    /**
     * Retrieves the XMLConnection object that produced this XMLCollection.
     * @return the XMLConnection object that produced this XMLCollection.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public XMLConnection getConnection() throws XMLDBCException;

    /****************************************************************/
    /*************************** Document management ***************/
    /****************************************************************/

    /** Creates a "filer" object that can be used to insert a new XML document
     * into the XML collection. A single filer can store several
     * documents sequentially, however it is an error to try to use a filer
     * before the current storage is finished.<BR>
     * <p><B>Warning:</B> Asuming an {@link XMLConnection} is monothread,
     * being able to create multiple XMLDocumentFiler does not mean one can use
     * XMLDocumentFiler created from the same XMLCollection in different
     * threads.</p>
     * @return an XMLDocumentFiler object.
     * @throws XMLDBCException if a database access error occurs.
     */
    public XMLDocumentFiler getDocumentFiler() throws XMLDBCException;

    /**
     * Retrieves an XML document in the XML collection.
     * <p>Note: implementations may accept for performance reasons the
     * XMLDocument creation even if the document does not really exist. In that
     * situation, an exception will be raised on subsequent operations.</p>
     * @param documentID the string used to identify the XML document in the collection.
     * @return an object representing the XML document, or null if not found.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public XMLDocument getDocument(String documentID) throws XMLDBCException;

    /**
     * Removes the named XML document from the XML collection.
     * @param documentID the string used to identify the XML document in the collection.
     * @return true if the XML document has been found, false otherwise.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public boolean removeDocument(String documentID) throws XMLDBCException;

    /**
     * Removes all XML documents from the XML collection.
     * @return the number of XML documents removed.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public int removeAllDocuments() throws XMLDBCException;

    /**
     * Checks if a named XML document exists in this XML collection.
     * @param documentID the string used to identify the XML document in the collection.
     * @return true if the document exists in this XML collection, false otherwise.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public boolean containsDocument(String documentID) throws XMLDBCException;

    /**
     * Renames a document.
     * @param oldID the old string used to identify the XML document in the collection.
     * @param newID the new string used to identify the XML document in the collection. 
     * @return true if the document has been found, false otherwise.
     * @throws XMLDBCException if an error or operation not supported.
     */
    public boolean renameDocument(String oldID, String newID) throws XMLDBCException;

    /**
     * Returns a list of documents identifiers available in the XML collection.
     * Because of potential updates to the data source, the iterator
     * does not necessarily return valid identifiers, nor does it
     * necessarily return all the documents in the collection.
     * @return a list of documents identifiers as String objects.
     * @throws XMLDBCException if data source access error occurs or operation not supported.
     */
    public List getIdentifierList() throws XMLDBCException;
}
