/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.xml.xdbc;

import java.util.List;
import java.util.Map;

/**
 * This interface provides comprehensive information about the "data source".
 * Since this interface accesses all sorts of data sources, the term data source is loosely applied.
 * Information provided depend on driver vendors implementation.<BR>
 * An XMLDBCNotSupportedException will be thrown if a driver does not support a metadata method.<BR><BR>
 */
public interface XMLDataSourceMetaData {
 


    /**
     * Inner interface of XMLDataSourceMetaData interface.<BR><BR>
     */
    public static interface Entry {

	/**
	 * Gets the input specification as an XPath expression.
	 */
	public java.lang.String getInput();

	/**
	 * Gets the output specification as an XPath expression.
	 */
	public java.lang.String getOutput();

	/**
	 * Sets the input specification
	 * @param xpath XPath expression that is valid over the schema
	 */
	public void setInput(java.lang.String xpath);

	/**
	 * Sets the output specification
	 * @param xpath XPath expression that is valid over the schema
	 */
	public void setOutput(java.lang.String xpath);
    }

    /**
     * Get the properties that describe this data source.
     * <p>
     * Valid properties are listed in the table.<br>
     * The type "int" means the property value string is a valid
     * integer.<br>
     * The type "string" means the property value is a string.<br>
     * The type "boolean" means the property value is "true" or "false".<br>
     * <p>
     * <table border=1>
     * <tr> <td><b>Property</b></td><td><b>Definition</b></td><td><b>Type</b></td></tr>
     * <tr> <td>xmldbc.maxConnections</td> <td>Maximum active connections</td> <td>int</td> </tr>
     * <tr> <td>xmldbc.maxStatementLength</td> <td>Maximum length for a statement</td> <td>int</td> </tr>
     * <tr> <td>xmldbc.maxActiveStatements</td> <td>Maximum number of active statements</td> <td>int</td> </tr>
     * <tr> <td>xmldbc.maxResultSetLength</td> <td>Maximum length of a result set</td> <td>int</td> </tr>
     * <tr> <td>xmldbc.maxUserNameLength</td> <td>Maximum length of a user name</td> <td>int</td> </tr>
     * <tr> <td>xmldbc.dataSourceVersion</td> <td>Version number of the data source</td> <td>string</td> </tr>
     * <tr> <td>xmldbc.vendor.name</td> <td>Name of vendor of underlying source</td> <td>string</td> </tr>
     * <tr> <td>xmldbc.vendor.product</td> <td>Product name of vendor of underlying source</td> <td>string</td> </tr>
     * <tr> <td>xmldbc.vendor.version</td> <td>Version string of vendor of underlying source</td> <td>string</td> </tr>
     * <tr> <td>xmldbc.driver.version</td> <td>Version string of this XMLDBC driver</td> <td>string</td> </tr>
     * <tr> <td>xmldbc.driver.product</td> <td>Product name of this XMLDBC driver</td> <td>string</td> </tr>
     * <tr> <td>xmldbc.xpath</td> <td>Supports XPath</td> <td>boolean</td> </tr>
     * <tr> <td>xmldbc.xpath.version</td> <td>Version of XPath supported</td> <td>string</td> </tr>
     * <tr> <td>xmldbc.xquery</td> <td>Supports XQuery</td> <td>boolean</td> </tr>
     * <tr> <td>xmldbc.xquery.version</td> <td>Version of XQuery supported</td> <td>string</td> </tr>
     * <tr> <td>xmldbc.xquery.outerJoin</td> <td>Supports XQuery outer join</td> <td>boolean</td> </tr>
     * <tr> <td>xmldbc.twoPhaseCommit</td> <td>Supports two phase commit</td> <td>boolean</td> </tr>
     * <tr> <td>xmldbc.xquery.equality</td> <td>Supports equality operator in WHERE clause</td> <td>boolean</td> </tr>
     * <tr> <td>xmldbc.xquery.result</td> <td>Supports construction in RESULT clause</td> <td>boolean</td> </tr>
     * <tr> <td>xmldbc.xmlCollection</td> <td>Supports at least part of XMLCollection</td> <td>boolean</td> </tr>
     * <tr> <td>xmldbc.xmlCollection.id</td> <td>Supports methods that use XML ID String</td> <td>boolean</td> </tr>
     * <tr> <td>xmldbc.xmlCollection.iterator</td> <td>Supports iterator() method</td> <td>boolean</td> </tr>
     * <tr> <td>xmldbc.xquery.multiCollection</td> <td>An XQUERY expression may reference multiple XMLCollections</td> <td>boolean</td> </tr>
     * <tr> <td></td> <td></td> <td></td> </tr>
     * </table>
     * <p>
     * For example, a data source that supports all of XQuery without restriction
     * would define true for the xmldb.xquery property.  This property implies all other xmldbc.xquery.*
     * properties are supported. A data source
     * that provided no query functionality at all and simply dumped
     * its contents (e.g. an XML file), would not provide any
     * properties at all. A data source may return other properties not listed.
     * For developers, note that no set of capabilities can completely describe all data sources,
     * so a particular data source may throw XMLDBCNotSupportedException in rare circumstances.
     * @return the property set describing this data source
     */
    String[] getPropertyList();

    /**
     * Get the property with the specified name
     * @return the value of the property
     * @throws XMLDBCNotRecognizedException if the property is not known
     */
    Object getProperty(String propertyId) throws XMLDBCNotRecognizedException;
    
    //*************************** Connection informations:

    /**
     * To get the XMLConnection object that produced this XMLDataSourceMetaData object.
     * @return the XMLConnection object that produced this object.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public XMLConnection getConnection() throws XMLDBCException;

    //*************************** Other metadata informations

//    /**
//     * Retrieves the name of all XML collections in this data source.
//     * @return a list of all XML collection names as String objects.
//     * @throws XMLDBCException if a data source access error occurs.
//     * @throws XMLDBCNotSupportedException if the data source does not support this operation (typically, when the data source is read-only)
//     */
    public List getCollectionNames() throws XMLDBCException, XMLDBCNotSupportedException;

    /**
     * Retrieves the available schema namespaces in the data source.
     * @return the target namespaces of all schemas available in the data source
     * in a list of String objects.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public List getSchemaNamespaces() throws XMLDBCException;

//    /**
//     * Retrieves the XML document describing all existing paths in all documents 
//     * of the specified XML collection.
//     * @param colName the name of the XML collection.
//     * @return an XML document containing the path set.
//     * @throws XMLDBCException if a data source access error occurs.
//     */
    public XMLDocument getPathSet(String colName) throws XMLDBCException;
    
    public List getMetaModules() throws XMLDBCException;
    /**
     * Retrieves the XML document describing the XML schema for the given namespace.
     * @param targetNamespace the target namespace of the schema.
     * @return an XML document containing the XML schema.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public XMLDocument getSchema(String targetNamespace) throws XMLDBCException;

    /**
     * Retrieves the XML document describing the complete metadata of the data source.
     * This includes all available schemas and the path set of all available XML collections.
     * @return an XML document containing the data source metadata.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public XMLDocument getMetaData() throws XMLDBCException;

    /**
     * To get the set of capabilities of the data source.  Each element is a
     * DataSourceMetaData.Entry object.  The input is an XPath
     * expression that requires a value in the
     * WHERE clause.  The output is an XPath expression describing the
     * output. An empty set indicates no retrictions on the queries.
     */
    public java.util.Set getCapabilities();
}

