/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.xml.xdbc;

/**
 * This interface is used to insert XML documents in XML collections.
 * A single filer can store several documents sequentially, however it is an 
 * error to try to use a filer before the current storage is finished.
 * Objects implementing this interface can be plugged into a SAX2 XMLReader
 * to process XML data as SAX2 events. Both org.xml.sax.ContentHandler,
 * org.xml.sax.ext.LexicalHandler and org.xml.sax.ErrorHandler are supported,
 * so that comments and processing instructions are not lost during storage.
 * <p><B>Notes:</B></p>
 * <ol>
 * <li>The error handler must absolutely be plugged so that the
 * application could have a consistent behavior in case of parsing error due to
 * to a malformed XML document.</li>
 * <li>This object is likely to use resources that can be released using the 
 * {@link #close()} method.</li>
 * </ol>
 */

public interface XMLDocumentFiler 
extends org.xml.sax.ContentHandler, org.xml.sax.ErrorHandler,
        org.xml.sax.ext.LexicalHandler, Configurable
   {
 


  /**
   * Closes the object and frees its associated resources.
   * @throws XMLDBCException if a data source access error occurs.
   */
  public void close() throws XMLDBCException;
  
  /**
   * Returns the XML collection that produced this object.
   * @return the XML collection that produced this object.
   */
  public XMLCollection getCollection();
  
  /**
   * Sets the identifier of the next document to be stored through this object.
   * The identifier is reset to null after each storage.
   * @param id the identifier of the next document to be stored.
   * @throws XMLDBCException if a data source access error occurs.
   */
  public void setDocumentId(String id) throws XMLDBCException;

  /**
   * Gets the identifier to be used by the next document to be stored through this object.
   * If this is null, an internal identifier will be used.
   * @return the identifier to be used by the next document to be stored.
   */
  public String getDocumentId();
  
  /**
   * Inserts an XML document in the filer's collection using the SAX method. 
   * The SAX parser is automatically allocated using JAXP. This method uses the
   * identifier stored through the setDocumentId() method if available, or an
   * internal identifier otherwise.
   * @param input the SAX InputSource.
   * @return the stored XML document.
   * @throws XMLDBCException if a data source access error occurs.
   * @throws org.xml.sax.SAXException if a parse exception occurs.
   */
  public XMLDocument insertDocument(org.xml.sax.InputSource input) throws XMLDBCException, org.xml.sax.SAXException;
  
  /**
   * Inserts an XML document in the filer's collection using the SAX method. 
   * The SAX parser is automatically allocated using JAXP.
   * @param input the SAX InputSource.
   * @param id the document identifier in the collection.
   * @return the stored XML document.
   * @throws XMLDBCException if a data source access error occurs.
   * @throws org.xml.sax.SAXException if a parse exception occurs.
   */
  public XMLDocument insertDocument(org.xml.sax.InputSource input, String id) throws XMLDBCException, org.xml.sax.SAXException;
  
  /**
   * Inserts an XML document in the filer's collection using the SAX method. 
   * This method uses the identifier stored through the setDocumentId() method
   * if available, or an internal identifier otherwise.
   * @param parser the SAX 2 XMLReader implementation.
   * @param input the SAX InputSource.
   * @return the stored XML document.
   * @throws XMLDBCException if a data source access error occurs.
   * @throws org.xml.sax.SAXException if a parse exception occurs.
   */
  public XMLDocument insertDocument(org.xml.sax.XMLReader parser, org.xml.sax.InputSource input) throws XMLDBCException, org.xml.sax.SAXException;
  
  /**
   * Inserts an XML document in the filer's collection using the SAX method. 
   * @param parser the SAX 2 XMLReader implementation.
   * @param input the SAX InputSource.
   * @param id the document identifier in the collection.
   * @return the stored XML document.
   * @throws XMLDBCException if a data source access error occurs.
   * @throws org.xml.sax.SAXException if a parse exception occurs.
   */
  public XMLDocument insertDocument(org.xml.sax.XMLReader parser, org.xml.sax.InputSource input, String id) throws XMLDBCException, org.xml.sax.SAXException;
   
    /**
   * Inserts the XML document in the filer's collection. This method uses the identifier
   * stored through the setDocumentId() method if available, or an internal identifier
   * otherwise.
   * @param doc the XML document as a DOM2 org.w3c.dom.Document.
   * @return the stored XML document.
   * @throws XMLDBCException if a data source access error occurs.
   */
  public XMLDocument insertDocument(org.w3c.dom.Document doc) throws XMLDBCException;
  
  /**
   * Inserts the XML document in the filer's collection with the specified identifier.
   * @param doc the XML document as a DOM2 org.w3c.dom.Document.
   * @param id the document identifier in the collection.
   * @return the stored XML document.
   * @throws XMLDBCException if a data source access error occurs.
   */
  public XMLDocument insertDocument(org.w3c.dom.Document doc, String id) throws XMLDBCException;

  /**
   * Inserts the XML document contained in the string in the filer's collection. 
   * This method uses the identifier stored through the setDocumentId() method 
   * if available, or an internal identifier otherwise.
   * @param doc a string containing the XML document.
   * @return the stored XML document.
   * @throws XMLDBCException if a data source access error occurs.
   * @throws org.xml.sax.SAXException if a parse exception occurs.
   */
  public XMLDocument insertDocument(String doc) throws XMLDBCException, org.xml.sax.SAXException;
  
  /**
   * Inserts the XML document contained in the string in the filer's collection
   * with the specified identifier.
   * @param doc a string containing the XML document.
   * @param id the document identifier in the collection.
   * @return the stored XML document.
   * @throws XMLDBCException if a data source access error occurs.
   * @throws org.xml.sax.SAXException if a parse exception occurs.
   */
  public XMLDocument insertDocument(String doc, String id) throws XMLDBCException, org.xml.sax.SAXException;
  
  /**
   * Turns on or off the automatic buffering mode. If this mode is off
   * XML documents are buffered by the filer until the flushBuffer() method is
   * called providing a kind of 'manual' mode for controlling data flush.
   * If the batch mode is on (the default) data is automatically flushed when
   * a certain amount of data is reached or when a commit is performed on the
   * connection.
   * @param activated 'true' is the default and means that autoflush is on.
   * @throws XMLDBCException if a data source access error occurs.
   */
  public void setAutoFlush(boolean activated) throws XMLDBCException;
  
  /**
   * Retrieves the current buffering mode.
   * @see #setAutoFlush(boolean)
   * @return true if the 'automatic' buffering mode is activated.
   */
  public boolean getAutoFlush();
  
  /**
   * Sends buffered XML data to the data source. This method is to be called
   * only when the automatic buffering mode is off.
   * @see #setAutoFlush(boolean)
   * @throws XMLDBCException if a data source access error occurs.
   */
  public void flushBuffer() throws XMLDBCException;

  /**
   * Discards the XML data buffered by the filer since the last call to
   * flushBuffer() method. This method is to be called only when the automatic
   * buffering mode is off.
   * @see #setAutoFlush(boolean)
   * @throws XMLDBCException if a data source access error occurs.
   */
    public void clearBuffer() throws XMLDBCException;

    /**
     * Allow an application to register an error event handler for non-fatal
     * errors.
     *
     * <p>If the application does not register an error handler, all warning
     * and non-fatal error events reported by the filer will be silently
     * ignored and storage will continue.</p>
     *
     * <p>Applications may register a new or different handler in the
     * middle of a storage operation, and the filer must begin using the new
     * handler immediately.</p>
     *
     * @param handler The error handler.
     * @exception java.lang.NullPointerException If the handler 
     *            argument is null.
     * @see #getErrorHandler
     */
    public void setErrorHandler (XMLErrorHandler handler);


    /**
     * Return the current error handler.
     *
     * @return The current error handler, or null if none
     *         has been registered.
     * @see #setErrorHandler
     */
    public XMLErrorHandler getErrorHandler ();
}
