/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.xml.xdbc;


/**
 * This class is used for managing a set of drivers and/or connections.<BR><BR>
 */
public class XMLDriverManager {



    private static java.util.Set drivers = java.util.Collections.synchronizedSet(new java.util.HashSet());
    private static java.io.PrintWriter logWriter = null;
    private static int loginTimeout = 0;
  
    private XMLDriverManager() {}
  
  /**
   * Attempts to locate a driver that understands the given URI.
   * @param uri an URI corresponding to an existing driver.
   * @return a driver that understands the given URI.
   * @throws XMLDBCException if a data source access error occurs.
   */
  public static XMLDriver getDriver(String uri)
  throws XMLDBCException {
    java.util.Iterator it = getDrivers();
    while (it.hasNext()) {
      XMLDriver driver = (XMLDriver)it.next();
      if (driver.acceptsURI(uri))
        return driver;
    }
    return null;
  }
  
  /**
   * Attempts to establish a connection to the given data source URI.
   * @param uri an URI corresponding to an existing driver.
   * @return a connection to the data source.
   * @throws XMLDBCException if a data source access error occurs.
   */
  public static XMLConnection getConnection(String uri)
  throws XMLDBCException {
    return getDataSource(uri).getConnection();
  }
  
  /**
   * Attempts to establish a connection to the given data source URI with a login/password.
   * @param uri an URI corresponding to an existing driver.
   * @param user a login to access data source account.
   * @param password a password to access data source account.
   * @return a connection to the data source.
   * @throws XMLDBCException if a data source access error occurs.
   */
  public static XMLConnection getConnection(String uri, String user, String password)
  throws XMLDBCException {
	XMLDriver driver = getDriver(uri);
	if (driver == null) throw new XMLDBCException("No driver found for uri: "+uri);
	// Try first with all parameters
	XMLDataSource dataSource = driver.getDataSource(uri, user, password);
	if (dataSource != null) {
		dataSource.setLogWriter(logWriter);
		dataSource.setLoginTimeout(loginTimeout);
	    return dataSource.getConnection();
	}
	// Try with uri alone
	dataSource = driver.getDataSource(uri);
	if (dataSource != null) {
		dataSource.setLogWriter(logWriter);
		dataSource.setLoginTimeout(loginTimeout);
		return dataSource.getConnection(user, password);
	} else {
		throw new XMLDBCException("No data source found for uri: "+uri);
	}
  }
  
  /**
   * Returns a handle to a data source with the specified URI.
   * @param uri an URI corresponding to a data source.
   * @return an instance of the specified data source,
   *         or null if the driver does not know this type of data source
   * @throws XMLDBCException if a data source access error occurs.
   */
  public static XMLDataSource getDataSource(String uri)
  throws XMLDBCException {
    XMLDriver driver = getDriver(uri);
    if (driver == null) throw new XMLDBCException("No driver found for uri: "+uri);
    XMLDataSource dataSource = driver.getDataSource(uri);
    if (dataSource == null) throw new XMLDBCException("No data source found for uri: "+uri);
    dataSource.setLogWriter(logWriter);
    dataSource.setLoginTimeout(loginTimeout);
    return dataSource;
  }
  
  /**
   * Returns a handle to a data source with the specified URI.
   * @param uri a URI corresponding to a data source.
   * @param user a user name
   * @param password a user password
   * @return an instance of the specified data source,
   *         or null if the driver does not know this type of data source
   * @throws XMLDBCException if a data source access error occurs.
   */
  public static XMLDataSource getDataSource(String uri, String user, String password)
  throws XMLDBCException {
	XMLDriver driver = getDriver(uri);
	if (driver == null) throw new XMLDBCException("No driver found for uri: "+uri);
	XMLDataSource dataSource = driver.getDataSource(uri, user, password);
	if (dataSource == null) throw new XMLDBCException("No data source found for uri: "+uri);
	dataSource.setLogWriter(logWriter);
	dataSource.setLoginTimeout(loginTimeout);
	return dataSource;
  }
  
  /** 
   * Sets the default logging/tracing java.io.PrintWriter object that is used
   * by all data source objects
   * obtained from the manager.
   * @param writer the new logging/tracing java.io.PrintWriter object, or null to disable logging and tracing
   */
  public static void setLogWriter(java.io.PrintWriter writer) {
    logWriter = writer;
  }
  
  /** 
   * Retrieves the default log writer currently used by all data source objects
   * obtained from the manager.
   * @return the java.io.PrintWriter object used to log data source messages, or null if none is defined
   */
  public java.io.PrintWriter getLogWriter() {
    return logWriter;
  }
  
  /**
   * Sets the maximum time in seconds that all data source objects obtained
   * from the manager will wait when attempting to log in to a data source.
   * Setting the timeout to 0 will disable it.
   * @param seconds The login time limit in seconds
   */
  public void setLoginTimeout(int seconds) {
    loginTimeout = seconds;
  }
  
  /** 
   * Gets the maximum time in seconds that all data source objects obtained
   * from the manager can wait when attempting to log in to a data source
   * @return the data source login timeout in seconds, 0 if no timeout is defined
   */
  public int getLoginTimeout() {
    return loginTimeout;
  }
  
  /**
   * Adds the given driver in the driver manager list.
   * @param driver the new XMLDBC driver that must be registered with the 
   * driver manager.
   */
  public static void registerDriver(XMLDriver driver) {
    drivers.add(driver);
  }
  
  /**
   * Removes the specified driver from the driver manager list.
   * @param driver the XMLDBC driver that must be dropped from the driver manager.
   */
  public static void deregisterDriver(XMLDriver driver) {
    drivers.remove(driver);
  }
  
  /**
   * Returns an iterator on registered XMLDBC drivers.
   * @return an enumeration of XMLDBC drivers.
   */
  public static java.util.Iterator getDrivers() {
    return drivers.iterator();
  }
}
