/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.xpath;

import java.util.*;

/**
 * This tree structure implements the XPath data model.
 *
 * Each <code>Node</code> in the <code>Tree</code> is identified by a
 * <code>org.xquark.xpath.LocationExpression</code> representing the full
 * access path.
 */
public class XTree
{
	private static final String RCSRevision = "$Revision: 1.2 $";
	private static final String RCSName = "$Name: XQ1_1_1 $";

	protected XTreeNode root;

	protected HashMap locationNodeMap = new HashMap();

	/**
	 * Accessor to the tree root node.
	 * @return an XTreeNode of type ROOT
	 * @see Type
	 */
	public XTreeNode getRoot()
	{
		return root;
	}

	/**
	 * Setter for the tree root node.
	 * @param root an XTreeNode of type ROOT
	 * @see Type
	 */
	void setRoot(XTreeNode root)
	{
		this.root = root;
	}

	/**
	 * Remove a branch from the tree.
	 * @param location a location expression identifying the root node of the
	 * subtree to prune. 
	 */
	public void remove(PathExpr location)
	{
		throw new UnsupportedOperationException("Not implemented yet.");
	}

	/**
	 * Returns the node corresponding to the path location passed as a
	 * parameter.
	 *
	 * <p><B>The path expression must not contains wildcards.</B> Once you
	 * get an XTreeNode (you can also use getRoot), use navigate(Step step)</p>
	 * @param location the location expression used for node retrieval. 
	 */
	public XTreeNode getNode(PathExpr location)
	{
		int hash = location.hashCode();
		return (XTreeNode) locationNodeMap.get(location);
	}

	/**
	 * Returns the nodes corresponding to an odd path location.
	 * @param location the location expression used for nodes search. 
	 */
	public Collection getNodeSet(ArrayList location)
	{
		return prune(location);
	}

	/**
	 * Compute the LocationExpressions matching given regular location
	 * expression.
	 * @param location the location expression used for nodes search.
	 * @return a collection of XTreeNode objects
	 */
	public final Collection prune(ArrayList location)
	{
		return root.prune(location);
	}

	/**
	 * Compute the LocationExpressions matching given regular location
	 * expression building a collection of collections in order to
	 * regroup same level (for using in PathResolver and next for typing))
	 * @param location a location expression
	 * @return a collection collection of collections containing XTreeNode
	 * objects matching the regular expression. First collection are results
	 * for the level of the root node.
	 */
	public final Collection pruneGroup(ArrayList location)
	{
		return root.pruneGroup(location);
	}

	/**
	 * Returns an iterator browsing the tree using a depth-first algorithm.
	 * @return a standard java iterator.
	 */
	public Iterator iterator()
	{
		return new XTreeIterator(this);
	}

	/**
	 * Add an XTree node to the tree index. Should be called for every node
	 * added to the tree
	 * @param node the new attached node.
	 * @return XTreeNode the new attached node.
	 */
	public XTreeNode register(XTreeNode node)
	{
		PathExpr path = node.getLocation();
		if (path != null)
			locationNodeMap.put(path, node);
		return node;
	}

	/**
	 * Return a string representation of this tree.
	 * @return the string representation of the tree.
	 */
	public String toString()
	{
		return getRoot().toString(0);
	}
}
