/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.xpath;

import java.util.Iterator;

/**
 * This class allows the construction of specific implementations of XTree.
 *
 * <p>Adding data to XTreeNode implementations is made possible by derivation
 * of this object.</p>
 * <p></B>Some methods in this base class may be overriden but they must absolutely
 * be called by derived implementations because it performs registration.</p>
 *  <p>Methods in this base class are synchronized because there is only one
 * instance of XTreeBuilder per tree so in case of thread shared building.</p>
 *
 */
public abstract class XTreeBuilder
{
	private static final String RCSRevision = "$Revision: 1.2 $";
	private static final String RCSName = "$Name: XQ1_1_1 $";
	
	protected XTree tree = null;

	protected XNode pattern = new XNode();

	/**
	 * Default constructor.
	 */
	public XTreeBuilder()
	{}

	/**
	 * Constructor for adding nodes to an existing tree.
	 * @param tree an existing tree to enrich.
	 */
	public XTreeBuilder(XTree tree)
	{
		this.tree = tree;
	}

	/**
	 * XTree factory. Calls {@link #allocateTree()};
	 * @return a new tree or the builder's tree if already initialized.
	 */
	public XTree createTree()
	{
		if (tree == null)
		{
			this.tree = allocateTree();
			tree.setRoot(createNode(null, NodeKind.NONE));
		}
		return tree;
	}

	/**
	 * Accessor to the builder's tree.
	 * @return XTree
	 */
	public XTree getTree()
	{
		return tree;
	}
	
	/**
	 * Creates an instance of <code>XTreeNode</code> or of its sub-class
	 * for qualified element and attributes nodes.
	 * @param parent the parent node of the new node
	 * @param namespace the namespace uri of the new node.
	 * @param localName the local name of the new node.
	 * @param type the type of the new node
	 */
	public synchronized XTreeNode createNamedNode(
		XTreeNode parent,
		String namespace,
		String localName,
		byte type)
	{
		return tree.register(allocateNode(parent, namespace, localName, type));
	}

	/**
	 * Creates an instance of <code>XTreeNode</code> or of its sub-class
	 * for unqualified element and attributes nodes.
	 * @param parent the parent node of the new node
	 * @param localName the local name of the new node.
	 * @param type the type of the new node
	 */
	public synchronized XTreeNode createNamedNode(
		XTreeNode parent,
		String localName,
		byte type)
	{
		return createNamedNode(parent, null, localName, type);
	}

	/**
	 * Creates an instance of <code>XTreeNode</code> or of its sub-class
	 * for unqualified element and attributes nodes if the node does not
	 * already exist.
	 * @param parent the parent node of the new node
	 * @param localName the local name of the new node.
	 * @param type the type of the new node
	 */
	public synchronized XTreeNode createNamedNodeIfNotExist(
		XTreeNode parent,
		String localName,
		byte type)
	{
		pattern.set(null, localName, type);
		XTreeNode node = parent.getChild(pattern);
		if (node == null)
			return createNamedNode(parent, localName, type);
		else
			return node;
	}

	/**
	 * Creates an instance of <code>XTreeNode</code> or of its sub-class
	 * for qualified element and attributes nodes if the node does not
	 * already exist.
	 * @param parent the parent node of the new node
	 * @param namespace the namespace uri of the new node.
	 * @param localName the local name of the new node.
	 * @param type the type of the new node
	 */
	public synchronized XTreeNode createNamedNodeIfNotExist(
		XTreeNode parent,
		String namespace,
		String localName,
		byte type)
	{
		pattern.set(namespace, localName, type);
		XTreeNode node = parent.getChild(pattern);
		if (node == null)
			return createNamedNode(parent, namespace, localName, type);
		else
			return node;
	}

	/**
	 * Creates an instance of <code>XTreeNode</code> or of its sub-class
	 * for root, comment & text nodes.
	 * @param parent the parent node of the new node
	 * @param type ROOT, COMMENT or TEXT
	 */
	public synchronized XTreeNode createNode(XTreeNode parent, byte type)
	{
		return tree.register(allocateNode(parent, null, null, type));
	}

	/**
	 * Creates an instance of <code>XTreeNode</code> or of its sub-class
	 * for Processing instructions nodes.
	 * @param parent the parent node of the new node
	 * @param target the PI target name or null
	 */
	public synchronized XTreeNode createPI(XTreeNode parent, String target)
	{
		return tree.register(allocateNode(parent, target, null, NodeKind.PI));
	}

	/**
	 * Creates an instance of <code>XTreeNode</code> or of its sub-class
	 * for Processing instructions nodes if a node does not already exist
	 * for this target.
	 * @param parent the parent node of the new node
	 * @param target the PI target name or null
	 */
	public synchronized XTreeNode createPIIfNotExist(
		XTreeNode parent,
		String target)
	{
		pattern.set(target);
		XTreeNode node = parent.getChild(pattern);
		if (node == null)
			return createPI(parent, target);
		else
			return node;
	}

	/**
	 * XTree factory to implement. Implementation should create a new XTree
	 * derived from XTree base implementation and calling the "super"
	 * constructor.
	 * @return XTreeNode
	 */
	public abstract XTree allocateTree();
	
	/**
	 * XTreeNode factory to implement. Implementation should create a new node
	 * derived from the base XTreeNode implementation and calling the "super"
	 * constructor that performs parent/child registration.
	 * @param parent parent node of the new node
	 * @param decl the XML schema declaration that will be used to create
	 * @return the new TypedXTreeNode (upcasted)
	 */
	public abstract XTreeNode allocateNode(
											XTreeNode parent,
											String namespace,
											String localName,
											byte type
											);

	/**
	 * Creates if necessary all the nodes of a given location expression.
	 * Location expression is normalized during process.
	 * @param location any location expression
	 */
	public void add(PathExpr location)
	{
		int index = 0;
		StepExpr step;
		XTreeNode parent = tree.getRoot();
		//Iterator it = location.normalize().iterator();
		Iterator it = location.getSteps().iterator();

		while (it.hasNext())
		{
			step = (StepExpr) it.next();
			parent =
				createNamedNodeIfNotExist(
					parent,
					step.getNameSpace(),
					step.getLocalName(),
					step.getTypeTest());
		}
	}

	/**
	 * Create an homothetic XTree using a distinct builder.
	 * @param builder the buider for homothetic XTree
	 */
	public void duplicate(XTreeBuilder builder)
	{
		tree.getRoot().duplicate(
			builder.getTree().getRoot().customizeDuplicate(tree.getRoot()),
			builder);
	}

}
