/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */


package org.xquark.xpath;

import java.util.Iterator;

import org.xml.sax.*;
import org.xml.sax.helpers.AttributesImpl;
import org.xquark.util.DefaultXMLReader;
import org.xquark.util.SAXConstants;

/**
 *  XTree SAX2 reader producing an XML document or fragment conform to the
 * collection element of the XMLDBC/Metadata schema.
 * <p>Special Features:</p>
 * <code>sax/features/fragment</code> disables SAX
 * startDocument() and endDocument() events triggering,
 * <code>sax/features/trace</code> calls the custom
 * method toXML(ContentHandler) on XTreeNode implementation to add trace
 * information.
 */
public class XTreeReader extends DefaultXMLReader implements SAXConstants
{
	private static final String RCSRevision = "$Revision: 1.2 $";
	private static final String RCSName = "$Name: XQ1_1_1 $";

	public static final String SOURCE_METADATA_NAMESPACE =
		"http://www.xquark.org/XMLDBC/Metadata";
	public static final String SOURCE_METADATA_PREFIX = "xm";
	public static final String NS_ATTR = "ns";
	public static final String NAME_ATTR = "name";
	public static final String TYPE_ATTR = "type";
	public static final String ATTRIBUTE_TYPE_VALUE = "attribute";
	public static final String ELEMENT_TYPE_VALUE = "element";
	public static final String STEP_ELEMENT_NAME = "step";
	public static final String COLLECTION_ELEMENT_NAME = "collection";

	private XTree tree;
	private boolean fragment = false;
	private boolean trace = false;
	private AttributesImpl atts = new AttributesImpl();

	/**
	 * Constructor.
	 * @param tree the XTree from which SAX events will be generated
	 */
	public XTreeReader(XTree tree)
	{
		this.tree = tree;
	}

	/**
	 * @see org.xml.sax.XMLReader#getFeature(String)
	 */
	public boolean getFeature(String name)
		throws SAXNotRecognizedException, SAXNotSupportedException
	{
		if (name.equals(FRAGMENT_FEATURE))
			return fragment;
		else if (name.equals(TRACE_FEATURE))
			return trace;
		else
			return super.getFeature(name);
	}

	/**
	 * @see org.xml.sax.XMLReader#setFeature(String, boolean)
	 */
	public void setFeature(String name, boolean value)
		throws SAXNotRecognizedException, SAXNotSupportedException
	{
		if (name.equals(FRAGMENT_FEATURE))
			fragment = value;
		else if (name.equals(TRACE_FEATURE))
			trace = value;
		else
			super.setFeature(name, value);
	}

	/**
	 * @see org.xml.sax.XMLReader#setProperty(String, Object)
	 */
	public void setProperty(String name, Object value)
		throws SAXNotRecognizedException, SAXNotSupportedException
	{
		if (name.equals(SAXConstants.SAX_LEXICAL_PROPERTY)) // ignored
			return;
		else
			super.setProperty(name, value);
	}

	/**
	 * Triggers SAX events generation.
	 * @param input is ignored.
	 */
	public void parse(InputSource input)
		throws SAXException, java.io.IOException
	{
		parse("");
	}

	/**
	 * Triggers SAX events generation.
	 * @param collectionName is passed as a root attribute during parsing.
	 */
	public void parse(String collectionName) throws SAXException, java.io.IOException
	{
		if (!fragment)
			contentHandler.startDocument();

		contentHandler.startPrefixMapping(
			SOURCE_METADATA_PREFIX,
			SOURCE_METADATA_NAMESPACE);

		if (collectionName != null)
			atts.addAttribute("", NAME_ATTR, "", "CDATA", collectionName);
		contentHandler.startElement(
			SOURCE_METADATA_NAMESPACE,
			COLLECTION_ELEMENT_NAME,
			"",
			atts);
		atts.clear();

		XTreeNode root = tree.getRoot();
		if (root.getChildren() != null)
		{
			Iterator it = root.getChildren().iterator();
			while (it.hasNext())
			{
				browseBranch((XTreeNode) it.next());
			}
		}
		contentHandler.endElement(
			SOURCE_METADATA_NAMESPACE,
			COLLECTION_ELEMENT_NAME,
			"");
		//        contentHandler.endPrefixMapping(WRAPPERMETADATA_PREFIX, WRAPPERMETADATA_NAMESPACE);

		contentHandler.endPrefixMapping(SOURCE_METADATA_PREFIX);

		if (!fragment)
			contentHandler.endDocument();
	}

	private void browseBranch(XTreeNode node)
		throws SAXException, java.io.IOException
	{
		if (node.getNamespace() != null)
			atts.addAttribute("", NS_ATTR, "", "CDATA", node.getNamespace());

		atts.addAttribute("", NAME_ATTR, "", "CDATA", node.getLocalName());

		if (node.getType() == NodeKind.ATTRIBUTE)
			atts.addAttribute("", TYPE_ATTR, "", "CDATA", ATTRIBUTE_TYPE_VALUE);
		// else : element is default

		contentHandler.startElement(
			SOURCE_METADATA_NAMESPACE,
			STEP_ELEMENT_NAME,
			"",
			atts);
		atts.clear();

		if (trace)
			node.toXML(contentHandler);

		// browse children
		if ((node.getType() == NodeKind.ELEMENT) && (node.getChildren() != null))
		{
			Iterator it = node.getChildren().iterator();
			while (it.hasNext())
			{
				browseBranch((XTreeNode) it.next());
			}
		}
		contentHandler.endElement(
			SOURCE_METADATA_NAMESPACE,
			STEP_ELEMENT_NAME,
			"");
	}

}
