/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Odonata.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.xpath.datamodel;

import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Date;

import org.xquark.schema.SchemaException;
import org.xquark.schema.datatypes.PrimitiveType;

/**
 *  <code>SchemaFormater</code> is a class for formatting Schema primitive types into user-friendly strings and backward.
 *  <P><strong>Examples Using the french Locale:</strong>
 *  <P>
 *  <pre>
 *  Type        Schema String           User-friendly String
 *  --------------------------------------------------------
 *  date        2001-01-15          >>  15 janvier 2001
 *              2001-01-15          <<  15/01/01
 *  time        13:30:25+02:00      >>  12:30:25 CET
 *  gMonth      --01--              >>  janvier
 *              --01--              <<  01
 *              --01--              <<  1
 *              --01--              <<  janv.
 *  </pre>
 *
 * @created    20 novembre 2001
 */
public class SchemaFormater {
	public static final String 	RCSRevision 	= "$Revision: 1.2 $";
	public static final String 	RCSName     	= "$Name: XQ1_1_1 $";

    /**
     *  Used for formatting primitive type <code>date</code> into user-friendly string.
     */
    protected static DateFormat[] dateFormats;
    /**
     *  Used for formatting user-friendly strings into primitive type <code>date</code>.
     */
    protected static DateFormat ufDateFormat = DateFormat.getDateInstance(DateFormat.LONG);

    /**
     *  Used for formatting primitive type <code>time</code> into user-friendly string.
     */
    protected static DateFormat[] timeFormats;
    /**
     *  Used for formatting user-friendly strings into primitive type <code>time</code>.
     */
    protected static DateFormat ufTimeFormat = DateFormat.getTimeInstance(DateFormat.LONG);

    /**
     *  Used for formatting primitive type <code>dateTime</code> into user-friendly string.
     */
    protected static DateFormat[] dateTimeFormats;
    /**
     *  Used for formatting user-friendly strings into primitive type <code>dateTime</code>.
     */
    protected static DateFormat ufDateTimeFormat = DateFormat.getDateTimeInstance(DateFormat.LONG, DateFormat.LONG);

    /**
     *  Used for formatting primitive type <code>gMonth</code> into user-friendly string.
     */
    protected static DateFormat[] gMonthFormats;
    /**
     *  Used for formatting user-friendly strings into primitive type <code>gMonth</code>.
     */
    protected static DateFormat ufGMonthFormat = new SimpleDateFormat("MMMM");

    /**
     *  Used for formatting primitive type <code>gDay</code> into user-friendly string.
     */
    protected static DateFormat[] gDayFormats;
    /**
     *  Used for formatting user-friendly strings into primitive type <code>gDay</code>.
     */
    protected static DateFormat ufGDayFormat = new SimpleDateFormat("d");

    /**
     *  Used for formatting primitive type <code>gYear</code> into user-friendly string.
     */
    protected static DateFormat[] gYearFormats;


    /**
     *  Formats a string entered by the user into a form conforming the given schema type.
     *
     * @param  input  The string entered by the user.
     * @param  type   The given schema type.
     * @return        The formatted string.
     */
    public static String toSchemaForm(String input, PrimitiveType type) {
        switch (type.getType()) {
            case PrimitiveType.DATE:
                for (int i = 0; i < dateFormats.length; i++) {
                    try {
                        return type.toXMLString(dateFormats[i].parse(input), null);
                    }
                    catch (ParseException e) {
                    }
                }
                break;
            case PrimitiveType.TIME:
                for (int i = 0; i < timeFormats.length; i++) {
                    try {
                        return type.toXMLString(timeFormats[i].parse(input), null);
                    }
                    catch (ParseException e) {
                    }
                }
                break;
            case PrimitiveType.DATE_TIME:
                for (int i = 0; i < dateTimeFormats.length; i++) {
                    try {
                        return type.toXMLString(dateTimeFormats[i].parse(input), null);
                    }
                    catch (ParseException e) {
                    }
                }
                break;
            case PrimitiveType.GMONTH:
                for (int i = 0; i < gMonthFormats.length; i++) {
                    try {
                        return type.toXMLString(gMonthFormats[i].parse(input), null);
                    }
                    catch (ParseException e) {
                    }
                }
                break;
            case PrimitiveType.GDAY:
                for (int i = 0; i < gDayFormats.length; i++) {
                    try {
                        return type.toXMLString(gDayFormats[i].parse(input), null);
                    }
                    catch (ParseException e) {
                    }
                }
                break;
            case PrimitiveType.GYEAR:
                for (int i = 0; i < gYearFormats.length; i++) {
                    try {
                        return type.toXMLString(gYearFormats[i].parse(input), null);
                    }
                    catch (ParseException e) {
                    }
                }
                break;
        }
        return input;
    }


    /**
     *  Formats the representation of a schema type into a user-friendly string.
     *
     * @param  input  The representation.
     * @param  type   The given schema type.
     * @return        The formatted string.
     */
    public static String toUserFriendlyForm(String input, PrimitiveType type) {
        switch (type.getType()) {
            case PrimitiveType.DATE:
                try {
                    return ufDateFormat.format((Date) type.convert(input, false, null));
                }
                catch (SchemaException e) {
                }
                break;
            case PrimitiveType.TIME:
                try {
                    return ufTimeFormat.format((Date) type.convert(input, false, null));
                }
                catch (SchemaException e) {
                }
                break;
            case PrimitiveType.DATE_TIME:
                try {
                    return ufDateTimeFormat.format((Date) type.convert(input, false, null));
                }
                catch (SchemaException e) {
                }
                break;
            case PrimitiveType.GMONTH:
                try {
                    return ufGMonthFormat.format((Date) type.convert(input, false, null));
                }
                catch (SchemaException e) {
                }
                break;
            case PrimitiveType.GDAY:
                try {
                    return ufGDayFormat.format((Date) type.convert(input, false, null));
                }
                catch (SchemaException e) {
                }
                break;
        }
        return input;
    }

    static {
        int[] styles = new int[]{DateFormat.FULL, DateFormat.LONG, DateFormat.MEDIUM, DateFormat.SHORT};
        int numStyles = styles.length;
        int i;
        int j;

        //dateFormats
        dateFormats = new DateFormat[numStyles];
        for (i = 0; i < numStyles; i++) {
            dateFormats[i] = DateFormat.getDateInstance(styles[i]);
        }

        //timeFormats
        timeFormats = new DateFormat[numStyles];
        for (i = 0; i < numStyles; i++) {
            timeFormats[i] = DateFormat.getTimeInstance(styles[i]);
        }

        //dateTimeFormats
        dateTimeFormats = new DateFormat[numStyles * numStyles];
        for (i = 0; i < numStyles; i++) {
            for (j = 0; j < numStyles; j++) {
                dateTimeFormats[i + j * numStyles] = DateFormat.getDateTimeInstance(styles[i], styles[j]);
            }
        }

        //gMonthFormats
        gMonthFormats = new DateFormat[]{new SimpleDateFormat("M"), new SimpleDateFormat("MMM"), new SimpleDateFormat("MMMM")};

        //gDayFormats
        gDayFormats = new DateFormat[]{new SimpleDateFormat("d")};

        //gYearFormats
        gYearFormats = new DateFormat[]{new SimpleDateFormat("y")};

    }

}
