/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.xpath.schema;

import org.xquark.schema.AttributeDeclaration;
import org.xquark.schema.Declaration;
import org.xquark.schema.ElementDeclaration;
import org.xquark.xpath.NodeKind;
import org.xquark.xpath.TypedXTreeNode;
import org.xquark.xpath.XTree;

/**
 * Implementation of XTreeNode for XML schema access through XPath indexing.
 */
public class SchemaNode extends TypedXTreeNode
{
	private static final String RCSRevision = "$Revision: 1.2 $";
	private static final String RCSName = "$Name: XQ1_1_1 $";
	
	Declaration declaration;

	/**
	 * Constructor for element and attributes nodes using names.
	 * Calls its ancestor constructor for registrations. Retrieves and sets
	 * the node declaration from the parent's one.
	 * @param tree the owner tree
	 * @param parent the parent node
	 * @param namespace the namespace URI of the new node
	 * @param localName the local name of the new node
	 * @param the type of the new node
	 */
	public SchemaNode(
		XTree tree,
		SchemaNode parent,
		String namespace,
		String localName,
		byte type)
	{
		super(tree, parent, namespace, localName, type);
		if ((type != NodeKind.NONE)
			&& (parent != null)
			&& (parent.getDeclaration() != null))
		{
			org.xquark.schema.Type parentType =
				parent.getDeclaration().getType();
			switch (type)
			{
				case NodeKind.ATTRIBUTE :
					declaration =
						parentType.getAttributeDeclaration(
							namespace,
							localName);
					break;
				case NodeKind.ELEMENT :
					declaration =
						parentType.getElementDeclaration(namespace, localName);
					break;
			}
		}
	}

	/**
	 * Constructor for element and attributes nodes using declaration.
	 * Calls its ancestor constructor for registrations. Sets name and type
	 * from the declaration passed as a parameter.
	 * @param tree the owner tree
	 * @param parent the parent node
	 * @param decl the node declaration defining the new node
	 */
	public SchemaNode(XTree tree, SchemaNode parent, Declaration decl)
	{
		super(tree, parent, decl.getNamespace(), decl.getName(), NodeKind.NODE);
		if (decl instanceof AttributeDeclaration)
			XModelNode.set(NodeKind.ATTRIBUTE);
		else if (decl instanceof ElementDeclaration)
			XModelNode.set(NodeKind.ELEMENT);
		declaration = decl;
	}

	/**
	 * Accessor to the node XML schema declaration.
	 * @return null if not available
	 */
	public Declaration getDeclaration()
	{
		return declaration;
	}

	/**
	 * Builds a String representation of the node.
	 * @returna String representation of the node.
	 */
	public String toString()
	{
		return super.toString() + "[Decl = " + declaration + "]";
	}
}
