/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.xquery.metadata;

import java.io.IOException;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.HashMap;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.*;
import org.xml.sax.helpers.AttributesImpl;
import org.xml.sax.helpers.DefaultHandler;
import org.xquark.schema.Schema;
import org.xquark.schema.loader.Loader;
import org.xquark.schema.validation.SchemaValidationContext;
import org.xquark.schema.validation.ValidatingSchemaFilter;
import org.xquark.util.NamespaceContextStack;
import org.xquark.xquery.DBModule;
import org.xquark.xquery.parser.*;
import org.xquark.xquery.typing.QType;
import org.xquark.xquery.typing.TypeException;
import org.xquark.xquery.typing.TypeVisitor;

// ***************************************************************************
// * Class 'MetaParser'
// ***************************************************************************
/**
 * Class to use to parse metadata represented as an XML file conforming to schema /org/xquark/xml/xdbc/resources/XMLDBCMetaData.xsd
 *  
 */
public class MetaParser extends DefaultHandler {

    // **********************************************************************
    // * Constant
    // **********************************************************************
    private static final String RCSRevision = "$Revision: 1.8 $";

    private static final String RCSName = "$Name: XQ1_1_1 $";

    // labels and attributes name of tag that we parse.
    private final static String TAG_METADATA = "metadata";

    //private final static String TAG_SCHEMAS = "schemas";
    private final static String TAG_SCHEMA = "schema";

    //private final static String TAG_MODULES = "modules";
    private final static String TAG_MODULE = "module";

    private final static String ATT_TARGETNAMESPACE = "targetNamespace";

    private final static String ATT_CATALOGNAME = "catalogName";

    private final static String ATT_SCHEMANAME = "schemaName";

    private final static String TAG_VARIABLE = "variable";

    private final static String TAG_FUNCTION = "function";

    private final static String TAG_XQUERYTYPE = "xquerytype";

    private final static String TAG_DECLARATION = "declaration";

    private final static String TAG_PARAMETERS = "parameters";

    private final static String TAG_PARAMETER = "parameter";

    private final static String ATT_SOURCE_NAME = "source";

    private final static String ATT_NAME = "name";

    private final static String ATT_DECLARED = "declared";

    // only for compatibility purposes
    private final static String TAG_COLLECTIONS = "collections";

    private final static String TAG_COLLECTION = "collection";

    private final static String TAG_STEP = "step";

    private final static String ATT_STEP_NS = "ns";

    private final static String ATT_STEP_TYPE = "type";

    private final static String ATT_STEP_TYPE_ATT = "attribute";

    private final static String ATT_STEP_TYPE_ELEMENT = "element";

    // **********************************************************************
    // * Fields
    // **********************************************************************
    // a reference of the metadata where we will insert metawrapper
    // and new schema we are going to discover.
    private MetaDataImpl metadata = null;

    // only for compatibility purposes
    // the current metacollection being parsed.
    private MetaCollection metacollection = null;

    // the current metawrapper being parsed.
    private MetaWrapper metawrapper = null;

    // the source name of the wrapper.
    private String sourcename = null;

    // Module
    private DBModule module = null;

    // Variable
    private Variable variable = null;

    // Function
    private FunctionDeclaration funcDecl = null;

    // declaration list (ordered)
    private ArrayList declList = null;

    // parameter list (ordered)
    private ArrayList paramList = null;

    // variables map
    private HashMap variables = null;

    // declaration list (ordered)
    private HashMap functions = null;

    // xquery type buffer
    private StringBuffer xqueryBuffer = new StringBuffer();

    // current qtype
    private QType qtype = null;

    // current SequenceType
    SequenceType seqType = null;

    // type parser
    XQueryParser parser = null;

    //VarCounter varCounter = null;

    // use during the parsing to see if we are in the context that we are
    // inside the parsing of a schema
    private boolean inSchema = false;

    private boolean inDataTobeParsed = false;

    private boolean isDeclared = false;

    // facilities when parsing schemas contained in the XMLDocument.
    // As prefix mapping are callbacked, *before* the callback
    // 'startElement' "schema", we should keep it for when we will see
    // the startElement 'schema'. In the preix mapping, we consider
    // prefixes and uri's. So we need 2 arraylist of strings to
    // keep them.
    private ArrayList tmppref = null;
    private ArrayList tmpuri = null;

    // probably temporary as well as schema doesn't correcty serialize.
    // keep the string representation of the schemas that we will parse.
    // private String strschema = null ;
    private StringBuffer bufSchema = new StringBuffer();

    // for generic and static use of SAX parser.
    private static SAXParserFactory spf = null;
    static {
        // Create a JAXP SAXParserFactory and configure it
        spf = SAXParserFactory.newInstance();
        spf.setNamespaceAware(true);
        spf.setValidating(false);
    }

    private XMLReader reader = null;

    private Loader loader = null;

    private SchemaValidationContext svContext = null;

    // **********************************************************************
    // * Constructor
    // **********************************************************************
    /**
     * Construct the metaparser of a wrapper XML description. As a wrapper has to be inserted into a metadata, keep a reference on metadata.
     */
    public MetaParser(MetaDataImpl metadata) {
        this.metadata = metadata;
        svContext = new SchemaValidationContext(metadata.getLoadingManager());
        tmppref = new ArrayList();
        tmpuri = new ArrayList();
    }

    // **********************************************************************
    // * Methods
    // **********************************************************************
    public void error(SAXParseException spe) {
        System.err.println("Caught SAXParseException : " + spe.getMessage());
    }

    private void createReader() throws SAXException, ParserConfigurationException {
        // Create a JAXP SAXParser
        SAXParser saxParser = spf.newSAXParser();

        // Get the encapsulated SAX XMLReader
        reader = saxParser.getXMLReader();
    }

    // **********************************************************************
    // Create a MetaWrapper
    // **********************************************************************
    /**
     * Create a wrapper from the SAX callbacks.
     */
    //    public MetaWrapper createWrapper(String uri) throws MetadataException {
    //        try {
    //            metawrapper = new MetaWrapper(metadata);
    //            createReader() ;
    //            
    //            ValidatingSchemaFilter filter = new ValidatingSchemaFilter(reader, svContext);
    //            filter.setContentHandler(this);
    //            filter.setErrorHandler(this);
    //            
    //            filter.parse(uri);
    //            
    //            return metawrapper;
    //        }
    //        catch (org.xml.sax.SAXParseException e) {}
    //        catch (SAXException e) {}
    //        catch (IOException e) {}
    //        catch (ParserConfigurationException e) {}
    //        throw new MetadataException();
    //    }
    /**
     * Create a wrapper from the SAX callbacks.
     */
    public MetaWrapper createWrapper(InputSource inputSource) throws MetadataException {
        try {
            metawrapper = new MetaWrapper(metadata);
            createReader();

            ValidatingSchemaFilter filter = new ValidatingSchemaFilter(reader, svContext);
            filter.setContentHandler(this);
            filter.setErrorHandler(this);
            filter.parse(inputSource);
            return metawrapper;
        } catch (org.xml.sax.SAXParseException e) {
            throw new MetadataException(e.getMessage());
        } catch (SAXException e) {
            throw new MetadataException(e.getMessage());
        } catch (IOException e) {
            throw new MetadataException(e.getMessage());
        } catch (ParserConfigurationException e) {
            throw new MetadataException(e.getMessage());
        }
    }

    public MetaWrapper createWrapper() throws MetadataException {
        try {
            metawrapper = new MetaWrapper(metadata);
            createReader();

            ValidatingSchemaFilter filter = new ValidatingSchemaFilter(reader, svContext);
            filter.setContentHandler(this);
            filter.setErrorHandler(this);

            return metawrapper;
        } catch (org.xml.sax.SAXParseException e) {
            throw new MetadataException(e.getMessage());
        } catch (SAXException e) {
            throw new MetadataException(e.getMessage());
        } catch (ParserConfigurationException e) {
            throw new MetadataException(e.getMessage());
        }
    }

    // **********************************************************************
    /**
     * Receive notification of the beginning of a document.
     * 
     * <p>
     * The SAX reader will invoke this method only once, before any other methods in this interface or in DTDHandler (except for setDocumentLocator).
     * </p>
     * 
     * @exception SAXException
     *                Any SAX exception, possibly wrapping another exception.
     */
    public void startDocument() throws SAXException {
        
    }

    public void startPrefixMapping(String prefix, String uri) throws SAXException {
        if (loader != null && inSchema) {
            loader.startPrefixMapping(prefix, uri);
        } else {
            tmppref.add(prefix);
            tmpuri.add(uri);
        }
    }

    public void endPrefixMapping(String prefix) throws SAXException {
        if (loader != null && inSchema) {
            loader.endPrefixMapping(prefix);
        } else {
            int index = tmppref.lastIndexOf(prefix);
            if (index != -1) {
                tmppref.remove(index);
                tmpuri.remove(index);
            }
        }
    }

    /**
     * Receive notification of the beginning of an element.
     */
    public void startElement(String uri, String local, String raw, Attributes attrs) throws SAXException {

        if (local.equalsIgnoreCase(TAG_SCHEMA)) {
            bufSchema.setLength(0);
            inSchema = true;
            loader = new Loader(metadata.getSchemaManager());
            loader.startDocument();

            // <serialisation of the strschema to keep in the cache>
            bufSchema.append("<?xml version=\"1.0\" encoding=\"iso-8859-1\"?>");
            bufSchema.append("<");
            if (uri != null) {
                int indexPref = tmpuri.indexOf(uri);
                if (indexPref >= 0) {
                    String pref = (String) tmppref.get(indexPref);
                    if (pref != null && pref.length() > 0) {
                        bufSchema.append(pref);
                        bufSchema.append(":");
                    }
                }
            }
            bufSchema.append(local);
            for (int i = 0; i < tmppref.size(); i++) {

                bufSchema.append(" xmlns");
                if (tmppref.get(i) != null && ((String) tmppref.get(i)).length() != 0)
                    bufSchema.append(":");
                bufSchema.append(tmppref.get(i));
                bufSchema.append("=\"");
                bufSchema.append(tmpuri.get(i));
                bufSchema.append("\"");

                loader.startPrefixMapping((String) tmppref.get(i), (String) tmpuri.get(i));
            }
            for (int i = 0; i < attrs.getLength(); i++) {
                bufSchema.append(" ");
                bufSchema.append(attrs.getLocalName(i));
                bufSchema.append("=\"");
                bufSchema.append(attrs.getValue(i));
                bufSchema.append("\"");
            }
            bufSchema.append(">");
            // </serialisation of the strschema to keep in the cache>

            loader.startElement(uri, local, (raw == null) ? "" : raw, (attrs == null) ? new AttributesImpl() : attrs);

            return;
        }

        if (inSchema) {
            bufSchema.append("<");
            if (uri != null) {
                int indexPref = tmpuri.indexOf(uri);
                if (indexPref >= 0) {
                    String pref = (String) tmppref.get(indexPref);
                    if (pref != null && pref.length() > 0) {
                        bufSchema.append(pref);
                        bufSchema.append(":");
                    }
                }
            }
            bufSchema.append(local);
            for (int i = 0; i < attrs.getLength(); i++) {
                bufSchema.append(" ");
                bufSchema.append(attrs.getLocalName(i));
                bufSchema.append("=\"");
                bufSchema.append(attrs.getValue(i));
                bufSchema.append("\"");
            }
            bufSchema.append(">");

            loader.startElement(uri, local, (raw == null) ? "" : raw, (attrs == null) ? new AttributesImpl() : attrs);
            return;
        }

        if (local.equalsIgnoreCase(TAG_METADATA)) {
            sourcename = attrs.getValue("", ATT_SOURCE_NAME);
            metawrapper.setSourceName(sourcename);
            //        } else if (local.equalsIgnoreCase(TAG_SCHEMAS)) {
            //        } else if (local.equalsIgnoreCase(TAG_MODULES)) {
        } else if (local.equalsIgnoreCase(TAG_MODULE)) {
            String declared = attrs.getValue("", ATT_DECLARED);
            isDeclared = (declared != null && declared.equalsIgnoreCase("true"));
            if (!isDeclared) {
                String targetnamespace = attrs.getValue("", ATT_TARGETNAMESPACE);
                module = (DBModule) metadata.getModuleManager().getModule(targetnamespace);
                if (module != null) {
                    throw new SAXException("Module with targetNamespace " + targetnamespace + " already exists");
                }
                module = new DBModule();
                StaticContext staticcontext = new StaticContext(metadata, metadata.getSchemaManager(), metadata.getModuleManager(), new VarCounter());
                try {
                    staticcontext.setTypeVisitor(new TypeVisitor(staticcontext, false));
                } catch (TypeException te) {
                    throw new SAXException(te);
                }
                module.setStaticContext(staticcontext);
                module.setCatalogName(attrs.getValue("", ATT_CATALOGNAME));
                module.setSchemaName(attrs.getValue("", ATT_SCHEMANAME));
                module.setNamespace(targetnamespace);
                module.setDefaultFunctionNamespace(targetnamespace);
                //                if (!tmppref.isEmpty()) {
                //                    NamespaceContextStack context = new NamespaceContextStack();
                //                    for (int i=0;i<tmppref.size();i++) {
                //                        context.declarePrefix((String)tmppref.get(i),(String)tmpuri.get(i));
                //                    }
                //                    module.setDeclarations(context);
                //                }
                metadata.getModuleManager().putModule(module);
            }
        } else if (local.equalsIgnoreCase(TAG_PARAMETER) || local.equalsIgnoreCase(TAG_VARIABLE)) {
            String name = attrs.getValue("", ATT_NAME);
            if (name != null) {
                int indexColon = name.indexOf(':');
                QName varName = null;
                try {
                    if (indexColon == -1)
                        varName = new QName(name, module);
                    else {
                        String qnamePrefix = name.substring(0, indexColon);
                        int indexPrefix = tmppref.indexOf(qnamePrefix);
                        if (indexPrefix == -1) {
                            int i=0;
                        }
                        String qnameUri = (String) tmpuri.get(indexPrefix);
                        varName = new QName(qnameUri, qnamePrefix, name.substring(indexColon + 1), module, null);
                    }
                    variable = new Variable(varName, module);
                } catch (TypeException te) {
                    throw new SAXException(te);
                } catch (XQueryException xqe) {
                    throw new SAXException(xqe);
                }
            }
        } else if (local.equalsIgnoreCase(TAG_FUNCTION)) {
            String name = attrs.getValue("", ATT_NAME);
            if (name != null) {
                int indexColon = name.indexOf(':');
                QName funcName = null;
                try {
                    if (indexColon == -1)
                        funcName = new QName(name, module);
                    else {
                        String qnamePrefix = name.substring(0, indexColon);
                        int indexPrefix = tmppref.indexOf(qnamePrefix);
                        String qnameUri = (String) tmpuri.get(indexPrefix);
                        funcName = new QName(qnameUri, qnamePrefix, name.substring(indexColon + 1), module, null);
                    }
                    funcDecl = new FunctionDeclaration(funcName, null, null, null, module);
                } catch (TypeException te) {
                    throw new SAXException(te);
                } catch (XQueryException xqe) {
                    throw new SAXException(xqe);
                }
            }
        } else if (local.equalsIgnoreCase(TAG_XQUERYTYPE)) {
            inDataTobeParsed = true;
        } else if (local.equalsIgnoreCase(TAG_DECLARATION)) {
            inDataTobeParsed = true;
        } else if (local.equalsIgnoreCase(TAG_PARAMETERS)) {
            paramList = new ArrayList();
            // only for compatibility purposes
        } else if (local.equalsIgnoreCase(TAG_COLLECTION)) {
            metacollection = new MetaCollection(metadata, attrs.getValue("", ATT_NAME));
        } else if (local.equalsIgnoreCase(TAG_STEP)) {
            String stepType = attrs.getValue("", ATT_STEP_TYPE);
            metacollection.addMetaElement(attrs.getValue("", ATT_STEP_NS), attrs.getValue("", ATT_NAME), stepType != null && stepType.equalsIgnoreCase(ATT_STEP_TYPE_ATT));
        }
    }

    /**
     * Receive notification of the end of an element.
     * 
     * <p>
     * The SAX reader will invoke this method at the end of every element in the XML document; there will be a corresponding startElement () event for every endElement () event (even when the element is empty).
     * </p>
     * 
     * <p>
     * For information on the names, see startElement.
     * </p>
     * 
     * @param uri
     *            The Namespace URI, or the empty string if the element has no Namespace URI or if Namespace processing is not being performed.
     * @param localName
     *            The local name (without prefix), or the empty string if Namespace processing is not being performed.
     * @param rawName
     *            The raw XML 1.0 name (with prefix), or the empty string if raw names are not available.
     * @exception SAXException
     *                Any SAX exception, possibly wrapping another exception.
     */
    public void endElement(String uri, String local, String raw) throws SAXException {
        if (local.equalsIgnoreCase(TAG_SCHEMA)) {
            bufSchema.append("</");
            if (uri != null) {
                int indexPref = tmpuri.indexOf(uri);
                if (indexPref >= 0) {
                    String pref = (String) tmppref.get(indexPref);
                    if (pref != null && pref.length() > 0) {
                        bufSchema.append(pref);
                        bufSchema.append(":");
                    }
                }
            }
            bufSchema.append(local);
            bufSchema.append(">");
            loader.endElement(uri, local, raw);
            inSchema = false;

            tmppref.clear();
            tmpuri.clear();

            if (loader != null) {
                loader.endDocument();
                Schema schema = loader.getSchema();
                if (metadata != null)
                    metadata.putSchema(schema, bufSchema.toString());
            }
            return;
        }

        if (inSchema) {
            loader.endElement(uri, local, raw);
            bufSchema.append("</");
            if (uri != null) {
                int indexPref = tmpuri.indexOf(uri);
                if (indexPref >= 0) {
                    String pref = (String) tmppref.get(indexPref);
                    if (pref != null && pref.length() > 0) {
                        bufSchema.append(pref);
                        bufSchema.append(":");
                    }
                }
            }
            bufSchema.append(local);
            bufSchema.append(">");
            return;
        }

        if (local.equalsIgnoreCase(TAG_MODULE)) {
            if (!isDeclared) {
                NamespaceContextStack context = new NamespaceContextStack();
                for (int i=0;i<tmppref.size();i++) {
                    context.declarePrefix((String)tmppref.get(i), (String)tmpuri.get(i));
                }
                module.setDeclarations(context);
                module.setDeclList(declList);
                if (variables != null && !variables.isEmpty()) {
                    HashMap mapVariables = new HashMap();
                    mapVariables.put(module.getNamespace(), variables);
                    module.setVariables(mapVariables);
                }
                if (functions != null && !functions.isEmpty()) {
                    HashMap mapFunctions = new HashMap();
                    mapFunctions.put(module.getNamespace(), functions);
                    module.setFunctions(mapFunctions);
                }
                metawrapper.addMetaModule(module);
                declList = null;
                variables = null;
                functions = null;
                module = null;
            }
        } else if (local.equalsIgnoreCase(TAG_PARAMETER) || local.equalsIgnoreCase(TAG_VARIABLE)) {
            if (qtype != null)
                variable.setQType(qtype);
            if (seqType != null)
                variable.setTypeDeclaration(seqType);
            if (paramList == null) {
                if (variables == null)
                    variables = new HashMap();
                variables.put(variable.getVarName().getLocalName().toString(), variable);
                if (declList == null)
                    declList = new ArrayList();
                declList.add(variable);
            } else {
                paramList.add(variable);
            }
            variable = null;
            qtype = null;
        } else if (local.equalsIgnoreCase(TAG_FUNCTION)) {
            if (qtype != null)
                funcDecl.setQType(qtype);
            if (seqType != null)
                funcDecl.setReturnType(seqType);
            funcDecl.setParameters(paramList);
            if (functions == null)
                functions = new HashMap();
            functions.put(funcDecl.getFuncName().getLocalName().toString(), funcDecl);
            if (declList == null)
                declList = new ArrayList();
            declList.add(funcDecl);
            paramList = null;
            funcDecl = null;
            qtype = null;
        } else if (local.equalsIgnoreCase(TAG_XQUERYTYPE)) {
            inDataTobeParsed = false;
            // create qtype from buffer
            try {
                if (parser == null) {
                    parser = new XQueryParser(new StringReader("dummy"));
                }
                parser.init(metadata, metadata.getSchemaManager(), module);
                parser.ReInit(new StringReader(xqueryBuffer.toString()));
                parser.token_source.switchState(XQueryParserConstants.ITEMTYPE);
                NamespaceContextStack context = parser.getNamespaceContextStack();
                context.pushContext();
                for (int i = 0; i < tmppref.size(); i++)
                    context.declarePrefix((String) tmppref.get(i), (String) tmpuri.get(i));
                seqType = parser.SequenceType();
                parser.token_source.popState();
                context.popContext();
                qtype = seqType.getQType();
            } catch (ParseException pe) {
                throw new SAXException(pe);
            } catch (XQueryException xqe) {
                throw new SAXException(xqe);
            }
            xqueryBuffer.setLength(0);
        } else if (local.equalsIgnoreCase(TAG_DECLARATION)) {
            inDataTobeParsed = false;
            // create declaration from buffer
            try {
                if (parser == null)
                    parser = new XQueryParser(new StringReader("dummy"));
                parser.ReInit(new StringReader(xqueryBuffer.toString()));
                XQueryModule tmpModule = parser.Start(metadata, metadata.getSchemaManager(), metadata.getModuleManager(), new VarCounter());
                metadata.getModuleManager().putModule(tmpModule);
                metawrapper.addMetaModule(tmpModule);
            } catch (ParseException pe) {
                throw new SAXException(pe);
            } catch (XQueryException xqe) {
                throw new SAXException(xqe);
            }
            xqueryBuffer.setLength(0);
        } else if (local.equalsIgnoreCase(TAG_COLLECTION)) {
            metawrapper.addMetaCollection(metacollection);
        } else if (local.equalsIgnoreCase(TAG_STEP)) {
            metacollection.addEndElement();
        }
    }

    /**
     * Receive notification of character data.
     * 
     * <p>
     * The Parser will call this method to report each chunk of character data. SAX readers may return all contiguous character data in a single chunk, or they may split it into several chunks; however, all of the characters in any single event must come from the same external entity so that the Locator provides useful information.
     * </p>
     * 
     * <p>
     * The application must not attempt to read from the array outside of the specified range.
     * </p>
     * 
     * <p>
     * Note that some readers will report whitespace using the ignorableWhitespace () method rather than this one (validating readers <em>must</em> do so).
     * </p>
     * 
     * @param ch
     *            The characters from the XML document.
     * @param start
     *            The start position in the array.
     * @param length
     *            The number of characters to read from the array.
     * @exception SAXException
     *                Any SAX exception, possibly wrapping another exception.
     * @see #ignorableWhitespace
     * @see org.xml.sax.Locator
     */
    public void characters(char ch[], int start, int length) throws SAXException {
        if (inDataTobeParsed)
            xqueryBuffer.append(ch, start, length);
    }

    /**
     * Receive notification of the end of a document.
     * 
     * <p>
     * The SAX reader will invoke this method only once, and it will be the last method invoked during the parse. The reader shall not invoke this method until it has either abandoned parsing (because of an unrecoverable error) or reached the end of input.
     * </p>
     * 
     * @exception SAXException
     *                Any SAX exception, possibly wrapping another exception.
     */
    public void endDocument() throws SAXException {
    }
}