/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Odonata.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.xquery.metadata;

import java.util.*;

import org.xquark.schema.SchemaManager;
import org.xquark.schema.datatypes.URI;
import org.xquark.util.NamespaceContextStack;
import org.xquark.xpath.StepExpr;
import org.xquark.xpath.XNode;
import org.xquark.xpath.XTreeNode;
import org.xquark.xquery.ModuleManager;
import org.xquark.xquery.metadata.resolver.CollectionMetadata;
import org.xquark.xquery.metadata.resolver.MetadataAccess;
import org.xquark.xquery.metadata.resolver.SourceMetadata;
import org.xquark.xquery.parser.*;
import org.xquark.xquery.typing.*;

public class StaticContext {

	public static final String XQUERY_FUNCTIONS_URI = "http://www.w3.org/2002/11/xquery-functions";
	public static final String XML_SCHEMA_URI = "http://www.w3.org/2001/XMLSchema";
	public static final String XML_SCHEMA_INSTANCE_URI = "http://www.w3.org/2001/XMLSchema-instance";

	private NamespaceContextStack prefixes = new NamespaceContextStack();
	private String defaultFunctionNamespace = XQUERY_FUNCTIONS_URI;
	private SchemaManager schemaManager;
	private ModuleManager moduleManager;
	private MetadataAccess metadataAccess = null;
	private HashMap functions = new HashMap();
	private HashMap collations = new HashMap();
	private String defaultCollation = null;
	private String baseURI = null;
	private boolean stripWhitespace = true;
	private TypeVisitor typeVisitor = null;
	
	private VarCounter varCounter = null;
	
	public StaticContext(MetadataAccess resolver, SchemaManager schemamanager, ModuleManager modulemanager, VarCounter varCounter) {
		this.metadataAccess = resolver;
		this.varCounter = varCounter;
		schemaManager = new SchemaManager(schemamanager);
		moduleManager = new ModuleManager(modulemanager);
		moduleManager.setStaticContext(this);
	}
	
	public VarCounter getVarCounter() {
		return varCounter;
	}	

	/**
	 * Returns the baseURI.
	 * @return String
	 */
	public String getBaseURI() {
		return baseURI;
	}

	/**
	 * Returns the defaultCollation.
	 * @return String
	 */
	public String getDefaultCollation() {
		return defaultCollation;
	}

	/**
	 * Returns the defaultFunctionNamespace.
	 * @return String
	 */
	public String getDefaultFunctionNamespace() {
		return defaultFunctionNamespace;
	}

	/**
	 * Returns the schemaManager.
	 * @return SchemaManager
	 */
	public SchemaManager getSchemaManager() {
		return schemaManager;
	}
	
	public void  setTypeVisitor(TypeVisitor typeVisitor) {
		this.typeVisitor = typeVisitor;
	}
	public TypeVisitor getTypeVisitor() {
		return typeVisitor;
	}

	/**
	 * Sets the baseURI.
	 * @param baseURI The baseURI to set
	 */
	public void setBaseURI(String baseURI) {
		this.baseURI = baseURI;
	}

	/**
	 * Sets the defaultCollation.
	 * @param defaultCollation The defaultCollation to set
	 */
	public void setDefaultCollation(String defaultCollation) {
		this.defaultCollation = defaultCollation;
	}

	/**
	 * Sets the defaultFunctionNamespace.
	 * @param defaultFunctionNamespace The defaultFunctionNamespace to set
	 */
	public void setDefaultFunctionNamespace(String defaultFunctionNamespace) {
		this.defaultFunctionNamespace = defaultFunctionNamespace;
	}

	/**
	 * Returns the collations.
	 * @return HashMap
	 */
	public HashMap getCollations() {
		return collations;
	}

	/**
	 * Returns the functions.
	 * @return HashMap
	 */
	public HashMap getFunctions() {
		return functions;
	}

	/**
	 * Returns the prefixes.
	 * @return NamespaceContextStack
	 */
	public NamespaceContextStack getPrefixes() {
		return prefixes;
	}

	/**
	 * Returns the module manager.
	 * @return ModuleManager
	 */
	public ModuleManager getModuleManager() {
		return moduleManager;
	}
//	public void setModuleManager(ModuleManager modulemanager) {
//		this.modulemanager = modulemanager;
//	}

	/**
	 * Returns the variables.
	 * @return HashMap
	 */
	/*
	public ScopeManager getVariables() {
		return variables;
	}
	*/

	/**
	 * Returns the stripWhitespace.
	 * @return boolean
	 */
	public boolean isStripWhitespace() {
		return stripWhitespace;
	}

	/**
	 * Sets the stripWhitespace.
	 * @param stripWhitespace The stripWhitespace to set
	 */
	public void setStripWhitespace(boolean stripWhitespace) {
		this.stripWhitespace = stripWhitespace;
	}

	// XPathResolver interface

	public ArrayList resolveXPath(ArrayList xnodes, ArrayList steps) {
		if (steps == null || steps.isEmpty())
			return xnodes;
		for (int i = 0; i < steps.size(); i++) {
			xnodes = resolveXPath(xnodes, (StepExpr) steps.get(i));
		}
		return xnodes;
	}

	public ArrayList resolveXPath(ArrayList xnodes, StepExpr step) {
		if (xnodes == null || xnodes.isEmpty())
			return null;
		ArrayList reslist = new ArrayList();
		for (int i = 0; i < xnodes.size(); i++) {
			Collection col = ((XTreeNode) xnodes.get(i)).navigate(step);
			reslist.addAll(col);
		}
		if (reslist.isEmpty())
			reslist = null;
		return reslist;
	}

	// this method assumes that steps are simple
	public QType resolveQType(byte context, QType qtype, StepExpr step, boolean inPredicate) throws TypeException, XQueryException {
		//return resolver.resolveQType(qtype,step);
		return Utils.getSubType(context, qtype, step, schemaManager, inPredicate);
	}

	private final static String COLON = ":";
	private final static String STAR = "*";
	private final static String DOT = ".";

	// defined for compatibility purposes
	public ArrayList getCollectionRootNodes(String sourceName, String collectionName) {

		ArrayList al = new ArrayList();
		SourceMetadata mw = null;
		mw = metadataAccess.getSourceMetadata(sourceName);
		if (mw != null) {
			CollectionMetadata metacol = mw.getCollectionMetadata(collectionName);
			if (metacol != null)
				al.add(metacol.getXTree().getRoot());
			else { // treat case where collection name ends with .collectionName
				Collection coll = mw.getCollectionsMetadata();
				for (Iterator it = coll.iterator(); it.hasNext();) {
					metacol = (CollectionMetadata) it.next();
					String colName = metacol.getCollectionName();
					if (STAR.equals(collectionName) || colName.endsWith(DOT + collectionName)) {
						//metacol = mw.getCollectionMetadata(colName);
						al.add(metacol.getXTree().getRoot());
						//break;
					}
				}
			}
		} else {
			Collection sources = metadataAccess.getSources();
			for (Iterator e = sources.iterator(); e.hasNext();) {
				mw = (SourceMetadata) e.next();
				CollectionMetadata metacol = mw.getCollectionMetadata(collectionName);
				if (metacol != null)
					al.add(metacol.getXTree().getRoot());
				else { // treat case where collection name ends with .collectionName
					Collection coll = mw.getCollectionsMetadata();
					if (coll != null) {
						for (Iterator it = coll.iterator(); it.hasNext();) {
							metacol = (CollectionMetadata) it.next();
							String colName = metacol.getCollectionName();
							if (STAR.equals(collectionName) || colName.endsWith(DOT + collectionName)) {
								//metacol = mw.getCollectionMetadata(colName);
								al.add(metacol.getXTree().getRoot());
								//break;
							}
						}
					}
				}
			}
		}
		if (al.isEmpty())
			al = null;
		return al;
	}

	public MetadataAccess getMetaData() {
		return metadataAccess;
	}

	public XNode getDocumentRootNode(URI documentURI) {
		return null;
	}

	public ArrayList getInputRootNodes() {
		return null;
	}
	
	private final String VAR_NAME = "var";

	public Variable createVariable(int bindingType, XQueryExpression expr, byte creator) throws XQueryException {
		if (typeVisitor != null && expr != null && expr.getQType() == null)
			expr.accept(typeVisitor);
		Variable var = new Variable(getVarName(expr.getParentModule()), expr.getParentModule());
		var.setBindingType(bindingType);
		var.setExpression(expr,true);
		var.setCreator(creator);
		return var;
	}
	public Variable createVariable(XQueryModule module, int bindingType, XQueryExpression expr, byte creator) throws XQueryException {
		if (typeVisitor != null && expr != null && expr.getQType() == null)
			expr.accept(typeVisitor);
		Variable var = new Variable(getVarName(module), module);
		var.setBindingType(bindingType);
		var.setExpression(expr,true);
		var.setCreator(creator);
		return var;
	}
	public QName getVarName(XQueryModule parentModule) throws XQueryException {
		return new QName(parentModule.getNamespace(), parentModule.getPrefix(), VAR_NAME + (++varCounter.lastVarIndex), parentModule, parentModule.getStaticContext().getPrefixes());
	}

}
