/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.xquery.normalize;

/**
 * This class is used to store and retrieve a context informations inside the expression normalize visitor.<BR>
 * Context information are stored in a hashmap: the key is the question (or the order), the value is an object that
 * is the (possibly) answer of the question (or the data to use for the order).<BR>
 * IMPORTANT: the question is a string like &lt;depth&gt;'.'&lt;node name&gt;["."&lt;question&gt;] where:<BR><UL>
 * <LI>&lt;depth&gt; is the absolute depth where is the node that must answer the question. The wildcard '*' can be used to
 * mean that the depth is indifferent.</LI>
 * <LI>&lt;node name&gt; is the corresponding Class name that must answer the question (if its depth equals &lt;depth&gt;).</LI>
 * <LI>&lt;question&gt; a possible question. If there is no question, the implicite question is "is there a &lt;node name&gt;
 * at the given &lt;depth&gt;?".</LI>
 * </UL>Examples of questions:<BR><UL>
 * <LI>"5.FLWRExpression": is there a FLWRExpresion node in the absolute depth 5? Can return true or false (for example).</LI>
 * <LI>"*.FLWRExpression": is there a FLWRExpresion node in any depth? Can return true or false (for example).</LI>
 * <LI>"3.XQueryExpressionSequence.containsOnlyOneChild": is there a XQueryExpressionSequence in the absolute depth 3 and
 * have it only one child expression? Can return true or false (for example).</LI>
 * <LI>"10.Variable.getExpression": If there is a Variable node in the absolute depth 10, get me the associeted expression.
 * Can return null or an expression.</LI>
 * </UL><BR>
 */
public class VisitorContext {

    /** CVS/QA stuff. */


/*************************** Package visibility constants *******************************/

    /** Constants for specific keys. */

    static final int _XQUERYEXPRESSIONSEQUENCE_EXISTS = 1;
    //static final int _XQUERYEXPRESSIONSEQUENCE_ELEMENT_CHILDREN = 2;

    static final int _VARIABLE_DECLARATION = 3;
    static final int _VARIABLE_SUBSTITUTION4QUANTIFIED = 4;
    static final int _VARIABLE_SUBSTITUTION4FLWR = 5;
    static final int _VARIABLE_SUBSTITUTION4LETFLWR = 6;
    static final int _VARIABLE_WASSUBSTITUTELETFLWR = 7;
    static final int _VARIABLE_CHANGEEXPRESSION4FLWR = 8;

    static final int _FLWREXPRESSION_GETFLWR = 9;
	static final int _FLWREXPRESSION_INNERFLWR = 10;
//    static final int _FLWREXPRESSION_GETVARCLAUSE = 10;
//    static final int _FLWREXPRESSION_GETWHERECLAUSE = 11;
//	static final int _FLWREXPRESSION_GETORDERBYCLAUSE = 12;
//    static final int _FLWREXPRESSION_GETRETURNCLAUSE = 13;
	static final int _QUANTIFIED_INNERFLWR = 14;
//    static final int _FLWREXPRESSION_GETVARCLAUSE4QUANTIFIED = 14;
//    static final int _FLWREXPRESSION_GETWHERECLAUSE4QUANTIFIED = 15;
//    static final int _FLWREXPRESSION_GETRETURNCLAUSE4QUANTIFIED = 16;

    static final int _FLWREXPRESSION_GETANDREPLACEFLWRINWHERE = 17;

    static final int _ELEMENT_EXISTS = 18;
    //static final int _ELEMENT_CHILDREN = 19;

    //static final int _LOCATEDEXPRESSION_DONOTNORMALIZE = 18;
    static final int _LOCATEDEXPRESSION_INPREDICATE = 20;

    static final int _LISTOPUNIONEXPRESSION_GETEXPRESSIONS = 21;

    static final int _INCLUDEINPARENTHESIS = 22;

    static final int _STEP_ISPARENTAXISSTEP = 23;
    static final int _STEP_GETPREDICATE = 24;

    static final int _INAGGREGATEFUNCTION = 25;

    static final int _INTERNALFUNCTIONCALL_NAMEDUNORDERED_PARAMETERS = 26;

    static final int _MULTIVALUED_NEEDEXISTSFUNCTION = 27;

	static final int _FLWR_RETURNS_TAG = 28;
	// add LARS 16/04/04
	static final int _STEP_GETPREDICATE_VAR = 29;
	
    /**
     * The maximum number of differents questions.
     * MUST BE greater (strictly) than the followings number of constants.
     */
    static final int MAX_QUESTIONS = 32;


    /** To specify any depth in structure. */
    static final int _ANY_DEPTH = 0;

/*************************** Protected properties *******************************/

    /** A hashmap to store couple of (constants string, value), the key is a question, the value its response. */
    protected java.util.HashMap memoMap = null;

/*************************** Constructors *******************************/

    /**
     * To construct a new empty visitor context.
     * @throws NormalizeException
     */
    public VisitorContext() throws NormalizeException {
        //trace("<o>Constructor VisitorContext 1.",0);
        this.memoMap = new java.util.HashMap();
    }

    /**
     * To construct a new visitor context with the given memo.
     * @param depth the destination depth in the structure (0 for any depth).
     * @param question the key (a question) of a memo couple (see class header).
     * @param response the value (the response) or a memo couple.
     * @throws NormalizeException
     */
    public VisitorContext(int depth, int question, java.lang.Object response) throws NormalizeException {
        //trace("<o>Constructor VisitorContext 1.",0);
        this.memoMap = new java.util.HashMap();
        this.setMemo(depth, question, response, true);
    }

/*************************** Public methods *******************************/

    /**
     * To get a memo value.
     * @param depth the destination depth in the structure (0 for any depth).
     * @param question the key (a question) of a memo couple (see class header).
     * @return the value (an object) of the specified memo couple. Caution: return 'null' when
     * the key does not exist or if the value is 'null'. Use existsMemo(...) to know if a key exists.
     */
    public java.lang.Object getMemo(int depth, int question) throws NormalizeException {
        Integer key = new Integer(depth * VisitorContext.MAX_QUESTIONS + question);
        if (this.memoMap.containsKey(key)) {
            return this.memoMap.get(key);
        } else {
            return null;
        }
    }

    /**
     * To know if a memo couple exits in the current context.
     * @param depth the destination depth in the structure (0 for any depth).
     * @param question the key (a question) of a memo couple (see class header).
     * @return true if the memo couple exists, false otherwise.
     */
    public boolean existsMemo(int depth, int question) throws NormalizeException {
        Integer key = new Integer(depth * VisitorContext.MAX_QUESTIONS + question);
        return this.memoMap.containsKey(key);
    }

    /**
     * To remove a memo couple of the current context.
     * @param depth the destination depth in the structure (0 for any depth).
     * @param question the key (a question) of a memo couple (see class header).
     * @return true if the memo couple is deleting, false otherwise.
     */
    public boolean removeMemo(int depth, int question) throws NormalizeException {
        Integer key = new Integer(depth * VisitorContext.MAX_QUESTIONS + question);
        if (this.memoMap.containsKey(key)) {
            this.memoMap.remove(key);
            return true;
        } else {
            return false;
        }
    }

    /**
     * To set a memo in this context.
     * @param depth the destination depth in the structure (0 for any depth).
     * @param question the key (a question) of a memo couple (see class header).
     * @param response the value (the response) of the specified memo couple (can be 'null').
     * @param overwrite if true, overwrite the previous value if exist. If false, this method throws an
     * NormalizeException if the memo already exist, else create a new memo.
     */
    public void setMemo(int depth, int question, java.lang.Object response, boolean overwrite) throws NormalizeException {
        Integer key = new Integer(depth * VisitorContext.MAX_QUESTIONS + question);
        if ((!overwrite) && (this.memoMap.containsKey(key))) {
            throw new NormalizeException("The memo couple \""+depth+"."+question+"\" already exist!");
        } else {
            this.memoMap.put(key, response);
        }
    }

    /**
     * To set a memo in this context.
     * @param key a number that is formed with depth and question integers.
     * @param response the value (the response) of the specified memo couple (can be 'null').
     * @param overwrite if true, overwrite the previous value if exist. If false, this method throws an
     * NormalizeException if the memo already exist, else create a new memo.
     */
    private void setMemo(Integer key, java.lang.Object response, boolean overwrite) throws NormalizeException {
        if ((!overwrite) && (this.memoMap.containsKey(key))) {
            int aKey = key.intValue();
            int depth = (int)(aKey / VisitorContext.MAX_QUESTIONS);
            int question = (int)(aKey % VisitorContext.MAX_QUESTIONS);
            throw new NormalizeException("The memo couple \""+depth+"."+question+"\" already exist!");
        } else {
            this.memoMap.put(key, response);
        }
    }

    /**
     * Add all given context memo couples in the current context memo.
     * @param ctx a context to merge with the current context.
     * @param keepPrevious if a memo couple already exist and has a different value of a given memo couple, and if
     * keepPrevious is true, we keep the value of the current memo couple. Otherwise, we put the given memo
     * couple value (create or replace).
     */
    public void merge(VisitorContext ctx, boolean keepPrevious) throws NormalizeException {
        Object[] keys = ctx.memoMap.keySet().toArray();
        for (int i = 0; i < keys.length; i++) {
            Integer aKey = (Integer)keys[i];
            if ((!this.memoMap.containsKey(aKey)) || (!keepPrevious)) {
                this.memoMap.put(aKey, ctx.memoMap.get(aKey));
            }
        }
    }

    /**
     * Update the values of all context memo couples of the current context with the given context
     * (but do not create new memo couple in current context).
     * @param ctx a context to use to update the current context values.
     */
    public void update(VisitorContext ctx) throws NormalizeException {
        Object[] keys = this.memoMap.keySet().toArray();
        for (int i = 0; i < keys.length; i++) {
            Integer aKey = (Integer)keys[i];
            if (ctx.memoMap.containsKey(aKey)) {
                if (aKey.intValue() == VisitorContext._FLWREXPRESSION_GETANDREPLACEFLWRINWHERE) {
                    //Merge two ArrayList :
                    java.util.ArrayList content1 = (java.util.ArrayList)this.memoMap.get(aKey);
                    java.util.ArrayList content2 = (java.util.ArrayList)ctx.memoMap.get(aKey);
                    if ((content1 != null) && (content2 != null)) {
                        for (int j = 0; j < content2.size(); j++) {
                            if (!content1.contains(content2.get(j))) {
                                content1.add(content2.get(j));
                            }
                        }
                    }
                    this.memoMap.put(aKey, content1);
                } else {
                    this.memoMap.put(aKey, ctx.memoMap.get(aKey));
                }
            }
        }
    }

    /**
     * To filter the current context i.e. to remove all the memo couples of which the depth (in the key) is
     * stricly smaller than the given depth (the depth 0 for 'any depth' is considered greater than all depth).<BR><BR>
     * The questions must have the following prototype: &lt;depth&gt;'.'&lt;node name&gt;["."&lt;question&gt;](see class header).<BR><BR>
     * It is use to optimize the context.
     * @param currentDepth the current depth of the visitor.
     */
    public void filter(int currentDepth) throws NormalizeException {
       Object[] keys = this.memoMap.keySet().toArray();
        for (int i = 0; i < keys.length; i++) {
            Integer aKey = (Integer)keys[i];
            int depth = (int)(aKey.intValue() / VisitorContext.MAX_QUESTIONS);
            if ((depth != 0) && (depth < currentDepth)) {
                this.memoMap.remove(aKey);
            }
        }
    }

    /**
     * Removes all memo couple of the current context. Cleans the current Context.
     */
    public void clear() {
        this.memoMap.clear();
    }

    /**
     * To get the current memoMap to string.
     * @return a string that represents the current memoMap.
     */
    public java.lang.String toString() {
        java.lang.String ret = "";
        Object[] keys = this.memoMap.keySet().toArray();
        for (int i = 0; i < keys.length; i++) {
            Integer aKey = (Integer)keys[i];
            int aKeyAsInt = aKey.intValue();
            int depth = (int)(aKeyAsInt / VisitorContext.MAX_QUESTIONS);
            int question = (int)(aKeyAsInt % VisitorContext.MAX_QUESTIONS);
            ret += depth+"."+question+" - "+this.memoMap.get(aKey)+"\n";
        }
        if (ret.equals("")) {
            return "EMPTY CONTEXT";
        }
        return ret;
    }

    /**
     * Clones the current context.
     * @return a clone of the current context.
     */
    public VisitorContext cloneContext() throws NormalizeException {
        VisitorContext clone = new VisitorContext();
        Object[] keys = this.memoMap.keySet().toArray();
        for (int i = 0; i < keys.length; i++) {
            Integer aKey = (Integer)keys[i];
            int aKeyAsInt = aKey.intValue();
            int depth = (int)(aKeyAsInt / VisitorContext.MAX_QUESTIONS);
            int question = (int)(aKeyAsInt % VisitorContext.MAX_QUESTIONS);
            clone.setMemo(depth, question,this.memoMap.get(aKey), false);
        }
        return clone;
    }
}
