/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.xquery.parser;

import java.util.ArrayList;

import org.xquark.xquery.parser.hinttree.HintTree;
import org.xquark.xquery.parser.util.Constants;
import org.xquark.xquery.typing.TypeException;

public class FLWRExpression extends XQueryExpression implements Cloneable {

	private static final String RCSRevision = "$Revision: 1.13 $";
	private static final String RCSName = "$Name: XQ1_1_1 $";

	// cannot be null or empty
	protected ArrayList variables = null;
	protected XQueryExpression whereClause = null;
	// cannot be null
	protected XQueryExpression returnClause = null;
	// can be null
	protected ArrayList orderby = null;
	protected ArrayList listWhereAdded = new ArrayList();
	protected HintTree hinttree = null;
	// #############################################################################
	// VISITOR STUFF
	// #############################################################################

	public void accept(ParserVisitor visitor) throws XQueryException {
		visitor.visit(this);
	}

	// #############################################################################
	// CONSTRUCTOR STUFF
	// #############################################################################

	public FLWRExpression(ArrayList variables, XQueryExpression whereClause, XQueryExpression returnClause, XQueryModule parentModule) throws TypeException, XQueryException {
		setVariables(variables);
		setWhereClause(whereClause);
		setReturnClause(returnClause,false);
		setParentModule(parentModule);
		if (parentModule != null && parentModule.getStaticContext().getTypeVisitor() != null)
			accept(parentModule.getStaticContext().getTypeVisitor());
	}

	public FLWRExpression(ArrayList variables, ArrayList orderby, XQueryExpression whereClause, XQueryExpression returnClause, XQueryModule parentModule) throws TypeException, XQueryException {
		setVariables(variables);
		setOrderBy(orderby);
		setWhereClause(whereClause);
		setReturnClause(returnClause,false);
		setParentModule(parentModule);
		if (parentModule != null && parentModule.getStaticContext().getTypeVisitor() != null)
			accept(parentModule.getStaticContext().getTypeVisitor());
	}

	public FLWRExpression(ArrayList variables, HintTree hintClause, XQueryExpression whereClause, ArrayList orderby, XQueryExpression returnClause, XQueryModule parentModule) throws TypeException, XQueryException {
		setVariables(variables);
		setHintTree(hintClause);
		setWhereClause(whereClause);
		setOrderBy(orderby);
		setReturnClause(returnClause,false);
		setParentModule(parentModule);
		if (parentModule != null && parentModule.getStaticContext().getTypeVisitor() != null)
			accept(parentModule.getStaticContext().getTypeVisitor());
	}

	public FLWRExpression(ArrayList variables, HintTree hintClause, ArrayList orderby, XQueryExpression whereClause, XQueryExpression returnClause, XQueryModule parentModule) throws TypeException, XQueryException {
		setVariables(variables);
		setHintTree(hintClause);
		setOrderBy(orderby);
		setWhereClause(whereClause);
		setReturnClause(returnClause,false);
		setParentModule(parentModule);
		if (parentModule != null && parentModule.getStaticContext().getTypeVisitor() != null)
			accept(parentModule.getStaticContext().getTypeVisitor());
	}

	// #############################################################################
	// ATTRIBUTE ACCESS STUFF
	// #############################################################################

	public ArrayList getVariables() {
		return variables;
	}
	public void setVariables(ArrayList variables) throws XQueryException {
		// cannot be null or empty
		if (variables == null || variables.isEmpty())
			throw new XQueryException("variables of FLWRExpression cannot be null or empty");
		this.variables = variables;
		for (int i = 0; i < this.variables.size(); i++) {
			Variable var = (Variable) this.variables.get(i);
			var.setParentFLWRExpression(this);
			var.setParentModule(parentModule);
		}
	}
	public void insertVariable(Variable variable) {
		// the existence of the same variable could be tested !!!
		variable.setParentFLWRExpression(this);
		variable.setParentModule(parentModule);
		variables.add(0, variable);
	}
	public void addVariable(Variable variable) {
		// the existence of the same variable could be tested !!!
		variable.setParentFLWRExpression(this);
		variable.setParentModule(parentModule);
		variables.add(variable);
	}
	public void addVariables(ArrayList varlist) {
		if (varlist == null || varlist.isEmpty())
			return;
		for (int i = 0; i < varlist.size(); i++)
			addVariable((Variable) varlist.get(i));
	}
	public void addVariable(int index, Variable variable) {
		// the existence of the same variable could be tested !!!
		variable.setParentFLWRExpression(this);
		variable.setParentModule(parentModule);
		variables.add(index, variable);
	}
	public void addVariables(int index, ArrayList varlist) {
		if (varlist == null || varlist.isEmpty())
			return;
		for (int i = 0; i < varlist.size(); i++)
			addVariable(index+i, (Variable) varlist.get(i));
	}

	public ArrayList getOrderBy() {
		return orderby;
	}
	public void setOrderBy(ArrayList orderby) {
		this.orderby = orderby;
	}

	public void addOrderBy(XQueryExpression expr) {
		// the existence of the same variable could be tested !!!
		expr.setParentExpression(this);
		expr.setParentModule(parentModule);
		if (orderby == null)
			orderby = new ArrayList(1);
		orderby.add(expr);
	}

	public void addOrderBy(int index, XQueryExpression expr) {
		// the existence of the same variable could be tested !!!
		expr.setParentExpression(this);
		expr.setParentModule(parentModule);
		if (orderby == null)
			orderby = new ArrayList(1);
		orderby.add(index, expr);
	}

	public XQueryExpression getReturnClause() {
		return returnClause;
	}
	public void setReturnClause(XQueryExpression returnClause) throws XQueryException {
		setReturnClause(returnClause,true);
	}
	public void setReturnClause(XQueryExpression returnClause, boolean typing) throws XQueryException {
		// cannot be null
		if (returnClause == null)
			throw new XQueryException("returnClause of FWLRExpression cannot be null");
		this.returnClause = returnClause;
		this.returnClause.setParentExpression(this);
		if (typing && parentModule != null && parentModule.getStaticContext().getTypeVisitor() != null) {
			qtype = null;
			accept(parentModule.getStaticContext().getTypeVisitor());
		}
	}

	public XQueryExpression getWhereClause() {
		return whereClause;
	}
	public void setWhereClause(XQueryExpression whereClause) {
		this.whereClause = whereClause;
		if (this.whereClause != null) {
			this.whereClause.setParentExpression(this);
			this.whereClause.setParentModule(parentModule);
		}
	}

	public int getVarPosition(Variable var) {
		if (variables == null)
			return -1;
		return variables.indexOf(var);
	}

	public ArrayList getXTrees() {
		return returnClause.getXTrees();
	}

	public HintTree getHintTree() {
		return hinttree;
	}

	public void setHintTree(HintTree hinttree) {
		this.hinttree = hinttree;
	}

	private int GetFirstForVariable(int index) {
		for (int i = index; i < variables.size(); i++) {
			if (((Variable) variables.get(i)).getBindingType() == Constants.FOR_BINDINGTYPE)
				return i;
		}
		return -1;
	}

	// #############################################################################
	// CLONE STUFF
	// #############################################################################

	public void addParentExpression(XQueryExpression parentExpression) {
		addParentExpression(parentExpression);

		if (variables != null)
			for (int i = 0; i < variables.size(); i++)
				 ((Variable) variables.get(i)).addParentExpression(parentExpression);
		if (whereClause != null)
			whereClause.addParentExpression(parentExpression);
		if (returnClause != null)
			returnClause.addParentExpression(parentExpression);
	}

	// #############################################################################
	// RESTRUCTURE STUFF
	// #############################################################################
	public void insertVariableAtPosition(Variable var, int position) {
		variables.add(position, var);
		var.setParentFLWRExpression(this);
		var.setParentModule(parentModule);
	}

	public void addWhereClause(XQueryExpression addWhereClause) throws XQueryException {
		if (addWhereClause == null)
			return;
		if (whereClause == null) {
			listWhereAdded.add(addWhereClause);
			setWhereClause(addWhereClause);
			return;
		}
		if (!listWhereAdded.contains(addWhereClause)) {
			listWhereAdded.add(addWhereClause);
			if (whereClause instanceof BinOpORExpression)
			    whereClause = new ParenthesizedExpression(whereClause, parentModule);
			if (addWhereClause instanceof BinOpORExpression)
			    addWhereClause = new ParenthesizedExpression(addWhereClause, parentModule);
			setWhereClause(new BinOpANDExpression(whereClause, addWhereClause, parentModule));
		}
	}

}
