/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.xquery.parser;

import java.util.ArrayList;

import org.xquark.schema.Type;
import org.xquark.xpath.*;
import org.xquark.xquery.typing.TypeException;

public class Step extends XQueryUnaryOperatorExpression implements StepExpr {
	private static final String RCSRevision = "$Revision: 1.14 $";
	private static final String RCSName = "$Name: XQ1_1_1 $";

	public static final int NONE					= 0;
	
	public static final int SELF_QNAME 					= 0x10;
	public static final int CHILD_QNAME 				= 0x20;
	public static final int PARENT_QNAME 				= 0x30;
	public static final int ANCESTOR_QNAME 				= 0x40;
	public static final int ANCESTOR_OR_SELF_QNAME 		= 0x50;
	public static final int DESCENDANT_QNAME 			= 0x60;
	public static final int DESCENDANT_OR_SELF_QNAME	= 0x70;
	public static final int PRECEDING_QNAME 			= 0x80;
	public static final int PRECEDING_SIBLING_QNAME 	= 0x90;
	public static final int FOLLOWING_QNAME 			= 0xa0;
	public static final int FOLLOWING_SIBLING_QNAME 	= 0xb0;
	
	// attention order is important ... see ReconstructionVisitor

	public static final int SELF_NODE 				= 0x11;
	public static final int CHILD_NODE 				= 0x21;
	public static final int PARENT_NODE 			= 0x31;
	public static final int ANCESTOR_NODE 			= 0x41;
	public static final int ANCESTOR_OR_SELF_NODE 	= 0x51;
	public static final int DESCENDANT_NODE 		= 0x61;
	public static final int DESCENDANT_OR_SELF_NODE = 0x71;
	public static final int PRECEDING_NODE 			= 0x81;
	public static final int PRECEDING_SIBLING_NODE 	= 0x91;
	public static final int FOLLOWING_NODE 			= 0xa1;
	public static final int FOLLOWING_SIBLING_NODE 	= 0xb1;

	public static final int SELF_DOCUMENT 				= 0x12;
	public static final int CHILD_DOCUMENT 				= 0x22;
	public static final int PARENT_DOCUMENT 			= 0x32;
	public static final int ANCESTOR_DOCUMENT 			= 0x42;
	public static final int ANCESTOR_OR_SELF_DOCUMENT 	= 0x52;
	public static final int DESCENDANT_DOCUMENT 		= 0x62;
	public static final int DESCENDANT_OR_SELF_DOCUMENT = 0x72;
	public static final int PRECEDING_DOCUMENT 			= 0x82;
	public static final int PRECEDING_SIBLING_DOCUMENT 	= 0x92;
	public static final int FOLLOWING_DOCUMENT 			= 0xa2;
	public static final int FOLLOWING_SIBLING_DOCUMENT 	= 0xb2;
	
	public static final int SELF_ATTRIBUTE 					= 0x13;
	public static final int CHILD_ATTRIBUTE 				= 0x23;
	public static final int PARENT_ATTRIBUTE 				= 0x33;
	public static final int ANCESTOR_ATTRIBUTE 				= 0x43;
	public static final int ANCESTOR_OR_SELF_ATTRIBUTE 		= 0x53;
	public static final int DESCENDANT_ATTRIBUTE 			= 0x63;
	public static final int DESCENDANT_OR_SELF_ATTRIBUTE 	= 0x73;
	public static final int PRECEDING_ATTRIBUTE 			= 0x83;
	public static final int PRECEDING_SIBLING_ATTRIBUTE 	= 0x93;
	public static final int FOLLOWING_ATTRIBUTE 			= 0xa3;
	public static final int FOLLOWING_SIBLING_ATTRIBUTE 	= 0xb3;

	public static final int SELF_PI 				= 0x14;
	public static final int CHILD_PI 				= 0x24;
	public static final int PARENT_PI 				= 0x34;
	public static final int ANCESTOR_PI 			= 0x44;
	public static final int ANCESTOR_OR_SELF_PI 	= 0x54;
	public static final int DESCENDANT_PI 			= 0x64;
	public static final int DESCENDANT_OR_SELF_PI 	= 0x74;
	public static final int PRECEDING_PI 			= 0x84;
	public static final int PRECEDING_SIBLING_PI 	= 0x94;
	public static final int FOLLOWING_PI 			= 0xa4;
	public static final int FOLLOWING_SIBLING_PI 	= 0xb4;

	public static final int SELF_COMMENT 				= 0x15;
	public static final int CHILD_COMMENT 				= 0x25;
	public static final int PARENT_COMMENT 				= 0x35;
	public static final int ANCESTOR_COMMENT 			= 0x45;
	public static final int ANCESTOR_OR_SELF_COMMENT 	= 0x55;
	public static final int DESCENDANT_COMMENT 			= 0x65;
	public static final int DESCENDANT_OR_SELF_COMMENT 	= 0x75;
	public static final int PRECEDING_COMMENT 			= 0x85;
	public static final int PRECEDING_SIBLING_COMMENT 	= 0x95;
	public static final int FOLLOWING_COMMENT 			= 0xa5;
	public static final int FOLLOWING_SIBLING_COMMENT 	= 0xb5;

	public static final int SELF_TEXT 				= 0x16;
	public static final int CHILD_TEXT 				= 0x26;
	public static final int PARENT_TEXT 			= 0x36;
	public static final int ANCESTOR_TEXT 			= 0x46;
	public static final int ANCESTOR_OR_SELF_TEXT 	= 0x56;
	public static final int DESCENDANT_TEXT 		= 0x66;
	public static final int DESCENDANT_OR_SELF_TEXT = 0x76;
	public static final int PRECEDING_TEXT 			= 0x86;
	public static final int PRECEDING_SIBLING_TEXT 	= 0x96;
	public static final int FOLLOWING_TEXT 			= 0xa6;
	public static final int FOLLOWING_SIBLING_TEXT 	= 0xb6;

	public static final int SELF_NAMESPACE 					= 0x17;
	public static final int CHILD_NAMESPACE 				= 0x27;
	public static final int PARENT_NAMESPACE 				= 0x37;
	public static final int ANCESTOR_NAMESPACE 				= 0x47;
	public static final int ANCESTOR_OR_SELF_NAMESPACE 		= 0x57;
	public static final int DESCENDANT_NAMESPACE 			= 0x67;
	public static final int DESCENDANT_OR_SELF_NAMESPACE	= 0x77;
	public static final int PRECEDING_NAMESPACE 			= 0x87;
	public static final int PRECEDING_SIBLING_NAMESPACE 	= 0x97;
	public static final int FOLLOWING_NAMESPACE 			= 0xa7;
	public static final int FOLLOWING_SIBLING_NAMESPACE 	= 0xb7;

	public static final int SELF_ELEMENT 				= 0x18;
	public static final int CHILD_ELEMENT 				= 0x28;
	public static final int PARENT_ELEMENT 				= 0x38;
	public static final int ANCESTOR_ELEMENT 			= 0x48;
	public static final int ANCESTOR_OR_SELF_ELEMENT 	= 0x58;
	public static final int DESCENDANT_ELEMENT 			= 0x68;
	public static final int DESCENDANT_OR_SELF_ELEMENT 	= 0x78;
	public static final int PRECEDING_ELEMENT 			= 0x88;
	public static final int PRECEDING_SIBLING_ELEMENT 	= 0x98;
	public static final int FOLLOWING_ELEMENT 			= 0xa8;
	public static final int FOLLOWING_SIBLING_ELEMENT 	= 0xb8;

	protected ArrayList predicates = null;
	/**
	 * 
	 */
	private boolean hasSeparator = false;

	/**
	 * 
	 */
	private byte axis = Axis.NONE;

	/**
	 * 
	 */
	private String localName = null;

	/**
	 * 
	 */
	private String nameSpace = null;

	/**
	 * 
	 */
	private byte wild = StepExpr.NO_WILDCARD;

	/**
	 * 
	 */
	private byte kindtest = NodeKind.NONE;

	/**
	 * 
	 */
	private Type type = null;

	/**
	 * 
	 */
	private boolean isTypeNillable = false;

	/**
	 * 
	 */
	private int stepKind = NONE;

	// #############################################################################
	// VISITOR STUFF
	// #############################################################################

	public void accept(ParserVisitor visitor) throws XQueryException {
		visitor.visit(this);
	}

	// #############################################################################
	// CONSTRUCTOR STUFF
	// #############################################################################


	public Step(boolean hasSeparator, byte axis, XQueryExpression expression, ArrayList predicates, XQueryModule parentModule) throws TypeException, XQueryException {
		super(expression);
		setHasSeparator(hasSeparator);
		setAxis(axis);
		setPredicates(predicates);
		setInterfaceProperties();
		setParentModule(parentModule);
//		if (parentModule != null && parentModule.getStaticContext().getTypeVisitor() != null)
//			accept(parentModule.getStaticContext().getTypeVisitor());
	}

	// #############################################################################
	// ATTRIBUTE ACCESS STUFF
	// #############################################################################

	public void setExpression(XQueryExpression expr) throws XQueryException {
	    if (expr instanceof EnclosedExpression)
	        expr = ((EnclosedExpression)expr).getExpression();
	    super.setExpression(expr);
	}
	private void setInterfaceProperties() {
		// depending on expression		
		// add LARS 07/04/04
		stepKind = NONE;
		if (expression instanceof QName) {
			QName qname = (QName) expression;
			localName = qname.getLocalName();
			nameSpace = qname.getNameSpace();
			kindtest = NodeKind.ELEMENT;
			byte nodeKind = NodeKind.ELEMENT;
			switch (axis) {
				case Axis.ATTRIBUTE :
					nodeKind = NodeKind.ATTRIBUTE;
					break;
				case Axis.NAMESPACE :
					nodeKind = NodeKind.NAMESPACE;
					break;
			}
			xnode = new XNode(nameSpace, localName, nodeKind);
			// set wild
			if (NAME_WILDCARD.equals(qname.getName()))
				wild = StepExpr.WILDCARD;
			else if (NAME_WILDCARD.equals(localName))
				wild = StepExpr.LOCALNAME_WILDCARD;
			else if (NAME_WILDCARD.equals(nameSpace))
				wild = StepExpr.NAMESPACE_WILDCARD;
			// add LARS 07/04/04
			stepKind = axis*16;
		} else if (expression instanceof NodeTest) {
			// set kindtest
			kindtest = ((NodeTest) expression).getKind();
			xnode = new XNode(kindtest);
			// add LARS 07/04/04
			stepKind = axis*16+kindtest;
		}
	}

	public void setParentModule(XQueryModule parentUnit) {
		if (parentUnit == null)
			return;
		this.parentModule = parentUnit;
		super.setParentModule(parentUnit);
		if (predicates != null)
			for (int i = 0; i < predicates.size(); i++)
				 ((XQueryExpression) predicates.get(i)).setParentModule(parentUnit);
	}

	public boolean hasSeparator() {
		return hasSeparator;
	}
	public void setHasSeparator(boolean hasSeparator) {
		this.hasSeparator = hasSeparator;
	}

	public byte getAxis() {
		return axis;
	}
	public void setAxis(byte axis) {
		this.axis = axis;
	}

	public ArrayList getPredicates() {
		return predicates;
	}
	public void setPredicates(ArrayList predicates) {
		this.predicates = predicates;
		if (predicates != null)
			for (int i = 0; i < this.predicates.size(); i++) {
				XQueryExpression expr = (XQueryExpression) this.predicates.get(i);
				expr.setParentModule(parentModule);
				expr.setParentExpression(this);
			}
	}

	public void addPredicates(ArrayList predicates) {
		if (predicates == null || predicates.isEmpty())
			return;
		if (this.predicates == null || this.predicates.isEmpty()) {
			setPredicates(predicates);
			return;
		}
		this.predicates.addAll(predicates);
		setPredicates(this.predicates);
	}

	public void addPredicate(XQueryExpression predicate) {
		if (predicate == null)
			return;
		if (predicates == null)
			predicates = new ArrayList(1);
		predicates.add(predicate);
		predicate.setParentModule(parentModule);
		predicate.setParentExpression(this);
	}

	public void addPredicate(int index, XQueryExpression predicate) {
		if (predicates == null)
			predicates = new ArrayList(1);
		predicates.add(index, predicate);
		predicate.setParentModule(parentModule);
		predicate.setParentExpression(this);
	}

	//	public String getName() {
	//		return expression.getName();
	//	}

	public String toHashCodeString() {
		return toString();
	}

	// #############################################################################
	// CLONE STUFF
	// #############################################################################
	public void addParentExpression(XQueryExpression parentExpression) {
		addParentExpression(parentExpression);
		expression.addParentExpression(parentExpression);
		if (predicates != null)
			for (int i = 0; i < predicates.size(); i++)
				 ((XQueryExpression) predicates.get(i)).addParentExpression(parentExpression);
	}

	// #############################################################################
	// RESTRUCTURATION STUFF
	// #############################################################################

	public boolean hasPredicates() {
		if (predicates != null && !predicates.isEmpty())
			return true;
		return false;
	}

	public boolean hasPositionPredicates() {
		if (predicates != null)
			for (int i = 0; i < predicates.size(); i++) {
				XQueryExpression predExpr = (XQueryExpression) predicates.get(i);
				// test the type of the predicates
				if (predExpr instanceof ValueInteger)
					return true;
			}
		return false;
	}

	public int getStepKind() {
		return stepKind;
	}
	
	// #############################################################################
	// INTERFACE STUFF
	// #############################################################################

	/**
	 * 
	 * @return the namespace if any null otherwise
	 */
	public String getNameSpace() {
		return nameSpace;
	}

	/**
	 * 
	 * @return the localname if any null otherwise
	 */
	public String getLocalName() {
		return localName;
	}

	/**
	 * 
	 * @return the constant representing the wilcard
	 */
	public byte getWildCardType() {
		return wild;
	}

	/**
	 * 
	 * @return false if no wildcard is present
	 */
	public boolean hasWildcard() {
		return wild != StepExpr.NO_WILDCARD;
	}

	/**
	 * 
	 * @return true if step if *
	 */
	public boolean isWildcard() {
		return wild == StepExpr.WILDCARD;
	}

	/**
	 * 
	 * @return true if step is qname with * as namespace
	 */
	public boolean isNameSpaceWildcard() {
		return isWildcard() || wild == StepExpr.NAMESPACE_WILDCARD;
	}

	/**
	 * 
	 * @return true if step is qname with * as localname
	 */
	public boolean isLocalNameWildcard() {
		return isWildcard() || wild == StepExpr.LOCALNAME_WILDCARD;
	}

	/**
	 * 
	 * @return the constant representing the kind of the step
	 */
	public byte getKindTest() {
		return kindtest;
	}

	/**
	 * 
	 * @return the (schema) type used in element or attribute test
	 */
	public org.xquark.schema.Type getType() {
		return type;
	}

	/**
	 * 
	 * @return true if the type used in element test is nillable
	 */
	public boolean isTypeNillable() {
		return isTypeNillable;
	}

	//	methods for path resolving

	public static final String NAME_WILDCARD = "*";
	public XNode xnode = null;

	//	public StepExprImpl(byte axis, XNode xnode) {
	//		this.axis = axis;
	//		this.xnode = xnode;
	//	}
	//
	//	public StepExprImpl(XNode xnode) {
	//		this.axis = Axis.CHILD;
	//		this.xnode = xnode;
	//	}

	public XNode getXNode() {
		return xnode;
	}

	public boolean match(XNode node) {
		if (node == null)
			return false;
		if (xnode == null)
		    return false;
		if (isTypeTest()) { // true if local name == null
			return typeMatch(node);
		} else // namespace matching
			{
			if (typeMatch(node) == false) {
				return false;
			}
			if (getNamespace() == null) {
				if (node.getNamespace() == null) {
					// namespace match
				} else {
					return false;
				}
			} else if (isNamespaceWildcard()) {
				// namespace match
			} else if (node.getNamespace() == null) {
				return false;
			} else if (getNamespace().equals(node.getNamespace())) {
				// namespace match
			} else {
				return false;
			}

			// matching local name
			if (this.getLocalName() == null) {
				if (node.getLocalName() == null) {
					// local name match
				} else {
					return false;
				}
			} else if (this.isLocalNameWildcard()) {
				// local name match
			} else if (node.getLocalName() == null) {
				return false;
			} else if (this.getLocalName().equals(node.getLocalName())) {
				// local name match
			} else {
				return false;
			}
			return true;
		}
	}
	public boolean isTypeTest() {
		return xnode.getLocalName() == null;
	}
	private boolean typeMatch(XNode node) {
		return ((getTypeTest() == node.getType()) || (getTypeTest() == NodeKind.NODE));
	}
	public byte getTypeTest() {
		return xnode.getType();
	}
	public String getNamespace() {
		return xnode.getNamespace();
	}
	public boolean isNamespaceWildcard() {
		return isNameTest() && (xnode.getNamespace() != null) && xnode.getNamespace().equals(NAME_WILDCARD);
	}
	public boolean isNameTest() {
		return xnode.getLocalName() != null;
	}

}
