/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.xquery.parser;

import java.util.*;

import org.xquark.schema.SchemaManager;
import org.xquark.util.NamespaceContextStack;
import org.xquark.xquery.metadata.MetaModule;
import org.xquark.xquery.metadata.StaticContext;
import org.xquark.xquery.normalize.*;
import org.xquark.xquery.parser.util.Constants;
import org.xquark.xquery.typing.TypeVisitor;

/*
[1]    Module    ::=    VersionDecl? (MainModule | LibraryModule) 
[2]    VersionDecl    ::=    <"xquery" "version"> StringLiteral ("encoding" StringLiteral)? Separator 
[3]    MainModule    ::=    Prolog QueryBody 
[4]    LibraryModule    ::=    ModuleDecl Prolog 
[5]    ModuleDecl    ::=    <"module" "namespace"> NCName "=" URILiteral Separator 
[6]    Prolog    ::=    ((Setter | Import | NamespaceDecl | DefaultNamespaceDecl) Separator)* ((VarDecl | FunctionDecl | OptionDecl) Separator)* 
[7]    Setter    ::=    BoundarySpaceDecl | DefaultCollationDecl | BaseURIDecl | ConstructionDecl | OrderingModeDecl | EmptyOrderDecl | CopyNamespacesDecl 
[8]    Import    ::=    SchemaImport | ModuleImport 
[9]    Separator    ::=    ";" 
[10]    NamespaceDecl    ::=    <"declare" "namespace"> NCName "=" URILiteral 
[11]    BoundarySpaceDecl    ::=    <"declare" "boundary-space"> ("preserve" | "strip") 
[12]    DefaultNamespaceDecl    ::=    (<"declare" "default" "element"> | <"declare" "default" "function">) "namespace" URILiteral 
[13]    OptionDecl    ::=    <"declare" "option"> QName StringLiteral 
[14]    OrderingModeDecl    ::=    <"declare" "ordering"> ("ordered" | "unordered") 
[15]    EmptyOrderDecl    ::=    <"declare" "default" "order"> (<"empty" "greatest"> | <"empty" "least">) 
[16]    CopyNamespacesDecl    ::=    <"declare" "copy-namespaces"> PreserveMode "," InheritMode 
[17]    PreserveMode    ::=    "preserve" | "no-preserve" 
[18]    InheritMode    ::=    "inherit" | "no-inherit" 
[19]    DefaultCollationDecl    ::=    <"declare" "default" "collation"> URILiteral 
[20]    BaseURIDecl    ::=    <"declare" "base-uri"> URILiteral 
*/

public class XQueryModule extends MetaModule {
	private static final String RCSRevision = "$Revision: 1.42 $";
	private static final String RCSName = "$Name: XQ1_1_1 $";

//	private static Log log = LogFactory.getLog("internal." + XQueryModule.class.getName());

	protected String version = null;
	protected String prefix = null;
	protected String defaultFunctionNamespace = null;
	protected int boundaryspace = Constants.NOTHING;
	protected int validation = Constants.NOTHING;
	protected int construction = Constants.NOTHING;
	protected String defaultCollation = null;
	protected int defaultOrdering = Constants.NOTHING;
	protected int defaultEmptyOrdering = Constants.NOTHING;
	protected int copyNamespacePreserveMode = Constants.NOTHING;
	protected int copyNamespaceInheritMode = Constants.NOTHING;
	protected String baseURI = null;
	protected String encoding = null;
	protected HashMap options = null;
	
	protected HashMap hashExtVarList = null;
	protected HashMap importSchemas = null;
	protected ArrayList schemaList = null;
	protected HashMap importModules = null;
	protected ArrayList moduleList = null;

	protected ArrayList expressions = null;

	protected SchemaManager schemaManager = null;
	protected StaticContext staticContext = null;

	// #############################################################################
	// VISITOR STUFF
	// #############################################################################

	public void accept(ParserVisitor visitor) throws XQueryException {
		visitor.visit(this);
	}

	// #############################################################################
	// CONSTRUCTOR STUFF
	// #############################################################################

	public XQueryModule() {
	}

	// #############################################################################
	// ATTRIBUTE ACCESS STUFF
	// #############################################################################

	// expressions
	public ArrayList getExpressions() {
		return expressions;
	}
	public void setExpressions(ArrayList expressions) throws XQueryException {
		this.expressions = expressions;
		if (expressions != null)
			for (int i = 0; i < this.expressions.size(); i++) {
				((XQueryExpression) this.expressions.get(i)).setParentModule(this);
				((XQueryExpression) this.expressions.get(i)).setParentExpressions(null);
			}
	}

	public void setElementAt(XQueryExpression expr, int index) {
		expr.setParentModule(this);
		expressions.set(index, expr);
	}

	public String getNameSpace(String prefixname) {
		if (prefixname == null)
			return null;
		if (declarations == null)
			return null;
		String str = declarations.getNamespaceURI(prefixname);
		if (str == null)
			return null;
		int len = str.length();
		if (str.indexOf("\"") == 0)
			return str.substring(1, len - 2);
		else
			return str;
	}

	// variables
//	public HashMap getVariables() {
//		return hashVarList;
//	}
//	public void setDeclList(ArrayList declList) {
//		this.declList = declList;
//	}
//	public ArrayList getDeclList() {
//		return declList;
//	}
	public void setVariables(HashMap hashVarList) {
		this.hashVarList = hashVarList;
		if (hashVarList != null) {
			Collection varValues = hashVarList.values();
			Iterator itValues = varValues.iterator();
			while (itValues.hasNext()) {
				HashMap map = (HashMap) itValues.next();
				Collection mapValues = map.values();
				Iterator itMapValues = mapValues.iterator();
				while (itMapValues.hasNext()) {
					((Variable) itMapValues.next()).setParentModule(this);
				}
			}
		}
	}

//	public Variable getVariableDefinition(String name) {
//		if (name == null || hashVarList == null)
//			return null;
//		return (Variable) hashVarList.get(name);
//	}

	// external variables
	public HashMap getExternalVariables() {
		return hashExtVarList;
	}
	public void setExternalVariables(HashMap hashExtVarList) {
		this.hashExtVarList = hashExtVarList;
	}

	// functions
//	public HashMap getFunctions() {
//		return hashFuncList;
//	}
	public void setFunctions(HashMap hashFuncList) {
		this.hashFuncList = hashFuncList;
		if (hashFuncList != null) {
			Collection funcValues = hashFuncList.values();
			Iterator itValues = funcValues.iterator();
			while (itValues.hasNext()) {
				HashMap map = (HashMap) itValues.next();
				Collection mapValues = map.values();
				Iterator itMapValues = mapValues.iterator();
				while (itMapValues.hasNext()) {
					((FunctionDeclaration) itMapValues.next()).setParentModule(this);
				}
			}
		}
	}

	// options
	public void setOptions(HashMap options) {
	    this.options = options;
	}
	
	public HashMap getOptions() {
	    return options;
	}
	
	// encoding
	public void setEncoding(String encoding) {
	    this.encoding = encoding;
	}
	
	public String getEncoding() {
	    return encoding;
	}
	
	// default ordering
	public void setDefaultOrdering(int defaultOrdering) {
		this.defaultOrdering = defaultOrdering;
	}

	public int getDefaultOrdering() {
		return defaultOrdering;
	}

	// default empty ordering
	public void setDefaultEmptyOrdering(int defaultEmptyOrdering) {
		this.defaultEmptyOrdering = defaultEmptyOrdering;
	}

	public int getDefaultEmptyOrdering() {
		return defaultEmptyOrdering;
	}

	// copy namespace preserve mode
	public void setCopyNamespacePreserveMode(int copyNamespacePreserveMode) {
	    this.copyNamespacePreserveMode = copyNamespacePreserveMode;
	}
	
	public int getCopyNamespacePreserveMode() {
	    return copyNamespacePreserveMode;
	}

	// copy namespace inherit mode
	public void setCopyNamespaceInheritMode(int copyNamespaceInheritMode) {
	    this.copyNamespaceInheritMode = copyNamespaceInheritMode;
	}
	
	public int getCopyNamespaceInheritMode() {
	    return copyNamespaceInheritMode;
	}

	// default collation
	public void setDefaultCollation(String defaultCollation) {
		this.defaultCollation = defaultCollation;
	}

	public String getDefaultCollation() {
		return defaultCollation;
	}

	// xquery version
	public void setVersion(String version) {
		this.version = version;
	}

	public String getVersion() {
		return version;
	}

	// module prefix
	public void setPrefix(String prefix) {
		this.prefix = prefix;
	}

	public String getPrefix() {
		return prefix;
	}

	// module location
//	public void setNamespace(String location) {
//		this.namespace = location;
//	}

//	public String getNamespace() {
//		return namespace;
//	}

	// default function namespace
	public void setDefaultFunctionNamespace(String namespace) {
		this.defaultFunctionNamespace = namespace;
	}

	public String getDefaultFunctionNamespace() {
		return defaultFunctionNamespace;
	}

	// base uri
	public void setBaseURI(String uri) {
		this.baseURI = uri;
	}

	public String getBaseURI() {
		return baseURI;
	}

	// xmlspace
	public void setBoundarySpace(int boundaryspace) {
		this.boundaryspace = boundaryspace;
	}

	public int getBoundarySpace() {
		return boundaryspace;
	}

	// validation
	public void setValidation(int validation) {
		this.validation = validation;
	}

	public int getValidation() {
		return validation;
	}

	// construction
	public void setConstruction(int construction) {
		this.construction = construction;
	}

	public int getConstruction() {
		return construction;
	}

	// import schemas
	public HashMap getImportSchemas() {
		return importSchemas;
	}
	public ArrayList getSchemaList() {
		return schemaList;
	}
	public void setSchemaList(ArrayList schemaList) {
		this.schemaList = schemaList;
	}
	public void setImportSchemas(HashMap importSchemas) {
		this.importSchemas = importSchemas;
		if (importSchemas != null) {
			Collection importSchemaValues = importSchemas.values();
			Iterator itValues = importSchemaValues.iterator();
			while (itValues.hasNext())
				 ((ImportSchema) itValues.next()).setParentModule(this);
		}
	}

	public ImportSchema getImportSchema(String name) {
		if (name == null || importSchemas == null)
			return null;
		return (ImportSchema) importSchemas.get(name);
	}

	// import modules
	public HashMap getImportModules() {
		return importModules;
	}
	public void setModuleList(ArrayList moduleList) {
		this.moduleList = moduleList;
	}
	public ArrayList getModuleList() {
		return moduleList;
	}
	public void setImportModules(HashMap importModules) {
		this.importModules = importModules;
		if (importModules != null) {
			Collection importModuleValues = importModules.values();
			Iterator itValues = importModuleValues.iterator();
			while (itValues.hasNext())
				 ((ImportModule) itValues.next()).setParentModule(this);
		}
	}

	public ImportModule getImportModule(String name) {
		if (name == null || importModules == null)
			return null;
		return (ImportModule) importModules.get(name);
	}

	// schemaManager
	public void setSchemaManager(SchemaManager schemaManager) {
		this.schemaManager = schemaManager;
	}
	public SchemaManager getSchemaManager() {
		return schemaManager;
	}

	// staticContext
	public void setStaticContext(StaticContext staticContext) {
		this.staticContext = staticContext;
	}
	public StaticContext getStaticContext() {
		return staticContext;
	}

	// #############################################################################
	// NORMALIZE STUFF
	// #############################################################################

	public void normalize(TypeVisitor typeVisitor) throws XQueryException {
		normalize(typeVisitor, null);
	}

	public void normalize(TypeVisitor typeVisitor, String sourceName) throws XQueryException {
//		if (log.isTraceEnabled())
//			log.trace("START NORMALIZE ");
		try {
			ExpressionNormalizationVisitor normalizeVisitor = new ExpressionNormalizationVisitor(this, typeVisitor, sourceName);
			accept(normalizeVisitor);
		} catch (XQueryException ne) {
			if (namespace != null)
				throw new XQueryException(ne.getMessage() + " in module " + namespace);
			else
				throw new XQueryException(ne.getMessage() + " in main module");
		}
//		if (log.isTraceEnabled()) {
//			log.trace("STOP NORMALIZE ");
//			log.trace(toString(true, true));
//		}

//		if (log.isTraceEnabled())
//			log.trace("START RESTRUCTURE ");
		try {
			if (expressions != null) {
				RestructureVisitor restructureVisitor = new RestructureVisitor(this, typeVisitor, (sourceName != null));
				accept(restructureVisitor);
			}
		} catch (XQueryException xqe) {
			if (namespace != null)
				throw new XQueryException(xqe.getMessage() + " in module " + namespace);
			else
				throw new XQueryException(xqe.getMessage() + " in main module");
		}
//		if (log.isTraceEnabled()) {
//			log.trace("STOP RESTRUCTURE ");
//			log.trace(toString(true, true));
//		}
	}

	public TransformXQueryExpression canonize(XQueryExpression expression, TypeVisitor typeVisitor) throws ParseException {
//		if (log.isTraceEnabled())
//			log.trace("START CANONIZE ");
		ArrayList QDB = null;
		XQueryExpression QMEM = null;
		try {
			CanonizeVisitor canonizeVisitor = new CanonizeVisitor(this, typeVisitor);
			expression.accept(canonizeVisitor);
//			if (log.isTraceEnabled())
//				log.trace("STOP  CANONIZE ");
			QDB = canonizeVisitor.getQDB();
			QMEM = canonizeVisitor.getQMEM();
		} catch (XQueryException e) {
			throw new ParseException("Invalid expression : " + this +" -> " + e.getMessage());
		}
		if (QDB == null)
			throw new ParseException("Invalid expression : " + this);
//		if (log.isTraceEnabled()) {
//			for (int si = 0; si < QDB.size(); si++) {
//				log.trace("QDB[" + si + "]: " + ((Variable) QDB.get(si)).getExpression());
//			}
//			log.trace("QMEM: " + QMEM);
//		}
		return new TransformXQueryExpression(QDB, QMEM);
	}

	// #############################################################################
	// SOURCES STUFF
	// #############################################################################

	public Set getSourceNames() {
		Set sources = null;
		// add lars 03/05/05 to add eval source for external variables if needed
		if (hashExtVarList != null) {
		    for (Iterator it = hashExtVarList.values().iterator();it.hasNext();) {
		        Variable extVar = (Variable)it.next();
		        if (extVar.getQType() == null || !extVar.getQType().isAtom()) {
		            if (sources == null) {
		                sources = new HashSet();
		            }
		            // todo modify SPECIAL !!!!
		            sources.add("SPECIAL");
		        }
		    }
		}
		if (expressions != null) {
			for (int i = 0; i < expressions.size(); i++) {
				Set addSet = ((XQueryExpression) expressions.get(i)).getSourceNames();
				if (addSet != null) {
					if (sources == null)
						sources = new HashSet();
					sources.addAll(addSet);
				}
			}
		}
		return sources;
	}

	public Set getUrls() {
		Set urls = null;
		if (expressions != null) {
			for (int i = 0; i < expressions.size(); i++) {
				Set addSet = ((XQueryExpression) expressions.get(i)).getUrls();
				if (addSet != null) {
					if (urls == null)
						urls = new HashSet();
					urls.addAll(addSet);
				}
			}
		}
		return urls;
	}

	// #############################################################################
	// PRINT STUFF
	// #############################################################################

	public String toString(boolean doLineFeeds, boolean doImports, boolean doFuncDecl) {
		PrintVisitor pv = new PrintVisitor();
		pv.reset(doLineFeeds, doImports, doFuncDecl, false, true);
		pv.setNamespaceContextStack(declarations);
		try {
			this.accept(pv);
		} catch (XQueryException qe) {
		}
		return pv.getString();
	}

	public String toString() {
		PrintVisitor pv = new PrintVisitor();
		pv.reset(false, true, false, false, true);
		try {
			this.accept(pv);
		} catch (XQueryException qe) {
		}
		return pv.getString();
	}
}
