/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Odonata.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.xquery.typing;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.xquark.schema.*;
import org.xquark.xquery.parser.QName;
import org.xquark.xquery.parser.XQueryException;

public class OccurrenceVisitor extends DefaultSchemaVisitor {
	private static final String RCSRevision = "$Revision: 1.5 $";
	private static final String RCSName = "$Name: XQ1_1_1 $";
	private static final String STAR = "*";

	private int[] occurs = new int[2];
	private String namespace = null;
	private String name = null;
	private List matches = new ArrayList();
	private boolean matchesWildcard = false;
	private SchemaManager schemamanager = null;

	public void reset(String namespace, String name, SchemaManager schemamanager) {
		occurs[0] = occurs[1] = 0;
		matches.clear();
		matchesWildcard = false;
		this.namespace = namespace;
		this.name = name;
		this.schemamanager = schemamanager;
	}

	public QType getQType() throws XQueryException {
		// calculate occurence
		byte occurrence = QType.OCC_0_0;
		if (occurs[0] == 0) {
			if (occurs[1] == 0)
				occurrence = QType.OCC_0_0;
			else if (occurs[1] == 1)
				occurrence = QType.OCC_0_1;
			else
				occurrence = QType.OCC_0_N;
		} else if (occurs[0] == 1) {
			if (occurs[1] == 1)
				occurrence = QType.OCC_1_1;
			else
				occurrence = QType.OCC_1_N;
		}
		if (matchesWildcard) {
			return new QTypeElement(new QName(namespace, null, name, null, null), schemamanager.getType(SchemaConstants.XMLSCHEMA_URI, TypeVisitor.SC_ANYTYPE), occurrence);
		}
		if (!matches.isEmpty()) {
			if (matches.size() == 1) {
				return getQTypeFromDeclaration((Declaration) matches.get(0), occurrence);
			} else {
				ArrayList qtypes = new ArrayList(matches.size());
				for (int i = 0; i < matches.size(); i++) {
					qtypes.add(getQTypeFromDeclaration((Declaration) matches.get(i), QType.OCC_1_1));
				}
				return new QTypeUnion(qtypes, occurrence);
			}
		}
		return null;
	}

	//	public int[] getOccurrenceCount() {
	//		return occurs;
	//	}
	//	public List getMatches() {
	//		return matches;
	//	}

	private void setValues(long min, long max) {
		if (max > Integer.MAX_VALUE)
			max = Integer.MAX_VALUE;
		occurs[0] = (int) min;
		occurs[1] = (int) max;
	}

	public void visit(Particle particle) throws SchemaException {
		((SchemaVisitable) particle.getTerm()).accept(this);
		long min = occurs[0] * particle.getMinOccurs();
		long max = occurs[1] * particle.getMaxOccurs();
		setValues(min, max);
	}

	public void visit(ModelGroup group) throws SchemaException {
		long min = 0;
		long max = 0;
		Iterator it = group.iterator();
		while (it.hasNext()) {
			occurs[0] = occurs[1] = 0;
			((SchemaVisitable) it.next()).accept(this);
			min += occurs[0];
			max += occurs[1];
		}
		setValues(min, max);
	}

	public void visit(ChoiceModelGroup group) throws SchemaException {
		long min = Integer.MAX_VALUE;
		long max = 0;
		Iterator it = group.iterator();
		while (it.hasNext()) {
			occurs[0] = occurs[1] = 0;
			((SchemaVisitable) it.next()).accept(this);
			if (occurs[0] < min)
				min = occurs[0];
			if (occurs[1] > max)
				max = occurs[1];
		}
		setValues(min, max);
	}

	public void visit(ElementDeclaration decl) throws SchemaException {
		if (namespace != null && namespace.equals(STAR)) {
			if (name.equals(STAR)) {
				matches.add(decl);
				if (decl.getSubstitutionElementList() != null)
					matches.addAll(decl.getSubstitutionElementList());
			} else {
				if (name.equals(decl.getName())) {
					matches.add(decl);
				}
				List decllist = decl.getSubstitutionElementList();
				if (decllist != null && !decllist.isEmpty()) {
					for (int i = 0; i < decllist.size(); i++) {
						if (name.equals(((Declaration) decllist.get(i)).getName())) {
							matches.add(decllist.get(i));
						}
					}
				}
			}
		} else if (name.equals(STAR)) {
			if (namespace != null && namespace.equals(decl.getNamespace())) {
				matches.add(decl);
			}
			List decllist = decl.getSubstitutionElementList();
			if (decllist != null && !decllist.isEmpty()) {
				for (int i = 0; i < decllist.size(); i++) {
					if (namespace != null && namespace.equals(((Declaration) decllist.get(i)).getNamespace())) {
						matches.add(decllist.get(i));
					}
				}
			}
		}
		if (decl.hasName(namespace, name) || (decl = decl.getSubstitutionElement(namespace, name)) != null) {
			occurs[0] = occurs[1] = 1;
			matches.add(decl);
		}
	}

	public void visit(Wildcard wildcard) throws SchemaException {
		if (wildcard.isAllowed(namespace)) {
			occurs[0] = occurs[1] = 1;
			matchesWildcard = true;
		}
	}

	// private methods
	private QType getQTypeFromDeclaration(Declaration decl, byte occurrence) throws XQueryException {
		QName qname = new QName(decl.getNamespace(), null, decl.getName(), null, null);
		if (decl instanceof AttributeDeclaration) {
			return new QTypeAttribute(qname, (AttributeDeclaration)decl, (SimpleType) decl.getType());
		} else if (decl instanceof ElementDeclaration) {
			return new QTypeElement(qname, decl.getType(), occurrence);
		}
		return null;
	}

}
