/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Odonata.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.xquery.typing;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.xquark.xquery.parser.XQueryException;

/*
prime(ItemType)			=   ItemType 
prime(empty)   			=   none 
prime(none)   			=   none 
prime(Type 1 , Type2)   =   prime(Type 1) | prime(Type 2) 
prime(Type 1 & Type2)   =   prime(Type 1) | prime(Type 2) 
prime(Type 1 | Type2)   =   prime(Type 1) | prime(Type 2) 
prime(Type?)   			=   prime(Type) 
prime(Type*)   			=   prime(Type) 
prime(Type+)   			=   prime(Type) 

Semantics

The quantifier function is defined by induction as follows.

quantifier(ItemType)   		=   1 
quantifier(empty)   		=   ? 
quantifier(none)   			=   1 
quantifier(Type 1 , Type2)	=   quantifier(Type 1) , quantifier(Type 2) 
quantifier(Type 1 & Type2)	=   quantifier(Type 1) , quantifier(Type 2) 
quantifier(Type 1 | Type2)	=   quantifier(Type 1) | quantifier(Type 2) 
quantifier(Type?)   		=   quantifier(Type)  ? 
quantifier(Type*)   		=   quantifier(Type)  * 
quantifier(Type+)   		=   quantifier(Type)  + 

This definition uses the sum (Occurrence1 , Occurrence2), the choice (Occurrence1 | Occurrence2), and the product (Occurrence1  Occurrence2) of two occurrence indicators Occurrence1, Occurrence2, which are defined by the following tables.

 ,   1   ?   +   *  
 1   +   +   +   +  
 ?   +   *   +   *  
 +   +   +   +   +  
 *   +   *   +   *  

 |   1   ?   +   *  
 1   1   ?   +   *  
 ?   ?   ?   *   *  
 +   +   *   +   *  
 *   *   *   *   *  

    1   ?   +   *  
 1   1   ?   +   *  
 ?   ?   ?   *   *  
 +   +   *   +   *  
 *   *   *   *   *  
 

Examples

For example, here are the result of applying prime and quantifier on a few simple types.

  prime(element a+)                         = element a
  prime(element a | empty)                  = element a
  prime(element a?,element b?)              = element a | element b
  prime(element a | element b+, element c*) = element a | element b | element c

  quantifier(element a+)                         = +
  quantifier(element a | empty)                  = ?
  quantifier(element a?,element b?)              = *
  quantifier(element a | element b+, element d*) = +

Note that the last occurrence indicator should be '+', since the regular expression is such that there must be at least one element in the sequence (this element being an 'a' element or a 'b' element).

Note

Note that prime(Type)  quantifier(Type) is always a super type of the original type Type I.e., prime(Type)  quantifier(Type) <: Type always holds. Therefore, it is appropriate to used it as an approximation for the type of an expression. This property is required for the soundness of the static type analysis.

*/

public abstract class QTypePrime extends QType {

	// specific stuff for prime type
	/**
	 * computed list
	 */
	protected PrimeList primeList = null;
	/**
	 * computed occurence
	 */
	protected byte primeOccurence = OCC_1_1;

	// constructor
	QTypePrime(List list, byte subclass) {
		this.subclass = subclass;
		// compute primelist
		computePrimeList(list);
		// compute occurence
	}

	/**
	 * method to get the list
	 */
	public List getList() {
		return primeList;
	}
	
	public List getPrimeList() {
		return primeList;
	}

	public byte getOccurence() {
		return primeOccurence;
	}

	private void computePrimeList(List list) {
		primeList = new PrimeList();
		if (list == null || list.isEmpty())
			return;
		//primeOccurence = OCC_0_0;
		if (list == null || list.isEmpty())
			return;
		for (int i = 0; i < list.size(); i++) {
			QType qtypei = (QType) list.get(i);
			List listi = ((QType) list.get(i)).getList();
			if (listi == null) { // QTypeItem
				primeList.add(qtypei);
			} else {
				// QTypePrime				
				primeList.addAll(listi);
			}
			computeOccurence(qtypei.getOccurence());
		}
	}

	private void computeOccurence(byte compOccurence) {
		switch (subclass) {
			case SEQUENCE :
				switch (primeOccurence) {
					case OCC_1_1 :
					case OCC_1_N :
						primeOccurence = OCC_1_N;
						break;
					case OCC_0_1 :
					case OCC_0_N :
						switch (compOccurence) {
							case OCC_1_1 :
							case OCC_1_N :
								primeOccurence = OCC_1_N;
								break;
							case OCC_0_1 :
							case OCC_0_N :
								primeOccurence = OCC_0_N;
								break;
						}
						break;
				}
				break;
			case UNION :
				primeOccurence = computeOccurence(primeOccurence, compOccurence);
				break;
		}
	}

	// equals
	public boolean equals(Object qtype) {
		if (!(qtype instanceof QTypePrime))
			return false;
//		if (((QTypePrime) qtype).getOccurence() != occurence)
//			return false;
		List list2 = ((QTypePrime) qtype).getPrimeList();
		int size1 = primeList.size();
		int size2 = list2.size();
		if (size1 != size2)
			return false;
		for (int i = 0; i < size1; i++) {
			QType qtype1 = (QType) primeList.get(i);
			boolean found = false;
			for (int j = 0; j < size2; j++)
				if (found = qtype1.equals((QType) list2.get(j)))
					break;
			if (!found)
				return false;
		}
		return true;
	}

	// tells whether qtype is numeric
	public boolean isNumeric() {
		Iterator it = primeList.iterator();
		while (it.hasNext())
			if (((QType) it.next()).isNumeric())
				return true;
		return false;
	}

	// tells whether qtype is boolean
	public boolean isBoolean() {
		int size = primeList.size();
		for (int i = 0; i < size; i++)
			if (((QType) primeList.get(i)).isBoolean())
				return true;
		return false;
	}

	// tells whether qtype is integer
	public boolean isInteger() {
		int size = primeList.size();
		for (int i = 0; i < size; i++)
			if (((QType) primeList.get(i)).isInteger())
				return true;
		return false;
	}

	// tells whether qtype is string
	public boolean isString() {
		int size = primeList.size();
		for (int i = 0; i < size; i++)
			if (((QType) primeList.get(i)).isString())
				return true;
		return false;
	}

	// tells whether qtype is qname
	public boolean isQName() {
		int size = primeList.size();
		for (int i = 0; i < size; i++)
			if (((QType) primeList.get(i)).isQName())
				return true;
		return false;
	}

	// tells whether qtype is an atom
	public boolean isAtom() {
		int size = primeList.size();
		for (int i = 0; i < size; i++)
			if (((QType) primeList.get(i)).isAtom())
				return true;
		return false;
	}

	// tells whether qtype can react to text()
	public boolean isText() {
		Iterator it = primeList.iterator();
		while (it.hasNext())
			if (((QType) it.next()).isText())
				return true;
		return false;
	}

	// tells whether qtype is a date or collection of dates
	public boolean isDate() {
		Iterator it = primeList.iterator();
		while (it.hasNext())
			if (((QType) it.next()).isDate())
				return true;
		return false;
	}

	// tells whether qtype is a attribte or collection or sequence
	public boolean isAttribute() {
		Iterator it = primeList.iterator();
		while (it.hasNext())
			if (((QType) it.next()).isAttribute())
				return true;
		return false;
	}

	// tells whether qtype is a attribte or collection or sequence
	public boolean isOnlyAttribute() {
		Iterator it = primeList.iterator();
		while (it.hasNext())
			if (!((QType) it.next()).isOnlyAttribute())
				return false;
		return true;
	}

	// tells whether qtype is a ID, IDREF or IDREFS type
	public boolean isIDAttribute() {
		Iterator it = primeList.iterator();
		while (it.hasNext())
			if (((QType) it.next()).isIDAttribute())
				return true;
		return false;
	}

	// tells whether qtype is a node
	public boolean isNode() {
		Iterator it = primeList.iterator();
		while (it.hasNext())
			if (((QType) it.next()).isNode())
				return true;
		return false;
	}

	// tells whether qtype is a node with simple type
	public boolean isSimpleTypeNode() {
		Iterator it = primeList.iterator();
		while (it.hasNext())
			if (((QType) it.next()).isSimpleTypeNode())
				return true;
		return false;
	}

	// tells whether qtype is a document
	public boolean isDocument() {
		Iterator it = primeList.iterator();
		while (it.hasNext())
			if (((QType) it.next()).isDocument())
				return true;
		return false;
	}

	// tells whether qtype is multiple
//	public boolean isMultiple() {
//		Iterator it = primeList.iterator();
//		while (it.hasNext())
//			if (((QType) it.next()).isMultiple())
//				return true;
//		return false;
//	}

	// tells whether qtype is atomic (atom or element with simple content)
	public boolean isAtomic() {
		Iterator it = primeList.iterator();
		while (it.hasNext())
			if (((QType) it.next()).isAtomic())
				return true;
		return false;
	}

	// tells whether qtype is mixed
	public boolean isMixed() {
		Iterator it = primeList.iterator();
		while (it.hasNext())
			if (((QType) it.next()).isMixed())
				return true;
		return false;
	}

	// tells whether qtype is element
	public boolean isElement() {
		Iterator it = primeList.iterator();
		while (it.hasNext())
			if (((QType) it.next()).isElement())
				return true;
		return false;
	}

	public boolean canBeCastedInto(int castType) {
		Iterator it = primeList.iterator();
		while (it.hasNext())
			if (((QType) it.next()).canBeCastedInto(castType))
				return true;
		return false;
	}

	public boolean canBeComparedTo(QType compType) {
		if (compType == null)
			return false;
		Iterator it = primeList.iterator();
		while (it.hasNext())
			if (((QType) it.next()).canBeComparedTo(compType))
				return true;
		return false;
	}

	public QType applyDATAFunction() throws TypeException {
		ArrayList newList = new ArrayList();
		Iterator it = primeList.iterator();
		while (it.hasNext()) {
			QType qtype = (QType) it.next();
			qtype = qtype.applyDATAFunction();
			if (qtype != null)
				newList.add(qtype);
		}
		if (!newList.isEmpty()) {
			if (newList.size() == 1)
				return (QType) newList.get(0);
			else
				return new QTypeUnion(newList);
		}
		return null;
	}
	
	public QType changeOccurence(byte occurence) {
		QTypePrime qtype = null;
		try {
			qtype = (QTypePrime) clone();
		} catch (CloneNotSupportedException cnse) {}
		qtype.primeOccurence = occurence;
		return qtype;
	}

	public void accept(QTypeVisitor visitor) throws XQueryException {
		visitor.visit(this);
	}

}
