/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * Copyright (C) 2003 Odonata.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.xquery.typing;

import java.util.*;

import org.xml.sax.SAXException;
import org.xquark.schema.*;
import org.xquark.schema.datatypes.PrimitiveType;
import org.xquark.xpath.*;
import org.xquark.xquery.ModuleLocator;
import org.xquark.xquery.ModuleManager;
import org.xquark.xquery.metadata.StaticContext;
import org.xquark.xquery.parser.*;
import org.xquark.xquery.parser.primitivefunctions.fnfunctions.*;
import org.xquark.xquery.parser.primitivefunctions.xsfunctions.*;
import org.xquark.xquery.parser.util.Constants;

public class TypeVisitor extends DefaultParserVisitor {
	private static final String RCSRevision = "$Revision: 1.19 $";
	private static final String RCSName = "$Name: XQ1_1_1 $";

	private static final String STAR = "*";
	private static final String XML_SCHEMA = "http://www.w3.org/2001/XMLSchema";

	public static final String SC_STRING = "string";
	public static final String SC_BOOLEAN = "boolean";
	public static final String SC_DOUBLE = "double";
	public static final String SC_DECIMAL = "decimal";
	public static final String SC_INTEGER = "integer";
	public static final String SC_FLOAT = "float";
	public static final String SC_DATE = "date";
	public static final String SC_TIME = "time";
	public static final String SC_DATETIME = "dateTime";
	public static final String SC_ANYTYPE = "anyType";
	//public static final String SC_ANYSIMPLETYPE = "anySimpleType";
	public static final String SC_QNAME = "QName";
	private static final String TEMP_NAME = "TEMP_NAME";

	private StaticContext context = null;

	// new context stuff
	private ArrayList contextXNodes = null;
	private QType contextQType = null;
	private byte contextType = QTypeDocument.ANY_ORIGIN;
	private LocatedExpression contextLocated = null;
	private int contextLocatedStepIndex = -1;

	// context stuff
	private boolean inPredicate = false;
	private boolean noForce = true;

	private SchemaManager typingschemamanager = null;
	private Schema typingschema = null;
	// for qa only
	boolean qaOnly = false;

	// constructor

	public TypeVisitor(StaticContext context, boolean qaOnly) throws TypeException {
		this.context = context;
		this.qaOnly = qaOnly;
		if (typingschemamanager == null) {
			this.typingschemamanager = new SchemaManager();
			this.typingschema = new Schema("http://www.xquark.org/EXECUTIONPLAN", null, typingschemamanager);
			try {
				typingschemamanager.putSchema(typingschema);
			} catch (SAXException saxe) {
				throw new TypeException("Could not add schema to manager");
			}
		}
	}

	public TypeVisitor(StaticContext context) throws TypeException {
		this.context = context;
		if (typingschemamanager == null) {
			this.typingschemamanager = new SchemaManager();
			this.typingschema = new Schema("http://www.xquark.org/EXECUTIONPLAN", null, typingschemamanager);
			try {
				typingschemamanager.putSchema(typingschema);
			} catch (SAXException saxe) {
				throw new TypeException("Could not add schema to manager");
			}
		}
	}

	public StaticContext getStaticContext() {
		return context;
	}

	public SchemaManager getSchemaManager() {
		return context.getSchemaManager();
	}

	public SchemaLocator getSchemaLocator() {
		return context.getSchemaManager();
	}

	public ModuleManager getModuleManager() {
		return context.getModuleManager();
	}

	public ModuleLocator getModuleLocator() {
		return context.getModuleManager();
	}

	public SchemaManager getTypingSchemaManager() {
		return typingschemamanager;
	}

	public void setNoForce(boolean noForce) {
		this.noForce = noForce;
	}

	// new context stuff

	private void setTypingContext(ArrayList xnodes, QType qtype, LocatedExpression contextLocated, int contextLocatedStepIndex) throws XQueryException {
		this.contextXNodes = xnodes;
		this.contextQType = qtype;
		this.contextLocated = contextLocated;
		this.contextLocatedStepIndex = contextLocatedStepIndex;
	}

	// visitors

	//	public void visit(AggregateFunctionCall arg) throws XQueryException
	/*
	[102]    CompAttrConstructor    ::=    (("attribute" QName "{") | ("attribute" "{" Expr "}" "{")) Expr? "}" 
	
	The name expression of a computed attribute constructor is processed as follows:
	
	Atomization is applied to the value of the name expression. If the result of atomization is not a single atomic value of type xs:QName, 
	xs:string, or xdt:untypedAtomic, a type error is raised.[err:XP0004][err:XP0006]
	
	If the atomized value of the name expression is of type xs:QName, that value is used as the name of the constructed attribute.
	
	If the atomized value of the name expression is of type xs:string or xdt:untypedAtomic, that value is cast to the type xs:QName using 
	the rules in [XQuery 1.0 and XPath 2.0 Functions and Operators]. The resulting value is used as the name of the constructed attribute. 
	A dynamic error is raised if the cast is not successful.[err:XP0021] In addition, a dynamic error is raised if the URI 
	part of the resulting QName is http://www.w3.org/TR/REC-xml-names.[err:XQ0044]
	
	The content expression of a computed attribute constructor is processed as follows:
	
	Atomization is applied to the value of the content expression, converting it to a sequence of atomic values.
	
	If the result of atomization is an empty sequence, the value of the attribute is the zero-length string. 
	Otherwise, each atomic value in the atomized sequence is cast into a string. 
	If any of these atomic values cannot be cast into a string, a dynamic error [err:XQ0052] is raised.
	
	The individual strings resulting from the previous step are merged into a single string by concatenating them 
	with a single space character between each pair. The resulting string is the string value of the attribute.
	
	A computed attribute constructor does not perform any automatic validation of the constructed attribute. 
	However, if the computed attribute constructor is inside an element constructor, the attribute will be validated during validation 
	of its parent element. The type annotation of an unvalidated attribute node is xdt:untypedAtomic.
	*/
	public void visit(AttributeValuePair arg) throws XQueryException {
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		XQueryExpression attributeName = arg.getAttributeName(); // cannot be null
		XQueryExpression attributeValue = arg.getAttributeValue(); // cannot be null
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null) {
			if (attributeValue.getQType() == null)
				return;
		}
		if (/*!noForce ||*/
			attributeValue.getQType() == null)
			attributeValue.accept(this);
		// type checking
		if (attributeValue.getQType() == null)
			throw new TypeException("Expression " + attributeValue + " could not be typed."); //return;

		QType nameQType = attributeName.getQType();
		if (nameQType != null && (!nameQType.isQName() || nameQType.isMultiple())) {
			nameQType = nameQType.applyDATAFunction();
			if (nameQType == null || !nameQType.isQName() || nameQType.isMultiple())
				throw new TypeException("Expression '" + attributeName + "', of attribute value pair has incorrect type.");
		}
		QType valueQType = attributeValue.getQType();
		if (valueQType != null && !valueQType.canBeCastedInto(QType.PRIMITIVE_string) && !valueQType.isString()) {
			valueQType = valueQType.applyDATAFunction();
			if (valueQType == null || (!valueQType.canBeCastedInto(QType.PRIMITIVE_string) && !valueQType.isString()))
				throw new TypeException("Expression '" + attributeValue + "', of attribute value pair has incorrect type.");
		}
		QType attQType = attributeValue.getQType();
		if (arg.getAttributeName() instanceof QName)
			arg.setQType(new QTypeAttribute((QName)arg.getAttributeName(), attQType.getSimpleType()));
		else if (arg.isXmlns())
			arg.setQType(new QTypeAttribute(new QName(null, null, "xmlns", arg.getParentModule(), null), attQType.getSimpleType()));
		else if (arg.getAttributeName() instanceof ValueString)
			arg.setQType(new QTypeAttribute(new QName(null, null, ((ValueString)arg.getAttributeName()).getValue(), arg.getParentModule(), null), attQType.getSimpleType()));
		else
			arg.setQType(new QTypeAttribute(new QName("*", "*", "*", arg.getParentModule(), null), attQType.getSimpleType()));
	}

	public void visit(BinOpANDExpression arg) throws XQueryException {
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		XQueryExpression expr1 = arg.getExpression1();
		XQueryExpression expr2 = arg.getExpression2();
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null) {
			if (expr1.getQType() == null)
				return;
			if (expr2.getQType() == null)
				return;
		}
		if (/*!noForce ||*/
			expr1.getQType() == null)
			expr1.accept(this);
		if (/*!noForce ||*/
			expr2.getQType() == null)
			expr2.accept(this);
		// type verification
		if (expr1.getQType() == null)
			throw new TypeException("Expression " + expr1 + " could not be typed."); //return;
		if (expr2.getQType() == null)
			throw new TypeException("Expression " + expr2 + " could not be typed."); //return;
		// type setting
		arg.setQType(new QTypeAtom(context.getSchemaManager().getBooleanType()));
	}

	public void visit(BinOpORExpression arg) throws XQueryException {
		XQueryExpression expr1 = arg.getExpression1();
		XQueryExpression expr2 = arg.getExpression2();
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null) {
			if (expr1.getQType() == null)
				return;
			if (expr2.getQType() == null)
				return;
		}
		if (/*!noForce ||*/
			expr1.getQType() == null)
			expr1.accept(this);
		if (/*!noForce ||*/
			expr2.getQType() == null)
			expr2.accept(this);
		// type verification
		if (expr1.getQType() == null)
			throw new TypeException("Expression " + expr1 + " could not be typed."); //return;
		if (expr2.getQType() == null)
			throw new TypeException("Expression " + expr2 + " could not be typed."); //return;
		// type setting
		arg.setQType(new QTypeAtom(context.getSchemaManager().getBooleanType()));
	}

	public void visit(CastTreatExpression arg) throws XQueryException {
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		XQueryExpression expr = arg.getExpression();
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (/*!noForce ||*/
			expr.getQType() == null)
			expr.accept(this);
		// type checking
		if (expr.getQType() == null)
			throw new TypeException("Expression " + expr + " could not be typed."); //return;
		// type setting
		QType qtype = arg.getSequenceType().getQType();
		arg.setQType(qtype);
	}

	public void visit(CData arg) throws XQueryException {
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_STRING)));
	}

	/*
	[101]    CompNSConstructor    ::=    ("namespace" NCName "{") Expr "}" 
	
	A computed namespace constructor (CompNSConstructor) constructs a new namespace node with its own node identity. 
	The immediately enclosing expression of the computed namespace constructor must be a computed element constructor; 
	otherwise a static error is raised.[err:XQ0042] The constructed namespace node is attached to the element node constructed by the enclosing expression.
	
	A constructed namespace node is the dynamic equivalent of a namespace declaration attribute. 
	It binds a namespace prefix represented as NCName in the syntax) to a URI and adds the namespace prefix to the in-scope namespaces for its enclosing element.
	
	The content expression of a computed namespace constructor is processed as follows:
	
	Atomization is applied to the value of the content expression, converting it to a sequence of atomic values.
	
	If the result of atomization is an empty sequence, it is replaced by a zero-length string. Otherwise, each atomic value in the atomized sequence is cast into a string.
	
	The individual strings resulting from the previous step are merged into a single string by concatenating them with a single space character between each pair. 
	The resulting string becomes the content (URI) of the constructed namespace node.
	*/

	public void visit(ComputedNamespace arg) throws XQueryException {
		if (noForce && arg.getQType() != null)
			return;
		XQueryExpression expr = arg.getExpression();
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (/*!noForce ||*/
			expr.getQType() == null)
			expr.accept(this);
		// type checking
		if (expr.getQType() == null)
			throw new TypeException("Computed namespace " + arg + " argument could not be typed."); //return;
		QType valueQType = expr.getQType();
		if (!valueQType.canBeCastedInto(QType.PRIMITIVE_string) && !valueQType.isString()) {
			valueQType = valueQType.applyDATAFunction();
			if (valueQType == null || (!valueQType.canBeCastedInto(QType.PRIMITIVE_string) && !valueQType.isString()))
				throw new TypeException("Computed namespace " + arg + " argument has incorrect type.");
		}
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_STRING)));
	}

	/*
	[105]    CompTextConstructor    ::=    "text" "{" Expr? "}" 
	
	All text node constructors are computed constructors. The result of a text node constructor is a new text node, with its own node identity.
	The content expression of a text node constructor is processed as follows:
	
	Atomization is applied to the value of the content expression, converting it to a sequence of atomic values.
	
	If the result of atomization is an empty sequence, no text node is constructed. Otherwise, each atomic value in the atomized sequence is cast into a string.
	
	The individual strings resulting from the previous step are merged into a single string by concatenating them with a single space character between each pair. 
	The resulting string becomes the content of the constructed text node.
	*/

	public void visit(ComputedText arg) throws XQueryException {
		if (noForce && arg.getQType() != null)
			return;
		XQueryExpression expr = arg.getExpression();
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (/*!noForce ||*/
			expr.getQType() == null)
			expr.accept(this);
		// type checking
		if (expr.getQType() == null)
			throw new TypeException("Computed text " + arg + " argument could not be typed."); //return;
		QType valueQType = expr.getQType();
		if (!valueQType.canBeCastedInto(QType.PRIMITIVE_string) && !valueQType.isString()) {
			valueQType = valueQType.applyDATAFunction();
			if (valueQType == null || (!valueQType.canBeCastedInto(QType.PRIMITIVE_string) && !valueQType.isString()))
				throw new TypeException("Computed text " + arg + " argument has incorrect type.");
		}
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_STRING)));
	}

	//	public void visit(ContextDeclaration arg) throws XQueryException;

	/*
	3.7.3.3 Document Node Constructors
	[99]    CompDocConstructor    ::=    "document" "{" Expr "}" 
	
	All document node constructors are computed constructors. The result of a document node constructor 
	is a new document node, with its own node identity.
	
	The content expression of a document node constructor is processed as follows:
	
	For each node returned by the content expression, a new deep copy of the node is constructed, 
	including its children, attributes, and namespace nodes (if any). Each copied node has a new node identity. 
	Copied element nodes are given the type annotation xs:anyType, and copied attribute nodes are given the type annotation xs:anySimpleType. 
	For each adjacent sequence of one or more atomic values returned by the content expression, a new text node is constructed, 
	containing the result of casting each atomic value to a string, with a single blank character inserted between adjacent values. 
	The resulting sequence of nodes is called the content sequence.
	
	If the content sequence contains a document, attribute, or namespace node, a type error is raised.[err:XQ0028]
	
	The resulting sequence of nodes becomes the children of the new document node.
	
	The base URI of a constructed document node is taken from the static context.
	
	No schema validation is performed on the constructed document. The [XML 1.0] rules that govern the structure of an XML document 
	(for example, the document node must have exactly one child that is an element node) are not enforced by the XQuery document node constructor.
	*/

	public void visit(Document arg) throws XQueryException {
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		XQueryExpression expr = arg.getExpression();
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (/*!noForce ||*/
			expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Computed document '" + arg + "', argument could not be typed"); // return;
		else {
			if (!expr.getQType().isElement())
				throw new TypeException("Computed document '" + arg + "', argument should be an element"); // return;
		}
		// type setting
		arg.setQType(new QTypeDocument(expr.getQType(), QTypeDocument.DOCUMENT_ORIGIN));
	}

	/*
	3.7.3.1 Computed Element Constructors
	[100]    CompElemConstructor    ::=    (("element" QName "{") | ("element" "{" Expr "}" "{")) Expr? "}" 
	
	The name expression of a computed element constructor is processed as follows:
	Atomization is applied to the value of the name expression. If the result of atomization is not a single atomic value of type xs:QName, xs:string, or xdt:untypedAtomic, a type error is raised.[err:XP0004][err:XP0006]
	If the atomized value of the name expression is of type xs:QName, that value is used as the name of the constructed element.
	If the atomized value of the name expression is of type xs:string or xdt:untypedAtomic, that value is cast to the type xs:QName using the rules in [XQuery 1.0 and XPath 2.0 Functions and Operators]. 
	The resulting value is used as the name of the constructed element. A dynamic error is raised if the cast is not successful.[err:XP0021]
	
	The content expression of a computed element constructor is processed as follows:
	For each node returned by the content expression, a new deep copy of the node is constructed, including all its children, attributes, and namespace nodes (if any). 
	Each copied node has a new node identity. Copied element nodes are given the type annotation xs:anyType, and copied attribute nodes are given the type annotation xs:anySimpleType. 
	For each adjacent sequence of one or more atomic values returned by the content expression, a new text node is constructed, containing the result of casting each atomic value to a string, with a single blank character inserted between adjacent values. 
	If any of these atomic values cannot be cast into a string, a dynamic error [err:XQ0052] is raised. The resulting sequence of nodes is called the content sequence. 
	Any sequence of adjacent text nodes in the content sequence is merged into a single text node.
	If the content sequence contains a document node, a type error is raised.[err:XQ0023]
	If the content sequence contains a namespace node following a node that is not a namespace node, a type error is raised.[err:XQ0040] Namespace nodes occurring in the content sequence are attached to the constructed element node.
	If the content sequence contains an attribute node following a node that is not an attribute node or a namespace node, a type error is raised.[err:XQ0024] Attribute nodes occurring in the content sequence become attributes of the new element node. 
	If two or more of these attribute nodes have the same name, an error is raised.[err:XQ0025]
	Element, text, comment, and processing instruction nodes in the content sequence become the children of the constructed element node.
	*/
	public void visit(Element arg) throws XQueryException {
		if (noForce && arg.getQType() != null)
			return;
		ArrayList attributes = arg.getAttributes();
		ArrayList expressions = arg.getSubExpressions();
		XQueryExpression startTag = arg.getStartTag();
		// forward type checking
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null) {
			if (attributes != null)
				for (int i = 0; i < attributes.size(); i++)
					if (((AttributeValuePair) attributes.get(i)).getQType() == null)
						return;
			if (expressions != null)
				for (int j = 0; j < expressions.size(); j++)
					if (((XQueryExpression) expressions.get(j)).getQType() == null)
						return;
		}
		if (attributes != null)
			for (int i = 0; i < attributes.size(); i++)
				if (/*!noForce ||*/
					 ((AttributeValuePair) attributes.get(i)).getQType() == null) {
					((AttributeValuePair) attributes.get(i)).accept(this);
					if (((AttributeValuePair) attributes.get(i)).getQType() == null)
						throw new TypeException("Expression " + (AttributeValuePair) attributes.get(i) + " could not be typed."); //return;
				}
		if (expressions != null)
			for (int j = 0; j < expressions.size(); j++) {
				XQueryExpression exprj = (XQueryExpression) expressions.get(j);
				if (/*!noForce ||*/
					exprj.getQType() == null) {
					exprj.accept(this);
					if (exprj.getQType() == null)
						throw new TypeException("Expression " + exprj + " could not be typed."); //return;
				}
				if (exprj.getQType().isBoolean()) {
					if (exprj instanceof XQueryExpressionSequence)
						exprj = (XQueryExpression) ((XQueryExpressionSequence) exprj).getSubExpressions().get(0);
					if (!(exprj instanceof LocatedExpression) && !(exprj instanceof Variable))
						throw new TypeException("Incorrect boolean typed expression in element " + exprj); //return;
				}
			}
		// type checking
		QType tagQType = startTag.getQType();
		if (tagQType != null && (!tagQType.isQName() || tagQType.isMultiple())) {
			tagQType = tagQType.applyDATAFunction();
			if (tagQType == null || !tagQType.isQName() || tagQType.isMultiple())
				throw new TypeException("Expression '" + startTag + "', of element constructor has incorrect type.");
		}
		// type setting
		Set allElements = new HashSet();
		if (arg.getAttributes() != null)
			allElements.addAll(arg.getAttributes());
		if (arg.getSubExpressions() != null)
			allElements.addAll(arg.getSubExpressions());
		arg.setQType(buildQTypeElementFromSubElements(arg.getStartTag(), allElements));
	}

//	public void visit(ExternalVariable arg) throws XQueryException {
//		Variable var = (Variable) arg.getObjectValue();
//		var.accept(this);
//		arg.setQType(var.getQType());
//		// test if qtype is null; any type by default ....
//		if (arg.getQType() == null) {
//			ArrayList list = new ArrayList(6);
//			list.add(new QTypeDocument(null, QTypeDocument.ANY_ORIGIN));
//			list.add(new QTypeProcessingInstruction());
//			list.add(new QTypeComment());
//			list.add(new QTypeText((SimpleType) getType(SC_STRING)));
//			list.add(new QTypeElement(new QName("*", null), getType(SC_ANYTYPE)));
//			list.add(new QTypeAttribute(new QName("*", null), context.getSchemaManager().getAnySimpleType()));
//			arg.setQType(new QTypeUnion(list));
//		}
//	}

	public void visit(FLWRExpression arg) throws XQueryException {
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		ArrayList variables = arg.getVariables();
		XQueryExpression whereClause = arg.getWhereClause();
		XQueryExpression returnClause = arg.getReturnClause();
		// forward type checking
		boolean multiple = false;
		for (int i = 0; i < variables.size(); i++) {
			Variable vari = (Variable) variables.get(i);
			if (/*!noForce ||*/
				vari.getQType() == null)
				vari.accept(this);
			if (!multiple)
				if (vari.getBindingType() == Constants.FOR_BINDINGTYPE && vari.getExpression().getQType().isMultiple())
					multiple = true;

		}
		if (whereClause != null && (/*!noForce ||*/
			whereClause.getQType() == null))
			whereClause.accept(this);
		if (/*!noForce ||*/
			returnClause.getQType() == null)
			returnClause.accept(this);
		// type verification
		if (returnClause.getQType() == null)
			throw new TypeException("Incorrect 'return' clause : [ " + returnClause + " ], it could not be typed.");
		if (whereClause != null && whereClause.getQType() == null)
			throw new TypeException("Incorrect 'where' clause : [ " + whereClause + " ], it could not be typed.");
		// verify sort clause typing
		if (arg.getOrderBy() != null) {
			for (int i = 0; i < arg.getOrderBy().size(); i++) {
				XQueryExpression sortiexpr = (XQueryExpression) arg.getOrderBy().get(i);
				if (sortiexpr.getOrder(0) != Constants.PLACE_ORDER) {
					QType qtypei = sortiexpr.getQType();
					if (qtypei == null || !qtypei.isAtomic() || qtypei.isMultiple())
						throw new TypeException("Invalid order by clause : " + sortiexpr);
				}
			}
		}
		// type setting
		arg.setQType((QType) arg.getReturnClause().getQType().changeOccurence(multiple ? QType.OCC_0_N : QType.OCC_1_1));
	}

	// public void visit(FunctionCall arg) throws XQueryException;

	public void visit(FunctionDeclaration arg) throws XQueryException {
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// ArrayList expressions = arg.getExpressions();
		SequenceType returnType = arg.getReturnType();
		// forward type checking --> cannot be done because of variables not binded in parameters
		// type setting
		QType qtype = null;
		// TODO verify that retrun type and calculated type are compatible
		if (returnType == null) {
			if (arg.getExpression() != null) {
			    qtype = arg.getExpression().getQType();
			}
			if (qtype == null) {
				SimpleType type = context.getSchemaManager().getAnySimpleType();
				//type.setPrimitive(PrimitiveType.createType(SC_ANYSIMPLETYPE));
				ArrayList list = new ArrayList(3);
				list.add(new QTypeAtom(type));
				list.add(new QTypeAttribute(new QName("*", "*", "*", arg.getParentModule(), null), type));
				list.add(new QTypeElement(new QName("*", "*", "*", arg.getParentModule(), null), getType(SC_ANYTYPE)));
				qtype = new QTypeUnion(list, QType.OCC_0_N);
			}
		} else {
			qtype = returnType.getQType();
		}
		arg.setQType(qtype);
	}

	//	public void visit(InstanceOfExpression arg) throws XQueryException;

	public void visit(InternalFunctionCall arg) throws XQueryException {
		SimpleType type = context.getSchemaManager().getAnySimpleType();
		ArrayList list = new ArrayList(3);
		list.add(new QTypeAtom(type));
		list.add(new QTypeAttribute(new QName("*", "*", "*", arg.getParentModule(), null), type));
		list.add(new QTypeElement(new QName("*", "*", "*", arg.getParentModule(), null), getType(SC_ANYTYPE)));
		QType qtype = new QTypeUnion(list, QType.OCC_0_N);
		arg.setQType(qtype);
	}

	public void visit(ITEExpression arg) throws XQueryException {
		if (noForce && arg.getQType() != null)
			return;
		XQueryExpression ifExpression = arg.getThenExpression();
		XQueryExpression thenExpression = arg.getThenExpression();
		XQueryExpression elseExpression = arg.getElseExpression();
		// forward type checking
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null) {
			if (ifExpression.getQType() == null)
				return;
			if (thenExpression.getQType() == null)
				return;
			if (elseExpression.getQType() == null)
				return;
		}
		if (ifExpression.getQType() == null)
			ifExpression.accept(this);
		if ((thenExpression.getQType() == null))
			thenExpression.accept(this);
		if ((elseExpression.getQType() == null))
			elseExpression.accept(this);
		// type checking
		if (ifExpression.getQType() == null)
			throw new TypeException("Expression " + ifExpression + " could not be typed."); //return;
		if (thenExpression.getQType() == null)
			throw new TypeException("Expression " + thenExpression + " could not be typed."); //return;
		if (elseExpression.getQType() == null)
			throw new TypeException("Expression " + elseExpression + " could not be typed."); //return;
		// type setting
		ArrayList tmpList = new ArrayList(2);
		tmpList.add(thenExpression.getQType());
		tmpList.add(elseExpression.getQType());
		arg.setQType(makeUnion(tmpList));
	}

	/*
	[127]    ItemType    ::=    AtomicType | KindTest | ("item" "(" ")") 
	[126]    AtomicType    ::=    QName 
	[128]    KindTest    ::=    DocumentTest | ElementTest | AttributeTest | PITest | CommentTest | TextTest | AnyKindTest
	*/
	/*
	QNames appearing in a SequenceType have their prefixes expanded to namespace URIs by means of the in-scope namespaces and the default element/type namespace. 
	It is a static error [err:XP0008] to use a TypeName in an ElementTest or AttributeTest if that name is not found in the in-scope type definitions. 
	It is a static error [err:XP0008] to use an ElementName in an ElementTest if that name is not found in the in-scope element definitions unless a TypeNameOrWildcard is specified. 
	It is a static error [err:XP0008] to use a (SchemaContextPath ElementName) pair in an ElementTest if the ElementName can not be located from the in-scope element definitions using the SchemaContextPath. 
	It is a static error [err:XP0008] to use an AttributeName in an AttributeTest if that name is not found in the in-scope attribute definitions unless a TypeNameOrWildcard is specified. 
	It is a static error [err:XP0008] to use a (SchemaContextPath AttributeName) pair in an AttributeTest if the AttributeName can not be located from the in-scope attribute definitions using the SchemaContextPath. 
	If a QName that is used as an AtomicType is not defined as an atomic type in the in-scope type definitions, a static error is raised. [err:XP0051]
	
	Here are some examples of SequenceTypes that might be used in XQuery expressions:
	xs:date refers to the built-in Schema type date
	attribute()? refers to an optional attribute
	element() refers to any element
	element(po:shipto, po:address) refers to an element that has the name po:shipto (or is in the substitution group of that element), and has the type annotation po:address (or a subtype of that type)
	element(po:shipto, *) refers to an element named po:shipto (or in the substitution group of po:shipto), with no restrictions on its type
	element(*, po:address) refers to an element of any name that has the type annotation po:address (or a subtype of po:address). If the keyword nillable were used following po:address, that would indicate that the element may have empty content and the attribute xsi:nil="true", even though the declaration of the type po:address has required content.
	node()* refers to a sequence of zero or more nodes of any type
	item()+ refers to a sequence of one or more nodes or atomic values
	*/
	public void visit(ItemType arg) throws XQueryException {
		XQueryExpression expr = arg.getExpression();
		QType qtype = null;
		if (expr == null) {
			ArrayList list = new ArrayList(7);
			list.add(new QTypeAtom(context.getSchemaManager().getAnySimpleType()));
			list.add(new QTypeDocument(null, QTypeDocument.ANY_ORIGIN));
			list.add(new QTypeProcessingInstruction());
			list.add(new QTypeComment());
			list.add(new QTypeText((SimpleType) getType(SC_STRING)));
			list.add(new QTypeElement(new QName("*", null), getType(SC_ANYTYPE)));
			list.add(new QTypeAttribute(new QName("*", null), context.getSchemaManager().getAnySimpleType()));
			qtype = new QTypeUnion(list);
		} else if (expr instanceof QName) {
			QName qname = (QName) expr;
			Type st = null;
			if (qname.getPrefixName() != null && qname.getPrefixName().equals("xs") && qname.getLocalName().equals("anySimpleType")) {
				st = context.getSchemaManager().getAnySimpleType();
			} else {
				// generate QTypeAtom			
				st = context.getSchemaManager().getType(((QName) expr).getNameSpace(), ((QName) expr).getLocalName());
			}
			if (st != null && st instanceof SimpleType)
				qtype = new QTypeAtom((SimpleType) st);
			else
				throw new TypeException("Expression " + expr + " is not a valid atomic type.");
		} else if (expr instanceof NodeTest) {
			NodeTest nodeType = (NodeTest) expr;
			XQueryExpression ex = nodeType.getArg();
			int kind = nodeType.getKind();
			switch (kind) {
				case NodeKind.DOCUMENT :
					qtype = new QTypeDocument((ex == null) ? null : ex.getQType(), QTypeDocument.ANY_ORIGIN);
					break;
				case NodeKind.PI :
					qtype = new QTypeProcessingInstruction();
					break;
				case NodeKind.COMMENT :
					qtype = new QTypeComment();
					break;
				case NodeKind.TEXT :
					qtype = new QTypeText((SimpleType) getType(SC_STRING));
					break;
				case NodeKind.NODE :
					ArrayList list = new ArrayList(6);
					list.add(new QTypeDocument(null, QTypeDocument.ANY_ORIGIN));
					list.add(new QTypeProcessingInstruction());
					list.add(new QTypeComment());
					list.add(new QTypeText((SimpleType) getType(SC_STRING)));
					list.add(new QTypeElement(new QName("*", null), getType(SC_ANYTYPE)));
					list.add(new QTypeAttribute(new QName("*", null), context.getSchemaManager().getAnySimpleType()));
					qtype = new QTypeUnion(list);
					break;
			}
		} else if (expr instanceof ElementTest || expr instanceof AttributeTest) {
			qtype = expr.getQType();
		}
		arg.setQType(qtype);
	}

	public void visit(ElementTest arg) throws XQueryException {
		Type type = null;
		QName typeQName = arg.getTypeQName();
		if (typeQName != null) {
			type = context.getSchemaManager().getType(typeQName.getNameSpace(), typeQName.getLocalName());
			if (type == null)
				throw new TypeException("Could not type ElementTest '" + arg + "' .");
		}
		if (arg.getElementQName() == null || arg.getElementQName().getLocalName().equals("*")) {
			if (type == null)
				type = getType(SC_ANYTYPE);
			arg.setQType(new QTypeElement(new QName("*", null), type));
		} else {
			QName eltQName = arg.getElementQName();
			ElementDeclaration eltDecl = context.getSchemaManager().getElementDeclaration(eltQName.getNameSpace(), eltQName.getLocalName());
			if (eltDecl == null)
				throw new TypeException("Could not type ElementTest '" + arg + "' .");
			// verifying type is compatible
			if (type == null) {
				type = eltDecl.getType();
			} else if (eltDecl.getType() != type && !eltDecl.getType().isDerivedFrom(type))
				throw new TypeException("Could not type ElementTest '" + arg + "' .");
			arg.setQType(new QTypeElement(eltQName, eltDecl, type));
		}
	}

	public void visit(AttributeTest arg) throws XQueryException {
		SimpleType stype = null;
		QName typeQName = arg.getTypeQName();
		if (typeQName != null) {
			Type type = context.getSchemaManager().getType(typeQName.getNameSpace(), typeQName.getLocalName());
			if (type == null || !(type instanceof SimpleType))
				throw new TypeException("Could not type AttributeTest '" + arg + "' .");
			stype = (SimpleType) type;
		}
		if (arg.getAttributeQName() == null || arg.getAttributeQName().equals("*")) {
			if (stype == null)
				stype = context.getSchemaManager().getAnySimpleType();
			arg.setQType(new QTypeAttribute(new QName("*", null), stype));
		} else {
			QName qname = arg.getAttributeQName();
			AttributeDeclaration attDecl = context.getSchemaManager().getAttributeDeclaration(qname.getNameSpace(), qname.getLocalName());
			if (attDecl == null)
				throw new TypeException("Could not type AttributeTest '" + arg + "' .");
			// verifying type is compatible
			if (stype == null) {
				stype = attDecl.getType().getValueType();
			} else if (attDecl.getType() != stype && !attDecl.getType().isDerivedFrom(stype))
				throw new TypeException("Could not type AttributeTest '" + arg + "' .");
			arg.setQType(new QTypeAttribute(qname, attDecl, stype));
		}
	}

	public void visit(SchemaElementTest arg) throws XQueryException {
		Type type = null;
		if (arg.getElementQName() == null || arg.getElementQName().equals("*")) {
			if (type == null)
				type = getType(SC_ANYTYPE);
			arg.setQType(new QTypeElement(new QName("*", null), type));
		} else {
			QName eltQName = arg.getElementQName();
			ElementDeclaration eltDecl = context.getSchemaManager().getElementDeclaration(eltQName.getNameSpace(), eltQName.getLocalName());
			if (eltDecl == null)
				throw new TypeException("Could not type SchemaElementTest '" + arg + "' .");
			// verifying type is compatible
			if (type == null) {
				type = eltDecl.getType();
			} else if (eltDecl.getType() != type && !eltDecl.getType().isDerivedFrom(type))
				throw new TypeException("Could not type SchemaElementTest '" + arg + "' .");
			arg.setQType(new QTypeElement(eltQName, eltDecl, type));
		}
	}

	public void visit(SchemaAttributeTest arg) throws XQueryException {
		SimpleType stype = null;
		if (arg.getAttributeQName() == null || arg.getAttributeQName().equals("*")) {
			if (stype == null)
				stype = context.getSchemaManager().getAnySimpleType();
			arg.setQType(new QTypeAttribute(new QName("*", null), stype));
		} else {
			QName qname = arg.getAttributeQName();
			AttributeDeclaration attDecl = context.getSchemaManager().getAttributeDeclaration(qname.getNameSpace(), qname.getLocalName());
			if (attDecl == null)
				throw new TypeException("Could not type SchemaAttributeTest '" + arg + "' .");
			// verifying type is compatible
			if (stype == null) {
				stype = attDecl.getType().getValueType();
			} else if (attDecl.getType() != stype && !attDecl.getType().isDerivedFrom(stype))
				throw new TypeException("Could not type SchemaAttributeTest '" + arg + "' .");
			arg.setQType(new QTypeAttribute(qname, attDecl, stype));
		}
	}

	public void visit(LibraryFunctionCall arg) throws XQueryException {
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		FunctionDeclaration funcdef = arg.getFunctionDefinition(); // cannot be null
		// forward type checking
		if (funcdef.getQType() == null)
			funcdef.accept(this);
		// type setting
		arg.setQType(funcdef.getQType());
	}

	public void visit(ListOpAFTERExpression arg) throws XQueryException {
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// TYPE BUILDING
		// the type returned is a sequence of elements that have the first sequence type

		XQueryExpression expr1 = arg.getExpression1();
		XQueryExpression expr2 = arg.getExpression2();
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null) {
			if (expr1.getQType() == null)
				return;
			if (expr2.getQType() == null)
				return;
		}
		if (/*!noForce ||*/
			expr1.getQType() == null)
			expr1.accept(this);
		if (/*!noForce ||*/
			expr2.getQType() == null)
			expr2.accept(this);
		// type verification
		if (expr1.getQType() == null)
			throw new TypeException("Expression " + expr1 + " could not be typed."); //return;
		if (expr2.getQType() == null)
			throw new TypeException("Expression " + expr2 + " could not be typed."); //return;
		if (!expr1.getQType().isNode() || !expr2.getQType().isNode())
			throw new TypeException("Expression '" + arg + "', at least one of the operands must be a node or collection of node.");
		// type setting
		arg.setQType(expr1.getQType());
	}

	/*
	An arithmetic expression is evaluated by applying the following rules, in order, until an error is raised or a value is computed:
	
	Atomization is applied to each operand.
	If either operand is now an empty sequence, the result of the operation is an empty sequence.
	If either operand is now a sequence of length greater than one, a type error is raised.[err:XP0006]
	If either operand is now of type xdt:untypedAtomic, it is cast to the default type for the given operator. 
	The default type for the idiv operator is xs:integer; the default type for all other arithmetic operators is xs:double. 
	If the cast fails, a dynamic error is raised.[err:XP0021]
	If the operand types are now valid for the given operator, the operator is applied to the operands, resulting in an atomic value or a dynamic error (for example, an error might result from dividing by zero.) 
	The combinations of atomic types that are accepted by the various arithmetic operators, and their respective result types, are listed in B.2 
	Operator Mapping together with the functions in [XQuery 1.0 and XPath 2.0 Functions and Operators] that define the semantics of the operation for each type.
	
	If the operand types are still not valid for the given operator, a type error is raised.
	
	XQuery supports two division operators named div and idiv. When invoked with two integer operands, div returns a value of type xs:decimal, but idiv returns a value of type xs:integer.
	*/
	/*
	 The following functions define the semantics of operators defined in [XQuery 1.0: An XML Query Language] and [XPath 2.0] on these numeric types.
	Operators Meaning 
	op:numeric-add Addition 
	op:numeric-subtract Subtraction 
	op:numeric-multiply Multiplication 
	op:numeric-divide Division 
	op:numeric-integer-divide Integer division 
	op:numeric-mod Modulus 
	op:numeric-unary-plus Unary plus 
	op:numeric-unary-minus Unary minus (negation) 
	
	The parameters and return types for the above operators are the basic numeric types: xs:integer, xs:decimal, xs:float and xs:double, 
	and types derived from them. The word "numeric" in function signatures signifies these four types. 
	For simplicity, each operator is defined to operate on operands of the same type and to return the same type. 
	The one exception is op:numeric-divide, which returns an xs:decimal if called with two xs:integer operands.)
	
	Operands of type xdt:untypedAtomic are converted to xs:double, except for arguments to 6.2.5 op:numeric-integer-divide which are converted to xs:integer. 
	If the two operands are not of the same type, subtype substitution and type promotion are used to obtain two operands of the same type. 
	Section B.1 Type PromotionXQ describes the semantics of these operations in detail.
	
	Subtype substitution: A derived type may substitute for its base type. In particular, xs:integer may be used where xs:decimal is expected.
	
	Type promotion: xs:decimal may be promoted to xs:float, and xs:float may be promoted to xs:double.
	
	The result type of operations depends on their argument datatypes and is defined in the following table:
	
	Operator Returns 
	op:operation(xs:integer, xs:integer) xs:integer (except for op:numeric-divide(integer, integer), which returns xs:decimal) 
	op:operation(xs:decimal, xs:decimal) xs:decimal 
	op:operation(xs:float, xs:float) xs:float 
	op:operation(xs:double, xs:double) xs:double 
	op:operation(xs:integer) xs:integer 
	op:operation(xs:decimal) xs:decimal 
	op:operation(xs:float) xs:float 
	op:operation(xs:double) xs:double 
	
	*/

	public void visit(ListOpArithExpression arg) throws XQueryException {
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		XQueryExpression expr1 = arg.getExpression1();
		XQueryExpression expr2 = arg.getExpression2();
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null) {
			if (expr1.getQType() == null)
				return;
			if (expr2.getQType() == null)
				return;
		}
		if (/*!noForce ||*/
			expr1.getQType() == null)
			expr1.accept(this);
		if (/*!noForce ||*/
			expr2.getQType() == null)
			expr2.accept(this);
		// type verification
		QType qtype1 = expr1.getQType();
		QType qtype2 = expr2.getQType();
		if (qtype1 == null)
			throw new TypeException("Expression " + expr1 + " could not be typed."); //return;
		if (qtype2 == null)
			throw new TypeException("Expression " + expr2 + " could not be typed."); //return;
		if (!qtype1.isAtomic())
			qtype1 = qtype1.applyDATAFunction();
		if (!qtype2.isAtomic())
			qtype2 = qtype2.applyDATAFunction();
		if (qtype1 == null || (!qtype1.isAtomic() /*&& !inPredicate*/))
			throw new TypeException("Expression '" + arg + "', first operand is not an atomic value.");
		if (qtype2 == null || (!qtype2.isAtomic() /*&& !inPredicate*/))
			throw new TypeException("Expression '" + arg + "', second operand is not an atomic value.");
		if (!qtype1.isNumeric())
			throw new TypeException("Expression '" + arg + "', first operand is not numeric.");
		if (!qtype2.isNumeric())
			throw new TypeException("Expression '" + arg + "', second operand is not numeric.");
		// type setting
		int primType1 = ((SimpleType) qtype1.getType()).getPrimitive().getType();
		int primType2 = ((SimpleType) qtype2.getType()).getPrimitive().getType();
		if (primType1 == PrimitiveType.DOUBLE || primType2 == PrimitiveType.DOUBLE)
			arg.setQType(new QTypeAtom((SimpleType) getType(SC_DOUBLE), (expr1.getQType().isMultiple() || expr2.getQType().isMultiple()) ? QType.OCC_0_N : QType.OCC_1_1));
		else if (primType1 == PrimitiveType.FLOAT || primType2 == PrimitiveType.FLOAT)
			arg.setQType(new QTypeAtom((SimpleType) getType(SC_FLOAT), (expr1.getQType().isMultiple() || expr2.getQType().isMultiple()) ? QType.OCC_0_N : QType.OCC_1_1));
		else if (primType1 == PrimitiveType.DECIMAL || primType2 == PrimitiveType.DECIMAL)
			arg.setQType(new QTypeAtom((SimpleType) getType(SC_DECIMAL), (expr1.getQType().isMultiple() || expr2.getQType().isMultiple()) ? QType.OCC_0_N : QType.OCC_1_1));
		else
			arg.setQType(new QTypeAtom((SimpleType) getType(SC_INTEGER), (expr1.getQType().isMultiple() || expr2.getQType().isMultiple()) ? QType.OCC_0_N : QType.OCC_1_1));
	}

	public void visit(ListOpBEFOREExpression arg) throws XQueryException {
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		XQueryExpression expr1 = arg.getExpression1();
		XQueryExpression expr2 = arg.getExpression2();
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null) {
			if (expr1.getQType() == null)
				return;
			if (expr2.getQType() == null)
				return;
		}
		if (/*!noForce ||*/
			expr1.getQType() == null)
			expr1.accept(this);
		if (/*!noForce ||*/
			expr2.getQType() == null)
			expr2.accept(this);
		// type verification
		if (expr1.getQType() == null)
			throw new TypeException("Expression " + expr1 + " could not be typed."); //return;
		if (expr2.getQType() == null)
			throw new TypeException("Expression " + expr2 + " could not be typed."); //return;
		if (!expr1.getQType().isNode() || !expr2.getQType().isNode())
			throw new TypeException("Expression '" + arg + "', at least one of the operands must be a node or collection of node.");
		// type setting
		arg.setQType(expr1.getQType());
	}

	public void visit(ListOpCompExpression arg) throws XQueryException {
		// TYPE VERIFICATION
		// both expressions must be sequences of elements of the same document
		// TYPE BUILDING
		// The =, !=, <, <=, >, and >= operators perform value comparisons.
		// If both operands are simple values of the same type, the result is straightforward.
		// If the operands are simple values of compatible types, the operand of the less-inclusive type is converted to the more-inclusive type for the purpose of comparison (for example, an integer might be converted to a float in order to be compared with a float.)
		// If one operand is a node and the other is a simple value, the content of the node is extracted by an implicit invocation of the data function before the comparison is performed.
		// If both operands are nodes, the string-values of the nodes are compared, as defined in [XPath 1.0].

		if (noForce && arg.getQType() != null)
			return;
		//arg.setQType(new QTypeAtom((SimpleType) getType(SC_BOOLEAN)));

		XQueryExpression expr1 = arg.getExpression1();
		XQueryExpression expr2 = arg.getExpression2();
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null) {
			if (expr1.getQType() == null)
				return;
			if (expr2.getQType() == null)
				return;
		}
		if (/*!noForce ||*/
			expr1.getQType() == null)
			expr1.accept(this);
		if (/*!noForce ||*/
			expr2.getQType() == null)
			expr2.accept(this);
		// type checking
		QType qtype1 = expr1.getQType();
		QType qtype2 = expr2.getQType();
		if (qtype1 == null)
			throw new TypeException("Expression " + expr1 + " could not be typed."); //return;
		if (qtype2 == null)
			throw new TypeException("Expression " + expr2 + " could not be typed."); //return;
		if (qtype1 != null && qtype2 != null) {
			if (!qtype1.isAtom())
				qtype1 = qtype1.applyDATAFunction();
			if (!qtype2.isAtom())
				qtype2 = qtype2.applyDATAFunction();
			// test for value comp
			if (arg.getOperator() >= Constants.EQ_VALUECOMP && arg.getOperator() <= Constants.NE_VALUECOMP) {
				// verify occurence is one
				if (qtype1.getOccurence() != QType.OCC_0_1 && qtype1.getOccurence() != QType.OCC_1_1)
					throw new TypeException("Expression " + expr1 + " has wrong occurence.");
				if (qtype2.getOccurence() != QType.OCC_0_1 && qtype2.getOccurence() != QType.OCC_1_1)
					throw new TypeException("Expression " + expr2 + " has wrong occurence.");
			}
			// if (!(qtype1.isNumeric() && qtype2.isNumeric()) && !(qtype1.canBeComparedTo(qtype2)))
			if (qtype1 == null || !(qtype1.canBeComparedTo(qtype2)))
				throw new TypeException("Could not compare types of expression --> " + arg);
		}
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_BOOLEAN)));
	}

	public void visit(ListOpEXCEPTExpression arg) throws XQueryException {
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		XQueryExpression expr1 = arg.getExpression1();
		XQueryExpression expr2 = arg.getExpression2();
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null) {
			if (expr1.getQType() == null)
				return;
			if (expr2.getQType() == null)
				return;
		}
		if (/*!noForce ||*/
			expr1.getQType() == null)
			expr1.accept(this);
		if (/*!noForce ||*/
			expr2.getQType() == null)
			expr2.accept(this);
		// type verification
		if (expr1.getQType() == null)
			throw new TypeException("Expression " + expr1 + " could not be typed."); //return;
		if (expr2.getQType() == null)
			throw new TypeException("Expression " + expr2 + " could not be typed."); //return;
		// type setting
		arg.setQType(expr1.getQType());
	}

	public void visit(ListOpINTERSECTExpression arg) throws XQueryException {
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		XQueryExpression expr1 = arg.getExpression1();
		XQueryExpression expr2 = arg.getExpression2();
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null) {
			if (expr1.getQType() == null)
				return;
			if (expr2.getQType() == null)
				return;
		}
		if (/*!noForce ||*/
			expr1.getQType() == null)
			expr1.accept(this);
		if (/*!noForce ||*/
			expr2.getQType() == null)
			expr2.accept(this);
		// type verification
		if (expr1.getQType() == null)
			throw new TypeException("Expression " + expr1 + " could not be typed."); //return;
		if (expr2.getQType() == null)
			throw new TypeException("Expression " + expr2 + " could not be typed."); //return;
		// type setting
		arg.setQType(expr1.getQType());
	}

	public void visit(ListOpUNIONExpression arg) throws XQueryException {
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		XQueryExpression expr1 = arg.getExpression1();
		XQueryExpression expr2 = arg.getExpression2();
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null) {
			if (expr1.getQType() == null)
				return;
			if (expr2.getQType() == null)
				return;
		}
		if (/*!noForce ||*/
			expr1.getQType() == null)
			expr1.accept(this);
		if (/*!noForce ||*/
			expr2.getQType() == null)
			expr2.accept(this);
		// type verification
		if (expr1.getQType() == null)
			throw new TypeException("Expression " + expr1 + " could not be typed."); //return;
		if (expr2.getQType() == null)
			throw new TypeException("Expression " + expr2 + " could not be typed."); //return;
		// type setting
		ArrayList list = new ArrayList();
		list.add(expr1.getQType());
		list.add(expr2.getQType());
		arg.setQType(new QTypeUnion(list));
	}

	public void visit(LocatedExpression arg) throws XQueryException {
		ArrayList steps = arg.getSteps();
		// checking force stuff
		if (noForce) {
			boolean alltyped = true;
			for (int i = 0; i < steps.size(); i++) {
				Step stepi = (Step) steps.get(i);
				if (stepi.getQType() == null || stepi.getAxis() == Axis.DESCENDANT || stepi.getAxis() == Axis.DESCENDANT_OR_SELF) {
					alltyped = false;
					break;
				}
			}
			// try to simplify if all steps have qtype
			if (alltyped) {
				arg.setQType(arg.getStepNum(steps.size() - 1).getQType());
				arg.setXTrees(arg.getStepNum(steps.size() - 1).getXTrees());
				return;
			}
		} else {
			arg.setQType(null);
			arg.setXTrees(null);
		}

		// look for simple paths to be typed only with their context
		boolean simplePath = arg.isRelative();

		// remember the current context 
		ArrayList tmpContextXNodes = null;
		QType tmpContextQType = null;
		LocatedExpression tmpContextLocated = null;
		int tmpContextLocatedStepIndex = -1;
		// for simple paths chek context and set current context
		if (simplePath) {
			if (arg.getStepNum(0).getAxis() == Axis.NONE)
				arg.getStepNum(0).setAxis(Axis.CHILD);
			if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null) {
				/*
				if (arg.getInPredicate())
					return;
				else
					throw new TypeException(arg + " could not be typed because context does not exist.");
				*/
				return;
			}
			tmpContextXNodes = contextXNodes;
			tmpContextQType = contextQType;
			tmpContextLocated = contextLocated;
			tmpContextLocatedStepIndex = contextLocatedStepIndex;
			arg.setContext(true);
		}
		// if not simple path set current context anyway
		else {
			if (contextXNodes != null || contextQType != null) {
				tmpContextXNodes = contextXNodes;
				tmpContextQType = contextQType;
				tmpContextLocated = contextLocated;
				tmpContextLocatedStepIndex = contextLocatedStepIndex;
				contextXNodes = null;
				contextQType = null;
				tmpContextLocated = null;
				contextLocatedStepIndex = -1;
			}
		}

		// forward checking and typing

		byte occurence = QType.OCC_1_1;

		boolean done = false;
		QType qtype = null;
		Step stepi = null;

		for (int i = 0; i < steps.size(); i++) {
			// initialisation
			qtype = null;
			done = false;
			stepi = (Step) steps.get(i);
			XQueryExpression stepiExpr = stepi.getExpression();
			while (stepiExpr instanceof ParenthesizedExpression)
			    stepiExpr = ((ParenthesizedExpression)stepiExpr).getExpression();

			// test to set the contextType for the first step
			if (i == 0 && contextType == QTypeDocument.ANY_ORIGIN) {
				if (stepiExpr.getQType() instanceof QTypeDocument && ((QTypeDocument) stepiExpr.getQType()).isFromDocument())
					contextType = QTypeDocument.DOCUMENT_ORIGIN;
				else
					contextType = QTypeDocument.COLLECTION_ORIGIN;
			}

			// test for existing information
			if (noForce) {
				if (i == 0 && stepi.getQType() != null) {
					contextXNodes = stepi.getXTrees();
					contextQType = stepi.getQType();
					qtype = stepi.getQType();
					done = true;
				}
			} else {
				stepi.setQType(null);
				stepi.setXTrees(null);
			}

			// handle special case of first step
			if (!done) {
				if (i == 0 && stepi.getAxis() == Axis.NONE) {
					qtype = stepiExpr.getQType();
					if (noForce && qtype != null) {
						stepi.setXTrees(stepiExpr.getXTrees());
						done = true;
					}
				}
			}

			// handle special case of step expression being a ListOpUNIONExpression
			if (!done && stepiExpr instanceof ListOpUNIONExpression) {
				this.setTypingContext(contextXNodes, contextQType, arg, i - 1);
				stepiExpr.accept(this);
				qtype = stepiExpr.getQType();
				if (qtype != null) {
					stepi.setXTrees(stepiExpr.getXTrees());
					done = true;
				}
			}

			// special treatment to avoid bug on navigation (does not navigate on text nodes)
			boolean doNodes = true;
			boolean isTextNode = false;
			if (!done && stepiExpr instanceof NodeTest) {
				if (((NodeTest) stepiExpr).getKind() == NodeKind.TEXT) {
					isTextNode = true;
					doNodes = false;
				} else if (((NodeTest) stepiExpr).getKind() == NodeKind.NODE)
					doNodes = false;
			}

			// get next step type
			if (!done && doNodes && contextXNodes != null) {
				ArrayList xnodes = context.resolveXPath(contextXNodes, stepi);
				if (xnodes != null) {
					qtype = this.buildQTypeFromXNodes(xnodes, occurence);
					if (qtype != null) {
						stepi.setXTrees(xnodes);
						done = true;
					}
				}
			}
			// get next step type if it could not be resolved with pathresolver
			if (!done && contextQType != null) {
				qtype = context.resolveQType(contextType, contextQType, stepi, inPredicate);
				if (qtype != null) {
					if (isTextNode)
						stepi.setXTrees(contextXNodes);
					else
						stepi.setXTrees(null);
					done = true;
				}

			}

			// verification of type and collection setting
			if (done) {
				occurence = qtype.getOccurence();
			} else {
				// change LARS 22/09/2003
				if (stepi.getAxis() == Axis.PARENT) {
					if (arg.startsWithVariable() || (arg.getParentExpression() != null && !arg.getParentExpression().isEmpty() && arg.getParentExpression().get(0) instanceof ListOpUNIONExpression)) {
						// add LARS 22/09/2003
						// special code to be able to type parent::node when root expression is variable
						XQueryExpression tmpExpr = arg;
						XQueryExpression firstExpr = arg.getStepNum(0).getExpression();
						ArrayList stepList = new ArrayList();
						int stepIndex = -1;
						try {
							while (true) {
								if (tmpExpr instanceof LocatedExpression) {
									LocatedExpression tmpLoc = (LocatedExpression) tmpExpr;

									if (stepIndex == -1)
										stepIndex = tmpLoc.getSteps().size() - 1;
									for (int j = stepIndex; j >= 0; j--) {
										Step clonedStep = (Step) tmpLoc.getStepNum(j).clone(arg.getParentModule());
										clonedStep.setPredicates(null);
										stepList.add(0, clonedStep);
									}
									if (simplePath && (tmpLoc.getInPredicate() || (tmpLoc.getParentExpression() != null && !tmpLoc.getParentExpression().isEmpty() && tmpLoc.getParentExpression().get(0) instanceof ListOpUNIONExpression))) {
										tmpExpr = contextLocated;
										stepIndex = contextLocatedStepIndex;
										((Step) stepList.get(0)).setHasSeparator(true);
									} else {
										stepIndex = -1;
										firstExpr = tmpLoc.getStepNum(0).getExpression();
										if (firstExpr instanceof Variable) {
											tmpExpr = ((Variable) firstExpr).getExpression();
											stepList.remove(0);
										} else {
											break;
										}
									}
								} else if (tmpExpr instanceof Variable) {
									tmpExpr = ((Variable) tmpExpr).getExpression();
								} else if (tmpExpr instanceof FunctionCall) {
									stepList.add(0, tmpExpr);
									break;
								}
							}

						} catch (CloneNotSupportedException cnse) {
							stepList.clear();
						}

						if (!stepList.isEmpty() && stepList.size() != steps.size()) {
							LocatedExpression loc = new LocatedExpression(stepList, false, arg.getParentModule(), true);
							if (loc.getQType() != null)
								qtype = loc.getQType();
						}
					} else if (arg.getParentExpression() != null && !arg.getParentExpression().isEmpty() && arg.getParentExpression().get(0) instanceof ListOpUNIONExpression) {
						boolean prev = inPredicate;
						inPredicate = true;
						arg.accept(this);
						inPredicate = prev;
					}
				}
			}
			if (qtype == null)
				throw new TypeException("Could not find path : " + arg.toString(null,false,false,true));

			// set typing context
			stepi.setQType(qtype);
			contextXNodes = stepi.getXTrees();
			contextQType = stepi.getQType();

			// handle predicates if any
			ArrayList predicates = stepi.getPredicates();
			if (predicates != null && !predicates.isEmpty()) {
				QType prevContextQType = null;
				if (contextQType != null && contextQType.isMultiple())
					prevContextQType = contextQType.changeOccurence(QType.OCC_1_1);
				else
					prevContextQType = contextQType;
				this.setTypingContext(contextXNodes, prevContextQType, arg, i);
				for (int j = 0; j < predicates.size(); j++) {
					XQueryExpression predicate = (XQueryExpression) predicates.get(j);
					if (/*!noForce ||*/
						predicate.getQType() == null) {
						boolean prev = inPredicate;
						inPredicate = true;
						predicate.accept(this);
						inPredicate = prev;
					}
				}
				if (prevContextQType != null)
					contextQType = prevContextQType;
			}
		}
		// type checking
		//		if (noForce && arg.getQType() != null) {
		//			this.setTypingContext(tmpContextXNodes, tmpContextQType);
		//			return;
		//		}
		if (qtype != null) {
			arg.setQType(qtype);
			//		} else if (noForce && arg.getQType() != null) {
			//			// do nothing

		} else {
			throw new TypeException("Could not find path : " + arg.toString(null,false,false,true));
		}

		// reset the context type
		contextType = QTypeDocument.ANY_ORIGIN;
		// reset the context
		this.setTypingContext(tmpContextXNodes, tmpContextQType, tmpContextLocated, tmpContextLocatedStepIndex);
	}

	public void visit(NodeTest arg) throws XQueryException {
		// type checking
		if (noForce && arg.getQType() != null)
			return;

		if (contextQType == null)
			return;

		QType qtype = Utils.getSubType(contextType, contextQType, new Step(true, Axis.CHILD, arg, null, null), context.getSchemaManager(), inPredicate);
		arg.setQType(qtype);
		// type setting
		// remove   04/06/2003
		//arg.setQType(new QTypeAtom((SimpleType) getType(SC_STRING)));
	}

	//	public void visit(PrimitiveFunctionCall arg) throws XQueryException;

	public void visit(QName arg) throws XQueryException {
		if (contextXNodes != null && contextQType != null && arg.getParentExpression() != null && arg.getParentExpression().get(0) instanceof SortedExpression) {
			QType qtype = null;
			//setContextNodeStepListType();
			//else setContextNodeStepListType();
			if (contextQType != null) {
				qtype = Utils.getSubType(contextType, contextQType, new Step(false, Axis.CHILD, arg, null, null), context.getSchemaManager(), inPredicate);
			}
			if (qtype == null)
				throw new TypeException("Could not type expression : " + arg.toString(null,false,false,true));
			arg.setQType(qtype);
			return;
		}
		QType qtype = null;
		if (contextQType != null) {
			qtype = Utils.getSubType(contextType, contextQType, new Step(false, Axis.CHILD, arg, null, null), context.getSchemaManager(), inPredicate);
		}
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null) {
			arg.setQType(new QTypeAtom((SimpleType) getType(SC_QNAME)));
			return;
		}
		//		if (qtype == null) {
		//			qtype = buildQTypeFromXNodes(contextXNodes, new Step(false, Axis.CHILD, arg, null, null), true);
		//		}
		if (qtype == null)
			throw new TypeException("Could not type expression : " + arg.toString(null,false,false,true));
		arg.setQType(qtype);
	}

	public void visit(QuantifiedExpression arg) throws XQueryException {
		// TYPE VERIFICATION
		// SATISFIES expression must retunr a boolean
		// TYPE BUILDING
		// type of the direct path (there should be only one!!!)

		// type checking
		if (noForce && arg.getQType() != null)
			return;
		//arg.setQType(new QTypeAtom((SimpleType) getType(SC_BOOLEAN)));

		ArrayList variables = arg.getVariables();
		XQueryExpression expr = arg.getExpression();
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null) {
			for (int i = 0; i < variables.size(); i++)
				if (((Variable) variables.get(i)).getExpression().getQType() == null)
					return;
			if (expr.getQType() == null)
				return;
		}
		for (int i = 0; i < variables.size(); i++)
			if (/*!noForce ||*/
				 ((Variable) variables.get(i)).getExpression().getQType() == null) {
				((Variable) variables.get(i)).getExpression().accept(this);
				if (((Variable) variables.get(i)).getExpression().getQType() == null)
					throw new TypeException("Expression " + ((Variable) variables.get(i)).getExpression() + " could not be typed."); //return;
			}
		if (/*!noForce ||*/
			expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Expression " + expr + " could not be typed."); //return;
		// type verification
		if (!expr.getQType().isBoolean())
			throw new TypeException("SATISFIES clause must be a boolean.");
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_BOOLEAN)));
	}

	public void visit(RangeExpression arg) throws XQueryException {
		// TYPE VERIFICATION
		// both expression must be integers
		// TYPE BUILDING
		// --> NONE

		XQueryExpression expr1 = arg.getExpression1();
		XQueryExpression expr2 = arg.getExpression2();
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null) {
			if (expr1.getQType() == null)
				return;
			if (expr2.getQType() == null)
				return;
		}
		if (/*!noForce ||*/
			expr1.getQType() == null)
			expr1.accept(this);
		if (/*!noForce ||*/
			expr2.getQType() == null)
			expr2.accept(this);
		// type verification
		if (expr1.getQType() == null)
			throw new TypeException("Expression " + expr1 + " could not be typed."); //return;
		if (expr2.getQType() == null)
			throw new TypeException("Expression " + expr2 + " could not be typed."); //return;
		if (!expr1.getQType().isInteger())
			throw new TypeException("Expression '" + arg + "', first expression is not an integer.");
		if (expr2 != null && !expr2.getQType().isInteger())
			throw new TypeException("Expression '" + arg + "', second expression is not an integer.");
	}

	//	public void visit(RootNodeFunctionCall arg) throws XQueryException;


	/*
	QNames appearing in a SequenceType have their prefixes expanded to namespace URIs by means of the in-scope namespaces and the default element/type namespace. 
	It is a static error [err:XP0008] to use a TypeName in an ElementTest or AttributeTest if that name is not found in the in-scope type definitions. 
	It is a static error [err:XP0008] to use an ElementName in an ElementTest if that name is not found in the in-scope element definitions unless a TypeNameOrWildcard is specified. 
	It is a static error [err:XP0008] to use a (SchemaContextPath ElementName) pair in an ElementTest if the ElementName can not be located from the in-scope element definitions using the SchemaContextPath. 
	It is a static error [err:XP0008] to use an AttributeName in an AttributeTest if that name is not found in the in-scope attribute definitions unless a TypeNameOrWildcard is specified. 
	It is a static error [err:XP0008] to use a (SchemaContextPath AttributeName) pair in an AttributeTest if the AttributeName can not be located from the in-scope attribute definitions using the SchemaContextPath. 
	If a QName that is used as an AtomicType is not defined as an atomic type in the in-scope type definitions, a static error is raised. [err:XP0051]
	
	Here are some examples of SequenceTypes that might be used in XQuery expressions:
	xs:date refers to the built-in Schema type date
	attribute()? refers to an optional attribute
	element() refers to any element
	element(po:shipto, po:address) refers to an element that has the name po:shipto (or is in the substitution group of that element), and has the type annotation po:address (or a subtype of that type)
	element(po:shipto, *) refers to an element named po:shipto (or in the substitution group of po:shipto), with no restrictions on its type
	element(*, po:address) refers to an element of any name that has the type annotation po:address (or a subtype of po:address). If the keyword nillable were used following po:address, that would indicate that the element may have empty content and the attribute xsi:nil="true", even though the declaration of the type po:address has required content.
	node()* refers to a sequence of zero or more nodes of any type
	item()+ refers to a sequence of one or more nodes or atomic values
	*/
	public void visit(SequenceType arg) throws XQueryException {
		QType qtype = null;
		if (arg.getItemType() != null)
			qtype = arg.getItemType().getQType();
		else {
			qtype = new QTypeSequence(null);
		}
		switch (arg.getOccurrence()) {
			case Constants.OCCURRENCE_PLUS :
				qtype = qtype.changeOccurence(QType.OCC_1_N);
				break;
			case Constants.OCCURRENCE_QUESTION_MARK :
				qtype = qtype.changeOccurence(QType.OCC_0_1);
				break;
			case Constants.OCCURRENCE_STAR :
				qtype = qtype.changeOccurence(QType.OCC_0_N);
				break;
		}
		if (qtype == null)
			throw new TypeException("Expression " + arg + " could not be typed."); //return;
		arg.setQType(qtype);
	}

	public void visit(SortedExpression arg) throws XQueryException {
		// TYPE VERIFICATION
		// sort clauses must be simple xpaths expressions
		//		ArrayList sortClauses = arg.getSortClauses();
		//		for (int i = 0; i < sortClauses.size(); i++)
		//		{
		//			((XQueryExpression) sortClauses.get(i)).accept(this);
		//		}
		// TYPE BUILDING
		// build chooice type!!!

		// type checking
		if (noForce && arg.getQType() != null)
			return;
		XQueryExpression expr = arg.getExpression();
		ArrayList sortClauses = arg.getSortClauses();
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null) {
			if (expr.getQType() == null)
				return;
		}
		if (/*!noForce ||*/
			expr.getQType() == null)
			expr.accept(this);
		// set the context (but only the type is present ... not so easy!)
		// not done for now
		if (expr.getQType() == null)
			throw new TypeException("Expression " + expr + " could not be typed."); //return;

		//		if (expr.getQType() != null) {
		//			QType tmpContextType = contextQType;
		//			setContextQType(expr.getQType());
		//			for (int i = 0; i < sortClauses.size(); i++)
		//				if (/*!noForce ||*/ ((XQueryExpression) sortClauses.get(i)).getQType() == null)
		//					 ((XQueryExpression) sortClauses.get(i)).accept(this);
		//			setContextQType(tmpContextType);
		//		}
		// type setting
		arg.setQType(expr.getQType());
	}

	// do nothing (but implement otherwise it fails)
	public void visit(Step arg) throws XQueryException {
	}

	//	public void visit(TypeSwitchExpression arg) throws XQueryException;

	public void visit(UnOpMinusExpression arg) throws XQueryException {
		// TYPE VERIFICATION
		// expression must be numeric
		// TYPE BUILDING

		// type checking
		if (noForce && arg.getQType() != null)
			return;
		XQueryExpression expr = arg.getExpression();
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null) {
			if (expr.getQType() == null)
				return;
		}
		if (/*!noForce ||*/
			expr.getQType() == null)
			expr.accept(this);
		// type verification
		if (expr.getQType() == null)
			throw new TypeException("Expression " + expr + " could not be typed."); //return;
		if (expr.getQType() != null) {
			if (!expr.getQType().isAtomic() /*&& !inPredicate*/)
				throw new TypeException("Expression '" + arg + "', operand is not an atomic value.");
			if (!expr.getQType().isNumeric())
				throw new TypeException("Expression '" + arg + "', operand is not numeric.");
		}
		// type setting
		QType qtypedata = expr.getQType().applyDATAFunction();
		if (qtypedata != null)
			arg.setQType(qtypedata);
		else
			arg.setQType(expr.getQType());
	}

	public void visit(ValidateExpression arg) throws XQueryException {
		if (noForce && arg.getQType() != null)
			return;
		XQueryExpression expr = arg.getExpression();
		arg.setQType(expr.getQType());
	}

	//	public void visit(Value arg) throws XQueryException;

	public void visit(ValueBoolean arg) throws XQueryException {
		// TYPE VERIFICATION
		// --> NONE
		// TYPE BUILDING
		// the return type is  boolean

		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom(context.getSchemaManager().getBooleanType()));
	}

	public void visit(ValueDecimal arg) throws XQueryException {
		// TYPE VERIFICATION
		// --> NONE
		// TYPE BUILDING
		// the return type is decimal

		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom(context.getSchemaManager().getDecimalType()));
	}

	public void visit(ValueDouble arg) throws XQueryException {
		// TYPE VERIFICATION
		// --> NONE
		// TYPE BUILDING
		// the return type is double

		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom(context.getSchemaManager().getDoubleType()));
	}

	public void visit(ValueFloat arg) throws XQueryException {
		// TYPE VERIFICATION
		// --> NONE
		// TYPE BUILDING
		// the return type is float

		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom(context.getSchemaManager().getFloatType()));
	}

	public void visit(ValueInteger arg) throws XQueryException {
		// TYPE VERIFICATION
		// --> NONE
		// TYPE BUILDING
		// the return type is integer

		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom(context.getSchemaManager().getIntegerType()));
	}

	public void visit(ValueString arg) throws XQueryException {
		// TYPE VERIFICATION
		// --> NONE
		// TYPE BUILDING
		// the return type is string

		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom(context.getSchemaManager().getStringType()));
	}

	public void visit(ValueText arg) throws XQueryException {
		// TYPE VERIFICATION
		// --> NONE
		// TYPE BUILDING
		// the return type is string

		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeText((SimpleType) getType(SC_STRING)));
	}

	public void visit(Variable arg) throws XQueryException {
		// TYPE VERIFICATION
		// --> NONE
		// TYPE BUILDING
		// the return type is the type of the associated expression, if any

		if (noForce && arg.getQType() != null)
			return;
		XQueryExpression expression = arg.getExpression();
		if (expression != null) {
			// forward checking and typing
			if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null) {
				if (expression.getQType() == null)
					return;
			}
			if (/*!noForce ||*/
				expression.getQType() == null)
				expression.accept(this);
			// type checking
			if (expression.getQType() == null)
				throw new TypeException("Expression " + expression + " could not be typed."); //return;
			if (expression instanceof ParenthesizedExpression && ((ParenthesizedExpression)expression).getExpression() == null)
				throw new TypeException("Variable " + arg + " could not be bound to existing path."); //return;			
			/*
			if (expression instanceof LocatedExpression && !arg.hasDocumentRoot() && (expression.getXTrees() == null || expression.getXTrees().isEmpty()))
				throw new TypeException("Variable " + arg + " could not be bound to existing path."); //return;
			*/
			// type setting
			if (arg.getBindingType() == Constants.LET_BINDINGTYPE)
				arg.setQType(expression.getQType());
			else if (arg.getBindingType() == Constants.FOR_BINDINGTYPE || arg.getBindingType() == Constants.QUANTIFIER_BINDINGTYPE || arg.getBindingType() == Constants.OUTERFOR_BINDINGTYPE) {
				QType exprqtype = expression.getQType();
				if (exprqtype instanceof QTypeSequence)
					exprqtype = new QTypeUnion(exprqtype.getList());
				if (exprqtype.isMultiple()) {
					exprqtype = exprqtype.changeOccurence(QType.OCC_1_1);
				}
				arg.setQType(exprqtype);

			}
			//else if (arg.getBindingType() == Constants.FUNCTION_BINDINGTYPE)
			//	arg.setQType(new QTypeAtom((SimpleType)getType((QName)expression)));
			else
				arg.setQType(expression.getQType());
		} else if (arg.getTypeDeclaration() != null) {
			arg.setQType(arg.getTypeDeclaration().getQType());
		}
		// for external variables without any type
		if (arg.isExternal() && arg.getQType() == null) {
			ArrayList list = new ArrayList(6);
			list.add(new QTypeDocument(null, QTypeDocument.ANY_ORIGIN));
			list.add(new QTypeProcessingInstruction());
			list.add(new QTypeComment());
			list.add(new QTypeText((SimpleType) getType(SC_STRING)));
			list.add(new QTypeElement(new QName("*", null), getType(SC_ANYTYPE)));
			list.add(new QTypeAttribute(new QName("*", null), context.getSchemaManager().getAnySimpleType()));
			arg.setQType(new QTypeUnion(list));
		}
		// else arg.setQType(null);
	}

	/*
	[104]    CompXmlComment    ::=    "comment" "{" Expr "}" 
	
	A computed comment constructor (CompXMLComment) constructs a new comment node with its own node identity. 
	The content expression of a computed comment constructor is processed as follows:
	
	Atomization is applied to the value of the content expression, converting it to a sequence of atomic values.
	
	If the result of atomization is an empty sequence, it is replaced by a zero-length string. 
	Otherwise, each atomic value in the atomized sequence is cast into a string.
	
	The individual strings resulting from the previous step are merged into a single string by concatenating them with 
	a single space character between each pair. The resulting string becomes the content of the constructed comment.
	*/
	public void visit(XMLComment arg) throws XQueryException {
		if (noForce && arg.getQType() != null)
			return;
		XQueryExpression comment = arg.getComment(); // cannot be null
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null) {
			if (comment.getQType() == null)
				return;
		}
		if (/*!noForce ||*/
			comment.getQType() == null)
			comment.accept(this);
		// type checking
		if (comment.getQType() == null)
			throw new TypeException("Expression " + comment + " could not be typed."); //return;

		QType valueQType = comment.getQType();
		if (!valueQType.canBeCastedInto(QType.PRIMITIVE_string) && !valueQType.isString()) {
			valueQType = valueQType.applyDATAFunction();
			if (valueQType == null || (!valueQType.canBeCastedInto(QType.PRIMITIVE_string) && !valueQType.isString()))
				throw new TypeException("Expression '" + comment + "', of attribute value pair has incorrect type.");
		}
		// type setting
		arg.setQType(new QTypeComment());
	}

	/*
	[103]    CompXmlPI    ::=    (("processing-instruction" NCName "{") | ("processing-instruction" "{" Expr "}" "{")) Expr? "}" 
	
	A computed processing instruction constructor (CompXmlPI) constructs a new processing instruction node with its own node identity. 
	The name expression of a computed processing instruction constructor is processed as follows:
	
	If the name expression returns an expanded QName: If the URI part of the QName is empty, 
	the local part of the QName is used as the name (target) of the processing instruction;	otherwise a dynamic error is raised.[err:XQ0041]
	
	If the name expression returns a string, that string is cast to a QName, which is then treated as in the previous item. 
	If the cast fails, a dynamic error is raised.[err:XP0021]
	
	If the name expression does not return a QName or a string, a type error is raised.[err:XP0004][err:XP0006]
	
	The content expression of a computed processing instruction constructor is processed as follows:
	
	Atomization is applied to the value of the content expression, converting it to a sequence of atomic values.
	
	If the result of atomization is an empty sequence, it is replaced by a zero-length string. 
	Otherwise, each atomic value in the atomized sequence is cast into a string.
	
	The individual strings resulting from the previous step are merged into a single string by concatenating them with a single space 
	character between each pair. The resulting string becomes the content of the constructed processing instruction.
	*/

	public void visit(XMLProcessingInstruction arg) throws XQueryException {
		if (noForce && arg.getQType() != null)
			return;
		XQueryExpression piName = arg.getExpression1(); // cannot be null
		XQueryExpression piValue = arg.getExpression2(); // cannot be null
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null) {
			//			if (!(piName instanceof QName) && piName.getQType() == null)
			//				return;
			if (piValue.getQType() == null)
				return;
		}
		//		if (!(piName instanceof QName) && (/*!noForce ||*/ piName.getQType() == null))
		//			piName.accept(this);
		// type checking
		//		if (!(piName instanceof QName) && piName.getQType() == null)
		//			throw new TypeException("Expression " + piName + " could not be typed.");
		if (/*!noForce ||*/
			piValue.getQType() == null)
			piValue.accept(this);
		// type checking
		if (piValue.getQType() == null)
			throw new TypeException("Expression " + piValue + " could not be typed."); //return;

		QType nameQType = piName.getQType();
		if (nameQType != null && (!nameQType.isQName() || nameQType.isMultiple())) {
			nameQType = nameQType.applyDATAFunction();
			if (nameQType == null || !nameQType.isQName() || nameQType.isMultiple())
				throw new TypeException("Expression '" + piName + "', of processing instruction value pair has incorrect type.");
		}
		QType valueQType = piValue.getQType();
		if (valueQType != null && !valueQType.canBeCastedInto(QType.PRIMITIVE_string) && !valueQType.isString()) {
			valueQType = valueQType.applyDATAFunction();
			if (valueQType == null || (!valueQType.canBeCastedInto(QType.PRIMITIVE_string) && !valueQType.isString()))
				throw new TypeException("Expression '" + piValue + "', of processing instruction pair has incorrect type.");
		}
		// CHANGE   this is to restrictive
		//		if (attributeValue.getQType().isMultiple())
		//			throw new TypeException("QType of an attribute cannot be multiple.");
		// type setting
		QType attQType = piValue.getQType();
		arg.setQType(new QTypeProcessingInstruction());
	}

	//	public void visit(XQueryBinaryOperatorExpression arg) throws XQueryException;
	//	public void visit(XQueryBooleanOperatorExpression arg) throws XQueryException;
	public void visit(XQueryExpression arg) throws XQueryException {
		throw new TypeException("Typing of expression : " + arg + " of class " + arg.getClass().getName() + " is not supported");
	}

	public void visit(XQueryExpressionSequence arg) throws XQueryException {
		// TYPE VERIFICATION
		// --> NONE
		// TYPE BUILDING
		// the return type is the sequence of the types

		// type checking
		if (noForce && arg.getQType() != null)
			return;
		ArrayList subExpressions = arg.getSubExpressions();
		// forward checking and typing
		for (int i = 0; i < subExpressions.size(); i++)
			if (/*!noForce ||*/
				 ((XQueryExpression) subExpressions.get(i)).getQType() == null) {
				((XQueryExpression) subExpressions.get(i)).accept(this);
				if (((XQueryExpression) subExpressions.get(i)).getQType() == null)
					throw new TypeException("Expression " + (XQueryExpression) subExpressions.get(i) + " could not be typed."); //return;
			}
		// type setting
		if (subExpressions.size() == 1)
			arg.setQType(((XQueryExpression) subExpressions.get(0)).getQType());
		else {
			ArrayList list = new ArrayList(subExpressions.size());
			boolean unique = true;
			for (int i = 0; i < subExpressions.size(); i++) {
				XQueryExpression expr = (XQueryExpression) subExpressions.get(i);
				QType qtype = expr.getQType();
				if (i != 0 && unique && !((QType) list.get(i - 1)).equals(qtype))
					unique = false;
				list.add(qtype);
			}
			if (unique)
				arg.setQType(((QType) list.get(0)).changeOccurence(QType.OCC_0_N));
			//				arg.setQType(new QTypeCollection((QType) list.get(0)));
			else
				arg.setQType(new QTypeSequence(list));
		}
	}

	//	public void visit(XQueryFile arg) throws XQueryException;
	//	public void visit(XQueryListBinaryOperatorExpression arg) throws XQueryException;
	//	public void visit(XQueryListUnaryOperatorExpression arg) throws XQueryException;
	//  public void visit(XQueryUnaryOperatorExpression arg) throws XQueryException;
	
	public void visit(ParenthesizedExpression arg) throws XQueryException {
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		XQueryExpression expr = arg.getExpression();
		if (expr == null) {
			arg.setQType(new QTypeSequence(null));
			return;
		}
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null) {
			if (expr.getQType() == null)
				return;
		}
		if (/*!noForce ||*/
			expr.getQType() == null)
			expr.accept(this);
		// type setting
		arg.setQType(expr.getQType());
	}

	public void visit(EnclosedExpression arg) throws XQueryException {
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		XQueryExpression expr = arg.getExpression();
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null) {
			if (expr.getQType() == null)
				return;
		}
		if (/*!noForce ||*/
			expr.getQType() == null)
			expr.accept(this);
		// type setting
		arg.setQType(expr.getQType());
	}

	// primitive functions

	public void visit(FunctionABS arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be numeric
		// TYPE BUILDING
		// the return type is a numeric

		// type checking
		if (noForce && arg.getQType() != null)
			return;
		XQueryExpression expr = (XQueryExpression) arg.getArguments().get(0);
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (/*!noForce ||*/
			expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Function call abs '" + arg + "', argument could not be typed"); // return;
		// type verification
		QType qtype = expr.getQType();
		if (!qtype.isNumeric())
			qtype = qtype.applyDATAFunction();
		if (qtype == null || !qtype.isNumeric())
			throw new TypeException("Function call abs '" + arg + "',  argument is not numeric");
		// type setting
		arg.setQType(new QTypeAtom(expr.getQType().getSimpleType()));
	}

	public void visit(FunctionAVG arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be numeric (or can be casted into a numeric)
		// each argument must be either
		//	- atom
		//	- attribute (hopefully the value can be casted into a numeric)
		//	- union (with at least a member being numeric)
		//	- collection (of one of the previous types)
		// TYPE BUILDING
		// the return type is a float

		// type checking
		if (noForce && arg.getQType() != null)
			return;
		//ArrayList arguments = arg.getArguments();
		XQueryExpression expr = (XQueryExpression) arg.getArguments().get(0);
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (/*!noForce ||*/
			expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Function call avg '" + arg + "', argument could not be typed"); // return;
		// type verification
		QType qtype = expr.getQType();
		/*
		if (!(qtype instanceof QTypeAtom) || (!qtype.isNumeric() && !qtype.isDuration()))
			qtype = qtype.applyDATAFunction();
		if (qtype == null || (!qtype.isNumeric() && !qtype.isDuration()))
			throw new TypeException("Function call avg '" + arg + "',  argument is not numeric or duration");
		*/
		if (!qtype.isAtom())
			qtype = qtype.applyDATAFunction();
		if (qtype == null || !qtype.isAtom())
			throw new TypeException("Function call avg '" + arg + "',  argument has incorrect type");
		// type setting
		if (qtype.getSubClass() == QType.SEQUENCE || qtype.getSubClass() == QType.UNION)
			arg.setQType(new QTypeAtom((SimpleType) getType(SC_DOUBLE)));
		else {
			PrimitiveType primType = ((SimpleType) qtype.getType()).getPrimitive();
			int prim = primType.getType();
			switch (prim) {
				case PrimitiveType.FLOAT :
					arg.setQType(new QTypeAtom((SimpleType) getType(SC_FLOAT)));
					break;
				case PrimitiveType.DECIMAL :
					arg.setQType(new QTypeAtom((SimpleType) getType(SC_DECIMAL)));
					break;
				case PrimitiveType.DOUBLE :
					arg.setQType(new QTypeAtom((SimpleType) getType(SC_DOUBLE)));
					break;
				default :
					arg.setQType(new QTypeAtom((SimpleType) getType(SC_DOUBLE)));
					break;
			}
		}
		//arg.setQType(new QTypeAtom((SimpleType) getType(SC_DOUBLE)));
	}

	public void visit(FunctionCEILING arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be numeric
		// TYPE BUILDING
		// the return type is a numeric

		// type checking
		if (noForce && arg.getQType() != null)
			return;
		XQueryExpression expr = (XQueryExpression) arg.getArguments().get(0);
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (/*!noForce ||*/
			expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Function call ceiling '" + arg + "', argument could not be typed"); // return;
		// type verification
		QType qtype = expr.getQType();
		if (!qtype.isNumeric())
			qtype = qtype.applyDATAFunction();
		if (qtype == null || !qtype.isNumeric())
			throw new TypeException("Function call ceiling '" + arg + "',  argument is not numeric");
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_INTEGER)));
	}

	public void visit(FunctionCOLLECTION arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be string
		// TYPE BUILDING
		// the return type is a node list

		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		ArrayList xnodes = context.getCollectionRootNodes(arg.getSourceName(), arg.getCollectionName());
		arg.setXTrees(xnodes);
		QType qtype = null;
		if (xnodes != null) {
			ArrayList tmplist = new ArrayList();
			for (int i = 0; i < xnodes.size(); i++) {
				Collection childreni = ((XTreeNode) xnodes.get(i)).getChildren();
				if (childreni != null && !childreni.isEmpty()) {
					ArrayList childlist = new ArrayList(1);
					childlist.addAll(childreni);
					QType qtypei = buildQTypeFromXNodes(childlist, QType.OCC_1_1, arg, true);
					if (qtypei != null)
						tmplist.add(new QTypeDocument(qtypei, QType.OCC_0_N, QTypeDocument.COLLECTION_ORIGIN));
				}
			}
			if (!tmplist.isEmpty()) {
				if (tmplist.size() == 1)
					qtype = (QType) tmplist.get(0);
				else
					qtype = new QTypeUnion(tmplist, QType.OCC_0_N);
			}
		}

		//ArrayList vectTypes = getDeclarationsFromSteps(arg);
		//QType qtype = buildTypeFromPathResolver(vectTypes, arg, true);
		// add   04/06/2003
		// ATTENTION for mediatorxql only
		if (qaOnly && qtype == null)
			qtype = new QTypeDocument(new QTypeElement(new QName("*", "*", "*", arg.getParentModule(), null), getType(SC_ANYTYPE)), QType.OCC_0_N, QTypeDocument.DOCUMENT_ORIGIN);
		// end add
		arg.setQType(qtype);

	}

	public void visit(FunctionCONCAT arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be collections
		// TYPE BUILDING
		// the return type is a integer

		// type checking
		if (noForce && arg.getQType() != null)
			return;
		ArrayList arguments = arg.getArguments();
		// forward checking and typing, and verification
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			for (int i = 0; i < arguments.size(); i++)
				if (((XQueryExpression) arguments.get(i)).getQType() == null)
					return;
		for (int i = 0; i < arguments.size(); i++) {
			XQueryExpression expri = (XQueryExpression) arguments.get(i);
			if (/*!noForce ||*/
				expri.getQType() == null)
				expri.accept(this);
			QType qtypei = expri.getQType();
			if (qtypei == null) //return;
				throw new TypeException("Function call concat '" + arg + "', argument " + (i + 1) + " could not be typed"); // return;
			if (!qtypei.isString())
				throw new TypeException("Function call concat '" + arg + "', argument " + (i + 1) + " is not a string.");
			if (!qtypei.isAtomic() /*&& !inPredicate*/)
				throw new TypeException("Function call concat '" + arg + "', argument " + (i + 1) + " is not an atomic value.'");
		}
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_STRING)));
	}

	public void visit(FunctionCONTAINS arg) throws XQueryException {
		// we treat now only signature like xf:contains(string $operand1, string $operand2) => boolean
		// but it exists signature like xf:contains(string $operand1, string $operand2, anyURI $collation) => boolean, later...
		// TYPE VERIFICATION
		// 2 arguments, and must be string
		// TYPE BUILDING
		// the return type is a boolean


		// type checking
		if (noForce && arg.getQType() != null)
			return;

		ArrayList arguments = arg.getArguments();
		// forward checking and typing, and verification
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			for (int i = 0; i < arguments.size(); i++)
				if (((XQueryExpression) arguments.get(i)).getQType() == null)
					return;
		for (int i = 0; i < arguments.size(); i++) {
			XQueryExpression expri = (XQueryExpression) arguments.get(i);
			if (/*!noForce ||*/
				expri.getQType() == null)
				expri.accept(this);
			if (expri.getQType() == null) //return;
				throw new TypeException("Function call contains '" + arg + "', argument " + (i + 1) + " could not be typed"); // return;
			if (!expri.getQType().isAtomic() /*&& !inPredicate*/)
				throw new TypeException("Function call contains '" + arg + "', argument " + (i + 1) + " is not an atomic value.'");
			if (!expri.getQType().isString())
				throw new TypeException("Function call contains '" + arg + "', argument " + (i + 1) + " is not a string.");
		}
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_BOOLEAN)));
	}

	public void visit(FunctionCOUNT arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be collections
		// TYPE BUILDING
		// the return type is a integer

		// type checking
		if (noForce && arg.getQType() != null)
			return;
		XQueryExpression expr = (XQueryExpression) arg.getArguments().get(0);
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (/*!noForce ||*/
			expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Function call count '" + arg + "', argument could not be typed"); // return;
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_INTEGER)));
	}

	public void visit(FunctionCURRENT_DATE arg) throws XQueryException {
		// but it exists signature like xf:current-date() => date
		// TYPE VERIFICATION
		// 0 arguments
		// TYPE BUILDING
		// the return type is a simpletype dateTime

		// ArrayList arguments = arg.getArguments();
		// forward checking and typing, and verification
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_DATE)));
	}

	public void visit(FunctionCURRENT_DATETIME arg) throws XQueryException {
		// but it exists signature like xf:current-dateTime() => date
		// TYPE VERIFICATION
		// 0 arguments
		// TYPE BUILDING
		// the return type is a simpletype dateTime

		// ArrayList arguments = arg.getArguments();
		// forward checking and typing, and verification
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_DATETIME)));
	}

	public void visit(FunctionCURRENT_TIME arg) throws XQueryException {
		// but it exists signature like xf:current-time() => date
		// TYPE VERIFICATION
		// 0 arguments
		// TYPE BUILDING
		// the return type is a simpletype dateTime

		// ArrayList arguments = arg.getArguments();
		// forward checking and typing, and verification
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_TIME)));
	}

	public void visit(FunctionDATA arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be simple type node
		// The built-in data function is implicitly invoked to extract the content of the node as a typed value.
		// If data is invoked with no argument, its implicit argument is the current (context) node.
		// For example, if the content of the current node is the integer 47, then data() has type integer and value 47.
		// If the content of its argument node cannot be expressed as a value of a simple type, the data function raises an error.
		// TYPE BUILDING
		// the return type depends on the node content

		// type checking
		if (noForce && arg.getQType() != null)
			return;
		XQueryExpression expr = (XQueryExpression) arg.getArguments().get(0);
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (/*!noForce ||*/
			expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Function call count '" + arg + "', argument could not be typed"); // return;
		QType qtype = expr.getQType();
		//		if (qtype == null) throw new TypeException("Could not type Function DATA argument.");
		// DATA function must be applied successfully on qtye
		QType argqtype = null;
		if (qtype.isAtom())
			argqtype = qtype;
		else
			argqtype = qtype.applyDATAFunction();
		if (argqtype == null)
			throw new TypeException("Function call data '" + arg + "', could not apply data function to type '" + qtype + "'.");
		// type setting
		arg.setQType(argqtype);
	}

	public void visit(FunctionDATE arg) throws XQueryException {
		// but it exists signature like xf:current-dateTime() => date
		// TYPE VERIFICATION
		// 0 arguments
		// TYPE BUILDING
		// the return type is a simpleType date

		// type checking
		if (noForce && arg.getQType() != null)
			return;
		XQueryExpression argument = (XQueryExpression) arg.getArguments().get(0);
		// forward checking and typing, and verification
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (argument.getQType() == null)
				return;
		if (!(argument instanceof ValueString))
			throw new TypeException("Function call date '" + arg + "', argument is not a string literal");
		try {
			SimpleType simpleType = context.getSchemaManager().getDateType();
			String dateValue = simpleType.normalizedValue(((ValueString) argument).getValue());
			simpleType.validate(dateValue, false, null);
			((ValueString) argument).setValue(dateValue);
		} catch (SchemaException se) {
			throw new TypeException("Function call date '" + arg + "', argument has invalid date format");
		}
		// type setting
		arg.setQType(new QTypeAtom(context.getSchemaManager().getDateType()));
	}

	public void visit(FunctionDATETIME arg) throws XQueryException {
		// but it exists signature like xf:current-dateTime() => date
		// TYPE VERIFICATION
		// 0 arguments
		// TYPE BUILDING
		// the return type is a simpleType date

		// type checking
		if (noForce && arg.getQType() != null)
			return;
		XQueryExpression argument = (XQueryExpression) arg.getArguments().get(0);
		// forward checking and typing, and verification
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (argument.getQType() == null)
				return;
		if (!(argument instanceof ValueString))
			throw new TypeException("Function call datetime '" + arg + "', argument is not a string literal");
		try {
			SimpleType simpleType = context.getSchemaManager().getDateTimeType();
			String dateValue = simpleType.normalizedValue(((ValueString) argument).getValue());
			simpleType.validate(dateValue, false, null);
			((ValueString) argument).setValue(dateValue);
		} catch (SchemaException se) {
			throw new TypeException("Function call datetime '" + arg + "', argument has invalid datetime format");
		}
		// type setting
		arg.setQType(new QTypeAtom(context.getSchemaManager().getDateTimeType()));
	}

	public void visit(FunctionDECIMAL arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be numeric
		// TYPE BUILDING
		// the return type is a numeric

		// type checking
		if (noForce && arg.getQType() != null)
			return;
		XQueryExpression expr = (XQueryExpression) arg.getArguments().get(0);
		// forward checking and typing, and verification
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (/*!noForce ||*/
			expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Function call decimal '" + arg + "', argument could not be typed"); // return;
		// CHANGE  
		QType qtype = expr.getQType();
		if (!qtype.isAtom()) {
			qtype = qtype.applyDATAFunction();
			if (!qtype.isAtom())
				throw new TypeException("Function call decimal '" + arg + "', argument has invalid type, not an atom");
		}
		if (!qtype.canBeCastedInto(QType.PRIMITIVE_decimal))
			throw new TypeException("Function call decimal '" + arg + "', argument has type that can not be casted");
		if (expr instanceof ValueString) {
			try {
				SimpleType simpleType = context.getSchemaManager().getDecimalType();
				String decimalValue = simpleType.normalizedValue(((ValueString) expr).getValue());
				simpleType.validate(decimalValue, false, null);
				((ValueString) expr).setValue(decimalValue);
			} catch (SchemaException se) {
				throw new TypeException("Function call decimal '" + arg + "', argument is string with invalid format");
			}
		}
		// type setting
		arg.setQType(new QTypeAtom(context.getSchemaManager().getDecimalType()));
	}

	public void visit(FunctionDISTINCT_VALUES arg) throws XQueryException {
		// TYPE VERIFICATION
		// argument must be node list
		// TYPE BUILDING
		// the return type is a node list

		// type checking
		if (noForce && arg.getQType() != null)
			return;
		XQueryExpression expr = (XQueryExpression) arg.getArgument(0);
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (/*!noForce ||*/
			expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Function call distinct-values '" + arg + "', argument could not be typed"); // return;
		// type checking
		// REMOVE   16/04/2003
		QType qtype = expr.getQType();
		if (!qtype.isAtom() /*&& !qtype.isSimpleTypeNode()*/
			) {
			qtype = qtype.applyDATAFunction();
			if (qtype == null || !qtype.isAtom() /*|| !qtype.isSimpleTypeNode()*/
				)
				throw new TypeException("Function call distinct-values '" + arg + "', argument is not a collection of simple values");
		}
		// type setting
		arg.setQType(qtype);
	}

	public void visit(FunctionDOC arg) throws XQueryException {

		// TYPE VERIFICATION
		// arguments must be string
		// TYPE BUILDING
		// the return type is a node list


		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		//arg.setQType(new QTypeDocument(null, QTypeDocument.DOCUMENT_ORIGIN));
		arg.setQType(new QTypeDocument(new QTypeElement(new QName("*", "*", "*", arg.getParentModule(), null), this.getType(SC_ANYTYPE)), QTypeDocument.DOCUMENT_ORIGIN));

	}

	public void visit(FunctionDOUBLE arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be numeric
		// TYPE BUILDING
		// the return type is a numeric

		//ArrayList arguments = arg.getArguments();
		XQueryExpression expr = (XQueryExpression) arg.getArguments().get(0);
		// forward checking and typing, and verification
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (/*!noForce ||*/
			expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Function call double '" + arg + "', argument could not be typed"); // return;
		QType qtype = expr.getQType();
		if (!qtype.isAtom()) {
			qtype = qtype.applyDATAFunction();
			if (qtype == null || !qtype.isAtom())
				throw new TypeException("Function call double '" + arg + "', argument has invalid type, not an atom");
		}
		if (!qtype.canBeCastedInto(QType.PRIMITIVE_double))
			throw new TypeException("Function call double '" + arg + "', argument has type that can not be casted");
		if (expr instanceof ValueString) {
			try {
				SimpleType simpleType = context.getSchemaManager().getDoubleType();
				String value = simpleType.normalizedValue(((ValueString) expr).getValue());
				simpleType.validate(value, false, null);
				((ValueString) expr).setValue(value);
			} catch (SchemaException se) {
				throw new TypeException("Function call double '" + arg + "', argument is string with invalid format");
			}
		}
		// type setting
		arg.setQType(new QTypeAtom(context.getSchemaManager().getDoubleType()));
	}

	public void visit(FunctionEMPTY arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be sequence
		// TYPE BUILDING
		// the return type is a boolean


		//arg.setQType(new QTypeAtom((SimpleType) getType(SC_BOOLEAN)));

		// type checking
		if (noForce && arg.getQType() != null)
			return;
		XQueryExpression expr = arg.getArgument(0);
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (/*!noForce ||*/
			expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Function call empty '" + arg + "', argument could not be typed"); // return;
		//		if (!expr.getQType().isNode())
		//			throw new TypeException("Function call empty '" + arg + "', argument is not a node or collection of nodes");
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_BOOLEAN)));
	}

	public void visit(FunctionENDS_WITH arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be sequence
		// TYPE BUILDING
		// the return type is a boolean

		//arg.setQType(new QTypeAtom((SimpleType) getType(SC_BOOLEAN)));

		// type checking
		if (noForce && arg.getQType() != null)
			return;
		ArrayList arguments = arg.getArguments();
		// forward checking and typing, and verification
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			for (int i = 0; i < arguments.size(); i++)
				if (((XQueryExpression) arguments.get(i)).getQType() == null)
					return;
		for (int i = 0; i < arguments.size(); i++) {
			XQueryExpression expri = (XQueryExpression) arguments.get(i);
			if (/*!noForce ||*/
				expri.getQType() == null)
				expri.accept(this);
			if (expri.getQType() == null) //return;
				throw new TypeException("Function call ends-with '" + arg + "', argument " + (i + 1) + " could not be typed"); // return;
			if (!expri.getQType().isString())
				throw new TypeException("Function call ends-with '" + arg + "', argument " + (i + 1) + " is not a string");
		}
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_BOOLEAN)));
	}

	public void visit(FunctionEXISTS arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be sequence
		// TYPE BUILDING
		// the return type is a boolean

		// type checking
		if (noForce && arg.getQType() != null)
			return;

		XQueryExpression expr = (XQueryExpression) arg.getArguments().get(0);
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (/*!noForce ||*/
			expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Function call empty '" + arg + "', argument could not be typed"); // return;
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_BOOLEAN)));
	}

	public void visit(FunctionEXPANDED_QNAME arg) throws XQueryException {
		if (noForce && arg.getQType() != null)
			return;

		ArrayList arguments = arg.getArguments();
		// forward checking and typing, and verification
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			for (int i = 0; i < arguments.size(); i++)
				if (((XQueryExpression) arguments.get(i)).getQType() == null)
					return;
		for (int i = 0; i < arguments.size(); i++) {
			XQueryExpression expri = (XQueryExpression) arguments.get(i);
			if (/*!noForce ||*/
				expri.getQType() == null)
				expri.accept(this);
			if (expri.getQType() == null) //return;
				throw new TypeException("Function call expanded-QName '" + arg + "', argument " + (i + 1) + " could not be typed"); // return;
			if (!expri.getQType().isAtom() || !expri.getQType().isString())
				throw new TypeException("Function call expanded-QName '" + arg + "', argument " + (i + 1) + " is not an atomic string.'");
		}
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_QNAME)));
	}

	public void visit(FunctionFALSE arg) throws XQueryException {
		// TYPE VERIFICATION
		// TYPE BUILDING
		// the return type is a boolean atom

		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_BOOLEAN)));
	}

	public void visit(FunctionFLOAT arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be numeric
		// TYPE BUILDING
		// the return type is a numeric

		// type checking
		if (noForce && arg.getQType() != null)
			return;
		XQueryExpression expr = (XQueryExpression) arg.getArguments().get(0);
		// forward checking and typing, and verification
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (/*!noForce ||*/
			expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Function call float '" + arg + "', argument could not be typed"); // return;
		QType qtype = expr.getQType();
		if (!qtype.isAtom()) {
			qtype = qtype.applyDATAFunction();
			if (qtype == null || !qtype.isAtom())
				throw new TypeException("Function call float '" + arg + "', argument has invalid type, not an atom");
		}
		if (!qtype.canBeCastedInto(QType.PRIMITIVE_float))
			throw new TypeException("Function call float '" + arg + "', argument has type that can not be casted");
		if (expr instanceof ValueString) {
			try {
				SimpleType simpleType = context.getSchemaManager().getFloatType();
				String value = simpleType.normalizedValue(((ValueString) expr).getValue());
				simpleType.validate(value, false, null);
				((ValueString) expr).setValue(value);
			} catch (SchemaException se) {
				throw new TypeException("Function call float '" + arg + "', argument is string with invalid format");
			}
		}
		// type setting
		arg.setQType(new QTypeAtom(context.getSchemaManager().getFloatType()));
	}

	public void visit(FunctionFLOOR arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be numeric
		// TYPE BUILDING
		// the return type is a numeric

		// type checking
		if (noForce && arg.getQType() != null)
			return;
		XQueryExpression expr = (XQueryExpression) arg.getArguments().get(0);
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (/*!noForce ||*/
			expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Function call floor '" + arg + "', argument could not be typed"); // return;
		// type verification
		QType qtype = expr.getQType();
		if (!qtype.isNumeric())
			qtype = qtype.applyDATAFunction();
		if (qtype == null || !qtype.isNumeric())
			throw new TypeException("Function call floor '" + arg + "',  argument is not numeric");
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_INTEGER)));
	}

	public void visit(FunctionGET_MONTH_FROM_DATE arg) throws XQueryException {
		// TYPE VERIFICATION
		// argument must be a date
		// TYPE BUILDING
		// the return type is an integer

		// type checking
		if (noForce && arg.getQType() != null)
			return;
		ArrayList arguments = arg.getArguments();
		XQueryExpression expr = (XQueryExpression) arguments.get(0);
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (/*!noForce ||*/
			expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Function call lower-case '" + arg + "', argument could not be typed"); // return;
		// CHANGE  
		QType qtype = expr.getQType();
		if (!qtype.isAtom()) {
			qtype = qtype.applyDATAFunction();
			if (qtype == null || !qtype.isAtom())
				throw new TypeException("Function call get-month-from-date '" + arg + "', argument has invalid type, not an atom");
		}
		if (!qtype.canBeCastedInto(QType.PRIMITIVE_date))
			throw new TypeException("Function call get-month-from-date '" + arg + "', argument has type that can not be casted");

		//if (!(argument instanceof ValueString))
		//	throw new TypeException("Function call decimal '" + arg + "', argument is not a string literal");
		if (expr instanceof ValueString) {
			try {
				SimpleType simpleType = context.getSchemaManager().getDateType();
				String dateValue = simpleType.normalizedValue(((ValueString) expr).getValue());
				simpleType.validate(dateValue, false, null);
				((ValueString) expr).setValue(dateValue);
			} catch (SchemaException se) {
				throw new TypeException("Function call get-month-from-date '" + arg + "', argument is string with invalid format");
			}
		}
		// type setting
		arg.setQType(new QTypeAtom(context.getSchemaManager().getIntegerType()));
	}

	public void visit(FunctionGET_YEAR_FROM_DATE arg) throws XQueryException {
		// TYPE VERIFICATION
		// argument must be a date
		// TYPE BUILDING
		// the return type is an integer

		// type checking
		if (noForce && arg.getQType() != null)
			return;
		ArrayList arguments = arg.getArguments();
		XQueryExpression expr = (XQueryExpression) arguments.get(0);
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (/*!noForce ||*/
			expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Function call lower-case '" + arg + "', argument could not be typed"); // return;
		// CHANGE  
		QType qtype = expr.getQType();
		if (!qtype.isAtom()) {
			qtype = qtype.applyDATAFunction();
			if (qtype != null || !qtype.isAtom())
				throw new TypeException("Function call get-month-from-date '" + arg + "', argument has invalid type, not an atom");
		}
		if (!qtype.canBeCastedInto(QType.PRIMITIVE_date))
			throw new TypeException("Function call get-month-from-date '" + arg + "', argument has type that can not be casted");

		//if (!(argument instanceof ValueString))
		//	throw new TypeException("Function call decimal '" + arg + "', argument is not a string literal");
		if (expr instanceof ValueString) {
			try {
				SimpleType simpleType = context.getSchemaManager().getDateType();
				String dateValue = simpleType.normalizedValue(((ValueString) expr).getValue());
				simpleType.validate(dateValue, false, null);
				((ValueString) expr).setValue(dateValue);
			} catch (SchemaException se) {
				throw new TypeException("Function call get-month-from-date '" + arg + "', argument is string with invalid format");
			}
		}
		// type setting
		arg.setQType(new QTypeAtom(context.getSchemaManager().getIntegerType()));
	}

	public void visit(FunctionINTEGER arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be numeric
		// TYPE BUILDING
		// the return type is a numeric

		// type checking
		if (noForce && arg.getQType() != null)
			return;
		ArrayList arguments = arg.getArguments();
		XQueryExpression expr = (XQueryExpression) arguments.get(0);
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (/*!noForce ||*/
			expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Function call integer '" + arg + "', argument could not be typed"); // return;
		QType qtype = expr.getQType();
		if (!qtype.canBeCastedInto(QType.PRIMITIVE_decimal))
			throw new TypeException("Function call integer '" + arg + "', argument has type that can not be casted");
		if (expr instanceof ValueString) {
			try {
				SimpleType simpleType = context.getSchemaManager().getIntegerType();
				String value = simpleType.normalizedValue(((ValueString) expr).getValue());
				simpleType.validate(value, false, null);
				((ValueString) expr).setValue(value);
			} catch (SchemaException se) {
				throw new TypeException("Function call float '" + arg + "', argument is string with invalid format");
			}
		}
		// type setting
		arg.setQType(new QTypeAtom(context.getSchemaManager().getIntegerType()));
	}

	public void visit(FunctionLOCAL_NAME arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be sequence
		// TYPE BUILDING
		// the return type is a string

		// type checking
		if (noForce && arg.getQType() != null)
			return;
		ArrayList arguments = arg.getArguments();
		XQueryExpression expr = (XQueryExpression) arguments.get(0);
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (/*!noForce ||*/
			expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Function call local-name'" + arg + "', argument could not be typed"); // return;
		// type verification
		QType qtype = expr.getQType();
		if (!qtype.isNode())
			throw new TypeException("Function call local-name '" + arg + "', argument is not a node");
		if (!qtype.isAtomic() /*&& !inPredicate*/)
			throw new TypeException("Function call local-name '" + arg + "', argument is not an atomic value.'");
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_STRING)));
	}

	public void visit(FunctionLOWER_CASE arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be sequence
		// TYPE BUILDING
		// the return type is a string

		// type checking
		if (noForce && arg.getQType() != null)
			return;
		ArrayList arguments = arg.getArguments();
		XQueryExpression expr = (XQueryExpression) arguments.get(0);
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (/*!noForce ||*/
			expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Function call lower-case '" + arg + "', argument could not be typed"); // return;
		// type verification
		QType qtype = expr.getQType();
		if (!qtype.isString())
			throw new TypeException("Function call lower-case '" + arg + "', argument is not a string");
		if (!qtype.isAtomic() /*&& !inPredicate*/)
			throw new TypeException("Function call lower-case '" + arg + "', argument is not an atomic value.'");
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_STRING)));
	}

	public void visit(FunctionMATCHES arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be sequence
		// TYPE BUILDING
		// the return type is a string

		// type checking
		if (noForce && arg.getQType() != null)
			return;
		ArrayList arguments = arg.getArguments();
		// forward checking and typing, and verification
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			for (int i = 0; i < arguments.size(); i++)
				if (((XQueryExpression) arguments.get(i)).getQType() == null)
					return;
		for (int i = 0; i < arguments.size(); i++) {
			XQueryExpression expri = (XQueryExpression) arguments.get(i);
			if (/*!noForce ||*/
				expri.getQType() == null)
				expri.accept(this);
			if (expri.getQType() == null) //return;
				throw new TypeException("Function call match '" + arg + "', argument " + (i + 1) + " could not be typed"); // return;
			if (!expri.getQType().isString())
				throw new TypeException("Function call match '" + arg + "', argument " + (i + 1) + " is not a string");
			if (!expri.getQType().isAtomic() /*&& !inPredicate*/)
				throw new TypeException("Function call match '" + arg + "', argument " + (i + 1) + " is not an atomic value.'");
		}
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_INTEGER), QType.OCC_0_N));
	}

	public void visit(FunctionMAX arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be numeric
		// TYPE BUILDING
		// the return type is a numeric

		// type checking
		if (noForce && arg.getQType() != null)
			return;
		ArrayList arguments = arg.getArguments();
		XQueryExpression expr = (XQueryExpression) arguments.get(0);
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (/*!noForce ||*/
			expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Function call max '" + arg + "', argument could not be typed"); // return;
		// type verification
		QType qtype = expr.getQType();
		/*
		if (!(qtype instanceof QTypeAtom) || (!qtype.isNumeric() && !qtype.isDate() && !qtype.isString() && !qtype.isDuration()))
			qtype = qtype.applyDATAFunction();
		if (qtype == null || (!qtype.isNumeric() && !qtype.isDate() && !qtype.isString() && !qtype.isDuration()))
			throw new TypeException("Function call max '" + arg + "',  argument is not numeric, date, duration or string");
		*/
		if (!qtype.isAtom())
			qtype = qtype.applyDATAFunction();
		if (qtype == null || !qtype.isAtom())
			throw new TypeException("Function call max '" + arg + "',  argument has incorrect type");
		// type setting
		//arg.setQType(new QTypeAtom((SimpleType) getType(SC_DOUBLE)));
		if (qtype.getSubClass() == QType.SEQUENCE || qtype.getSubClass() == QType.UNION)
			arg.setQType(new QTypeAtom((SimpleType) getType(SC_DOUBLE)));
		else {
			if (qtype.getOccurence() != QType.OCC_1_1)
				qtype = qtype.changeOccurence(QType.OCC_1_1);
			arg.setQType(qtype);
		}
	}

	public void visit(FunctionMIN arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be numeric
		// TYPE BUILDING
		// the return type is a numeric

		// type checking
		if (noForce && arg.getQType() != null)
			return;
		ArrayList arguments = arg.getArguments();
		XQueryExpression expr = (XQueryExpression) arguments.get(0);
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (/*!noForce ||*/
			expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Function call min '" + arg + "', argument could not be typed"); // return;
		// type verification
		QType qtype = expr.getQType();
		/*
		if (!(qtype instanceof QTypeAtom) || (!qtype.isNumeric() && !qtype.isDate() && !qtype.isString() && !qtype.isDuration()))
			qtype = qtype.applyDATAFunction();
		if (qtype == null || (!qtype.isNumeric() && !qtype.isDate() && !qtype.isString() && !qtype.isDuration()))
			throw new TypeException("Function call min '" + arg + "',  argument is not numeric, date, duration or string");
		*/
		if (!qtype.isAtom())
			qtype = qtype.applyDATAFunction();
		if (qtype == null || !qtype.isAtom())
			throw new TypeException("Function call min '" + arg + "',  argument has incorrect type");
		// type setting
		//arg.setQType(new QTypeAtom((SimpleType) getType(SC_DOUBLE)));
		if (qtype.getSubClass() == QType.SEQUENCE || qtype.getSubClass() == QType.UNION)
			arg.setQType(new QTypeAtom((SimpleType) getType(SC_DOUBLE)));
		else {
			if (qtype.getOccurence() != QType.OCC_1_1)
				qtype = qtype.changeOccurence(QType.OCC_1_1);
			arg.setQType(qtype);
		}
	}

	public void visit(FunctionNAMESPACE_URI arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be sequence
		// TYPE BUILDING
		// the return type is a string

		// type checking
		if (noForce && arg.getQType() != null)
			return;
		ArrayList arguments = arg.getArguments();
		XQueryExpression expr = (XQueryExpression) arguments.get(0);
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (/*!noForce ||*/
			expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Function call namespace-uri'" + arg + "', argument could not be typed"); // return;
		// type verification
		QType qtype = expr.getQType();
		if (!qtype.isNode())
			throw new TypeException("Function call namespace-uri '" + arg + "', argument is not a node");
		if (!qtype.isAtomic() /*&& !inPredicate*/)
			throw new TypeException("Function call local-name '" + arg + "', argument is not an atomic value.'");
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_STRING)));
	}

	public void visit(FunctionNOT arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be boolean
		// TYPE BUILDING
		// the return type is a boolean

		// type checking
		if (noForce && arg.getQType() != null)
			return;
		ArrayList arguments = arg.getArguments();
		XQueryExpression expr = (XQueryExpression) arguments.get(0);
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (/*!noForce ||*/
			expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Function call not '" + arg + "', argument could not be typed"); // return;
		// type verification
		if (!expr.getQType().isBoolean())
			throw new TypeException("Function call not '" + arg + "',  argument is not a boolean");
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_BOOLEAN)));
	}

	public void visit(FunctionNUMBER arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be an element
		// TYPE BUILDING
		// returns the content of an element expressed as a number

		// type checking
		if (noForce && arg.getQType() != null)
			return;
		ArrayList arguments = arg.getArguments();
		if (arguments != null) {
			XQueryExpression expr = (XQueryExpression) arguments.get(0);
			// forward checking and typing
			if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
				if (expr.getQType() == null)
					return;
			if (/*!noForce ||*/
				expr.getQType() == null)
				expr.accept(this);
			if (expr.getQType() == null)
				throw new TypeException("Function call number '" + arg + "', argument could not be typed"); // return;
			// type verification
			if (!expr.getQType().isSimpleTypeNode())
				throw new TypeException("Function call number '" + arg + "',  argument is not a simple node or collection of simple nodes");
		}
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_FLOAT)));
	}

	//	public void visit(FunctionPOSITION arg) throws XQueryException {
	//		// TYPE VERIFICATION
	//		// arguments must be an element
	//		// TYPE BUILDING
	//		// returns the content of an element expressed as a number
	//
	//		ArrayList arguments = arg.getArguments();
	//		// type checking
	//		if (noForce && arg.getQType() != null)
	//			return;
	//		// type setting
	//		arg.setQType(new QTypeAtom((SimpleType) getType(SC_INTEGER)));
	//	}

	public void visit(FunctionQNAME arg) throws XQueryException {
		if (noForce && arg.getQType() != null)
			return;

		ArrayList arguments = arg.getArguments();
		XQueryExpression expr = (XQueryExpression) arguments.get(0);
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (/*!noForce ||*/
			expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null) //return;
			throw new TypeException("Function call QName '" + arg + "', argument could not be typed"); // return;
		if (!expr.getQType().isAtom() || !expr.getQType().isString())
			throw new TypeException("Function call QName '" + arg + "', argument is not an atomic string.'");
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_QNAME)));
	}

	public void visit(FunctionROUND arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be numeric
		// TYPE BUILDING
		// the return type is a numeric

		// type checking
		if (noForce && arg.getQType() != null)
			return;
		ArrayList arguments = arg.getArguments();
		XQueryExpression expr = (XQueryExpression) arguments.get(0);
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (/*!noForce ||*/
			expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Function call round '" + arg + "', argument could not be typed"); // return;
		// type verification
		QType qtype = expr.getQType();
		if (!qtype.isNumeric())
			qtype = qtype.applyDATAFunction();
		if (qtype == null || !qtype.isNumeric())
			throw new TypeException("Function call round '" + arg + "',  argument is not numeric");
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_INTEGER)));
	}

	public void visit(FunctionSTARTS_WITH arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be sequence
		// TYPE BUILDING
		// the return type is a boolean

		// type checking
		if (noForce && arg.getQType() != null)
			return;
		ArrayList arguments = arg.getArguments();
		// forward checking and typing, and verification
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			for (int i = 0; i < arguments.size(); i++)
				if (((XQueryExpression) arguments.get(i)).getQType() == null)
					return;
		for (int i = 0; i < arguments.size(); i++) {
			XQueryExpression expri = (XQueryExpression) arguments.get(i);
			if (/*!noForce ||*/
				expri.getQType() == null)
				expri.accept(this);
			if (expri.getQType() == null) //return;
				throw new TypeException("Function call starts-with '" + arg + "', argument " + (i + 1) + " could not be typed"); // return;
			if (!expri.getQType().isString())
				throw new TypeException("Function call starts-with '" + arg + "', argument " + (i + 1) + " is not a string");
			if (!expri.getQType().isAtomic() /*&& !inPredicate*/)
				throw new TypeException("Function call starts-with '" + arg + "', argument " + (i + 1) + " is not an atomic value.'");
		}
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_BOOLEAN)));
	}

	public void visit(FunctionSTRING arg) throws XQueryException {
		// TYPE VERIFICATION
		/*
		The string function converts its argument to a string according to the following rules
		A number is converted to a string as follows:
			NaN is converted to the string NaN.
			Positive zero is converted to the string 0.
			Negative zero is converted to the string 0.
			Positive infinity is converted to the string Infinity.
			Negative infinity is converted to the string -Infinity.
			If the number is an integer, the number is represented in decimal form with no decimal point and no leading zeros, preceded by a minus sign (-) if the number is negative.
			Otherwise, the number is represented in decimal form including a decimal point with at least one digit before the decimal point and at least one digit after the decimal point, preceded by a minus sign (-) if the number is negative; there must be no leading zeros before the decimal point apart possibly from the one required digit immediately before the decimal point; beyond the one required digit after the decimal point there must be as many, but only as many, more digits as are needed to uniquely distinguish the number from all other IEEE 754 numeric values.
		The boolean false value is converted to the string false. The boolean true value is converted to the string true.
		An object of a type other than the above types is converted to a string in a way that is dependent on that type.
		If the argument is omitted, it defaults to a node-set with the context node as its only member
		 */
		// each argument must be either
		//	- element
		//	- attribute
		//	- union (with at least a member being node)
		//	- collection (of one of the previous types)
		// TYPE BUILDING
		// returns the content of an element expressed as a string

		// type checking
		if (noForce && arg.getQType() != null)
			return;
		ArrayList arguments = arg.getArguments();
		XQueryExpression expr = (XQueryExpression) arguments.get(0);
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (/*!noForce ||*/
			expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Function call string '" + arg + "', argument could not be typed"); // return;
		// type verification
		QType qtype = expr.getQType();
		// type verification
		//if (!qtype.isAtom()) qtype = qtype.applyDATAFunction();
		if (qtype == null || (!qtype.isAtomic() && !inPredicate && !qtype.isSimpleTypeNode()))
			throw new TypeException("Function call string '" + arg + "', argument must be an atomic value or simple content element.");
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_STRING), qtype.isMultiple() ? QType.OCC_0_N : QType.OCC_1_1));
		//		if (qtype instanceof QTypeCollection)
		//			arg.setQType(new QTypeCollection(new QTypeAtom((SimpleType) getType(SC_STRING))));
		//		else
		//			arg.setQType(new QTypeAtom((SimpleType) getType(SC_STRING)));
	}

	public void visit(FunctionSTRING_LENGTH arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be sequence
		// TYPE BUILDING
		// the return type is a boolean

		// type checking
		if (noForce && arg.getQType() != null)
			return;
		ArrayList arguments = arg.getArguments();
		XQueryExpression expr = (XQueryExpression) arguments.get(0);
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		// type verification
		if (/*!noForce ||*/
			expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null) //return;
			throw new TypeException("Function call string-length '" + arg + "', argument could not be typed"); // return;
		if (!expr.getQType().isString())
			throw new TypeException("Function call string-length '" + arg + "', argument is not a string");
		if (!expr.getQType().isAtomic() /*&& !inPredicate*/)
			throw new TypeException("Function call string-length '" + arg + "', argument is not an atomic value.'");
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_INTEGER)));
	}

	public void visit(FunctionSUBSTRING arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be sequence
		// TYPE BUILDING
		// the return type is a string

		// type checking
		if (noForce && arg.getQType() != null)
			return;
		ArrayList arguments = arg.getArguments();
		// forward checking and typing, and verification
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			for (int i = 0; i < arguments.size(); i++)
				if (((XQueryExpression) arguments.get(i)).getQType() == null)
					return;
		for (int i = 0; i < arguments.size(); i++) {
			XQueryExpression expri = (XQueryExpression) arguments.get(i);
			if (/*!noForce ||*/
				expri.getQType() == null)
				expri.accept(this);
			if (expri.getQType() == null) //return;
				throw new TypeException("Function call substring '" + arg + "', argument " + (i + 1) + " could not be typed"); // return;
			if (i == 0 && !expri.getQType().isString())
				throw new TypeException("Function call substring '" + arg + "', argument " + (i + 1) + " is not a string");
			if (i != 0 && !expri.getQType().isNumeric())
				throw new TypeException("Function call substring '" + arg + "', argument " + (i + 1) + " is not numeric");
		}
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_STRING)));
	}

	public void visit(FunctionSUM arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be numeric (or can be casted into a numeric)
		// each argument must be either
		//	- atom
		//	- attribute (hopefully the value can be casted into a numeric)
		//	- union (with at least a member being numeric)
		//	- collection (of one of the previous types)
		// TYPE BUILDING
		// the return type is a float

		// type checking
		if (noForce && arg.getQType() != null)
			return;
		ArrayList arguments = arg.getArguments();
		XQueryExpression expr = (XQueryExpression) arguments.get(0);
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (/*!noForce ||*/
			expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Function call sum '" + arg + "', argument could not be typed"); // return;
		// type verification
		QType qtype = expr.getQType();
		/*
		if (!(qtype instanceof QTypeAtom) || (!qtype.isNumeric() && !qtype.isDuration()))
			qtype = qtype.applyDATAFunction();
		if (qtype == null || (!qtype.isNumeric() && !qtype.isDuration()))
			throw new TypeException("Function call sum '" + arg + "',  argument is not numeric or duration");
		*/
		if (!qtype.isAtom())
			qtype = qtype.applyDATAFunction();
		if (qtype == null || !qtype.isAtom())
			throw new TypeException("Function call sum '" + arg + "',  argument has incorrect type");
		// type setting
		if (qtype.getSubClass() == QType.SEQUENCE || qtype.getSubClass() == QType.UNION)
			arg.setQType(new QTypeAtom((SimpleType) getType(SC_DOUBLE)));
		else {
			PrimitiveType primType = ((SimpleType) qtype.getType()).getPrimitive();
			int prim = primType.getType();
			switch (prim) {
				case PrimitiveType.FLOAT :
					arg.setQType(new QTypeAtom((SimpleType) getType(SC_FLOAT)));
					break;
				case PrimitiveType.DECIMAL :
					if (primType.getFractionDigits() == 0)
						arg.setQType(new QTypeAtom((SimpleType) getType(SC_INTEGER)));
					else
						arg.setQType(new QTypeAtom((SimpleType) getType(SC_DECIMAL)));
					break;
				case PrimitiveType.DOUBLE :
					arg.setQType(new QTypeAtom((SimpleType) getType(SC_DOUBLE)));
					break;
				default :
					arg.setQType(new QTypeAtom((SimpleType) getType(SC_DOUBLE)));
					break;
			}
		}
		//arg.setQType(new QTypeAtom((SimpleType) getType(SC_DOUBLE)));
	}

	public void visit(FunctionTIME arg) throws XQueryException {
		// but it exists signature like xf:current-dateTime() => date
		// TYPE VERIFICATION
		// 0 arguments
		// TYPE BUILDING
		// the return type is a simpleType date

		// type checking
		if (noForce && arg.getQType() != null)
			return;
		XQueryExpression argument = (XQueryExpression) arg.getArguments().get(0);
		// forward checking and typing, and verification
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (argument.getQType() == null)
				return;
		if (!(argument instanceof ValueString))
			throw new TypeException("Function call time '" + arg + "', argument is not a string literal");
		try {
			SimpleType simpleType = context.getSchemaManager().getTimeType();
			String timeValue = simpleType.normalizedValue(((ValueString) argument).getValue());
			simpleType.validate(timeValue, false, null);
			((ValueString) argument).setValue(timeValue);
		} catch (SchemaException se) {
			throw new TypeException("Function call time '" + arg + "', argument has invalid datetime format");
		}
		// type setting
		arg.setQType(new QTypeAtom(context.getSchemaManager().getTimeType()));
	}

	public void visit(FunctionTRUE arg) throws XQueryException {
		// TYPE VERIFICATION
		// TYPE BUILDING
		// the return type is a boolean atom

		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_BOOLEAN)));
	}

	public void visit(FunctionUPPER_CASE arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be sequence
		// TYPE BUILDING
		// the return type is a string

		// type checking
		if (noForce && arg.getQType() != null)
			return;
		ArrayList arguments = arg.getArguments();
		XQueryExpression expr = (XQueryExpression) arguments.get(0);
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (/*!noForce ||*/
			expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Function call upper-case '" + arg + "', argument could not be typed"); // return;
		// type verification
		QType qtype = expr.getQType();
		if (!qtype.isString())
			throw new TypeException("Function call upper-case '" + arg + "', argument is not a string");
		if (!qtype.isAtomic() /*&& !inPredicate*/)
			throw new TypeException("Function call upper-case '" + arg + "', argument is not an atomic value.'");
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_STRING)));
	}

	public void visit(FunctionDEEP_EQUALS arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be sequence
		// TYPE BUILDING
		// the return type is a boolean

		// type checking
		if (noForce && arg.getQType() != null)
			return;

		ArrayList arguments = arg.getArguments();
		// forward checking and typing, and verification
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			for (int i = 0; i < arguments.size(); i++)
				if (((XQueryExpression) arguments.get(i)).getQType() == null)
					return;
		for (int i = 0; i < arguments.size(); i++) {
			XQueryExpression expri = (XQueryExpression) arguments.get(i);
			if (/*!noForce ||*/
				expri.getQType() == null)
				expri.accept(this);
			if (expri.getQType() == null) //return;
				throw new TypeException("Function call deep-equals '" + arg + "', argument " + (i + 1) + " could not be typed"); // return;
			if (!expri.getQType().isAtomic() /*&& !inPredicate*/)
				throw new TypeException("Function call deep-equals '" + arg + "', argument " + (i + 1) + " is not an atomic value.'");
		}
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_BOOLEAN)));
	}

	//*******************************************************************************************
	// UTLITIES
	//*******************************************************************************************
	private Type getType(String typeStr) {
		return context.getSchemaManager().getType(XML_SCHEMA, typeStr);
	}

	private QType buildQTypeFromXNodes(ArrayList schemaNodes, byte occurence) throws XQueryException {
		return buildQTypeFromXNodes(schemaNodes, occurence, null, false);
	}

	private QType buildQTypeFromXNodes(ArrayList schemaNodes, byte occurence, XQueryExpression expr, boolean verify) throws XQueryException {
		if (schemaNodes == null || schemaNodes.isEmpty())
			return null;

		QType retQType = null;
		ArrayList list = new ArrayList(schemaNodes.size());
		QType tmpQType = null;
		for (int i = 0; i < schemaNodes.size(); i++) {
			TypedXTreeNode typedXTreeNode = (TypedXTreeNode) schemaNodes.get(i);
			Declaration decl = typedXTreeNode.getDeclaration();
			if (decl == null) {
				if (typedXTreeNode.getLocalName() != null) {
					//					// TODO check if test is correct
					if (typedXTreeNode.getChildren() == null) {
						SimpleType type = context.getSchemaManager().getAnySimpleType();
						tmpQType = new QTypeAttribute(new QName(typedXTreeNode.getNamespace(), null, typedXTreeNode.getLocalName(), null, null), type, QType.computeOccurence(occurence, QType.OCC_0_1));
					} else
						tmpQType = new QTypeElement(new QName(typedXTreeNode.getNamespace(), null, typedXTreeNode.getLocalName(), null, null), getType(SC_ANYTYPE), QType.OCC_0_N);
				} else {
					// TODO find out why this case happens --> collection("ITEMS")//DESCRIPTION
					// for now do nothing
				}
			} else
				tmpQType = getQTypeFromDeclaration(decl, occurence);
			list.add(tmpQType);
		}
		if (list.isEmpty())
			if (verify)
				throw new TypeException("Could not find declaration(s) of expression '" + expr + "'");
			else
				return null;
		else if (list.size() == 1)
			return (QType) list.get(0);
		return makeUnion(list);
	}

	private QType getQTypeFromDeclaration(Declaration decl, byte occurence) throws XQueryException {
		if (decl == null)
			return null;
		// throw new TypeException("Expression not found in metadata");
		QType retQType = null;
		QName qname = new QName(decl.getNamespace(), null, decl.getName(), null, null);
		if (decl instanceof AttributeDeclaration)
			retQType = new QTypeAttribute(qname, (AttributeDeclaration) decl, (SimpleType) decl.getType(), QType.computeOccurence(occurence, (((AttributeDeclaration) decl).getUse() == AttributeDeclaration.REQUIRED) ? QType.OCC_1_1 : QType.OCC_0_1));
		if (decl instanceof ElementDeclaration) {
			byte elemOccurence = QType.OCC_1_1;
			ElementDeclaration eltDecl = (ElementDeclaration) decl;
			SchemaScope scopeType = eltDecl.getScope(); // see zhang yun for always true!
			if (scopeType != null && scopeType instanceof ComplexType) {
				int size[] = ((ComplexType) scopeType).getDeclarationOccurrence(eltDecl.getNamespace(), eltDecl.getName());
				elemOccurence = QType.getOccurence(size[0], size[1]);
			}
			retQType = new QTypeElement(qname, decl.getType(), QType.computeOccurence(occurence, elemOccurence));
			//			if (multi)
			//				retQType = new QTypeCollection(new QTypeElement(qname, decl.getType()));
			//			else
			//				retQType = new QTypeElement(qname, decl.getType());
		}
		return retQType;
	}

	private QType makeUnion(ArrayList list) throws XQueryException {
		if (list == null || list.isEmpty())
			return null;
		// check special case : only simple types with same name and diffrent namespaces
		for (int i = 0; i < list.size(); i++) {
			QType qtypei = (QType) list.get(i);
			Type typei = qtypei.getType();
			if (typei != null) {
			}
		}
		for (int i = 0; i < list.size() - 1; i++) {
			QType qtypei = (QType) list.get(i);
			for (int j = i + 1; j < list.size(); j++) {
				if (qtypei.equals((QType) list.get(j))) {
					list.remove(j);
					j--;
				}
			}
		}
		if (list.size() == 1)
			return (QType) list.get(0);
		return new QTypeUnion(list);
	}

	private QTypeElement buildQTypeElementFromSubElements(XQueryExpression name, Set allElements) throws XQueryException {
		int size = allElements.size();
		ElementTools eTools = new ElementTools(size, name);
		// case of <tag/>
		if (size == 0)
			return new QTypeElement(name, getType(SC_ANYTYPE));
		ComplexType complexType = new ComplexType(this.typingschema, null, null);
		//complexType.setRequiredAndDefaultAttributes(new ArrayList());
		
		SequenceModelGroup sequenceModel = new SequenceModelGroup();
		QTypeElement qte = null;
		//		try {
		for (Iterator it = allElements.iterator();it.hasNext();) {
			QType qtype = ((XQueryExpression) it.next()).getQType();
			int subclass = qtype.getSubClass();
			if (subclass == QType.ATOM) {
				// case of <tag> [simpleType] </tag>, like <a> 23 </a>
				if (size == 1)
					return new QTypeElement(name, qtype.getSimpleType() == null ? getType(SC_ANYTYPE) : qtype.getSimpleType());
				else
					buildQTypeAtom(eTools, qtype);
			} else if (subclass == QType.TEXT) {
				if (size == 1)
					return new QTypeElement(name, qtype.getSimpleType() == null ? getType(SC_ANYTYPE) : qtype.getSimpleType());
				else
					buildQTypeText(eTools, qtype);
			} else if (subclass == QType.ATTRIBUTE) {
				buildQTypeAttribute(eTools, (QTypeAttribute)qtype, complexType);
			} else if (subclass == QType.ELEMENT) {
				buildQTypeElement(eTools, qtype, complexType, sequenceModel, qtype.getMinOccurence() == 1, qtype.isMultiple());
			} else if (subclass == QType.UNION) {
				buildQTypeUnion(eTools, qtype, complexType, sequenceModel);
			} else if (subclass == QType.SEQUENCE) {
				qte = buildQTypeSequence(eTools, qtype, complexType, sequenceModel);
				if (qte != null)
					return qte;
			}
		}
		// case of <tag> [simpleType]+ </tag>, like <a> 23 toto</a>
		if (eTools.getCountAttribute() == 0 && eTools.getSize() == eTools.getCountSimpleType())
			return new QTypeElement(name, (SimpleType) getType(SC_STRING));
		// case of <tag (attr)* > [simpleType]+ </tag>, like <a b="1"> 23 toto </a>
		if ((eTools.getCountAttribute() == eTools.getSize() - eTools.getCountSimpleType()) && (eTools.getMixed())) {
			complexType.setBaseType((SimpleType) getType(SC_STRING));
			complexType.setExtension(true);
			complexType.setContentModel(new ContentModel((SimpleType) getType(SC_STRING)));
		}
		// case of <tag (attr)* > [simpleType] </tag>, like <a b="1"> 23 </a>
		else if ((eTools.getCountAttribute() == eTools.getSize() - 1) && (eTools.getMixed())) {
			complexType.setBaseType(eTools.getSimpleType());
			complexType.setExtension(true);
			complexType.setContentModel(new ContentModel(eTools.getSimpleType()));

		}
		// all others cases...
		else {
			ContentModel contentModel = new ContentModel();
			contentModel.setModel(new Particle(sequenceModel));
			if (eTools.getMixed())
				contentModel.setContentType(SchemaConstants.MIXED);
			else
				contentModel.setContentType(SchemaConstants.ELEMENT_ONLY);
			complexType.setContentModel(contentModel);
		}
		//		} catch (SchemaException se) {
		//			complexType = (ComplexType) this.getType(SC_ANYTYPE);
		//		}
		return new QTypeElement(name, complexType);
	}

	private void buildQTypeAtom(ElementTools eTools, QType qtype) throws XQueryException {
		eTools.setSimpleType(qtype.getSimpleType());
		eTools.setMixed(true);
		eTools.incCountSimpleType();
		eTools.incCountSimpleList();
	}

	private void buildQTypeText(ElementTools eTools, QType qtype) throws XQueryException {
		eTools.setSimpleType(qtype.getSimpleType());
		eTools.setMixed(true);
		eTools.incCountSimpleType();
		eTools.incCountSimpleList();
	}

	private void buildQTypeAttribute(ElementTools eTools, QTypeAttribute qtype, ComplexType complexType) throws XQueryException {
		String name = null;
		XQueryExpression qname = qtype.getName();
		if (qname instanceof QName)
			name = ((QName) qtype.getName()).getName();
		else
			name = getTemporaryName();
		// add LARS 05/07/04
		if (name.equals("xmlns"))
		    return;
		eTools.incCountAttribute();
		AttributeDeclaration attribute = new AttributeDeclaration(this.typingschema, name, complexType);
		attribute.setType(qtype.getSimpleType());
		try {
			complexType.register(attribute);
		} catch (SchemaException saxe) {
			throw new TypeException("Error in register(ElementDeclaration): " + saxe.getMessage());
		}
	}

	private byte index = 0;
	private String getTemporaryName() {
		return TEMP_NAME + index++;
	}

	private void buildQTypeElement(ElementTools eTools, QType qtype, ComplexType complexType, ModelGroup model, boolean hasOne, boolean unbounded) throws XQueryException {
		String name = null;
		String uri = null;
		Schema tmpschema = this.typingschema;
		try {
			QName qname = (QName) ((QTypeElement) qtype).getName();
			uri = qname.getNameSpace();
			if (uri != null && !uri.equals(STAR)) {
				tmpschema = this.typingschemamanager.getSchema(uri);
				if (tmpschema == null) {
					tmpschema = new Schema(uri, null, typingschemamanager);
					try {
						typingschemamanager.putSchema(tmpschema);
					} catch (SAXException saxe) {
						throw new TypeException("Could not add schema to manager");
					}
				}
			}
			name = qname.getName();
		} catch (ClassCastException cce) {
			name = "TagDefineInEvaluation";
		}
		Particle particle = null;
		if (!name.equals(STAR)) {
			ElementDeclaration element = null;
			// test if element already exists
			element = complexType.getElementDeclaration(uri, name);
			if (element == null) {
				SchemaScope scope = null;
				if (uri != null && !uri.equals(STAR)) {
					scope = tmpschema;
					element = new ElementDeclaration(tmpschema, name, scope);
					element.setQualified(true);
				} else {
					scope = complexType;
					element = new ElementDeclaration(tmpschema, name, scope);
				}
				element.setType(qtype.getType());
				try {
					if (uri != null && !uri.equals(STAR))
						complexType.register(element);
					else
						scope.register(element);
				} catch (SchemaException se) {
					throw new XQueryException(se.getMessage());
				}
			} else {
				element.setType(tmpschema.getType(SC_ANYTYPE));
				/*
				Type elementType = element.getType();
				if (elementType.getValueType() != null)
				*/
			}
			particle = new Particle(element);
		} else { // name = STAR	
			// TODO? make a choice of different wildcards?
			particle = eTools.getWildCard();
			if (particle == null) {
				particle = new Particle(new Wildcard(tmpschema,true));
				eTools.addWildCard(particle);
			}			
		}
		if (!hasOne)
			particle.setMinOccurs(0);
		if (unbounded)
			particle.setMaxOccursUnbounded();
		model.add(particle);
	}

	private QTypeElement buildQTypeSequence(ElementTools eTools, QType qtype, ComplexType complexType, SequenceModelGroup sequenceModel) throws XQueryException {
		List list = qtype.getList();
		if (list == null)
			return null;
		int size = list.size();
		QTypeElement qte = null;
		for (int i = 0; i < size; i++) {
			QType stype = (QType) list.get(i);
			int subclass = stype.getSubClass();
			if (subclass == QType.ATOM) {
				// case of <tag> [simpleType] </tag>, like <a> 23 </a>
				if (size == 1 && eTools.getSize() == 1)
					return new QTypeElement(eTools.getName(), qtype.getSimpleType() == null ? getType(SC_ANYTYPE) : qtype.getSimpleType());
				else
					buildQTypeAtom(eTools, stype);
			} else if (subclass == QType.TEXT) {
				buildQTypeText(eTools, stype);
			} else if (subclass == QType.ATTRIBUTE) {
				buildQTypeAttribute(eTools, (QTypeAttribute)stype, complexType);
			} else if (subclass == QType.ELEMENT) {
				buildQTypeElement(eTools, stype, complexType, sequenceModel, stype.getMinOccurence() == 1, stype.isMultiple());
			} else if (subclass == QType.UNION) {
				buildQTypeUnion(eTools, stype, complexType, sequenceModel);
			} else if (subclass == QType.SEQUENCE) {
				qte = buildQTypeSequence(eTools, stype, complexType, sequenceModel);
				if (qte != null)
					return qte;
			}

		}
		return null;
	}

	private void buildQTypeUnion(ElementTools eTools, QType qtype, ComplexType complexType, SequenceModelGroup sequenceModel) throws XQueryException {
		List list = qtype.getList();
		ChoiceModelGroup choiceModel = null;
		for (int i = 0; i < list.size(); i++) {
			if (i != 0)
				eTools.incSize();
			QType utype = (QType) list.get(i);
			int subclass = utype.getSubClass();
			boolean atom = false;
			switch (subclass) {
				case QType.ATOM :
					if (!atom) {
						buildQTypeAtom(eTools, utype);
						atom = true;
					}
					break;
				case QType.ATTRIBUTE :
					buildQTypeAttribute(eTools, (QTypeAttribute)utype, complexType);
					break;
				case QType.ELEMENT :
					if (choiceModel == null)
						choiceModel = new ChoiceModelGroup();
					buildQTypeElement(eTools, utype, complexType, choiceModel, utype.getMinOccurence() == 1, utype.isMultiple());
				    break;
				case QType.SEQUENCE :
					this.buildQTypeSequence(eTools, utype, complexType, sequenceModel);
				    break;
				case QType.TEXT :
					buildQTypeText(eTools, utype);
				    break;
				case QType.PROCESSING :
				case QType.COMMENT :
				case QType.DOCUMENT :
				case QType.UNION :
				    throw new TypeException("Could not type, incorrect element for element content.");
			}
			/*
			if (subclass == QType.ATOM) {
				if (!atom) {
					buildQTypeAtom(eTools, utype);
					atom = true;
				}
			} else if (subclass == QType.TEXT) {
				buildQTypeText(eTools, utype);
			} else if (subclass == QType.ATTRIBUTE) {
				buildQTypeAttribute(eTools, (QTypeAttribute)utype, complexType);
			} else if (subclass == QType.ELEMENT) {
				if (choiceModel == null)
					choiceModel = new ChoiceModelGroup();
				buildQTypeElement(eTools, utype, complexType, choiceModel, utype.getMinOccurence() == 1, utype.isMultiple());
			} else if (subclass == QType.SEQUENCE) {
				this.buildQTypeSequence(eTools, utype, complexType, sequenceModel);
			} else if (subclass == QType.DOCUMENT) {
			    
			} else
				throw new TypeException("We have no specification for this case (NYI), definition: union { (atom | element | attribute)+ }.");
			*/
		}
		if (choiceModel != null)
			sequenceModel.add(new Particle(choiceModel));
	}

	private boolean isAnyTypeKind(QType qType) {
		Type type = qType.getType();
		if (type != null && type.getNamespace() != null && type.getNamespace().equals(SchemaConstants.XMLSCHEMA_URI) && type.getName() != null && (type.getName().equals(SchemaConstants.ANY_TYPE) || type.getName().equals(SchemaConstants.SIMPLE_UR_TYPE)))
			return true;
		return false;
	}
}
