/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.xquery.xdbc;

import java.io.StringWriter;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Map;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;
import org.xml.sax.*;
import org.xml.sax.ext.LexicalHandler;
import org.xquark.schema.SchemaManager;
import org.xquark.schema.validation.PSVInfoSetProvider;
import org.xquark.schema.validation.SchemaValidationContext;
import org.xquark.schema.validation.ValidatingSchemaHandler;
import org.xquark.serialize.XMLSerializer;
import org.xquark.util.DOMBuilder;
import org.xquark.util.NamespaceContextStack;
import org.xquark.xml.xdbc.*;
import org.xquark.xquery.metadata.DynamicContext;
import org.xquark.xquery.parser.*;
import org.xquark.xquery.parser.util.Constants;
import org.xquark.xquery.reconstruction.ReconstructionVisitor;

/**
 * Common base implementation of the XBDC result set. It is used as this
 * by the Extractor.
 */
public class XResultSetImpl implements XMLResultSet {
	private static final String RCSRevision = "$Revision: 1.26 $";
	private static final String RCSName = "$Name: XQ1_1_1 $";
	protected ContentHandler contentHandler = null;
	protected LexicalHandler lexicalhandler = null;
	protected ErrorHandler errorhandler = null;
	protected XMLStatement statement = null;
	protected XDBCResultSetInterface xdbcResulSet = null;
	protected ReconstructionVisitor reconstructionvisitor = null;
	protected XQueryExpression reconstructExpression = null;
	private boolean noNext = false;

	private int validationMode = Constants.VALIDATION_SKIP;
	private SchemaManager schemamanager = null;
	private ValidatingSchemaHandler vsh = null;

	private ArrayList varList = null;

	private int position = 0;
	private boolean firstHasNext = true;

	private XMLSerializer serializer = null;
	private DOMBuilder builder = null;
	private Document document = null;
	private static DocumentBuilder DOMCreator;
	
	// for dynamic context
	private DynamicContext dc = null;

	private PSVInfoSetProvider psvisp = null;
	/**********************************************************************************************/

	public XResultSetImpl(XDBCResultSetInterface xdbcResulSet, XQueryExpression reconstructExpression, ArrayList varList, XMLStatement statement) throws XQueryException {
		this.xdbcResulSet = xdbcResulSet;
		this.reconstructExpression = reconstructExpression;
		// changed not available anymore
		if (reconstructExpression.getParentModule().getValidation() != Constants.NOTHING)
			this.validationMode = reconstructExpression.getParentModule().getValidation();
		this.schemamanager = reconstructExpression.getParentModule().getSchemaManager();
		this.varList = varList;
		this.statement = statement;
		this.dc = new DynamicContext();
		// fill dynamic context
		XQueryModule module = reconstructExpression.getParentModule();
		Map map = module.getImportModules();
		if (map != null) {
			for (Iterator it = map.values().iterator();it.hasNext();) {
				dc.addImportModule(((ImportModule)it.next()).getModule());
			}
		}
		if (validationMode == Constants.VALIDATION_PRESERVE)
			psvisp = (PSVInfoSetProvider)new SchemaValidationContext(schemamanager);
		this.reconstructionvisitor = new ReconstructionVisitor(this, varList);
		this.reconstructionvisitor.setDynamicContext(dc);
	}

	/**********************************************************************************************/

	/**
	 * @see XMLResultSet#setContentHandler
	 */
	public void setContentHandler(ContentHandler contentHandler) {
		if (contentHandler == null)
			throw new IllegalArgumentException("ContentHandler cannot be null.");
		this.contentHandler = contentHandler;
		if (validationMode == Constants.VALIDATION_STRICT || validationMode == Constants.VALIDATION_LAX) {
			// Creates new manager to remove schema without namespace (that contains the default relational schema for Bridge)
			// Thus, elements with no namespace are not validated during reconstruction
			SchemaManager reconstructionManager = new SchemaManager(schemamanager);
			reconstructionManager.removeSchema(null);
			vsh = new ValidatingSchemaHandler(new SchemaValidationContext(reconstructionManager), validationMode == Constants.VALIDATION_STRICT);
			vsh.setContentHandler(contentHandler);
			if (errorhandler == null) {
				vsh.setErrorHandler(new ErrorHandler() {
					public void fatalError(SAXParseException spe) throws SAXException {
					};
					public void warning(SAXParseException spe) throws SAXException {
					};
					public void error(SAXParseException spe) throws SAXException {
						throw spe;
					};
				});
			} else {
				vsh.setErrorHandler(errorhandler);
			}
		}
		if (xdbcResulSet != null)
			xdbcResulSet.setContentHandler((vsh == null) ? contentHandler : vsh);
		reconstructionvisitor.setContentHandler((vsh == null) ? contentHandler : vsh);
	}

	/**
	 * @see ResultSet#setLexicalHandler
	 */
	public void setLexicalHandler(LexicalHandler lexicalhandler) {
		if (lexicalhandler == null)
			throw new IllegalArgumentException("LexicalHandler cannot be null.");
		this.lexicalhandler = lexicalhandler;
		if (xdbcResulSet != null)
			xdbcResulSet.setLexicalHandler(lexicalhandler);
		reconstructionvisitor.setLexicalHandler(lexicalhandler);
	}

	/**
	 * @see ResultSet#setErrorHandler
	 */
	public void setErrorHandler(ErrorHandler errorhandler) {
		if (errorhandler == null)
			throw new IllegalArgumentException("ErrorHandler cannot be null.");
		this.errorhandler = errorhandler;
		if (xdbcResulSet != null)
			xdbcResulSet.setErrorHandler(errorhandler);
		reconstructionvisitor.setErrorHandler(errorhandler);
		if (vsh != null)
			vsh.setErrorHandler(errorhandler);
	}

	/**
	 * @see ResultSet#setErrorHandler
	 */
	public ContentHandler getContentHandler() {
		return (vsh == null) ? contentHandler : vsh;
	}

	/**
	 * @see ResultSet#setErrorHandler
	 */
	public LexicalHandler getLexicalHandler() {
		return lexicalhandler;
	}

	/**
	 * @see ResultSet#setErrorHandler
	 */
	public ErrorHandler getErrorHandler() {
		return errorhandler;
	}

	/**********************************************************************************************/

	/**
	 * @see XMLResultSet#getStatement
	 */
	public XMLStatement getStatement() throws XMLDBCException {
		return statement;
	}

	/**
	 * @see XMLResultSet#close
	 */
	public void close() throws XMLDBCException {
		if (xdbcResulSet != null) {
			xdbcResulSet.close();
			xdbcResulSet = null;
		}
	}

	/**********************************************************************************************/

	/**
	 * @see XMLResultSet#isBeforeFirst
	 */
	public boolean isBeforeFirst() throws XMLDBCException {
		return 0 == position ? true : false;
	}

	/**
	 * @see XMLResultSet#nextAsDOM
	 */
	public Document nextAsDocument() throws XMLDBCException {
		try {
			createDOMBuilder();
			Document document = DOMCreator.newDocument();
			if (builder == null)
				builder = new DOMBuilder(document);
			else
				builder.setDocument(document);
			setContentHandler(builder);
			setLexicalHandler(builder);
			nextAsSAX();
			return document;
		} catch (SAXException e) {
			throw new XMLDBCException(e.getMessage(), e);
		}
	}

	public org.w3c.dom.Node nextAsDOM() throws XMLDBCException {
		return nextAsDocument().getFirstChild();
	}

	public void nextAsDOM(org.w3c.dom.Element parent) throws XMLDBCException {
		try {
			if (builder == null)
				builder = new DOMBuilder(parent);
			else
				builder.setElement(parent);
			setContentHandler(builder);
			setLexicalHandler(builder);
			nextAsSAX();
		} catch (SAXException e) {
			throw new XMLDBCException(e.getMessage(), e);
		}
	}

	/**
	 * @see XMLResultSet#nextAsString
	 */
	public java.lang.String nextAsString() throws XMLDBCException {
		StringWriter stringWriter = new StringWriter();
		nextAsStream(stringWriter);
		return stringWriter.toString();
	}

	public void nextAsStream(Writer out) throws XMLDBCException {
		try {
			if (serializer == null)
				serializer = new XMLSerializer(out);
			else {
				serializer.reset();
				serializer.setWriter(out);
			}
			setContentHandler(serializer);
			setLexicalHandler(serializer);
			nextAsSAX();
		} catch (SAXException e) {
			throw new XMLDBCException(e.getMessage(), e);
		}
	}

	/**
	 * @see XMLResultSet#nextAsSAX
	 */
	public void nextAsSAX() throws XMLDBCException, SAXException {
		firstHasNext = false;
		boolean add = false;
		try {
			if (!reconstructionvisitor.getNoNext() && hasNext()) {
				xdbcResulSet.next();
				add = true;
			}
			if (reconstructExpression != null && reconstructionvisitor != null && !reconstructionvisitor.getNoReconstructor()) {
				// for prefix mapping
				NamespaceContextStack declarations = reconstructExpression.getParentModule().getDeclarations();
				if (contentHandler != null && declarations != null) {
					reconstructionvisitor.setNamespaceDecl(declarations);
				}
				reconstructExpression.accept(reconstructionvisitor);
				reconstructionvisitor.reInit();
				add = true;
			}
			if (add) position++;
		} catch (XQueryException e) {
			throw new XMLDBCException(e.getMessage(), e);
		}
	}

	/**
	 * @see XMLResultSet#hasNext
	 */
	public boolean hasNext() throws XMLDBCException {
		boolean hasnext = false;
		if (xdbcResulSet != null)
			hasnext = xdbcResulSet.hasNext();
		if (hasnext == false && reconstructExpression.getRoot() && firstHasNext) {
			reconstructionvisitor.setNoResultSet(true);
			hasnext = true;
		}
		return hasnext;
	}

	/**
	 * @see XMLResultSet#getPosition
	 */
	public int getPosition() throws XMLDBCException {
		return 0 == position ? -1 : position;
	}

	/**********************************************************************************************/

	public XDBCResultSetInterface getXDBCResultSet() {
		return xdbcResulSet;
	}

	public boolean hasRootTag() {
		if (varList != null && !varList.isEmpty()) {
			Variable tmpVar = (Variable) varList.get(0);
			XQueryExpression tmpExpr = tmpVar.getExpression();
			if ((reconstructExpression instanceof Element) && ((tmpExpr instanceof ValueInteger) || reconstructExpression.getLoop()))
				return true;
		} else if (reconstructExpression instanceof Element && (reconstructExpression.getParentExpression() == null || reconstructExpression.getParentExpression().isEmpty()))
			return true;
		return false;
	}

	public boolean isDocument() {
		if (reconstructExpression instanceof org.xquark.xquery.parser.Document)
			return true;
		return false;
	}

	public PSVInfoSetProvider getPSVInfoSetProvider() {
		return psvisp;
	}

	public void  setPSVInfoSetProvider(PSVInfoSetProvider psvisp) {
		this.psvisp = psvisp;
	}
	
	/**
	 * Returns the result set prefix map (as prefix-namespace pairs).
	 * This map is used to produce each result in the set.
	 * @return the result set prefix map.
	 * @throws XMLDBCException if a data source access error occurs.
	 * @deprecated
	 */
	public java.util.Map getPrefixMap() throws XMLDBCException {
		return java.util.Collections.EMPTY_MAP;
	}

	/**
	 * Returns the fragments contained in the result set as a single XML document, with
	 * a document root element having the specified namespace, local name and qualified name.
	 * Note: this convenience method can only be called if the cursor is before the
	 * first element. It will process the complete result set in one step.
	 * @param namespace the namespace of the document root element.
	 * @param localName the local name of the document root element.
	 * @param qName the qualified name of the document root element.
	 * @return an XML document containing the fragments of the result set, as children of
	 * the specified document root element.
	 * @throws XMLDBCException if a data source access error occurs.
	 */
	public XMLDocument getFragmentsAsDocument(String namespace, String localName, String qName) throws XMLDBCException {
		if (isBeforeFirst()) {
			XMLDocumentImpl xmlDocument = new XMLDocumentImpl(this, namespace, localName, qName);
			xmlDocument.setContentHandler(contentHandler);
			xmlDocument.setLexicalHandler(lexicalhandler);
			xmlDocument.setErrorHandler(errorhandler);
			return xmlDocument;
		} else {
			throw new XMLDBCException("The cursor moved, it is not anymore before the first element.");
		}
	}

	/**
	 * Returns the result set metadata (the XML Schema that models the returned data)
	 * as an XMLDocument.
	 * This schema describes the XML type of each result in the set.
	 * @return the result set metadata (a XML Schema).
	 * @throws XMLDBCException if a data source access error occurs.
	 */
	public XMLDocument getMetaData() throws XMLDBCException {
		return null;
	}

	private synchronized void createDOMBuilder() {
		if (DOMCreator == null) {
			DocumentBuilderFactory DOMFactory = DocumentBuilderFactory.newInstance();
			DOMFactory.setNamespaceAware(true);
			try {
				DOMCreator = DOMFactory.newDocumentBuilder();
			} catch (ParserConfigurationException pce) {
				throw new RuntimeException("Could not create an empty DOM level 2 document using JAXP. Check a JAXP DOM implementation is in your classpath.");
			}
		}
	}
}
